# @file
# 
# - Find CXXLinking
# This module determines how to link C and Fortran applications
# with a C++ library that used libstdc++. 
# It will set the following variables:
#  MARMOT_C_LINKER_LANGUAGE = Language to use for linking C applications (C / CXX)
#  MARMOT_FORTRAN_LINKER_LANGUAGE = Language to use for linking C applications (Fortran / CXX)
#  MARMOT_C_LINK_EXTRA_LIBS = if needed, MARMOT_STDCXX_LIBS (see below)
#  MARMOT_C_LINK_EXTRA_LIBS_CMDLINE = commandline usable version of MARMOT_C_LINK_EXTRA_LIBS
#  MARMOT_FORTRAN_LINK_EXTRA_LIBS = if needed, MARMOT_STDCXX_LIBS/MARMOT_FORTRAN_RUNTIME_LIBS (see below)
#  MARMOT_FORTRAN_LINK_EXTRA_LIBS_CMDLINE = commandline usable version of MARMOT_FORTRAN_LINK_EXTRA_LIBS
#
# Internally it used the follwing extra variables to store state and let 
# the user pass extra information.
#  MARMOT_C_LINK_TYPE = {NATIV_NO_LIB, NATIV_STDCXX, CXX_NO_LIB, CXX_EXTRA_LIBS}
#  MARMOT_FORTRAN_LINK_TYPE = {NATIV_NO_LIB, NATIV_STDCXX, CXX_NO_LIB, CXX_EXTRA_LIBS}
#  MARMOT_STDCXX_LIBS = Library list to use to link against the C++ STL (with full path if not in default path, e.g. /usr/lib/stdc++.so.5)
#  MARMOT_FORTRAN_RUNTIME_LIBS = Library list to use for Fortran runtime (when linking fortran with c++) 
#  
# @author Tobias Hilbrich
#

##
## The individual tests use a C++ library that uses STL.
## Afterwards each tests performs one attempt to link this 
## library to a C or Fortran program.
## The different attempts are:
##  (a) try to link with just C/Fortran compiler
##  (b) try to link with C/Fortran compiler and a detected STDC++ lib
##      (only available if such a lib is found or manually specified)
##  (c) try to link with the C++ compiler without using any extra libs
##  (d) (Fortran only) try to link with the C++ compiler and a detected
##      Fortran runtime library
##      (only available if such a lib is found or manually specified)  
##

SET (CXXLinking_FOUND TRUE)

##### Try to find a lib stdc++ #####
## I tried around with FIND_LIBARARY but it seems to be super broken
## FIND_LIBRARY(MARMOT_STDCXX_LIBS NAMES ${SEARCH_LIBS} DOC "Libraries (Semicolon seperated list) that are needed to link agains a C++ library, e.g., \"/usr/lib/libstdc++.so;/usr/local/lib/libspecial.a\".")
##
## So we provide known configs, if none of them works the user has to do the job himself.
SET (MARMOT_STDCXX_LIBS "" CACHE STRING "Libraries (Semicolon seperated list) that are needed to link agains an STL using C++ library, e.g., \"/usr/lib/libstdc++.so;/usr/lib/libstdc++extra.so\".")
SET (STDCXX_LIBS_1 ${CMAKE_CXX_IMPLICIT_LINK_LIBRARIES}) ##Comes with CMake 2.8, should not harm for older versions
SET (STDCXX_LIBS_2 "stdc++") #gnu, intel
SET (STDCXX_LIBS_3 "stlport") #STLPort Project
SET (STDCXX_LIBS_4 "-library=stlport4;-lstlport;-lCrun") #TODO: STLPort Project with Sun compilers (uses extra flags)
#AIX -> don't add stuff for AIX, use C++ compiler to link there!
SET (STDCXX_LIBS "STDCXX_LIBS_1" "STDCXX_LIBS_2" "STDCXX_LIBS_3" "STDCXX_LIBS_4")

##### Try to find Fortran runtime libraries #####
## I tried around with FIND_LIBARARY but it seems to be super broken
## FIND_LIBRARY(MARMOT_FORTRAN_RUNTIME_LIBS NAMES "gfortran" PATHS "/usr/lib" DOC "Libraries (Semicolon seperated list) that are needed to link agains Fortran applications with the C++ compiler, e.g., \"/usr/lib/libgfortran.so;/usr/lib/libgfortranbegin.so\".")
##
## So we provide known configs, if none of them works the user has to do the job himself. 
SET (MARMOT_FORTRAN_RUNTIME_LIBS "" CACHE STRING "Libraries (Semicolon seperated list) that are needed to link against Fortran applications with the C++ compiler, e.g., \"/usr/lib/libgfortran.so;/usr/lib/libgfortranbegin.so\" or \"-lgfortran;-lgfortranbegin\".")
SET (FORTRAN_RUNTIME_1 ${CMAKE_Fortran_IMPLICIT_LINK_LIBRARIES}) ##Comes with CMake 2.8, should not harm for older versions 
SET (FORTRAN_RUNTIME_2 "gfortran;gfortranbegin") #gfortran
SET (FORTRAN_RUNTIME_3 "-Bstatic;-lifport;-lifcore;-limf") #intel
SET (FORTRAN_RUNTIME_4 "-lxlf90;-lxlopt;-lxlf") #IBM AIX
SET (FORTRAN_RUNTIMES "FORTRAN_RUNTIME_1" "FORTRAN_RUNTIME_2" "FORTRAN_RUNTIME_3" "FORTRAN_RUNTIME_4")

###################
##### First C #####
###################
SET (MARMOT_C_LINK_TYPE)

#Try linking without any modifications
IF (NOT MARMOT_C_LINK_TYPE)
    TRY_COMPILE(result "${PROJECT_BINARY_DIR}/FindTemp/no-lib-C" "${PROJECT_SOURCE_DIR}/CMAKEMODULES/FeatureTests/CXXLibLink/no-lib-C" no-lib-C CMAKE_FLAGS -DCMAKE_C_COMPILER=${CMAKE_C_COMPILER} -DCMAKE_CXX_COMPILER=${CMAKE_CXX_COMPILER} -DCMAKE_Fortran_COMPILER=${CMAKE_Fortran_COMPILER} OUTPUT_VARIABLE output)
    
    IF (result)
        SET (MARMOT_C_LINK_TYPE "NATIV_NO_LIB")
        SET (MARMOT_C_LINKER_LANGUAGE "C" CACHE INTERNAL "Linker language to be used for C applications." FORCE)
        SET (MARMOT_C_LINK_EXTRA_LIBS "" CACHE INTERNAL "Extra libraries needed to link C applications." FORCE)
        MESSAGE (STATUS "Linking C application with C++ Library: Works without any modifications.")
    ENDIF (result)
ENDIF (NOT MARMOT_C_LINK_TYPE)

#Try linking with extra libraries (used as libstdc++)
IF (NOT MARMOT_C_LINK_TYPE)
    #Use user specified(or empty) libraries
    TRY_COMPILE(result "${PROJECT_BINARY_DIR}/FindTemp/with-lib-C" "${PROJECT_SOURCE_DIR}/CMAKEMODULES/FeatureTests/CXXLibLink/with-lib-C" with-lib-C CMAKE_FLAGS -DCMAKE_C_COMPILER=${CMAKE_C_COMPILER} -DCMAKE_CXX_COMPILER=${CMAKE_CXX_COMPILER} -DCMAKE_Fortran_COMPILER=${CMAKE_Fortran_COMPILER} "-DMARMOT_DETECTED_CXX_LIB=${MARMOT_STDCXX_LIBS}" OUTPUT_VARIABLE output)
    
    IF (result)
        SET (MARMOT_C_LINK_TYPE "NATIV_STDCXX")
        SET (MARMOT_C_LINKER_LANGUAGE "C" CACHE INTERNAL "Linker language to be used for C applications." FORCE)
        
        BC_LIST_REMOVE_BSTATIC_BDYNAMIC (MARMOT_STDCXX_LIBS "C++ libraries to use when linking a C programm with a C++ library" temp)
        SET (MARMOT_C_LINK_EXTRA_LIBS "${temp}" CACHE INTERNAL "Extra libraries needed to link C applications." FORCE)
        MESSAGE (STATUS "Linking C application with C++ Library: Works with C++ STL and runtime libs (${MARMOT_C_LINK_EXTRA_LIBS}).")
    ELSE (result)
        #If user did not specify anything, try the different default sets
        FOREACH (libset ${STDCXX_LIBS})
            IF (NOT MARMOT_STDCXX_LIBS)
                TRY_COMPILE(result "${PROJECT_BINARY_DIR}/FindTemp/with-lib-C-${libset}" "${PROJECT_SOURCE_DIR}/CMAKEMODULES/FeatureTests/CXXLibLink/with-lib-C" with-lib-C CMAKE_FLAGS -DCMAKE_C_COMPILER=${CMAKE_C_COMPILER} -DCMAKE_CXX_COMPILER=${CMAKE_CXX_COMPILER} -DCMAKE_Fortran_COMPILER=${CMAKE_Fortran_COMPILER} "-DMARMOT_DETECTED_CXX_LIB=${${libset}}" OUTPUT_VARIABLE output)
                
                IF (result)
                    SET (MARMOT_C_LINK_TYPE "NATIV_STDCXX")
                    SET (MARMOT_C_LINKER_LANGUAGE "C" CACHE INTERNAL "Linker language to be used for C applications." FORCE)
                    
                    BC_LIST_REMOVE_BSTATIC_BDYNAMIC (${libset} "C++ libraries to use when linking a C programm with a C++ library" temp)
                    SET (MARMOT_STDCXX_LIBS "${temp}" CACHE STRING "Libraries (Semicolon seperated list) that are needed to link agains an STL using C++ library, e.g., \"/usr/lib/libstdc++.so;/usr/lib/libstdc++extra.so\"." FORCE)
                    SET (MARMOT_C_LINK_EXTRA_LIBS "${MARMOT_STDCXX_LIBS}" CACHE INTERNAL "Extra libraries needed to link C applications." FORCE)
                    MESSAGE (STATUS "Linking C application with C++ Library: Works with C++ STL and runtime libs (${MARMOT_STDCXX_LIBS}).")
                ENDIF (result)
            ENDIF (NOT MARMOT_STDCXX_LIBS)
        ENDFOREACH (libset)
    ENDIF (result)    
ENDIF (NOT MARMOT_C_LINK_TYPE)

#Try linking with C++ compiler
IF (NOT MARMOT_C_LINK_TYPE)
    TRY_COMPILE(result "${PROJECT_BINARY_DIR}/FindTemp/with-CXX-C" "${PROJECT_SOURCE_DIR}/CMAKEMODULES/FeatureTests/CXXLibLink/with-CXX-C" with-CXX-C CMAKE_FLAGS -DCMAKE_C_COMPILER=${CMAKE_C_COMPILER} -DCMAKE_CXX_COMPILER=${CMAKE_CXX_COMPILER} -DCMAKE_Fortran_COMPILER=${CMAKE_Fortran_COMPILER} OUTPUT_VARIABLE output)
    
    IF (result)
        SET (MARMOT_C_LINK_TYPE "CXX_NO_LIB")
        SET (MARMOT_C_LINKER_LANGUAGE "CXX" CACHE INTERNAL "Linker language to be used for C applications." FORCE)
        SET (MARMOT_C_LINK_EXTRA_LIBS "" CACHE INTERNAL "Extra libraries needed to link C applications." FORCE)
        MESSAGE (STATUS "Linking C application with C++ Library: Works with C++ compiler as linker.")
    ENDIF (result)
ENDIF (NOT MARMOT_C_LINK_TYPE)

#Report error if no working configuration found
IF (NOT MARMOT_C_LINK_TYPE)
    IF (CXXLinking_FIND_REQUIRED)
        MESSAGE (FATAL_ERROR "Can't link a C application with a C++ library, try to specify the C++ STL and runtime libraries in MARMOT_STDCXX_LIBS, contact your local admistrator or the Marmot developers for further advice.")
    ELSE (CXXLinking_FIND_REQUIRED)
        MARK_AS_ADVANCED(FORCE MARMOT_STDCXX_LIBS)
        MESSAGE (STATUS "Can't link a C application with a C++ library, try to specify the C++ STL and runtime libraries in MARMOT_STDCXX_LIBS")
    ENDIF (CXXLinking_FIND_REQUIRED)
ELSE (NOT MARMOT_C_LINK_TYPE)    
    MARK_AS_ADVANCED(FORCE MARMOT_STDCXX_LIBS)
ENDIF (NOT MARMOT_C_LINK_TYPE)

#########################################
##### Second Fortran (if activated) #####
#########################################
IF (MARMOT_ENABLE_FORTRAN)
    SET (MARMOT_FORTRAN_LINK_TYPE)

    #Try linking without any modifications
    IF (NOT MARMOT_FORTRAN_LINK_TYPE)
        TRY_COMPILE(result "${PROJECT_BINARY_DIR}/FindTemp/no-lib-F" "${PROJECT_SOURCE_DIR}/CMAKEMODULES/FeatureTests/CXXLibLink/no-lib-F" no-lib-F CMAKE_FLAGS -DCMAKE_C_COMPILER=${CMAKE_C_COMPILER} -DCMAKE_CXX_COMPILER=${CMAKE_CXX_COMPILER} -DCMAKE_Fortran_COMPILER=${CMAKE_Fortran_COMPILER} OUTPUT_VARIABLE output)
        
        IF (result)
            SET (MARMOT_FORTRAN_LINK_TYPE "NATIV_NO_LIB")
            SET (MARMOT_FORTRAN_LINKER_LANGUAGE "Fortran" CACHE INTERNAL "Linker language to be used for Fortran applications." FORCE)
            SET (MARMOT_FORTRAN_LINK_EXTRA_LIBS "" CACHE INTERNAL "Extra libraries needed to link Fortran applications." FORCE)
            MESSAGE (STATUS "Linking Fortran application with C++ Library: Works without any modifications.")
        ENDIF (result)
    ENDIF (NOT MARMOT_FORTRAN_LINK_TYPE)
    
    #Try linking with extra libraries (used as libstdc++)
    IF (NOT MARMOT_FORTRAN_LINK_TYPE)
        #Use user specified(or empty) libraries
        TRY_COMPILE(result "${PROJECT_BINARY_DIR}/FindTemp/with-lib-F" "${PROJECT_SOURCE_DIR}/CMAKEMODULES/FeatureTests/CXXLibLink/with-lib-F" with-lib-F CMAKE_FLAGS -DCMAKE_C_COMPILER=${CMAKE_C_COMPILER} -DCMAKE_CXX_COMPILER=${CMAKE_CXX_COMPILER} -DCMAKE_Fortran_COMPILER=${CMAKE_Fortran_COMPILER} "-DMARMOT_DETECTED_CXX_LIB=${MARMOT_STDCXX_LIBS}" OUTPUT_VARIABLE output)
        
        IF (result)
            SET (MARMOT_FORTRAN_LINK_TYPE "NATIV_STDCXX")
            SET (MARMOT_FORTRAN_LINKER_LANGUAGE "Fortran" CACHE INTERNAL "Linker language to be used for Fortran applications." FORCE)
            
            BC_LIST_REMOVE_BSTATIC_BDYNAMIC (MARMOT_STDCXX_LIBS "C++ libraries to use when linking a Fortran programm with a C++ library" temp)
            SET (MARMOT_FORTRAN_LINK_EXTRA_LIBS "${temp}" CACHE INTERNAL "Extra libraries needed to link Fortran applications." FORCE)
            MESSAGE (STATUS "Linking Fortran application with C++ Library: Works with C++ STL and runtime libs (${MARMOT_FORTRAN_LINK_EXTRA_LIBS}).")
        ELSE (result)
            #If user did not specify anything, try the different default sets
            FOREACH (libset ${STDCXX_LIBS})
                IF (NOT MARMOT_STDCXX_LIBS)
                    TRY_COMPILE(result "${PROJECT_BINARY_DIR}/FindTemp/with-lib-F-${libset}" "${PROJECT_SOURCE_DIR}/CMAKEMODULES/FeatureTests/CXXLibLink/with-lib-F" with-lib-F CMAKE_FLAGS -DCMAKE_C_COMPILER=${CMAKE_C_COMPILER} -DCMAKE_CXX_COMPILER=${CMAKE_CXX_COMPILER} -DCMAKE_Fortran_COMPILER=${CMAKE_Fortran_COMPILER} "-DMARMOT_DETECTED_CXX_LIB=${${libset}}" OUTPUT_VARIABLE output)
                    
                    IF (result)
                        SET (MARMOT_FORTRAN_LINK_TYPE "NATIV_STDCXX")
                        SET (MARMOT_FORTRAN_LINKER_LANGUAGE "Fortran" CACHE INTERNAL "Linker language to be used for Fortran applications." FORCE)
                        
                        BC_LIST_REMOVE_BSTATIC_BDYNAMIC (${libset} "C++ libraries to use when linking a Fortran programm with a C++ library" temp)
                        SET (MARMOT_STDCXX_LIBS "${temp}" CACHE STRING "Libraries (Semicolon seperated list) that are needed to link agains an STL using C++ library, e.g., \"/usr/lib/libstdc++.so;/usr/lib/libstdc++extra.so\"." FORCE)
                        SET (MARMOT_FORTRAN_LINK_EXTRA_LIBS "${MARMOT_STDCXX_LIBS}" CACHE INTERNAL "Extra libraries needed to link Fortran applications." FORCE)
                        MESSAGE (STATUS "Linking Fortran application with C++ Library: Works with C++ STL and runtime libs (${MARMOT_STDCXX_LIBS}).")
                    ENDIF (result)
                ENDIF (NOT MARMOT_STDCXX_LIBS)
            ENDFOREACH (libset)
        ENDIF (result)    
    ENDIF (NOT MARMOT_FORTRAN_LINK_TYPE)
    
    #Try linking with C++ compiler
    IF (NOT MARMOT_FORTRAN_LINK_TYPE)
        TRY_COMPILE(result "${PROJECT_BINARY_DIR}/FindTemp/with-CXX-F" "${PROJECT_SOURCE_DIR}/CMAKEMODULES/FeatureTests/CXXLibLink/with-CXX-F" with-CXX-F CMAKE_FLAGS -DCMAKE_C_COMPILER=${CMAKE_C_COMPILER} -DCMAKE_CXX_COMPILER=${CMAKE_CXX_COMPILER} -DCMAKE_Fortran_COMPILER=${CMAKE_Fortran_COMPILER} OUTPUT_VARIABLE output)
        
        IF (result)
            SET (MARMOT_FORTRAN_LINK_TYPE "CXX_NO_LIB")
            SET (MARMOT_FORTRAN_LINKER_LANGUAGE "CXX" CACHE INTERNAL "Linker language to be used for Fortran applications." FORCE)
            SET (MARMOT_FORTRAN_LINK_EXTRA_LIBS "" CACHE INTERNAL "Extra libraries needed to link Fortran applications." FORCE)
            MESSAGE (STATUS "Linking Fortran application with C++ Library: Works with C++ compiler as linker.")
        ENDIF (result)
    ENDIF (NOT MARMOT_FORTRAN_LINK_TYPE)
    
    #Try linking with C++ compiler AND extra libraries (used as Fortran runtime)
    IF (NOT MARMOT_FORTRAN_LINK_TYPE)
        #Use user specified(or empty) libraries
        TRY_COMPILE(result "${PROJECT_BINARY_DIR}/FindTemp/with-CXX-with-lib-F" "${PROJECT_SOURCE_DIR}/CMAKEMODULES/FeatureTests/CXXLibLink/with-CXX-with-lib-F" with-CXX-with-lib-F CMAKE_FLAGS -DCMAKE_C_COMPILER=${CMAKE_C_COMPILER} -DCMAKE_CXX_COMPILER=${CMAKE_CXX_COMPILER} -DCMAKE_Fortran_COMPILER=${CMAKE_Fortran_COMPILER} "-DMARMOT_DETECTED_F_EXTRA_LIB=${MARMOT_FORTRAN_RUNTIME_LIBS}" OUTPUT_VARIABLE output)
        
        IF (result)
            SET (MARMOT_FORTRAN_LINK_TYPE "CXX_EXTRA_LIB")
            SET (MARMOT_FORTRAN_LINKER_LANGUAGE "CXX" CACHE INTERNAL "Linker language to be used for Fortran applications." FORCE)
            
            BC_LIST_REMOVE_BSTATIC_BDYNAMIC (MARMOT_FORTRAN_RUNTIME_LIBS "Fortran libraries to use when linking a Fortran programm with a C++ compiler" temp)
            SET (MARMOT_FORTRAN_LINK_EXTRA_LIBS "${temp}" CACHE INTERNAL "Extra libraries needed to link Fortran applications." FORCE)
            MESSAGE (STATUS "Linking Fortran application with C++ Library: Works with C++ compiler as linker and Fortran runtime libs (${MARMOT_FORTRAN_LINK_EXTRA_LIBS}).")
        ELSE (result)
            #If user did not specify anything, try the different default sets
            FOREACH (libset ${FORTRAN_RUNTIMES})
                IF (NOT MARMOT_FORTRAN_RUNTIME_LIBS)
                    TRY_COMPILE(result "${PROJECT_BINARY_DIR}/FindTemp/with-CXX-with-lib-F-${libset}" "${PROJECT_SOURCE_DIR}/CMAKEMODULES/FeatureTests/CXXLibLink/with-CXX-with-lib-F" with-CXX-with-lib-F CMAKE_FLAGS -DCMAKE_C_COMPILER=${CMAKE_C_COMPILER} -DCMAKE_CXX_COMPILER=${CMAKE_CXX_COMPILER} -DCMAKE_Fortran_COMPILER=${CMAKE_Fortran_COMPILER} "-DMARMOT_DETECTED_F_EXTRA_LIB=${${libset}}" OUTPUT_VARIABLE output)

                    IF (result)
                        SET (MARMOT_FORTRAN_LINK_TYPE "CXX_EXTRA_LIB")
                        SET (MARMOT_FORTRAN_LINKER_LANGUAGE "CXX" CACHE INTERNAL "Linker language to be used for Fortran applications." FORCE)
                        
                        BC_LIST_REMOVE_BSTATIC_BDYNAMIC (${libset} "Fortran libraries to use when linking a Fortran programm with a C++ compiler" temp)
                        SET (MARMOT_FORTRAN_RUNTIME_LIBS "${temp}" CACHE STRING "Libraries (Semicolon seperated list) that are needed to link against Fortran applications with the C++ compiler, e.g., \"/usr/lib/libgfortran.so;/usr/lib/libgfortranbegin.so\" or \"-lgfortran;-lgfortranbegin\"." FORCE)
                        SET (MARMOT_FORTRAN_LINK_EXTRA_LIBS "${MARMOT_FORTRAN_RUNTIME_LIBS}" CACHE INTERNAL "Extra libraries needed to link Fortran applications." FORCE)
                        MESSAGE (STATUS "Linking Fortran application with C++ Library: Works with C++ compiler as linker and Fortran runtime libs (${MARMOT_FORTRAN_RUNTIME_LIBS}).")
                    ENDIF (result)
                ENDIF (NOT MARMOT_FORTRAN_RUNTIME_LIBS)
            ENDFOREACH (libset)
        ENDIF (result)    
    ENDIF (NOT MARMOT_FORTRAN_LINK_TYPE)
    
    #Report error if no working configuration found
    IF (NOT MARMOT_FORTRAN_LINK_TYPE)
        IF (CXXLinking_FIND_REQUIRED)
            MESSAGE (FATAL_ERROR "Can't link a Fortran application with a C++ library, try to specify Fortran runtime libraries that are needed when linking with a C++ compiler. Specify them in MARMOT_FORTRAN_RUNTIME_LIBS, contact your local admistrator or the Marmot developers for further advice.")
        ELSE (CXXLinking_FIND_REQUIRED)
            MARK_AS_ADVANCED(FORCE  MARMOT_FORTRAN_RUNTIME_LIBS MARMOT_STDCXX_LIBS)
            MESSAGE (STATUS "Can't link a Fortran application with a C++ library, try to specify Fortran runtime libraries that are needed when linking with a C++ compiler. Specify them in MARMOT_FORTRAN_RUNTIME_LIBS.")
        ENDIF (CXXLinking_FIND_REQUIRED)
    ELSE (NOT MARMOT_FORTRAN_LINK_TYPE)
        MARK_AS_ADVANCED(FORCE MARMOT_FORTRAN_RUNTIME_LIBS MARMOT_STDCXX_LIBS)
    ENDIF (NOT MARMOT_FORTRAN_LINK_TYPE)
ENDIF (MARMOT_ENABLE_FORTRAN)

###############################################
##### Last, Command line strings for libs #####
###############################################
BC_LIBLIST_TO_LINKCOMMANDLINE (MARMOT_FORTRAN_LINK_EXTRA_LIBS temp)
SET (MARMOT_FORTRAN_LINK_EXTRA_LIBS_CMDLINE "${temp}" CACHE INTERNAL "" FORCE)

BC_LIBLIST_TO_LINKCOMMANDLINE (MARMOT_C_LINK_EXTRA_LIBS temp)
SET (MARMOT_C_LINK_EXTRA_LIBS_CMDLINE "${temp}" CACHE INTERNAL "" FORCE)
