# @file
# 
# - Find Cube
# This module looks for Cube support (generic lookup). It will set the following variables:
#  CUBE_INCLUDE_DIR     = where Cube.h can be found
#  CUBE_LIBRARY         = the Cube library to link in
#  CUBE_LIBRARIES       = these are all the needed libraries for Cube this is identical to ${CUBE_LIBRARY}
#
# Note: If CUBE_HOME is set by the user this is the place where the FIND_* macros
#       will look first
#
# @author Blasius Czink

include(MarmotHelperMacros)
include(FindPackageHandleStandardArgs) 

# where to look for the include
SET(CUBE_INC_SEARCH_DIRS 
  "$ENV{EXTERNLIBS}/cube/include"
  /usr/include 
  /usr/local/include 
  "$ENV{ProgramFiles}/cube/include"
)

# where to look for the libraries
SET(CUBE_LIB_SEARCH_DIRS 
  "$ENV{EXTERNLIBS}/cube/lib"
  /usr/lib
  /usr/local/lib
  "$ENV{ProgramFiles}/cube/lib"
)

# show an option for CUBE_HOME
IF (CUBE_HOME)
  IF (NOT "${CUBE_HOME}" STREQUAL "${CUBE_HOME_OLD}") 
    # user selected something
    SET(CUBE_HOME ${CUBE_HOME} CACHE PATH "CUBE_HOME")
    # set variables to ...-NOTFOUND to reset (Is there a more elegant way to do that?)
    BC_RESET(CUBE_INCLUDE_DIR)
    BC_RESET(CUBE_LIBRARY)
    BC_RESET(CUBE_LIBRARIES)
    BC_RESET(CUBE_LIBRARY_DEBUG)
  ENDIF (NOT "${CUBE_HOME}" STREQUAL "${CUBE_HOME_OLD}")
  # only look in CUBE_HOME if given
  SET(CUBE_INC_SEARCH_DIRS "${CUBE_HOME}/include")
  SET(CUBE_LIB_SEARCH_DIRS "${CUBE_HOME}/lib") 
ELSE (CUBE_HOME)
  # nothing, so just create an empty field
  SET(CUBE_HOME "" CACHE PATH "CUBE_HOME")
  SET(CUBE_HOME_OLD "" CACHE INTERNAL "") 
ENDIF (CUBE_HOME)

IF (CUBE_VERSION EQUAL "2")
  SET (CUBE_LIB_NAME "cube")
  # only need libxml2 for cube version 2
  FIND_PACKAGE (LibXml2)
ELSE (CUBE_VERSION EQUAL "2")
  SET (CUBE_LIB_NAME "cube3")
ENDIF (CUBE_VERSION EQUAL "2")

# windows
IF (WIN32)
  
  FIND_PATH (CUBE_INCLUDE_DIR Cube.h
    PATHS
    ${CUBE_INC_SEARCH_DIRS}
    PATH_SUFFIXES cube3
    DOC "Cube Header path"
    NO_DEFAULT_PATH
  )

  FIND_LIBRARY (CUBE_LIBRARY
    NAMES ${CUBE_LIB_NAME}
    PATHS
    ${CUBE_LIB_SEARCH_DIRS}
    DOC "Cube Library"
    NO_DEFAULT_PATH
  )

  IF (MSVC)
    FIND_LIBRARY (CUBE_LIBRARY_DEBUG
          NAMES "${CUBE_LIB_NAME}d"
          PATHS
          ${CUBE_LIB_SEARCH_DIRS}
          DOC "Cube Library (debug)"
          NO_DEFAULT_PATH
    )

    IF (MSVC_IDE)
      # VS build
      IF (CUBE_LIBRARY AND CUBE_LIBRARY_DEBUG)
        SET(CUBE_LIBRARIES optimized ${CUBE_LIBRARY} debug ${CUBE_LIBRARY_DEBUG})
      ENDIF (CUBE_LIBRARY AND CUBE_LIBRARY_DEBUG)
  
    ELSE (MSVC_IDE)
      # nmake
      STRING(TOLOWER ${CMAKE_BUILD_TYPE} CMAKE_BUILD_TYPE_TOLOWER)
      IF(CMAKE_BUILD_TYPE_TOLOWER MATCHES debug)
        IF(CUBE_LIBRARY_DEBUG)
         SET(CUBE_LIBRARIES ${CUBE_LIBRARY_DEBUG})
        ENDIF(CUBE_LIBRARY_DEBUG)
      ELSE(CMAKE_BUILD_TYPE_TOLOWER MATCHES debug)
        IF(CUBE_LIBRARY)
         SET(CUBE_LIBRARIES ${CUBE_LIBRARY})
        ENDIF(CUBE_LIBRARY)
      ENDIF(CMAKE_BUILD_TYPE_TOLOWER MATCHES debug)
    ENDIF (MSVC_IDE)
    
    FIND_PACKAGE_HANDLE_STANDARD_ARGS(CUBE "Could not locate CUBE" CUBE_LIBRARY CUBE_LIBRARY_DEBUG CUBE_INCLUDE_DIR)
  ELSE (MSVC)
    
    # mingw
    FIND_LIBRARY (CUBE_SCZ_LIBRARY NAMES sc.z PATHS ${CUBE_LIB_SEARCH_DIRS} DOC "Cube sc.z library" NO_DEFAULT_PATH)
    
    SET(CUBE_LIBRARIES ${CUBE_LIBRARY} ${CUBE_SCZ_LIBRARY})
    
    FIND_PACKAGE_HANDLE_STANDARD_ARGS(CUBE "Could not locate CUBE" CUBE_LIBRARY CUBE_SCZ_LIBRARY CUBE_INCLUDE_DIR)
  
  ENDIF (MSVC)
  
ELSE (WIN32)

  FIND_PROGRAM(CUBE_CONFIG NAMES cube-config PATHS "${CUBE_HOME}/bin" "$ENV{CUBE_HOME}/bin")
  IF (CUBE_CONFIG)
    EXEC_PROGRAM(${CUBE_CONFIG} ARGS --prefix
                 OUTPUT_VARIABLE CUBE_PREFIX
                 RETURN_VALUE CUBE_CONFIG_RETURN)

    IF (CUBE_CONFIG_RETURN EQUAL 0)
      FIND_PATH (CUBE_INCLUDE_DIR Cube.h
        PATHS "${CUBE_PREFIX}/include"
        PATH_SUFFIXES cube3
        DOC "Cube Header path"
        NO_DEFAULT_PATH
      )
      FIND_LIBRARY (CUBE_LIBRARY
        NAMES ${CUBE_LIB_NAME}
        PATHS ${CUBE_PREFIX}/lib
        DOC "Cube Library"
        NO_DEFAULT_PATH
      )
    ENDIF (CUBE_CONFIG_RETURN EQUAL 0)
    
    EXEC_PROGRAM(${CUBE_CONFIG} ARGS --libs
                 OUTPUT_VARIABLE CUBE_LIBS_CMDLINE
                 RETURN_VALUE CUBE_CONFIG_RETURN)
		 
    IF (CUBE_CONFIG_RETURN EQUAL 0)
      # extract linker path from the link command line
      string(REGEX MATCHALL "-L([^\" ]+|\"[^\"]+\")" CUBE_ALL_LINK_PATHS "${CUBE_LIBS_CMDLINE}")
      set(CUBE_LINK_PATH)
      foreach(LPATH ${CUBE_ALL_LINK_PATHS})
        string(REGEX REPLACE "^-L" "" LPATH ${LPATH})
        string(REGEX REPLACE "//" "/" LPATH ${LPATH})
        list(APPEND CUBE_LINK_PATH ${LPATH})
      endforeach(LPATH)
      
      # extract the set of libraries
      string(REGEX MATCHALL "-l([^\" ]+|\"[^\"]+\")" CUBE_LIBNAMES "${CUBE_LIBS_CMDLINE}")
      # message("CUBE_LINK_PATH = ${CUBE_LINK_PATH}")
      # message("CUBE_LIBNAMES = ${CUBE_LIBNAMES}")
      
      # get full path names for all the libraries
      set(CUBE_LIBRARIES)
      foreach(LIB ${CUBE_LIBNAMES})
        string(REGEX REPLACE "^-l" "" LIB ${LIB})
        set(CUBE_LIB "CUBE_LIB-NOTFOUND" CACHE FILEPATH "Cleared" FORCE)
        find_library(CUBE_LIB ${LIB} HINTS ${CUBE_LINK_PATH})
        if (CUBE_LIB)
          list(APPEND CUBE_LIBRARIES ${CUBE_LIB})
        else (CUBE_LIB)
          message(SEND_ERROR "Unable to find CUBE library ${LIB}")
        endif (CUBE_LIB)
      endforeach(LIB)
      set(CUBE_LIB "CUBE_LIB-NOTFOUND" CACHE INTERNAL "Scratch variable for CUBE detection" FORCE)
    ENDIF (CUBE_CONFIG_RETURN EQUAL 0)
  ENDIF (CUBE_CONFIG)
  
  FIND_PACKAGE_HANDLE_STANDARD_ARGS(CUBE "Could not locate CUBE" CUBE_LIBRARIES CUBE_INCLUDE_DIR)
  
  # message("CUBE_LIBRARIES = ${CUBE_LIBRARIES}")
  
ENDIF (WIN32)

IF (CUBE_FOUND)
  ADD_DEFINITIONS (${LIBXML2_DEFINITIONS})
  SET (CUBE_LIBRARIES ${CUBE_LIBRARIES} ${LIBXML2_LIBRARIES})
ENDIF (CUBE_FOUND)
