# @file
# 
# - Find MARMOT
# This module looks for MARMOT support (and is intended for users of marmot)
# it will define the following values
#  MARMOT_INCLUDE_DIR = where marmot.h can be found
#  MARMOT_LIBRARY     = the library to link in
#  MARMOT_LIBRARIES   = all the needed libraries
#
# Note: If MARMOT_HOME is set by the user this is the place where the FIND_* macros
#       will look first
#
# @author Blasius Czink
#

include (MarmotHelperMacros)
include (MarmotConfig)

# try to find the MPI package that was used to compile marmot
# note that the cmake-modules have to be made locatable in the main CMakeLists.txt file
# with a SET(CMAKE_MODULE_PATH "${MARMOT_INSTALL_DIR}/share/cmakemodules;${CMAKE_MODULE_PATH}")
# where MARMOT_INSTALL_DIR is set to the marmot installation directory 
FIND_PACKAGE(${MARMOT_WITH_MPI_PACKAGE} REQUIRED)

IF (MARMOT_WITH_CUBE)
  FIND_PACKAGE(Cube REQUIRED)
ENDIF (MARMOT_WITH_CUBE)

IF (NOT DEFINED MARMOT_HOME)
  # not defined! create empty cache entry
  SET(MARMOT_HOME "" CACHE PATH "MARMOT_HOME")
ELSE (NOT DEFINED MARMOT_HOME)
  # it's defined. check whether path exists
  IF (EXISTS "${MARMOT_HOME}")
    # reset include and library variables
    BC_RESET(MARMOT_INCLUDE_DIR)
    BC_RESET(MARMOT_LIBRARY)
    BC_RESET(MARMOT_LIBRARIES)
    BC_RESET(MARMOT_LIBRARY_RELEASE)
    BC_RESET(MARMOT_LIBRARY_DEBUG)
    BC_RESET(MARMOT_profile_LIBRARY_RELEASE)
    BC_RESET(MARMOT_profile_LIBRARY_DEBUG)
  ELSE (EXISTS "${MARMOT_HOME}")
    MESSAGE("Given path '${MARMOT_HOME}' for MARMOT_HOME does not exist!") 
  ENDIF (EXISTS "${MARMOT_HOME}")
ENDIF (NOT DEFINED MARMOT_HOME)

FIND_PATH(MARMOT_INCLUDE_DIR marmot.h 
          ${MARMOT_HOME}/include
          "$ENV{MARMOT_HOME}/include"
          /usr/include 
          /usr/local/include 
          /usr/local/marmot/include
          /opt/marmot/include
          "$ENV{EXTERNLIBS}/marmot/include"
          "$ENV{ProgramFiles}/marmot/include"
          "C:/Program Files/marmot/include"
          [HKEY_LOCAL_MACHINE\\SOFTWARE\\HLRS\\Marmot]/include
          NO_DEFAULT_PATH
)

FIND_LIBRARY(MARMOT_LIBRARY_RELEASE
          NAMES marmot-core
          PATHS ${MARMOT_HOME}/lib "$ENV{MARMOT_HOME}/lib" /usr/lib /usr/local/lib /usr/local/marmot/lib /opt/marmot/lib 
          "$ENV{EXTERNLIBS}/marmot/lib"
          "$ENV{ProgramFiles}/marmot/lib"
          "C:/Program Files/marmot/lib"
          [HKEY_LOCAL_MACHINE\\SOFTWARE\\HLRS\\Marmot]/lib
)
FIND_LIBRARY(MARMOT_LIBRARY_DEBUG
          NAMES marmot-cored
          PATHS ${MARMOT_HOME}/lib "$ENV{MARMOT_HOME}/lib" /usr/lib /usr/local/lib /usr/local/marmot/lib /opt/marmot/lib 
          "$ENV{EXTERNLIBS}/marmot/lib"
          "$ENV{ProgramFiles}/marmot/lib"
          "C:/Program Files/marmot/lib"
          [HKEY_LOCAL_MACHINE\\SOFTWARE\\HLRS\\Marmot]/lib
)

FIND_LIBRARY(MARMOT_profile_LIBRARY_RELEASE
          NAMES marmot-profile
          PATHS ${MARMOT_HOME}/lib "$ENV{MARMOT_HOME}/lib" /usr/lib /usr/local/lib /usr/local/marmot/lib /opt/marmot/lib 
          "$ENV{EXTERNLIBS}/marmot/lib"
          "$ENV{ProgramFiles}/marmot/lib"
          "C:/Program Files/marmot/lib"
          [HKEY_LOCAL_MACHINE\\SOFTWARE\\HLRS\\Marmot]/lib
)
FIND_LIBRARY(MARMOT_profile_LIBRARY_DEBUG
          NAMES marmot-profiled
          PATHS ${MARMOT_HOME}/lib "$ENV{MARMOT_HOME}/lib" /usr/lib /usr/local/lib /usr/local/marmot/lib /opt/marmot/lib 
          "$ENV{EXTERNLIBS}/marmot/lib"
          "$ENV{ProgramFiles}/marmot/lib"
          "C:/Program Files/marmot/lib"
          [HKEY_LOCAL_MACHINE\\SOFTWARE\\HLRS\\Marmot]/lib
)

FIND_LIBRARY(MARMOT_1sided_LIBRARY_RELEASE
          NAMES marmot-1sided
          PATHS ${MARMOT_HOME}/lib "$ENV{MARMOT_HOME}/lib" /usr/lib /usr/local/lib /usr/local/marmot/lib /opt/marmot/lib 
          "$ENV{EXTERNLIBS}/marmot/lib"
          "$ENV{ProgramFiles}/marmot/lib"
          "C:/Program Files/marmot/lib"
          [HKEY_LOCAL_MACHINE\\SOFTWARE\\HLRS\\Marmot]/lib
)
FIND_LIBRARY(MARMOT_1sided_LIBRARY_DEBUG
          NAMES marmot-1sidedd
          PATHS ${MARMOT_HOME}/lib "$ENV{MARMOT_HOME}/lib" /usr/lib /usr/local/lib /usr/local/marmot/lib /opt/marmot/lib 
          "$ENV{EXTERNLIBS}/marmot/lib"
          "$ENV{ProgramFiles}/marmot/lib"
          "C:/Program Files/marmot/lib"
          [HKEY_LOCAL_MACHINE\\SOFTWARE\\HLRS\\Marmot]/lib
)

SET(MARMOT_FOUND "NO")
# was include path found at least?
IF (MARMOT_INCLUDE_DIR)
  IF (WIN32)
    # needed for gethostname on windows
    SET(EXTRA_SOCKET_LIBS "Ws2_32.lib")
  ENDIF (WIN32)

  # special case sun
  IF(CMAKE_SYSTEM_NAME STREQUAL "SunOS")
    SET(EXTRA_SOCKET_LIBS socket nsl)
  ENDIF(CMAKE_SYSTEM_NAME STREQUAL "SunOS")
  
  # adjust libvars & check
  BC_ADJUST_LIB_VARS(MARMOT)
  BC_ADJUST_LIB_VARS(MARMOT_profile)
  BC_ADJUST_LIB_VARS(MARMOT_1sided)
  # check
  IF (MARMOT_LIBRARY AND MARMOT_profile_LIBRARY)
    SET(MARMOT_FOUND "YES")
    # add all the necessary libraries
    SET(MARMOT_LIBRARIES ${MARMOT_profile_LIBRARY} ${MARMOT_1sided_LIBRARY} ${MARMOT_LIBRARY} ${CUBE_LIBRARIES} ${MPI_LIBRARIES} ${EXTRA_SOCKET_LIBS} CACHE STRING "The MARMOT_LIBRARIES" FORCE)
  ENDIF (MARMOT_LIBRARY AND MARMOT_profile_LIBRARY)

  # dump the info
  #BC_DUMP_LIB_SETUP(MARMOT)
  #BC_DUMP_LIB_SETUP(MARMOT_profile)
  
ENDIF (MARMOT_INCLUDE_DIR)

IF (MARMOT_FOUND)
  IF (NOT MARMOT_FIND_QUIETLY)
     MESSAGE(STATUS "Found MARMOT")
  ENDIF (NOT MARMOT_FIND_QUIETLY)
ELSE (MARMOT_FOUND)
  IF (MARMOT_FIND_REQUIRED)
     MESSAGE(FATAL_ERROR "Could not find MARMOT")
  ENDIF (MARMOT_FIND_REQUIRED)
ENDIF (MARMOT_FOUND)
# MARK_AS_ADVANCED(MARMOT_INCLUDE_PATH MARMOT_LIBRARY)
