# @file
#
# - Find OpenMPI
# This module looks for OpenMPI support. It will set the following variables:
#  OPENMPI_INCLUDE_DIR = where mpi.h can be found
#  OPENMPI_LIBRARY     = the library to link in
#  OPENMPI_LIBRARIES   = all the needed libraries
#  OPENMPI_mpicxx_LIBRARY_[RELEASE/DEBUG]
#  OPENMPI_orte_LIBRARY_[RELEASE/DEBUG]
#  OPENMPI_pal_LIBRARY_[RELEASE/DEBUG]
#
# -lmpi_cxx -lmpi -lopen-rte -lopen-pal
# Note: If OPENMPI_HOME is set by the user this is the place where the FIND_* macros
#       will look first
#
# @author Blasius Czink
#

# include (MarmotHelperMacros)
include(FindPackageHandleStandardArgs)

# get paths from windows registry (if possible)
GET_FILENAME_COMPONENT(OPENMPI_REG_PATH_BIN "[HKEY_LOCAL_MACHINE\\SOFTWARE\\Open\ MPI;OPAL_BINDIR]" ABSOLUTE)
GET_FILENAME_COMPONENT(OPENMPI_REG_PATH_INC "[HKEY_LOCAL_MACHINE\\SOFTWARE\\Open\ MPI;OPAL_INCLUDEDIR]" ABSOLUTE)
GET_FILENAME_COMPONENT(OPENMPI_REG_PATH_LIB "[HKEY_LOCAL_MACHINE\\SOFTWARE\\Open\ MPI;OPAL_LIBDIR]" ABSOLUTE)

# where to look for compile-wrapper stuff (optional)
SET(OPENMPI_COMPILE_WRAPPER_PATHS
  "$ENV{EXTERNLIBS}/openmpi/bin"
  "/usr/bin"
  "/usr/local/bin"
  "/usr/local/mpi/bin"
  "/opt/openmpi/bin"
  "/opt/mpi/bin"
  "$ENV{ProgramFiles}/OpenMPI/bin"
  "${OPENMPI_REG_PATH_BIN}"
)

# where to look for the include
SET(OPENMPI_INC_SEARCH_DIRS
  "$ENV{EXTERNLIBS}/openmpi/include"
  /usr/include 
  /usr/include/mpi
  /usr/local/include 
  /usr/local/mpi/include
  /opt/openmpi/include
  /opt/mpi/include
  "$ENV{ProgramFiles}/OpenMPI/include"
  "${OPENMPI_REG_PATH_INC}"
)

# where to look for the libraries
SET(OPENMPI_LIB_SEARCH_DIRS
  "$ENV{EXTERNLIBS}/openmpi/lib"
  /usr/lib
  /usr/local/lib
  /usr/local/mpi/lib
  /opt/openmpi/lib
  /opt/mpi/lib
  "$ENV{ProgramFiles}/OpenMPI/lib"
  "${OPENMPI_REG_PATH_LIB}"
)

# show an option for OPENMPI_HOME
IF (OPENMPI_HOME)
  IF (NOT "${OPENMPI_HOME}" STREQUAL "${OPENMPI_HOME_OLD}")
    # user changed OPENMPI_HOME
    # set variables to ...-NOTFOUND to reset (Is there a more elegant way to do that?)
    BC_RESET(MPI_CC)
    BC_RESET(MPI_CXX)
    BC_RESET(MPI_F77)
    BC_RESET(MPI_F90)
    BC_RESET(OPENMPI_INCLUDE_DIR)
    BC_RESET(OPENMPI_LIBRARY)
    BC_RESET(OPENMPI_mpicxx_LIBRARY)
    BC_RESET(OPENMPI_orte_LIBRARY)
    BC_RESET(OPENMPI_pal_LIBRARY)
    BC_RESET(OPENMPI_mpif77_LIBRARY)
    SET(OPENMPI_HOME_OLD "${OPENMPI_HOME}" CACHE INTERNAL "" FORCE)
  ENDIF (NOT "${OPENMPI_HOME}" STREQUAL "${OPENMPI_HOME_OLD}")
  # only look in OPENMPI_HOME if given
  SET(OPENMPI_COMPILE_WRAPPER_PATHS "${OPENMPI_HOME}/bin")
  SET(OPENMPI_INC_SEARCH_DIRS "${OPENMPI_HOME}/include")
  SET(OPENMPI_LIB_SEARCH_DIRS "${OPENMPI_HOME}/lib")
ELSE (OPENMPI_HOME)
  # nothing, so just create an empty field
  SET(OPENMPI_HOME "" CACHE PATH "Directory of the OpenMPI installation (change this if you installed OpenMPI in a nonstandard location")
  SET(OPENMPI_HOME_OLD "" CACHE INTERNAL "")
ENDIF (OPENMPI_HOME)

# cache variable for additional libs
IF (NOT OPENMPI_EXTRA_LIBS)
  SET(OPENMPI_EXTRA_LIBS "" CACHE STRING "Specify extra OPENMPI libraries to link against (use semicolon to delimit)")
ENDIF (NOT OPENMPI_EXTRA_LIBS)

FIND_PROGRAM(MPI_CC 
  NAMES mpicc mpcc mpcc_r mpxlc mpxlc_r mpiicc 
  PATHS ${OPENMPI_COMPILE_WRAPPER_PATHS}
  DOC "mpicc compile wrapper (optional)"
  NO_DEFAULT_PATH
)
FIND_PROGRAM(MPI_CXX 
  NAMES mpicxx mpcxx mpcxx_r mpxlcxx mpxlcxx_r mpiicxx
        mpic++ mpc++ mpc++_r mpxlc++ mpxlc++_r mpiic++
        mpiCC mpCC mpCC_r mpxlCC mpxlCC_r mpiiCC 
  PATHS ${OPENMPI_COMPILE_WRAPPER_PATHS}
  DOC "mpicxx compile wrapper (optional)"
  NO_DEFAULT_PATH
)
FIND_PROGRAM(MPI_F77 
  NAMES mpif77 mpf77 mpf77_r mpxlf77 mpxlf77_r mpiif77
  PATHS ${OPENMPI_COMPILE_WRAPPER_PATHS}
  DOC "mpif77 compile wrapper (optional)"
  NO_DEFAULT_PATH
)
FIND_PROGRAM(MPI_F90 
  NAMES mpif90 mpf90 mpf90_r mpxlf90 mpxlf90_r mpiif90
  PATHS ${OPENMPI_COMPILE_WRAPPER_PATHS}
  DOC "mpif90 compile wrapper (optional)"
  NO_DEFAULT_PATH
)
MARK_AS_ADVANCED(MPI_CC MPI_CXX MPI_F77 MPI_F90)

FIND_PATH(OPENMPI_INCLUDE_DIR mpi.h
  PATHS
  ${OPENMPI_INC_SEARCH_DIRS}
  DOC "OPENMPI include directory"
  NO_DEFAULT_PATH
)

FIND_LIBRARY(OPENMPI_LIBRARY
  NAMES mpi libmpi
  PATHS
  ${OPENMPI_LIB_SEARCH_DIRS}
  DOC "OPENMPI mpi library"
  NO_DEFAULT_PATH
)

FIND_LIBRARY(OPENMPI_mpicxx_LIBRARY
  NAMES mpi_cxx mpicxx libmpi_cxx libmpicxx
  PATHS
  ${OPENMPI_LIB_SEARCH_DIRS}
  DOC "OPENMPI mpicxx library"
  NO_DEFAULT_PATH
)

FIND_LIBRARY(OPENMPI_orte_LIBRARY
  NAMES open-rte libopen-rte
  PATHS
  ${OPENMPI_LIB_SEARCH_DIRS}
  DOC "OPENMPI orte library (optional)"
  NO_DEFAULT_PATH
)

FIND_LIBRARY(OPENMPI_pal_LIBRARY
  NAMES open-pal libopen-pal
  PATHS
  ${OPENMPI_LIB_SEARCH_DIRS}
  DOC "OPENMPI pal library (optional)"
  NO_DEFAULT_PATH
)

FIND_LIBRARY(OPENMPI_mpif77_LIBRARY
  NAMES libmpi_f77 mpi_f77
  PATHS
  ${OPENMPI_LIB_SEARCH_DIRS}
  DOC "OPENMPI mpi_f77 library (optional)"
  NO_DEFAULT_PATH
)

# mark some of the libs as advanced
MARK_AS_ADVANCED(OPENMPI_mpicxx_LIBRARY OPENMPI_orte_LIBRARY OPENMPI_pal_LIBRARY OPENMPI_mpif77_LIBRARY)

# only for windows & MS-VisualStudio
IF (WIN32 AND MSVC)
  FIND_LIBRARY(OPENMPI_LIBRARY_DEBUG
    NAMES mpid libmpid
    PATHS
    ${OPENMPI_LIB_SEARCH_DIRS}
    DOC "OPENMPI mpid (debug) library"
    NO_DEFAULT_PATH
  )
  
  FIND_LIBRARY(OPENMPI_mpicxx_LIBRARY_DEBUG
    NAMES mpi_cxxd mpicxxd libmpi_cxxd libmpicxxd
    PATHS
    ${OPENMPI_LIB_SEARCH_DIRS}
    DOC "OPENMPI mpicxxd (debug) library"
    NO_DEFAULT_PATH
  )
  
  FIND_LIBRARY(OPENMPI_orte_LIBRARY_DEBUG
    NAMES open-rted libopen-rted
    PATHS
    ${OPENMPI_LIB_SEARCH_DIRS}
    DOC "OPENMPI orte (debug) library (optional)"
    NO_DEFAULT_PATH
  )

  FIND_LIBRARY(OPENMPI_pal_LIBRARY_DEBUG
    NAMES open-pald libopen-pald
    PATHS
    ${OPENMPI_LIB_SEARCH_DIRS}
    DOC "OPENMPI pal (debug) library (optional)"
    NO_DEFAULT_PATH
  )
  
  FIND_LIBRARY(OPENMPI_mpif77_LIBRARY_DEBUG
    NAMES libmpi_f77d mpi_f77d
    PATHS
    ${OPENMPI_LIB_SEARCH_DIRS}
    DOC "OPENMPI mpi_f77d (debug) library (optional)"
    NO_DEFAULT_PATH
  )
  
  MARK_AS_ADVANCED(OPENMPI_mpicxx_LIBRARY_DEBUG OPENMPI_orte_LIBRARY_DEBUG OPENMPI_pal_LIBRARY_DEBUG OPENMPI_mpif77_LIBRARY_DEBUG)
  
  IF (MSVC_IDE)
    # VS-IDE
    IF (OPENMPI_LIBRARY AND OPENMPI_LIBRARY_DEBUG AND OPENMPI_mpicxx_LIBRARY AND OPENMPI_mpicxx_LIBRARY_DEBUG)
       SET(OPENMPI_LIBRARIES optimized ${OPENMPI_LIBRARY} optimized ${OPENMPI_mpicxx_LIBRARY} debug ${OPENMPI_LIBRARY_DEBUG} debug ${OPENMPI_mpicxx_LIBRARY_DEBUG})
    ENDIF (OPENMPI_LIBRARY AND OPENMPI_LIBRARY_DEBUG AND OPENMPI_mpicxx_LIBRARY AND OPENMPI_mpicxx_LIBRARY_DEBUG)
    #IF (OPENMPI_orte_LIBRARY AND OPENMPI_orte_LIBRARY_DEBUG)
    #   SET(OPENMPI_LIBRARIES ${OPENMPI_LIBRARIES} optimized ${OPENMPI_orte_LIBRARY} debug ${OPENMPI_orte_LIBRARY_DEBUG})
    #ENDIF (OPENMPI_orte_LIBRARY AND OPENMPI_orte_LIBRARY_DEBUG)
    #IF (OPENMPI_pal_LIBRARY AND OPENMPI_pal_LIBRARY_DEBUG)
    #   SET(OPENMPI_LIBRARIES ${OPENMPI_LIBRARIES} optimized ${OPENMPI_pal_LIBRARY} debug ${OPENMPI_pal_LIBRARY_DEBUG})
    #ENDIF (OPENMPI_pal_LIBRARY AND OPENMPI_pal_LIBRARY_DEBUG)
    # optional fortran stuff
    IF (MARMOT_ENABLE_FORTRAN AND OPENMPI_mpif77_LIBRARY AND OPENMPI_mpif77_LIBRARY_DEBUG)
      SET(OPENMPI_LIBRARIES ${OPENMPI_LIBRARIES} optimized ${OPENMPI_mpif77_LIBRARY} debug ${OPENMPI_mpif77_LIBRARY_DEBUG})
    ENDIF (MARMOT_ENABLE_FORTRAN AND OPENMPI_mpif77_LIBRARY AND OPENMPI_mpif77_LIBRARY_DEBUG)
  ELSE (MSVC_IDE)
    # nmake
    STRING(TOLOWER ${CMAKE_BUILD_TYPE} CMAKE_BUILD_TYPE_TOLOWER)
    IF(CMAKE_BUILD_TYPE_TOLOWER MATCHES debug)
      SET(OPENMPI_LIBRARIES ${OPENMPI_LIBRARY_DEBUG} ${OPENMPI_mpicxx_LIBRARY_DEBUG})
      #IF(OPENMPI_orte_LIBRARY_DEBUG)
      #  SET(OPENMPI_LIBRARIES ${OPENMPI_LIBRARIES} ${OPENMPI_orte_LIBRARY_DEBUG})
      #ENDIF(OPENMPI_orte_LIBRARY_DEBUG)
      #IF(OPENMPI_pal_LIBRARY_DEBUG)
      #  SET(OPENMPI_LIBRARIES ${OPENMPI_LIBRARIES} ${OPENMPI_pal_LIBRARY_DEBUG})
      #ENDIF(OPENMPI_pal_LIBRARY_DEBUG)
      IF (MARMOT_ENABLE_FORTRAN AND OPENMPI_mpif77_LIBRARY_DEBUG)
        SET(OPENMPI_LIBRARIES ${OPENMPI_LIBRARIES} ${OPENMPI_mpif77_LIBRARY_DEBUG})
      ENDIF (MARMOT_ENABLE_FORTRAN AND OPENMPI_mpif77_LIBRARY_DEBUG)
    ELSE(CMAKE_BUILD_TYPE_TOLOWER MATCHES debug)
      SET(OPENMPI_LIBRARIES ${OPENMPI_LIBRARY} ${OPENMPI_mpicxx_LIBRARY})
      #IF(OPENMPI_orte_LIBRARY)
      #  SET(OPENMPI_LIBRARIES ${OPENMPI_LIBRARIES} ${OPENMPI_orte_LIBRARY})
      #ENDIF(OPENMPI_orte_LIBRARY)
      #IF(OPENMPI_pal_LIBRARY)
      #  SET(OPENMPI_LIBRARIES ${OPENMPI_LIBRARIES} ${OPENMPI_pal_LIBRARY})
      #ENDIF(OPENMPI_pal_LIBRARY)
      IF (MARMOT_ENABLE_FORTRAN AND OPENMPI_mpif77_LIBRARY)
        SET(OPENMPI_LIBRARIES ${OPENMPI_LIBRARIES} ${OPENMPI_mpif77_LIBRARY})
      ENDIF (MARMOT_ENABLE_FORTRAN AND OPENMPI_mpif77_LIBRARY)
    ENDIF(CMAKE_BUILD_TYPE_TOLOWER MATCHES debug)
  ENDIF (MSVC_IDE)

  FIND_PACKAGE_HANDLE_STANDARD_ARGS(OPENMPI "Could not locate OpenMPI" OPENMPI_LIBRARY OPENMPI_LIBRARY_DEBUG OPENMPI_mpicxx_LIBRARY OPENMPI_mpicxx_LIBRARY_DEBUG OPENMPI_INCLUDE_DIR)

ELSE (WIN32 AND MSVC)  

  # nothing special to be done for the rest of the world
  SET(OPENMPI_LIBRARIES ${OPENMPI_LIBRARY} ${OPENMPI_mpicxx_LIBRARY})
  # append optional libs
  #IF(OPENMPI_orte_LIBRARY)
  #  SET(OPENMPI_LIBRARIES ${OPENMPI_LIBRARIES} ${OPENMPI_orte_LIBRARY})
  #ENDIF(OPENMPI_orte_LIBRARY)
  #IF(OPENMPI_pal_LIBRARY)
  #  SET(OPENMPI_LIBRARIES ${OPENMPI_LIBRARIES} ${OPENMPI_pal_LIBRARY})
  #ENDIF(OPENMPI_pal_LIBRARY)
  # for fortran
  IF (MARMOT_ENABLE_FORTRAN AND OPENMPI_mpif77_LIBRARY)
    SET(OPENMPI_LIBRARIES ${OPENMPI_LIBRARIES} ${OPENMPI_mpif77_LIBRARY})
  ENDIF (MARMOT_ENABLE_FORTRAN AND OPENMPI_mpif77_LIBRARY)
  
  FIND_PACKAGE_HANDLE_STANDARD_ARGS(OPENMPI "Could not locate OpenMPI" OPENMPI_LIBRARY OPENMPI_mpicxx_LIBRARY OPENMPI_INCLUDE_DIR)

ENDIF (WIN32 AND MSVC)

# Set path for mpif.h (If Fortran enabled)
IF (MARMOT_ENABLE_FORTRAN)
    SET (MPI_F_INCLUDE_DIR "${MPI_INCLUDE_DIR}" CACHE PATH "Path to the Fortran header of MPI.")
ENDIF (MARMOT_ENABLE_FORTRAN)

IF (OPENMPI_FOUND)
  # add extra libs specified by user
  IF (OPENMPI_EXTRA_LIBS)
    SET(OPENMPI_LIBRARIES ${OPENMPI_LIBRARIES} ${OPENMPI_EXTRA_LIBS})
  ENDIF (OPENMPI_EXTRA_LIBS)
  # set variables MPI_INCLUDE_DIR and MPI_LIBRARIES which will be used by marmot
  SET (MPI_INCLUDE_DIR ${OPENMPI_INCLUDE_DIR} CACHE PATH "The MPI-Header path" FORCE)
  SET (MPI_LIBRARIES ${OPENMPI_LIBRARIES} CACHE STRING "The MPI-Libraries" FORCE)
  MARK_AS_ADVANCED (MPI_LIBRARIES MPI_INCLUDE_DIR)
ENDIF (OPENMPI_FOUND)
