# @file
# 
# Provides two helper macros:
#
# DETERMINE_FORTRAN_MANGLING_SCHEME ()
#  - Tries to figure out the used fortran-compiler's name mangling scheme
#  - It provides an option MARMOT_DETECT_FORTRAN_NM (defaults to "ON")
#    If you set this option to "OFF" you can specify the name manglig scheme manually
#
# FORTRAN_MANGLING_TRANSFORM (src_file dest_file)
#  - generates a file according to the set name mangling scheme
#    in the source-file you have entries like "#define MPI_ABORT"
#    which gets expanded to something like "#define MPI_ABORT mpi_abort_" in the destination file
#
# @author Blasius Czink
#

MACRO (DETERMINE_FORTRAN_MANGLING_SCHEME)

 # ENABLE_LANGUAGE(Fortran)

 OPTION (MARMOT_DETECT_FORTRAN_NM "Disable to manually set name mangling scheme" ON)
 MARK_AS_ADVANCED (MARMOT_DETECT_FORTRAN_NM)
 
 IF (MARMOT_DETECT_FORTRAN_NM)
   
   # Make sure that the following tests use the C and Fortran flags corresponding
   # to the current build type. These flags are stored in the variables TMP_C_FLAGS 
   # and TMP_Fortran_FLAGS, respectively, and are used in the generated CMakeLists files.
    IF(NOT CMAKE_BUILD_TYPE)
      SET(TMP_C_FLAGS       ${CMAKE_C_FLAGS})
      SET(TMP_Fortran_FLAGS ${CMAKE_Fortran_FLAGS})
    ENDIF(NOT CMAKE_BUILD_TYPE)
    IF(CMAKE_BUILD_TYPE MATCHES "Default")
      SET(TMP_C_FLAGS       ${CMAKE_C_FLAGS})
      SET(TMP_Fortran_FLAGS ${CMAKE_Fortran_FLAGS})
    ENDIF(CMAKE_BUILD_TYPE MATCHES "Default")
    IF(CMAKE_BUILD_TYPE MATCHES "Release")
      SET(TMP_C_FLAGS       ${CMAKE_C_FLAGS_RELEASE})
      SET(TMP_Fortran_FLAGS ${CMAKE_Fortran_FLAGS_RELEASE})
    ENDIF(CMAKE_BUILD_TYPE MATCHES "Release")
    IF(CMAKE_BUILD_TYPE MATCHES "Debug")
      SET(TMP_C_FLAGS       ${CMAKE_C_FLAGS_DEBUG})
      SET(TMP_Fortran_FLAGS ${CMAKE_Fortran_FLAGS_DEBUG})
    ENDIF(CMAKE_BUILD_TYPE MATCHES "Debug")
    IF(CMAKE_BUILD_TYPE MATCHES "RelWithDebInfo")
      SET(TMP_C_FLAGS       ${CMAKE_C_FLAGS_RELWITHDEBINFO})
      SET(TMP_Fortran_FLAGS ${CMAKE_Fortran_FLAGS_RELWITHDEBINFO})
    ENDIF(CMAKE_BUILD_TYPE MATCHES "RelWithDebInfo")
    IF(CMAKE_BUILD_TYPE MATCHES "MinSizeRel")
      SET(TMP_C_FLAGS       ${CMAKE_C_FLAGS_MINSIZE})
      SET(TMP_Fortran_FLAGS ${CMAKE_Fortran_FLAGS_MINSIZE})
    ENDIF(CMAKE_BUILD_TYPE MATCHES "MinSizeRel")
  
    # Create a CMakeLists.txt file which will generate the "flib" library
    FILE(WRITE ${PROJECT_BINARY_DIR}${CMAKE_FILES_DIRECTORY}/CMakeTmp/CMakeLists.txt
      "CMAKE_MINIMUM_REQUIRED(VERSION 2.4)\n"
      "PROJECT(FortranTest Fortran)\n"
      "SET(CMAKE_VERBOSE_MAKEFILE ON)\n"
      "SET(CMAKE_Fortran_FLAGS \"${TMP_Fortran_FLAGS}\")\n"
      "ADD_LIBRARY(flib ftest.f)\n"
      )
    
    # Create a simple Fortran source which defines two subroutines, "mysub" and "my_sub"
    FILE(WRITE ${PROJECT_BINARY_DIR}${CMAKE_FILES_DIRECTORY}/CMakeTmp/ftest.f
      "        SUBROUTINE mysub\n"
      "        RETURN\n"
      "        END\n"
      "        SUBROUTINE my_sub\n"
      "        RETURN\n"
      "        END\n"
      )
    
    # Use TRY_COMPILE to make the target "flib"
    TRY_COMPILE(
      FTEST_OK
      ${PROJECT_BINARY_DIR}${CMAKE_FILES_DIRECTORY}/CMakeTmp
      ${PROJECT_BINARY_DIR}${CMAKE_FILES_DIRECTORY}/CMakeTmp
      flib
      OUTPUT_VARIABLE MY_OUTPUT)
  
    # message("${MY_OUTPUT}")
    
    # Initialize the name mangling schemes for symbol names 
    # with and without underscores
    SET (FORTRAN_SCHEME_NO_UNDERSCORES "")
    SET (FORTRAN_SCHEME_WITH_UNDERSCORES "")
    
    # Continue only if we were successful in creating the "flib" library
    IF(FTEST_OK)
      
      # CASE 1: symbol names WITHOUT undersores
      # ---------------------------------------
      
      # Overwrite CMakeLists.txt with one which will generate the "ctest1" executable
      FILE(WRITE ${PROJECT_BINARY_DIR}${CMAKE_FILES_DIRECTORY}/CMakeTmp/CMakeLists.txt
        "CMAKE_MINIMUM_REQUIRED(VERSION 2.4)\n"
        "PROJECT(FortranTest C)\n"
        "SET(CMAKE_VERBOSE_MAKEFILE ON)\n"
        "SET(CMAKE_C_FLAGS \"${TMP_C_FLAGS}\")\n"
        "ADD_EXECUTABLE(ctest1 ctest1.c)\n"
        "FIND_LIBRARY(FLIB flib ${PROJECT_BINARY_DIR}${CMAKE_FILES_DIRECTORY}/CMakeTmp)\n"
        "TARGET_LINK_LIBRARIES(ctest1 \${FLIB})\n")
      
      # Define the list "options" of all possible schemes that we want to consider
      # Get its length and initialize the counter "iopt" to zero
      SET(options mysub mysub_ mysub__ MYSUB MYSUB_ MYSUB__)
      LIST(LENGTH options imax)
      SET(iopt 0)
      
      # We will attempt to sucessfully generate the "ctest" executable as long as
      # there still are entries in the "options" list
      WHILE(${iopt} LESS ${imax})   
        
        # Get the current list entry (current scheme)
        LIST(GET options ${iopt} opt)
        
        # Generate C source which calls the "mysub" function using the current scheme
        FILE(WRITE ${PROJECT_BINARY_DIR}${CMAKE_FILES_DIRECTORY}/CMakeTmp/ctest1.c
          "int main(){${opt}();return(0);}\n")
        
        # Use TRY_COMPILE to make the "ctest1" executable from the current C source
        # and linking to the previously created "flib" library.
        TRY_COMPILE(
          CTEST_OK
          ${PROJECT_BINARY_DIR}${CMAKE_FILES_DIRECTORY}/CMakeTmp
          ${PROJECT_BINARY_DIR}${CMAKE_FILES_DIRECTORY}/CMakeTmp
          ctest1
          OUTPUT_VARIABLE MY_OUTPUT)
        
        # To ensure we do not use stuff from the previous attempts, we must remove the
        # CMakeFiles directory.
        # ??? I didn't think I'll have to do this, but it doesn't work otherwise
        FILE(REMOVE_RECURSE ${PROJECT_BINARY_DIR}${CMAKE_FILES_DIRECTORY}/CMakeTmp/CMakeFiles)
        
        # Test if we successfully created the "ctest" executable.
        # If yes, flag that we have successfuly determined the name mangling scheme,
        # save the current scheme, and set the counter "iopt" to "imax" so that we
        # exit the while loop.
        # Otherwise, increment the counter "iopt" and go back in the while loop.
        IF(CTEST_OK)
          SET(FORTRAN_SCHEME_NO_UNDERSCORES ${opt})
          SET(iopt ${imax})
        ELSE(CTEST_OK)
          MATH(EXPR iopt ${iopt}+1)
        ENDIF(CTEST_OK)
        
      ENDWHILE(${iopt} LESS ${imax})   
      
      # CASE 2: symbol names WITH undersores
      # ------------------------------------
      
      FILE(WRITE ${PROJECT_BINARY_DIR}${CMAKE_FILES_DIRECTORY}/CMakeTmp/CMakeLists.txt
        "CMAKE_MINIMUM_REQUIRED(VERSION 2.4)\n"
        "PROJECT(FortranTest C)\n"
        "SET(CMAKE_VERBOSE_MAKEFILE ON)\n"
        "SET(CMAKE_C_FLAGS \"${TMP_C_FLAGS}\")\n"
        "ADD_EXECUTABLE(ctest2 ctest2.c)\n"
        "FIND_LIBRARY(FLIB flib ${PROJECT_BINARY_DIR}${CMAKE_FILES_DIRECTORY}/CMakeTmp)\n"
        "TARGET_LINK_LIBRARIES(ctest2 \${FLIB})\n")
      
      SET(options my_sub my_sub_ my_sub__ MY_SUB MY_SUB_ MY_SUB__)
      LIST(LENGTH options imax)
      SET(iopt 0)
      
      WHILE(${iopt} LESS ${imax})   
        LIST(GET options ${iopt} opt)
        FILE(WRITE ${PROJECT_BINARY_DIR}${CMAKE_FILES_DIRECTORY}/CMakeTmp/ctest2.c
          "int main(){${opt}();return(0);}\n")
        TRY_COMPILE(
          CTEST_OK
          ${PROJECT_BINARY_DIR}${CMAKE_FILES_DIRECTORY}/CMakeTmp
          ${PROJECT_BINARY_DIR}${CMAKE_FILES_DIRECTORY}/CMakeTmp
          ctest2
          OUTPUT_VARIABLE MY_OUTPUT)
        FILE(REMOVE_RECURSE ${PROJECT_BINARY_DIR}${CMAKE_FILES_DIRECTORY}/CMakeTmp/CMakeFiles)
        IF(CTEST_OK)
          SET(FORTRAN_SCHEME_WITH_UNDERSCORES ${opt})
          SET(iopt ${imax})
        ELSE(CTEST_OK)
          MATH(EXPR iopt ${iopt}+1)
        ENDIF(CTEST_OK)
      ENDWHILE(${iopt} LESS ${imax})   
      
    ENDIF(FTEST_OK)
  
    # If the name mangling scheme of symbol names not containing underscores
    # was successfully determined, set the appropriate regexp for replacement
  
    SET (FORTRAN_FUNC_NAMES_CAPITAL FALSE)
    SET (FORTRAN_NM_SCHEME_REPLACE  "")
    SET (FORTRAN_NM_SCHEME_REPLACE_ "")
    
    IF (FORTRAN_SCHEME_NO_UNDERSCORES)
  
      IF (FORTRAN_SCHEME_NO_UNDERSCORES MATCHES "mysub")
        SET (FORTRAN_NM_SCHEME_REPLACE "\\1")
        # SET(CPP_macro "#define F77_FUNC(name,NAME) name")
      ENDIF (FORTRAN_SCHEME_NO_UNDERSCORES MATCHES "mysub")
      
      IF (FORTRAN_SCHEME_NO_UNDERSCORES MATCHES "mysub_")
        SET (FORTRAN_NM_SCHEME_REPLACE "\\1_")
        # SET(CPP_macro "#define F77_FUNC(name,NAME) name ## _")
      ENDIF (FORTRAN_SCHEME_NO_UNDERSCORES MATCHES "mysub_")      
      
      IF (FORTRAN_SCHEME_NO_UNDERSCORES MATCHES "mysub__")
        SET (FORTRAN_NM_SCHEME_REPLACE "\\1__")
        # SET(CPP_macro "#define F77_FUNC(name,NAME) name ## __")
      ENDIF (FORTRAN_SCHEME_NO_UNDERSCORES MATCHES "mysub__")
      
      IF (FORTRAN_SCHEME_NO_UNDERSCORES MATCHES "MYSUB")
        SET (FORTRAN_NM_SCHEME_REPLACE "\\1")
        SET (FORTRAN_FUNC_NAMES_CAPITAL TRUE)
        # SET(CPP_macro "#define F77_FUNC(name,NAME) NAME")
      ENDIF (FORTRAN_SCHEME_NO_UNDERSCORES MATCHES "MYSUB")
      
      IF (FORTRAN_SCHEME_NO_UNDERSCORES MATCHES "MYSUB_")
        SET (FORTRAN_NM_SCHEME_REPLACE "\\1_")
        SET (FORTRAN_FUNC_NAMES_CAPITAL TRUE)
        # SET(CPP_macro "#define F77_FUNC(name,NAME) NAME ## _")
      ENDIF (FORTRAN_SCHEME_NO_UNDERSCORES MATCHES "MYSUB_")
      
      IF (FORTRAN_SCHEME_NO_UNDERSCORES MATCHES "MYSUB__")
        SET (FORTRAN_NM_SCHEME_REPLACE "\\1__")
        SET (FORTRAN_FUNC_NAMES_CAPITAL TRUE)
        # SET(CPP_macro "#define F77_FUNC(name,NAME) NAME ## __")
      ENDIF (FORTRAN_SCHEME_NO_UNDERSCORES MATCHES "MYSUB__")
  
    ENDIF (FORTRAN_SCHEME_NO_UNDERSCORES)
  
    # SET(DEFINE_F77_FUNC ${CPP_macro} CACHE STRING "CPP macro for name mangling scheme of symbols without underscores")
    
    IF(FORTRAN_SCHEME_NO_UNDERSCORES)
      # MESSAGE("Name mangling scheme for symbol names without underscores:\n"
      #  "   mysub  ->  ${FORTRAN_SCHEME_NO_UNDERSCORES}\n"
      #  "   ${DEFINE_F77_FUNC}")
    ELSE(FORTRAN_SCHEME_NO_UNDERSCORES)
      MESSAGE("Unable to determine name mangling scheme for symbol names without underscores!")
    ENDIF(FORTRAN_SCHEME_NO_UNDERSCORES)
    
    # If the name mangling scheme of symbol names containing underscores
    # was successfully determined, set the appropriate C preprocessor macro
    
    IF (FORTRAN_SCHEME_WITH_UNDERSCORES)
      
      IF (FORTRAN_SCHEME_WITH_UNDERSCORES MATCHES "my_sub")
        SET (FORTRAN_NM_SCHEME_REPLACE_ "\\1")
        # SET(CPP_macro "#define F77_FUNC_(name,NAME) name")
      ENDIF (FORTRAN_SCHEME_WITH_UNDERSCORES MATCHES "my_sub")
      
      IF (FORTRAN_SCHEME_WITH_UNDERSCORES MATCHES "my_sub_")
        SET (FORTRAN_NM_SCHEME_REPLACE_ "\\1_")
        # SET(CPP_macro "#define F77_FUNC_(name,NAME) name ## _")
      ENDIF (FORTRAN_SCHEME_WITH_UNDERSCORES MATCHES "my_sub_")      
      
      IF (FORTRAN_SCHEME_WITH_UNDERSCORES MATCHES "my_sub__")
        SET (FORTRAN_NM_SCHEME_REPLACE_ "\\1__")
        # SET(CPP_macro "#define F77_FUNC_(name,NAME) name ## __")
      ENDIF (FORTRAN_SCHEME_WITH_UNDERSCORES MATCHES "my_sub__")
      
      IF (FORTRAN_SCHEME_WITH_UNDERSCORES MATCHES "MY_SUB")
        SET (FORTRAN_NM_SCHEME_REPLACE_ "\\1")
        SET (FORTRAN_FUNC_NAMES_CAPITAL TRUE)
        # SET(CPP_macro "#define F77_FUNC_(name,NAME) NAME")
      ENDIF (FORTRAN_SCHEME_WITH_UNDERSCORES MATCHES "MY_SUB")
      
      IF (FORTRAN_SCHEME_WITH_UNDERSCORES MATCHES "MY_SUB_")
        SET (FORTRAN_NM_SCHEME_REPLACE_ "\\1_")
        SET (FORTRAN_FUNC_NAMES_CAPITAL TRUE)
        # SET(CPP_macro "#define F77_FUNC_(name,NAME) NAME ## _")
      ENDIF (FORTRAN_SCHEME_WITH_UNDERSCORES MATCHES "MY_SUB_")
      
      IF (FORTRAN_SCHEME_WITH_UNDERSCORES MATCHES "MY_SUB__")
        SET (FORTRAN_NM_SCHEME_REPLACE_ "\\1__")
        SET (FORTRAN_FUNC_NAMES_CAPITAL TRUE)
        # SET(CPP_macro "#define F77_FUNC_(name,NAME) NAME ## __")
      ENDIF (FORTRAN_SCHEME_WITH_UNDERSCORES MATCHES "MY_SUB__")
      
    ENDIF (FORTRAN_SCHEME_WITH_UNDERSCORES)
    
    # SET (DEFINE_F77_FUNC_ "${CPP_macro}" CACHE STRING "CPP macro for name mangling scheme of symbols with underscores")
  
    IF(FORTRAN_SCHEME_WITH_UNDERSCORES)
      # MESSAGE("Name mangling scheme for symbol names with underscores:\n"
      #   "   my_sub ->  ${FORTRAN_SCHEME_WITH_UNDERSCORES}\n"
      #   "   ${DEFINE_F77_FUNC_}")
    ELSE(FORTRAN_SCHEME_WITH_UNDERSCORES)
      MESSAGE("Unable to determine name mangling scheme for symbol names with underscores!")
    ENDIF(FORTRAN_SCHEME_WITH_UNDERSCORES)

  ENDIF(MARMOT_DETECT_FORTRAN_NM)

  SET (FORTRAN_FUNC_NAMES_CAPITAL ${FORTRAN_FUNC_NAMES_CAPITAL} CACHE BOOL "TRUE if fortran function-names are capital" FORCE)
  SET (FORTRAN_SCHEME_NO_UNDERSCORES "${FORTRAN_SCHEME_NO_UNDERSCORES}" CACHE STRING "Name mangling scheme (symbol names without underscores)" FORCE)
  SET (FORTRAN_SCHEME_WITH_UNDERSCORES "${FORTRAN_SCHEME_WITH_UNDERSCORES}" CACHE STRING "Name mangling scheme (symbol names with underscores)" FORCE)
  SET (FORTRAN_NM_SCHEME_REPLACE  "${FORTRAN_NM_SCHEME_REPLACE}" CACHE STRING "CMake replacement string used in name-mangling (func-name without underscore)" FORCE)
  SET (FORTRAN_NM_SCHEME_REPLACE_ "${FORTRAN_NM_SCHEME_REPLACE_}" CACHE STRING "CMake replacement string used in name-mangling (func-name with underscore)" FORCE)
  MARK_AS_ADVANCED(FORTRAN_FUNC_NAMES_CAPITAL FORTRAN_NM_SCHEME_REPLACE FORTRAN_NM_SCHEME_REPLACE_
                   FORTRAN_SCHEME_NO_UNDERSCORES FORTRAN_SCHEME_WITH_UNDERSCORES)

  # setting define macros like the following might be helpful too ...
  # #define F77_FUNC(name,NAME) name ## _
  # #define F77_FUNC_(name,NAME) name ## __

ENDMACRO (DETERMINE_FORTRAN_MANGLING_SCHEME)

# match capture regexp  ... #define BLAH
# "#define\\s+([A-Z,a-z,_]*)"

MACRO (FORTRAN_MANGLING_TRANSFORM src_file dest_file)
  # a simple safety check
  IF (NOT ${src_file} STREQUAL ${dest_file})
    IF (EXISTS ${src_file})
      # read source-file and write out a transformed version according to regexp set by previous macro
      FILE (READ ${src_file} src_file_content)
      # open destination file
      FILE (WRITE ${dest_file} "/*  generated from ${src_file} */\n")
      IF (src_file_content)
        # create a list of functions from the define lines
        # MESSAGE ("SRC: ${src_file_content}")
        # create list with matching defines
        STRING (REGEX MATCHALL "#define[ ]+([A-Z,a-z,_,0-9]+)" tmp_dest_list "${src_file_content}")
        # work on the list
        LIST(LENGTH tmp_dest_list dest_list_max)
        FOREACH (myarg ${tmp_dest_list})
          STRING (REGEX REPLACE "#define[ ]+([A-Z,a-z,_,0-9]+)" "\\1" fortran_func_name ${myarg})
          IF (FORTRAN_FUNC_NAMES_CAPITAL)
            # if capital
            STRING (TOUPPER ${fortran_func_name} uc_func_name)
            # does if contain underscores?
            STRING (REGEX MATCH "_"  has_underscores ${fortran_func_name})
            IF (has_underscores)
              STRING (REGEX REPLACE "([A-Z,a-z,_,0-9]+)" "${FORTRAN_NM_SCHEME_REPLACE_}" uc_func_name ${uc_func_name})
            ELSE (has_underscores)
              STRING (REGEX REPLACE "([A-Z,a-z,_,0-9]+)" "${FORTRAN_NM_SCHEME_REPLACE}" uc_func_name ${uc_func_name})
            ENDIF (has_underscores)
            # create line in destination file
            FILE (APPEND ${dest_file} "#define ${fortran_func_name} ${uc_func_name}\n")
          ELSE (FORTRAN_FUNC_NAMES_CAPITAL)
            # not capital
            STRING (TOLOWER ${fortran_func_name} lc_func_name)
            # does if contain underscores?
            STRING (REGEX MATCH "_"  has_underscores ${fortran_func_name})
            IF (has_underscores)
              STRING (REGEX REPLACE "([A-Z,a-z,_,0-9]+)" "${FORTRAN_NM_SCHEME_REPLACE_}" lc_func_name ${lc_func_name})
            ELSE (has_underscores)
              STRING (REGEX REPLACE "([A-Z,a-z,_,0-9]+)" "${FORTRAN_NM_SCHEME_REPLACE}" lc_func_name ${lc_func_name})
            ENDIF (has_underscores)
            # create line in destination file
            FILE (APPEND ${dest_file} "#define ${fortran_func_name} ${lc_func_name}\n")
          ENDIF (FORTRAN_FUNC_NAMES_CAPITAL)
          #LIST (APPEND my_funcs ${fortran_func_name})
        ENDFOREACH (myarg)
        # MESSAGE ("MATCHED: ${dest_list_max} entries\n")
      ELSE (src_file_content)
        MESSAGE ("FORTRAN_MANGLING_TRANSFORM() - ${src_file} empty?")
      ENDIF (src_file_content)
      
    ELSE (EXISTS ${src_file})
      MESSAGE ("FORTRAN_MANGLING_TRANSFORM - Error: ${src_file} does not exist!")
    ENDIF (EXISTS ${src_file})
  ELSE (NOT ${src_file} STREQUAL ${dest_file})
    MESSAGE ("FORTRAN_MANGLING_TRANSFORM - Error: The same file was passed as src- and dst-file!")
  ENDIF (NOT ${src_file} STREQUAL ${dest_file})
ENDMACRO (FORTRAN_MANGLING_TRANSFORM src_file dest_file)
