# @file HaveMPI2Helpers.cmake
# 
# Support Macros for detection of MPI2 Availability.
# (Fine grained) 
#
# @author Tobias Hilbrich
#
# Contains:
#   * HaveMPI_Call2FTName
#       -> MPI Call name to filepath of feature test of call
#
#   * HaveMPI_Call2HaveVariable
#       -> returns name of the CMake variable that determines
#          whether given MPI2 call is available or not
#
#   * HaveMPI_GetMPI2CallList
#       -> returns list of all MPI2 calls
#
#   * HaveMPI_IsMPI2CallAvailable
#       -> returns whether feature test for this call was successfull
#
#   * HaveMPI_IsFullMPI2
#       -> returns TRUE if this is a full MPI2 (after feature tests 
#          were run)
#
#   * HaveMPI_GetUnsupportedCallList
#       -> returns list of MPI2 calls not supported by this MPI implem.
#
#   * MARMOT_GEN_CREATE_WRAPPER
#       -> hook macro used to generate feature tests
#
#   * directly executed script
#       -> starts generation of MPI2 feature tests

##Needed to parse headers !!!
INCLUDE (generation/GenerateMacros)

#===================================
# Macro: HaveMPI_Call2FTName
#===================================
#Takes the name of an MPI2 call and
#returns the filepath of its associated
#feature test.
#
MACRO (HaveMPI_Call2FTName callname out_filepath)
    STRING(TOLOWER "${callname}" HaveMPI_Call2FTName_temp)
    SET (${out_filepath} "${MARMOT_MPI2FT_DIR}/Have_${HaveMPI_Call2FTName_temp}.cc")
ENDMACRO (HaveMPI_Call2FTName)

#===================================
# Macro: HaveMPI_Call2HaveVariable
#===================================
#Takes the name of an MPI2 call and
#returns the filepath of its associated
#feature test.
#
MACRO (HaveMPI_Call2HaveVariable callname out_havevar)
    STRING(TOUPPER "${callname}" HaveMPI_Call2HaveVariable_temp)
    SET (${out_havevar} "IS_${HaveMPI_Call2HaveVariable_temp}_PRESENT")
ENDMACRO (HaveMPI_Call2HaveVariable)

#===================================
# Macro: HaveMPI_GetMPI2CallList
#===================================
#Sets given variable to a list of
#all MPI2 calls.
MACRO (HaveMPI_GetMPI2CallList out_var)
    SET (${out_var} "${MARMOT_MPI2_CALLS}")
ENDMACRO (HaveMPI_GetMPI2CallList)

#===================================
# Macro: HaveMPI_IsMPI2CallAvailable
#===================================
#Must only be called after having run
#feature tests for all MPI2 calls.
#
# Returns in out_var TRUE if the MPI2
# call is supported, FALSE otherwise.
MACRO (HaveMPI_IsMPI2CallAvailable callname out_var)
    HaveMPI_Call2HaveVariable ("${callname}" HaveMPI_IsMPI2CallAvailable_temp)
    IF (${${HaveMPI_IsMPI2CallAvailable_temp}}) 
        SET (${out_var} TRUE)
    ELSE (${${HaveMPI_IsMPI2CallAvailable_temp}})
        SET (${out_var} FALSE)
    ENDIF (${${HaveMPI_IsMPI2CallAvailable_temp}})
ENDMACRO (HaveMPI_IsMPI2CallAvailable)

#===================================
# Macro: HaveMPI_IsFullMPI2
#===================================
#Must only be called after having run
#feature tests for all MPI2 calls.
#
# Returns in out_var TRUE if full MPI2
# support is given, FALSE otherwise.
MACRO (HaveMPI_IsFullMPI2 out_var)
    HaveMPI_GetMPI2CallList (HaveMPI_IsFullMPI2_list)
    SET (${out_var} TRUE)
    FOREACH (call ${HaveMPI_IsFullMPI2_list})
        HaveMPI_Call2HaveVariable ("${call}" HaveMPI_IsFullMPI2_temp)
        IF (NOT ${HaveMPI_IsFullMPI2_temp}) 
            SET (${out_var} FALSE)
        ENDIF (NOT ${HaveMPI_IsFullMPI2_temp})
    ENDFOREACH (call)
ENDMACRO (HaveMPI_IsFullMPI2)

#===================================
# Macro: HaveMPI_GetUnsupportedCallList
#===================================
#Must only be called after having run
#feature tests for all MPI2 calls.
#
# Returns in out_var TRUE if full MPI2
# support is given, FALSE otherwise.
MACRO (HaveMPI_GetUnsupportedCallList out_var)
    HaveMPI_GetMPI2CallList (HaveMPI_GetUnsupportedCallList_list)
    SET (${out_var} "")
    FOREACH (call ${HaveMPI_GetUnsupportedCallList_list})
        HaveMPI_Call2HaveVariable ("${call}" HaveMPI_GetUnsupportedCallList_temp)
        IF (NOT ${HaveMPI_GetUnsupportedCallList_temp})
             SET (${out_var} ${${out_var}} "${call}")
        ENDIF (NOT ${HaveMPI_GetUnsupportedCallList_temp})
    ENDFOREACH (call)
ENDMACRO (HaveMPI_GetUnsupportedCallList)

#===================================
# Macro: MARMOT_GEN_CREATE_WRAPPER
#===================================
#This is a hook name, as the parser calls
#a macro of that name.
#
# Used to generate the feature tests.
MACRO (MARMOT_GEN_CREATE_WRAPPER outdir callname retval p_args_n_types)
    SET (args_n_types ${${p_args_n_types}})

    #Add to list of MPI2 calls
    SET (MARMOT_MPI2_CALLS "${MARMOT_MPI2_CALLS};${callname}" CACHE INTERNAL "" FORCE)

    #Text for the wrapper function
    SET (text "")
    
    #Header for FeatureTest
    SET (text "${text}#include <mpi.h>\n\nint main (int argc, char** argv)\n{\n  MPI_Request req;\n")
             
    #Call to function
    SET (text_t "${callname} (")
    
    IF (args_n_types)
        #For calls with arguments, loop over them and print
        SET (type)
        FOREACH (t_n_a ${args_n_types})
            IF (type)
                SET (text_t "${text_t}, ")
            ENDIF (type)

            MARMOT_GEN_SPLIT_T_N_A ("${t_n_a}" type arg)
                
            #print argument and type
            IF (${type} MATCHES "[*]")
                IF (${type} MATCHES "MPI_Request" AND 
                    NOT ${type} MATCHES "[*][^*]*[*]")
                   #For LAM MPI we need to put in an actual request pointer for 
                   #MPI_Request*
                   SET (text_t "${text_t}&req") #not NULL ?!
                ELSE (${type} MATCHES "MPI_Request" AND 
                    NOT ${type} MATCHES "[*][^*]*[*]")
                    SET (text_t "${text_t}0") #not NULL ?!
                ENDIF (${type} MATCHES "MPI_Request" AND 
                    NOT ${type} MATCHES "[*][^*]*[*]")
            ELSEIF (${type} MATCHES "int")
                SET (text_t "${text_t}0")
            ELSEIF (${type} MATCHES "MPI_Datatype")
                SET (text_t "${text_t}MPI_DATATYPE_NULL")
            ELSEIF (${type} MATCHES "MPI_Op")
                SET (text_t "${text_t}MPI_OP_NULL")
            ELSEIF (${type} MATCHES "MPI_Win")
                SET (text_t "${text_t}MPI_WIN_NULL")
            ELSEIF (${type} MATCHES "MPI_Info")
                SET (text_t "${text_t}MPI_INFO_NULL")
            ELSEIF (${type} MATCHES "MPI_Comm")
                SET (text_t "${text_t}MPI_COMM_NULL")
            ELSEIF (${type} MATCHES "MPI_Group")
                SET (text_t "${text_t}MPI_GROUP_NULL")
            ELSEIF (${type} MATCHES "MPI_File")
                SET (text_t "${text_t}MPI_FILE_NULL")
            ELSEIF (${type} MATCHES "MPI_Offset")
                SET (text_t "${text_t}0")
            ELSEIF (${type} MATCHES "MPI_Errhandler")
                SET (text_t "${text_t}MPI_ERRHANDLER_NULL")
            ELSEIF (${type} MATCHES "MPI_Request")
                SET (text_t "${text_t}MPI_REQUEST_NULL")
            ELSE (${type} MATCHES "[*]")
                #ERROR unknown type !
                MESSAGE (FATAL_ERROR "Unknown type found in feature test generation. Call=${callname}, type=${type}, arg=${arg}. Contact Marmot support !")            
            ENDIF (${type} MATCHES "[*]")                
        ENDFOREACH (t_n_a)
    ENDIF (args_n_types)
            
    #Trailer         
    SET (text_t "${text_t});\n")
    
    #2 versions of call
    SET (text "${text}  ${text_t}  P${text_t}")
    
    SET (text "${text}  return 0;\n}\n\n\n")
    
    #Write to output file
    HaveMPI_Call2FTName (${callname} filepath)
    FILE (WRITE "${filepath}" "${text}")         
       
ENDMACRO (MARMOT_GEN_CREATE_WRAPPER outdir callname retval p_args_n_types)

#===================================
# Actual script 
#===================================
MESSAGE (STATUS "Building MPI2 feature tests")

SET (INFILES ${INFILES} "${PROJECT_SOURCE_DIR}/SRC/INCLUDE/mpi2_1sided.h")
SET (INFILES ${INFILES} "${PROJECT_SOURCE_DIR}/SRC/INCLUDE/mpi2_extcoll.h")
SET (INFILES ${INFILES} "${PROJECT_SOURCE_DIR}/SRC/INCLUDE/mpi2_file.h")
SET (INFILES ${INFILES} "${PROJECT_SOURCE_DIR}/SRC/INCLUDE/mpi2_standard.h")
SET (INFILES ${INFILES} "${PROJECT_SOURCE_DIR}/SRC/INCLUDE/mpi2_proc.h")

SET (MARMOT_MPI2FT_DIR "${PROJECT_BINARY_DIR}/generated/ft_tests" CACHE INTERNAL "Directory that contains generated feature tests for MPI2.")

#Set internal variable to empty, hook function adds calls to it
SET (MARMOT_MPI2_CALLS "" CACHE INTERNAL "" FORCE)

FOREACH (INFILE ${INFILES})
    MARMOT_GEN_PARSE_HEADER ("${INFILE}" "${MARMOT_MPI2FT_DIR}" "")
ENDFOREACH (INFILE ${INFILES})
