# @file
#
# Provides helper macros for build control:
#
#  BC_ADJUST_LIB_VARS (basename)
#    - this is a support macro for cmake find modules (it simply sets/adjusts library-variables
#      depending on what was previously found for example by FIND_LIBRARY()
#
#  BC_RESET (varname)
#    - sets a variable to "not found"
#
#  BC_SET_FTPARAM (env_var_name env_var_value)
#    - set temporary environment variables for BC_TEST_FEATURE macro
#      (in case of libraries the debug versions will get filtered out)
#      Example: BC_SET_FTPARAM (MARMOT_FT_LIB "${MPICH_LIBRARIES}")
#
#  BC_TEST_FEATURE (feature_dest_var feature_test_name my_output)
#    - Compiles a small program given in "feature_test_name" and sets the variable "feature_dest_var"
#      if the compile/link process was successful
#      The full output from the compile/link process is returned in "my_output"
#      This macro expects the "feature-test"-files in CM_FEATURE_TESTS_DIR which is preset to
#      ${CMAKE_SOURCE_DIR}/CMAKEMODULES/FeatureTests
#
#      Example: BC_TEST_FEATURE (MARMOT_THREADED ft_marmot_threaded.c MY_OUTPUT)
#
#  BC_COPY_TARGET_PDB (target_name pdb_dest)
#    - gets the targets .pdb file and deploys it to the given location during install
#    - only the pdb files for debug versions of a library are installed
#      (this macro is windows specific)
#
#  BC_DUMP_LIB_SETUP (basename)
#    - dumps the different library-variable contents to a file
# 
#  BC_LIBLIST_TO_LINKCOMMANDLINE (liblist commlinelink)
#    - takes name of a library list variable as liblist
#    - return value is a string, stored in variable of commlinelink
#    - loops over the liblist and creates from each element 
#      a representation that can be used to link on the command line 
#
#  BC_SET_ON_VAR_CHANGE (var var_to_check1 var_to_check2 ...)
#    - sets the variable "var" to TRUE if one of the following variables changes it's value
#      between cmake-configure-runs otherwise to FALSE.
#
#  BC_LIST_REMOVE_BSTATIC_BDYNAMIC (list desc return)
#    Rational: "-Bstatic", "-Bdynamic" has to be removed from library
#              lists, otherwise unexpected behavior may appear when 
#              using MPI compiler wrappers.
#    - Loops over the list with name <list> and removes all entries
#      named "-Bstatic" or "-Bdynamic" from it
#    - The output is stored in the variable with name <return>
#    - If it detects one of the two strings it gives a warning
#      that the string was removed from the library list with 
#      description <desc>
#    - Uses a variable to override this replacement behavior 
#
# @author Blasius Czink, Tobias Hilbrich
#

# Macro to adjust library-variables
MACRO (BC_ADJUST_LIB_VARS basename)
    # if only the release version was found, set the debug variable also to the release version
    IF (${basename}_LIBRARY_RELEASE AND NOT ${basename}_LIBRARY_DEBUG)
      SET(${basename}_LIBRARY_DEBUG ${${basename}_LIBRARY_RELEASE} CACHE FILEPATH "${basename}_LIBRARY_DEBUG" FORCE)
      SET(${basename}_LIBRARY       ${${basename}_LIBRARY_RELEASE} CACHE FILEPATH "${basename}_LIBRARY" FORCE)
      SET(${basename}_LIBRARIES     ${${basename}_LIBRARY_RELEASE} CACHE FILEPATH "${basename}_LIBRARIES" FORCE)
    ENDIF (${basename}_LIBRARY_RELEASE AND NOT ${basename}_LIBRARY_DEBUG)

    # if only the debug version was found, set the release variable also to the debug version
    IF (${basename}_LIBRARY_DEBUG AND NOT ${basename}_LIBRARY_RELEASE)
      SET(${basename}_LIBRARY_RELEASE ${${basename}_LIBRARY_DEBUG} CACHE FILEPATH "${basename}_LIBRARY_RELEASE" FORCE)
      SET(${basename}_LIBRARY         ${${basename}_LIBRARY_DEBUG} CACHE FILEPATH "${basename}_LIBRARY" FORCE)
      SET(${basename}_LIBRARIES       ${${basename}_LIBRARY_DEBUG} CACHE FILEPATH "${basename}_LIBRARIES" FORCE)
    ENDIF (${basename}_LIBRARY_DEBUG AND NOT ${basename}_LIBRARY_RELEASE)
    
    # if both are set...
    IF (${basename}_LIBRARY_DEBUG AND ${basename}_LIBRARY_RELEASE)
      # if the generator supports configuration types then set
      # optimized and debug libraries, or if the CMAKE_BUILD_TYPE has a value
      IF (CMAKE_CONFIGURATION_TYPES OR CMAKE_BUILD_TYPE)
        SET(${basename}_LIBRARY
            optimized ${${basename}_LIBRARY_RELEASE}
            debug     ${${basename}_LIBRARY_DEBUG}
        )
      ELSE(CMAKE_CONFIGURATION_TYPES OR CMAKE_BUILD_TYPE)
        # if there are no configuration types and CMAKE_BUILD_TYPE has no value
        # then just use the release libraries
        SET(${basename}_LIBRARY ${${basename}_LIBRARY_RELEASE} )
      ENDIF(CMAKE_CONFIGURATION_TYPES OR CMAKE_BUILD_TYPE)

      SET(${basename}_LIBRARIES
          optimized ${${basename}_LIBRARY_RELEASE}
          debug ${${basename}_LIBRARY_DEBUG}
          CACHE FILEPATH "${basename}_LIBRARIES" FORCE
      )
      
    ENDIF (${basename}_LIBRARY_DEBUG AND ${basename}_LIBRARY_RELEASE)

    SET(${basename}_LIBRARY "${${basename}_LIBRARY}" CACHE FILEPATH "The ${basename} library" FORCE)
    # do not show those in normal view
    MARK_AS_ADVANCED(FORCE
      ${basename}_LIBRARY
      ${basename}_LIBRARY_RELEASE
      ${basename}_LIBRARY_DEBUG
      ${basename}_LIBRARIES
    )
ENDMACRO (BC_ADJUST_LIB_VARS)


# helper to reset variables to ...-NOTFOUND so that the FIND...-macros run again
MACRO (BC_RESET varname)
  IF (DEFINED ${varname})
    SET (${varname} "${varname}-NOTFOUND")
  ENDIF (DEFINED ${varname})
  IF (DEFINED ${varname}_DEBUG)
    SET (${varname}_DEBUG "${varname}_DEBUG-NOTFOUND")
  ENDIF (DEFINED ${varname}_DEBUG)
ENDMACRO (BC_RESET varname)

# helper to dump the lib-values to a simple text-file
MACRO (BC_DUMP_LIB_SETUP basename)
    SET (dump_file "${CMAKE_BINARY_DIR}/${basename}_lib_setup.txt")
    FILE(WRITE  ${dump_file} "${basename}_LIBRARY        is set to ${${basename}_LIBRARY}\n")
    FILE(APPEND ${dump_file} "${basename}_LIBRARY_RELESE is set to ${${basename}_LIBRARY_RELEASE}\n")
    FILE(APPEND ${dump_file} "${basename}_LIBRARY_DEBUG  is set to ${${basename}_LIBRARY_DEBUG}\n")
    FILE(APPEND ${dump_file} "${basename}_LIBRARIES      is set to ${${basename}_LIBRARIES}\n")
ENDMACRO (BC_DUMP_LIB_SETUP basename)


MACRO (BC_SET_FTPARAM env_var_name env_var_value)
  SET (FOUT "")

  #MESSAGE("orig mystring is ${env_var_value}")

  # get rid of debug libs RegExp ([Dd][Ee][Bb][Uu][Gg];[^;]*;?)
  STRING (REGEX REPLACE "[Dd][Ee][Bb][Uu][Gg];[^;]*;?" "" FOUT "${env_var_value}")
  #MESSAGE("1.mystring is ${FOUT}")

  # get rid of "optimized"
  STRING (REPLACE "optimized;" "" FOUT "${FOUT}")
  #MESSAGE("2.mystring is ${FOUT}")

  # add backslash before ";"
  STRING (REPLACE ";" "\\;" FOUT "${FOUT}")
  #MESSAGE("3.mystring is ${FOUT}")

  SET(ENV{${env_var_name}} "${FOUT}")
ENDMACRO (BC_SET_FTPARAM env_var_name env_var_value)


MACRO (BC_CLEAN_DEBUG_LIBS var_name var_value)
  SET (FOUT "")
  # get rid of debug libs RegExp ([Dd][Ee][Bb][Uu][Gg];[^;]*;?)
  STRING (REGEX REPLACE "[Dd][Ee][Bb][Uu][Gg];[^;]*;?" "" FOUT "${var_value}")
  # get rid of "optimized"
  STRING (REPLACE "optimized;" "" FOUT "${FOUT}")
  # replace backslash with space
  STRING (REPLACE ";" " " FOUT "${FOUT}")
  SET(${var_name} "${FOUT}")
ENDMACRO (BC_CLEAN_DEBUG_LIBS var_name var_value)


MACRO (BC_TEST_FEATURE feature_dest_var feature_test_name my_output)
  MESSAGE (STATUS "Checking for ${feature_test_name}")
  
  #Add MPI skip flags for cc (C++ files) too, should never harm I hope ...
  IF (${feature_test_name} MATCHES "[.]cc[ ]*$" OR 
      ${feature_test_name} MATCHES "[.]cpp[ ]*$")
      BC_LIST_TO_STRING (MARMOT_USED_SKIP_FLAGS string)
  ELSE (${feature_test_name} MATCHES "[.]cc[ ]*$")
      SET (string "")
  ENDIF (${feature_test_name} MATCHES "[.]cc[ ]*$" OR 
         ${feature_test_name} MATCHES "[.]cpp[ ]*$")
         
  IF (CM_FEATURE_TESTS_DIR STREQUAL "")
    SET(hstr_feature_test_name "${feature_test_name}")
  ELSE (CM_FEATURE_TESTS_DIR STREQUAL "")
    SET(hstr_feature_test_name "${CM_FEATURE_TESTS_DIR}/${feature_test_name}")
  ENDIF (CM_FEATURE_TESTS_DIR STREQUAL "")
    
  TRY_COMPILE (${feature_dest_var}
               ${CMAKE_BINARY_DIR}/CMakeTemp
               ${hstr_feature_test_name}
               CMAKE_FLAGS
               "-DCOMPILE_DEFINITIONS=${string}"
               -DINCLUDE_DIRECTORIES=$ENV{MARMOT_FT_INC}
               -DLINK_LIBRARIES=$ENV{MARMOT_FT_LIB}
               OUTPUT_VARIABLE ${my_output}
  )
  
  IF (NOT ${feature_dest_var})
     #Second try with second set of libraries
     IF ($ENV{MARMOT_FT_LIB2_WANTED} MATCHES "TRUE")
         TRY_COMPILE (${feature_dest_var}
                   ${CMAKE_BINARY_DIR}/CMakeTemp
                   ${hstr_feature_test_name}
                   CMAKE_FLAGS
                   "-DCOMPILE_DEFINITIONS=${string}"
                   -DINCLUDE_DIRECTORIES=$ENV{MARMOT_FT_INC}
                   -DLINK_LIBRARIES=$ENV{MARMOT_FT_LIB2}
                   OUTPUT_VARIABLE ${my_output}
                   )
     ENDIF ($ENV{MARMOT_FT_LIB2_WANTED} MATCHES "TRUE")
  ENDIF (NOT ${feature_dest_var})               
  
  # Feature test failed
  IF (${feature_dest_var})
     MESSAGE (STATUS "Checking for ${feature_test_name} - succeeded")
  ELSE (${feature_dest_var})
     
     MESSAGE (STATUS "Checking for ${feature_test_name} - feature not available")
  ENDIF (${feature_dest_var})
ENDMACRO (BC_TEST_FEATURE feature_dest_var feature_test_name my_output)


MACRO (BC_COPY_TARGET_PDB target_name)
  IF(WIN32 AND MSVC)
    GET_TARGET_PROPERTY(TARGET_LOC ${target_name} DEBUG_LOCATION)
    IF (TARGET_LOC)
      GET_FILENAME_COMPONENT(TARGET_HEAD "${TARGET_LOC}" NAME_WE)
      # GET_FILENAME_COMPONENT(FNABSOLUTE  "${TARGET_LOC}" ABSOLUTE)
      GET_FILENAME_COMPONENT(TARGET_PATH "${TARGET_LOC}" PATH)
      SET(TARGET_PDB "${TARGET_PATH}/${TARGET_HEAD}.pdb")
      message(STATUS "PDB-File of ${target_name} is ${TARGET_PDB}")
      GET_TARGET_PROPERTY(TARGET_TYPE ${target_name} TYPE)
      IF (TARGET_TYPE)
        IF (${TARGET_TYPE} STREQUAL "STATIC_LIBRARY")
          SET (pdb_dest lib)
        ELSE (${TARGET_TYPE} STREQUAL "STATIC_LIBRARY")
          SET (pdb_dest bin)
        ENDIF (${TARGET_TYPE} STREQUAL "STATIC_LIBRARY")
        INSTALL(FILES ${TARGET_PDB} DESTINATION ${pdb_dest} CONFIGURATIONS "Debug") 
      ENDIF(TARGET_TYPE)
    ENDIF (TARGET_LOC)
  ENDIF(WIN32 AND MSVC)
ENDMACRO (BC_COPY_TARGET_PDB target_name)


MACRO (BC_INVERT_BOOL var)
  IF (${var})
    SET (${var} FALSE)
  ELSE (${var})
    SET (${var} TRUE)
  ENDIF (${var})
ENDMACRO (BC_INVERT_BOOL)


MACRO (BC_LIST_CONTAINS var value)
  SET (${var})
  FOREACH (value2 ${ARGN})
    STRING (TOUPPER ${value}  str1)
    STRING (TOUPPER ${value2} str2)
    IF (${str1} STREQUAL ${str2})
      SET (${var} TRUE)
    ENDIF (${str1} STREQUAL ${str2})
  ENDFOREACH (value2)
ENDMACRO (BC_LIST_CONTAINS)


MACRO (BC_LIST_CONTAINS_CS var value)
  SET (${var})
  FOREACH (value2 ${ARGN})
    IF (${value} STREQUAL ${value2})
      SET (${var} TRUE)
    ENDIF (${value} STREQUAL ${value2})
  ENDFOREACH (value2)
ENDMACRO (BC_LIST_CONTAINS_CS)


MACRO (BC_ADD_GLOBAL_LFLAG given_flag)
  foreach(flag_var CMAKE_EXE_LINKER_FLAGS CMAKE_SHARED_LINKER_FLAGS CMAKE_MODULE_LINKER_FLAGS)
    if (NOT ${flag_var} MATCHES "${given_flag}")
      set(${flag_var} "${${flag_var}} ${given_flag}" CACHE STRING "Linker flags" FORCE)
    endif(NOT ${flag_var} MATCHES "${given_flag}")
  endforeach(flag_var)
ENDMACRO (BC_ADD_GLOBAL_LFLAG given_flag)

MACRO (BC_LIBLIST_TO_LINKCOMMANDLINE liblist commlinelink)
    #Set initial output ( empty )
    SET (${commlinelink} "")
        
    #Loop over each lib
    FOREACH (lib ${${liblist}})
        
        IF (NOT ${commlinelink} STREQUAL "")
            SET (${commlinelink} "${${commlinelink}} ")
        ENDIF (NOT ${commlinelink} STREQUAL "")
    
        ##TODO: Likely has to be adapted for different OS (Win, MAC, ...)
        ##      But currently only used for linux/AIX based systems anyways
        IF (lib MATCHES "^[/]")
            #It is an absoulte path with full information -> everything is fine
            SET (${commlinelink} "${${commlinelink}}${lib}")
        ELSE (lib MATCHES "^[/]")
            IF (lib MATCHES "^[-]")
                #It is the common linux "-lX" format to load libX -> everything is fine
                SET (${commlinelink} "${${commlinelink}}${lib}")
            ELSE (lib MATCHES "^[-]")
                #It is the shortened CMake form for a lib
                SET (${commlinelink} "${${commlinelink}}-l${lib}")
            ENDIF (lib MATCHES "^[-]")
        ENDIF (lib MATCHES "^[/]")
    ENDFOREACH (lib)
ENDMACRO (BC_LIBLIST_TO_LINKCOMMANDLINE liblist commlinelink)

MACRO (BC_LIST_TO_STRING list string)
    #Set initial output ( empty )
    SET (${string} "")
        
    #Loop over each lib
    FOREACH (elem ${${list}})
        SET (${string} "${${string}} ${elem}")
    ENDFOREACH (elem)
ENDMACRO (BC_LIST_TO_STRING)

MACRO (BC_SET_ON_VAR_CHANGE var)
  SET(${var} FALSE)
  FOREACH (arg ${ARGN})
    IF (NOT DEFINED ${arg}_OLD)
      SET(${arg}_OLD "" CACHE INTERNAL "")
      SET (${var} TRUE)
    ELSE (NOT DEFINED ${arg}_OLD)
      IF (NOT "${${arg}}" STREQUAL "${${arg}_OLD}")
        SET (${var} TRUE)
	SET (${arg}_OLD "${${arg}}" CACHE INTERNAL "")
      ENDIF (NOT "${${arg}}" STREQUAL "${${arg}_OLD}")
    ENDIF (NOT DEFINED ${arg}_OLD)
  ENDFOREACH (arg)
ENDMACRO (BC_SET_ON_VAR_CHANGE var)

MACRO (BC_LIST_REMOVE_BSTATIC_BDYNAMIC list desc return)
  SET (${return} "")
  FOREACH (l ${${list}})
    IF (l STREQUAL "-Bstatic")
      IF (NOT MARMOT_NO_BSTATIC_BDYNAMIC_REMOVE)
        MESSAGE (STATUS "Warning: ${desc} contains a \"-Bstatic\", which was removed to avoid problems with MPI compiler wrappers, if you want to disable this behavior clear the cache rerun cmake with -DMARMOT_NO_BSTATIC_BDYNAMIC_REMOVE.")
      ELSE (NOT MARMOT_NO_BSTATIC_BDYNAMIC_REMOVE)
        SET (${return} ${${return}} ${l})
      ENDIF (NOT MARMOT_NO_BSTATIC_BDYNAMIC_REMOVE)
    ELSEIF(l STREQUAL "-Bdynamic")
      IF (NOT MARMOT_NO_BSTATIC_BDYNAMIC_REMOVE)
        MESSAGE (STATUS "Warning: ${desc} contains a \"-Bdynamic\", which was removed to avoid problems with MPI compiler wrappers, if you want to disable this behavior clear the cache and rerun cmake with -DMARMOT_NO_BSTATIC_BDYNAMIC_REMOVE.")
      ELSE (NOT MARMOT_NO_BSTATIC_BDYNAMIC_REMOVE)
        SET (${return} ${${return}} ${l})
      ENDIF (NOT MARMOT_NO_BSTATIC_BDYNAMIC_REMOVE)
    ELSE (l STREQUAL "-Bstatic")
      SET (${return} ${${return}} ${l})
    ENDIF (l STREQUAL "-Bstatic")     
  ENDFOREACH (l)
ENDMACRO (BC_LIST_REMOVE_BSTATIC_BDYNAMIC list desc return)

MACRO (BC_PLATFORM_CORRECT_LIB libvar)
    ## Compiler/System specific stuff 
    #BG/P -> enable linking with shared libs
    IF (CMAKE_C_COMPILER MATCHES "bgxlc" OR 
        CMAKE_C_COMPILER MATCHES "bgxlc_r")
        STRING (REGEX REPLACE "[.]so[ \t]*$" ".a" ${libvar} "${${libvar}}")
    ENDIF (CMAKE_C_COMPILER MATCHES "bgxlc" OR 
           CMAKE_C_COMPILER MATCHES "bgxlc_r")
ENDMACRO (BC_PLATFORM_CORRECT_LIB)

# ----------------
#  Unused stuff
# ----------------

# MACRO(BC_FILTER_OUT FILTERS INPUTS OUTPUT)
#   # Arguments:
#   #  FILTERS - list of patterns that need to be removed
#   #  INPUTS  - list of inputs that will be worked on
#   #  OUTPUT  - the filtered list to be returned
#   # 
#   # Example: 
#   #  SET(MYLIST this that and the other)
#   #  SET(FILTS this that)
#   #
#   #  FILTER_OUT("${FILTS}" "${MYLIST}" OUT)
#   #  MESSAGE("OUTPUT = ${OUT}")
#   #
#   # The output - 
#   #   OUTPUT = and;the;other
#   #
#   SET(FOUT "")
#   FOREACH(INP ${INPUTS})
#      SET(FILTERED 0)
#      FOREACH(FILT ${FILTERS})
#          IF(${FILTERED} EQUAL 0)
#              IF("${FILT}" STREQUAL "${INP}")
#                  SET(FILTERED 1)
#              ENDIF("${FILT}" STREQUAL "${INP}")
#          ENDIF(${FILTERED} EQUAL 0)
#      ENDFOREACH(FILT ${FILTERS})
#      IF(${FILTERED} EQUAL 0)
#          SET(FOUT ${FOUT} ${INP})
#      ENDIF(${FILTERED} EQUAL 0)
#   ENDFOREACH(INP ${INPUTS})
#   SET(${OUTPUT} ${FOUT})
# ENDMACRO(BC_FILTER_OUT FILTERS INPUTS OUTPUT)

# MACRO (CHECK_LIBRARY LIB_NAME LIB_DESC LIB_TEST_SOURCE)
#   SET (HAVE_${LIB_NAME})
#   MESSAGE (STATUS "Checking for ${LIB_DESC} library...")
#   TRY_COMPILE (HAVE_${LIB_NAME} ${CMAKE_BINARY_DIR}/.cmake_temp ${CMAKE_SOURCE_DIR}/cmake/${LIB_TEST_SOURCE})
#   IF (HAVE_${LIB_NAME})
#      # Don't need one
#      MESSAGE(STATUS "Checking for ${LIB_DESC} library... none needed")
#   ELSE (HAVE_${LIB_NAME})
#      # Try to find a suitable library
#      FOREACH (lib ${ARGN})
#          TRY_COMPILE (HAVE_${LIB_NAME} ${CMAKE_BINARY_DIR}/.cmake_temp
#                      ${CMAKE_SOURCE_DIR}/cmake/${LIB_TEST_SOURCE}
#                      CMAKE_FLAGS -DLINK_LIBRARIES:STRING=${lib})
#          IF (HAVE_${LIB_NAME})
#            MESSAGE (STATUS "Checking for ${LIB_DESC} library... ${lib}")
#            SET (HAVE_${LIB_NAME}_LIB ${lib})
#          ENDIF (HAVE_${LIB_NAME})
#      ENDFOREACH (lib)
#   ENDIF (HAVE_${LIB_NAME})
#   # Unable to find a suitable library
#   IF (NOT HAVE_${LIB_NAME})
#      MESSAGE (STATUS "Checking for ${LIB_DESC} library... not found")
#   ENDIF (NOT HAVE_${LIB_NAME})
# ENDMACRO (CHECK_LIBRARY)
# 
# ...the LIB_NAME param controls the name of the variable(s) that are set 
# if it finds the library. LIB_DESC is used for the status messages. An 
# example usage might be:
# 
# CHECK_LIBRARY(SCK socket scktest.c socket xnet ws2_32)
