# @file GenerateCWrapper.cmake
# 
# Generats the C wrappers for Marmot. 
#
# Based on mkwrapper.sh.in (r921), which had the following authors:
#    * Katrin Bidmon
#    * Bettina Krammer
#    * Matthias Mueller
#    * Tobias Hilbrich
#
# Uses the following other CMake variables that have to be defined before
# calling this script:
#    * MPI_C_HEADER   --- full path to the MPI C header
#
# @author Tobias Hilbrich
#

#===================================
# Macro: MARMOT_GEN_CREATE_WRAPPER
#===================================
#This is a hook name, as the parser calls
#a macro of that name.
MACRO (MARMOT_GEN_CREATE_WRAPPER outfile callname retval p_args_n_types)
    SET (args_n_types ${${p_args_n_types}})

    #Text for the wrapper function
    SET (text "")
    
    #Is it implemented in Marmot ?
    MARMOT_GEN_IS_IMPLEMENTED ("${callname}" implemented)
    
    #This loop creates two different versions of the calls one with aditional 
    #parameters for line and file name and one without
    FOREACH(e "reg" "sourceinfo") 
        # Start with function name of wrapper
        IF (e MATCHES "reg")
            SET (text "${text}\nextern \"C\" ${retval} MPIAPI ${callname} (")
        ELSE (e MATCHES "reg")
            STRING (REGEX REPLACE "MPI_" "MPI_SOURCEINFO_" sourceinfo_name ${callname})
            SET (text "${text}\nextern \"C\" ${retval} ${sourceinfo_name} (")
        ENDIF (e MATCHES "reg")
          
        SET (mpocall ${callname})
        
        IF (args_n_types)
            #For calls with arguments, loop over them and print
            SET (type)
            FOREACH (t_n_a ${args_n_types})
                IF (type)
                  SET (text "${text}, ")
                ENDIF (type)

                MARMOT_GEN_SPLIT_T_N_A ("${t_n_a}" type arg)
                
                #print argument and type
                SET (text "${text}${type} ${arg}")                
            ENDFOREACH (t_n_a)
        ENDIF (args_n_types)
            
        IF (e MATCHES "sourceinfo")
             IF (type)
                 SET (text "${text},int line,const char *scfile")
             ELSE (type)
                 SET (text "${text}int line,const char *scfile")
             ENDIF (type)
        ENDIF (e MATCHES "sourceinfo")     
                    
        #Finish the function line                
        SET (text "${text})\n")
        
        #we are done with the prototype, now we start with the real function:
        SET (text "${text}{\n")
        
        STRING (REPLACE "MPI_" "MPO_" mpocall ${mpocall})
        
        # we are done with the prototype, now we start with the real function:
        SET (text "${text}  ${retval} rvalue=0;\n")

        #We use this one in order to get the mpo id and callname of an mpo object.
        #Otherwise we would have to create the object for this call before the
        #enterMARMOT call which causes problems with non threadsafe implementations
        #of memory allocation.
        IF (implemented)
            SET (text "${text}   static ${mpocall} temp;\n")
        ELSE (implemented)
            #create some temp object, not of much interest which one
            SET (text "${text}   static MPO_Comm_rank temp;\n")
        ENDIF (implemented)
  
        # now test whether we need registration of requests
        FOREACH (t_n_a ${args_n_types})
            MARMOT_GEN_SPLIT_T_N_A ("${t_n_a}" type arg)
            PRINT_REQUEST_REGISTER (text "${type}" "${arg}" "${callname}" "${e}" "NO")  
        ENDFOREACH (t_n_a)
        
        # now test whether we need registration of collective comms
        FOREACH (t_n_a ${args_n_types})
            MARMOT_GEN_SPLIT_T_N_A ("${t_n_a}" type arg)
            PRINT_COLLCOMM_REGISTER (text "${type}" "${arg}" "${callname}" "${e}" "NO")  
        ENDFOREACH (t_n_a)
        
        #synchronisation    
        SET (text "${text}#ifdef MARMOT_USE_THREADSYNC\n")
        IF (e MATCHES "sourceinfo")
            SET (text "${text}      bool enter_succeeded = MPO_Space::getSyncObject()->enterMARMOT(line,scfile);\n")
        ELSE (e MATCHES "sourceinfo") 
            SET (text "${text}      bool enter_succeeded = MPO_Space::getSyncObject()->enterMARMOT(0,\"\");\n")
        ENDIF (e MATCHES "sourceinfo")
        SET (text "${text}#endif\n")
        
        #print the creation of the MPO Object and set the source location
        #Extra for MPI_Init that calls PMPI during object creation !
        IF ("${callname}" MATCHES "MPI_Init[^A-Za-z0-9_]*$" OR
            "${callname}" MATCHES "MPI_Init_thread[^A-Za-z0-9_]*$")
      	    SET (text "${text}MPO_in_profile=1;\n")
        ENDIF ("${callname}" MATCHES "MPI_Init[^A-Za-z0-9_]*$" OR
               "${callname}" MATCHES "MPI_Init_thread[^A-Za-z0-9_]*$")
  
        IF (implemented)
            SET (text "${text}   ${mpocall} my_obj")
            
            IF (args_n_types)
              SET (text "${text}(")
              SET (type)
              FOREACH (t_n_a ${args_n_types})
                IF (type)
                    SET (text "${text}, ")
                ENDIF (type)
                MARMOT_GEN_SPLIT_T_N_A ("${t_n_a}" type arg)
                
                #Remove any existing array brackets (group_range_incl)
                STRING (REGEX REPLACE "\\[.*\\]" "" conv_name "${arg}")
                SET (text "${text}${conv_name}")
              ENDFOREACH (t_n_a)
              SET (text "${text})")
            ENDIF (args_n_types)
            SET (text "${text};\n")
             
            #this if is necessary for MPI implems like MPICH that perform MPI calls during MPI_Init
            IF ("${callname}" MATCHES "MPI_Init[^A-Za-z0-9_]*$" OR
                "${callname}" MATCHES "MPI_Init_thread[^A-Za-z0-9_]*$")
          	    SET (text "${text}MPO_in_profile=0;\n")
            ENDIF ("${callname}" MATCHES "MPI_Init[^A-Za-z0-9_]*$" OR
                   "${callname}" MATCHES "MPI_Init_thread[^A-Za-z0-9_]*$")
            
            IF (e MATCHES "sourceinfo")
                SET (text "${text}   my_obj.setSourceLocation(line,scfile);\n")
            ENDIF (e MATCHES "sourceinfo")

            SET (text "${text}#ifdef MARMOT_USE_THREADSYNC\n")
            SET (text "${text}   if(enter_succeeded ){\n")
            SET (text "${text}#else\n")
            SET (text "${text}   if(! MPO_in_profile ){\n")
            SET (text "${text}#endif\n")
            SET (text "${text}      MPO_in_profile=1;\n")
            SET (text "${text}      rvalue = my_obj.callDebugServer();\n")
            SET (text "${text}      MPO_in_profile=0;\n")
            SET (text "${text}   }\n")
            SET (text "${text}   else{\n")
            SET (text "${text}      rvalue = P${callname}(")
            SET (type)
            FOREACH (t_n_a ${args_n_types})
                IF (type)
                    SET (text "${text}, ")
                ENDIF (type)
                MARMOT_GEN_SPLIT_T_N_A ("${t_n_a}" type arg)
                
                #Remove any existing array brackets (group_range_incl)
                STRING (REGEX REPLACE "\\[.*\\]" "" conv_name "${arg}")
                                
                #Is it a non-implemented call with a communicator ?
                #Replace MPI_COMM_WORLD with our_comm_world in that case
                IF (${type} MATCHES "MPI_Comm([^_]+|$)" AND
                    (${arg} MATCHES "_CLASS_SINGLE_IN"  OR 
                     NOT ${type} MATCHES "[*]") ) 
                    SET (conv_name "MPO_Comm(${conv_name})")
                ENDIF (${type} MATCHES "MPI_Comm([^_]+|$)" AND
                       (${arg} MATCHES "_CLASS_SINGLE_IN"  OR 
                        NOT ${type} MATCHES "[*]") )
                
                SET (text "${text}${conv_name}")
            ENDFOREACH (t_n_a)
            SET (text "${text});\n")
            SET (text "${text}   }\n")
        ELSE (implemented)
            SET (text "${text}#ifdef MARMOT_USE_THREADSYNC\n")
            SET (text "${text}  if(enter_succeeded )\n")
            SET (text "${text}#else\n")
            SET (text "${text}  if(! MPO_in_profile )\n")
            SET (text "${text}#endif\n")
            SET (text "${text}  {\n")
            SET (text "${text}      static bool not_implemented_warned = false;\n")
            SET (text "${text}      if (!not_implemented_warned)\n")
            SET (text "${text}      {\n")
            SET (text "${text}          std::cerr << \"WARNING: Application issued an MPI call not implemented by Marmot, false positives and non-reported errors are possible. (Call: ${callname})\" << std::endl;\n")
            SET (text "${text}          not_implemented_warned = true;\n")
            SET (text "${text}      }\n")
            SET (text "${text}  }\n")
            SET (text "${text}  rvalue = P${callname}(")
            SET (type)
            FOREACH (t_n_a ${args_n_types})
                IF (type)
                    SET (text "${text}, ")
                ENDIF (type)
                MARMOT_GEN_SPLIT_T_N_A ("${t_n_a}" type arg)
                
                #Remove any existing array brackets (group_range_incl)
                STRING (REGEX REPLACE "\\[.*\\]" "" conv_name "${arg}")
                                
                #Is it a non-implemented call with a communicator ?
                #Replace MPI_COMM_WORLD with our_comm_world in that case
                IF (${type} MATCHES "MPI_Comm([^_]+|$)" AND
                    (${arg} MATCHES "_CLASS_SINGLE_IN"  OR 
                     NOT ${type} MATCHES "[*]") ) 
                    SET (conv_name "MPO_Comm(${conv_name})")
                ENDIF (${type} MATCHES "MPI_Comm([^_]+|$)" AND
                       (${arg} MATCHES "_CLASS_SINGLE_IN"  OR 
                        NOT ${type} MATCHES "[*]") )
                        
                SET (text "${text}${conv_name}")
            ENDFOREACH (t_n_a)
            SET (text "${text});\n")            
        ENDIF (implemented)
      
        SET (text "${text}#ifdef MARMOT_USE_THREADSYNC\n")
        IF (e MATCHES "sourceinfo")
            SET (text "${text}      MPO_Space::getSyncObject()->leaveMARMOT(line,scfile,enter_succeeded);\n")
        ELSE (e MATCHES "sourceinfo")
            SET (text "${text}      MPO_Space::getSyncObject()->leaveMARMOT(0,\"\",enter_succeeded);\n")
        ENDIF (e MATCHES "sourceinfo")
        SET (text "${text}#endif\n")

        # now test whether we need registration of collective comms
        FOREACH (t_n_a ${args_n_types})
            MARMOT_GEN_SPLIT_T_N_A ("${t_n_a}" type arg)
            PRINT_COLLCOMM_UNREGISTER (text "${type}" "${arg}" "${callname}" "${e}" "NO")  
        ENDFOREACH (t_n_a)
  
        # now test whether we have to unregister requests
        FOREACH (t_n_a ${args_n_types})
            MARMOT_GEN_SPLIT_T_N_A ("${t_n_a}" type arg)
            PRINT_REQUEST_UNREGISTER (text "${type}" "${arg}" "${callname}" "${e}" "NO")  
        ENDFOREACH (t_n_a)
    
        SET (text "${text}   return rvalue;\n")
        SET (text "${text}}\n\n\n")
        
    ENDFOREACH (e)
    
    FILE (APPEND ${outfile} "${text}") #|${callname}|" ${args_n_types} "\n"
        
ENDMACRO (MARMOT_GEN_CREATE_WRAPPER outfile callname retval p_args_n_types)

#===================================
# Actual script 
#===================================
MESSAGE (STATUS "Building MPI wrappers for C bindings")

SET (INFILES "${PROJECT_SOURCE_DIR}/SRC/INCLUDE/mpi_standard.h")
SET (INFILES ${INFILES} "${PROJECT_SOURCE_DIR}/SRC/INCLUDE/mpi2_1sided.h")
SET (INFILES ${INFILES} "${PROJECT_SOURCE_DIR}/SRC/INCLUDE/mpi2_extcoll.h")
SET (INFILES ${INFILES} "${PROJECT_SOURCE_DIR}/SRC/INCLUDE/mpi2_file.h")
SET (INFILES ${INFILES} "${PROJECT_SOURCE_DIR}/SRC/INCLUDE/mpi2_standard.h")
SET (INFILES ${INFILES} "${PROJECT_SOURCE_DIR}/SRC/INCLUDE/mpi2_proc.h")

SET (OUTFILE "${PROJECT_BINARY_DIR}/generated/c_wrapper.cc")
CONFIGURE_FILE("${PROJECT_SOURCE_DIR}/CMAKEMODULES/generation/CHeader.cc.in" "${OUTFILE}" @ONLY)

SET (IGNOREFILE "${PROJECT_BINARY_DIR}/generated/ignore-list.txt")
CONFIGURE_FILE("${PROJECT_SOURCE_DIR}/CMAKEMODULES/generation/ignore-list.txt" "${IGNOREFILE}" COPYONLY)
HaveMPI_GetUnsupportedCallList(unsuported_list)
FOREACH (call ${unsuported_list})
    FILE (APPEND "${IGNOREFILE}" "\n${call}")
ENDFOREACH (call)

FOREACH (INFILE ${INFILES})
    MARMOT_GEN_PARSE_HEADER ("${INFILE}" "${OUTFILE}" "${IGNOREFILE}")
ENDFOREACH (INFILE ${INFILES})
