# @file GenerateFortranWrapper.cmake
# 
# Generats the Fortran wrappers for Marmot. 
#
# Based on mk_fortran_wrapper.sh.in (r972), which had the following authors:
#    * Rainer Keller
#    * Bettina Krammer
#    * Matthias Mueller
#    * Tobias Hilbrich
#
# Uses the following other CMake variables that have to be defined before
# calling this script:
#    * HANDLE_CONVERT \in {YES, NO}
#      -> set with value from MPI_HANDLE_CONVERT_PRESENT
#    * REQUEST_CONVERT \in {YES, NO}
#      -> set with value from MPI_REQUEST_CONVERT_PRESENT
#    * STATUS_CONVERT \in {YES, NO}
#      -> set with value from MPI_STATUS_CONVERT_PRESENT
#    * MPI_C_HEADER   --- full path to the MPI C header
#
# @author Tobias Hilbrich
#

#===================================
# Macro: FORTRAN_INIT_PRE
#===================================
MACRO (FORTRAN_INIT_PRE p_text)
  SET (temp
      "  int argc=1;\n"
      "#ifdef MARMOT_MPICH_P4\n"
      " /*\n"
      "  ** this part of the code is taken from mpich-1.2.4/src/fortran/initf.c\n"
      "  ** the goal is to recover the calling arguments for MPICH with the p4 device,\n"
      "  ** since mpich needs it for a proper initialization\n"
      "  */\n"
      "  int i;\n"
      "  char *p;\n"
      "  int  ArgcSave;           /* Save the argument count */\n"
      "  char **ArgvSave,         /* Save the pointer to the argument vector */\n"
      "    **ArgvValSave;      /* Save the ENTRIES in the argument vector */\n"
      "  int argsize=1024;\n"
      "  char **argv;\n"
      "\n"
      "  /* Recover the args with the Fortran routines iargc_ and getarg_ */\n"
      "  ArgcSave    = argc = MPIR_IARGC() + 1;\n"
      "  ArgvSave    = argv = new char* [argc];\n"
      "  ArgvValSave = new char* [argc];\n"
      "  if (argv == NULL || ArgvValSave == NULL)\n"
      "    {\n"
      "      fprintf (stderr, \"MPIf_Init: couldn't allocate memory\");\n"
      "      exit (1);\n"
      "    }\n"
      "\n"
      "  for (i=0; i < argc; i++)\n"
      "    {\n"
      "      ArgvValSave[i] = argv[i] = new char [argsize + 1];\n"
      "      if (argv[i] == NULL)\n"
      "        {\n"
      "          fprintf (stderr, \"MPIf_Init: couldn't allocate memory\");\n"
      "          exit (1);\n"
      "        }\n"
      "\n"
      "      MPIR_GETARG (&i, argv[i], argsize);\n"
      "      /* Trim trailing blanks */\n"
      "      p = argv[i] + argsize - 1;\n"
      "      while (p >= argv[i]) {\n"
      "        if (*p != ' ' && *p) {\n"
      "          p[1] = '\\0';\n"
      "          break;\n"
      "        }\n"
      "        p--;\n"
      "      }\n"
      "    }\n"
      "#else\n"
      "  /*\n"
      "  ** We have to simulate a C-startup of MPI, so we have to engage\n"
      "  ** the variables argc and argv .\n"
      "  */\n"
      "\n"
      "  char *argx[] = {new char[256], new char[256], NULL};\n"
      "  sprintf (argx[0],\"Fortran_Init\");\n"
      "  sprintf (argx[1],\"\");\n"
      "  char ** argv = argx;\n"
      "#endif\n"
      "\n"
      "  static char interface_mode[64];\n"
      "  sprintf (interface_mode, \"MARMOT_INTERFACE_MODE=1\");\n"
      "  putenv (interface_mode);\n"
      )
    
    FOREACH (t ${temp})
        IF ("${t}" MATCHES "^[ ]*\n")
            SET (${p_text} "${${p_text}};${t}")
        ELSE ("${t}" MATCHES "^[ ]*\n")
            SET (${p_text} "${${p_text}}${t}")
        ENDIF ("${t}" MATCHES "^[ ]*\n")                        
    ENDFOREACH (t)
ENDMACRO (FORTRAN_INIT_PRE p_text)
  
#===================================
# Macro: FORTRAN_INIT_POST
#===================================
MACRO (FORTRAN_INIT_POST p_text)  
    SET (temp
          "#ifdef MARMOT_MPICH_P4\n"
          "  /* Recover space */\n"
          "  for (i=0; i < ArgcSave; i++)\n"
          "    free (ArgvValSave[i]);\n"
          "\n"
          "  free (ArgvSave);\n"
          "  free (ArgvValSave);\n"
          "#else\n"
          "  delete [] argx[0];\n"
          "  delete [] argx[1];\n"
          "#endif\n"
          )
    
    FOREACH (t ${temp})
        IF ("${t}" MATCHES "^[ ]*\n")
            SET (${p_text} "${${p_text}};${t}")
        ELSE ("${t}" MATCHES "^[ ]*\n")
            SET (${p_text} "${${p_text}}${t}")
        ENDIF ("${t}" MATCHES "^[ ]*\n")                        
    ENDFOREACH (t)
ENDMACRO (FORTRAN_INIT_POST p_text)

#===================================
# Macro: CONVERTTYPE
#===================================
MACRO (CONVERTTYPE type out callname)
    IF (HANDLE_CONVERT MATCHES "NO")
        SET (${out} "${type}")
    ELSE (HANDLE_CONVERT MATCHES "NO")
        SET (ret "${type}")
        
        IF (ret MATCHES "MPI_Comm([^_]+|$)")
            STRING (REPLACE "MPI_Comm" "MPI_Fint" ret "${ret}")
        ENDIF (ret MATCHES "MPI_Comm([^_]+|$)")
        
        IF (ret MATCHES "MPI_Datatype([^_]+|$)")
            STRING (REPLACE "MPI_Datatype" "MPI_Fint" ret "${ret}")
        ENDIF (ret MATCHES "MPI_Datatype([^_]+|$)")
            
        IF (ret MATCHES "MPI_Group([^_]+|$)")
            STRING (REPLACE "MPI_Group" "MPI_Fint" ret "${ret}")
        ENDIF (ret MATCHES "MPI_Group([^_]+|$)")
            
        IF (ret MATCHES "MPI_Op([^_]+|$)")
            STRING (REPLACE "MPI_Op" "MPI_Fint" ret "${ret}")
        ENDIF (ret MATCHES "MPI_Op([^_]+|$)")
   
        IF (REQUEST_CONVERT MATCHES "YES" AND
            ret MATCHES "MPI_Request([^_]+|$)")
            STRING (REPLACE "MPI_Request" "MPI_Fint" ret "${ret}")
        ENDIF (REQUEST_CONVERT MATCHES "YES" AND
            ret MATCHES "MPI_Request([^_]+|$)")
        
        IF (STATUS_CONVERT MATCHES "YES" AND
            ret MATCHES "MPI_Status([^_]+|$)")
            STRING (REPLACE "MPI_Status" "MPI_Fint" ret "${ret}")
        ENDIF (STATUS_CONVERT MATCHES "YES" AND
            ret MATCHES "MPI_Status([^_]+|$)")
        
        IF (INFO_CONVERT MATCHES "YES" AND
            ret MATCHES "MPI_Info([^_]+|$)")
            STRING (REPLACE "MPI_Info" "MPI_Fint" ret "${ret}")
        ENDIF (INFO_CONVERT MATCHES "YES" AND
            ret MATCHES "MPI_Info([^_]+|$)")
        
        IF (WIN_CONVERT MATCHES "YES" AND
            ret MATCHES "MPI_Win([^_]+|$)")
            STRING (REPLACE "MPI_Win" "MPI_Fint" ret "${ret}")
        ENDIF (WIN_CONVERT MATCHES "YES" AND
            ret MATCHES "MPI_Win([^_]+|$)")
        
        IF (FILE_CONVERT MATCHES "YES" AND
            ret MATCHES "MPI_File([^_]+|$)")
            STRING (REPLACE "MPI_File" "MPI_Fint" ret "${ret}")
        ENDIF (FILE_CONVERT MATCHES "YES" AND
            ret MATCHES "MPI_File([^_]+|$)")
        
        SET (${out} "${ret}")
    ENDIF (HANDLE_CONVERT MATCHES "NO")
    
    
    #Special case for a very few calls where MPI_Aint is represented with 
    #MPI_Fint in Fortran (32 vs 64 bit)
    IF ("${callname}" MATCHES "MPI_Type_hindexed" OR
        "${callname}" MATCHES "MPI_Type_hvector" OR
        "${callname}" MATCHES "MPI_Type_struct" OR
        "${callname}" MATCHES "MPI_Address" OR 
        "${callname}" MATCHES "MPI_Type_extent" OR
        "${callname}" MATCHES "MPI_Type_lb" OR
        "${callname}" MATCHES "MPI_Type_ub")
        IF (type MATCHES "MPI_Aint[*]")
            SET (${out} "MPI_Fint*")
        ELSEIF (type MATCHES "MPI_Aint")
            SET (${out} "MPI_Fint")
        ENDIF (type MATCHES "MPI_Aint[*]")
    ENDIF ("${callname}" MATCHES "MPI_Type_hindexed" OR
        "${callname}" MATCHES "MPI_Type_hvector" OR
        "${callname}" MATCHES "MPI_Type_struct" OR
        "${callname}" MATCHES "MPI_Address" OR 
        "${callname}" MATCHES "MPI_Type_extent" OR
        "${callname}" MATCHES "MPI_Type_lb" OR
        "${callname}" MATCHES "MPI_Type_ub")
ENDMACRO (CONVERTTYPE type out callname)

#===================================
# Macro: PRINT_CONVERT_FUNCTION
#===================================
MACRO (PRINT_CONVERT_FUNCTION p_text type arg)
    IF (HANDLE_CONVERT MATCHES "YES")
        # a SINGLE value for input purposes
        IF ( NOT ${type} MATCHES "[*]" OR
             ${arg} MATCHES "_CLASS_SINGLE_IN" OR
             ${arg} MATCHES "_CLASS_SINGLE_IO")
            #all of these classes need temp parameter and conversion
            SET (pointer "*")
            
            IF (${type} MATCHES "char[*]([^*]+|$)")
                IF (${arg} MATCHES "_CLASS_SINGLE_IO")
                    MESSAGE (FATAL_ERROR "ERROR: \"char*\" in-out arguments are not supported by the Fortran wrapper generator, extend it!")
                ENDIF (${arg} MATCHES "_CLASS_SINGLE_IO")
            
                SET (${p_text} "${${p_text}}  char *${arg}C;\n")
                SET (${p_text} "${${p_text}}  marmot_string_f2c(${arg}, ${arg}_len, &${arg}C);\n")
            ENDIF (${type} MATCHES "char[*]([^*]+|$)")
            
            IF (${type} MATCHES "MPI_Comm([^_]+|$)")
                SET (${p_text} "${${p_text}}  MPI_Comm ${arg}C = MPI_Comm_f2c(${pointer}${arg});\n")
            ENDIF (${type} MATCHES "MPI_Comm([^_]+|$)")
                
            IF (${type} MATCHES "MPI_Datatype([^_]+|$)")
                SET (${p_text} "${${p_text}}  MPI_Datatype ${arg}C = MPI_Type_f2c(${pointer}${arg});\n")
            ENDIF (${type} MATCHES "MPI_Datatype([^_]+|$)")
                
            IF (${type} MATCHES "MPI_Group([^_]+|$)")
                SET (${p_text} "${${p_text}}  MPI_Group ${arg}C = MPI_Group_f2c(${pointer}${arg});\n")
            ENDIF (${type} MATCHES "MPI_Group([^_]+|$)")
                
            IF (${type} MATCHES "MPI_Op([^_]+|$)")
                SET (${p_text} "${${p_text}}  MPI_Op ${arg}C = MPI_Op_f2c(${pointer}${arg});\n")
            ENDIF (${type} MATCHES "MPI_Op([^_]+|$)")
                
            IF (REQUEST_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Request([^_]+|$)")
                SET (${p_text} "${${p_text}}  MPI_Request ${arg}C = MPI_Request_f2c(${pointer}${arg});\n")
            ENDIF (REQUEST_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Request([^_]+|$)")
            
            IF (INFO_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Info([^_]+|$)")
                SET (${p_text} "${${p_text}}  MPI_Info ${arg}C = MPI_Info_f2c(${pointer}${arg});\n")
            ENDIF (INFO_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Info([^_]+|$)")
            
            IF (WIN_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Win([^_]+|$)")
                SET (${p_text} "${${p_text}}  MPI_Win ${arg}C = MPI_Win_f2c(${pointer}${arg});\n")
            ENDIF (WIN_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Win([^_]+|$)")
            
            IF (FILE_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_File([^_]+|$)")
                SET (${p_text} "${${p_text}}  MPI_File ${arg}C = MPI_File_f2c(${pointer}${arg});\n")
            ENDIF (FILE_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_File([^_]+|$)")
                
            IF (STATUS_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Status([^_]+|$)")
                # it is erroneous to pass MPI_STATUS_IGNORE as an input argument !
                # this design enables us to also cause this error within the C interface !
                SET (${p_text} "${${p_text}}  MPI_Status* ${arg}C;\n")
                SET (${p_text} "${${p_text}}  if ((${arg} != MPI_F_STATUS_IGNORE) && (${arg} != MPI_F_STATUSES_IGNORE)) {\n")
                SET (${p_text} "${${p_text}}  \t${arg}C = new MPI_Status;\n")
                SET (${p_text} "${${p_text}}  \tMPI_Status_f2c(${arg},${arg}C);\n")             
                SET (${p_text} "${${p_text}}  \t}\n")
                SET (${p_text} "${${p_text}}  else if (${arg} == MPI_F_STATUS_IGNORE) {\n")
                SET (${p_text} "${${p_text}}  \t${arg}C = MPI_STATUS_IGNORE;\n")
                SET (${p_text} "${${p_text}}  \t}\n")
                SET (${p_text} "${${p_text}}  else {\n")
                SET (${p_text} "${${p_text}}  \t${arg}C = MPI_STATUSES_IGNORE;\n")
                SET (${p_text} "${${p_text}}  \t}\n")
            ENDIF (STATUS_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Status([^_]+|$)")
        ENDIF ( NOT ${type} MATCHES "[*]" OR
                ${arg} MATCHES "_CLASS_SINGLE_IN" OR
                ${arg} MATCHES "_CLASS_SINGLE_IO")
        
        # a Single output argument
        IF (${arg} MATCHES "_CLASS_SINGLE_OUT") 
        
            #Here we just need to create a temp parameter
            IF (${type} MATCHES "char[*]([^*]+|$)")
                SET (${p_text} "${${p_text}}  char ${arg}C[1024];\n")
            ENDIF (${type} MATCHES "char[*]([^*]+|$)")            
            
            IF (${type} MATCHES "MPI_Comm")
                SET (${p_text} "${${p_text}}  MPI_Comm ${arg}C = MPI_COMM_NULL;\n")
            ENDIF (${type} MATCHES "MPI_Comm")
                
            IF (${type} MATCHES "MPI_Datatype")
                SET (${p_text} "${${p_text}}  MPI_Datatype ${arg}C = MPI_DATATYPE_NULL;\n")
            ENDIF (${type} MATCHES "MPI_Datatype")
                
            IF (${type} MATCHES "MPI_Group")
                SET (${p_text} "${${p_text}}  MPI_Group ${arg}C = MPI_GROUP_NULL;\n")
            ENDIF (${type} MATCHES "MPI_Group")
                
            IF (${type} MATCHES "MPI_Op")
                SET (${p_text} "${${p_text}}  MPI_Op ${arg}C = MPI_OP_NULL;\n")
            ENDIF (${type} MATCHES "MPI_Op")
                
            IF (REQUEST_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Request")
                SET (${p_text} "${${p_text}}  MPI_Request ${arg}C = MPI_REQUEST_NULL;\n")
            ENDIF (REQUEST_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Request")
            
            IF (INFO_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Info")
                SET (${p_text} "${${p_text}}  MPI_Info ${arg}C = MPI_INFO_NULL;\n")
            ENDIF (INFO_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Info")
            
            IF (WIN_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Win")
                SET (${p_text} "${${p_text}}  MPI_Win ${arg}C = MPI_WIN_NULL;\n")
            ENDIF (WIN_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Win")
            
            IF (FILE_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_File")
                SET (${p_text} "${${p_text}}  MPI_File ${arg}C = MPI_FILE_NULL;\n")
            ENDIF (FILE_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_File")
            
            IF (STATUS_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Status")
                SET (${p_text} "${${p_text}}  MPI_Status* ${arg}C;\n")
                SET (${p_text} "${${p_text}}  if (${arg} == MPI_F_STATUS_IGNORE)\n")
                SET (${p_text} "${${p_text}}  \t${arg}C = MPI_STATUS_IGNORE;\n")  #case 1 using MPI_STATUS_IGNORE
                SET (${p_text} "${${p_text}}  else if (${arg} == MPI_F_STATUSES_IGNORE)\n")
                SET (${p_text} "${${p_text}}  \t${arg}C = MPI_STATUSES_IGNORE;\n") #case 2 user erroneously passes MPI_STATUSES_IGNORE, we pass this error on, For Matthias Jurenz: you might want to handle this differently 
                SET (${p_text} "${${p_text}}  else ${arg}C = new MPI_Status;\n")  #case 3 normal status that is used
            ENDIF (STATUS_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Status")
        ENDIF (${arg} MATCHES "_CLASS_SINGLE_OUT")
        
        # an array for input and/or output purposes
        IF (${arg} MATCHES "_CLASS_ARRAY_")
            #We need for all array cases a temporary array
    
            #extract the name of the length parameter
            MARMOT_GEN_GET_LEN_PARAMETER (${arg} len_parameter)
            
            #now we have to handle all the different handles        
            IF (${type} MATCHES "MPI_Comm")
                SET (${p_text} "${${p_text}}  MPI_Comm* ${arg}C = new MPI_Comm [(*${len_parameter})];\n")
                SET (${p_text} "${${p_text}}  for (int i=0;i<*${len_parameter};i++) ${arg}C[i] = MPI_COMM_NULL;\n")
            ENDIF (${type} MATCHES "MPI_Comm")        
                
            IF (${type} MATCHES "MPI_Datatype")
                SET (${p_text} "${${p_text}}  MPI_Datatype* ${arg}C = new MPI_Datatype [(*${len_parameter})];\n")
                SET (${p_text} "${${p_text}}  for (int i=0;i<*${len_parameter};i++) ${arg}C[i] = MPI_DATATYPE_NULL;\n")
            ENDIF (${type} MATCHES "MPI_Datatype")
                
            IF (${type} MATCHES "MPI_Group")
                SET (${p_text} "${${p_text}}  MPI_Group* ${arg}C = new MPI_Group [(*${len_parameter})];\n")
                SET (${p_text} "${${p_text}}  for (int i=0;i<*${len_parameter};i++) ${arg}C[i] = MPI_GROUP_NULL;\n")
            ENDIF (${type} MATCHES "MPI_Group")
                
            IF (${type} MATCHES "MPI_Op")
                SET (${p_text} "${${p_text}}  MPI_Op* ${arg}C = new MPI_Op [(*${len_parameter})];\n")
                SET (${p_text} "${${p_text}}  for (int i=0;i<*${len_parameter};i++) ${arg}C[i] = MPI_OP_NULL;\n")
            ENDIF (${type} MATCHES "MPI_Op")    
    
            IF (REQUEST_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Request")
                SET (${p_text} "${${p_text}}  MPI_Request* ${arg}C = new MPI_Request [(*${len_parameter})];\n")
                SET (${p_text} "${${p_text}}  for (int i=0;i<*${len_parameter};i++) ${arg}C[i] = MPI_REQUEST_NULL;\n")
            ENDIF (REQUEST_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Request")
            
            IF (INFO_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Info")
                SET (${p_text} "${${p_text}}  MPI_Info* ${arg}C = new MPI_Info [(*${len_parameter})];\n")
                SET (${p_text} "${${p_text}}  for (int i=0;i<*${len_parameter};i++) ${arg}C[i] = MPI_INFO_NULL;\n")
            ENDIF (INFO_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Info")
            
            IF (WIN_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Win")
                SET (${p_text} "${${p_text}}  MPI_Win* ${arg}C = new MPI_Win [(*${len_parameter})];\n")
                SET (${p_text} "${${p_text}}  for (int i=0;i<*${len_parameter};i++) ${arg}C[i] = MPI_WIN_NULL;\n")
            ENDIF (WIN_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Win")
            
            IF (FILE_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_File")
                SET (${p_text} "${${p_text}}  MPI_File* ${arg}C = new MPI_File [(*${len_parameter})];\n")
                SET (${p_text} "${${p_text}}  for (int i=0;i<*${len_parameter};i++) ${arg}C[i] = MPI_FILE_NULL;\n")
            ENDIF (FILE_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_File")
            
            IF (STATUS_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Status")
                SET (${p_text} "${${p_text}}  MPI_Status* ${arg}C;\n")
                SET (${p_text} "${${p_text}}  if ((${arg} != MPI_F_STATUSES_IGNORE)  && (${arg} != MPI_F_STATUS_IGNORE))\n")
                SET (${p_text} "${${p_text}}  \t${arg}C = new MPI_Status [(*${len_parameter})];\n")
                SET (${p_text} "${${p_text}}  else if (${arg} == MPI_F_STATUSES_IGNORE)\n")
                SET (${p_text} "${${p_text}}  \t${arg}C = MPI_STATUSES_IGNORE;\n")
                SET (${p_text} "${${p_text}}  else  ${arg}C = MPI_STATUS_IGNORE;\n") #User made an error !!!, For Matthias Jurenz: you might to handle this differently !!
            ENDIF (STATUS_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Status")
        ENDIF (${arg} MATCHES "_CLASS_ARRAY_")
        
        #an Array for input purposes
        IF ( ${arg} MATCHES "_CLASS_ARRAY_IN_" OR
             ${arg} MATCHES "_CLASS_ARRAY_IO_")
            #both of these classes need temp array parameter and conversion
            
            #extract the name of the length parameter
            MARMOT_GEN_GET_LEN_PARAMETER (${arg} len_parameter)
            
            #now we have to handle all the different handles        
            IF (${type} MATCHES "MPI_Comm")
                SET (${p_text} "${${p_text}}  for (int i = 0;i < (*${len_parameter});i++) \n")
                SET (${p_text} "${${p_text}}  \t${arg}C[i]=MPI_Comm_f2c(${arg}[i]);\n")
            ENDIF (${type} MATCHES "MPI_Comm")
                
            IF (${type} MATCHES "MPI_Datatype")
                SET (${p_text} "${${p_text}}  for (int i = 0;i < (*${len_parameter});i++) \n")
                SET (${p_text} "${${p_text}}  \t${arg}C[i]=MPI_Type_f2c(${arg}[i]);\n")
            ENDIF (${type} MATCHES "MPI_Datatype")
                
            IF (${type} MATCHES "MPI_Group")
                SET (${p_text} "${${p_text}}  for (int i = 0;i < (*${len_parameter});i++) \n")
                SET (${p_text} "${${p_text}}  \t${arg}C[i]=MPI_Group_f2c(${arg}[i]);\n")
            ENDIF (${type} MATCHES "MPI_Group")
                
            IF (${type} MATCHES "MPI_Op")
                SET (${p_text} "${${p_text}}  for (int i = 0;i < (*${len_parameter});i++) \n")
                SET (${p_text} "${${p_text}}  \t${arg}C[i]=MPI_Op_f2c(${arg}[i]);\n")
            ENDIF (${type} MATCHES "MPI_Op")
                
            IF (REQUEST_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Request")
                SET (${p_text} "${${p_text}}  for (int i = 0;i < (*${len_parameter});i++) \n")
                SET (${p_text} "${${p_text}}  \t${arg}C[i]=MPI_Request_f2c(${arg}[i]);\n")
            ENDIF (REQUEST_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Request")
            
            IF (INFO_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Info")
                SET (${p_text} "${${p_text}}  for (int i = 0;i < (*${len_parameter});i++) \n")
                SET (${p_text} "${${p_text}}  \t${arg}C[i]=MPI_Info_f2c(${arg}[i]);\n")
            ENDIF (INFO_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Info")
            
            IF (WIN_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Win")
                SET (${p_text} "${${p_text}}  for (int i = 0;i < (*${len_parameter});i++) \n")
                SET (${p_text} "${${p_text}}  \t${arg}C[i]=MPI_Win_f2c(${arg}[i]);\n")
            ENDIF (WIN_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Win")
            
            IF (FILE_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_File")
                SET (${p_text} "${${p_text}}  for (int i = 0;i < (*${len_parameter});i++) \n")
                SET (${p_text} "${${p_text}}  \t${arg}C[i]=MPI_File_f2c(${arg}[i]);\n")
            ENDIF (FILE_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_File")
                
            IF (STATUS_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Status")
                SET (${p_text} "${${p_text}}  if ((${arg} != MPI_F_STATUSES_IGNORE) && (${arg} != MPI_F_STATUS_IGNORE)) {\n")
                SET (${p_text} "${${p_text}}  \tfor (int i = 0;i < (*${len_parameter});i++) \n")
                SET (${p_text} "${${p_text}}  \t\tMPI_Status_f2c(&${arg}[i],&${arg}C[i*MPO_Space::getStatusSizeFortran()]);\n")
                SET (${p_text} "${${p_text}}  }\n")
            ENDIF (STATUS_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Status")
        ENDIF ( ${arg} MATCHES "_CLASS_ARRAY_IN_" OR
                ${arg} MATCHES "_CLASS_ARRAY_IO_")       
    ENDIF (HANDLE_CONVERT MATCHES "YES") 
ENDMACRO (PRINT_CONVERT_FUNCTION p_text type arg)

#===================================
# Macro: PRINT_BACK_CONVERT_FUNCTION
#===================================
MACRO (PRINT_BACK_CONVERT_FUNCTION p_text type arg)
    
    IF (HANDLE_CONVERT MATCHES "YES")
        # SINGLE ARGUMENT BACK CONVERSION    
        IF (${arg} MATCHES "_CLASS_SINGLE_OUT" OR
            ${arg} MATCHES "_CLASS_SINGLE_IO")
            #this class needs back conversion of a single argument
            
            IF (${type} MATCHES "char[*]([^*]+|$)")
                SET (${p_text} "${${p_text}}  marmot_string_c2f(${arg}C, ${arg}, ${arg}_len);\n")
            ENDIF (${type} MATCHES "char[*]([^*]+|$)")
            
            IF (${type} MATCHES "MPI_Comm")
                SET (${p_text} "${${p_text}}  *${arg} = MPI_Comm_c2f(${arg}C);\n")
            ENDIF (${type} MATCHES "MPI_Comm")
                
            IF (${type} MATCHES "MPI_Datatype")
                SET (${p_text} "${${p_text}}  *${arg} = MPI_Type_c2f(${arg}C);\n")
            ENDIF (${type} MATCHES "MPI_Datatype")
                
            IF (${type} MATCHES "MPI_Group")
                SET (${p_text} "${${p_text}}  *${arg} = MPI_Group_c2f(${arg}C);\n")
            ENDIF (${type} MATCHES "MPI_Group")
                
            IF (${type} MATCHES "MPI_Op")
                SET (${p_text} "${${p_text}}  *${arg} = MPI_Op_c2f(${arg}C);\n")
            ENDIF (${type} MATCHES "MPI_Op")
            
            IF (REQUEST_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Request")
                SET (${p_text} "${${p_text}}  *${arg} = MPI_Request_c2f(${arg}C);\n")
            ENDIF (REQUEST_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Request")
            
            IF (INFO_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Info")
                SET (${p_text} "${${p_text}}  *${arg} = MPI_Info_c2f(${arg}C);\n")
            ENDIF (INFO_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Info")
            
            IF (WIN_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Win")
                SET (${p_text} "${${p_text}}  *${arg} = MPI_Win_c2f(${arg}C);\n")
            ENDIF (WIN_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Win")
            
            IF (FILE_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_File")
                SET (${p_text} "${${p_text}}  *${arg} = MPI_File_c2f(${arg}C);\n")
            ENDIF (FILE_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_File")
                
            IF (STATUS_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Status")
                SET (${p_text} "${${p_text}}  if ((${arg} != MPI_F_STATUS_IGNORE) && (${arg} != MPI_F_STATUSES_IGNORE))\n")
                SET (${p_text} "${${p_text}}  \tMPI_Status_c2f(${arg}C,${arg});\n")
            ENDIF (STATUS_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Status")
        ENDIF (${arg} MATCHES "_CLASS_SINGLE_OUT" OR
               ${arg} MATCHES "_CLASS_SINGLE_IO")
    
        #ARRAY BACK CONVERSION
        IF ( ${arg} MATCHES "_CLASS_ARRAY_OUT_" OR
             ${arg} MATCHES "_CLASS_ARRAY_IO_")
            #both of these classes need array back conversion
            
            #extract the name of the length parameter
            MARMOT_GEN_GET_LEN_PARAMETER (${arg} len_parameter)
    
            #now we have to handle all the imaginable handles        
            IF (${type} MATCHES "MPI_Comm")
                SET (${p_text} "${${p_text}}  if (*${len_parameter} != MPI_UNDEFINED)\n")
                SET (${p_text} "${${p_text}}    for (int i = 0;i < (*${len_parameter});i++) \n")
                SET (${p_text} "${${p_text}}      ${arg}[i]=MPI_Comm_c2f(${arg}C[i]);\n")
            ENDIF (${type} MATCHES "MPI_Comm")
                
            IF (${type} MATCHES "MPI_Datatype")
                SET (${p_text} "${${p_text}}  if (*${len_parameter} != MPI_UNDEFINED)\n")
                SET (${p_text} "${${p_text}}    for (int i = 0;i < (*${len_parameter});i++) \n")
                SET (${p_text} "${${p_text}}      ${arg}[i]=MPI_Type_c2f(${arg}C[i]);\n")
            ENDIF (${type} MATCHES "MPI_Datatype")
                
            IF (${type} MATCHES "MPI_Group")
                SET (${p_text} "${${p_text}}  if (*${len_parameter} != MPI_UNDEFINED)\n")
                SET (${p_text} "${${p_text}}    for (int i = 0;i < (*${len_parameter});i++) \n")
                SET (${p_text} "${${p_text}}      ${arg}[i]=MPI_Group_c2f(${arg}C[i]);\n")
            ENDIF (${type} MATCHES "MPI_Group")
                
            IF (${type} MATCHES "MPI_Op")
                SET (${p_text} "${${p_text}}  if (*${len_parameter} != MPI_UNDEFINED)\n")
                SET (${p_text} "${${p_text}}    for (int i = 0;i < (*${len_parameter});i++) \n")
                SET (${p_text} "${${p_text}}      ${arg}[i]=MPI_Op_c2f(${arg}C[i]);\n")
            ENDIF (${type} MATCHES "MPI_Op")
                
            IF (REQUEST_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Request")
                SET (${p_text} "${${p_text}}  if (*${len_parameter} != MPI_UNDEFINED)\n")
                SET (${p_text} "${${p_text}}    for (int i = 0;i < (*${len_parameter});i++) \n")
                SET (${p_text} "${${p_text}}      ${arg}[i]=MPI_Request_c2f(${arg}C[i]);\n")
            ENDIF (REQUEST_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Request")
            
            IF (INFO_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Info")
                SET (${p_text} "${${p_text}}  if (*${len_parameter} != MPI_UNDEFINED)\n")
                SET (${p_text} "${${p_text}}    for (int i = 0;i < (*${len_parameter});i++) \n")
                SET (${p_text} "${${p_text}}      ${arg}[i]=MPI_Info_c2f(${arg}C[i]);\n")
            ENDIF (INFO_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Info")
            
            IF (WIN_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Win")
                SET (${p_text} "${${p_text}}  if (*${len_parameter} != MPI_UNDEFINED)\n")
                SET (${p_text} "${${p_text}}    for (int i = 0;i < (*${len_parameter});i++) \n")
                SET (${p_text} "${${p_text}}      ${arg}[i]=MPI_Win_c2f(${arg}C[i]);\n")
            ENDIF (WIN_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Win")
            
            IF (FILE_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_File")
                SET (${p_text} "${${p_text}}  if (*${len_parameter} != MPI_UNDEFINED)\n")
                SET (${p_text} "${${p_text}}    for (int i = 0;i < (*${len_parameter});i++) \n")
                SET (${p_text} "${${p_text}}      ${arg}[i]=MPI_File_c2f(${arg}C[i]);\n")
            ENDIF (FILE_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_File")
                
            IF (STATUS_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Status")
                SET (${p_text} "${${p_text}}  if ((*${len_parameter} != MPI_UNDEFINED) &&\n")
                SET (${p_text} "${${p_text}}      (${arg} != MPI_F_STATUSES_IGNORE) && (${arg} != MPI_F_STATUS_IGNORE)) {\n")
                SET (${p_text} "${${p_text}}    for (int i = 0;i < (*${len_parameter});i++) \n")
                SET (${p_text} "${${p_text}}      MPI_Status_c2f(&${arg}C[i],&${arg}[i*MPO_Space::getStatusSizeFortran()]);\n")
                SET (${p_text} "${${p_text}}  }\n")
            ENDIF (STATUS_CONVERT MATCHES "YES" AND ${type} MATCHES "MPI_Status")
        ENDIF (${arg} MATCHES "_CLASS_ARRAY_OUT_" OR
               ${arg} MATCHES "_CLASS_ARRAY_IO_")
        
        #free the array memory for all handles but statuses !
        IF ( ${arg} MATCHES "_CLASS_ARRAY_" AND
             NOT ${type} MATCHES "MPI_Status")
            IF (HANDLE_CONVERT MATCHES "YES")
                IF (${type} MATCHES "MPI_Comm" OR
                    ${type} MATCHES "MPI_Datatype" OR
                    ${type} MATCHES "MPI_Group" OR
                    ${type} MATCHES "MPI_Op")
                    SET (${p_text} "${${p_text}}  delete [] (${arg}C);\n")
                ENDIF (${type} MATCHES "MPI_Comm" OR
                    ${type} MATCHES "MPI_Datatype" OR
                    ${type} MATCHES "MPI_Group" OR
                    ${type} MATCHES "MPI_Op")    
                    
                IF (${type} MATCHES "MPI_Request" AND REQUEST_CONVERT MATCHES "YES")
                    SET (${p_text} "${${p_text}}  delete [] (${arg}C);\n")
                ENDIF (${type} MATCHES "MPI_Request" AND REQUEST_CONVERT MATCHES "YES")
                
                IF (${type} MATCHES "MPI_Info" AND INFO_CONVERT MATCHES "YES")
                    SET (${p_text} "${${p_text}}  delete [] (${arg}C);\n")
                ENDIF (${type} MATCHES "MPI_Info" AND INFO_CONVERT MATCHES "YES")
                
                IF (${type} MATCHES "MPI_Win" AND WIN_CONVERT MATCHES "YES")
                    SET (${p_text} "${${p_text}}  delete [] (${arg}C);\n")
                ENDIF (${type} MATCHES "MPI_Win" AND WIN_CONVERT MATCHES "YES")
                
                IF (${type} MATCHES "MPI_File" AND FILE_CONVERT MATCHES "YES")
                    SET (${p_text} "${${p_text}}  delete [] (${arg}C);\n")
                ENDIF (${type} MATCHES "MPI_File" AND FILE_CONVERT MATCHES "YES")
            ENDIF (HANDLE_CONVERT MATCHES "YES")
        ENDIF ( ${arg} MATCHES "_CLASS_ARRAY_" AND
             NOT ${type} MATCHES "MPI_Status")
        
        #free the array memory for statuses
        IF ( ${arg} MATCHES "_CLASS_ARRAY_" AND
             ${type} MATCHES "MPI_Status" AND
             STATUS_CONVERT MATCHES "YES")
            SET (${p_text} "${${p_text}}  if ((${arg} != MPI_F_STATUSES_IGNORE) && (${arg} != MPI_F_STATUS_IGNORE))\n")
            SET (${p_text} "${${p_text}}    delete [] (${arg}C);\n")
        ENDIF ( ${arg} MATCHES "_CLASS_ARRAY_" AND
                ${type} MATCHES "MPI_Status" AND
                STATUS_CONVERT MATCHES "YES")
            
        #free Status memory (for single status values)
        IF ( ${arg} MATCHES "_CLASS_SINGLE_" AND
             ${type} MATCHES "MPI_Status" AND
             STATUS_CONVERT MATCHES "YES")
            SET (${p_text} "${${p_text}}  if ((${arg} != MPI_F_STATUSES_IGNORE) && (${arg} != MPI_F_STATUS_IGNORE))\n")
            SET (${p_text} "${${p_text}}    delete (${arg}C);\n")
        ENDIF ( ${arg} MATCHES "_CLASS_SINGLE_" AND
                ${type} MATCHES "MPI_Status" AND
                STATUS_CONVERT MATCHES "YES")
                
        #free memory for single input strings
        IF (${type} MATCHES "char[*]([^*]+|$)" AND ${arg} MATCHES "_CLASS_SINGLE_IN")
            SET (${p_text} "${${p_text}}  if (${arg}C) delete (${arg}C);\n")
        ENDIF (${type} MATCHES "char[*]([^*]+|$)" AND ${arg} MATCHES "_CLASS_SINGLE_IN")
    ENDIF (HANDLE_CONVERT MATCHES "YES")
ENDMACRO (PRINT_BACK_CONVERT_FUNCTION p_text type arg)

#===================================
# Macro: CONVERT_OUTPUT_ARRAY_INDICES
#===================================
MACRO (CONVERT_OUTPUT_ARRAY_INDICES p_text type arg)
    #We have three different cases:
    # * _CLASS_ARRAYINDEX_OUT_SINGLE_NOCONDITION
    # * _CLASS_ARRAYINDEX_OUT_SINGLE_CONDITION_conditionargument
    # * _CLASS_ARRAYINDEX_OUT_ARRAY_sizeargument
    
    #is it the "_CLASS_ARRAYINDEX_OUT_SINGLE_NOCONDITION" case ?
    IF (${arg} MATCHES "_CLASS_ARRAYINDEX_OUT_SINGLE_NOCONDITION")
        SET (${p_text} "${${p_text}}  if ( (*${arg} != MPI_UNDEFINED) && (*${arg} >= 0)) \n  {\n")
        SET (${p_text} "${${p_text}}      *${arg}=*${arg}+1;\n  }\n")
    ENDIF (${arg} MATCHES "_CLASS_ARRAYINDEX_OUT_SINGLE_NOCONDITION")
    
    #is it the "_CLASS_ARRAYINDEX_OUT_SINGLE_CONDITION_<ARG>" case ?
    IF (${arg} MATCHES "_CLASS_ARRAYINDEX_OUT_SINGLE_CONDITION_")
        #extract the name of the "conditionargument" parameter
        MARMOT_GEN_GET_LEN_PARAMETER ("${arg}" cond)
        
        SET (${p_text} "${${p_text}}  if ( (*${cond}) && (*${arg} != MPI_UNDEFINED) && (*${arg} >= 0)) \n  {\n")
        SET (${p_text} "${${p_text}}      *${arg}=*${arg}+1;\n  }\n")
    ENDIF (${arg} MATCHES "_CLASS_ARRAYINDEX_OUT_SINGLE_CONDITION_")
    
    #is it the "_CLASS_ARRAYINDEX_OUT_ARRAY_<ARG>" case ?
    IF (${arg} MATCHES "_CLASS_ARRAYINDEX_OUT_ARRAY_")
        #extract the name of the "conditionargument" parameter
        MARMOT_GEN_GET_LEN_PARAMETER ("${arg}" count)
        
        SET (${p_text} "${${p_text}}  if ( *${count} != MPI_UNDEFINED )\n  {\n") 
        SET (${p_text} "${${p_text}}      for (int i = 0; i < *${count};i++)\n      {\n")
        SET (${p_text} "${${p_text}}          if ( (${arg}[i] != MPI_UNDEFINED) && (${arg}[i] >= 0))\n          {\n")
        SET (${p_text} "${${p_text}}              ${arg}[i]=${arg}[i]+1;\n          }\n      }\n  }\n")
    ENDIF (${arg} MATCHES "_CLASS_ARRAYINDEX_OUT_ARRAY_")
ENDMACRO (CONVERT_OUTPUT_ARRAY_INDICES p_text type arg)

#===================================
# Macro: PRINT_FINT_TO_AINT
#===================================
MACRO (PRINT_FINT_TO_AINT p_text type arg callname)
    IF ("${callname}" MATCHES "MPI_Type_hindexed" OR
        "${callname}" MATCHES "MPI_Type_hvector" OR
        "${callname}" MATCHES "MPI_Type_struct" OR
        "${callname}" MATCHES "MPI_Address" OR 
        "${callname}" MATCHES "MPI_Type_extent" OR
        "${callname}" MATCHES "MPI_Type_lb" OR
        "${callname}" MATCHES "MPI_Type_ub" )
        IF (type MATCHES "MPI_Aint[*]")
            
            IF (${arg} MATCHES "_CLASS_SINGLE_IN" OR
                ${arg} MATCHES "_CLASS_SINGLE_IO")
                SET (${p_text} "${${p_text}} MPI_Aint ${arg}C = (MPI_Aint)(*${arg});\n")
            ENDIF (${arg} MATCHES "_CLASS_SINGLE_IN" OR
                   ${arg} MATCHES "_CLASS_SINGLE_IO")
                   
            IF (${arg} MATCHES "_CLASS_SINGLE_OUT")
                SET (${p_text} "${${p_text}} MPI_Aint ${arg}C = 0;\n")            
            ENDIF (${arg} MATCHES "_CLASS_SINGLE_OUT")
            
            IF ( ${arg} MATCHES "_CLASS_ARRAY_IN_" OR
                 ${arg} MATCHES "_CLASS_ARRAY_IO_")
                #extract the name of the length parameter
                MARMOT_GEN_GET_LEN_PARAMETER (${arg} len_parameter)
                        
                SET (${p_text} "${${p_text}} MPI_Aint* ${arg}C = NULL;\n")
                SET (${p_text} "${${p_text}} if (*${len_parameter} > 0)\n")
                SET (${p_text} "${${p_text}} {\n")
                SET (${p_text} "${${p_text}}     ${arg}C = new MPI_Aint [*${len_parameter}];\n")
                SET (${p_text} "${${p_text}}     for (int i = 0; i < *${len_parameter}; i++)\n")
                SET (${p_text} "${${p_text}}         ${arg}C[i] = (MPI_Aint)${arg}[i];\n")
                SET (${p_text} "${${p_text}} }\n")
            ENDIF ( ${arg} MATCHES "_CLASS_ARRAY_IN_" OR
                    ${arg} MATCHES "_CLASS_ARRAY_IO_")
                    
            IF ( ${arg} MATCHES "_CLASS_ARRAY_OUT_")
                #extract the name of the length parameter
                MARMOT_GEN_GET_LEN_PARAMETER (${arg} len_parameter)
                        
                SET (${p_text} "${${p_text}} MPI_Aint* ${arg}C = NULL;\n")
                SET (${p_text} "${${p_text}} if (*${len_parameter} > 0)\n")
                SET (${p_text} "${${p_text}}     ${arg}C = new MPI_Aint [*${len_parameter}];\n")
            ENDIF ( ${arg} MATCHES "_CLASS_ARRAY_OUT_")
        ELSEIF (type MATCHES "MPI_Aint")
            SET (${p_text} "${${p_text}} MPI_Aint ${arg}C = (MPI_Aint)(*${arg});\n")
        ENDIF (type MATCHES "MPI_Aint[*]")
    ENDIF ("${callname}" MATCHES "MPI_Type_hindexed" OR
        "${callname}" MATCHES "MPI_Type_hvector" OR
        "${callname}" MATCHES "MPI_Type_struct" OR
        "${callname}" MATCHES "MPI_Address" OR 
        "${callname}" MATCHES "MPI_Type_extent" OR
        "${callname}" MATCHES "MPI_Type_lb" OR
        "${callname}" MATCHES "MPI_Type_ub")
    
ENDMACRO (PRINT_FINT_TO_AINT)

#===================================
# Macro: FREE_FINT_TO_AINT
#===================================
MACRO (FREE_FINT_TO_AINT p_text type arg callname)
    IF ("${callname}" MATCHES "MPI_Type_hindexed" OR
        "${callname}" MATCHES "MPI_Type_hvector" OR
        "${callname}" MATCHES "MPI_Type_struct" OR
        "${callname}" MATCHES "MPI_Address" OR 
        "${callname}" MATCHES "MPI_Type_extent" OR
        "${callname}" MATCHES "MPI_Type_lb" OR
        "${callname}" MATCHES "MPI_Type_ub")
        IF (type MATCHES "MPI_Aint[*]")
        
            IF (${arg} MATCHES "_CLASS_SINGLE_OUT" OR
                ${arg} MATCHES "_CLASS_SINGLE_IO")
                SET (${p_text} "${${p_text}} *${arg} = (MPI_Fint)${arg}C;\n")
            ENDIF (${arg} MATCHES "_CLASS_SINGLE_OUT" OR
                   ${arg} MATCHES "_CLASS_SINGLE_IO")
                   
            IF ( ${arg} MATCHES "_CLASS_ARRAY_OUT_" OR
                 ${arg} MATCHES "_CLASS_ARRAY_IO_")
                #extract the name of the length parameter
                MARMOT_GEN_GET_LEN_PARAMETER (${arg} len_parameter)
                        
                SET (${p_text} "${${p_text}} if (*${len_parameter} > 0)\n")
                SET (${p_text} "${${p_text}} {\n")
                SET (${p_text} "${${p_text}}     for (int i = 0; i < *${len_parameter}; i++)\n")
                SET (${p_text} "${${p_text}}         ${arg}[i] = (MPIFint)${arg}C[i];\n")
                SET (${p_text} "${${p_text}} }\n")
            ENDIF ( ${arg} MATCHES "_CLASS_ARRAY_OUT_" OR
                    ${arg} MATCHES "_CLASS_ARRAY_IO_")
                    
            IF ( ${arg} MATCHES "_CLASS_ARRAY_")
                SET (${p_text} "${${p_text}} if (${arg}C)\n")
                SET (${p_text} "${${p_text}}     delete [] ${arg}C;\n")
            ENDIF ( ${arg} MATCHES "_CLASS_ARRAY_")
        
        ENDIF (type MATCHES "MPI_Aint[*]")
    ENDIF ("${callname}" MATCHES "MPI_Type_hindexed" OR
        "${callname}" MATCHES "MPI_Type_hvector" OR
        "${callname}" MATCHES "MPI_Type_struct" OR
        "${callname}" MATCHES "MPI_Address" OR 
        "${callname}" MATCHES "MPI_Type_extent" OR
        "${callname}" MATCHES "MPI_Type_lb" OR
        "${callname}" MATCHES "MPI_Type_ub")    
ENDMACRO (FREE_FINT_TO_AINT)

#===================================
# Macro: MARMOT_GEN_CREATE_WRAPPER
#===================================
#This is a hook name, as the parser calls
#a macro of that name.
MACRO (MARMOT_GEN_CREATE_WRAPPER outfile callname retval p_args_n_types)
    SET (args_n_types ${${p_args_n_types}})

    #Text for the wrapper function
    SET (text "")
    
    #Is it implemented in Marmot ?
    MARMOT_GEN_IS_IMPLEMENTED ("${callname}" implemented)
    
    #This loop creates two different versions of the calls one with aditional 
    #parameters for line and file name and one without
    #
    # WARNING: as handling of Strings differs between different Fortran versions
    #          we have to implement a handling that matches the version dependent behavior
    #          actually we only handle one of those different possibilities !
    #
    FOREACH(e "reg" "sourceinfo")   
       
        # Start with function name of wrapper
        STRING (TOUPPER ${callname} fn_name)
        IF (e MATCHES "reg")
            SET (text "${text}\nextern \"C\" void ${fn_name} (")
        ELSE (e MATCHES "reg")
            STRING (REGEX REPLACE "MPI_" "MPI_SOURCEINFO_" sourceinfo_name ${fn_name})
            SET (text "${text}\nextern \"C\" void ${sourceinfo_name} (")
        ENDIF (e MATCHES "reg")
          
        SET (mpocall ${callname})
        SET (strings)
        SET (numstring "0")
        
        IF (args_n_types)
            #For calls with arguments, loop over them and print
            SET (type)
            FOREACH (t_n_a ${args_n_types})
                IF (type)
                  SET (text "${text}, ")
                ENDIF (type)

                MARMOT_GEN_SPLIT_T_N_A ("${t_n_a}" type arg)
                CONVERTTYPE ("${type}" conv_type "${callname}")
                
                #Is it a init call ? Then skip "argc" and "argv"
                IF (fn_name MATCHES "MPI_INIT[^A-Za-z0-9_]*$" OR
                    fn_name MATCHES "MPI_INIT_THREAD[^A-Za-z0-9_]*$")
                    IF (arg MATCHES "argc" OR arg MATCHES "argv")
                        SET (type)
                        SET (arg)
                        SET (conv_type)
                    ENDIF (arg MATCHES "argc" OR arg MATCHES "argv")
                ENDIF (fn_name MATCHES "MPI_INIT[^A-Za-z0-9_]*$" OR
                       fn_name MATCHES "MPI_INIT_THREAD[^A-Za-z0-9_]*$")
                
                #print argument and type
                IF (type)
                    #Count number of strings, and their name
                    IF (type MATCHES "char[*]([^*]+|$)")
                        #special case for char* !
                        LIST(APPEND strings "${arg}")
                        MATH (EXPR numstring "${numstring} + 1")
                    ENDIF (type MATCHES "char[*]([^*]+|$)")
                    
                    IF (type MATCHES "[*]" OR arg MATCHES "\\[.*\\]")
                        #case for pointers and arrays
                        SET (text "${text}${conv_type} ${arg}")
                    ELSE (type MATCHES "[*]" OR arg MATCHES "\\[.*\\]")
                        #case for single arguments
                        SET (text "${text}${conv_type} * ${arg}")
                    ENDIF (type MATCHES "[*]" OR arg MATCHES "\\[.*\\]")
                    
                ENDIF (type)
            ENDFOREACH (t_n_a)
            
            IF (type)
                SET (text "${text}, ")
            ENDIF (type)
        ENDIF (args_n_types)
        
        #ierror and line+file for sourceinfo
        IF (e MATCHES "reg")
             SET (text "${text}MPI_Fint* ierror")
        ELSE (e MATCHES "reg")
             SET (text "${text}int* line,char *scfile,MPI_Fint* ierror")
        ENDIF (e MATCHES "reg")     
        
        #len arguments resulting from fortran strings
        FOREACH (string ${strings})
            SET (text "${text}, int ${string}_len")
        ENDFOREACH (string) 
        
        #len argument for file name (in case of sourceinfo call)
        IF (e MATCHES "sourceinfo")
             SET (text "${text}, int len")
        ENDIF (e MATCHES "sourceinfo")

        #Finish the function line                
        SET (text "${text})\n")
        
        #we are done with the prototype, now we start with the real function:
        SET (text "${text}{\n")
        
        STRING (REPLACE "MPI_" "MPO_" mpocall ${mpocall})

        #We use this one in order to get the mpo id and callname of an mpo object.
        #Otherwise we would have to create the object for this call before the
        #enterMARMOT call which causes problems with non threadsafe implementations
        #of memory allocation.
        IF (implemented)
            SET (text "${text}   static ${mpocall} temp;\n")
        ELSE (implemented)
            #create some temp object, not of much interest which one
            SET (text "${text}   static MPO_Comm_rank temp;\n")
        ENDIF (implemented)
        
        #convert fortran source file name string to c string
        IF (e MATCHES "sourceinfo")
            SET (text "${text}  char cBuf[256];\n")
            SET (text "${text}  if (len > 255) len = 255;\n")
            SET (text "${text}  cBuf[len] = 0;\n")
            SET (text "${text}  while (len > 0){len--;cBuf[len] = scfile[len];}\n")        
        ENDIF (e MATCHES "sourceinfo")
        
        #print special conversion line for MPI_Waitsome, MPI_Testsome
        #For these calls it is necessary to allocate an amount of statuses
        #equal to the value speciefied in incount and to convert as many 
        #statuses as defined by the return value outcount. In order to avoid
        #special handling the initial value for outcount is set to the value 
        #of incount.
        IF (fn_name MATCHES "MPI_WAITSOME" OR
            fn_name MATCHES "MPI_TESTSOME" )
            SET (text "${text}  *outcount = *incount;\n")
        ENDIF (fn_name MATCHES "MPI_WAITSOME" OR
               fn_name MATCHES "MPI_TESTSOME" )
              
        # MPI_Init for Fortran has some extra handling
        IF (fn_name MATCHES "MPI_INIT[^A-Za-z0-9_]*$" OR
            fn_name MATCHES "MPI_INIT_THREAD[^A-Za-z0-9_]*$")
            FORTRAN_INIT_PRE (text)
        ENDIF (fn_name MATCHES "MPI_INIT[^A-Za-z0-9_]*$" OR
               fn_name MATCHES "MPI_INIT_THREAD[^A-Za-z0-9_]*$")
                
        # do the handle conversion if possible
        FOREACH (t_n_a ${args_n_types})
            MARMOT_GEN_SPLIT_T_N_A ("${t_n_a}" type arg)
            PRINT_CONVERT_FUNCTION (text "${type}" "${arg}")   
        ENDFOREACH (t_n_a)
            
        # now test whether we need registration of requests
        FOREACH (t_n_a ${args_n_types})
            MARMOT_GEN_SPLIT_T_N_A ("${t_n_a}" type arg)
            PRINT_REQUEST_REGISTER (text "${type}" "${arg}" "${callname}" "${e}" "YES")  
        ENDFOREACH (t_n_a)
        
        # now test whether we need registration of collective comms
        FOREACH (t_n_a ${args_n_types})
            MARMOT_GEN_SPLIT_T_N_A ("${t_n_a}" type arg)
            PRINT_COLLCOMM_REGISTER (text "${type}" "${arg}" "${callname}" "${e}" "YES")  
        ENDFOREACH (t_n_a)
        
        #now test for MPI_Fint to MPI_Aint conversion (only a very few calls)
        FOREACH (t_n_a ${args_n_types})
            MARMOT_GEN_SPLIT_T_N_A ("${t_n_a}" type arg)
            PRINT_FINT_TO_AINT (text "${type}" "${arg}" "${callname}")  
        ENDFOREACH (t_n_a)

        #synchronisation    
        SET (text "${text}#ifdef MARMOT_USE_THREADSYNC\n")
        IF (e MATCHES "sourceinfo")
            SET (text "${text}      bool enter_succeeded = MPO_Space::getSyncObject()->enterMARMOT(*line,cBuf);\n")
        ELSE (e MATCHES "sourceinfo") 
            SET (text "${text}      bool enter_succeeded = MPO_Space::getSyncObject()->enterMARMOT(0,\"\");\n")
        ENDIF (e MATCHES "sourceinfo")
        SET (text "${text}#endif\n")
        
        SET (text "${text}  INTERFACE_MODE = 1;\n")
        
        #print the creation of the MPO Object and set the source location
        #Extra for MPI_Init that calls PMPI during object creation !
        IF ("${callname}" MATCHES "MPI_Init[^A-Za-z0-9_]*$" OR
            "${callname}" MATCHES "MPI_Init_thread[^A-Za-z0-9_]*$")
      	    SET (text "${text}MPO_in_profile=1;\n")
        ENDIF ("${callname}" MATCHES "MPI_Init[^A-Za-z0-9_]*$" OR
               "${callname}" MATCHES "MPI_Init_thread[^A-Za-z0-9_]*$")
          
        IF (implemented)
            SET (text "${text}  ${mpocall} my_obj")
            IF (args_n_types)
              SET (text "${text}(")
              SET (type)
              FOREACH (t_n_a ${args_n_types})
                IF (type)
                    SET (text "${text}, ")
                ENDIF (type)
                MARMOT_GEN_SPLIT_T_N_A ("${t_n_a}" type arg)
                
                # If the original type already needed a pointer, pass as is,
                # otherwise dereference the parameter
                GET_CONVERTED_NAME ("${arg}" "${type}" conv_name "YES" "${callname}")
                
                #Is it a init call ? Then skip add "&" to "argc" and "argv"
                IF (fn_name MATCHES "MPI_INIT[^A-Za-z0-9_]*$" OR
                    fn_name MATCHES "MPI_INIT_THREAD[^A-Za-z0-9_]*$")
                    IF (arg MATCHES "argc" OR arg MATCHES "argv")
                        SET (conv_name "&${arg}")
                    ENDIF (arg MATCHES "argc" OR arg MATCHES "argv")
                ENDIF (fn_name MATCHES "MPI_INIT[^A-Za-z0-9_]*$" OR
                       fn_name MATCHES "MPI_INIT_THREAD[^A-Za-z0-9_]*$")
                
                SET (text "${text}${conv_name}")
              ENDFOREACH (t_n_a)
              SET (text "${text})")
            ENDIF (args_n_types)
            SET (text "${text};\n")
            
            #this if is necessary for MPI implems like MPICH that perform MPI calls during MPI_Init
            IF ("${callname}" MATCHES "MPI_Init[^A-Za-z0-9_]*$" OR
                "${callname}" MATCHES "MPI_Init_thread[^A-Za-z0-9_]*$")
          	    SET (text "${text}MPO_in_profile=0;\n")
            ENDIF ("${callname}" MATCHES "MPI_Init[^A-Za-z0-9_]*$" OR
                   "${callname}" MATCHES "MPI_Init_thread[^A-Za-z0-9_]*$")
            
            #add the source location to the object if possible
            IF (e MATCHES "sourceinfo")
                SET (text "${text}  my_obj.setSourceLocation(*line,cBuf); \n")
            ENDIF (e MATCHES "sourceinfo")
                        
            SET (text "${text}#ifdef MARMOT_USE_THREADSYNC\n")
            SET (text "${text}   if(enter_succeeded ){\n")
            SET (text "${text}#else\n")
            SET (text "${text}   if(! MPO_in_profile ){\n")
            SET (text "${text}#endif\n")
            SET (text "${text}      MPO_in_profile=1;\n")
            SET (text "${text}      *ierror = (MPI_Fint) my_obj.callDebugServer();\n")
            SET (text "${text}      MPO_in_profile=0;\n")
            SET (text "${text}   }\n")
            SET (text "${text}   else\n")
            SET (text "${text}      *ierror = P${callname}(")
        ELSE (implemented)
            SET (text "${text}  static bool not_implemented_warned = false;\n")
            SET (text "${text}#ifdef MARMOT_USE_THREADSYNC\n")
            SET (text "${text}  if(enter_succeeded ){\n")
            SET (text "${text}#else\n")
            SET (text "${text}  if(! MPO_in_profile ){\n")
            SET (text "${text}#endif\n")
            SET (text "${text}      if (!not_implemented_warned)\n")
            SET (text "${text}      {\n")
            SET (text "${text}          std::cerr << \"WARNING: Application issued an MPI call not implemented by Marmot, false positives and non-reported errors are possible. (Call: ${callname})\" << std::endl;\n")
            SET (text "${text}          not_implemented_warned = true;\n")
            SET (text "${text}      }\n")
            SET (text "${text}  }\n")
            SET (text "${text}  *ierror = P${callname}(")
        ENDIF (implemented)
        
        SET (type)
        FOREACH (t_n_a ${args_n_types})
            IF (type)
                SET (text "${text}, ")
            ENDIF (type)
            MARMOT_GEN_SPLIT_T_N_A ("${t_n_a}" type arg)
                
            # If the original type already needed a pointer, pass as is,
            # otherwise dereference the parameter
            GET_CONVERTED_NAME ("${arg}" "${type}" conv_name "YES" "${callname}")
            
            #Is it a init call ? Then skip add "&" to "argc" and "argv"
            IF (fn_name MATCHES "MPI_INIT[^A-Za-z0-9_]*$" OR
                fn_name MATCHES "MPI_INIT_THREAD[^A-Za-z0-9_]*$")
                IF (arg MATCHES "argc" OR arg MATCHES "argv")
                    SET (conv_name "&${arg}")
                ENDIF (arg MATCHES "argc" OR arg MATCHES "argv")
            ENDIF (fn_name MATCHES "MPI_INIT[^A-Za-z0-9_]*$" OR
                   fn_name MATCHES "MPI_INIT_THREAD[^A-Za-z0-9_]*$")
                   
            #Is it a non-implemented call with a communicator ?
            #Replace MPI_COMM_WORLD with out_comm_world in that case
            IF (${type} MATCHES "MPI_Comm([^_]+|$)" AND
                (${arg} MATCHES "_CLASS_SINGLE_IN"  OR 
                 NOT ${type} MATCHES "[*]") ) 
                SET (conv_name "MPO_Comm(${conv_name})")
            ENDIF (${type} MATCHES "MPI_Comm([^_]+|$)" AND
                   (${arg} MATCHES "_CLASS_SINGLE_IN"  OR 
                    NOT ${type} MATCHES "[*]") )
                
            #Remove any existing array brackets (group_range_incl)
            SET (text "${text}${conv_name}")
        ENDFOREACH (t_n_a)
        SET (text "${text});\n")
                
        SET (text "${text}  INTERFACE_MODE = 0;\n")
        
        #now test for freeing of arrays for MPI_Fint to MPI_Aint conversion (only a very few calls)
        FOREACH (t_n_a ${args_n_types})
            MARMOT_GEN_SPLIT_T_N_A ("${t_n_a}" type arg)
            FREE_FINT_TO_AINT (text "${type}" "${arg}" "${callname}")  
        ENDFOREACH (t_n_a)
        
        # now test whether we need de-registration of collective comms
        FOREACH (t_n_a ${args_n_types})
            MARMOT_GEN_SPLIT_T_N_A ("${t_n_a}" type arg)
            PRINT_COLLCOMM_UNREGISTER (text "${type}" "${arg}" "${callname}" "${e}" "YES")  
        ENDFOREACH (t_n_a)
        
        # finaly do the handle back conversion if possible/necessary
        FOREACH (t_n_a ${args_n_types})
            MARMOT_GEN_SPLIT_T_N_A ("${t_n_a}" type arg)
            PRINT_BACK_CONVERT_FUNCTION (text "${type}" "${arg}")  
        ENDFOREACH (t_n_a)
        
        #also dont forget to do convert output array indices
        FOREACH (t_n_a ${args_n_types})
            MARMOT_GEN_SPLIT_T_N_A ("${t_n_a}" type arg)
            CONVERT_OUTPUT_ARRAY_INDICES (text "${type}" "${arg}")  
        ENDFOREACH (t_n_a)
        
        #Special handling for MPI_Init and MPI_Init_thread 
        IF (fn_name MATCHES "MPI_INIT[^A-Za-z0-9_]*$" OR
            fn_name MATCHES "MPI_INIT_THREAD[^A-Za-z0-9_]*$")
            FORTRAN_INIT_POST (text)
        ENDIF (fn_name MATCHES "MPI_INIT[^A-Za-z0-9_]*$" OR
               fn_name MATCHES "MPI_INIT_THREAD[^A-Za-z0-9_]*$")
       
        SET (text "${text}#ifdef MARMOT_USE_THREADSYNC\n")
        IF (e MATCHES "sourceinfo")
            SET (text "${text}      MPO_Space::getSyncObject()->leaveMARMOT(*line,cBuf,enter_succeeded);\n")
        ELSE (e MATCHES "sourceinfo")
            SET (text "${text}      MPO_Space::getSyncObject()->leaveMARMOT(0,\"\",enter_succeeded);\n")
        ENDIF (e MATCHES "sourceinfo")
        SET (text "${text}#endif\n")
        
        # now test whether we have to unregister requests
        FOREACH (t_n_a ${args_n_types})
            MARMOT_GEN_SPLIT_T_N_A ("${t_n_a}" type arg)
            PRINT_REQUEST_UNREGISTER (text "${type}" "${arg}" "${callname}" "${e}" "YES")  
        ENDFOREACH (t_n_a)
        
        SET (text "${text}}\n\n\n")
    ENDFOREACH (e)
        
    FILE (APPEND ${outfile} "${text}") #|${callname}|" ${args_n_types} "\n"

ENDMACRO (MARMOT_GEN_CREATE_WRAPPER outfile callname retval args_n_types)

#===================================
# Actual script 
#===================================
MESSAGE (STATUS "Building MPI wrappers for Fortran bindings")

IF (MPI_HANDLE_CONVERT_PRESENT)
    SET (HANDLE_CONVERT "YES")
ELSE (MPI_HANDLE_CONVERT_PRESENT)
    SET (HANDLE_CONVERT "NO")
ENDIF (MPI_HANDLE_CONVERT_PRESENT)

IF (MPI_REQUEST_CONVERT_PRESENT)
    SET (REQUEST_CONVERT "YES")
ELSE (MPI_REQUEST_CONVERT_PRESENT)
    SET (REQUEST_CONVERT "NO")
ENDIF (MPI_REQUEST_CONVERT_PRESENT)

IF (MPI_STATUS_CONVERT_PRESENT)
    SET (STATUS_CONVERT "YES")
ELSE (MPI_STATUS_CONVERT_PRESENT)
    SET (STATUS_CONVERT "NO")
ENDIF (MPI_STATUS_CONVERT_PRESENT)

IF (MPI_INFO_CONVERT_PRESENT)
    SET (INFO_CONVERT "YES")
ELSE (MPI_INFO_CONVERT_PRESENT)
    SET (INFO_CONVERT "NO")
ENDIF (MPI_INFO_CONVERT_PRESENT)

IF (MPI_WIN_CONVERT_PRESENT)
    SET (WIN_CONVERT "YES")
ELSE (MPI_WIN_CONVERT_PRESENT)
    SET (WIN_CONVERT "NO")
ENDIF (MPI_WIN_CONVERT_PRESENT)

IF (MPI_FILE_CONVERT_PRESENT)
    SET (FILE_CONVERT "YES")
ELSE (MPI_FILE_CONVERT_PRESENT)
    SET (FILE_CONVERT "NO")
ENDIF (MPI_FILE_CONVERT_PRESENT)

SET (INFILES "${PROJECT_SOURCE_DIR}/SRC/INCLUDE/mpi_standard.h")
SET (INFILES ${INFILES} "${PROJECT_SOURCE_DIR}/SRC/INCLUDE/mpi2_1sided.h")
SET (INFILES ${INFILES} "${PROJECT_SOURCE_DIR}/SRC/INCLUDE/mpi2_extcoll.h")
SET (INFILES ${INFILES} "${PROJECT_SOURCE_DIR}/SRC/INCLUDE/mpi2_file.h")
SET (INFILES ${INFILES} "${PROJECT_SOURCE_DIR}/SRC/INCLUDE/mpi2_standard.h")
SET (INFILES ${INFILES} "${PROJECT_SOURCE_DIR}/SRC/INCLUDE/mpi2_proc.h")

SET (OUTFILE "${PROJECT_BINARY_DIR}/generated/fortran_wrapper.cc")
CONFIGURE_FILE("${PROJECT_SOURCE_DIR}/CMAKEMODULES/generation/FortranHeader.cc.in" "${OUTFILE}" @ONLY)

SET (IGNOREFILE "${PROJECT_BINARY_DIR}/generated/ignore-list.txt")
CONFIGURE_FILE("${PROJECT_SOURCE_DIR}/CMAKEMODULES/generation/ignore-list.txt" "${IGNOREFILE}" COPYONLY)
HaveMPI_GetUnsupportedCallList(unsuported_list)
FOREACH (call ${unsuported_list})
    FILE (APPEND "${IGNOREFILE}" "\n${call}")
ENDFOREACH (call)

FOREACH (INFILE ${INFILES})
    MARMOT_GEN_PARSE_HEADER ("${INFILE}" "${OUTFILE}" "${IGNOREFILE}")
ENDFOREACH (INFILE ${INFILES})
