# @file GenerateMacros.cmake
# 
# Common Macros for generation. 
#
# @author Tobias Hilbrich
#

#===================================
# Macro: Parse header 
#===================================
#Detects MPI calls in given header and splits them
#into: <callname> {<arg_type> <arg_name>}*
MACRO (MARMOT_GEN_PARSE_HEADER infile outfile ignorefile)
  FILE (READ ${infile} CONTENT)
  IF (NOT "${ignorefile}" STREQUAL "")
      FILE (READ ${ignorefile} ignore)
  ELSE (NOT "${ignorefile}" STREQUAL "")
      SET (ignore "\n")
  ENDIF (NOT "${ignorefile}" STREQUAL "")
    
  STRING (REPLACE "\n" "<BREAK><BREAK>" CONTENT ${CONTENT})
  STRING (REPLACE "\n" "<b>" ignore ${ignore})
  STRING (REPLACE "[ ]*" "" ignore ${ignore})

  #Loop over all the MPI calls
  STRING (REGEX REPLACE "[*][/]" "°" CONTENT "${CONTENT}")
  STRING (REGEX REPLACE "[/][*][^°]*[°]" "" CONTENT "${CONTENT}")
  STRING (REGEX REPLACE "[/][/][^<>]*" "" CONTENT "${CONTENT}")
  
  STRING (REGEX MATCH "(^|<BREAK>)[a-zA-Z0-9]*[ ]MPI[^<>]*($|<BREAK>)" mpicall "${CONTENT}")
  
  WHILE (mpicall MATCHES "MPI" )
    #remove the found string from content
    STRING (REPLACE "${mpicall}" "" CONTENT ${CONTENT})
    
    #clean mpicall
    STRING (REPLACE "<BREAK>" "" mpicall ${mpicall})
    
    #Get the MPI call name
    STRING (REGEX REPLACE "^[a-zA-Z0-9]*[ ](MPI_[^( ]*).*" "\\1" mpicallname ${mpicall})
  
    #Filter out if in ignore list
    IF (NOT ignore MATCHES "(^|<b>)${mpicallname}($|<b>)")
      
        #Get return value
        STRING (REGEX MATCH "^[a-zA-Z0-9]*[ ]" retval ${mpicall})
        STRING (REGEX REPLACE "^[a-zA-Z0-9]*[ ]MPI_[^(]*" "" mpicall ${mpicall})
        
        #Handle arguments
        SET(args_n_types "")
        IF (NOT ${mpicall} MATCHES "[(][ ]*void[ ]*[)]")
            STRING (REGEX REPLACE "^[^()]*[(]" "" mpicall ${mpicall})
            STRING (REGEX REPLACE "[)][^()]*$" "" mpicall ${mpicall})
            
            STRING (REGEX MATCH "[^,]+([,]|$)" argtype "${mpicall}")
            WHILE (argtype MATCHES "[^ ]+")
                STRING (REPLACE "${argtype}" "" mpicall "${mpicall}")
                STRING (REPLACE "," "" argtype ${argtype})
                
                STRING (REGEX REPLACE "([^ ]+)[ ][^ ]+" "\\1" type "${argtype}")
                STRING (REGEX REPLACE "[^ ]+[ ]([^ ]+)" "\\1" arg "${argtype}")
                
                SET(args_n_types ${args_n_types} "${type}|${arg}")
                
                #next iteration
                SET (argtype "") 
                STRING (REGEX MATCH "[^,]+([,]|$)" argtype "${mpicall}")
            ENDWHILE (argtype MATCHES "[^ ]+")
        ENDIF (NOT ${mpicall} MATCHES "[(][ ]*void[ ]*[)]")
       
        #Final clean
        STRING (REPLACE " " "" retval "${retval}")
        STRING (REPLACE " " "" mpicallname "${mpicallname}")
        STRING (REPLACE " " "" args_n_types "${args_n_types}")
       
        #Call generation of wrapper
        MARMOT_GEN_CREATE_WRAPPER (${outfile} "${mpicallname}" "${retval}" args_n_types)
    ENDIF (NOT ignore MATCHES "(^|<b>)${mpicallname}($|<b>)")
 
    #next iteration 
    STRING (REGEX MATCH "(^|<BREAK>)[a-zA-Z0-9]*[ ]MPI[^<>]*($|<BREAK>)" mpicall "${CONTENT}")
  ENDWHILE (mpicall MATCHES "MPI" )
ENDMACRO (MARMOT_GEN_PARSE_HEADER infile outfile)

#===================================
# Macro: MARMOT_GEN_SPLIT_T_N_A 
#===================================
#Takes a "TYPE|ARGUMENT" string and splits it into 
#TYPE and ARGUMENT
MACRO (MARMOT_GEN_SPLIT_T_N_A t_n_a type arg)
    STRING (REGEX MATCH "^[^|]+" ${type} "${t_n_a}")
    STRING (REPLACE "${${type}}|" "" ${arg} "${t_n_a}")
ENDMACRO (MARMOT_GEN_SPLIT_T_N_A t_n_a type arg)

#===================================
# Macro: MARMOT_GEN_GET_LEN_PARAMETER
#===================================
#Takes a "<argname>_<array-arg-descriptor>_<len_arg>" string  
#and returns the <len_arg> part
MACRO (MARMOT_GEN_GET_LEN_PARAMETER arg out)
    #extract the name of the length parameter
    SET (splitstring "_CLASS_ARRAY_IN_")
    IF (${arg} MATCHES "_CLASS_ARRAY_IO_")
        SET (splitstring "_CLASS_ARRAY_IO_")
    ENDIF (${arg} MATCHES "_CLASS_ARRAY_IO_")
    IF (${arg} MATCHES "_CLASS_ARRAY_OUT_")
        SET (splitstring "_CLASS_ARRAY_OUT_")
    ENDIF (${arg} MATCHES "_CLASS_ARRAY_OUT_")
    IF (${arg} MATCHES "_CLASS_ARRAYINDEX_OUT_SINGLE_CONDITION_")
        SET (splitstring "_CLASS_ARRAYINDEX_OUT_SINGLE_CONDITION_")
    ENDIF (${arg} MATCHES "_CLASS_ARRAYINDEX_OUT_SINGLE_CONDITION_")    
    IF (${arg} MATCHES "_CLASS_ARRAYINDEX_OUT_ARRAY_")
        SET (splitstring "_CLASS_ARRAYINDEX_OUT_ARRAY_")
    ENDIF (${arg} MATCHES "_CLASS_ARRAYINDEX_OUT_ARRAY_")
            
    STRING (REGEX MATCH "${splitstring}(.*)" ${out} "${arg}")
    STRING (REPLACE "${splitstring}" ""${out} "${${out}}")
ENDMACRO (MARMOT_GEN_GET_LEN_PARAMETER arg out)

#===================================
# Macro: GET_CONVERTED_NAME
#===================================
MACRO (GET_CONVERTED_NAME arg type out fortran callname)
    SET (${out})
    
    IF ("${fortran}" MATCHES "YES")
        SET (pointer)
        #If it is not yet a pointer type or an array (group_rankge_excl uses [][]) 
        #than make it a pointer as Fortran uses call by reference 
        IF (NOT ${type} MATCHES "[*]" AND NOT ${arg} MATCHES "\\[.*\\]")
            SET (pointer "*")
        ENDIF (NOT ${type} MATCHES "[*]" AND NOT ${arg} MATCHES "\\[.*\\]")
    
        #Handle conversion needed ?
        IF (HANDLE_CONVERT MATCHES "NO")
            SET (${out} "${pointer}${arg}")
        ELSE (HANDLE_CONVERT MATCHES "NO")
            #Find out wheter we have to pass adress operator "&"
            SET (adress)
            IF (${arg} MATCHES "_CLASS_SINGLE_") 
                IF (NOT STATUS_CONVERT MATCHES "YES" OR NOT ${type} MATCHES "MPI_Status") # TODO ???
                    SET (adress "&")
                ENDIF (NOT STATUS_CONVERT MATCHES "YES" OR NOT ${type} MATCHES "MPI_Status")
            ENDIF (${arg} MATCHES "_CLASS_SINGLE_")
        
            #handle conversion
            SET (handles "MPI_Comm([^_]+|$)" "MPI_Datatype([^_]+|$)" "MPI_Group([^_]+|$)" "MPI_Op([^_]+|$)")
            FOREACH (handle ${handles}) 
                IF (${type} MATCHES ${handle})
                    SET (${out} "${adress}${arg}C")
                ENDIF (${type} MATCHES ${handle})
            ENDFOREACH (handle ${handles}) 
            
            IF (${type} MATCHES "char[*]([^*]+|$)")
                SET (${out} "${arg}C")
            ENDIF (${type} MATCHES "char[*]([^*]+|$)")
                    
            IF (${type} MATCHES "MPI_Request([^_]+|$)" AND REQUEST_CONVERT MATCHES "YES")
                SET (${out} "${adress}${arg}C")
            ENDIF (${type} MATCHES "MPI_Request([^_]+|$)" AND REQUEST_CONVERT MATCHES "YES")
            
            IF (${type} MATCHES "MPI_Status([^_]+|$)" AND STATUS_CONVERT MATCHES "YES")
                SET (${out} "${adress}${arg}C")
            ENDIF (${type} MATCHES "MPI_Status([^_]+|$)" AND STATUS_CONVERT MATCHES "YES")
            
            IF (${type} MATCHES "MPI_Info([^_]+|$)" AND INFO_CONVERT MATCHES "YES")
                SET (${out} "${adress}${arg}C")
            ENDIF (${type} MATCHES "MPI_Info([^_]+|$)" AND INFO_CONVERT MATCHES "YES")
            
            IF (${type} MATCHES "MPI_Win([^_]+|$)" AND WIN_CONVERT MATCHES "YES")
                SET (${out} "${adress}${arg}C")
            ENDIF (${type} MATCHES "MPI_Win([^_]+|$)" AND WIN_CONVERT MATCHES "YES")
            
            IF (${type} MATCHES "MPI_File([^_]+|$)" AND FILE_CONVERT MATCHES "YES")
                SET (${out} "${adress}${arg}C")
            ENDIF (${type} MATCHES "MPI_File([^_]+|$)" AND FILE_CONVERT MATCHES "YES")
        
            #it is not a handle type so normal handling
            IF (NOT ${out})
                 SET (${out} "${pointer}${arg}")
            ENDIF (NOT ${out})
        ENDIF (HANDLE_CONVERT MATCHES "NO")
        
        #Is it a function pointer for a Fortran callback function ?
        ##TODO: currently we replace with a Marmot dummy errhandler,
        ##      as we would have to map the Fortran callback to a C
        #       callback
        IF (${type} MATCHES "MPI_Handler_function")
            SET (${out} "marmot_comm_dummy_errhandler_function")
        ENDIF (${type} MATCHES "MPI_Handler_function")
        
        IF (${type} MATCHES "MPI_Comm_errhandler_fn")
            SET (${out} "marmot_comm_dummy_errhandler_function")
        ENDIF (${type} MATCHES "MPI_Comm_errhandler_fn")
        
        IF (${type} MATCHES "MPI_File_errhandler_fn")
            SET (${out} "marmot_file_dummy_errhandler_function")
        ENDIF (${type} MATCHES "MPI_File_errhandler_fn")
        
        IF (${type} MATCHES "MPI_Win_errhandler_fn")
            SET (${out} "marmot_win_dummy_errhandler_function")
        ENDIF (${type} MATCHES "MPI_Win_errhandler_fn")
        
        #Is it a MPI_Aint in a function where it is actually a MPI_Fint ?
        IF ("${callname}" MATCHES "MPI_Type_hindexed" OR
            "${callname}" MATCHES "MPI_Type_hvector" OR
            "${callname}" MATCHES "MPI_Type_struct" OR
            "${callname}" MATCHES "MPI_Address" OR 
            "${callname}" MATCHES "MPI_Type_extent" OR
            "${callname}" MATCHES "MPI_Type_lb" OR
            "${callname}" MATCHES "MPI_Type_ub")
            IF (type MATCHES "MPI_Aint[*]")
                IF (${arg} MATCHES "_CLASS_SINGLE")
                    SET (${out} "&${arg}C")
                ELSE (${arg} MATCHES "_CLASS_SINGLE")
                    SET (${out} "${arg}C")
                ENDIF (${arg} MATCHES "_CLASS_SINGLE")
            ELSEIF (type MATCHES "MPI_Aint")
                SET (${out} "${arg}C")
            ENDIF (type MATCHES "MPI_Aint[*]")
        ENDIF ("${callname}" MATCHES "MPI_Type_hindexed" OR
            "${callname}" MATCHES "MPI_Type_hvector" OR
            "${callname}" MATCHES "MPI_Type_struct" OR
            "${callname}" MATCHES "MPI_Address" OR 
            "${callname}" MATCHES "MPI_Type_extent" OR
            "${callname}" MATCHES "MPI_Type_lb" OR
            "${callname}" MATCHES "MPI_Type_ub")
        
    ELSE ("${fortran}" MATCHES "YES")
        SET (${out} "${arg}")
    ENDIF ("${fortran}" MATCHES "YES")
    
    #Remove any existing array brackets (group_range_incl)
    STRING (REGEX REPLACE "\\[.*\\]" "" ${out} "${${out}}")
ENDMACRO (GET_CONVERTED_NAME arg type out fortran callname)

#===================================
# Macro: PRINT_REQUEST_REGISTER
#===================================
MACRO (PRINT_REQUEST_REGISTER p_text type arg callname e fortran)
    #is it a request 
    IF (${type} MATCHES "MPI_Request")
        #is it a WAIT / TEST call ?
        STRING (TOUPPER "${callname}" upper)
        IF ( upper MATCHES "MPI_WAIT" OR
             upper MATCHES "MPI_TEST")
            SET (pointer "")
            IF (NOT ${type} MATCHES "[*]")
                SET (pointer "*")
            ENDIF (NOT ${type} MATCHES "[*]")
            
            #select source location parameters
            IF (${e} MATCHES "sourceinfo")
                IF ("${fortran}" MATCHES "YES")
                    SET (line "*line")
                    SET (file "cBuf")
                ENDIF ("${fortran}" MATCHES "YES")
                
                IF ("${fortran}" MATCHES "UNIMCI")
                    SET (line "unimci_line")
                    SET (file "unimci_file")
                ENDIF ("${fortran}" MATCHES "UNIMCI")
                
                IF ("${fortran}" MATCHES "NO")
                    SET (line "line")
                    SET (file "scfile")
                ENDIF ("${fortran}" MATCHES "NO")
            ELSE (${e} MATCHES "sourceinfo")
                SET (line "0")
                SET (file "\"\"")
            ENDIF (${e} MATCHES "sourceinfo")
                
            #we have to check in case of:
            # _CLASS_SINGLE_IO and
            # _CLASS_ARRAY_IO_ 
            # as those are used for the request wait/test functions.
            
            IF ( ${arg} MATCHES "_CLASS_SINGLE_IO")
                GET_CONVERTED_NAME ("${arg}" "${type}" convname "${fortran}" "${callname}")
                
                SET (${p_text} "${${p_text}}#ifdef MARMOT_USE_THREADSYNC\n")
                SET (${p_text} "${${p_text}}  int myMARMOTRequestRegistrationID = MPO_Space::getSyncObject()->registerRequest (${convname}, ${line}, ${file}, MPO_Space::getSyncObject()->getThreadRank(), temp.getId(), temp.getCallname());\n")
                SET (${p_text} "${${p_text}}#endif\n")
            ENDIF ( ${arg} MATCHES "_CLASS_SINGLE_IO")
                
            IF ( ${arg} MATCHES "_CLASS_ARRAY_IO_")
                #extract the name of the length parameter
                MARMOT_GEN_GET_LEN_PARAMETER ("${arg}" len_parameter)
                GET_CONVERTED_NAME ("${arg}" "${type}" convname "${fortran}" "${callname}")
                
                IF ("${fortran}" MATCHES "YES")
                    SET (len_parameter "*${len_parameter}")
                ENDIF ("${fortran}" MATCHES "YES")
            
                SET (${p_text} "${${p_text}}#ifdef MARMOT_USE_THREADSYNC\n")
                SET (${p_text} "${${p_text}}  int myMARMOTRequestRegistrationID = MPO_Space::getSyncObject()->registerRequests (${convname}, ${len_parameter}, ${line}, ${file}, MPO_Space::getSyncObject()->getThreadRank(), temp.getId(), temp.getCallname());\n")
                SET (${p_text} "${${p_text}}#endif\n")
            ENDIF ( ${arg} MATCHES "_CLASS_ARRAY_IO_")
        ENDIF ( upper MATCHES "MPI_WAIT" OR
                upper MATCHES "MPI_TEST")
    ENDIF (${type} MATCHES "MPI_Request")
ENDMACRO (PRINT_REQUEST_REGISTER p_text type arg callname e fortran)

#===================================
# Macro: PRINT_REQUEST_UNREGISTER
#===================================
MACRO (PRINT_REQUEST_UNREGISTER p_text type arg callname e fortran)
    #is it a request 
    IF (${type} MATCHES "MPI_Request")
        #is it a WAIT / TEST call ?
        STRING (TOUPPER "${callname}" upper)
        IF ( upper MATCHES "MPI_WAIT" OR
             upper MATCHES "MPI_TEST")
            SET (pointer "")
            IF (NOT ${type} MATCHES "[*]")
                SET (pointer "*")
            ENDIF (NOT ${type} MATCHES "[*]")
                
            #we have to check in case of:
            # _CLASS_SINGLE_IO and
            # _CLASS_ARRAY_IO_ 
            # as those are used for the request wait/test functions.
            
            IF (${arg} MATCHES "_CLASS_SINGLE_IO" OR
                ${arg} MATCHES "_CLASS_ARRAY_IO_")
                SET (${p_text} "${${p_text}}#ifdef MARMOT_USE_THREADSYNC\n")
                SET (${p_text} "${${p_text}}  MPO_Space::getSyncObject()->unregisterRequests(myMARMOTRequestRegistrationID);\n")
                SET (${p_text} "${${p_text}}#endif\n")
            ENDIF (${arg} MATCHES "_CLASS_SINGLE_IO" OR
                ${arg} MATCHES "_CLASS_ARRAY_IO_")
        ENDIF ( upper MATCHES "MPI_WAIT" OR
                upper MATCHES "MPI_TEST")
    ENDIF (${type} MATCHES "MPI_Request")
ENDMACRO (PRINT_REQUEST_UNREGISTER p_text type arg callname e fortran)

#===================================
# Macro: PRINT_COLLCOMM_REGISTER
#===================================
MACRO (PRINT_COLLCOMM_REGISTER p_text type arg callname e fortran)
    #is it a comm
    IF (${type} MATCHES "MPI_Comm")
            
        #is it a collective call of MPI 1.2:
        # * MPI_BARRIER
        # * MPI_BCAST
        # * MPI_GATHER, MPI_GATHERV
        # * MPI_SCATTER, MPI_SCATTERV
        # * MPI_ALLGATHER, MPI_ALLGATHERV
        # * MPI_ALLTOALL, MPI_ALLTOALLV
        # * MPI_REDUCE
        # * MPI_ALLREDUCE
        # * MPI_REDUCE_SCATTER
        # * MPI_SCAN
        STRING (TOUPPER "${callname}" upper)
        IF ( ${upper} MATCHES "MPI_BARRIER" OR
             ${upper} MATCHES "MPI_BCAST" OR
             ${upper} MATCHES "MPI_GATHER" OR
             ${upper} MATCHES "MPI_SCATTER" OR
             ${upper} MATCHES "MPI_ALLGATHER" OR
             ${upper} MATCHES "MPI_ALLTOALL" OR
             ${upper} MATCHES "MPI_REDUCE" OR
             ${upper} MATCHES "MPI_ALLREDUCE" OR
             ${upper} MATCHES "MPI_REDUCE_SCATTER" OR
             ${upper} MATCHES "MPI_SCAN")
            #select source location parameters
            IF (${e} MATCHES "sourceinfo")
                IF ("${fortran}" MATCHES "YES")
                    SET (line "*line")
                    SET (file "cBuf")
                ENDIF ("${fortran}" MATCHES "YES")
                
                IF ("${fortran}" MATCHES "UNIMCI")
                    SET (line "unimci_line")
                    SET (file "unimci_file")
                ENDIF ("${fortran}" MATCHES "UNIMCI")
                
                IF ("${fortran}" MATCHES "NO")
                    SET (line "line")
                    SET (file "scfile")
                ENDIF ("${fortran}" MATCHES "NO")
            ELSE (${e} MATCHES "sourceinfo")
                SET (line "0")
                SET (file "\"\"")
            ENDIF (${e} MATCHES "sourceinfo")
        
            #no parameter distinction checks necessary here !
            GET_CONVERTED_NAME ("${arg}" "${type}" convname "${fortran}" "${callname}")
        
            SET (${p_text} "${${p_text}}#ifdef MARMOT_USE_THREADSYNC\n")
            SET (${p_text} "${${p_text}}     int myMARMOTCommRegistrationID = MPO_Space::getSyncObject()->registerCollComm (${convname}, ${line}, ${file}, MPO_Space::getSyncObject()->getThreadRank(), temp.getId(), temp.getCallname());\n")
            SET (${p_text} "${${p_text}}#endif\n")
        ENDIF ( ${upper} MATCHES "MPI_BARRIER" OR
                ${upper} MATCHES "MPI_BCAST" OR
                ${upper} MATCHES "MPI_GATHER" OR
                ${upper} MATCHES "MPI_SCATTER" OR
                ${upper} MATCHES "MPI_ALLGATHER" OR
                ${upper} MATCHES "MPI_ALLTOALL" OR
                ${upper} MATCHES "MPI_REDUCE" OR
                ${upper} MATCHES "MPI_ALLREDUCE" OR
                ${upper} MATCHES "MPI_REDUCE_SCATTER" OR
                ${upper} MATCHES "MPI_SCAN")
    ENDIF (${type} MATCHES "MPI_Comm") 
ENDMACRO (PRINT_COLLCOMM_REGISTER p_text type arg callname e fortran)

#===================================
# Macro: PRINT_COLLCOMM_UNREGISTER
#===================================
MACRO (PRINT_COLLCOMM_UNREGISTER p_text type arg callname e fortran)
    #is it a comm
    IF (${type} MATCHES "MPI_Comm")
            
        #is it a collective call of MPI 1.2:
        # * MPI_BARRIER
        # * MPI_BCAST
        # * MPI_GATHER, MPI_GATHERV
        # * MPI_SCATTER, MPI_SCATTERV
        # * MPI_ALLGATHER, MPI_ALLGATHERV
        # * MPI_ALLTOALL, MPI_ALLTOALLV
        # * MPI_REDUCE
        # * MPI_ALLREDUCE
        # * MPI_REDUCE_SCATTER
        # * MPI_SCAN
        STRING (TOUPPER "${callname}" upper)
        IF ( ${upper} MATCHES "MPI_BARRIER" OR
             ${upper} MATCHES "MPI_BCAST" OR
             ${upper} MATCHES "MPI_GATHER" OR
             ${upper} MATCHES "MPI_SCATTER" OR
             ${upper} MATCHES "MPI_ALLGATHER" OR
             ${upper} MATCHES "MPI_ALLTOALL" OR
             ${upper} MATCHES "MPI_REDUCE" OR
             ${upper} MATCHES "MPI_ALLREDUCE" OR
             ${upper} MATCHES "MPI_REDUCE_SCATTER" OR
             ${upper} MATCHES "MPI_SCAN")
        
            #no parameter distinction checks necessary here !
            
            SET (${p_text} "${${p_text}}#ifdef MARMOT_USE_THREADSYNC\n")
            SET (${p_text} "${${p_text}}     MPO_Space::getSyncObject()->unregisterCollComm(myMARMOTCommRegistrationID);\n")
            SET (${p_text} "${${p_text}}#endif\n")
        ENDIF ( ${upper} MATCHES "MPI_BARRIER" OR
                ${upper} MATCHES "MPI_BCAST" OR
                ${upper} MATCHES "MPI_GATHER" OR
                ${upper} MATCHES "MPI_SCATTER" OR
                ${upper} MATCHES "MPI_ALLGATHER" OR
                ${upper} MATCHES "MPI_ALLTOALL" OR
                ${upper} MATCHES "MPI_REDUCE" OR
                ${upper} MATCHES "MPI_ALLREDUCE" OR
                ${upper} MATCHES "MPI_REDUCE_SCATTER" OR
                ${upper} MATCHES "MPI_SCAN")
    ENDIF (${type} MATCHES "MPI_Comm") 
ENDMACRO (PRINT_COLLCOMM_UNREGISTER p_text type arg callname e fortran)

#===================================
# Macro: MARMOT_GEN_IS_IMPLEMENTED
# Finds out whether the given MPI call is 
# implemented by Marmot.
#===================================
MACRO (MARMOT_GEN_IS_IMPLEMENTED callname implemented)
    STRING(REPLACE "MPI_" "" callnompi "${callname}")
    FILE (GLOB mpo_implems1 "${PROJECT_SOURCE_DIR}/SRC/SRC/mpo_${callnompi}.cc")
    FILE (GLOB mpo_implems2 "${PROJECT_SOURCE_DIR}/SRC/SRC-1SIDED/mpo_${callnompi}.cc")
    FILE (GLOB mpo_implems3 "${PROJECT_SOURCE_DIR}/SRC/SRC-2-MISC/mpo_${callnompi}.cc")
    #FILE (GLOB mpo_implems4 "${PROJECT_SOURCE_DIR}/SRC/SRC-IO/mpo_${callnompi}.cc") ##TODO add when templates in SRC-IO get content !!!
    SET (mpo_implems ${mpo_implems1} ${mpo_implems2} ${mpo_implems3} ${mpo_implems4}) 
    
    IF (NOT mpo_implems)
        SET (${implemented} FALSE)
    ELSE (NOT mpo_implems)
        SET (${implemented} TRUE)
    ENDIF (NOT mpo_implems)
ENDMACRO (MARMOT_GEN_IS_IMPLEMENTED callname implemented)