# @file GenerateUniMCIWrapper.cmake
# 
# Generats the UniMCI wrappers for Marmot. 
#
# Based on mkunimcibindings.sh.in (r981), which had the following authors:
#    * Katrin Bidmon
#    * Bettina Krammer
#    * Matthias Mueller
#    * Tobias Hilbrich
#
# Uses the following other CMake variables that have to be defined before
# calling this script:
#    * MPI_C_HEADER   --- full path to the MPI C header
#
# @author Tobias Hilbrich
#

#===================================
# Macro: MARMOT_GEN_CREATE_WRAPPER
#===================================
#This is a hook name, as the parser calls
#a macro of that name.
MACRO (MARMOT_GEN_CREATE_WRAPPER outfile callname retval p_args_n_types)
    SET (args_n_types ${${p_args_n_types}})
    
    #Split the outfiles
    STRING (REGEX MATCH "^[^|]*" source_out "${outfile}")
    STRING (REGEX MATCH "[|][^|]*" header_out "${outfile}")
    STRING (REPLACE "|" "" header_out "${header_out}")
        
    #Text for the wrapper function
    SET (text "")
    
    #Is it implemented in Marmot ?
    MARMOT_GEN_IS_IMPLEMENTED ("${callname}" implemented)
    
 #This loop creates two different versions of the calls one with aditional 
    #parameters for line and file name and one without
    FOREACH(e "pre" "post")
        #Add a comment
        SET (text "${text}\n/* ${callname}* implemented */\n")
     
        # Start with function name of wrapper
        IF (e MATCHES "pre")
            STRING (REGEX REPLACE "MPI_" "MARMOT_check_pre__MPI_" unimci_name ${callname})
            SET (text "${text}extern \"C\" void ${unimci_name} (")
        ELSE (e MATCHES "pre")
            STRING (REGEX REPLACE "MPI_" "MARMOT_check_post__MPI_" unimci_name ${callname})
            SET (text "${text}extern \"C\" void ${unimci_name} (")
        ENDIF (e MATCHES "pre")
          
        SET (mpocall ${callname})
        
        SET (type)
        IF (args_n_types)
            #For calls with arguments, loop over them and print
            FOREACH (t_n_a ${args_n_types})
                IF (type)
                  SET (text "${text}, ")
                ENDIF (type)

                MARMOT_GEN_SPLIT_T_N_A ("${t_n_a}" type arg)
                
                #print argument and type
                SET (text "${text}${type} ${arg}")                
            ENDFOREACH (t_n_a)
        ENDIF (args_n_types)
            
        IF (type)
            SET (text "${text}, const char* unimci_file, unsigned int unimci_line,unsigned long long unimci_eventDescriptor")
        ELSE (type)
            SET (text "${text} const char* unimci_file, unsigned int unimci_line,unsigned long long unimci_eventDescriptor")
        ENDIF (type)
                    
        #Finish the function line                
        SET (text "${text})\n")
        
        #we are done with the prototype, now we start with the real function:
        SET (text "${text}{\n")
        
        STRING (REPLACE "MPI_" "MPO_" mpocall ${mpocall})
        
        #We use this one in order to get the mpo id and callname of an mpo object.
        #Otherwise we would have to create the object for this call before the
        #enterMARMOT call which causes problems with non threadsafe implementations
        #of memory allocation.
        IF (implemented)
            SET (text "${text}   static ${mpocall} temp;\n")
        ELSE (implemented)
            #create some temp object, not of much interest which one
            SET (text "${text}   static MPO_Comm_rank temp;\n")
        ENDIF (implemented)        
        
        IF (e MATCHES "pre")
            # --- PRE ---
            # now test whether we need registration of requests
            FOREACH (t_n_a ${args_n_types})
                MARMOT_GEN_SPLIT_T_N_A ("${t_n_a}" type arg)
                PRINT_REQUEST_REGISTER (text "${type}" "${arg}" "${callname}" "sourceinfo" "UNIMCI")  
            ENDFOREACH (t_n_a)
        
            # now test whether we need registration of collective comms
            FOREACH (t_n_a ${args_n_types})
                MARMOT_GEN_SPLIT_T_N_A ("${t_n_a}" type arg)
                PRINT_COLLCOMM_REGISTER (text "${type}" "${arg}" "${callname}" "sourceinfo" "UNIMCI")  
            ENDFOREACH (t_n_a)
            
            #synchronisation    
            SET (text "${text}#ifdef MARMOT_USE_THREADSYNC\n")
            SET (text "${text}   bool enter_succeeded = MPO_Space::getSyncObject()->enterMARMOT(unimci_line,unimci_file);\n")
            SET (text "${text}#endif\n")
          
            #this if is necessary for MPI implems like MPICH that perform MPI calls during MPI_Init
            IF ("${callname}" MATCHES "MPI_Init[^A-Za-z0-9_]*$" OR
                "${callname}" MATCHES "MPI_Init_thread[^A-Za-z0-9_]*$")
          	    SET (text "${text}   MPO_in_profile=1;\n")
                 
                #Special for UniMCI bindings:
                # * disable the debug server
                # * force the queue logger
                # * disable MPO_Init constructor to call PMPI_Init[_thread]
                SET (text "${text}   MPO_Space::disableDebugServer ();\n")
                SET (text "${text}   MPO_Space::forceLogType (LOGFILE_TYPE_QUEUE);\n")
                SET (text "${text}   MPO_Init::doNotCallInit ();\n")      
                SET (text "${text}   MPO_Callobject::SetPreDataRestore (true);\n")
            ENDIF ("${callname}" MATCHES "MPI_Init[^A-Za-z0-9_]*$" OR
                   "${callname}" MATCHES "MPI_Init_thread[^A-Za-z0-9_]*$")
        ELSE (e MATCHES "pre")
            # --- POST ---
            SET (text "${text}bool enter_succeeded = MPO_Space::getSyncObject()->didEnterSucceed();\n")
            SET (text "${text}if (!enter_succeeded)\n")
            SET (text "${text}{\n")
            SET (text "${text}#ifdef MARMOT_USE_THREADSYNC\n")
            SET (text "${text}   MPO_Space::getSyncObject()->leaveMARMOT(unimci_line,unimci_file,enter_succeeded);\n")
            SET (text "${text}#endif\n")
            SET (text "${text}   MPO_in_profile--;\n")
            SET (text "${text}   return;\n")
            SET (text "${text}}\n")
            SET (text "${text}\n")
            SET (text "${text}#ifdef MARMOT_USE_THREADSYNC\n")
            SET (text "${text}   MPO_Space::getSyncObject()->leavePMPI();\n")
            SET (text "${text}#endif\n")         
        ENDIF (e MATCHES "pre")
  
        IF (implemented)
            SET (text "${text}   ${mpocall} my_obj")
            
            IF (args_n_types)
              SET (text "${text}(")
              SET (type)
              FOREACH (t_n_a ${args_n_types})
                IF (type)
                    SET (text "${text}, ")
                ENDIF (type)
                MARMOT_GEN_SPLIT_T_N_A ("${t_n_a}" type arg)
                
                #Remove any existing array brackets (group_range_incl)
                STRING (REGEX REPLACE "\\[.*\\]" "" conv_name "${arg}")
                SET (text "${text}${conv_name}")
              ENDFOREACH (t_n_a)
              SET (text "${text})")
            ENDIF (args_n_types)
            SET (text "${text};\n")
        ELSE (implemented)
            IF (e MATCHES "pre")
                SET (text "${text}#ifdef MARMOT_USE_THREADSYNC\n")
                SET (text "${text}   if(enter_succeeded )\n")
                SET (text "${text}#else\n")
                SET (text "${text}   if(! MPO_in_profile )\n")
                SET (text "${text}#endif\n")
                SET (text "${text}   {\n")
                SET (text "${text}       static bool ${mpocall}_${e}_not_implemented_warned = false;\n")
                SET (text "${text}       if (!${mpocall}_${e}_not_implemented_warned)\n")
                SET (text "${text}       {\n")
                SET (text "${text}           int my_rank;\n")
                SET (text "${text}           PMPI_Comm_rank (MPI_COMM_WORLD,&my_rank);\n")
                SET (text "${text}           std::string text = \"WARNING: Application issued an MPI call not implemented by Marmot, false positives and non-reported errors are possible. (Call: ${callname})\";\n") 
                SET (text "${text}           MPO_MsgObject msg (MSG_TYPE_WARNING, unimci_eventDescriptor, my_rank, text, MARMOT_ERR_MARMOT);\n")
                SET (text "${text}           MPO_Space::getLogFile()->getMsg(msg);\n")
                SET (text "${text}           ${mpocall}_${e}_not_implemented_warned = true;\n")
                SET (text "${text}       }\n")
                SET (text "${text}   }\n")
            ENDIF (e MATCHES "pre")
        ENDIF (implemented)
        
        #this if is necessary for MPI implems like MPICH that perform MPI calls during MPI_Init
        IF (e MATCHES "pre")
            IF ("${callname}" MATCHES "MPI_Init[^A-Za-z0-9_]*$" OR
                "${callname}" MATCHES "MPI_Init_thread[^A-Za-z0-9_]*$")
          	    SET (text "${text}   MPO_in_profile=0;\n")
            ENDIF ("${callname}" MATCHES "MPI_Init[^A-Za-z0-9_]*$" OR
                   "${callname}" MATCHES "MPI_Init_thread[^A-Za-z0-9_]*$")
        ENDIF (e MATCHES "pre")    
        
        IF (implemented)
            SET (text "${text}   my_obj.setSourceLocation(unimci_line,unimci_file,unimci_eventDescriptor);\n")
        ENDIF (implemented)

        IF (e MATCHES "pre")
            # --- PRE ---
            SET (text "${text}#ifdef MARMOT_USE_THREADSYNC\n")
            SET (text "${text}   if(enter_succeeded )\n")
            SET (text "${text}#else\n")
            SET (text "${text}   if(! MPO_in_profile )\n")
            SET (text "${text}#endif\n")
            SET (text "${text}   {\n")
            SET (text "${text}      MPO_in_profile++;\n")
            SET (text "${text}\n")
            
            IF (implemented)
                SET (text "${text}      my_obj.checkPrerequisites();\n")
            ENDIF (implemented)
            
            SET (text "${text}\n")
            SET (text "${text}#ifdef MARMOT_USE_THREADSYNC\n")
            SET (text "${text}      MPO_Space::getSyncObject()->enterPMPI();\n")
            SET (text "${text}#endif\n")
            SET (text "${text}   }\n")
            SET (text "${text}   else\n")
            SET (text "${text}   {\n")
            SET (text "${text}       MPO_in_profile++;\n")
            SET (text "${text}   }\n")
        ELSE (e MATCHES "pre")
            # --- POST ---
            IF (implemented)
                SET (text "${text}   my_obj.checkPost();\n")
            ENDIF (implemented)
            SET (text "${text}\n")
            SET (text "${text}   MPO_in_profile--;\n")
            SET (text "${text}#ifdef MARMOT_USE_THREADSYNC\n")
            SET (text "${text}   MPO_Space::getSyncObject()->leaveMARMOT(unimci_line,unimci_file,enter_succeeded);\n")
            SET (text "${text}#endif\n")
      
            # now test whether we need registration of collective comms
            #TODO change registration mechanism
            #Currently a fix that allows compilation, but breaks the check
            SET (text "${text}int myMARMOTCommRegistrationID = 0;\n")
            FOREACH (t_n_a ${args_n_types})
                MARMOT_GEN_SPLIT_T_N_A ("${t_n_a}" type arg)
                PRINT_COLLCOMM_UNREGISTER (text "${type}" "${arg}" "${callname}" "sourceinfo" "UNIMCI")  
            ENDFOREACH (t_n_a)
  
            # now test whether we have to unregister requests
            #TODO change registration mechanism
            #Currently a fix that allows compilation, but breaks the check
            SET (text "${text}int myMARMOTRequestRegistrationID = 0;\n")
            FOREACH (t_n_a ${args_n_types})
                MARMOT_GEN_SPLIT_T_N_A ("${t_n_a}" type arg)
                PRINT_REQUEST_UNREGISTER (text "${type}" "${arg}" "${callname}" "sourceinfo" "UNIMCI")  
            ENDFOREACH (t_n_a)
        ENDIF (e MATCHES "pre")
   
        SET (text "${text}}\n\n\n")
        
    ENDFOREACH (e)
    
    FILE (APPEND ${source_out} "${text}") #|${callname}|" ${args_n_types} "\n"
    
    #Second step create the header for the UniMCI calls
    SET (text "")
    
    #This loop is used to create one PRE and one POST prototype per call
    FOREACH(e "pre" "post")
        #Print name and return value
        # Start with function name of wrapper
        IF (e MATCHES "pre")
            STRING (REGEX REPLACE "MPI_" "MARMOT_check_pre__MPI_" unimci_name ${callname})
            SET (text "${text}void ${unimci_name} (")
        ELSE (e MATCHES "pre")
            STRING (REGEX REPLACE "MPI_" "MARMOT_check_post__MPI_" unimci_name ${callname})
            SET (text "${text}void ${unimci_name} (")
        ENDIF (e MATCHES "pre")
          
        SET (type)
        IF (args_n_types)
            #For calls with arguments, loop over them and print
            FOREACH (t_n_a ${args_n_types})
                IF (type)
                  SET (text "${text}, ")
                ENDIF (type)

                MARMOT_GEN_SPLIT_T_N_A ("${t_n_a}" type arg)
                
                #print argument and type
                SET (text "${text}${type} ${arg}")                
            ENDFOREACH (t_n_a)
        ENDIF (args_n_types)
            
        IF (type)
            SET (text "${text}, const char* unimci_file, unsigned int unimci_line,unsigned long long unimci_eventDescriptor")
        ELSE (type)
            SET (text "${text} const char* unimci_file, unsigned int unimci_line,unsigned long long unimci_eventDescriptor")
        ENDIF (type)
                    
        #Finish the function line                
        SET (text "${text});\n")
    ENDFOREACH(e "pre" "post")
    
    FILE (APPEND ${header_out} "${text}") #|${callname}|" ${args_n_types} "\n"        
ENDMACRO (MARMOT_GEN_CREATE_WRAPPER outfile callname retval p_args_n_types)

#===================================
# Actual script 
#===================================
MESSAGE (STATUS "Building bindings for UniMCI")

SET (UNIMCI_HEADER "${CMAKE_BINARY_DIR}/SRC/INCLUDE/unimcicalls.h")
SET (OUTFILE "${PROJECT_BINARY_DIR}/generated/unimci_wrapper.cc|${UNIMCI_HEADER}")
CONFIGURE_FILE("${PROJECT_SOURCE_DIR}/CMAKEMODULES/generation/UniMCIHeader.cc.in" "${PROJECT_BINARY_DIR}/generated/unimci_wrapper.cc" @ONLY)
CONFIGURE_FILE("${PROJECT_SOURCE_DIR}/CMAKEMODULES/generation/UniMCIHeader.h.in" "${UNIMCI_HEADER}" @ONLY)

SET (IGNOREFILE "${PROJECT_BINARY_DIR}/generated/ignore-list.txt")
CONFIGURE_FILE("${PROJECT_SOURCE_DIR}/CMAKEMODULES/generation/ignore-list.txt" "${IGNOREFILE}" COPYONLY)
HaveMPI_GetUnsupportedCallList(unsuported_list)
FOREACH (call ${unsuported_list})
    FILE (APPEND "${IGNOREFILE}" "\n${call}")
ENDFOREACH (call)

SET (INFILES "${PROJECT_SOURCE_DIR}/SRC/INCLUDE/mpi_standard.h")
SET (INFILES ${INFILES} "${PROJECT_SOURCE_DIR}/SRC/INCLUDE/mpi2_1sided.h")
SET (INFILES ${INFILES} "${PROJECT_SOURCE_DIR}/SRC/INCLUDE/mpi2_extcoll.h")
SET (INFILES ${INFILES} "${PROJECT_SOURCE_DIR}/SRC/INCLUDE/mpi2_file.h")
SET (INFILES ${INFILES} "${PROJECT_SOURCE_DIR}/SRC/INCLUDE/mpi2_standard.h")
SET (INFILES ${INFILES} "${PROJECT_SOURCE_DIR}/SRC/INCLUDE/mpi2_proc.h")

FOREACH (INFILE ${INFILES})
    MARMOT_GEN_PARSE_HEADER ("${INFILE}" "${OUTFILE}" "${IGNOREFILE}")
ENDFOREACH (INFILE)

FILE (APPEND "${UNIMCI_HEADER}"
    "#if defined(__cplusplus)\n"
    "}\n"
    "#endif\n\n")
