/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
  * \file MPO_IThreadSynchronisation.h
  * This Headerfile describes the Interface used for thread synchronisation.
  * @author Tobias Hilbrich
  * @date 01.06.2007
  * $ID$
  */

#ifndef MPO_ITHREADSYNCHRONISATION_H
#define MPO_ITHREADSYNCHRONISATION_H

#include <string>
#include <mpi.h>
#include "mpi_undef_makros.h"
#include "mpo.h"
#include "mpo_ThreadViolationInfo.h"
#include "mpo_SourceLocation.h"
#include "config.h"

namespace MPO
{
    /**
     * We use a point on this object here.
     */
    class MPO_Callobject;

    /**
      * This Interface is used to synchronize MARMOT in Multithreaded environments.
      * This is an abstract class, implementations have to implement the abstract
      * methods. Possibly usefull implementations would be:
      *     + an implementation using OpenMP
      *     + an implementation using pthreds
      *     + a dummy implementation in case no synchronisation is necessary
      *       (i.e. only one thread of execution or guaranted MPI_THREAD_SERIALIZED)
      */
   class MPO_IThreadSynchronisation
    {
    public:

        /**
         * This method is used when a thread enters the marmot wrappers
         * in order to guarantie that only one thread is using marmots global
         * data at any given time.
         * This method also has to check for violations to the used level of thread
         * support.
         * @param line this value is used to get information on the MPI_Call made
         *               in case of an detected error.
         * @param file this value is used to get information on the MPI_Call made
         *               in case of an detected error.
         * @return returns true if this thread is not already in MARMOT and false otherwise.
         *         Note: this might happen if the Implementation of an MPI call uses other MPI calls.
         */
        virtual bool enterMARMOT (int line,const char * file) = 0;

        /**
         * This method is used when a thread leaves the marmot wrappers to
         * guarantie that only one thread is using marmots global data at
         * any given time.
         * @param line this value is used to get information on the MPI_Call made
         *               in case of an detected error.
         * @param file this value is used to get information on the MPI_Call made
         *               in case of an detected error.
         * @param enter_succeeded value returned by enterMARMOT.
         */
        virtual void leaveMARMOT (int line,const char * file,bool enter_succeeded) = 0;

        /**
         * This method is used after a thread made its PMPI call in order to
         * guarantie that only one thread is using marmots global data at any
         * given time.
         */
        virtual void enterPMPI () = 0;

        /**
         * This method is used before a thread makes the PMPI call to
         * guarantie that only one thread is using marmots global data
         * at any given time.
         */
        virtual void leavePMPI () = 0;

        /**
         * This method is usefull for determining whether thread parallelisation is
         * activated or not.
         * i.e. whether we are in a omp parallel region or not.
         * @return true if execution is in a parallel region, false otherwise.
         */
        virtual bool isInParallelRegion (void) = 0;

        /**
         * This method is used to return the number/id of the thread actually executing.
         * @return the id of the thread in execution or 0 if no parallel region is present.
         */
        virtual int getThreadRank (void) = 0;

        /**
         * Swichtes wrapping of MPI calls on for this
         * thread.
         * Must not be called from while inside Marmot
         */
        virtual void wrappingOn (void) = 0;

        /**
         * Swichtes wrapping of MPI calls off for this
         * thread.
         * Must not be called from while inside Marmot.
         */
        virtual void wrappingOff (void) = 0;

        /**
         * Returns whether the enterMARMOT succeeded.
         * Must only be called between Enter- and Leave-
         * MARMOT.
         */
        virtual bool didEnterSucceed (void) = 0;

        /**
         * This method returns the name of the used implementation of this interface.
         * Used for logging data about the used thread synchronisation.
         * @return the name of the implementation used.
         */
        virtual std::string getImplementationName (void) = 0;

        /**
         * This method returns true if another thread is waiting at enterMARMOT.
         * @return true if another thread is waiting at enterMARMOT, false otherwise.
         */
        virtual bool otherThreadsWaiting (void) = 0;

        /**
         * Registers the given request as currently used.
         * Checks for conflicts with another MPI-Call using the same requests
         * and reports an error in this case.
         * @param request the request to be registered.
         * @param sourceLine the line number in the source file.
         * @param sourceName the name of the source file.
         * @return an id used to unregister the requests after usage.
         */
        virtual int registerRequest (MPI_Request *request, int sourceLine,const char *sourceName, int threadNum, int callId, std::string callName) = 0;

        /**
         * Registers an array of Requests as currently used.
         * Checks for conflicts with another MPI-Call using one of the
         * given requests and reports an error in this case.
         * @param requests the request to be registered.
         * @param count the amount of requests in the given array.
         * @param location the sourcecode location of the call using these requests.
         * @return an id used to unregister the requests after usage.
         */
        virtual int registerRequests (MPI_Request *requests, int count, int sourceLine,const char *sourceName, int threadNum, int callId, std::string callName) = 0;

        /**
         * Removes the Requests associated with the given id from the currently
         * used list.
         * @param id the id of the requests to be unregistered.
         */
        virtual void unregisterRequests (int id) = 0;

#ifdef MARMOT_TCHECK
        /**
         * Used to remove a request registered in the request to int conversion for
         * TCheck conflict Variables.
         * Before calling make sure you are not within MARMOT i.e. by calling enterPMPI
         * and leavePMPI after the call !
         */
        virtual void unregisterRequestToIntEntry(MPI_Request *request) = 0;

        /**
         * Used to remove a comm registered in the request to int conversion for
         * TCheck conflict Variables.
         * Before calling make sure you are not within MARMOT i.e. by calling enterPMPI
         * and leavePMPI after the call !
         */
        virtual void unregisterCommToIntEntry(MPI_Comm comm) = 0;
#endif

        /**
         * Registers the given communicator of a collective call in an internal list
         * in order to detect the usage of one communicator in multiple collective
         * calls simultanously. If the given communicator is already used a error is
         * printed.
         * @param comm the communicator to be registered.
         * @param sourceLine the line number in the source file.
         * @param sourceName the name of the source file.
         * @param threadNum the id of the calling thread.
         * @param callId the call id of the associated call.
         * @param callName the name of the associated call.
         * @return an id used to unregister the communicator after usage.
         */
        virtual int registerCollComm (MPI_Comm comm, int sourceLine,const char *sourceName, int threadNum, int callId, std::string callName) = 0;

        /**
         * Removes the Communicator associated with the given id from the
         * list of currently used communicators.
         * @param id the id of the communicator to be unregistered.
         */
        virtual void unregisterCollComm (int id) = 0;

        /**
         * Returns the minimal level of thread support that was needed by this
         * application and run.
         * @return the minimal level of thread support neded for this application and run.
         */
        virtual int getMinThreadLvl (void) = 0;

        /**
         * Returns a MPO_ThreadViolationInfo for the violated thread level.
         * @param threadLevel one of MPI_THREAD_SINGLE, MPI_THREAD_FUNNELED,
         *                       MPI_THREAD_SERIALIZED where
         *                       threadLevel < getMinThreadLvl ()
         * @return if threadLevel < getMinThreadLvl () the thread violation
         *            info of the given thread level otherwise empty info
         *         (0 participants).
         */
        virtual MPO_ThreadViolationInfo getViolationInfoForThreadLvl (int threadLevel) = 0;

    };//end of class IThreadSynchronisation

}//end of namespace MPO

#endif







