/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 * @file
 * This Headerfile describes the MPO_LogFile interface.
 * This interface recieves MPO_MsgObject objects and creates a logfile from their data.
 * @author Tobias Hilbrich
 * @date 12.10.2005
 * $Id: MPO_LogFile.h 886 2008-11-10 15:21:03Z hpczink $ 
 */

#ifndef MPO_LOGFILE_H
#define MPO_LOGFILE_H

#include "mpo.h"
#include "mpo-enums.h"
#include "mpo_MsgObject.h"
#include "config.h"
#include <string>

class CTcpSocket;

using std::string;

namespace MPO 
{
    /**
    * @class   CSocketLogger
    * @brief   Socket logger singleton
    * If the environment variable MARMOT_LOG_HOST is set try to connect to MARMOT_LOG_HOST:30777
    * (MARMOT_LOG_PORT overrides the port) 
    * 
    * @author  Blasius Czink
    * @date    07.11.2008
    * @version 0.10
    * @since   0.10
    */
    class CSocketLogger
    {
      public:
        ~CSocketLogger();
        //! Get the instance of the singleton, create one if none exists yet
        static CSocketLogger *instance(void);
        //! Destroys the instance
        static void destroyInstance(void);
        //! Returns the socket
        CTcpSocket *getSocket() { return m_socket; };
        //! Wait for acknowledge-message from log-host
        bool waitForAck(void);

      private:
        // No standard and copy-constr.
        CSocketLogger();
        CSocketLogger(const CSocketLogger &) {};
        // No operator=
        CSocketLogger& operator=(const CSocketLogger &) { return *this; };
        static CSocketLogger *m_instance;

        std::string m_desthost;
        int         m_destport;
        CTcpSocket *m_socket;
        bool        m_do_ack;
    };

    /**
     * This Interface recieves MPO_MsgObject objects.
     * It defines how to receive them and nothing more. So each implementation 
     * of this interface can use the information in the MPO_MsgObject objects as
     * it wants.
     * 
     * Important: 	Each implementation of MPO_LogFile has to take care to
     * 				handle the global variable LOGFLUSH_TYPE correctly!
     * 
     */
    class MPO_LogFile
    {
    private:

        //==========================
        // Data
        //==========================
        /** \todo at Tobias from Tobias: Create Object for MPI-Standard referenciation and uncomment this line !*/
        int m_iDocPathLength;
        int m_iRelativPathLength;
        std::string m_strDocPath;
        std::string m_strRelPath;
        std::string m_strLocPath;
        
    public:
        //==========================
        // Constructors/Destructors
        //==========================
        MPO_LogFile();
        virtual ~MPO_LogFile (); /**< Standard destructor. */
         
        //==========================
        // Methods own
        //==========================
    public:	
        //! Initializes the MPI-Standardreferencer.
        bool initMPIStandardReferencer (void);
        //! Initializes connection to a MARMOT_LOG_HOST (if the environment variable is set and the given host is reachable)
        bool initSocket();
        //! Returns the initialized socket (use it with MPO_MsgObject::writeToSocket()) or NULL 
        CTcpSocket *getLogSocket() { return CSocketLogger::instance()->getSocket(); }
        //! Gets a message object, looks for MPI-Standard reference text and calls logMsg.
        virtual bool getMsg (MPO_MsgObject &pMsg) = 0;
        //! Calles to close the logfile and write it to disk.
        virtual bool closeLog (void) = 0;
        //! Returns a link to a HTML file that points to the MPI-Standard and describes the given ERROR-CODE, the html file is in the local doc directory.
        const char *getLocalPathToMPIStandard(MARMOT_ERR_CODES err);
        //! Returns a link to a HTML file that points to the MPI-Standard and describes the given ERROR-CODE, the string describes a address relativ to the doc directory in order to enable mirroring.
        const char *getRelativPathToMPIStandard (MARMOT_ERR_CODES err);
        //! Returns the local path to the doc directory.
        const char *getLocalDocPath();
    };//end of class MPO_LogFile

}//end of namespace MPO

#endif
