/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 * @file
 * This file declares and implements the MPO_ResourceGathererFull class.
 * @author Tobias Hilbrich
 * @date 28.11.2006
 * $Id: $
 */

#ifndef MPO_RESOURCEGATHERERFULL_H
#define MPO_RESOURCEGATHERERFULL_H

#include "MPO_ResourceGatherer.h"
#include "MPO_ResourceTrace.h"

namespace MPO
{
    /**
     * This class manages all the traces for a certain Resource type
     * like MPO_Datatype. It just maps the handle of these resources
     * to an object containing all the trace information.
     */
    template < class T_RESOURCE_TYPE > class MPO_ResourceGathererFull : public MPO_ResourceGatherer<T_RESOURCE_TYPE>
    {
    protected:
        /** Map with all the trace objects. */
        std::map < T_RESOURCE_TYPE, MPO_ResourceTrace* > my_resources;

        /** Flag used to describe whether usage of the associated resource is traced. */
        bool my_usage_flag;
        /** Flag used to describe whether freeing the resource associated is traced. */
        bool my_free_flag;
        /** Flag used to describe whether comiting the resource associated is traced. */
        bool my_commit_flag;
        /** Flag used to describe whether creating the resource associated is traced. */
        bool my_creation_flag;
        /** Flag used to describe whether dependencies of the resource associated are traced. */
        bool my_dependencies_flag;
        /** The name used for the class of resources (i.e. MPI_Datatype). */
        std::string my_resource_name;
        /** The maximum amount of resources known. */
        int my_max_num_ress;
        /** Creation time of this ResourceGatherer used for printing creation timings of resources. */
        double my_creation_time;

    public:
        /**
         * Constructor which sets the behavior of the resource traces.
         * @param resource_name the name of the resource-type i.e. "MPI_Datatype".
         * @param has_usage flag that sets whether resource usage is traced or not.
         * @param has_free flag that sets whether freeing resources is traced or not.
         * @param has_commit flag that sets whether commiting resources is traced or not.
         * @param has_creation flag that sets whether resource creation is traced or not.
         * @param has_dependencies flag that sets whether resource dependencies are traced or not.
         */
        MPO_ResourceGathererFull (std::string resource_name,bool has_usage,bool has_free,bool has_commit,bool has_creation,bool has_dependencies);

        /**
         * Destructor that frees all the resource traces.
         */
        ~MPO_ResourceGathererFull (void);

        /**
         * Searches for the trace associated with the given handle.
         * @param handle the handle identifing the resource.
         * @return the trace associated with the given handle when present, othereise NULL.
         */
        MPO_ResourceTrace* find (T_RESOURCE_TYPE handle);

        /**
         * Finds a resource object and sets its free info to the given
         * source location.
         * If there is no resource with the speciefied handle present the
         * operation is aborted!
         * If the source location represents an unknown location the operation will
         * be completly aborted.
         * @param handle the handle of the resource object to find.
         * @param loc the location of the free call.
         * @return the resource trace found or NULL if the location is invalid
         * 			or no trace with the given handle is present.
         */
        void findAndSetFree (T_RESOURCE_TYPE handle, MPO_SourceLocation loc);

        /**
         * Finds a resource object and sets its commit info to the given
         * source location.
         * If there is no resource with the speciefied handle present the
         * operation is aborted!
         * If the source location represents an unknown location the operation will
         * be completly aborted.
         * @param handle the handle of the resource object to find.
         * @param loc the location of the commit call.
         * @return the resource trace found or NULL if the location is invalid
         * 			or no trace with the given handle is present.
         */
        void findAndSetCommit (T_RESOURCE_TYPE handle, MPO_SourceLocation loc);

        /**
         * Finds a resource object and uses it to print its trace to the given
         * ostream.
         * If there is no resource with the speciefied handle present the
         * operation is aborted!
         * @param handle the handle of the resource object to find.
         * @param out the target for the print operation.
         */
        void findAndPrintTrace (T_RESOURCE_TYPE handle, std::ostream & out);

        /**
         * Finds a resource object and uses it to print its trace to the given
         * ostream.
         * If there is no resource with the speciefied handle present the
         * operation is aborted!
         * @param handle the handle of the resource object to find.
         * @param out the target for the print operation.
         * @param argument_name the name of the associated argument.
         */
        void findAndPrintTraceNamed (T_RESOURCE_TYPE handle, std::ostream & out, std::string argument_name);

        /**
         * Finds a resource object and an resource object it depends on.
         * If both are found the second one's creation location is added to the
         * first one's dependencies.
         * If one of the resources was not found the operation is aborted
         * i.e because the second resource was a standard resource that is not
         * traced !
         * @param handle the handle of the dependent resource object.
         * @param dependency the handle of the resource object representing the dependency.
         */
        void findAndTryAddDependendency (T_RESOURCE_TYPE handle, T_RESOURCE_TYPE dependency);

        /**
         * This method searches the resource trace for handle. If found it adds
         * loc to its usage list.
         * If not found it aborts the operation.
         * @param handle the handle of the resource to find.
         * @param loc the usage location that should be added to the resource trace of handle.
         * @return the resource trace found or NULL.
         */
        void findAndAddUsage (T_RESOURCE_TYPE handle, MPO_SourceLocation loc);

    protected:
        /**
         * Adds a resource object.
         * If there is already a resource with the same handle present, the old
         * one will be removed!
         * @param handle the handle of the new resource object.
         * @return the newly created resource trace.
         */
        MPO_ResourceTrace* add (T_RESOURCE_TYPE handle);

    public:
        /**
         * Adds a resource object and sets its creation info to the given
         * source location.
         * If there is already a resource with the same handle present, the old
         * one will be removed!
         * If the source location represents an unknown location the operation will
         * be completly aborted.
         * @param handle the handle of the new resource object.
         * @param loc the source location of the creation call.
         * @return the newly created resource trace or null if the location is invalid.
         */
        void addAndSetCreation (T_RESOURCE_TYPE handle, MPO_SourceLocation loc);

        /**
         * Removes the resource with the given handle.
         * @param handle the handle of the resource to be removed.
         */
        void remove (T_RESOURCE_TYPE handle);

        /**
         * This method returns the maximum amount of resources ever used.
         * @return the max amount of resources used.
         */
        int getMaxNumResourcesUsed (void);

        /**
         * Returns the amount of resources that are not freed.
         * If resources don`t have a free information the return
         * value is 0.
         * @return 	amount of resources not freed or 0 if this information
         * 			is not available.
         */
        int getNumResourcesNotFreed (void);

        /**
         * Prints information about the resources to the given ostream.
         * @param out ostream to store the information in.
         */
        void printResourcesNotFreed (std::ostream & out);

        /**
         * Prints information for the amount of Resources used.
         * The output is thougth to be processed by a tool like
         * gnuplot in order to gain information on resource usage
         * and its timing.
         * @param out the stream to store the data in.
         */
        void printResourcesUsedGraph(std::ostream & out);

        /**
         * Prints information for the amount of Resources used.
         * The output is thougth to be processed by a tool like
         * gnuplot in order to gain information on resource usage
         * and its timing.
         * The file won`t be created if no data is present.
         * @param file_name the name of the file to store the output in
         * 		  the file is only created if data is present.
         */
        void printResourcesUsedGraph(string file_name);

        /**
         * Returns the amount of resources that were used at the given time.
         * @param t the time point.
         * @return the amount of resources active at that instant.
         */
        int getAmountOfResourcesUsedAt (double t);

    }; //end of class MPO_ResourceGatherer

    //Method MPO_ResourceGatherer
    //========================================================================
    template < class T_RESOURCE_TYPE > MPO_ResourceGathererFull<T_RESOURCE_TYPE>::MPO_ResourceGathererFull(std::string resource_name, bool has_usage, bool has_free, bool has_commit, bool has_creation, bool has_dependencies) :
    	my_usage_flag(has_usage),
    	my_free_flag(has_free),
    	my_commit_flag(has_commit),
    	my_creation_flag(has_creation),
    	my_dependencies_flag(has_dependencies),
    	my_resource_name(resource_name),
    	my_max_num_ress(0),
    	my_creation_time(PMPI_Wtime())
    {
        //my_creation_time = PMPI_Wtime ();
    }

    //Method ~MPO_ResourceGatherer
    //========================================================================
    template < class T_RESOURCE_TYPE > MPO_ResourceGathererFull<T_RESOURCE_TYPE>::~MPO_ResourceGathererFull(void)
    {
    	//remove all the traces
    	typename std::map < T_RESOURCE_TYPE, MPO_ResourceTrace* >::iterator i;
    	MPO_ResourceTrace* p;

    	for (i = my_resources.begin(); i != my_resources.end(); i++)
    	{
            p = (*i).second;
            if (p != NULL)
                delete (p);

            (*i).second = NULL;
    	}

    }//end of method ~MPO_ResourceGatherer

    //Method find
    //========================================================================
    template < class T_RESOURCE_TYPE > MPO_ResourceTrace* MPO_ResourceGathererFull<T_RESOURCE_TYPE>::find(T_RESOURCE_TYPE handle)
    {
    	typename std::map < T_RESOURCE_TYPE, MPO_ResourceTrace* >::iterator i;
    	i = my_resources.find(handle);

        if (i == my_resources.end())
            return NULL;

        return i->second;

    } //end of methof find

    //Method findAndSetFree
    //========================================================================
    template < class T_RESOURCE_TYPE > void MPO_ResourceGathererFull<T_RESOURCE_TYPE>::findAndSetFree (T_RESOURCE_TYPE handle, MPO_SourceLocation loc)
    {
    	//if (!loc.isValidLocation())
        //    return;

    	MPO_ResourceTrace *ret = find (handle);

    	if (ret)
            ret->setFree(loc);
    }

    //Method findAndPrintTrace
    //========================================================================
    template < class T_RESOURCE_TYPE > void MPO_ResourceGathererFull<T_RESOURCE_TYPE>::findAndPrintTrace (T_RESOURCE_TYPE handle, std::ostream & out)
    {
    	MPO_ResourceTrace *ret = find (handle);

    	if (ret)
            ret->printTrace(out);
    }

    template < class T_RESOURCE_TYPE > void MPO_ResourceGathererFull<T_RESOURCE_TYPE>::findAndPrintTraceNamed (T_RESOURCE_TYPE handle, std::ostream & out, std::string argument_name)
    {
        MPO_ResourceTrace *ret = find (handle);

    	if (ret)
    	{
            out << std::endl << "Argument: " << argument_name << std::endl;
            ret->printTrace(out);
    	}
    }

    //Method findAndSetCommit
    //========================================================================
    template < class T_RESOURCE_TYPE > void MPO_ResourceGathererFull<T_RESOURCE_TYPE>::findAndSetCommit (T_RESOURCE_TYPE handle, MPO_SourceLocation loc)
    {
    	//if (!loc.isValidLocation())
        //    return;

    	MPO_ResourceTrace *ret = find (handle);

    	if (ret)
            ret->setCommit(loc);

    }

    //Method findAndTryAddDependendency
    //========================================================================
    template < class T_RESOURCE_TYPE > void MPO_ResourceGathererFull<T_RESOURCE_TYPE>::findAndTryAddDependendency (T_RESOURCE_TYPE handle, T_RESOURCE_TYPE dependency)
    {
    	MPO_ResourceTrace *ret = find (handle);
    	MPO_ResourceTrace *dep = find (dependency);

    	if (!ret || !dep)
            return;

    	MPO_SourceLocation l = dep->getCreation();

    	if (l.isValidLocation())
            ret->addDependency(l);

    }

    //Method findAndAddUsage
    //========================================================================
    template < class T_RESOURCE_TYPE > void MPO_ResourceGathererFull<T_RESOURCE_TYPE>::findAndAddUsage (T_RESOURCE_TYPE handle, MPO_SourceLocation loc)
    {
    	MPO_ResourceTrace *ret = find (handle);

    	if (ret != NULL)
            ret->addUsage(loc);

    }

    //Method add
    //========================================================================
    template < class T_RESOURCE_TYPE > MPO_ResourceTrace* MPO_ResourceGathererFull<T_RESOURCE_TYPE>::add(T_RESOURCE_TYPE handle)
    {
    	//remove when already present
    	/** \todo is this a good solution ? (for Tobias).*/
    	remove(handle);

    	MPO_ResourceTrace* t = new MPO_ResourceTrace (
            my_resource_name,
            my_usage_flag,
            my_free_flag,
            my_commit_flag,
            my_creation_flag,
            my_dependencies_flag);

    	my_resources.insert( std::make_pair(
                                 handle,t));

    	int numUsed = getNumResourcesNotFreed();
    	if (numUsed > my_max_num_ress)
            my_max_num_ress = numUsed;

    	return t;
    } //end of method add

    //Method addAndSetLocation
    //========================================================================
    template < class T_RESOURCE_TYPE > void MPO_ResourceGathererFull<T_RESOURCE_TYPE>::addAndSetCreation (T_RESOURCE_TYPE handle, MPO_SourceLocation loc)
    {
      	//if (loc.isValidLocation()) //Needs to be added always, otherwise non-freed handles will not be listed
    	//{
            MPO_ResourceTrace* t = add (handle);
            if (t)
                t->setCreation(loc);
    	//}
    }

    //Method remove
    //========================================================================
    template < class T_RESOURCE_TYPE > void MPO_ResourceGathererFull<T_RESOURCE_TYPE>::remove(T_RESOURCE_TYPE handle)
    {
    	MPO_ResourceTrace* p = find (handle);

    	if (p != NULL)
    	{
            my_resources.erase (handle);
            delete (p);
            p = NULL;
    	}
    }

    //Method getMaxNumResourcesUsed
    //========================================================================
    template < class T_RESOURCE_TYPE > int MPO_ResourceGathererFull<T_RESOURCE_TYPE>::getMaxNumResourcesUsed (void)
    {
    	return my_max_num_ress;
    }

    //Method getNumResourcesNotFreed
    //========================================================================
    template < class T_RESOURCE_TYPE > int MPO_ResourceGathererFull<T_RESOURCE_TYPE>::getNumResourcesNotFreed (void)
    {
    	if (!my_free_flag)
            return 0;

    	int rvalue = 0;

    	typename std::map < T_RESOURCE_TYPE, MPO_ResourceTrace* >::iterator i;

    	for (i = my_resources.begin(); i != my_resources.end();i++)
    	{
            MPO_ResourceTrace* t = i->second;

            if (!t)
                continue;

            if (!t->isFreed())
            {
                rvalue++;
            }
    	}

    	return rvalue;
    }

    //Method printResourcesNotFreed
    //========================================================================
    template < class T_RESOURCE_TYPE > void MPO_ResourceGathererFull<T_RESOURCE_TYPE>::printResourcesNotFreed (std::ostream & out)
    {
    	if (!my_free_flag)
            return;

    	typename std::map < T_RESOURCE_TYPE, MPO_ResourceTrace* >::iterator i;

    	for (i = my_resources.begin(); i != my_resources.end();i++)
    	{
            MPO_ResourceTrace* t = i->second;

            if (!t)
                continue;

            if (!t->isFreed())
            {
                t->printTrace(out);
                out << std::endl;
            }
    	}
    }

    //Method printResourcesUsedGraph
    //========================================================================
    template < class T_RESOURCE_TYPE > void MPO_ResourceGathererFull<T_RESOURCE_TYPE>::printResourcesUsedGraph(std::ostream & out)
    {
    	/** @note Maybe this is a bit slow (cubic in the amount of resources used). */
    	if (!my_creation_flag)
            return; //we cant do this work as we don`t know anything about creation

    	//first create a map with the creation times of the resources
    	//and the amount of resources used at this time
    	std::map < double, int > data;
    	MPO_ResourceTrace	*temp;
    	double 	creation_time,
            free_time,
            end_time = PMPI_Wtime();

    	typename std::map < T_RESOURCE_TYPE, MPO_ResourceTrace* >::iterator i;

    	for (i = my_resources.begin(); i != my_resources.end();i++)
    	{
            temp = i->second;

            if (!temp)
                continue; //something weird happened

            creation_time = temp->getCreationTime();
            //put the data tupel in the map
            data.insert( std::make_pair(creation_time,getAmountOfResourcesUsedAt(creation_time)));

            if (temp->isFreed())
            {
                free_time = temp->getFreeTime();
                //put the data tupel in the map
                data.insert( std::make_pair(free_time,getAmountOfResourcesUsedAt(free_time)));
            }

    	}

    	data.insert(std::make_pair(end_time,getNumResourcesNotFreed()));

    	//now we print the stuff to the ostream
    	out << "#Time\tResources-Used" << std::endl;

    	std::map < double, int >::iterator j;

    	for (j = data.begin();j != data.end();j++)
    	{
            out << (double)(j->first) << "\t" << j->second << std::endl;
            //out << (double)(i->first-my_creation_time) << "\t" << i->second << std::endl;
    	}
    }

    //Method printResourcesUsedGraph
    //========================================================================
    template < class T_RESOURCE_TYPE > void MPO_ResourceGathererFull<T_RESOURCE_TYPE>::printResourcesUsedGraph(string file_name)
    {
    	if (my_resources.size() == 0)
            return; //no data present

    	std::fstream out;
        out.open (file_name.c_str(), std::fstream::out);
        printResourcesUsedGraph(out);
        out.close();
    }

    //Method getAmountOfResourcesUsedAt
    //========================================================================
    template < class T_RESOURCE_TYPE > int MPO_ResourceGathererFull<T_RESOURCE_TYPE>::getAmountOfResourcesUsedAt (double t)
    {
    	int number_ress_used = 0;
    	MPO_ResourceTrace	*toTest;

    	//calculate the amount of resources used at this time
    	typename std::map < T_RESOURCE_TYPE, MPO_ResourceTrace* >::iterator j;

    	for (j = my_resources.begin(); j != my_resources.end();j++)
    	{
            toTest = j->second;
            if (!toTest)
                continue;
            if (toTest->wasUsedAtTime(t))
                number_ress_used++;
    	}

    	return number_ress_used;
    }

} //end of namespace MPO

#endif /*MPO_RESOURCEGATHERER_H*/
