/**
 * MPI-calls of MPI-2-standard (Miscellany and External Interfaces)
 *
 * @see MPI-2 standard
 *
 * @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *        Matthias Jurenz
 *
  * \note For Fortran we need additional information for the arguments.
 * 		 That is necessary for the handle conversion for MPI-Implementations like
 * 		 OpenMPI.
 * 		 We need for all MPI-Handles of pointer type information how the argument
 * 		 is used. That means we have to know whether it is:
 * 			* an out single value
 * 			* an in-out single value
 * 			* an array for input purpoeses
 * 			* an out array for output purposses
 * 			* an in-out array
 * 		For the arrays we additionally need to know their size !
 * 		We have to denote all this in the argument name, so we will append at the
 * 		end of these argument names:
 * 			* _CLASS_SINGLE_IN for an in single value (this is often used for requests)
 * 			* _CLASS_SINGLE_OUT for an out single value
 * 			* _CLASS_SINGLE_IO for an in-out single value
 * 			* _CLASS_ARRAY_IN_sizeargument an array for input purpoeses
 * 			* _CLASS_ARRAY_OUT_sizeargument an out array for output purposses
 * 			* _CLASS_ARRAY_IO_sizeargument an in-out array
 * 		Where "sizeargument" is the fixed numeric size or the argument speci-
 * 		fing the array size.
 * 		Otherwise we would have to create the wrapper manually.
 *
 * \note For Fortran we have to modify indizes into arrays.
 * 	     As Fortran arrays start with 1 and C arrays with 0.
 *       This affects the MPI-1.2 functions MPI_Wait[some,any] and MPI_Test[some,any].
 *       We define the following definitions that can be appended to an argument name:
 *          * _CLASS_ARRAYINDEX_OUT_SINGLE_NOCONDITION
 *             -> for a single index into an array which is used as a output argument
 *          * _CLASS_ARRAYINDEX_OUT_SINGLE_CONDITION_conditionargument
 *             -> for a single index into an array which is used as a output argument
 *                but only if the argument "*conditionargument" is true
 *                As this is a output value we assume conditionargument also to be a
 *                output argument !
 *          * _CLASS_ARRAYINDEX_OUT_ARRAY_sizeargument
 *             -> for an array of indices into a array which is used as a output argument
 *                the size of the index array is set by "*sizeargument"
 *                As this is a output value we assume sizeargument also to be a output
 *                argument !
 *       These definitions are only a subset of the imaginable cases and the ones currently
 *       needed for the MPI-1.2 functions.
 *
 **/

/* Miscellany */
int MPI_Alloc_mem(MPI_Aint size, MPI_Info info, void* baseptr);
int MPI_Comm_create_errhandler(MPI_Comm_errhandler_fn* function, MPI_Errhandler* errhandler_CLASS_SINGLE_OUT);
int MPI_Comm_get_errhandler(MPI_Comm comm, MPI_Errhandler* errhandler_CLASS_SINGLE_OUT);
int MPI_Comm_set_errhandler(MPI_Comm comm, MPI_Errhandler errhandler);
int MPI_Finalized(int* flag);
int MPI_Free_mem(void* base);
int MPI_Get_address(void* location, MPI_Aint* address_CLASS_SINGLE_OUT);
int MPI_Info_create(MPI_Info* info_CLASS_SINGLE_OUT);
int MPI_Info_delete(MPI_Info info, char* key_CLASS_SINGLE_IN);
int MPI_Info_dup(MPI_Info info, MPI_Info* newinfo_CLASS_SINGLE_OUT);
int MPI_Info_free(MPI_Info* info_CLASS_SINGLE_IO);
int MPI_Info_get(MPI_Info info, char* key_CLASS_SINGLE_IN, int valuelen, char* value_CLASS_SINGLE_OUT, int* flag);
int MPI_Info_get_nkeys(MPI_Info info, int* nkeys);
int MPI_Info_get_nthkey(MPI_Info info, int n, char* key_CLASS_SINGLE_OUT);
int MPI_Info_get_valuelen(MPI_Info info, char* key_CLASS_SINGLE_IN, int* valuelen, int* flag);
int MPI_Info_set(MPI_Info info, char* key_CLASS_SINGLE_IN, char* value_CLASS_SINGLE_IN);
int MPI_Pack_external(char* datarep_CLASS_SINGLE_IN, void* inbuf, int incount, MPI_Datatype datatype, void* outbuf, MPI_Aint outsize, MPI_Aint* position_CLASS_SINGLE_IO);
int MPI_Pack_external_size(char* datarep_CLASS_SINGLE_IN, int incount, MPI_Datatype datatype, MPI_Aint* size_CLASS_SINGLE_OUT);
int MPI_Request_get_status(MPI_Request request, int* flag, MPI_Status* status_CLASS_SINGLE_OUT);
int MPI_Type_create_darray(int size, int rank, int ndims, int* array_of_gsizes, int* array_of_distribs, int* array_of_dargs, int* array_of_psizes, int order, MPI_Datatype oldtype, MPI_Datatype* newtype_CLASS_SINGLE_OUT);
int MPI_Type_create_hindexed(int count, int* array_of_blocklengths, MPI_Aint* array_of_displacements_CLASS_ARRAY_IN_count, MPI_Datatype oldtype, MPI_Datatype* newtype_CLASS_SINGLE_OUT);
int MPI_Type_create_hvector(int count, int blocklength, MPI_Aint stride, MPI_Datatype oldtype, MPI_Datatype* newtype_CLASS_SINGLE_OUT);
int MPI_Type_create_indexed_block(int count, int blocklength, int* array_of_displacements, MPI_Datatype oldtype, MPI_Datatype* newtype_CLASS_SINGLE_OUT);
int MPI_Type_create_resized(MPI_Datatype oldtype, MPI_Aint lb, MPI_Aint extent, MPI_Datatype* newtype_CLASS_SINGLE_OUT);
int MPI_Type_create_struct(int count, int* array_of_blocklengths, MPI_Aint* array_of_displacements_CLASS_ARRAY_IN_count, MPI_Datatype* array_of_types_CLASS_ARRAY_IN_count, MPI_Datatype* newtype_CLASS_SINGLE_OUT);
int MPI_Type_create_subarray(int ndims, int* array_of_sizes, int* array_of_subsizes, int* array_of_starts, int order, MPI_Datatype oldtype, MPI_Datatype* newtype_CLASS_SINGLE_OUT);
int MPI_Type_get_extent(MPI_Datatype datatype, MPI_Aint* lb_CLASS_SINGLE_OUT, MPI_Aint* extent_CLASS_SINGLE_OUT);
int MPI_Type_get_true_extent(MPI_Datatype datatype, MPI_Aint* true_lb_CLASS_SINGLE_OUT, MPI_Aint* true_extent_CLASS_SINGLE_OUT);
int MPI_Unpack_external(char* datarep_CLASS_SINGLE_IN, void* inbuf, MPI_Aint insize, MPI_Aint* position_CLASS_SINGLE_IO, void* outbuf, int outcount, MPI_Datatype datatype);
int MPI_Win_create_errhandler(MPI_Win_errhandler_fn* function, MPI_Errhandler* errhandler_CLASS_SINGLE_OUT);
int MPI_Win_get_errhandler(MPI_Win win, MPI_Errhandler* errhandler_CLASS_SINGLE_OUT);
int MPI_Win_set_errhandler(MPI_Win win, MPI_Errhandler errhandler);
int MPI_File_call_errhandler(MPI_File fh, int errorcode);
int MPI_File_create_errhandler(MPI_File_errhandler_fn* function_CLASS_SINGLE_IN, MPI_Errhandler* errhandler_CLASS_SINGLE_OUT);
int MPI_File_get_errhandler(MPI_File file, MPI_Errhandler* errhandler_CLASS_SINGLE_OUT);
int MPI_File_set_errhandler(MPI_File file, MPI_Errhandler errhandler);


/* External Interfaces */
int MPI_Add_error_class(int* errorclass);
int MPI_Add_error_code(int errorclass, int* errorcode);
int MPI_Add_error_string(int errorcode, char* string_CLASS_SINGLE_IN);
int MPI_Comm_call_errhandler(MPI_Comm comm, int errorcode);
int MPI_Comm_create_keyval(MPI_Comm_copy_attr_function* comm_copy_attr_fn, MPI_Comm_delete_attr_function* comm_delete_attr_fn, int* comm_keyval, void* extra_state);
int MPI_Comm_delete_attr(MPI_Comm comm, int comm_keyval);
int MPI_Comm_free_keyval(int* comm_keyval);
int MPI_Comm_get_attr(MPI_Comm comm, int comm_keyval, void* attribute_val, int* flag);
int MPI_Comm_get_name(MPI_Comm comm, char* comm_name_CLASS_SINGLE_OUT, int* resultlen);
int MPI_Comm_set_attr(MPI_Comm comm, int comm_keyval, void* attribute_val);
int MPI_Comm_set_name(MPI_Comm comm, char* comm_name_CLASS_SINGLE_IN);
int MPI_Grequest_complete(MPI_Request request);
int MPI_Grequest_start(MPI_Grequest_query_function* query_fn, MPI_Grequest_free_function* free_fn, MPI_Grequest_cancel_function* cancel_fn, void* extra_state, MPI_Request* request_CLASS_SINGLE_OUT);
/*int MPI_Init_thread(int* argc, char*** argv, int required, int* provided);*/ /*Stays in mpi_standard.h for historic reasons!*/
int MPI_Is_thread_main(int* flag);
int MPI_Query_thread(int* provided);
int MPI_Status_set_cancelled(MPI_Status* status_CLASS_SINGLE_IO, int flag);
int MPI_Status_set_elements(MPI_Status* status_CLASS_SINGLE_IO, MPI_Datatype datatype, int count);
int MPI_Type_create_keyval(MPI_Type_copy_attr_function* type_copy_attr_fn, MPI_Type_delete_attr_function* type_delete_attr_fn, int* type_keyval, void* extra_state);
int MPI_Type_delete_attr(MPI_Datatype type, int type_keyval);
int MPI_Type_dup(MPI_Datatype type, MPI_Datatype* newtype_CLASS_SINGLE_OUT);
int MPI_Type_free_keyval(int* type_keyval);
int MPI_Type_get_attr(MPI_Datatype type, int type_keyval, void* attribute_val, int* flag);
int MPI_Type_get_contents(MPI_Datatype datatype, int max_integers, int max_addresses, int max_datatypes, int* array_of_integers, MPI_Aint* array_of_addresses_CLASS_ARRAY_OUT_max_addresses, MPI_Datatype* array_of_datatypes_CLASS_ARRAY_OUT_max_datatypes);
int MPI_Type_get_envelope(MPI_Datatype datatype, int* num_integers, int* num_addresses, int* num_datatypes, int* combiner);
int MPI_Type_get_name(MPI_Datatype type, char* type_name_CLASS_SINGLE_OUT, int* resultlen);
int MPI_Type_set_attr(MPI_Datatype type, int type_keyval, void* attribute_val);
int MPI_Type_set_name(MPI_Datatype type, char* type_name_CLASS_SINGLE_IN);
int MPI_Win_call_errhandler(MPI_Win win, int errorcode);
int MPI_Win_create_keyval(MPI_Win_copy_attr_function* win_copy_attr_fn, MPI_Win_delete_attr_function* win_delete_attr_fn, int* win_keyval, void* extra_state);
int MPI_Win_delete_attr(MPI_Win win, int win_keyval);
int MPI_Win_free_keyval(int* win_keyval);
int MPI_Win_get_attr(MPI_Win win, int win_keyval, void* attribute_val, int* flag);
int MPI_Win_get_name(MPI_Win win, char* win_name_CLASS_SINGLE_OUT, int* resultlen);
int MPI_Win_set_attr(MPI_Win win, int win_keyval, void* attribute_val);
int MPI_Win_set_name(MPI_Win win, char* win_name_CLASS_SINGLE_IN);
int MPI_Type_create_f90_complex(int p, int r, MPI_Datatype* newtype_CLASS_SINGLE_OUT);
int MPI_Type_create_f90_integer(int r, MPI_Datatype* newtype_CLASS_SINGLE_OUT);
int MPI_Type_create_f90_real(int p, int r, MPI_Datatype* newtype_CLASS_SINGLE_OUT);
int MPI_Type_match_size(int typeclass, int size, MPI_Datatype* type_CLASS_SINGLE_OUT);
int MPI_Register_datarep(char* datarep_CLASS_SINGLE_IN, MPI_Datarep_conversion_function* read_conversion_fn, MPI_Datarep_conversion_function* write_conversion_fn, MPI_Datarep_extent_function* dtype_file_extent_fn, void* extra_state);
