//#include "mpo.h"
#ifndef MPO_1SIDED_H
#define MPO_1SIDED_H

namespace MPO {
    //=======================================================================


    //=======================================================================
    // one-sided communication
    //=======================================================================


    //************************************************************************
    //
    //                          MPO_Winhandler
    //
    /**  This abstract base class represents calls handling windows.
     *
     *  All calls for window-handling are derived from this.
     *  Correctnes of used window is checked in implemented
     *  member function.
     */

    enum MPI_WIN_MARMOT_STATES{
    	MPI_WIN_FREE_ALLOWED,
    	MPI_WIN_FREE_NOT_ALLOWED
    } ;

    typedef enum MPI_WIN_MARMOT_STATES mpi_win_marmot_states_t;


    //************************************************************************
    //
    //                          MPO_Win
    //

    /** This class represents the type MPI_Win.
     */
    class MPO_Win : public MPO_Typeobject
    {
    public:
        //========================
        // Windows
        //========================

        /** Set to handle (known) windows.
         *
         *  Herewith we can check if a used window is valid or not:
         *  -  All newly created windows are automatically inserted in
         *     our_wins, and removed when freed.
         */
       static std::map<MPI_Win, mpi_win_marmot_states_t> our_wins;

        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Win();

        /** Constructor taking proper arguments.
         *
         *  @param  win    The MPI window
         */
        MPO_Win(MPI_Win win);

        /** The destructor is virtual because there are virtual
         *  members here.
         */
        virtual ~MPO_Win();

        //==========================
        // Methods
        //==========================

        /** @return  the unique type identity set by enum MPO_TYPE_ID
         */
        virtual int getId() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

        //==========================
        // Operators
        //==========================

        /** @return my_win    Our window */
        operator  MPI_Win() const;

    private:
        //==========================
        // Data
        //==========================

        /** Our window */
        MPI_Win my_win;

        /** Our debugserver name (only used on debugserver)*/
        std::string my_name;

    };                          // end class MPO_Win


    class MPO_Winhandler : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Winhandler();

        /** Constructor taking proper arguments.
         *
         *  @param  comm    communicator
         */
        MPO_Winhandler(MPO_Win win);

        /** The destructor is virtual because there are virtual
         *  members here.
         */
        virtual ~MPO_Winhandler();

        //==========================
        // Methods
        //==========================

        /** Checks prerequisites such as correctness of parameters.
         *  Checks whether
         *  - win
         *  is valid.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether window is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_WIN_NULL
         *           - MARMOT_ERR_WIN_NOT_VALID
         */
        virtual int checkWin();

        /** Returns the unique family id (the unique enumeration value)
         *  for the family of callobjects.
         */
        virtual int getFamilyId() const;

        /** Does the MPI call after checking it. */
        virtual int callNativeMPI() = 0;

        /** Returns the name of the MPI-call. */
        virtual string getCallname() const = 0;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const = 0;

    protected:
        //==========================
        // Data
        //==========================

        MPO_Win my_win;

    };                          // end class MPO_Winhandler



    class MPO_WinCommunicator : public MPO_Winhandler
    {
    public:
      MPO_WinCommunicator();
      MPO_WinCommunicator(void *origin_addr, int origin_count, MPO_Datatype origin_datatype, int target_rank, MPI_Aint target_disp,
                          int target_count, MPO_Datatype target_datatype, MPO_Win win);
      virtual ~MPO_WinCommunicator();
      int getFamilyId() const;
      int checkPrerequisites2();
      int checkCountOrigin();
      int checkDatatypeOrigin();
      int checkRankTarget();
      int checkDispTarget();
      int checkCountTarget();
      int checkDatatypeTarget();

    protected:
      void        *my_origin_addr;
      int          my_origin_count;
      MPO_Datatype my_origin_datatype;
      int          my_target_rank;
      MPI_Aint     my_target_disp;
      int          my_target_count;
      MPO_Datatype my_target_datatype;
    };

#ifdef IS_MPI_WIN_CREATE_PRESENT
    //************************************************************************
    //
    //                          MPO_Win_create
    //
    /** This class represents MPI_Win_create.
     *
     *  collective call
     */
    class MPO_Win_create : public MPO_Winhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Win_create();

        /** Constructor taking proper arguments.
         *
         *  @param  base       initial address of window (choice)
         *  @param  size       size of window in bytes (nonnegative integer)
         *  @param  disp_unit  local unit size for displacements,
         *                     in bytes (positive integer)
         *  @param  info       info argument (handle)
         *  @param  comm       communicator (handle)
         *  @param  win        window object returned by the call (handle)
         */
        MPO_Win_create(void *base, MPI_Aint size, int disp_unit,
                       MPI_Info info, MPO_Comm comm, MPI_Win *win);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Win_create();

        //==========================
        // Methods
        //==========================

        /** Checks correctness of
         *  - size of window
         *  - displacements
         *  - info?
         *  - communicator
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.*/
        virtual int checkPost();

        /** Checks whether size of window is correct.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_WIN_SIZE_NEG
         */
        virtual int checkWinSize();

        /** Checks whether unit size of displacements is correct.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_WIN_DISP_UNIT_SIZE_NONPOS
         */
        virtual int checkWinDispUnit();

        /** Checks whether communicator is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_NULL
         *           - MARMOT_ERR_COMM_NOT_VALID
         */
        virtual int checkComm();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Win_create
         */
        virtual int callNativeMPI();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    protected:
        //==========================
        // Data
        //==========================
        void *my_base;
        MPI_Aint my_size;
        int my_disp_unit;
        MPI_Info my_info;
        MPO_Comm my_comm;
        MPI_Win *mpo_win_create_Win;

    };                          // end class MPO_Win_create
#endif /*IS_MPI_WIN_CREATE_PRESENT*/

#ifdef IS_MPI_WIN_FREE_PRESENT
    //************************************************************************
    //
    //                          MPO_Win_free
    //
    /** This class represents MPI_Win_free.
     *
     *  collective call
     *  can only be called after communication is completed, i.e. after that
     *  process has called an MPI_WIN_ fence, MPI_Win_wait, MPI_Win_complete
     *  or MPI_Win_unlock call.
     */
    class MPO_Win_free : public MPO_Winhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Win_free();

        /** Constructor taking proper arguments.
         *  @param  win        window object (handle)
         */
        MPO_Win_free(MPI_Win *win);

        MPI_Win MPO_Win_getMPI_Win();
        void MPO_Win_setMPI_Win(MPI_Win win);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Win_free();

        //==========================
        // Methods
        //==========================

        /** Checks whether
         *  - win
         *  is valid.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether window is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_WARN_WIN_NULL
         *           - MARMOT_ERR_WIN_NOT_VALID
         */
        virtual int checkWin();

        virtual int checkPost();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Win_free
         */
        virtual int callNativeMPI();


        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================
        MPI_Win *mpo_win_free_Win;


    };                          // end class MPO_Win_free
#endif /*IS_MPI_WIN_CREATE_PRESENT*/

#ifdef IS_MPI_WIN_GET_GROUP_PRESENT
    //************************************************************************
    //
    //                          MPO_Win_get_group
    //
    /** This class represents MPI_Win_get_group. */
    class MPO_Win_get_group : public MPO_Winhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Win_get_group();

        /** Constructor taking proper arguments.
         *
         *  @param  win        window object (handle)
         *  @param  group      group of processes which share access to
         *                     the window (handle)
         */
        MPO_Win_get_group(MPO_Win win, MPI_Group *group);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Win_get_group();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Win_get_group
         */
        virtual int callNativeMPI();

        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Win my_win;
        MPI_Group *my_group;

    };                          // end class MPO_Win_get_group
#endif /*IS_MPI_WIN_GET_GROUP_PRESENT*/


#ifdef IS_MPI_PUT_PRESENT
    //************************************************************************
    //
    //                          MPO_Put
    //
    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_WIN
     *           - MARMOT_ERR_TYPE...
     *           - MARMOT_ERR_COUNT
     *           - MARMOT_WARN_COUNT_ZERO
     */

    /** This class represents MPI_Put. */
    class MPO_Put : public MPO_WinCommunicator
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Put();

        /** Constructor taking proper arguments.
         *
         *  @param  origin_addr     initial address of origin buffer (choice)
         *  @param  origin_count    number of entries in origin buffer
         *                          (nonnegative integer)
         *  @param  origin_datatype datatype of each entry
         *                          in origin buffer (handle)
         *  @param  target_rank     rank of target (nonnegative integer)
         *  @param  target_disp     displacement from start of window to
         *                          target buffer (nonnegative integer)
         *  @param  target_count    number of entries in target buffer
         *                          (nonnegative integer)
         *  @param  target_datatype datatype of each entry
         *                          in target buffer (handle)
         *  @parma  win             window object used for communication
         *                          (handle)
         */
        MPO_Put(void *origin_addr, int origin_count, MPO_Datatype
            origin_datatype, int target_rank, MPI_Aint target_disp,
            int target_count, MPO_Datatype target_datatype, MPI_Win
            win);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Put();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        virtual int checkPrerequisites();

        /** Checks whether window is valid.
         *
         *  @return unique error code set by enum MARMOT_ERR_CODES
         *          - MARMOT_SUCCESS
         *          - MARMOT_ERR_WIN_NULL
         *          - MARMOT_ERR_WIN_NOT_VALID
         */
        virtual int checkWin();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Put
         */
        virtual int callNativeMPI();

        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        // void *my_origin_addr;
        // int my_origin_count;
        // MPO_Datatype my_origin_datatype;
        // int my_target_rank;
        // MPI_Aint my_target_disp;
        // int my_target_count;
        // MPO_Datatype my_target_datatype;
        MPO_Win mpo_put_Win;

    };                          // end class MPO_Put
#endif /*IS_MPI_PUT_PRESENT*/

#ifdef IS_MPI_GET_PRESENT
    //************************************************************************
    //
    //                          MPO_Get
    //
    /** This class represents MPI_Get. */
    class MPO_Get : public MPO_WinCommunicator
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Get();

        /** Constructor taking proper arguments.
         *
         *  @param  origin_addr     address of the buffer in which to receive
         *                          the data
         *  @param  origin_count    number of entries in origin buffer
         *                          (nonnegative integer)
         *  @param  origin_datatype datatype of each entry
         *                          in origin buffer (handle)
         *  @param  target_rank     rank of target (nonnegative integer)
         *  @param  target_disp     displacement from window start to the
         *                          beginning of the target buffer
         *                          (nonnegative integer)
         *  @param  target_count    number of entries in target buffer
         *                          (nonnegative integer)
         *  @param  target_datatype datatype of each entry
         *                          in target buffer (handle)
         *  @param  win             window object used for communication (handle)
         */
        MPO_Get(void *origin_addr, int origin_count,
                MPO_Datatype origin_datatype, int target_rank,
                MPI_Aint target_disp, int target_count,
                MPO_Datatype target_datatype, MPI_Win win);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Get();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

         virtual int checkPrerequisites();

         /** Checks whether window is valid.
         *
         *  @return unique error code set by enum MARMOT_ERR_CODES
         *          - MARMOT_SUCCESS
         *          - MARMOT_ERR_WIN_NULL
         *          - MARMOT_ERR_WIN_NOT_VALID
         */
        virtual int checkWin();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Get
         */
        virtual int callNativeMPI();

        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        // void *my_origin_addr;
        // int my_origin_count;
        // MPO_Datatype my_origin_datatype;
        // int my_target_rank;
        // MPI_Aint my_target_disp;
        // int my_target_count;
        // MPO_Datatype my_target_datatype;
        MPO_Win mpo_get_Win;

    };                          // end class MPO_Get
#endif /*IS_MPI_PUT_PRESENT*/

#ifdef IS_MPI_ACCUMULATE_PRESENT
    //************************************************************************
    //
    //                          MPO_Accumulate
    //
    /** This class represents MPI_Accumulate. */
    class MPO_Accumulate : public MPO_WinCommunicator
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Accumulate();

        /** Constructor taking proper arguments.
         *
         *  @param  origin_addr     initial address of buffer (choice)
         *  @param  origin_count    number of entries in buffer
         *                          (nonnegative integer)
         *  @param  origin_datatype datatype of each buffer entry (handle)
         *  @param  target_rank     rank of target (nonnegative integer)
         *  @param  target_disp     displacement from window start to the
         *                          beginning of the target buffer
         *                          (nonnegative integer)
         *  @param  target_count    number of entries in target buffer
         *                          (nonnegative integer)
         *  @param  target_datatype datatype of each entry
         *                          in target buffer (handle)
         *  @param  op              predefined reduce operation (handle)
         *  @param  win             window object used for communication
         *                          (handle)
         */
        MPO_Accumulate(void *origin_addr, int origin_count,
                       MPO_Datatype origin_datatype, int target_rank,
                       MPI_Aint target_disp, int target_count,
                       MPO_Datatype target_datatype, MPO_Op op, MPO_Win win);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Accumulate();

        //==========================
        // Methods
        //==========================

        /** Checks whether op is valid.
         *
         *  @return unique error code set by enum MARMOT_ERR_CODES
         *          - MARMOT_SUCCESS
         *          - MARMOT_ERR_OP_NULL
         *          - MARMOT_ERR_OP_NOT_VALID
         */
        virtual int checkOp();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks whether window is valid.
         *
         *  @return unique error code set by enum MARMOT_ERR_CODES
         *          - MARMOT_SUCCESS
         *          - MARMOT_ERR_WIN_NULL
         *          - MARMOT_ERR_WIN_NOT_VALID
         */
        virtual int checkWin();

        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Accumulate
         */
        virtual int callNativeMPI();

        virtual int checkPost ();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        // void *my_origin_addr;
        // int my_origin_count;
        // MPO_Datatype my_origin_datatype;
        // int my_target_rank;
        // MPI_Aint my_target_disp;
        // int my_target_count;
        // MPO_Datatype my_target_datatype;
        MPO_Op mpo_accumulate_Op;
        MPO_Win mpo_accumulate_Win;

    };                          // end class MPO_Accumulate
#endif /*IS_MPI_ACCUMULATE_PRESENT*/

#ifdef IS_MPI_WIN_FENCE_PRESENT
    //************************************************************************
    //
    //                          MPO_Win_fence
    //
    /** This class represents MPI_Win_fence. */
    class MPO_Win_fence : public MPO_Winhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Win_fence();

        /** Constructor taking proper arguments.
         *
         *  @param  assert          program assertion (integer)
         *  @param  win             window object used for communication
         *                          (handle)
         */
        MPO_Win_fence(int assert, MPI_Win win);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Win_fence();

        //==========================
        // Methods
        //==========================

        /** Checks whether assert is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_ASSERT_NOT_VALID
         */
        virtual int checkAssert();

        virtual int checkWin();
        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        virtual int checkPrerequisites();

        virtual int checkPost();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Win_fence
         */
        virtual int callNativeMPI();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);


    private:
        //==========================
        // Data
        //==========================
        int my_assert;
        MPO_Win mpo_win_fence_Win;

    };                          // end class MPO_Win_fence
#endif /*IS_MPI_WIN_FENCE_PRESENT*/

#ifdef IS_MPI_WIN_START_PRESENT
    //************************************************************************
    //
    //                          MPO_Win_start
    //
    /** This class represents MPI_Win_start. */
    class MPO_Win_start : public MPO_Winhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Win_start();

        /** Constructor taking proper arguments.
         *
         *  @param  group           group of target processes (handle)
         *  @param  assert          Used to optimize this call;
         *                          zero may be used as a default.
         *  @param  win             window object (handle)
         */
        MPO_Win_start(MPO_Group group, int assert, MPO_Win win);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Win_start();

        //==========================
        // Methods
        //==========================

        /** Checks prerequisites such as correctness of parameters.
         *  Checks whether
         *  - group
         *  - assert
         *  - win
         *  are valid.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether group is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_GROUP_NULL
         *           - MARMOT_ERR_GROUP_NOT_VALID
         */
        virtual int checkGroup();

        /** Checks whether assert is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_ASSERT_NOT_VALID
         */
        virtual int checkAssert();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        virtual int checkPost();

        /** Checks whether window is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_WIN_NULL
         *           - MARMOT_ERR_WIN_NOT_VALID
         */
        virtual int checkWin();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Win_start
         */
        virtual int callNativeMPI();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================
        MPO_Group my_group;
        int my_assert;
        MPO_Win mpo_win_start_Win;

    };                          // end class MPO_Win_start
#endif /*IS_MPI_WIN_START_PRESENT*/

#ifdef IS_MPI_WIN_COMPLETE_PRESENT
    //************************************************************************
    //
    //                          MPO_Win_complete
    //
    /** This class represents MPI_Win_complete. */
    class MPO_Win_complete : public MPO_Winhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Win_complete();

        /** Constructor taking proper arguments.
         *  @param  win             window object (handle)
         */
        MPO_Win_complete(MPO_Win win);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Win_complete();

        //==========================
        // Methods
        //==========================

        /** Checks prerequisites such as correctness of parameters.
         *  Checks whether
         *  - win
         *  is valid.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        virtual int checkPost();

        /** Checks whether window is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_WIN_NULL
         *           - MARMOT_ERR_WIN_NOT_VALID
         */
        virtual int checkWin();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Win_complete
         */
        virtual int callNativeMPI();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        MPO_Win mpo_win_complete_Win;

    };                          // end class MPO_Win_complete
#endif /*IS_MPI_WIN_COMPLETE_PRESENT*/


#ifdef IS_MPI_WIN_POST_PRESENT
    //************************************************************************
    //
    //                          MPO_Win_post
    //
    /** This class represents MPI_Win_post. */
    class MPO_Win_post : public MPO_Winhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Win_post();

        /** Constructor taking proper arguments.
         *
         *  @param  group           group of origin processes (handle)
         *  @param  assert          used to optimize this call;
         *                          zero may be used as a default.
         *  @param  win             window object (handle)
         */
        MPO_Win_post(MPO_Group group, int assert, MPO_Win win);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Win_post();

        //==========================
        // Methods
        //==========================

        /** Checks prerequisites such as correctness of parameters.
         *  Checks whether
         *  - group
         *  - assert
         *  - win
         *  are valid.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether group is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_GROUP_NULL
         *           - MARMOT_ERR_GROUP_NOT_VALID
         */
        virtual int checkGroup();

        /** Checks whether assert is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_ASSERT_NOT_VALID
         */
        virtual int checkAssert();

        /** Checks whether window is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_WIN_NULL
         *           - MARMOT_ERR_WIN_NOT_VALID
         */
        virtual int checkWin();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        virtual int checkPost();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Win_post
         */
        virtual int callNativeMPI();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        MPO_Group my_group;
        int my_assert;
        MPO_Win mpo_win_post_Win;

    };                          // end class MPO_Win_post
#endif /*IS_MPI_WIN_POST_PRESENT*/

#ifdef IS_MPI_WIN_WAIT_PRESENT
    //************************************************************************
    //
    //                          MPO_Win_wait
    //
    /** This class represents MPI_Win_wait. */
    class MPO_Win_wait : public MPO_Winhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Win_wait();

        /** Constructor taking proper arguments.
         *
         *  @param  win             window object (handle)
         */
        MPO_Win_wait(MPO_Win win);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Win_wait();

        //==========================
        // Methods
        //==========================

        /** Checks prerequisites such as correctness of parameters.
         *  Checks whether
         *  - win
         *  is valid.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        virtual int checkPost();

        /** Checks whether window is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_WIN_NULL
         *           - MARMOT_ERR_WIN_NOT_VALID
         */
        virtual int checkWin();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Win_wait
         */
        virtual int callNativeMPI();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        MPO_Win mpo_win_wait_Win;

    };                          // end class MPO_Win_wait
#endif /*IS_MPI_WIN_WAIT_PRESENT*/


#ifdef IS_MPI_WIN_TEST_PRESENT
    //************************************************************************
    //
    //                          MPO_Win_test
    //
    /** This class represents MPI_Win_test. */
    class MPO_Win_test : public MPO_Winhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object.
         *
         *  @param  win             window object (handle)
         *  @param  flag            success flag (logical)
         */
        MPO_Win_test();

        /** Constructor taking proper arguments. */
        MPO_Win_test(MPO_Win win, int *flag);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Win_test();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks prerequisites such as correctness of parameters.
         *  Checks whether
         *  - win
         *  is valid.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        virtual int checkPost();

        /** Checks whether window is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_WIN_NULL
         *           - MARMOT_ERR_WIN_NOT_VALID
         */
        virtual int checkWin();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Win_test
         */
        virtual int callNativeMPI();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        MPO_Win mpo_win_test_Win;
        int *my_flag;

    };                          // end class MPO_Win_test
#endif /*IS_MPI_WIN_TEST_PRESENT*/

#ifdef IS_MPI_WIN_LOCK_PRESENT
    //************************************************************************
    //
    //                          MPO_Win_lock
    //
    /** Checks prerequisites such as correctness of parameters.
     *
     *  @return  unique error code set by enum MARMOT_ERR_CODES:
     *           - MARMOT_SUCCESS
     *           - MARMOT_ERR_RANK
     *           - MARMOT_ERR_WIN
     */


    /** This class represents MPI_Win_lock. */
    class MPO_Win_lock : public MPO_Winhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Win_lock();

        /** Constructor taking proper arguments.
         *
         *  @param  lock_type       indicates whether other processes may
         *                          access the target window at the same time
         *                          (if 'MPI_LOCK_SHARED') or not
         *                          ('MPI_LOCK_EXCLUSIVE')
         *  @param  rank            rank of locked window (nonnegative integer)
         *  @param  assert          Used to optimize this call;
         *                          zero may be used as a default.
         *  @param  win             window object (handle)
         */
        MPO_Win_lock(int lock_type, int rank, int assert, MPO_Win win);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Win_lock();

        //==========================
        // Methods
        //==========================

        /** Checks whether locktype is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_LOCKTYPE_NOT_VALID
         */
        virtual int checkLockType();

        /** Checks whether rank is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_RANK_NEG
         *           - MARMOT_ERR_RANK_ANY_SOURCE
         *           - MARMOT_ERR_RANK_PROC_NULL
         *           - MARMOT_ERR_RANK_TOO_BIG
         */
        virtual int checkRank();

        /** Checks whether assert is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_ASSERT_NOT_VALID
         */
        virtual int checkAssert();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Win_lock
         */
        virtual int callNativeMPI();

        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        int my_lock_type;
        int my_rank;
        int my_assert;
        // MPO_Win my_win;

    };                          // end class MPO_Win_lock
#endif /*IS_MPI_WIN_LOCK_PRESENT*/


#ifdef IS_MPI_WIN_UNLOCK_PRESENT
    //************************************************************************
    //
    //                          MPO_Win_unlock
    //
    /** This class represents MPI_Win_unlock. */
    class MPO_Win_unlock : public MPO_Winhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Win_unlock();

        /** Constructor taking proper arguments.
         *
         *  @param  rank            rank of window (nonnegative integer)
         *  @param  win             window object (handle)
         */
        MPO_Win_unlock(int rank, MPO_Win win);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Win_unlock();

        //==========================
        // Methods
        //==========================

        /** Checks whether rank is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_RANK_NEG
         *           - MARMOT_ERR_RANK_ANY_SOURCE
         *           - MARMOT_ERR_RANK_PROC_NULL
         *           - MARMOT_ERR_RANK_TOO_BIG
         */
        virtual int checkRank();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Win_unlock
         */
        virtual int callNativeMPI();

        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        int my_rank;
        // MPO_Win my_win;

    };                          // end class MPO_Win_unlock
#endif /*IS_MPI_WIN_UNLOCK_PRESENT*/

}


// end of MARMOT_1SIDED
#endif /*MPO_1SIDED_H*/
