#ifndef MPO_2_EXTCOLL_H
#define MPO_2_EXTCOLL_H

#include "config.h"
#include <iostream>
#include <string>
using std::string;

#include <mpi.h>
#include "mpi_undef_makros.h"

#include "mpo.h"


namespace MPO
{

#ifdef IS_MPI_ALLTOALLW_PRESENT
    //*************************************************************************
    //
    //                          MPO_Alltoallw
    //
    /** This class represents MPI_ Alltoallw. */
    class MPO_Alltoallw : public MPO_Collective
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Alltoallw();

        /** Constructor taking proper arguments. */
        MPO_Alltoallw(void *sendbuf, int *sendcounts, int *sdispls,
                      MPI_Datatype* sendtypes, void *recvbuf, int *recvcounts,
                      int *rdispls, MPI_Datatype* recvtypes, MPO_Comm comm);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Alltoallw();

        //==========================
        // Methods
        //==========================

        /** Checks whether sizes match. */
        virtual int checkPrerequisites2(int *counts,
                                        MPI_Datatype* datatype,
                                        int type);

        /** Checks array of displacements.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES:
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_DISPLACEMENT_OVERRIDES_DATA
         *           - MARMOT_NOTE_DISPLACEMENT_READS_DATA
         */
        virtual int checkDisplacementArrayOfReceiver();

        virtual int checkDatatypeOfSender(int *counts, MPI_Datatype* datatype,
                                          int type);

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Alltallv
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_sendbuf;
        int *my_sendcounts;
        int *my_sdispls;
        MPI_Datatype* my_sendtypes;
        void *my_recvbuf;
        int *my_recvcounts;
        int *my_rdispls;
        MPI_Datatype* my_recvtypes;
        //MPO_Comm my_comm;

    };                          // end class MPO_Alltoallw
#endif /*IS_MPI_ALLTOALLW_PRESENT*/

#ifdef IS_MPI_EXSCAN_PRESENT
    //*************************************************************************
    //
    //                          MPO_Exscan
    //
    /** This class represents MPI_Exscan. */
    class MPO_Exscan : public MPO_Collective
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Exscan();

        /** Constructor taking proper arguments. */
        MPO_Exscan(void *sendbuf, void *recvbuf, int count,
                 MPO_Datatype datatype,
                 MPO_Op op, MPO_Comm comm);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Exscan();

        //==========================
        // Methods
        //==========================

        /** Checks correctness of parameters.
         *  Checks if
         *  - count
         *  - datatype
         *  - operation
         *  are valid.
         *
         *  @return  0
         */
        virtual int checkPrerequisites2();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Exscan
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_sendbuf;
        void *my_recvbuf;
        int my_count;
        MPO_Datatype my_datatype;
        MPO_Op my_op;
        //MPO_Comm my_comm;

    };                          // end  class MPO_Exscan
#endif /*IS_MPI_EXSCAN_PRESENT*/

}                               // end of namespace MPO

#endif /*MPO_2_EXTCOLL_H*/

