/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Declaration of all classes representing MPI-2 calls for file I/O and
 *  one-sided communication.
 *
 *  The calls are currently in the same order as in the MPI-2 standard, I/O
 *  chapter 9, p. 209 pp, one-sided, chapter 6, 109pp. Unfortunately, calls
 *  for I/O and one-sided communication are also spread over the MPI-2
 *  standard document, e.g. errhandlers, etc.
 *
 *  @author Bettina Krammer
 *
 *
 *  $Id: mpo-2-objects.h 1112 2010-04-19 11:09:52Z tobias $
 */

#ifndef MPO_2_OBJECTS_H
#define MPO_2_OBJECTS_H

#include "config.h"
#include <iostream>
#include <string>
using std::string;

#include <mpi.h>
#include "mpi_undef_makros.h"


#include "mpo-objects.h"

/*TODO: IO completly disabled ATM*/
//#include "mpo-io.h"

#ifdef MARMOT_1SIDED
#include "mpo-1sided.h"
#endif

#include "mpo-2-misc.h"

#include "mpo-2-extcoll.h"

namespace MPO
{

#ifdef IS_MPI_ALLOC_MEM_PRESENT
    //************************************************************************
    //
    //                          MPO_Alloc_mem
    //
    /** This class represents MPI_Alloc_mem. It is related to MPI 1-sided, but is strictly not part of the MPI 1-sided communication, so it is declared here */
    class MPO_Alloc_mem : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Alloc_mem();

        /** Constructor taking proper arguments.
         *  @param  size             [in] size of memory segment in bytes (nonnegative integer)
         *  @param  info             [in] info argument (handle)
         *  @param  baseptr          [out] pointer to a pointer to the beginning of the memory segment allocated
         */
        MPO_Alloc_mem(MPI_Aint size, MPI_Info info, void *baseptr);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Alloc_mem();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        virtual int checkPrerequisites();

        virtual int checkPost();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Win_complete
         */
        virtual int callNativeMPI();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================
        MPI_Aint m_size;
        MPI_Info m_info;
        void *m_baseptr;
    };                          // end class MPO_Alloc_mem
#endif /*IS_MPI_ALLOC_MEM_PRESENT*/


#ifdef IS_MPI_FREE_MEM_PRESENT
    //************************************************************************
    //
    //                          MPO_Free_mem
    //
    /** This class represents MPI_Free_mem. */
    class MPO_Free_mem : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Free_mem();

        /** Constructor taking proper arguments.
         *  @param  baseptr          [in] initial address of memory segment allocated by MPI_ALLOC_MEM
         */
        MPO_Free_mem(void *baseptr);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Free_mem();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        virtual int checkPrerequisites();

        virtual int checkPost();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Win_complete
         */
        virtual int callNativeMPI();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================
        void *m_baseptr;
    };                          // end class MPI_Free_mem
#endif /*IS_MPI_FREE_MEM_PRESENT*/

#ifdef MARMOT_THREADED

    //*************************************************************************
    //
    //                          MPO_Init_thread
    //
    /** This class represents MPI_Init_thread. */
    class MPO_Init_thread : public MPO_Init
      {
      public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object.
         *  It should never be called anyhow, because MPI_Init is only called
         *  once. And the corresponding objects are never transferred.
         * \note we can`t use the MPI_Init constructor here because it would call PMPI_Init
         */
        MPO_Init_thread();

        /** Constructor taking proper arguments.
         *
         *  @param   argc - Pointer to the number of arguments
         *  @param   argv - Pointer to the argument vector
         *  @see     mk_fortran_wrapper.sh takes care that argument list is
         *           recovered when using mpich-p4.
         */
        MPO_Init_thread(int *argc, char ***argv,int required, int *provided);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Init_thread();

        //==========================
        // Methods
        //==========================


// TODO // currently no checks implemented.
        /** Checks correctness of parameters etc.
         *  Currently no checks implemented.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  my_rvalue
         */
        virtual int callNativeMPI();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** We override callDebugServer(), because we need special treatment
         *  in MPI_Init_thread.
         *
         *  @return  MPI_SUCCESS
         */
        virtual int callDebugServer();

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    protected:
        //==========================
        // Data
        //==========================

        int *my_provided;
        int  my_required;

    };                          // end class MPO_Init_thread

#else

    //*************************************************************************
    //
    //                          MPO_Init_thread
    // WE will use this dummy class for MPI_Init_thread. This is done to
    // aviod the necessity of 2 different Wrapper libraries (one with
    // a MPI_Init_thread function and one without)
    /** This class represents MPI_Init_thread. */
    class MPO_Init_thread : public MPO_Init
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** Constructor taking proper arguments.
         *
         *  @param   argc - Pointer to the number of arguments
         *  @param   argv - Pointer to the argument vector
         *  @see     mk_fortran_wrapper.sh takes care that argument list is
         *           recovered when using mpich-p4.
         */
        MPO_Init_thread(void);
        MPO_Init_thread(int *argc, char ***argv,int required, int *provided);
    };                          // end class MPO_Init_thread

#endif
// end of MARMOT_THREADED

}                               // end of namespace MPO

#endif /*MPO_2_OBJECTS_H*/
