/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Declaration of DebugServer class.
 *
 *  @bug According to MPI-1.2, it is allowed to call MPI_Get_version and
 *       MPI_Initialized before MPI_Init/after MPI_Finalize.
 *       According to MPI-2.0, it is additionally allowed to call
 *       MPI_Finalized before MPI_Init/after MPI_Finalize.
 *       Currently we can't check calls that are issued
 *       before MPI_Init/after MPI_Finalize.
 *       In fact, calls after MPI_Finalize lead
 *       to an abortion of the program if run with MARMOT.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller
 *
 *  $Id: mpo-debugserver.h 921 2009-02-09 15:54:33Z tobias $
 */

//bk// To be done: think about setting client states and timeouts..
// Accorcing to MPI-1.2, it is allowed to call MPI_Get_version and
// MPI_Initialized before MPI_Init/after MPI_Finalize.
// According to MPI-2.0, it is additionally allowed to call MPI_Finalized
// before MPI_Init/after MPI_Finalize.
// Currently we can't check calls that are issued
// before MPI_Init/after MPI_Finalize. In fact, calls after MPI_Finalize lead
// to an abortion of the program if run with MARMOT.


#ifndef MPO_DEBUGSERVER_H
#define MPO_DEBUGSERVER_H

#include <string>
#include <map>
#include <set>
#include <iostream>

#include "mpo-messenger.h"
#include "mpo-enums.h"
#include "mpo-envvar.h"

namespace MPO
{

//************************************************************************
//
//                          Declarations
//
class MPO_LogFile;

//***************************************************************************
//
//                              DebugServer
//
/** This class represents the server for debugging.
 *
 *  The debug server has several (global) tasks:
 *  - get/set environment variables for MAX_TIMEOUT, TRACE_CALLS,
 *    MAX_PEND_COUNT, SERIALIZE.
 *  - control the execution flow
 *  - detect deadlocks
 *  - check matching send/receive pairs for consistency
 *  - print statements to log execution.
 *
 *  Local tasks like verification of resources are performed on the clients.
 */
    class DebugServer
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** Constructor sets up server with provided communication_channel.
         */
        DebugServer(Messenger & communication_channel, char *ExeName);

        /** Destructor. */
        ~DebugServer();

        //==========================
        // Methods
        //==========================

        /** Gets/sets and prints all our environment variables used for
         *  configuration.
         *
         *  @return  0
         */
        int setAndPrintEnvironment() const;

        /** Prints the type of the utilized Syncronisation object to the log.
         *
         *  @return  0
         */
        void logThreadSyncInfo(void) const;

        /** Overloads << to print environment.
         *  Gets/sets and prints all our environment variables used for
         *  configuration.*/
        std::ostream & setAndPrintEnvironment(std::ostream &output) const;

        /** Sets the output of the error strings.
         *
         *  @return  0
         */
        static int setErrorMessages();

        /** Gets the output of the error strings.
         *
         *  @param   errorcodes unique error code set by enum MARMOT_ERR_CODES
         *  @return  0
         */
        static int getErrorMessages(int errorcodes);

        /** Overloads << to print error message. */
		//made this static to make it available everywhere
		//this method has no link to the internal state of the debugserver
        static std::ostream& outputErrorMessages(int errorcodes,
                                           std::ostream &output);


        /** The loop will never return, but exit from the application
         *  on request.
         */
        void runMainLoop();


    public:
        /** Does nothing but indicates that error has been detected.
         *
         *  @return  0
         */
        static int insertBreakpointMarmotError(const char *message);

        /** Does nothing but indicates that warning has been detected.
         *
         *  @return  0
         */
        static int insertBreakpointMarmotWarning(const char *message);

        /** Does nothing but indicates that note has been detected.
         *
         *  @return  0
         */
        static int insertBreakpointMarmotNote(const char *message);

    private:
       /** Initializes the Logfile of the DebugServer.
        * Must be called before any Message can be recieved.
        * @return always 0.
        */
       int initLogFile (void);


    private:
        //==========================
        // Data
        //==========================

        Messenger & my_channel;
        MPO_LogFile *my_log;
        std::string myExeName;

        /** Map to get error strings for output. */
        static
        std::map < enum MARMOT_ERR_CODES, std::string > MARMOT_ERR_STRINGS;
		static bool myErrorsSet;

    };                          // end class DebugServer


}                               // end of namespace MPO

#endif
