/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Declaration of all the environment variables we use in our namespace MPO
 *  and of all other global variables.
 *
 *  All our environment variables are declared here:
 *  - DEBUG_MODE, INTERFACE_MODE are set in the constructor of MPO_Init.
 *  - MAX_TIMEOUT, TRACE_CALLS, MAX_PEND_COUNT, SERIALIZE 
 *    (and maybe other test variables such as MAX_TIME_OUT_ONE etc.)
 *    are set in the constructor of DebugServer.
 *  All other global variables (e.g. MSG_FOO) are also declared here and 
 *  defined in mpo-envvar.cc.
 *
 *  HINT FOR DEVELOPERS:
 *  ==================== 
 *  DON'T USE HARD-CODED VALUES, DON'T USE MAGIC NUMBERS, 
 *  THEY ARE EVIL AND MAKE HORRIBLE CODE!
 *  Use descriptive enums or variable names and try to keep the complexity 
 *  of check functions low.
 *  If you declare a new environment variable, put it into THIS file.
 *
 *  @author Bettina Krammer, Tobias Hilbrich
 *
 *  $Id: mpo-envvar.h 838 2008-06-13 10:44:34Z hpchimml $
 */


#ifndef MPO_ENVVAR_H
#define MPO_ENVVAR_H

#include <string>
using std::string;

#include "mpi.h"
#include "mpi_undef_makros.h"


namespace MPO
{

    //==========================
    // DEBUG_MODE                          
    //==========================

    /** Global variable for debug-mode.
     *
     *  Global variable for debug-mode:
     *  -  negative value: Nothing is reported.
     *  -  values from enum MPO_DEBUG_MODE: level of verbosity
     *  -  default: DEBUG_MODE_DEFAULT
     *
     *  HINT FOR DEVELOPERS:
     *  ==================== 
     *  Don't use hard-coded values in our check functions, but the variables
     *  MSG_ERR, MSG_ERR_WARN and MSG_ERR_WARN_NOTE from enum MPO_DEBUG_MODE
     *  inside the conditions:
     *  -  MSG_ERR <= DEBUG_MODE
     *     i.e. 0 <= DEBUG_MODE,
     *     i.e. an error message is issued
     *  -  MSG_ERR_WARN <= DEBUG_MODE
     *     i.e. 1 <= DEBUG_MODE,
     *     i.e. a warning message is issued
     *  -  MSG_ERR_WARN_NOTE <= DEBUG_MODE
     *     i.e. 2 <= DEBUG_MODE,
     *     i.e. a remark message is issued
     *  Start with the variable MSG_*, this makes the code easier 
     *  to read.
     *
     *  @see  enum MPO_DEBUG_MODE
     *  @see  MPO_Init
     *  @see  DebugServer::setAndPrintEnvironment()
     */
    extern int DEBUG_MODE;

    /** Global variable for default value of DEBUG_MODE. */
    extern const int DEBUG_MODE_DEFAULT;


    //==========================
    // INTERFACE_MODE                          
    //==========================

    /** Global variable for interface-mode.
     *
     *  Global variable for interface-mode:
     *  -  values from enum MPO_INTERFACE_MODE: 
     *     C or Fortran interface (Fortran interface is set automatically 
     *     by mk_fortran_wrapper.sh)
     *  -  default: INTERFACE_MODE_DEFAULT
     *
     *  HINT FOR DEVELOPERS:
     *  ====================
     *  Don't use hard-coded values in our check functions, but the variables
     *  C_INTERFACE and F_INTERFACE from enum MPO_INTERFACE_MODE 
     *  inside the conditions:
     *  -  C_INTERFACE == INTERFACE_MODE
     *  -  F_INTERFACE == INTERFACE_MODE
     *  Start with the variable *_INTERFACE, this makes the code easier 
     *  to read.
     *
     *  @see enum MPO_INTERFACE_MODE
     *  @see  MPO_Init
     *  @see  DebugServer::setAndPrintEnvironment()
     */
    extern int INTERFACE_MODE;

    /** Global variable for default value of INTERFACE_MODE. */
    extern const int INTERFACE_MODE_DEFAULT;

//clean// for the cleaned up version we use instead of MAX_TIMEOUT
//clean// MAX_TIMEOUT_DEADLOCK  for MAX_TIMEOUT_SEVEN
//clean// MAX_TIMEOUT_SERIALIZE for MAX_TIMEOUT_SIX

//clean//  //! global variable for MAX_TIMEOUT
//clean//  /*! maximum message time in microseconds:
//clean//   *  default 1000 
//clean//   */
//clean//  extern long MAX_TIMEOUT;


    //==========================
    // SERIALIZE                       
    //==========================

    /** Global variable for SERIALIZE.
     *  Sets whether code is serialized or not:
     *  - SERIALIZE = 0: code is not serialized
     *  - SERIALIZE = 1: code is serialized
     *  - default   = SERIALIZE_DEFAULT.
     */
    extern int SERIALIZE;

    /** Global variable for default value of SERIALIZE. */
    extern const int SERIALIZE_DEFAULT;


    //==========================
    // TRACE_CALLS                      
    //==========================

    /** Global variable for TRACE_CALLS.
     *  Set whether calls are traced or not:
     *  - TRACE_CALLS = 1:  calls are traced with output to stderr, 
     *                      traceback in case of a deadlock is possible
     *  - TRACE_CALLS = 0:  calls are traced without output to stderr, 
     *                      traceback in case of a deadlock is possible
     *  - TRACE_CALLS = -1: calls are not traced, 
     *                      traceback in case of a deadlock is NOT possible)
     *  - default = TRACE_CALLS_DEFAULT 
     *    (number of calls to be traced back in case of deadlock 
     *    can be set via the environmental variable MAX_PEND_COUNT).
     */
    extern long TRACE_CALLS;

    /** Global variable for default value of TRACE_CALLS. */
    extern const long TRACE_CALLS_DEFAULT;
    
    
    //==========================
    // LOGFILE_TYPE                      
    //==========================
    /** Global variable for the type of Logfile.
     *  For logfile types see MPO_LOGFILE_TYPE (enum).
     *  Default is LOGFILE_TYPE_ASCII.
     */
    extern int LOGFILE_TYPE;
          
    /** Global variable for default value of LOGFILE_TYPE. */
    extern const int LOGFILE_TYPE_DEFAULT;
    
    //==========================
    // LOGFLUSH_TYPE                      
    //==========================
    /** Global variable used to specify wheter to log 
     *  everything immediately or not.
     */
    extern int LOGFLUSH_TYPE;
          
    /** Global variable for default value of LOGFILE_TYPE. */
    extern const int LOGFLUSH_TYPE_DEFAULT;
    
    //==========================
    // LOGFILTER_COUNT                      
    //==========================
    /** Global variable used to specify the amount of
     * occurrences needed to filter a certain message out. 
     */
    extern int LOGFILTER_COUNT;
             
    /** Global variable for default value of LOGFILTER_COUNT. */
    extern const int LOGFILTER_COUNT_DEFAULT;
    
    //==========================
    // RESOURCE_TRACE_TYPE                      
    //==========================
    /** Global variable for resource tracing.
     *  For types see MPO_RESOURCE_TRACE_SELECTION (enum).
     *  Default is RESOURCE_TRACE_ON.
     */
    extern int RESOURCE_TRACE_TYPE;
          
    /** Global variable for default value of RESOURCE_TRACE_TYPE. */
    extern const int RESOURCE_TRACE_TYPE_DEFAULT;
    
    //==========================
    // MAX_PEND_COUNT                       
    //==========================

    /** Global variable for MAX_PEND_COUNT.
     *  Maximum number of MPI calls that are traced back:
     *  default = MAX_PEND_COUNT_DEFAULT.
     */
    extern long MAX_PEND_COUNT;

    /** Global variable for default value of MAX_PEND_COUNT. */
    extern const long MAX_PEND_COUNT_DEFAULT;


    //==========================
    // MAX_TIMEOUT_DEADLOCK                         
    //==========================

    /** Global variable for MAX_TIMEOUT_DEADLOCK.
     *  Maximum message time in microseconds before deadlock warning is issued:
     *  default = MAX_TIMEOUT_DEADLOCK_DEFAULT. 
     */
    extern long MAX_TIMEOUT_DEADLOCK;

    /** Global variable for default value of MAX_TIMEOUT_DEADLOCK, 
     *  in microseconds.
     */
    extern const long MAX_TIMEOUT_DEADLOCK_DEFAULT;


    //==========================
    // MAX_TIMEOUT_SERIALIZE                       
    //==========================

    /** Global variable for MAX_TIMEOUT_SERIALIZE.
     *  Maximum message time in microseconds in case of serialization:
     *  default = MAX_TIMEOUT_SERIALIZE_DEFAULT.  
     */
    extern long MAX_TIMEOUT_SERIALIZE;

    /** Global variable for default value of MAX_TIMEOUT_SERIALIZE,
     *  in microseconds.
     */
    extern const long MAX_TIMEOUT_SERIALIZE_DEFAULT;


    //===========================================
    // additional variables for MAX_TIMEOUT 
    // to test performance and deadlock detection
    //===========================================

    // everything related to these test variables is commented out 
    // with "//test p&d//" to switch them off

    /** Global variable for sleeping time in microseconds.
     */
    extern long MAX_TIMEOUT_ONE;

    /** Global variable for default value of MAX_TIMEOUT_ONE in microseconds. 
     */
    extern const long MAX_TIMEOUT_ONE_DEFAULT;

    /** Global variable for sleeping time in microseconds.
     */
    extern long MAX_TIMEOUT_TWO;

    /** Global variable for default value of MAX_TIMEOUT_TWO in microseconds. 
     */
    extern const long MAX_TIMEOUT_TWO_DEFAULT;

//test p&d//   extern long MAX_TIMEOUT_THREE;
//test p&d//   extern long MAX_TIMEOUT_FOUR;
//test p&d//   extern long MAX_TIMEOUT_FIVE;
//test p&d//   extern long MAX_TIMEOUT_SIX;
//test p&d//   extern long MAX_TIMEOUT_SEVEN;
//  extern long MAX_FLIGHT_TIME;
//  extern long MAX_TIMEOUT_FIRST;
//  extern long MAX_TIMEOUT;


    //========================================================================
    // Constant strings MSG_FOO for output   
    //======================================================================== 
    // These variables replace the magic numbers that had been used previously
    // in the switch statements of our check functions.  
    // We use them to simplify the output, because sometimes we specify the
    // kind of parameter, e.g. the kind of count etc.               
    //========================================================================

    //==========================
    // Kind of message                      
    //==========================

    /** Global variable for the kind of message that is used ("ERROR"). 
     */
    extern const string MSG_ERROR;

    /** Global variable for the kind of message that is used ("WARNING"). 
     */
    extern const string MSG_WARNING;

    /** Global variable for the kind of message that is used ("NOTE"). 
     */
    extern const string MSG_NOTE;


    //==========================
    // Kind of count                       
    //==========================

    /** Global variable for the kind of count that is used ("Count"). 
     */
    extern const string MSG_COUNT;      // used to be magic number 0

    /** Global variable for the kind of count that is used ("Send count"). 
     */
    extern const string MSG_SENDCOUNT;  // used to be magic number 1

    /** Global variable for the kind of count that is used ("Receive count"). 
     */
    extern const string MSG_RECVCOUNT;  // used to be magic number 2

    /** Global variable for the kind of count that is used ("Incount"). 
     */
    extern const string MSG_INCOUNT;

    /** Global variable for the kind of count that is used ("Outcount"). 
     */
    extern const string MSG_OUTCOUNT;


    //==========================
    // Kind of size                      
    //==========================

    /** Global variable for the kind of size that is used ("Insize"). 
     */
    extern const string MSG_INSIZE;

    /** Global variable for the kind of size that is used ("Outsize"). 
     */
    extern const string MSG_OUTSIZE;


    //==========================
    // Kind of tag                       
    //==========================

    /** Global variable for the kind of count that is used ("Tag"). 
     */
    extern const string MSG_TAG;

    /** Global variable for the kind of count that is used ("Send tag"). 
     */
    extern const string MSG_SENDTAG;

    /** Global variable for the kind of count that is used ("Receive tag"). 
     */
    extern const string MSG_RECVTAG;


    //==========================
    // Kind of new group                       
    //==========================

    /** Global variable for the kind of new group that is created. 
     */
    extern const string MSG_NEW_GROUP_EQ_OLD_GROUP;

    /** Global variable for the kind of new group that is created. 
     */
    extern const string MSG_NEW_GROUP_EQ_EMPTY;

    extern string  MARMOT_LOGFILE_PATH;

    extern const string MARMOT_LOGFILE_PATH_DEFAULT;


}                               // end of namespace MPO 

#endif
