/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Declarations of MPO_State, CallTimestampPair and MPO_Clientstate.
 *
 *  This file contains all declarations of classes that take care of 
 *  state management of the clients.
 *
 *  @see mpo-state.cc (implementation)
 * 
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo-state.h 838 2008-06-13 10:44:34Z hpchimml $
 */


#ifndef MPO_STATE_H
#define MPO_STATE_H

#include <memory>
#include <queue>
#include <iostream>

#include "mpo.h"
#include "mpo-timer.h"
#include "mpo_SourceLocation.h"

namespace MPO
{

    //*************************************************************************
    //
    //                          MPO_State
    //   
    /** This class represents the state of an MPI process.
     *
     *  @see mpo-state.c (implementation)
     */
    class MPO_State : public MPO_Object
    {
    public:
        //==========================
        // Constructors/Destructors                            
        //==========================

        /** Constructor taking proper arguments. */
        MPO_State(MPO_STATE_ID state);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_State();

        //==========================
        // Methods                            
        //==========================

        /** Returns the unique family id (the unique enumeration value)  
         *  for the family of stateobjects. 
         */
        virtual int getFamilyId() const;

        /** Returns the unique id (the unique enumeration value) for 
         *  the object.
         */
        virtual int getId() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data                           
        //==========================

        int my_id;

    };                          // end class MPO_State



    //==========================
    // Operators                            
    //==========================

    bool operator==(const MPO_State & lhs, const MPO_State & rhs);
    bool operator!=(const MPO_State & lhs, const MPO_State & rhs);



    //*************************************************************************
    //
    //                          CallTimestampPair
    //   
    /** This class contains the data concerning the timestamp and the 
     *  calling objects.
     *  
     *  The data members are declared protected (used to be declared 
     *  public, but data should never be public). They must be accessible
     *  from within and from outside this class. That's why the classes 
     *  DebugServer and MPO_Clientstate are declared friends.
     * 
     *  
     *  @see mpo-state.cc (implementation)
     *  @see mpo-debugserver.cc (implementation of friend class)
     */
    class CallTimestampPair
    {
        friend class DebugServer;
        friend class MPO_Clientstate;

    public:
        //==========================
        // Constructors/Destructors                            
        //==========================

        /** The default constructor initializes an invalid object. */
        CallTimestampPair();

        /** Destructor. */
        ~CallTimestampPair();

    protected:
        //==========================
        // Data                           
        //==========================

        MPO_Callobject * my_call_p;
        MPO_Timestamp my_timestamp;
        MPO_SourceLocation my_location;

    };                          // end class CallTimestampPair



    //*************************************************************************
    //
    //                          MPO_Clientstate
    //   
    /** This class represents the detailed state of a client. 
     *
     *  This class takes the responsibility of the object. 
     *  We don't derive it from MPO_Object, because it is not intended to 
     *  send it between clients and server.
     *  We also forbid to copy and assign these objects by declaring 
     *  the copy constructor and assignment operator private.
     *  
     *  @see mpo-state.cc (implementation)
     */
    class MPO_Clientstate
    {
    public:
        //==========================
        // Constructors/Destructors                            
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Clientstate();

        /** Constructor taking proper arguments. */
        MPO_Clientstate(const MPO_Clientstate & rhs);

        /** Destructor. */
        ~MPO_Clientstate();

        //==========================
        // Methods                            
        //==========================

        /** Returns state of client. */
        MPO_State getState() const;

        /** Returns name of call. */
        const MPO_Callobject & getClientname() const;

        /** Returns timestamp. */
        MPO_Timestamp getTimestamp() const;

        /** Sets new state. */
        void setState(MPO_State new_state);

        /** Sets new call_pair.
         *
         *  We take the full responsibility of the object. 
         *  Every time we receive a new object we delete the old one. 
         */
        void setCallTimestampPair(CallTimestampPair new_call_pair);

        //==========================
        // Operators                            
        //==========================

        /** Assignment operator. */
        const MPO_Clientstate & operator=(const MPO_Clientstate & rhs);

    private:
        //==========================
        // Data                           
        //==========================

        MPO_State my_state;
        CallTimestampPair my_call_pair;

    };                          // end class MPO_Clientstate

}                               // end of namespace MPO

#endif
