/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 * @file
 * This Headerfile describes the MPO_GenericLogFilter class.
 * This class filters messages and passes all non filtered ones to another MPO_LogFile
 * implementation.
 * @author Tobias Hilbrich
 * @date 05.10.2007
 * $Id$
 */

#ifndef MPO_GENERICLOGFILTER_H
#define MPO_GENERICLOGFILTER_H

#include "MPO_LogFile.h"
#include <string>
#include <map>

namespace MPO 
{
    /**
     * This class filters messages and passes all non filtered ones to another MPO_LogFile
     * implementation.
     */
    template <class BaseLog>
    class MPO_GenericLogFilter : public BaseLog
    {
    private:
        /**
         * Datatype type for storing the count of msgs. 
         * Using 3 keys:
         *  1. ErrorID
         *  2. Source Line Number
         *  3. Source File Name
         */
        typedef std::map< std::string, int > My_File_Type;
        typedef std::map <int, My_File_Type > My_LineFile_Type;
        typedef std::map< MPO::MARMOT_ERR_CODES, My_LineFile_Type > My_OccurrenceCount_Type;
        
        /**
         * The count of msgs map.
         */
        My_OccurrenceCount_Type my_occs;
            
    public:
        MPO_GenericLogFilter (); /**< Constructor. */
        ~MPO_GenericLogFilter (); /**< Standard destructor. */

    public: 
        bool getMsg (MPO_MsgObject &pMsg); /**< Gets a message object. */
        bool closeLog (void); /**< Prepares the Log for Shutdown. */
        
    };//end of class MPO_LogToServer
    
    // Constructor
    //========================================================================
    template <class BaseLog>
    MPO_GenericLogFilter<BaseLog>::MPO_GenericLogFilter (void)
        : BaseLog ()
    {
        my_occs.clear();
    }
    
    // Destructor
    //========================================================================
    template <class BaseLog>
    MPO_GenericLogFilter<BaseLog>::~MPO_GenericLogFilter (void)
    {
        my_occs.clear();
    }
    
    // getMsg
    //========================================================================
    template <class BaseLog>
    bool MPO_GenericLogFilter<BaseLog>::getMsg (MPO_MsgObject &pMsg)
    {
        //Key 1: The ErrorCode
        My_OccurrenceCount_Type::iterator iErrorCode;
        iErrorCode = my_occs.find(pMsg.getErrorCode());
        
        if (iErrorCode == my_occs.end())
        {
            My_LineFile_Type x;
            x.clear();
            std::pair<My_OccurrenceCount_Type::iterator, bool> result = my_occs.insert (std::make_pair(pMsg.getErrorCode(),x));
            iErrorCode = result.first;
        }
        
        //Key 2: The line number
        My_LineFile_Type::iterator iSourceLine;
        int line = pMsg.getSourceLocation().getCodeline();
        iSourceLine = iErrorCode->second.find(line);
                
        if (iSourceLine == iErrorCode->second.end())
        {
            My_File_Type x;
            x.clear();
            std::pair<My_LineFile_Type::iterator, bool> result = iErrorCode->second.insert (std::make_pair(line,x));
            iSourceLine = result.first;
        }
                
        //key 3: The file name
        My_File_Type::iterator iSourceFile;
        std::string file = pMsg.getSourceLocation().getFileName();
        iSourceFile = iSourceLine->second.find(file);
                
        if (iSourceFile == iSourceLine->second.end())
        {
            std::pair<My_File_Type::iterator, bool> result = iSourceLine->second.insert (std::make_pair(file,0));
            iSourceFile = result.first;
        }
        
        //do the filtering
        int msg_count = iSourceFile->second;
        
        if (msg_count >= LOGFILTER_COUNT)
        {
            // Do not log the msg its amount of occurrences is exceeded.
            return true;
        }
        
        iSourceFile->second = iSourceFile->second + 1; 
        
        if (msg_count == LOGFILTER_COUNT-1)
        {
            //append a info msg to the last occurrence of this msg type
            std::string text = "";
            text+="\n\nThis message will not be repeated on this process as it has exceeded the MARMOT_LOG_FILTER_COUNT limit.\n"; 
            pMsg.appendToText(text);
        }
        
        return BaseLog::getMsg(pMsg);
    }
    
    
    // closeLog
    //========================================================================
    template <class BaseLog>
    bool MPO_GenericLogFilter<BaseLog>::closeLog (void)
    {
        return BaseLog::closeLog();
    }

}//end of namespace MPO

#endif







