/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 * @file
 * This Headerfile describes the MPO_SourceLocation class.
 * A MPO_SourceLocation object describes a source-code Location.
 * @author Tobias Hilbrich
 * @date 24.10.2005
 * $Id: mpo_SourceLocation.h 921 2009-02-09 15:54:33Z tobias $
 */

#ifndef MPO_SOURCELOCATION_H
#define MPO_SOURCELOCATION_H

#include "mpo-enums.h"
#include <string>
#include <iostream>

using std::string;

namespace MPO
{
    /**
     * This object identifies SourceCode locations.
     * A Sourcecode location is identified by a filename and a code line.
     */
    class MPO_SourceLocation //: public MPO_Object /**< this caused lots of trouble ...
    {
    private:

        //==========================
       	// Data
       	//==========================
        char *myFileName; /**< Reference to the filename (Memory will not be deleted). */
        int myLine; /**< Line-number. */
        char *myFileNameMem; /**< Array that stores a filename after a restore Operation (Memory is managed). */

        /**
         * An ID may be assigned to each source location.
         * The ID may be used by external tools to associate a
         * Marmot message (especally global asynchroneous ones)
         * with a specific MPI call.
         */
        unsigned long long my_extra_id;

    public:
        //==========================
        // Constructors/Destructors
        //==========================

        MPO_SourceLocation (char* fileName,int codeline); /**< Constructor initializes information by value. */
        MPO_SourceLocation (char* fileName,int codeline,unsigned long long extra_id); /**< Constructor initializes information by value. */
        MPO_SourceLocation (const char* fileName,int codeline); /**< Constructor initializes information by value. */
        MPO_SourceLocation (const char* fileName,int codeline,unsigned long long extra_id); /**< Constructor initializes information by value. */
        MPO_SourceLocation (std::string fileName,int codeline); /**< Constructor initializes information by value. */
        MPO_SourceLocation (std::string fileName,int codeline,unsigned long long extra_id); /**< Constructor initializes information by value. */
        MPO_SourceLocation (); /**< Constructor initializes as an invalid source location. */

        /**
         * Constructor that initalized this source location with a given exsiting source location.
         * To perform this operation member operator= is used.
         * @param l the source location whose content is used to initalize this object.
         */
        MPO_SourceLocation (const MPO_SourceLocation &l);

        ~MPO_SourceLocation (); /**< Standard destructor. */

        //==========================

        // Methods own
        //==========================
    public:
        const char* getFileName (void) const;
        int getCodeline (void) const;
        bool isValidLocation (void) const;
        unsigned long long getExtraID (void) const;

        //==========================
        // Methods inherited
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        int getId() const;

        /** @return  Family ID. */
        int getFamilyId() const;

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        string getCallname() const;

        /** Overloads << to print call-description. */
        std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        std::istream & restore(std::istream & in);

        /**
         *  Operator for copying a source location.
         * This method manages the memory needes for the file name and
         * creates an independent copy with its own memory.
         * @param l the location that will be copied into this object.
         */
        void operator = (const MPO_SourceLocation &l);

    };//end of class MPO_SourceLocation

}//end of namespace MPO

#endif







