/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 * @file 
 * This Headerfile describes and implements the MPO_TemplateTree template.
 * This template manages objects with ids in a tree and supports some accessors 
 * as they will be needed for CUBE logging.
 * @author Tobias Hilbrich
 * @date 13.11.2005
 * $Id: mpo_TemplateTree.h 922 2009-02-09 17:30:15Z hpczink $
 */

#ifndef MPO_TEMPLATETREE_H
#define MPO_TEMPLATETREE_H

#include <vector>
#include <stack>

namespace MPO 
{
    /** This template class is supposed to manage resources arranged in a tree.
     * It is designed for handling Pointer of objects. When you want also to handle 
     * somthing else than pointers than T should have a working = Operator.
     */
    template <class T>
    class MPO_TemplateTree
    {
        //==========================
        // Typedefs
        //==========================
        //Defines the id type we use
        typedef unsigned long MyIDType;
        
    private:
        //==========================
        // Data
        //==========================
        T   myObject; /**< The object associated with this node/leaf. */
        MyIDType myId; /**< The Id of this Object. */ 
        std::vector<MPO_TemplateTree<T>*> myChilds; /**< The Childs of this Node. */
        
    public:
        //==========================
        // Constructors and Destructors
        //==========================
            
        /** Initializes this Node without data.
         * By Tobias Hilbrich, Date 20.12.2005.
         */
        MPO_TemplateTree () 
            {
                myId = 0;
            }
            
        /** Initializes this Node.
         * By Tobias Hilbrich, Date 13.11.2005.
         * @param object the object associated with this node.
         * @param id the id of this node.
         */
        MPO_TemplateTree (T object,MyIDType id) 
            {
                myObject = object;
                myId = id;
            }
        
        /** Deletes this Node and all its Children.
         * By Tobias Hilbrich, Date 13.11.2005.
         */
        ~MPO_TemplateTree () 
            {
                for (unsigned int i = 0; i < myChilds.size();i++)
                {
                    if (myChilds[i])
                        delete (myChilds[i]);
                    myChilds[i] = NULL;
                }
            }
        
        //==========================
        // Methods
        //==========================
        
        /** Adds a child to this node.
         * By Tobias Hilbrich, Date 13.11.2005.
         * @param object the object associated with the child.
         * @param id the id of the child.
         * @return always true.
         */
        bool addChild (T object,MyIDType id)
            {
                myChilds.push_back(new MPO_TemplateTree<T> (object,id));
                return (true);
            }
        
        /** Returns the object of this node.
         * By Tobias Hilbrich, Date 13.11.2005.
         * @return the object associated with this node, when T is not a pointer type it will be only a copy.
         */
        T getObject (void) 
            {
                return myObject;
            }
        
        /** Returns the id of this node.
         * By Tobias Hilbrich, Date 13.11.2005.
         * @return the object id of this node.
         */
        MyIDType getId (void)
            {
                return myId;
            }
        
        /** Searches the Tree for the node with the given id.
         * By Tobias Hilbrich, Date 13.11.2005.
         * @param id the id of the node you are looking for.
         * @return a pointer to the node with the given id or NULL when no node with this id is present.
         */
        MPO_TemplateTree<T>* searchId (MyIDType id) 
            {
                if (myId == id)
                    return (this);
            
                MPO_TemplateTree<T>* ret;
            
                for (unsigned int i = 0; i < myChilds.size();i++)
                {
                    ret = myChilds[i]->searchId (id);
                
                    if (ret != NULL)
                        return ret;
                }//for childs
            
                return NULL;
            }
        
        /** This Method looks for the node with the given id and stores a path to this node in the given vector.
         * The first element of the vector will be the node with the given id the second its father and so on.
         * When theres no node with the given id the method will return false.
         * By Tobias Hilbrich, Date 13.11.2005.
         * @param id the id of the node you are looking for.
         * @param pPath a pointer to an exsiting std::vector that stores Pointers to MPO_TemplateTree<T> Nodes or NULL.
         * @return true when there is a node with the given id, false otherwise.
         */
        bool getPathToId (MyIDType id,std::vector<MPO_TemplateTree<T>*> *pPath) 
            {
                if (myId == id)
                {
                    if (pPath)
                        pPath->push_back(this);
                    return(true);
                }//If (myId == id)
                    
                for (int i = 0; i < myChilds.size();i++)
                {
                    if (myChilds[i]->getPathToId(id,pPath) == true)
                    {
                        if (pPath)
                            pPath->push_back(this);
                        return (true);
                    }//if Child has path to id
                }//for myChilds
                
                return (false);
            }
        
        /** Deletes all the myObject`s in the tree.
         * Only for pointer types.
         * By Tobias Hilbrich, Date 13.11.2005.
         * @return always true.
         */
        bool deleteAssociatedObjects (void)
            {
                for (unsigned int i = 0; i < myChilds.size();i++)
                {
                    myChilds[i]->deleteAssociatedObjects();
                }
            
                if (myObject)
                    delete (myObject);
                myObject = NULL;
            
                return true;
            }
        
        /** Returns the amount of childs of this node.
         * By Tobias Hilbrich, Date 10.12.2005.
         * @return the amount of childs of this node.
         */
        int getNumChilds (void)
            {
                return myChilds.size();
            }
        
        /** Returns the child with the given index.
         * By Tobias Hilbrich, Date 10.12.2005.
         * @param index the index of the child you want to get.
         * @return the child with the given index or NULL if there is no such child.
         */
        MPO_TemplateTree<T>* getChildIndexed (size_t index)
            {
                if (index >= myChilds.size())
                    return NULL;
            
                return myChilds[index];
            }
        
    };//template MPO_TemplateTree
    
    /** This template class iterates over a MPO_TemplateTree.
     */
    template <class T>
    class MPO_TemplateTreeIterator
    {
    private:
        std::stack<MPO_TemplateTree<T>*> myStack; /**< Stack that stores all the nodes that have to be handled. */
        MPO_TemplateTree<T> *myCurrentNode; /**< The Node that is actually set. */
        
    public:
        /** Constructor.
         * By Tobias Hilbrich, Date 10.12.2005.
         * @param first the root of the tree that shall be iterated.
         */
        MPO_TemplateTreeIterator (MPO_TemplateTree<T>* first) 
            {
                myCurrentNode = first;
            }
        
        /** Returns the next node or NULL when there are not further nodes.
         * By Tobias Hilbrich, Date 10.12.2005.
         * @return the next node or NULL.
         */
        MPO_TemplateTree<T>* next (void)
            {
                MPO_TemplateTree<T>* ret = myCurrentNode;
            
                if (myCurrentNode == NULL)
                    return NULL;
            
                for (int i = myCurrentNode->getNumChilds()-1;i >= 0;i--)
                {
                    myStack.push(myCurrentNode->getChildIndexed(i));
                }
            
                if (myStack.size() > 0)
                {
                    myCurrentNode = myStack.top ();
                    myStack.pop();
                }
                else
                {
                    myCurrentNode = NULL;
                }
            
                return ret;
            }
        
        /** Returns wheter there is a further node.
         * By Tobias Hilbrich, Date 10.12.2005.
         * @return true when ther is still a node left.
         */
        bool hasNext (void)
            {
                if (myCurrentNode == NULL)
                    return false;
                return true;
            }
                  
    };

}//namespace MPO
#endif
