/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of MPO_Accumulate class.
 *
 *  @author Bettina Krammer
 *
 *  $Id: mpo_Accumulate.cc 1112 2010-04-19 11:09:52Z tobias $
 */

#include "config.h"
#include "mpo-2-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"


using namespace MPO;

#ifdef IS_MPI_ACCUMULATE_PRESENT

// default constructor
//========================================================================
MPO_Accumulate::MPO_Accumulate()
    : MPO_WinCommunicator(0,
                          -1,
                          MPI_DATATYPE_NULL,
                          MPI_PROC_NULL,
                          -1,
                          -1,
                          MPI_DATATYPE_NULL,
                          MPI_WIN_NULL),
      mpo_accumulate_Op(MPI_OP_NULL)
{
}


// proper constructor
//========================================================================
MPO_Accumulate::MPO_Accumulate(void *origin_addr,
                              int origin_count,
                              MPO_Datatype origin_datatype,
                              int target_rank,
                              MPI_Aint target_disp,
                              int target_count,
                              MPO_Datatype target_datatype,
                              MPO_Op op,
                              MPO_Win win)
    : MPO_WinCommunicator(origin_addr,
                          origin_count,
                          origin_datatype,
                          target_rank,
                          target_disp,
                          target_count,
                          target_datatype,
                          win),
      mpo_accumulate_Op(op)
{
}


// destructor
//========================================================================
MPO_Accumulate::~MPO_Accumulate()
{
}


// checkOp
//========================================================================
int MPO_Accumulate::checkOp()
{
    // Return value is the unique error code set by enum MARMOT_ERR_CODES.

    int rvalue = MARMOT_SUCCESS;

    // (1) Check if op is null handle.
    if ((our_errcodes.at(MARMOT_ERR_OP_NULL))&&(mpo_accumulate_Op == MPI_OP_NULL) &&
        (MSG_ERR <= DEBUG_MODE))
    {
        std::stringstream stream;
        stream
            << "ERROR: "
            << getCallname()
            << ": Operation is MPI_OP_NULL!"
            << std::endl;
        rvalue = MARMOT_ERR_OP_NULL;

        MPO_String text;
        text.assign(&stream);
        MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
        our_log->getMsg(msg);
    }

    // (2) Check if op is valid, i.e. registered in our_ops.
    if ((our_errcodes.at(MARMOT_ERR_OP_NOT_VALID))&&(our_ops.find(mpo_accumulate_Op) == our_ops.end()) &&
        (MSG_ERR <= DEBUG_MODE))
    {
        std::stringstream stream;
        stream
            << "ERROR: "
            << getCallname()
            << ": Operation not valid!"
            << std::endl;
        rvalue = MARMOT_ERR_OP_NOT_VALID;

        MPO_String text;
        text.assign(&stream);
        MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
        our_log->getMsg(msg);
    }

    return rvalue;
}



// getId
//========================================================================
int MPO_Accumulate::getId() const
{
    // returns the unique call identity set by enum MPO_CALL_ID
    return MPO_ACCUMULATE_ID;
}


int MPO_Accumulate::checkPrerequisites()
{
    int rvalue = MARMOT_SUCCESS;

    if ( MARMOT_SUCCESS != ( rvalue = checkWin()) )
    {
        return rvalue;
    }
    else if( MARMOT_SUCCESS != ( rvalue = checkOp()) )
    {
        return rvalue;
    }

    return rvalue;
}


// checkWin
//========================================================================
int MPO_Accumulate::checkWin()
{
    // Return value is the unique error code set by enum MARMOT_ERR_CODES.
    int rvalue = MARMOT_SUCCESS;

    // (1) Check if window is null handle.
    if ((MPI_WIN_NULL == mpo_accumulate_Win) &&
        (MSG_ERR <= DEBUG_MODE))
    {
        std::stringstream stream;
        stream
            << "ERROR: "
            << getCallname()
            << ": window is MPI_WIN_NULL!"
            << std::endl;
        rvalue = MARMOT_ERR_WIN_NULL;

        MPO_String text;
        text.assign(&stream);
        MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
        our_log->getMsg(msg);
    }

    // (2) Check if window is valid, i.e. registered in our_wins.
    else if ((MPO_Win::our_wins.end() == MPO_Win::our_wins.find(mpo_accumulate_Win)) &&
             (MSG_ERR <= DEBUG_MODE))
    {
        std::stringstream stream;
        stream
            << "ERROR: "
            << getCallname()
            << " : Window is not valid!"
            << std::endl;
        rvalue = MARMOT_ERR_WIN_NOT_VALID;

        MPO_String text;
        text.assign(&stream);
        MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
        our_log->getMsg(msg);
    }

    return rvalue;
}


// callNativeMPI
//========================================================================
int MPO_Accumulate::callNativeMPI()
{
    int rvalue = 1;

#ifdef MARMOT_USE_THREADSYNC
    MPO_Space::getSyncObject()->enterPMPI();
#endif

    rvalue = PMPI_Accumulate(my_origin_addr, my_origin_count, my_origin_datatype, my_target_rank, my_target_disp, my_target_count, my_target_datatype, mpo_accumulate_Op, mpo_accumulate_Win);

#ifdef MARMOT_USE_THREADSYNC
    MPO_Space::getSyncObject()->leavePMPI();
#endif

    return rvalue;
}

// checkPost
//========================================================================
int MPO_Accumulate::checkPost()
{
	//TODO
    return MARMOT_SUCCESS;
}

// getCallname
//========================================================================
string MPO_Accumulate::getCallname() const
{
    // Distinguish between C and Fortran interface, default: C interface.
    string rvalue = "MPI_Accumulate";

    if (F_INTERFACE == INTERFACE_MODE)
    {
        rvalue = "MPI_ACCUMULATE";
    }

    return rvalue;
}


// debugout
//========================================================================
std::ostream & MPO_Accumulate::debugout(std::ostream & output) const
{
    // Distinguish between C and Fortran interface, default: C interface.
    output
      << getCallname()
      << "(*origin_addr"
      << ", origin_count = "
      << my_origin_count
      << ", origin_datatype = "
      << my_origin_datatype
      << ", target_rank = "
      << my_target_rank
      << ", target_disp = "
      << my_target_disp
      << ", target_count = "
      << my_target_count
      << ", target_datatype = "
      << my_target_datatype
      << ", op = "
      << mpo_accumulate_Op
      << ", win = "
      << mpo_accumulate_Win;

    if (F_INTERFACE == INTERFACE_MODE)
    {
        output << ", ierror";
    }

    output << ")" << std::endl;

    return output;
}

// save
//========================================================================
std::ostream & MPO_Accumulate::save(std::ostream & out) const
{
    // we don't save and restore pointers
    MPO_Callobject::save(out);
    // out < my_origin_addr;
    out < my_origin_count;
    out < my_origin_datatype;
    out < my_target_rank;
    out < my_target_disp;
    out < my_target_count;
    out < my_target_datatype;
    out < mpo_accumulate_Op;
    out < mpo_accumulate_Win;

    return out;
}


// restore
//========================================================================
std::istream & MPO_Accumulate::restore(std::istream & in)
{
    // we don't save and restore pointers
    MPO_Callobject::restore(in);
    // in > my_origin_addr;
    in > my_origin_count;
    in > my_origin_datatype;
    in > my_target_rank;
    in > my_target_disp;
    in > my_target_count;
    in > my_target_datatype;
    in > mpo_accumulate_Op;
    in > mpo_accumulate_Win;

    return in;
}


// mpich implementation
//========================================================================
// The following comment is copied from mpich2-1.0/src/mpi/rma/accumulate.c

/*@
   MPI_Accumulate - Accumulate data into the target process using remote
   memory access

   Input Parameters:
+ origin_addr - initial address of buffer (choice)
. origin_count - number of entries in buffer (nonnegative integer)
. origin_datatype - datatype of each buffer entry (handle)
. target_rank - rank of target (nonnegative integer)
. target_disp - displacement from start of window to beginning of target
  buffer (nonnegative integer)
. target_count - number of entries in target buffer (nonnegative integer)
. target_datatype - datatype of each entry in target buffer (handle)
. op - predefined reduce operation (handle)
- win - window object (handle)

   Notes:
The basic components of both the origin and target datatype must be the same
predefined datatype (e.g., all 'MPI_INT' or all 'MPI_DOUBLE_PRECISION').

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COUNT
.N MPI_ERR_RANK
.N MPI_ERR_OP

int MPI_Accumulate(void *origin_addr, int origin_count, MPI_Datatype
                   origin_datatype, int target_rank, MPI_Aint
                   target_disp, int target_count, MPI_Datatype
                   target_datatype, MPI_Op op, MPI_Win win)
@*/

#endif /*IS_MPI_ACCUMULATE_PRESENT*/

/*EOF*/
