/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of MPO_Put class.
 *
 *  @author Bettina Krammer
 *
 *  $Id: mpo_Put.cc 1112 2010-04-19 11:09:52Z tobias $
 */

#include "config.h"
#include "mpo-2-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"


using namespace MPO;

#ifdef IS_MPI_PUT_PRESENT

// default constructor
//========================================================================
MPO_Put::MPO_Put()
    : MPO_WinCommunicator(0,
                          -1,
                          MPI_DATATYPE_NULL,
                          MPI_PROC_NULL,
                          -1,
                          -1,
                          MPI_DATATYPE_NULL,
                          MPI_WIN_NULL)
{
}


// proper constructor
//========================================================================
MPO_Put::MPO_Put(void *origin_addr,
                int origin_count,
                MPO_Datatype origin_datatype,
                int target_rank,
                MPI_Aint target_disp,
                int target_count,
                MPO_Datatype target_datatype,
                MPI_Win win)
    : MPO_WinCommunicator(origin_addr,
                          origin_count,
                          origin_datatype,
                          target_rank,
                          target_disp,
                          target_count,
                          target_datatype,
                          win)
{
    mpo_put_Win = win;
}


// destructor
//========================================================================
MPO_Put::~MPO_Put()
{
}


// getId
//========================================================================
int MPO_Put::getId() const
{
    // returns the unique call identity set by enum MPO_CALL_ID
    return MPO_PUT_ID;
}

// checkPrerequisites
//========================================================================
int MPO_Put::checkPrerequisites()
{
    // Return value is the unique error code set by enum MARMOT_ERR_CODES.
    int rvalue = MARMOT_SUCCESS;

    rvalue = checkWin();

    return rvalue;
}


// checkWin
//========================================================================
int MPO_Put::checkWin()
{
    // Return value is the unique error code set by enum MARMOT_ERR_CODES.
    int rvalue = MARMOT_SUCCESS;

    // (1) Check if window is null handle.
    if ((MPI_WIN_NULL == mpo_put_Win) &&
        (MSG_ERR <= DEBUG_MODE))
    {
        std::stringstream stream;
        stream
            << "ERROR: "
            << getCallname()
            << ": window is MPI_WIN_NULL!"
            << std::endl;
        rvalue = MARMOT_ERR_WIN_NULL;

        MPO_String text;
        text.assign(&stream);
        MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
        our_log->getMsg(msg);
    }

    // (2) Check if window is valid, i.e. registered in our_wins.
    else if ((MPO_Win::our_wins.end() == MPO_Win::our_wins.find(mpo_put_Win)) &&
             (MSG_ERR <= DEBUG_MODE))
    {
        std::stringstream stream;
        stream
            << "ERROR: "
            << getCallname()
            << " : Window is not valid!"
            << std::endl;
        rvalue = MARMOT_ERR_WIN_NOT_VALID;

        MPO_String text;
        text.assign(&stream);
        MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
        our_log->getMsg(msg);
    }

    return rvalue;
}


// callNativeMPI
//========================================================================
int MPO_Put::callNativeMPI()
{
    int rvalue = 1;

#ifdef MARMOT_USE_THREADSYNC
    MPO_Space::getSyncObject()->enterPMPI();
#endif

    rvalue = PMPI_Put(my_origin_addr, my_origin_count, my_origin_datatype, my_target_rank, my_target_disp, my_target_count, my_target_datatype, mpo_put_Win);

#ifdef MARMOT_USE_THREADSYNC
    MPO_Space::getSyncObject()->leavePMPI();
#endif

    return rvalue;
}

// checkPost
//========================================================================
int MPO_Put::checkPost()
{
    //TODO
    return MARMOT_SUCCESS;
}


// getCallname
//========================================================================
string MPO_Put::getCallname() const
{
    // Distinguish between C and Fortran interface, default: C interface.
    string rvalue = "MPI_Put";

    if (F_INTERFACE == INTERFACE_MODE)
    {
      rvalue = "MPI_PUT";
    }

    return rvalue;
}


// debugout
//========================================================================
std::ostream & MPO_Put::debugout(std::ostream & output) const
{
    // Distinguish between C and Fortran interface, default: C interface.
    output
        << getCallname()
        << "(*origin_addr"
        << ", origin_count = "
        << my_origin_count
        << ", origin_datatype = "
        << my_origin_datatype
        << ", target_rank = "
        << my_target_rank
        << ", target_disp = "
        << my_target_disp
        << ", target_count = "
        << my_target_count
        << ", target_datatype = "
        << my_target_datatype
        << ", win = "
        << mpo_put_Win;

    if (F_INTERFACE == INTERFACE_MODE)
    {
    output << ", ierror";
    }

    output << ")" << std::endl;

    return output;
}

// save
//========================================================================
std::ostream & MPO_Put::save(std::ostream & out) const
{
    // we don't save and restore pointers
    MPO_Callobject::save(out);
    // out < my_origin_addr;
    out < my_origin_count;
    out < my_origin_datatype;
    out < my_target_rank;
    out < my_target_disp;
    out < my_target_count;
    out < my_target_datatype;
    out < mpo_put_Win;

    return out;
}


// restore
//========================================================================
std::istream & MPO_Put::restore(std::istream & in)
{
    // we don't save and restore pointers
    MPO_Callobject::restore(in);
    // in > my_origin_addr;
    in > my_origin_count;
    in > my_origin_datatype;
    in > my_target_rank;
    in > my_target_disp;
    in > my_target_count;
    in > my_target_datatype;
    in > mpo_put_Win;

    return in;
}


// mpich implementation
//========================================================================
// The following comment is copied from mpich2-1.0/src/mpi/rma/put.c

/*@
   MPI_Put - Put data into a memory window on a remote process

   Input Parameters:
+ origin_addr -initial address of origin buffer (choice)
. origin_count -number of entries in origin buffer (nonnegative integer)
. origin_datatype -datatype of each entry in origin buffer (handle)
. target_rank -rank of target (nonnegative integer)
. target_disp -displacement from start of window to target buffer (nonnegative integer)
. target_count -number of entries in target buffer (nonnegative integer)
. target_datatype -datatype of each entry in target buffer (handle)

- win - window object used for communication (handle)

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_WIN
.N MPI_ERR_ARG
.N MPI_ERR_TYPE
.N MPI_ERR_COUNT

int MPI_Put(void *origin_addr, int origin_count, MPI_Datatype
            origin_datatype, int target_rank, MPI_Aint target_disp,
            int target_count, MPI_Datatype target_datatype, MPI_Win
            win)
@*/

#endif /*IS_MPI_PUT_PRESENT*/

/*EOF*/

