/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of MPO_Win_complete class.
 *
 *  @author Bettina Krammer
 *
 *  $Id: mpo_Win_complete.cc 1112 2010-04-19 11:09:52Z tobias $
 */

#include "config.h"
#include "mpo-2-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"

#ifdef IS_MPI_WIN_COMPLETE_PRESENT

namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Win_complete::MPO_Win_complete()
        : MPO_Winhandler(MPI_WIN_NULL)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Win_complete::MPO_Win_complete(MPO_Win win)
        : MPO_Winhandler(win)
    {
        mpo_win_complete_Win = win;
    }


    // destructor
    //========================================================================
    MPO_Win_complete::~MPO_Win_complete()
    {
    }


    int MPO_Win_complete::checkPrerequisites()
    {
        int rvalue = checkWin();

        return rvalue;
    }


    int MPO_Win_complete::checkPost(){
        return MARMOT_SUCCESS;
    }


    // checkWin
    //========================================================================
    int MPO_Win_complete::checkWin()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if window is null handle.
        if ((MPI_WIN_NULL == mpo_win_complete_Win) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": window is MPI_WIN_NULL!"
                << std::endl;
            rvalue = MARMOT_ERR_WIN_NULL;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        // (2) Check if window is valid, i.e. registered in our_wins.
        else if ((MPO_Win::our_wins.end() == MPO_Win::our_wins.find(mpo_win_complete_Win)) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << " : Window is not valid!"
                << std::endl;
            rvalue = MARMOT_ERR_WIN_NOT_VALID;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }


    // getId
    //========================================================================
    int MPO_Win_complete::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_WIN_COMPLETE_ID;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Win_complete::callNativeMPI()
    {
        int rvalue = 1;

#ifdef MARMOT_USE_THREADSYNC
        MPO_Space::getSyncObject()->enterPMPI();
#endif

        rvalue = PMPI_Win_complete(mpo_win_complete_Win);

#ifdef MARMOT_USE_THREADSYNC
        MPO_Space::getSyncObject()->leavePMPI();
#endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Win_complete::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Win_complete";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_WIN_COMPLETE";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Win_complete::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(win = "
            << mpo_win_complete_Win;

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Win_complete::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < mpo_win_complete_Win;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Win_complete::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > mpo_win_complete_Win;

        return in;
    }

}                               // end of namespace MPO


// mpich implementation
//========================================================================
// The following comment is copied from mpich2-1.0/src/mpi/rma/win_complete.c

/*@
   MPI_Win_complete - Completes an RMA operations begun after an MPI_Win_start.

   Input Parameter:
. win - window object (handle)

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_WIN
.N MPI_ERR_OTHER

int MPI_Win_complete(MPI_Win win)
@*/

#endif /*IS_MPI_WIN_COMPLETE_PRESENT*/

/*EOF*/
