/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of MPO_Win_create class.
 *
 *  @author Bettina Krammer
 *
 *  $Id: mpo_Win_create.cc 1112 2010-04-19 11:09:52Z tobias $
 */

#include "config.h"
#include "mpo-2-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"


using namespace MPO;

#ifdef IS_MPI_WIN_CREATE_PRESENT

// default constructor
//========================================================================
MPO_Win_create::MPO_Win_create()
    : my_base(0),
      my_size(0),
      my_disp_unit(0),
      my_info(MPI_INFO_NULL),
      my_comm(MPI_COMM_NULL),
      mpo_win_create_Win(0)
{
}


// proper constructor
//========================================================================
MPO_Win_create::MPO_Win_create(void *base,
                               MPI_Aint size,
                               int disp_unit,
                               MPI_Info info,
                               MPO_Comm comm,
                               MPI_Win *win)
    : my_base(base),
      my_size(size),
      my_disp_unit(disp_unit),
      my_info(info),
      my_comm(comm)
{
	mpo_win_create_Win=win;
}


// destructor
//========================================================================
MPO_Win_create::~MPO_Win_create()
{
}


int MPO_Win_create::checkPrerequisites()
{
  int rvalue = MARMOT_SUCCESS;
  // TODO

  rvalue = checkComm();

  return rvalue;
}

int MPO_Win_create::checkPost()
{
  int rvalue = MARMOT_SUCCESS;
  // TODO

  return rvalue;
}

// checkWinSize
//========================================================================
int MPO_Win_create::checkWinSize()
{
  // Return value is the unique error code set by enum MARMOT_ERR_CODES.
  int rvalue = MARMOT_SUCCESS;

  return rvalue;
}


// checkWinDispUnit
//========================================================================
int MPO_Win_create::checkWinDispUnit()
{
  // Return value is the unique error code set by enum MARMOT_ERR_CODES.
  int rvalue = MARMOT_SUCCESS;


  return rvalue;
}


// checkComm
//========================================================================
int MPO_Win_create::checkComm()
{
  // Return value is the unique error code set by enum MARMOT_ERR_CODES.
  int rvalue = MARMOT_SUCCESS;

  // (1) Check if communicator is MPI_COMM_NULL.
  if ((MPI_COMM_NULL == my_comm) &&
      (MSG_ERR <= DEBUG_MODE))
  {
      std::stringstream stream;
      stream
          << "ERROR: "
          << getCallname()
          << ": Communicator is MPI_COMM_NULL!"
          << std::endl;
      rvalue = MARMOT_ERR_COMM_NULL;

      MPO_String text;
      text.assign(&stream);
      MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
      our_log->getMsg(msg);
  }

  // (2) Check if communicator is valid.
  else if ((our_comms.end() == our_comms.find(my_comm)) &&
           (MSG_ERR <= DEBUG_MODE))
  {
      std::stringstream stream;
      stream
          << "ERROR: "
          << getCallname()
          << " : Communicator is not valid!"
          << std::endl;
      rvalue = MARMOT_ERR_COMM_NOT_VALID;

      MPO_String text;
      text.assign(&stream);
      MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
      our_log->getMsg(msg);
  }

  return rvalue;
}                           // end checkComm


// getId
//========================================================================
int MPO_Win_create::getId() const
{
  // returns the unique call identity set by enum MPO_CALL_ID
  return MPO_WIN_CREATE_ID;
}


// callNativeMPI
//========================================================================
int MPO_Win_create::callNativeMPI()
{
  int rvalue = 1;

  #ifdef MARMOT_USE_THREADSYNC
         MPO_Space::getSyncObject()->enterPMPI();
  #endif
  rvalue = PMPI_Win_create(my_base, my_size, my_disp_unit, my_info, my_comm, mpo_win_create_Win);

  MPO_Win::our_wins.insert(std::pair<MPI_Win,mpi_win_marmot_states_t>(*mpo_win_create_Win,MPI_WIN_FREE_ALLOWED));

  #ifdef MARMOT_USE_THREADSYNC
         MPO_Space::getSyncObject()->leavePMPI();
  #endif
  return rvalue;
}


// getCallname
//========================================================================
string MPO_Win_create::getCallname() const
{
  // Distinguish between C and Fortran interface, default: C interface.
  string rvalue = "MPI_Win_create";

  if (F_INTERFACE == INTERFACE_MODE)
  {
      rvalue = "MPI_WIN_CREATE";
  }

  return rvalue;
}


// debugout
//========================================================================
std::ostream & MPO_Win_create::debugout(std::ostream & output) const
{
  // Distinguish between C and Fortran interface, default: C interface.
  output
      << getCallname()
      << "(*base, size ="
      << my_size
      << " disp_unit = "
      << my_disp_unit
      << "info = "
      << my_info
      << "(comm = "
      << my_comm
      << ", *win";

  if (F_INTERFACE == INTERFACE_MODE)
  {
    output << ", ierror";
  }

  output << ")" << std::endl;

  return output;
}


// save
//========================================================================
std::ostream & MPO_Win_create::save(std::ostream & out) const
{
  // we don't save and restore pointers
  MPO_Callobject::save(out);
  // out < my_base;
  out < my_size;
  out < my_disp_unit;
  //out < my_info;
  out < my_comm;
  // out < my_win;

  return out;
}


// restore
//========================================================================
std::istream & MPO_Win_create::restore(std::istream & in)
{
  // we don't save and restore pointers
  MPO_Callobject::restore(in);
  // in > my_base;
  in > my_size;
  in > my_disp_unit;
  //in > my_info;
  in > my_comm;
  // in > my_win;

  return in;
}


// mpich implementation
//========================================================================
// The following comment is copied from mpich2-1.0/src/mpi/rma/win_create.c

/*@
  MPI_Win_create - Create an MPI Window object for one-sided communication

  Input Parameters:
  + base - initial address of window (choice)
  . size - size of window in bytes (nonnegative integer)
  . disp_unit - local unit size for displacements, in bytes (positive integer)
  . info - info argument (handle)
  - comm - communicator (handle)

  Output Parameter:
  . win - window object returned by the call (handle)

  .N ThreadSafe
  .N Fortran

  .N Errors
  .N MPI_SUCCESS
  .N MPI_ERR_COMM
  .N MPI_ERR_INFO
  .N MPI_ERR_OTHER

  int MPI_Win_create(void *base, MPI_Aint size, int disp_unit, MPI_Info info,
  MPI_Comm comm, MPI_Win *win)

  @*/

#endif /*IS_MPI_WIN_CREATE_PRESENT*/

/*EOF*/

