/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of MPO_Win_lock class.
 *
 *  @author Bettina Krammer
 *
 *  $Id: mpo_Win_lock.cc 1112 2010-04-19 11:09:52Z tobias $
 */

#include "config.h"
#include "mpo-2-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"

#ifdef IS_MPI_WIN_LOCK_PRESENT

namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Win_lock::MPO_Win_lock()
        : MPO_Winhandler(MPI_WIN_NULL),
          my_lock_type(-1),
          my_rank(MPI_PROC_NULL),
          my_assert(-1)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Win_lock::MPO_Win_lock(int lock_type,
                               int rank,
                               int assert,
                               MPO_Win win)
        : MPO_Winhandler(win),
          my_lock_type(lock_type),
          my_rank(rank),
          my_assert(assert)
    {
    }


    // destructor
    //========================================================================
    MPO_Win_lock::~MPO_Win_lock()
    {
    }


    // checkLockType
    //========================================================================
    int MPO_Win_lock::checkLockType()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        // Check if locktype is valid.
        if ((MPI_LOCK_EXCLUSIVE != my_lock_type) &&
            (MPI_LOCK_SHARED != my_lock_type) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": lock type is not valid!"
                << std::endl;
            rvalue = MARMOT_ERR_LOCKTYPE_NOT_VALID;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }                           // end checkLockType


    // checkRank
    //========================================================================
    int MPO_Win_lock::checkRank()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if rank of source < 0.
        if ((my_rank != MPI_ANY_SOURCE) &&
            (my_rank != MPI_PROC_NULL) &&
            (my_rank < 0) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Rank = "
                << my_rank
                << " less than 0 !"
                << std::endl;
            rvalue = MARMOT_ERR_RANK_NEG;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        // (2) Check if rank of source is MPI_PROC_NULL.
        else if ((my_rank == MPI_PROC_NULL) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Rank = MPI_PROC_NULL !"
                << std::endl;
            rvalue = MARMOT_ERR_RANK_PROC_NULL;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        // (3) Check if MPI_ANY_SOURCE is used.
        else if ((my_rank == MPI_ANY_SOURCE) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": rank is MPI_ANY_SOURCE!"
                << std::endl;
            rvalue = MARMOT_ERR_RANK_ANY_SOURCE;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        // (4) Check if rank is out of communicator.
//TODO
//??
//my_comm
//         int size = -1;
//         PMPI_Comm_size(my_comm, &size);

//         if ((my_rank >= size) &&
//             (MSG_ERR <= DEBUG_MODE))
//         {
//             std::stringstream stream;
//             stream
//                 << "ERROR: "
//                 << getCallname()
//                 << ": Rank is out of communicator: rank = "
//                 << my_rank
//                 << ", size of communicator = "
//                 << size
//                 << std::endl;
//             rvalue = MARMOT_ERR_RANK_TOO_BIG;

//             MPO_String text;
//             text.assign(&stream);
//             MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
//             our_log->getMsg(msg);
//         }


        return rvalue;
    }                           // end checkRank


    // checkAssert
    //========================================================================
    int MPO_Win_lock::checkAssert()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        // Check if assert is valid.
        if ((0 != my_assert) &&
            (MPI_MODE_NOCHECK != my_assert) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": assert is not valid!"
                << " Legal values are only 0 or MPI_MODE_NOCHECK."
                << std::endl;
            rvalue = MARMOT_ERR_ASSERT_NOT_VALID;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }                           // end checkAssert


    // getId
    //========================================================================
    int MPO_Win_lock::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_WIN_LOCK_ID;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Win_lock::callNativeMPI()
    {
        int rvalue = 1;

        checkPrerequisites();
        checkLockType();
        checkRank();
        checkAssert();
        rvalue = PMPI_Win_lock(my_lock_type, my_rank, my_assert, my_win);

        return rvalue;
    }

    // checkPost
    //========================================================================
    int MPO_Win_lock::checkPost()
    {
    	//TODO
      return MARMOT_SUCCESS;
    }


    // getCallname
    //========================================================================
    string MPO_Win_lock::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Win_lock";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_WIN_LOCK";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Win_lock::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(locktype = "
            << my_lock_type
            << ", rank = "
            << my_rank
            << ", assert = "
            << my_assert
            << ", win = "
            << my_win;

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Win_lock::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_lock_type;
        out < my_rank;
        out < my_assert;
        out < my_win;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Win_lock::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_lock_type;
        in > my_rank;
        in > my_assert;
        in > my_win;

        return in;
    }

}                               // end of namespace MPO


// mpich implementation
//========================================================================
// The following comment is copied from mpich2-1.0/src/mpi/rma/win_lock.c

/*@
   MPI_Win_lock - Begin an RMA access epoch at the target process.

   Input Parameters:
+ lock_type - Indicates whether other processes may access the target
   window at the same time (if 'MPI_LOCK_SHARED') or not ('MPI_LOCK_EXCLUSIVE')
. rank - rank of locked window (nonnegative integer)
. assert - Used to optimize this call; zero may be used as a default.
  See notes. (integer)
- win - window object (handle)

   Notes:

   The name of this routine is misleading.  In particular, this
   routine need not block, except when the target process is the calling
   process.

   Implementations may restrict the use of RMA communication that is
   synchronized
   by lock calls to windows in memory allocated by 'MPI_Alloc_mem'. Locks can
   be used portably only in such memory.

   The 'assert' argument is used to indicate special conditions for the
   fence that an implementation may use to optimize the 'MPI_Win_fence'
   operation.  The value zero is always correct.  Other assertion values
   may be or''ed together.  Assertions that are valid for 'MPI_Win_fence' are\:

. MPI_MODE_NOCHECK - no other process holds, or will attempt to acquire a
  conflicting lock, while the caller holds the window lock. This is useful
  when mutual exclusion is achieved by other means, but the coherence
  operations that may be attached to the lock and unlock calls are still
  required.

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS

int MPI_Win_lock(int lock_type, int rank, int assert, MPI_Win win)
@*/

#endif /*IS_MPI_WIN_LOCK_PRESENT*/

/*EOF*/


