/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of MPO_Win_post class.
 *
 *  @author Bettina Krammer
 *
 *  $Id: mpo_Win_post.cc 1112 2010-04-19 11:09:52Z tobias $
 */


//#include <bitset> does not exist for all compilers :-(
#include "config.h"
#include "mpo-2-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"

#ifdef IS_MPI_WIN_POST_PRESENT

namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Win_post::MPO_Win_post()
        : MPO_Winhandler(MPI_WIN_NULL),
          my_group(MPI_GROUP_NULL),
          my_assert(-1)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Win_post::MPO_Win_post(MPO_Group group,
                               int assert,
                               MPO_Win win)
        : MPO_Winhandler(win),
          my_group(group),
          my_assert(assert)
    {
        mpo_win_post_Win = win;
    }


    // destructor
    //========================================================================
    MPO_Win_post::~MPO_Win_post()
    {
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Win_post::checkPrerequisites()
    {
        int rvalue = MARMOT_SUCCESS;
        if ( MARMOT_SUCCESS != ( rvalue = checkGroup()) )
        {
            return rvalue;
        }
        else if( MARMOT_SUCCESS != ( rvalue = checkAssert()) )
        {
            return rvalue;
        }
        else if( MARMOT_SUCCESS != ( rvalue = checkWin()) )
        {
            return rvalue;
        }

        return rvalue;
    }


    int MPO_Win_post::checkPost(){
        return MARMOT_SUCCESS;
    }


    // checkGroup
    //========================================================================
    int MPO_Win_post::checkGroup()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if group is null handle.
        if ((my_group == MPI_GROUP_NULL) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Group is MPI_GROUP_NULL!"
                << std::endl;
            rvalue = MARMOT_ERR_GROUP_NULL;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        // (2) Check if group is valid, i.e. registered in our_groups.
        if ((our_groups.find(my_group) == our_groups.end()) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Group not valid!"
                << std::endl;
            rvalue = MARMOT_ERR_GROUP_NOT_VALID;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }                           // end checkGroup


    // checkAssert
    //========================================================================
    int MPO_Win_post::checkAssert()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        if ( (0 != my_assert) &&
          (MPI_MODE_NOCHECK != my_assert) &&
          (MPI_MODE_NOSTORE != my_assert) &&
          (MPI_MODE_NOPUT != my_assert) &&
          (MSG_ERR <= DEBUG_MODE) )
        {
            std::stringstream stream;
            stream
             << "ERROR: "
             << getCallname()
             << ": assert is not valid!"
             << " Legal values are bit vectors OR of 0 or more of:"
             << "MPI_MODE_NOCHECK, "
             << "MPI_MODE_NOSTORE, "
             << "MPI_MODE_NOPUT."
             << std::endl;
            rvalue = MARMOT_ERR_ASSERT_NOT_VALID;
            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }


        // The nocheck mode can be specified if and only if it is specified
        // by each matching start call.
        // This is a global check!

        return rvalue;
    }                           // end checkAssert


    // checkWin
    //========================================================================
    int MPO_Win_post::checkWin()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if window is null handle.
        if ((MPI_WIN_NULL == mpo_win_post_Win) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": window is MPI_WIN_NULL!"
                << std::endl;
            rvalue = MARMOT_ERR_WIN_NULL;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        // (2) Check if window is valid, i.e. registered in our_wins.
        else if ((MPO_Win::our_wins.end() == MPO_Win::our_wins.find(mpo_win_post_Win)) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << " : Window is not valid!"
                << std::endl;
            rvalue = MARMOT_ERR_WIN_NOT_VALID;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }


    // getId
    //========================================================================
    int MPO_Win_post::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_WIN_POST_ID;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Win_post::callNativeMPI()
    {
        int rvalue = 1;

#ifdef MARMOT_USE_THREADSYNC
        MPO_Space::getSyncObject()->enterPMPI();
#endif


        //Debug Output for group:
/*        MPI_Group g;
        int size;
        int *ranks1, *ranks2;
        MPI_Comm_group (MPI_COMM_WORLD, &g);
        MPI_Comm_size (MPI_COMM_WORLD, &size);
        ranks1 = new int[size];
        ranks2 = new int[size];
        for (int i = 0; i < size; i++) ranks1[i] = i;
        MPI_Group_translate_ranks (g, size, ranks1, my_group, ranks2);

        if (ranks2[size-1] != MPI_UNDEFINED)
        {
			std::cout << "Erroneous group (WIN_POST): ";
			for (int i = 0; i < size; i++)
			{
				if (ranks2[i] != MPI_UNDEFINED)
					std::cout << i << ", ";
			}
			std::cout << std::endl;
        }
*/        //END Debug Output for group:

        rvalue = PMPI_Win_post(my_group, my_assert, mpo_win_post_Win);

#ifdef MARMOT_USE_THREADSYNC
        MPO_Space::getSyncObject()->leavePMPI();
#endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Win_post::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Win_post";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_WIN_POST";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Win_post::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(group = "
            << my_group
            << ", assert = "
            << my_assert
            << ", win = "
            << mpo_win_post_Win;

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Win_post::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_group;
        out < my_assert;
        out < mpo_win_post_Win;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Win_post::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_group;
        in > my_assert;
        in > mpo_win_post_Win;

        return in;
    }

}                               // end of namespace MPO


// mpich implementation
//========================================================================
// The following comment is copied from mpich2-1.0/src/mpi/rma/win_post.c

/*@
   MPI_Win_post - Start an RMA exposure epoch

   Input parameters:
+ group - group of origin processes (handle)
. assert - Used to optimize this call; zero may be used as a default.
  See notes. (integer)
- win - window object (handle)

   Notes:
   The 'assert' argument is used to indicate special conditions for the
   fence that an implementation may use to optimize the 'MPI_Win_post'
   operation.  The value zero is always correct.  Other assertion values
   may be or''ed together.  Assertions that are valid for 'MPI_Win_post' are\:

+ MPI_MODE_NOCHECK - the matching calls to 'MPI_WIN_START' have not yet
  occurred on any origin processes when the call to 'MPI_WIN_POST' is made.
  The nocheck option can be specified by a post call if and only if it is
  specified by each matching start call.
. MPI_MODE_NOSTORE - the local window was not updated by local stores (or
  local get or receive calls) since last synchronization. This may avoid
  the need for cache synchronization at the post call.
- MPI_MODE_NOPUT - the local window will not be updated by put or accumulate
  calls after the post call, until the ensuing (wait) synchronization. This
  may avoid the need for cache synchronization at the wait call.

.N Fortran

.N Errors
.N MPI_SUCCESS

int MPI_Win_post(MPI_Group group, int assert, MPI_Win win)
@*/

#endif /*IS_MPI_WIN_POST_PRESENT*/

/*EOF*/
