/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of MPO_Win_start class.
 *
 *  @author Bettina Krammer
 *
 *  $Id: mpo_Win_start.cc 1112 2010-04-19 11:09:52Z tobias $
 */

#include "config.h"
#include "mpo-2-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"

#ifdef IS_MPI_WIN_START_PRESENT

namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Win_start::MPO_Win_start()
        : MPO_Winhandler(MPI_WIN_NULL),
          my_group(MPI_GROUP_NULL),
          my_assert(-1)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Win_start::MPO_Win_start(MPO_Group group,
                                 int assert,
                                 MPO_Win win)
        : MPO_Winhandler(win),
          my_group(group),
          my_assert(assert)
    {
        mpo_win_start_Win = win;
    }


    // destructor
    //========================================================================
    MPO_Win_start::~MPO_Win_start()
    {
    }


    int MPO_Win_start::checkPrerequisites()
    {
        int rvalue = MARMOT_SUCCESS;
        if ( MARMOT_SUCCESS != ( rvalue = checkGroup()) )
        {
            return rvalue;
        }
        else if( MARMOT_SUCCESS != ( rvalue = checkAssert()) )
        {
            return rvalue;
        }
        else if( MARMOT_SUCCESS != ( rvalue = checkWin()) )
        {
            return rvalue;
        }

        return rvalue;
    }


    int MPO_Win_start::checkPost(){
        return MARMOT_SUCCESS;
    }


    // checkWin
    //========================================================================
    int MPO_Win_start::checkWin()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if window is null handle.
        if ((MPI_WIN_NULL == mpo_win_start_Win) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": window is MPI_WIN_NULL!"
                << std::endl;
            rvalue = MARMOT_ERR_WIN_NULL;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        // (2) Check if window is valid, i.e. registered in our_wins.
        else if ((MPO_Win::our_wins.end() == MPO_Win::our_wins.find(mpo_win_start_Win)) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << " : Window is not valid!"
                << std::endl;
            rvalue = MARMOT_ERR_WIN_NOT_VALID;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }


    // checkGroup
    //========================================================================
    int MPO_Win_start::checkGroup()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if group is null handle.
        if ((my_group == MPI_GROUP_NULL) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Group is MPI_GROUP_NULL!"
                << std::endl;
            rvalue = MARMOT_ERR_GROUP_NULL;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        // (2) Check if group is valid, i.e. registered in our_groups.
        if ((our_groups.find(my_group) == our_groups.end()) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Group not valid!"
                << std::endl;
            rvalue = MARMOT_ERR_GROUP_NOT_VALID;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }                           // end checkGroup


    // checkAssert
    //========================================================================
    int MPO_Win_start::checkAssert()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        // Check if assert is valid.
        if ((0 != my_assert) &&
            (MPI_MODE_NOCHECK != my_assert) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": assert is not valid!"
                << " Legal values are only 0 or MPI_MODE_NOCHECK."
                << std::endl;
            rvalue = MARMOT_ERR_ASSERT_NOT_VALID;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        // The nocheck mode can be specified if and only if it is specified
        // by each matching post call.
        // This is a global check!

        return rvalue;
    }                           // end checkAssert


    // getId
    //========================================================================
    int MPO_Win_start::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_WIN_START_ID;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Win_start::callNativeMPI()
    {
        int rvalue = 1;

#ifdef MARMOT_USE_THREADSYNC
        MPO_Space::getSyncObject()->enterPMPI();
#endif

        rvalue = PMPI_Win_start(my_group, my_assert, mpo_win_start_Win);

#ifdef MARMOT_USE_THREADSYNC
        MPO_Space::getSyncObject()->leavePMPI();
#endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Win_start::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Win_start";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_WIN_START";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Win_start::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(group = "
            << my_group
            << ", assert = "
            << my_assert
            << ", win = "
            << mpo_win_start_Win;

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Win_start::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_group;
        out < my_assert;
        out < mpo_win_start_Win;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Win_start::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_group;
        in > my_assert;
        in > mpo_win_start_Win;

        return in;
    }

}                               // end of namespace MPO


// mpich implementation
//========================================================================
// The following comment is copied from mpich2-1.0/src/mpi/rma/win_start.c

/*@
   MPI_Win_start - Start an RMA access epoch for MPI

  Input Parameters:
+ group - group of target processes (handle)
. assert - Used to optimize this call; zero may be used as a default.
  See notes. (integer)
- win - window object (handle)

   Notes:
   The 'assert' argument is used to indicate special conditions for the
   fence that an implementation may use to optimize the 'MPI_Win_start'
   operation.  The value zero is always correct.  Other assertion values
   may be or''ed together.  Assertions tha are valid for 'MPI_Win_start' are\:

. MPI_MODE_NOCHECK - the matching calls to 'MPI_WIN_POST' have already
  completed on all target processes when the call to 'MPI_WIN_START' is made.
  The nocheck option can be specified in a start call if and only if it is
  specified in each matching post call. This is similar to the optimization
  of ready-send that may save a handshake when the handshake is implicit in
  the code. (However, ready-send is matched by a regular receive, whereas
  both start and post must specify the nocheck option.)

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_WIN
.N MPI_ERR_OTHER

int MPI_Win_start(MPI_Group group, int assert, MPI_Win win)
@*/

#endif /*IS_MPI_WIN_START_PRESENT*/

/*EOF*/
