/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of MPO_Win_unlock class.
 *
 *  @author Bettina Krammer
 *
 *  $Id: mpo_Win_unlock.cc 1112 2010-04-19 11:09:52Z tobias $
 */

#include "config.h"
#include "mpo-2-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"

#ifdef IS_MPI_WIN_UNLOCK_PRESENT

namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Win_unlock::MPO_Win_unlock()
        : MPO_Winhandler(MPI_WIN_NULL),
          my_rank(MPI_PROC_NULL)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Win_unlock::MPO_Win_unlock(int rank,
                                   MPO_Win win)
        : MPO_Winhandler(win),
          my_rank(rank)
    {
    }


    // destructor
    //========================================================================
    MPO_Win_unlock::~MPO_Win_unlock()
    {
    }


    // checkRank
    //========================================================================
    int MPO_Win_unlock::checkRank()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if rank of source < 0.
        if ((my_rank != MPI_ANY_SOURCE) &&
            (my_rank != MPI_PROC_NULL) &&
            (my_rank < 0) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Rank = "
                << my_rank
                << " less than 0 !"
                << std::endl;
            rvalue = MARMOT_ERR_RANK_NEG;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        // (2) Check if rank of source is MPI_PROC_NULL.
        else if ((my_rank == MPI_PROC_NULL) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Rank = MPI_PROC_NULL !"
                << std::endl;
            rvalue = MARMOT_ERR_RANK_PROC_NULL;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        // (3) Check if MPI_ANY_SOURCE is used.
        else if ((my_rank == MPI_ANY_SOURCE) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": rank is MPI_ANY_SOURCE!"
                << std::endl;
            rvalue = MARMOT_ERR_RANK_ANY_SOURCE;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        // (4) Check if rank is out of communicator.
//TODO
//??
//my_comm
//         int size = -1;
//         PMPI_Comm_size(my_comm, &size);

//         if ((my_rank >= size) &&
//             (MSG_ERR <= DEBUG_MODE))
//         {
//             std::stringstream stream;
//             stream
//                 << "ERROR: "
//                 << getCallname()
//                 << ": Rank is out of communicator: rank = "
//                 << my_rank
//                 << ", size of communicator = "
//                 << size
//                 << std::endl;
//             rvalue = MARMOT_ERR_RANK_TOO_BIG;

//             MPO_String text;
//             text.assign(&stream);
//             MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
//             our_log->getMsg(msg);
//         }

        return rvalue;
    }                           // end checkRank


    // getId
    //========================================================================
    int MPO_Win_unlock::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_WIN_UNLOCK_ID;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Win_unlock::callNativeMPI()
    {
        int rvalue = 1;

        checkPrerequisites();
        checkRank();
        rvalue = PMPI_Win_unlock(my_rank, my_win);

        return rvalue;
    }

    // checkPost
    //========================================================================
    int MPO_Win_unlock::checkPost()
    {
    	//TODO
      return MARMOT_SUCCESS;
    }

    // getCallname
    //========================================================================
    string MPO_Win_unlock::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Win_unlock";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_WIN_UNLOCK";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Win_unlock::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(rank = "
            << my_rank
             << ", win = "
            << my_win;

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Win_unlock::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_rank;
        out < my_win;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Win_unlock::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_rank;
        in > my_win;

        return in;
    }

}                               // end of namespace MPO


// mpich implementation
//========================================================================
// The following comment is copied from mpich2-1.0/src/mpi/rma/win_unlock.c

/*@
   MPI_Win_unlock - Completes an RMA access epoch at the target process

   Input Parameters:
+ rank - rank of window (nonnegative integer)
- win - window object (handle)

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_WIN
.N MPI_ERR_OTHER

.seealso: MPI_Win_lock

int MPI_Win_unlock(int rank, MPI_Win win)
@*/

#endif /*IS_MPI_WIN_UNLOCK_PRESENT*/

/*EOF*/
