/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Type_create_darray.
 *
 *  @author Tobias Hilbrich
 *
 *  $Id$
 */


#include "mpo-objects.h"
#include "mpo-2-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"
#include <stdio.h>

#ifdef IS_MPI_TYPE_CREATE_DARRAY_PRESENT

namespace MPO
{
    // default constructor
    //========================================================================
	MPO_Type_create_darray::MPO_Type_create_darray()
		: MPO_Typeconstructors (-1, MPI_DATATYPE_NULL, -1)
	{
		//nothing to do
	}

	// proper constructor
	//========================================================================
	MPO_Type_create_darray::MPO_Type_create_darray(
    		int size,
    		int rank,
    		int ndims,
    		int* array_of_gsizes,
    		int* array_of_distribs,
    		int* array_of_dargs,
    		int* array_of_psizes,
    		int order,
    		MPI_Datatype oldtype,
    		MPI_Datatype* newtype
    		)
		: MPO_Typeconstructors (ndims, oldtype, MPO_TYPE_CREATE_DARRAY_ID),
		  my_size (size),
		  my_rank (rank),
		  my_array_of_gsizes (array_of_gsizes),
		  my_array_of_distribs (array_of_distribs),
		  my_array_of_dargs (array_of_dargs),
		  my_array_of_psizes (array_of_psizes),
		  my_order (order),
		  my_newtype (newtype)
	{
		//nothing to do
	}

    // default destructor
    //========================================================================
    MPO_Type_create_darray::~MPO_Type_create_darray()
    {
    	//Nothing to do
    }

    // checkPrerequisites
    //========================================================================
    int MPO_Type_create_darray::checkPrerequisites()
    {
    	int ret;

    	ret = check_size_and_rank ();
    	ret = check_gsizes ();
    	ret = check_distribs ();
    	ret = check_dargs ();
    	ret = check_psizes ();
    	ret = check_order ();

    	return ret;
    }

    // callNativeMPI
    //========================================================================
    int MPO_Type_create_darray::callNativeMPI()
    {
    	int rvalue = 1;

#ifdef MARMOT_USE_THREADSYNC
    	MPO_Space::getSyncObject()->enterPMPI();
#endif
    	rvalue = PMPI_Type_create_darray (
    			my_size,
    			my_rank,
    			my_count, /*=ndims (used in MPO_TypeConstructor as count)*/
    			my_array_of_gsizes,
    			my_array_of_distribs,
    			my_array_of_dargs,
    			my_array_of_psizes,
    			my_order,
    			my_oldtype,
    			my_newtype
    			);
#ifdef MARMOT_USE_THREADSYNC
    	MPO_Space::getSyncObject()->leavePMPI();
#endif

    	return rvalue;
    }

    // checkPost
    //========================================================================
   	int MPO_Type_create_darray::checkPost()
   	{
   		our_datatypes.insert(std::make_pair(MPO_Datatype(*my_newtype),
   				DATATYPE_CREATED_NOT_COMMITTED));

   		our_datatype_traces->addAndSetCreation(*my_newtype,getSourceLocation());
   		our_datatype_traces->findAndTryAddDependendency(*my_newtype,my_oldtype);

   		return 0;
   	}

    // debugout
    //========================================================================
  	std::ostream & MPO_Type_create_darray::debugout(std::ostream & output) const
  	{
  		output
  		<< getCallname()
  		<< "(size = "
  		<< my_size
  		<< ", rank = "
  		<< my_rank
  		<< ", ndims = "
  		<< my_count /*=ndims (used as count in MPO_TypeConstructor constructor)*/
  		<< ", *array_of_gsizes"
  		<< ", *array_of_distribs"
  		<< ", *array_of_dargs"
  		<< ", *array_of_psizes"
  		<< ", *array_of_gsizes"
  		<< ", order = "
  		<< my_order << "(int value)"
  		<< ", oldtype = "
  		<< my_oldtype
  		<< ", *newtype";

  		if (F_INTERFACE == INTERFACE_MODE)
  		{
  			output << ", ierror";
  		}

  		output
  		<< ")"
  		<< std::endl;

  		return output;
  	}

  	// save
  	//========================================================================
    std::ostream & MPO_Type_create_darray::save(std::ostream & out) const
    {
    	// we don't save and restore pointers
    	MPO_Callobject::save(out);
    	out < my_size;
    	out < my_rank;
    	out < my_count /*ndims*/;
    	//out < my_array_of_gsized;
    	//out < my_array_of_distribs;
    	//out < my_array_of_dargs;
    	//out < my_array_of_psizes;
    	out < my_order;
    	out < my_oldtype;
    	//out < my_newtype;

    	return out;
    }

    // restore
    //========================================================================
    std::istream & MPO_Type_create_darray::restore(std::istream & in)
    {
    	// we don't save and restore pointers
    	MPO_Callobject::restore(in);
    	in > my_size;
    	in > my_rank;
    	in > my_count /*ndims*/;
    	//in < my_array_of_gsized;
    	//in < my_array_of_distribs;
    	//in < my_array_of_dargs;
    	//in < my_array_of_psizes;
    	in > my_order;
    	in > my_oldtype;
    	//in < my_newtype;

    	return in;
    }

	// getCallname
	//========================================================================
	string MPO_Type_create_darray::getCallname() const
	{
		/*  Distinguish between C and Fortran interface, default: C interface.
		 */
		string rvalue = "MPI_Type_create_darray";

		if (F_INTERFACE == INTERFACE_MODE)
		{
			rvalue = "MPI_TYPE_CREATE_DARRAY";
		}

		return rvalue;
	}

	// getId
	//========================================================================
	int MPO_Type_create_darray::getId() const
	{
		// returns the unique call identity set by enum MPO_CALL_ID
		return MPO_TYPE_CREATE_DARRAY_ID;
	}

	// check_size_and_rank
	//========================================================================
	int MPO_Type_create_darray::check_size_and_rank (void)
	{
		int rvalue = MARMOT_SUCCESS;

		if ((my_size < 0) &&
			(MSG_ERR <= DEBUG_MODE))
		{
			std::stringstream stream;
			MPO_MSG_TYPE type = MSG_TYPE_ERROR;

			stream
				<< "ERROR: "
				<< getCallname()
				<< ": size is less than 0! (size = "
				<< my_size
				<< ")"
				<< std::endl;

			rvalue = MARMOT_ERR_SIZE_NEG;

			MPO_String text;
			text.assign(&stream);
			MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
			our_log->getMsg(msg);
		}
		else if  ((my_size == 0) &&
				  (MSG_ERR_WARN <= DEBUG_MODE))
		{
			std::stringstream stream;
			MPO_MSG_TYPE type = MSG_TYPE_WARNING;

			stream
			<< "Warning: "
			<< getCallname()
			<< ": size == 0!"
			<< std::endl;
			rvalue = MARMOT_WARN_SIZE_ZERO;

			MPO_String text;
			text.assign(&stream);
			MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
			our_log->getMsg(msg);
		}

		if  ((my_rank < 0) &&
				(MSG_ERR <= DEBUG_MODE))
		{
			std::stringstream stream;
			MPO_MSG_TYPE type = MSG_TYPE_ERROR;

			stream
			<< "Error: "
			<< getCallname()
			<< ": rank less than 0! (rank = "
			<< my_rank
			<< ")"
			<< std::endl;
			rvalue = MARMOT_ERR_RANK_NEG;

			MPO_String text;
			text.assign(&stream);
			MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
			our_log->getMsg(msg);
		}
		else if  ((my_rank >= my_size) &&
				  (MSG_ERR <= DEBUG_MODE))
		{
			std::stringstream stream;
			MPO_MSG_TYPE type = MSG_TYPE_ERROR;

			stream
			<< "Error: "
			<< getCallname()
			<< ": rank too high, is greater equal size (rank = "
			<< my_rank
			<< ", size = "
			<< my_size
			<< ")!"
			<< std::endl;
			rvalue = MARMOT_ERR_RANK_TOO_BIG;

			MPO_String text;
			text.assign(&stream);
			MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
			our_log->getMsg(msg);
		}

		return rvalue;
	}

	// check_gsizes
	//========================================================================
	int MPO_Type_create_darray::check_gsizes (void)
	{
		int rvalue = MARMOT_SUCCESS;

		if ((my_count > 0) && (!my_array_of_gsizes) &&
			(MSG_ERR <= DEBUG_MODE))
		{
			std::stringstream stream;
			MPO_MSG_TYPE type = MSG_TYPE_ERROR;

			stream
			<< "ERROR: "
			<< getCallname()
			<< ": array_of_gsizes is NULL!"
			<< std::endl;
			rvalue = MARMOT_ERR_POINTER_NULL;

			MPO_String text;
			text.assign(&stream);
			MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
			our_log->getMsg(msg);
		}
		else
		{
			for (int i = 0; i < my_count; i++)
			{
				if ((my_array_of_gsizes[i] < 0) &&
					(MSG_ERR <= DEBUG_MODE))
				{
					std::stringstream stream;
					MPO_MSG_TYPE type = MSG_TYPE_ERROR;

					stream
					<< "ERROR: "
					<< getCallname()
					<< ": array_of_gsizes["
					<< i
					<< "] is less than 0!"
					<< std::endl;
					rvalue = MARMOT_ERR_SIZE_NEG;

					MPO_String text;
					text.assign(&stream);
					MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
					our_log->getMsg(msg);
				}

				if ((my_array_of_gsizes[i] == 0) &&
					(MSG_ERR <= DEBUG_MODE))
				{
					std::stringstream stream;
					MPO_MSG_TYPE type = MSG_TYPE_ERROR;

					stream
					<< "Warning: "
					<< getCallname()
					<< ": array_of_gsizes["
					<< i
					<< "] == 0!"
					<< std::endl;
					rvalue = MARMOT_WARN_SIZE_ZERO;

					MPO_String text;
					text.assign(&stream);
					MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
					our_log->getMsg(msg);
				}
			}
		}

		return rvalue;

	}

	// check_distribs
	//========================================================================
	int MPO_Type_create_darray::check_distribs (void)
	{
		int rvalue = MARMOT_SUCCESS;

		if ((my_count > 0) && (!my_array_of_distribs) &&
		(MSG_ERR <= DEBUG_MODE))
		{
			std::stringstream stream;
			MPO_MSG_TYPE type = MSG_TYPE_ERROR;

			stream
			<< "ERROR: "
			<< getCallname()
			<< ": array_of_distribs is NULL!"
			<< std::endl;
			rvalue = MARMOT_ERR_POINTER_NULL;

			MPO_String text;
			text.assign(&stream);
			MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
			our_log->getMsg(msg);
		}
		else
		{
			for (int i = 0; i < my_count; i++)
			{
				if ((my_array_of_distribs[i] != MPI_DISTRIBUTE_BLOCK) &&
					(my_array_of_distribs[i] != MPI_DISTRIBUTE_CYCLIC) &&
					(my_array_of_distribs[i] != MPI_DISTRIBUTE_NONE) &&
					(MSG_ERR <= DEBUG_MODE))
				{
					std::stringstream stream;
					MPO_MSG_TYPE type = MSG_TYPE_ERROR;

					stream
					<< "ERROR: "
					<< getCallname()
					<< ": array_of_distribs is not one of MPI_DISTRIBUTE_BLOCK, MPI_DISTRIBUTE_CYCLIC, MPI_DISTRIBUTE_NONE!"
					<< std::endl;
					rvalue = MARMOT_ERR_DISTRIBUTION_INVALID;

					MPO_String text;
					text.assign(&stream);
					MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
					our_log->getMsg(msg);
				}
			}
		}

		return rvalue;
	}

	// check_dargs
	//========================================================================
	int MPO_Type_create_darray::check_dargs (void)
	{
		int rvalue = MARMOT_SUCCESS;

		if ((my_count > 0) && (!my_array_of_dargs) &&
		    (MSG_ERR <= DEBUG_MODE))
		{
			std::stringstream stream;
			MPO_MSG_TYPE type = MSG_TYPE_ERROR;

			stream
			<< "ERROR: "
			<< getCallname()
			<< ": array_of_dargs is NULL!"
			<< std::endl;
			rvalue = MARMOT_ERR_POINTER_NULL;

			MPO_String text;
			text.assign(&stream);
			MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
			our_log->getMsg(msg);
		}
		else
		{
			for (int i = 0; i < my_count; i++)
			{
				if (    (my_array_of_dargs[i] < 0) &&
						(my_array_of_dargs[i] != MPI_DISTRIBUTE_DFLT_DARG) &&
						(MSG_ERR <= DEBUG_MODE))
				{
					std::stringstream stream;
					MPO_MSG_TYPE type = MSG_TYPE_ERROR;

					stream
					<< "ERROR: "
					<< getCallname()
					<< ": array_of_dargs["<<i<<"] is less than 0, though it is required to be a positive integer (even for MPI_DISTRIBUTE_NONE)!"
					<< std::endl;
					rvalue = MARMOT_ERR_SIZE_NEG;

					MPO_String text;
					text.assign(&stream);
					MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
					our_log->getMsg(msg);
				}

				if ( my_array_of_distribs &&
					 my_array_of_gsizes &&
					 my_array_of_psizes &&
					 (my_array_of_distribs[i] == MPI_DISTRIBUTE_BLOCK) &&
					 (my_array_of_dargs[i] != MPI_DISTRIBUTE_DFLT_DARG) &&
					 (my_array_of_dargs[i] * my_array_of_psizes[i] < my_array_of_gsizes[i]) &&
					 (MSG_ERR <= DEBUG_MODE))
				{
					std::stringstream stream;
					MPO_MSG_TYPE type = MSG_TYPE_ERROR;

					stream
					<< "ERROR: "
					<< getCallname()
					<< ": block distribution is invalid, array_of_dargs["<<i<<"] * array_of_psizes["<<i<<"] less than array_of_gsizes["<<i<<"]!"
					<< std::endl;
					rvalue = MARMOT_ERR_BLOCK_DISTRIBUTION_TOO_SMALL;

					MPO_String text;
					text.assign(&stream);
					MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
					our_log->getMsg(msg);
				}

				if ( my_array_of_distribs &&
				     (my_array_of_distribs[i] == MPI_DISTRIBUTE_CYCLIC ||
				      my_array_of_distribs[i] == MPI_DISTRIBUTE_BLOCK   ) &&
					 (my_array_of_dargs[i] != MPI_DISTRIBUTE_DFLT_DARG) &&
					 (my_array_of_dargs[i] <= 0) &&
					 (MSG_ERR <= DEBUG_MODE))
				{
					std::stringstream stream;
					MPO_MSG_TYPE type = MSG_TYPE_ERROR;

					stream
					<< "ERROR: "
					<< getCallname()
					<< ": distribution argument invalid (less equal 0) at array_of_dargs["<<i<<"]!"
					<< std::endl;
					rvalue = MARMOT_ERR_DISTRIBUTION_ARGUMENT_INVALID;

					MPO_String text;
					text.assign(&stream);
					MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
					our_log->getMsg(msg);
				}
			}
		}

		return rvalue;
	}

	// check_psizes
	//========================================================================
	int MPO_Type_create_darray::check_psizes (void)
	{
		int rvalue = MARMOT_SUCCESS;

		if ((my_count > 0) && (!my_array_of_psizes) &&
		(MSG_ERR <= DEBUG_MODE))
		{
			std::stringstream stream;
			MPO_MSG_TYPE type = MSG_TYPE_ERROR;

			stream
			<< "ERROR: "
			<< getCallname()
			<< ": array_of_psizes is NULL!"
			<< std::endl;
			rvalue = MARMOT_ERR_POINTER_NULL;

			MPO_String text;
			text.assign(&stream);
			MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
			our_log->getMsg(msg);
		}
		else
		{
			int test_size = 1;
			for (int i = 0; i < my_count; i++)
			{
				test_size = test_size * my_array_of_psizes[i];
			}

			if ( 	(test_size != my_size) &&
					(MSG_ERR <= DEBUG_MODE))
			{
				std::stringstream stream;
				MPO_MSG_TYPE type = MSG_TYPE_ERROR;

				stream
				<< "ERROR: "
				<< getCallname()
				<< ": array_of_psizes is invalid, product of all of its entries must equal the size argument (product="<< test_size <<", size="<<my_size<<")!"
				<< std::endl;
				rvalue = MARMOT_ERR_PROCESS_GRID_INVALID;

				MPO_String text;
				text.assign(&stream);
				MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
				our_log->getMsg(msg);
			}
		}

		return rvalue;
	}

	// check_order
	//========================================================================
	int MPO_Type_create_darray::check_order (void)
	{
		int rvalue = MARMOT_SUCCESS;

		if ( (my_order != MPI_ORDER_C) &&
			 (my_order != MPI_ORDER_FORTRAN) &&
			 (MSG_ERR <= DEBUG_MODE))
		{
			std::stringstream stream;
			MPO_MSG_TYPE type = MSG_TYPE_ERROR;

			stream
			<< "ERROR: "
			<< getCallname()
			<< ": order argument is invalid (is neither MPI_ORDER_C, nor MPI_ORDER_FORTRAN)!"
			<< std::endl;
			rvalue = MARMOT_ERR_ARRAY_ORDER_INVALID;

			MPO_String text;
			text.assign(&stream);
			MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
			our_log->getMsg(msg);
		}

		return rvalue;
	}
}

#endif /*IS_MPI_TYPE_CREATE_DARRAY_PRESENT*/

/*EOF*/
