/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Type_create_subarray.
 *
 *  @author Tobias Hilbrich
 *
 *  $Id$
 */


#include "mpo-objects.h"
#include "mpo-2-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"
#include <stdio.h>

#ifdef IS_MPI_TYPE_CREATE_INDEXED_BLOCK_PRESENT

namespace MPO
{
    // default constructor
    //========================================================================
	MPO_Type_create_indexed_block::MPO_Type_create_indexed_block()
		: MPO_Typeconstructors (-1, MPI_DATATYPE_NULL, -1)
	{
		//nothing to do
	}

	// proper constructor
	//========================================================================
	MPO_Type_create_indexed_block::MPO_Type_create_indexed_block(
    		int count,
    		int blocklength,
    		int* array_of_displacements,
    		MPI_Datatype oldtype,
    		MPI_Datatype* newtype
    		)
		: MPO_Typeconstructors (count, oldtype, MPO_TYPE_CREATE_INDEXED_BLOCK_ID),
		  my_blocklength (blocklength),
		  my_array_of_displacements (array_of_displacements),
		  my_newtype (newtype)
	{
		//nothing to do
	}

    // default destructor
    //========================================================================
    MPO_Type_create_indexed_block::~MPO_Type_create_indexed_block()
    {
    	//Nothing to do
    }

    // checkPrerequisites
    //========================================================================
    int MPO_Type_create_indexed_block::checkPrerequisites()
    {
    	int ret;

    	ret = check_blocklength ();
    	ret = check_displacements ();

    	return ret;
    }

    // callNativeMPI
    //========================================================================
    int MPO_Type_create_indexed_block::callNativeMPI()
    {
    	int rvalue = 1;

#ifdef MARMOT_USE_THREADSYNC
    	MPO_Space::getSyncObject()->enterPMPI();
#endif
    	rvalue = PMPI_Type_create_indexed_block (
    			my_count,
    			my_blocklength,
    			my_array_of_displacements,
    			my_oldtype,
    			my_newtype
    			);
#ifdef MARMOT_USE_THREADSYNC
    	MPO_Space::getSyncObject()->leavePMPI();
#endif

    	return rvalue;
    }

    // checkPost
    //========================================================================
   	int MPO_Type_create_indexed_block::checkPost()
   	{
   		our_datatypes.insert(std::make_pair(MPO_Datatype(*my_newtype),
   				DATATYPE_CREATED_NOT_COMMITTED));

   		our_datatype_traces->addAndSetCreation(*my_newtype,getSourceLocation());
   		our_datatype_traces->findAndTryAddDependendency(*my_newtype,my_oldtype);

   		return 0;
   	}

    // debugout
    //========================================================================
  	std::ostream & MPO_Type_create_indexed_block::debugout(std::ostream & output) const
  	{
  		output
  		<< getCallname()
  		<< "(count = "
  		<< my_count
  		<< ", blocklength = "
  		<< my_blocklength
  		<< ", *array_of_displacements"
  		<< ", oldtype = "
  		<< my_oldtype
  		<< ", *newtype";

  		if (F_INTERFACE == INTERFACE_MODE)
  		{
  			output << ", ierror";
  		}

  		output
  		<< ")"
  		<< std::endl;

  		return output;
  	}

  	// save
  	//========================================================================
    std::ostream & MPO_Type_create_indexed_block::save(std::ostream & out) const
    {
    	// we don't save and restore pointers
    	MPO_Callobject::save(out);
    	out < my_count;
    	out < my_blocklength;
    	//out < my_array_of_displacements;
    	out < my_oldtype;
    	//out < my_newtype;

    	return out;
    }

    // restore
    //========================================================================
    std::istream & MPO_Type_create_indexed_block::restore(std::istream & in)
    {
    	// we don't save and restore pointers
    	MPO_Callobject::restore(in);
    	in > my_count;
    	in > my_blocklength;
    	//in < my_array_of_displacements;
    	in > my_oldtype;
    	//in < my_newtype;

    	return in;
    }

	// getCallname
	//========================================================================
	string MPO_Type_create_indexed_block::getCallname() const
	{
		/*  Distinguish between C and Fortran interface, default: C interface.
		 */
		string rvalue = "MPI_Type_create_indexed_block";

		if (F_INTERFACE == INTERFACE_MODE)
		{
			rvalue = "MPI_TYPE_CREATE_INDEXED_BLOCK";
		}

		return rvalue;
	}

	// getId
	//========================================================================
	int MPO_Type_create_indexed_block::getId() const
	{
		// returns the unique call identity set by enum MPO_CALL_ID
		return MPO_TYPE_CREATE_INDEXED_BLOCK_ID;
	}

	// check_blocklength
	//========================================================================
	int MPO_Type_create_indexed_block::check_blocklength (void)
	{
		int rvalue = MARMOT_SUCCESS;

		if ((my_count > 0) &&
			(my_blocklength < 0) &&
			(MSG_ERR <= DEBUG_MODE))
		{
			std::stringstream stream;
			MPO_MSG_TYPE type = MSG_TYPE_ERROR;

			stream
			<< "ERROR: "
			<< getCallname()
			<< ": blocklength is less than 0!"
			<< std::endl;
			rvalue = MARMOT_ERR_BLOCKLENGTH_NEG;

			MPO_String text;
			text.assign(&stream);
			MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
			our_log->getMsg(msg);
		}

		if (	(my_count > 0) &&
				(my_blocklength == 0) &&
				(MSG_ERR_WARN <= DEBUG_MODE))
		{
			std::stringstream stream;
			MPO_MSG_TYPE type = MSG_TYPE_WARNING;

			stream
			<< "Warning: "
			<< getCallname()
			<< ": blocklength == 0."
			<< std::endl;
			rvalue = MARMOT_WARN_BLOCKLENGTH_ZERO;

			MPO_String text;
			text.assign(&stream);
			MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
			our_log->getMsg(msg);
		}

		return rvalue;
	}

	// check_displacements
	//========================================================================
	int MPO_Type_create_indexed_block::check_displacements (void)
	{
		int rvalue = MARMOT_SUCCESS;

		if ((my_count > 0) &&
			(!my_array_of_displacements) &&
			(MSG_ERR <= DEBUG_MODE))
		{
			std::stringstream stream;
			MPO_MSG_TYPE type = MSG_TYPE_ERROR;

			stream
			<< "ERROR: "
			<< getCallname()
			<< ": array_of_displacements is NULL (not allocated)!"
			<< std::endl;
			rvalue = MARMOT_ERR_POINTER_NULL;

			MPO_String text;
			text.assign(&stream);
			MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
			our_log->getMsg(msg);
		}

		//TODO: displacements should be checked to determine whether blocks are non-overlapping
		//      but mpo_Type_indexed doesn't implements this, so that would be the starting
		//      point for such a check

		return rvalue;
	}
}

#endif /*IS_MPI_TYPE_CREATE_INDEXED_BLOCK_PRESENT*/

/*EOF*/
