/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Type_create_subarray.
 *
 *  @author Tobias Hilbrich
 *
 *  $Id$
 */


#include "mpo-objects.h"
#include "mpo-2-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"
#include <stdio.h>

#ifdef IS_MPI_TYPE_CREATE_SUBARRAY_PRESENT

namespace MPO
{
    // default constructor
    //========================================================================
	MPO_Type_create_subarray::MPO_Type_create_subarray()
		: MPO_Typeconstructors (-1, MPI_DATATYPE_NULL, -1)
	{
		//nothing to do
	}

	// proper constructor
	//========================================================================
	MPO_Type_create_subarray::MPO_Type_create_subarray(
    		int ndims,
    		int* array_of_sizes,
    		int* array_of_subsizes,
    		int* array_of_starts,
    		int order,
    		MPI_Datatype oldtype,
    		MPI_Datatype* newtype
    		)
		: MPO_Typeconstructors (ndims, oldtype, MPO_TYPE_CREATE_SUBARRAY_ID),
		  my_array_of_sizes (array_of_sizes),
		  my_array_of_subsizes (array_of_subsizes),
		  my_array_of_starts (array_of_starts),
		  my_order (order),
		  my_newtype (newtype)
	{
		//nothing to do
	}

    // default destructor
    //========================================================================
    MPO_Type_create_subarray::~MPO_Type_create_subarray()
    {
    	//Nothing to do
    }

    // checkPrerequisites
    //========================================================================
    int MPO_Type_create_subarray::checkPrerequisites()
    {
    	int ret;

    	ret = check_sizes ();
    	ret = check_subsizes ();
    	ret = check_starts ();
    	ret = check_order ();

    	return ret;
    }

    // callNativeMPI
    //========================================================================
    int MPO_Type_create_subarray::callNativeMPI()
    {
    	int rvalue = 1;

#ifdef MARMOT_USE_THREADSYNC
    	MPO_Space::getSyncObject()->enterPMPI();
#endif
    	rvalue = PMPI_Type_create_subarray (
    			my_count, /*=ndims (used in MPO_TypeConstructor as count)*/
    			my_array_of_sizes,
    			my_array_of_subsizes,
    			my_array_of_starts,
    			my_order,
    			my_oldtype,
    			my_newtype
    			);
#ifdef MARMOT_USE_THREADSYNC
    	MPO_Space::getSyncObject()->leavePMPI();
#endif

    	return rvalue;
    }

    // checkPost
    //========================================================================
   	int MPO_Type_create_subarray::checkPost()
   	{
   		our_datatypes.insert(std::make_pair(MPO_Datatype(*my_newtype),
   				DATATYPE_CREATED_NOT_COMMITTED));

   		our_datatype_traces->addAndSetCreation(*my_newtype,getSourceLocation());
   		our_datatype_traces->findAndTryAddDependendency(*my_newtype,my_oldtype);

   		return 0;
   	}

    // debugout
    //========================================================================
  	std::ostream & MPO_Type_create_subarray::debugout(std::ostream & output) const
  	{
  		output
  		<< getCallname()
  		<< "(ndims = "
  		<< my_count /*=ndims (used as count in MPO_TypeConstructor constructor)*/
  		<< ", *array_of_sizes"
  		<< ", *array_of_subsizes"
  		<< ", *array_of_starts"
  		<< ", order = "
  		<< my_order << "(int value)"
  		<< ", oldtype = "
  		<< my_oldtype
  		<< ", *newtype";

  		if (F_INTERFACE == INTERFACE_MODE)
  		{
  			output << ", ierror";
  		}

  		output
  		<< ")"
  		<< std::endl;

  		return output;
  	}

  	// save
  	//========================================================================
    std::ostream & MPO_Type_create_subarray::save(std::ostream & out) const
    {
    	// we don't save and restore pointers
    	MPO_Callobject::save(out);
    	out < my_count /*ndims*/;
    	//out < my_array_of_sizes;
    	//out < my_array_of_subsizes;
    	//out < my_array_of_starts;
    	out < my_order;
    	out < my_oldtype;
    	//out < my_newtype;

    	return out;
    }

    // restore
    //========================================================================
    std::istream & MPO_Type_create_subarray::restore(std::istream & in)
    {
    	// we don't save and restore pointers
    	MPO_Callobject::restore(in);
    	in > my_count /*ndims*/;
    	//in < my_array_of_sizes;
    	//in < my_array_of_subsizes;
    	//in < my_array_of_starts;
    	in > my_order;
    	in > my_oldtype;
    	//in < my_newtype;

    	return in;
    }

	// getCallname
	//========================================================================
	string MPO_Type_create_subarray::getCallname() const
	{
		/*  Distinguish between C and Fortran interface, default: C interface.
		 */
		string rvalue = "MPI_Type_create_subarray";

		if (F_INTERFACE == INTERFACE_MODE)
		{
			rvalue = "MPI_TYPE_CREATE_SUBARRAY";
		}

		return rvalue;
	}

	// getId
	//========================================================================
	int MPO_Type_create_subarray::getId() const
	{
		// returns the unique call identity set by enum MPO_CALL_ID
		return MPO_TYPE_CREATE_SUBARRAY_ID;
	}

	// check_sizes
	//========================================================================
	int MPO_Type_create_subarray::check_sizes (void)
	{
		int rvalue = MARMOT_SUCCESS;

		if ((my_count > 0) && (!my_array_of_sizes) &&
			(MSG_ERR <= DEBUG_MODE))
		{
			std::stringstream stream;
			MPO_MSG_TYPE type = MSG_TYPE_ERROR;

			stream
			<< "ERROR: "
			<< getCallname()
			<< ": array_of_sizes is NULL!"
			<< std::endl;
			rvalue = MARMOT_ERR_POINTER_NULL;

			MPO_String text;
			text.assign(&stream);
			MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
			our_log->getMsg(msg);
		}
		else
		{
			for (int i = 0; i < my_count; i++)
			{
				if ((my_array_of_sizes[i] < 1) &&
					(MSG_ERR <= DEBUG_MODE))
				{
					std::stringstream stream;
					MPO_MSG_TYPE type = MSG_TYPE_ERROR;

					stream
					<< "ERROR: "
					<< getCallname()
					<< ": array_of_sizes["<<i<<"] is less than 1!"
					<< std::endl;
					rvalue = MARMOT_ERR_SIZE_TOO_SMALL;

					MPO_String text;
					text.assign(&stream);
					MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
					our_log->getMsg(msg);
				}
			}
		}

		return rvalue;
	}

	// check_subsizes
	//========================================================================
	int MPO_Type_create_subarray::check_subsizes (void)
	{
		int rvalue = MARMOT_SUCCESS;

		if ((my_count > 0) && (!my_array_of_subsizes) &&
		    (MSG_ERR <= DEBUG_MODE))
		{
			std::stringstream stream;
			MPO_MSG_TYPE type = MSG_TYPE_ERROR;

			stream
			<< "ERROR: "
			<< getCallname()
			<< ": array_of_subsizes is NULL!"
			<< std::endl;
			rvalue = MARMOT_ERR_POINTER_NULL;

			MPO_String text;
			text.assign(&stream);
			MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
			our_log->getMsg(msg);
		}
		else
		{
			for (int i = 0; i < my_count; i++)
			{
				if ((my_array_of_subsizes[i] < 1) &&
				    (MSG_ERR <= DEBUG_MODE))
				{
					std::stringstream stream;
					MPO_MSG_TYPE type = MSG_TYPE_ERROR;

					stream
					<< "ERROR: "
					<< getCallname()
					<< ": array_of_subsizes["<<i<<"] is less than 1!"
					<< std::endl;
					rvalue = MARMOT_ERR_SIZE_TOO_SMALL;

					MPO_String text;
					text.assign(&stream);
					MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
					our_log->getMsg(msg);
				}

				if ( my_array_of_sizes &&
				    (my_array_of_subsizes[i] > my_array_of_sizes[i]) &&
					(MSG_ERR <= DEBUG_MODE))
				{
					std::stringstream stream;
					MPO_MSG_TYPE type = MSG_TYPE_ERROR;

					stream
					<< "ERROR: "
					<< getCallname()
					<< ": array_of_subsizes["<<i<<"] is greater than array_of_sizes["<<i<<"]!"
					<< std::endl;
					rvalue = MARMOT_ERR_SIZE_TOO_BIG;

					MPO_String text;
					text.assign(&stream);
					MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
					our_log->getMsg(msg);
				}
			}
		}

		return rvalue;
	}

	// check_starts
	//========================================================================
	int MPO_Type_create_subarray::check_starts (void)
	{
		int rvalue = MARMOT_SUCCESS;

		if ((my_count > 0) && (!my_array_of_starts) &&
				(MSG_ERR <= DEBUG_MODE))
		{
			std::stringstream stream;
			MPO_MSG_TYPE type = MSG_TYPE_ERROR;

			stream
			<< "ERROR: "
			<< getCallname()
			<< ": array_of_starts is NULL!"
			<< std::endl;
			rvalue = MARMOT_ERR_POINTER_NULL;

			MPO_String text;
			text.assign(&stream);
			MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
			our_log->getMsg(msg);
		}
		else
		{
			for (int i = 0; i < my_count; i++)
			{
				if ((my_array_of_starts[i] < 0) &&
					(MSG_ERR <= DEBUG_MODE))
				{
					std::stringstream stream;
					MPO_MSG_TYPE type = MSG_TYPE_ERROR;

					stream
					<< "ERROR: "
					<< getCallname()
					<< ": array_of_starts["<<i<<"] is less than 0!"
					<< std::endl;
					rvalue = MARMOT_ERR_INDEX_NEG;

					MPO_String text;
					text.assign(&stream);
					MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
					our_log->getMsg(msg);
				}
				else
				if ((my_array_of_starts[i] > my_array_of_sizes[i] - my_array_of_subsizes[i]) &&
					(MSG_ERR <= DEBUG_MODE))
				{
					std::stringstream stream;
					MPO_MSG_TYPE type = MSG_TYPE_ERROR;

					stream
					<< "ERROR: "
					<< getCallname()
					<< ": array_of_starts["<<i<<"] + array_of_subsizes["<<i<<"] is larger than array_of_sizes["<<i<<"]!"
					<< std::endl;
					rvalue = MARMOT_ERR_SIZE_TOO_SMALL;

					MPO_String text;
					text.assign(&stream);
					MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
					our_log->getMsg(msg);
				}
			}
		}

		return rvalue;
	}

	// check_order
	//========================================================================
	int MPO_Type_create_subarray::check_order (void)
	{
		int rvalue = MARMOT_SUCCESS;

		if ( (my_order != MPI_ORDER_C) &&
			 (my_order != MPI_ORDER_FORTRAN) &&
			 (MSG_ERR <= DEBUG_MODE))
		{
			std::stringstream stream;
			MPO_MSG_TYPE type = MSG_TYPE_ERROR;

			stream
			<< "ERROR: "
			<< getCallname()
			<< ": order argument is invalid (is neither MPI_ORDER_C, nor MPI_ORDER_FORTRAN)!"
			<< std::endl;
			rvalue = MARMOT_ERR_ARRAY_ORDER_INVALID;

			MPO_String text;
			text.assign(&stream);
			MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
			our_log->getMsg(msg);
		}

		return rvalue;
	}
}

#endif /*IS_MPI_TYPE_CREATE_SUBARRAY_PRESENT*/

/*EOF*/
