/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 * @file
 * This Sourcefile implements the MPO_LogFile interface.
 * This interface recieves MPO_MsgObject objects and creates a logfile from their data.
 * @author Tobias Hilbrich
 * @date 25.10.2005
 * $Id: MPO_LogFile.cc 955 2009-05-15 14:02:17Z hpczink $
 */

#include <cstring>
#include "SocketLib.h"
#ifdef _WIN32
#include <windows.h>
#endif
#include <assert.h>
#include <stdlib.h>
#include "MPO_LogFile.h"
#include "MPO_MPIStandardRef.h"

// workaround for MinGW
#if defined(__MINGW32__)
# ifndef KEY_WOW64_64KEY
#  define KEY_WOW64_64KEY 0x0100
# endif
# ifndef KEY_WOW64_32KEY
#  define KEY_WOW64_32KEY 0x0200
# endif
#endif

namespace MPO
{
    CSocketLogger *CSocketLogger::m_instance = NULL;
    CSocketLogger::CSocketLogger() : m_desthost(""), m_destport(30777), m_socket(NULL), m_do_ack(true)
    {
      // if env.var MARMOT_LOG_HOST is set try to connect to MARMOT_LOG_HOST:30777 (MARMOT_LOG_PORT overrides the port)
      char *port = NULL;
      char *host = getenv("MARMOT_LOG_HOST");
      if (host)
      {
        m_desthost = host;
        // get custom port
        port = getenv("MARMOT_LOG_PORT");
        if (port)
        {
          std::stringstream sstr;
          sstr << port;
          sstr >> m_destport;
        }
        // get ack override
        if (getenv("MARMOT_DONT_ACK") != NULL)
        {
          std::cerr << "Marmot: will not wait for acknowledge messages!" << std::endl;
          m_do_ack = false;
        }
        // create & connect
        m_socket = new CTcpSocket(m_desthost.c_str(), m_destport);
        if (m_socket && m_socket->isOk())
        {
          m_socket->setLinger(true, 1);
          std::cerr << "Marmot: Successfully connected to " << m_desthost << ":" << m_destport << std::endl;
        } else {
          std::cerr << "Marmot: Could not connect to " << m_desthost << ":" << m_destport << std::endl;
        }
      }
    }
    CSocketLogger::~CSocketLogger()
    {
      if (m_socket)
      {
        delete m_socket;  m_socket = NULL;
      }
      std::cerr << "Marmot: SocketLogger instance destroyed!" << std::endl;
    }
    CSocketLogger *CSocketLogger::instance(void)
    {
      if (m_instance == NULL) m_instance = new CSocketLogger;
      assert(m_instance != NULL);
      return m_instance;
    }
    void CSocketLogger::destroyInstance(void)
    {
      if (m_instance)
      {
        delete m_instance;  m_instance = NULL;
      }
    }
    bool CSocketLogger::waitForAck(void)
    {
      if (!m_socket || !m_do_ack)
        return false;
      int timeout = 5;
      std::string ack_mes;
      while (m_socket->isOk() && timeout > 0)
      {
        if (m_socket->checkSocketReadable())
        {
          ack_mes = m_socket->readln();
          if (ack_mes == "ACK")
          {
            // std::cerr << "Marmot: got ACK!" << std::endl;
            return true;
          }
        }
        timeout--;
      }
      return false;
    }


    MPO_LogFile::MPO_LogFile() : m_iDocPathLength(0), m_iRelativPathLength(0), m_strDocPath(""), m_strRelPath(""), m_strLocPath("")
    {
      m_strDocPath.reserve(512);
      m_strRelPath.reserve(512);
      m_strLocPath.reserve(512);
    }

    MPO_LogFile::~MPO_LogFile ()
    {
    }

    /** Initializes the MPI-Standardreferencer.
     * @return alway true.
     */
    bool MPO_LogFile::initMPIStandardReferencer (void)
    {
        char *docpath = getenv("MARMOT_HOME");
        if (docpath)
        {
          m_strDocPath = docpath;
          m_strDocPath.append("/share/doc/");
          m_strDocPath.append(PACKAGE_STRING);
        } else {
#ifdef _WIN32
          // try registry on windows if no env.var MARMOT_HOME set
          HKEY hKey;
          TCHAR szMarmotInstallationPath[MAX_PATH];
          DWORD dwBufLen=MAX_PATH*sizeof(TCHAR);
          LONG lRet = RegOpenKeyEx(HKEY_LOCAL_MACHINE, TEXT("SOFTWARE\\HLRS\\Marmot"), 0, KEY_QUERY_VALUE|KEY_WOW64_32KEY, &hKey);
          if (lRet != ERROR_SUCCESS) // try reflected
            LONG lRet = RegOpenKeyEx(HKEY_LOCAL_MACHINE, TEXT("SOFTWARE\\HLRS\\Marmot"), 0, KEY_QUERY_VALUE|KEY_WOW64_64KEY, &hKey);
          if (lRet == ERROR_SUCCESS)
          {
            lRet = RegQueryValueEx(hKey, NULL, NULL, NULL, (LPBYTE)szMarmotInstallationPath, &dwBufLen);
            if (lRet == ERROR_SUCCESS)
            {
              m_strDocPath = szMarmotInstallationPath;
              m_strDocPath.append("/share/doc/");
              m_strDocPath.append(PACKAGE_STRING);
            } else {
              m_strDocPath = DOCPATH;
            }
            RegCloseKey(hKey);
          }
#else
          m_strDocPath = DOCPATH;
#endif
        }

#ifdef _WIN32
        if (!m_strDocPath.empty())
        {
          for (string::size_type i=0, pos; (pos = m_strDocPath.find("\\", i)) != string::npos; i=pos + 1)
          {
            m_strDocPath.replace(pos, 1, "/");
          }
        }
#endif

        m_iDocPathLength = m_strDocPath.length();
        m_iRelativPathLength = strlen(HTMLDOCBASEDIR);

        return true;
    }

  /**
   * @brief  Initializes connection to a log-host
   * If the environment variable MARMOT_LOG_HOST is set and the given host is reachable. The default port is 30777 and can be overidden by setting
   * MARMOT_LOG_PORT.
   * @return Returns true if a connection could be set up successfully, false otherwise. False is also returned if MARMOT_LOG_HOST is not set.
   */
    bool MPO_LogFile::initSocket()
    {
      CSocketLogger *sl = CSocketLogger::instance();
      if (sl && sl->getSocket())
        return true;
      return false;
    }

    /** Returns a link to a HTML file that points to the MPI-Standard and describes the given ERROR-CODE, the html file is in the local doc directory.
     * @param err describes the errorcode that occured.
     * @return the function returns the path/filename of the HTML file that points to the MPI-Standard and describes the given ERROR-CODE. If case of MARMOT_LAST_ERR_CODE, NULL is returned.
     */
    const char *MPO_LogFile::getLocalPathToMPIStandard(MARMOT_ERR_CODES err)
    {
        //sanity check
        if (err == MARMOT_LAST_ERR_CODE)
        {
          return NULL;
        }

        m_strLocPath = m_strDocPath + HTMLDOCBASEDIR;
        m_strLocPath.append(g_MPIRefs[err]);

        return m_strLocPath.c_str();
    }

    /** Returns a link to a HTML file that points to the MPI-Standard and describes the given ERROR-CODE, the string describes a address relativ to the doc directory in order to enable mirroring.
     * @param err describes the errorcode that occured.
     * @return the function returns the relative path/filename of the HTML file that points to the MPI-Standard and describes the given ERROR-CODE. If case of MARMOT_LAST_ERR_CODE, NULL is returned.
     */
    const char *MPO_LogFile::getRelativPathToMPIStandard(MARMOT_ERR_CODES err)
    {
        //sanity check
        if (err == MARMOT_LAST_ERR_CODE)
        {
          return NULL;
        }

        m_strRelPath = HTMLDOCBASEDIR;
        m_strRelPath.append(g_MPIRefs[err]);

        return m_strRelPath.c_str();
    }

    /** Returns the local path to the doc directory.
     * @return Local path to the doc directory.
     */
    const char *MPO_LogFile::getLocalDocPath()
    {
      return m_strDocPath.c_str();
    }

}//end of namespace MPO
