/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of the base class MPO_Starts.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: MPO_Starts.cc 1071 2010-02-08 18:29:20Z shiqing $
 */


#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"
#include <stdio.h>

namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Starts::MPO_Starts()
        : my_count(-1),
          my_request_array(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Starts::MPO_Starts(int count,
                           MPI_Request * array)
        : my_count(count),
          my_request_array(array)
    {
    }


    // destructor
    //========================================================================
    MPO_Starts::~MPO_Starts()
    {
    }


    /* There are some pure virtual methods, see mpo.h.
     */


    // getFamilyId
    //========================================================================
    int MPO_Starts::getFamilyId() const
    {
        // returns the unique family identity set by enum MPO_FAMILY_ID
        return MPO_CALLOBJECT_ID;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Starts::checkPrerequisites()
    {
        // check correctness of parameters
        checkRequestCountArray();
        checkRequestArray();

        return 0;
    }


    // checkRequestCountArray
    //========================================================================
    int MPO_Starts::checkRequestCountArray()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if count is negative.
        if ((our_errcodes.at(MARMOT_ERR_REQUEST_COUNT_NEG))&&(my_count < 0) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": count of requests = "
                << my_count
                << " less than 0!"
                << std::endl;
            rvalue = MARMOT_ERR_REQUEST_COUNT_NEG;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        // (2) Check if count is 0.
        if ((our_errcodes.at(MARMOT_WARN_REQUEST_COUNT_ZERO))&&(my_count == 0) &&
            (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "WARNING: "
                << getCallname()
                << ": count of requests is zero!"
                << std::endl;
            rvalue = MARMOT_WARN_REQUEST_COUNT_ZERO;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }


    // checkRequestArray
    //========================================================================
    int MPO_Starts::checkRequestArray()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        std::map < MPO_Request, enum KIND_OF_REQUEST >::iterator pos;

        // Check if request_array is correct.
        for (int i = 0; i <= my_count - 1; i++)
        {
            pos = our_requests.find(&my_request_array[i]);

            // (1) Request is not valid, i.e. not registered in the map
            // of our_requests.
            if ((our_errcodes.at(MARMOT_ERR_REQUEST_NOT_VALID))&&(pos == our_requests.end()) &&
                (MSG_ERR <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname()
                    << ": Request "
                    << i
                    << " does not exist, is not / no more valid! "
                    << std::endl;
                rvalue = MARMOT_ERR_REQUEST_NOT_VALID;

                string s("request[");
                char buf[64];
                sprintf (buf,"%d",i);
                s.append(buf);
                s.append("]");
                our_request_traces->findAndPrintTraceNamed(&(my_request_array[i]),stream,s);

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                our_log->getMsg(msg);
            }

            // (2) Request is registered in the map of our_requests.
            // In this case, we have to check different violations
            // in the switch statement.
            else if ((pos != our_requests.end()) &&
                     (MSG_ERR <= DEBUG_MODE))
            {
                // all the cases from enum KIND_OF_REQUEST
                switch (pos->second)
                {
                    case (REQUEST_NULL):
                        if ((our_errcodes.at(MARMOT_WARN_REQUEST_NULL))&&(MSG_ERR_WARN <= DEBUG_MODE))
                        {
                            std::stringstream stream;
                            stream
                                << "WARNING: "
                                << getCallname()
                                << ": Request "
                                << i
                                << " is MPI_REQUEST_NULL! "
                                << std::endl;
                            rvalue = MARMOT_WARN_REQUEST_NULL;

                            string s("request[");
			                char buf[64];
			                sprintf (buf,"%d",i);
			                s.append(buf);
			                s.append("]");
			                our_request_traces->findAndPrintTraceNamed(&(my_request_array[i]),stream,s);

                            MPO_String text;
                            text.assign(&stream);
                            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                            our_log->getMsg(msg);
                        }
                        break;
                    case (REQUEST_NOT_PERSISTENT):
                    {
			    if(our_errcodes.at(MARMOT_ERR_REQUEST_NOT_PERSISTENT)&&(MSG_ERR <= DEBUG_MODE))
			    {
                        std::stringstream stream;
                        stream
                            << "ERROR: "
                            << getCallname()
                            << ": Request "
                            << i
                            << " is no persistent request! "
                            << std::endl;
                        rvalue = MARMOT_ERR_REQUEST_NOT_PERSISTENT;

                        string s("request[");
		                char buf[64];
		                sprintf (buf,"%d",i);
		                s.append(buf);
		                s.append("]");
		                our_request_traces->findAndPrintTraceNamed(&(my_request_array[i]),stream,s);

                        MPO_String text;
                        text.assign(&stream);
                        MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                        our_log->getMsg(msg);
			    }
                        break;
                    }
                    case (REQUEST_ACTIVE_PERSISTENT):
                        if ((our_errcodes.at(MARMOT_WARN_REQUEST_ACTIVE_PERSISTENT))&&(MSG_ERR_WARN <= DEBUG_MODE))
                        {
                            std::stringstream stream;
                            stream
                                << "WARNING: "
                                << getCallname()
                                << ": Request "
                                << i
                                << " is persistent request, but still active! "
                                << std::endl;
                            rvalue = MARMOT_WARN_REQUEST_ACTIVE_PERSISTENT;

                            string s("request[");
			                char buf[64];
			                sprintf (buf,"%d",i);
			                s.append(buf);
			                s.append("]");
			                our_request_traces->findAndPrintTraceNamed(&(my_request_array[i]),stream,s);

                            MPO_String text;
                            text.assign(&stream);
                            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                            our_log->getMsg(msg);
                        }
                        break;
                    case (REQUEST_INACTIVE_PERSISTENT):
                        // Everything is fine. Set request to active.
                        pos->second = REQUEST_ACTIVE_PERSISTENT;
                        break;
                    default:
                    {
			    if(our_errcodes.at(MARMOT_ERR_MARMOT))
			    {
                        std::stringstream stream;
                        stream
                            << "Check mapping of requests."
                            << std::endl
                            << "Check implementation of switch statement."
                            << std::endl;
                        rvalue = MARMOT_ERR_MARMOT;

                        string s("request[");
		                char buf[64];
		                sprintf (buf,"%d",i);
		                s.append(buf);
		                s.append("]");
		                our_request_traces->findAndPrintTraceNamed(&(my_request_array[i]),stream,s);

                        MPO_String text;
                        text.assign(&stream);
                        MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                        our_log->getMsg(msg);
			    }
                        break;
                    }           // end default
                }               // end switch
            }                   // end else if
        }                       // end for

        return rvalue;
    }                           // end checkRequestArray

}                               // end of namespace MPO
