/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
  * \file MPO_ThreadSynchronisationSyncBase.cc
  * This Sourcefile implements the MPO_ThreadSynchronisationSyncBase class.
  * @author Tobias Hilbrich
  * @date 27.06.2007
  * $Id$
  */

#include "mpo_SourceLocation.h"
#include "MPO_ThreadSynchronisationSyncBase.h"
#include "mpo_String.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "mpo-objects.h"
#include "mpo_TCheck_Races.h"

namespace MPO
{

    // MPO_ThreadSynchronisationSyncBase
    //========================================================================
    MPO_ThreadSynchronisationSyncBase::MPO_ThreadSynchronisationSyncBase(void)
    {
        my_registeredRequests.clear();
        my_registeredComms.clear();
        my_minThreadLevel = MPI_THREAD_SINGLE;

        //init one thread as not in MARMOT
        my_MPO_in_profiles.clear();
        my_MPO_in_profiles.resize(1);
        my_MPO_in_profiles[0] = 0;

#ifdef MARMOT_TCHECK
        my_nextRequestId = 0;
        my_requestToIntConversion.clear ();
        my_nextCommId = 0;
        my_commToIntConversion.clear();
#endif
    }

    // MPO_ThreadSynchronisationSyncBase
    //========================================================================
    MPO_ThreadSynchronisationSyncBase::~MPO_ThreadSynchronisationSyncBase(void)
    {
        my_registeredRequests.clear();
    }

    // enterMARMOT
    //========================================================================
    bool MPO_ThreadSynchronisationSyncBase::enterMARMOT (int line, const char *file)
    {

#ifdef MARMOT_TCHECK
    //write - write Data race for TCheck in order to detect MPI_THREAD_MULTIPLE usage
    MPO_TCheck_Races::getInstance()->write(APP_NEEDS_MPI_THREAD_MULTIPLE);

    //read - write Data race for TCheck in order to detect an MPI_FINALIZE
    //performed while other threads are still doing MPI
    MPO_TCheck_Races::getInstance()->read(APP_MPI_FINALIZE_TO_EARLY);
#endif
        //Start synchronisation, needed to determine whether already in Marmot
        beginCritical();
#ifdef MARMOT_USE_THREADSYNC
#ifdef MARMOT_THREADED

        /**
         * \todo the synchronisation used here might slow the enterMARMOT tests down
         *       maybe its beter to accept the existance of an relatively harmles race
         *       condition here for performances sake.
         */
        //beginCritical();

        //test for min level of thread support
        switch (my_minThreadLevel)
        {
            case MPI_THREAD_SINGLE:
            {
                //test for violation
                if (!isInParallelRegion())
                    break;

                if (getNumThreads() < 2)
                    break;

                //set new level and violation info
                my_minThreadLevel = MPI_THREAD_FUNNELED;
                MPO_ThreadViolationInfo violationInfo (getThreadRank(),MPO_SourceLocation(file,line));
                my_threadViolationInfos.insert(std::make_pair((int)(MPI_THREAD_SINGLE),violationInfo));

                //do we have to log an error ?
                int provided = MPO_Space::getProvidedThreadLevel();
                if (provided < my_minThreadLevel)
                {
                    std::stringstream stream;
                    stream
                        << "ERROR: "
                        << "Violation to the provided thread level ("
                        << MPO_Space::getThreadLevelName(provided)
                        << "). Conflict caused by thread "
                        << getThreadRank()
                        << " in line "
                        << line
                        << " of file "
                        << file
                        << std::endl
                        << "(Multiple threads were present)"
                        << std::endl
                        << "For this application you need at least MPI_THREAD_FUNNELED!"
                        << "(Further messages might specify a higher Thread level)";

                    MPO_String text;
                    text.assign(&stream);
                    MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, MARMOT_ERR_VIOLATED_THREAD_LEVEL, MPO_SourceLocation(file,line));

                    //we have to make full synchonisation here as we are going to use something global!
                    endCritical();
                    leavePMPI();
                    MPO_LogFile *log = MPO_Space::getLogFile();
                    if (log)
                        log->getMsg(msg);
                    enterPMPI();
                    beginCritical();
                }

                //do not break here as we might also violate MPI_THREAD_FUNNELED !
            }

            case MPI_THREAD_FUNNELED:
            {
                int main_thread;
                PMPI_Is_thread_main(&main_thread);

                if (main_thread)
                    break;

                my_minThreadLevel = MPI_THREAD_SERIALIZED;
                MPO_ThreadViolationInfo violationInfo (getThreadRank(),MPO_SourceLocation(file,line));
                my_threadViolationInfos.insert(std::make_pair((int)(MPI_THREAD_FUNNELED),violationInfo));

                //do we have to log an error ?
                int provided = MPO_Space::getProvidedThreadLevel();
                if (provided < my_minThreadLevel)
                {
                    std::stringstream stream;
                    stream
                        << "ERROR: "
                        << "Violation to the provided thread level ("
                        << MPO_Space::getThreadLevelName(provided)
                        << "). Conflict caused by thread "
                        << getThreadRank()
                        << " in line "
                        << line
                        << " of file "
                        << file
                        << std::endl
                        << "(This thread did not call MPI_Init[_thread])"
                        << std::endl
                        << "For this application you need at least MPI_THREAD_SERIALIZED!"
                        << "(Further messages might specify a higher Thread level)";

                    MPO_String text;
                    text.assign(&stream);
                    MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, MARMOT_ERR_VIOLATED_THREAD_LEVEL, MPO_SourceLocation(file,line));

                    //we have to make full synchonisation here as we are going to use something global!
                    endCritical();
                    leavePMPI();
                    MPO_LogFile *log = MPO_Space::getLogFile();
                    if (log)
                        log->getMsg(msg);
                    enterPMPI();
                    beginCritical();
                }

                //do not break here as we might also violate MPI_THREAD_SERIALIZED !
            }

            case MPI_THREAD_SERIALIZED:
            {
                if (!otherThreadsInMARMOT())
                    break;

                my_minThreadLevel = MPI_THREAD_MULTIPLE;
                MPO_ThreadViolationInfo violationInfo (getThreadRank(),MPO_SourceLocation(file,line));
                std::pair<int,MPO_SourceLocation> otherInfo = getInfoForThreadInMARMOT();
                violationInfo.addParticipant(otherInfo.first,otherInfo.second);
                my_threadViolationInfos.insert(std::make_pair((int)(MPI_THREAD_SERIALIZED),violationInfo));

                //do we have to log an error ?
                int provided = MPO_Space::getProvidedThreadLevel();
                if (provided < my_minThreadLevel)
                {
                    std::stringstream stream;
                    stream
                        << "ERROR: "
                        << "Violation to the provided thread level ("
                        << MPO_Space::getThreadLevelName(provided)
                        << "). Conflict caused by the following participants:"
                        << std::endl;

                    violationInfo.printParticipants(stream);

                    stream
                        << std::endl
                        << "(Multiple threads made MPI calls simultaneously)"
                        << std::endl
                        << "For this application you need at least MPI_THREAD_MULTIPLE!";

                    MPO_String text;
                    text.assign(&stream);
                    MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, MARMOT_ERR_VIOLATED_THREAD_LEVEL, MPO_SourceLocation(file,line));

                    //we have to make full synchonisation here as we are going to use something global!
                    endCritical();
                    leavePMPI();
                    MPO_LogFile *log = MPO_Space::getLogFile();
                    if (log)
                        log->getMsg(msg);
                    enterPMPI();
                    beginCritical();
                }

            }//end case MPI_THREAD_SERIALIZED

        }//end switch

        //endCritical();

#endif
#endif

        //test whether this call is done from within MPI
        //(MPI call issued by implementation of another MPI call)
        checkIsNewThread ();
        int current_entry = my_MPO_in_profiles[getThreadRank()];
        my_MPO_in_profiles[getThreadRank()] = current_entry + 1;
        endCritical();

        //abort enter MARMOT in case we are already in an MPI call catched with MARMOT
        //We should not need synchronisation here as MARMOT should not touch anything global
        //in this case.
        if (current_entry != 0)
        {
            return false;
        }

        //enter the syncronisation
        enterMARMOTSync(line,file);

        return true;
    }

    // leaveMARMOT
    //========================================================================
    void MPO_ThreadSynchronisationSyncBase::leaveMARMOT (int line, const char *file,bool enter_succeeded)
    {
        //(1) leave the "in-Marmot" synchronization
        if (enter_succeeded)
        {
            //leave the syncronisation
            leaveMARMOTSync(line,file);
        }

        //(2) reduce the in Marmot depth by one
        beginCritical();
        my_MPO_in_profiles[getThreadRank()] = my_MPO_in_profiles[getThreadRank()] - 1;
        endCritical();

    }

    // wrappingOn
    //========================================================================
    void MPO_ThreadSynchronisationSyncBase::wrappingOn (void)
    {
        //May only be called from outside Marmot,
        //so my_MPO_in_profiles is either 0 or 1
        beginCritical();
        checkIsNewThread ();
        int cur_val = my_MPO_in_profiles[getThreadRank()];
        if (cur_val == 1)
            my_MPO_in_profiles[getThreadRank()] = 0;
        endCritical();
    }

    // wrappingOf
    //========================================================================
    void MPO_ThreadSynchronisationSyncBase::wrappingOff (void)
    {
        //May only be called from outside Marmot,
        //so my_MPO_in_profiles is either 0 or 1
        beginCritical();
        checkIsNewThread ();
        int cur_val = my_MPO_in_profiles[getThreadRank()];
        if (cur_val == 0)
            my_MPO_in_profiles[getThreadRank()] = 1;
        endCritical();
    }

    // didEnterSucceed
    //========================================================================
    bool MPO_ThreadSynchronisationSyncBase::didEnterSucceed (void)
    {
        bool retval = false;

        beginCritical();
        int cur_val = my_MPO_in_profiles[getThreadRank()];
        if (cur_val == 1)
            retval = true;
        endCritical();

        return retval;
    }

    // checkIsNewThread
    //========================================================================
    void MPO_ThreadSynchronisationSyncBase::checkIsNewThread (void)
    {
        int oldsize = my_MPO_in_profiles.size();
        int numthreads = getNumThreads();
        if (oldsize <= numthreads)
        {
            my_MPO_in_profiles.resize(numthreads);

            //init newly created entries
            for (int i = oldsize;i < numthreads;i++)
                my_MPO_in_profiles[i] = 0;
        }
    }

    // registerRequest
    //========================================================================
    int MPO_ThreadSynchronisationSyncBase::registerRequest (MPI_Request *request, int sourceLine,const char *sourceName, int threadNum, int callId, std::string callName)
    {
        return registerRequests (request, 1, sourceLine, sourceName, threadNum, callId, callName);
    }

    // registerRequests
    //========================================================================
    int MPO_ThreadSynchronisationSyncBase::registerRequests (MPI_Request *requests, int count, int sourceLine,const char *sourceName, int threadNum, int callId, std::string callName)
    {
        int rvalue;

#ifdef MARMOT_TCHECK
        //write - write Data race for TCheck in order to detect usage of a request
        //in two thread-parallel MPI calls
        //therfore we have to shutdown race condition protection !!!
        beginCritical();

        int *iRequestIDs = new int[count];

        for (int i = 0; i < count;i++)
        {
            std::map<MPO_Request,int>::iterator itRequestPos = my_requestToIntConversion.find(&(requests[i]));
            if (itRequestPos != my_requestToIntConversion.end())
            {
                iRequestIDs[i] = itRequestPos->second;
                continue;
            }

            my_requestToIntConversion.insert(std::make_pair(&(requests[i]),my_nextRequestId));
            iRequestIDs[i] = my_nextRequestId;
            my_nextRequestId++;

            if (my_nextRequestId >= TCHECK_NUM_REQUEST_SLOTS)
                my_nextRequestId = 0;
        }
        endCritical();

        for (int i = 0; i < count;i++)
            MPO_TCheck_Races::getInstance()->writeIntForRequests(iRequestIDs[i]);

        beginCritical();
        if (iRequestIDs)
            delete (iRequestIDs);
        endCritical();
#endif

        //start critical section
        beginCritical();

        bool bRun = true;

        //is the request already used ?
        for (my_mapType::iterator i = my_registeredRequests.begin();i != my_registeredRequests.end() ;i++)
        {
            std::vector<MPI_Request*> *pReqs = &(i->second.first);
            for (std::vector<MPI_Request*>::iterator j = pReqs->begin();(j != pReqs->end()) && bRun;j++)
            {
                for (int r = 0;(r < count) && bRun;r++)
                {
                    if (*j == &(requests[r]))
                    {
                        logRequestUsageError(&(i->second.second), sourceLine, sourceName, threadNum, callId, callName);
                        bRun = false;
                    }
                }
            }
        }

        //insert an empty vector with the next free id and the source information
        std::vector<MPI_Request*> v;
        sourceInfo s;
        s.loc = MPO_SourceLocation(sourceName,sourceLine);
        s.threadId = threadNum;
        s.callId = callId;
        s.callName = callName;
        my_registeredRequests.insert(
            std::make_pair(
                threadNum,
                std::make_pair(v,s)
                )
            );

        rvalue = threadNum;

        //add the requests to the newly created entry
        for (int r = 0;r < count;r++)
            my_registeredRequests[rvalue].first.push_back(&(requests[r]));

        //end critical section
        endCritical();

        return rvalue;
    }

    // unregisterRequests
    //========================================================================
    void MPO_ThreadSynchronisationSyncBase::unregisterRequests (int id)
    {
        //start critical section
        beginCritical();

        //remove the entry
        my_registeredRequests.erase(id);

        //end critical section
        endCritical();
    }

#ifdef MARMOT_TCHECK
    // unregisterRequestToIntEntry
    //========================================================================
    void MPO_ThreadSynchronisationSyncBase::unregisterRequestToIntEntry(MPI_Request *request)
    {
        beginCritical();
        my_requestToIntConversion.erase(request);
        endCritical();
    }
#endif

    // logRequestUsageError
    //========================================================================
    void MPO_ThreadSynchronisationSyncBase::logRequestUsageError (sourceInfo *pInfo, int sourceLine,const char *sourceName, int threadNum, int callId, std::string callName)
    {
        //Request already used -> give an error
        //a) log first direction
        std::stringstream stream;
            stream
                << "Error on call "
                << callName
                << std::endl
                << "Same request was used in two MPI Calls simultaneously (on different threads)."
                << std::endl
                << "Conflicting with: Thread "
                << pInfo->threadId
                << " Call "
                << pInfo->callName
                << " ";

        pInfo->loc.debugout(stream);

        MPO_String text;
        text.assign(&stream);
        MPO_MsgObject msg1 (MSG_TYPE_ERROR,0,text, callName ,(MPO_CALL_ID) callId,MARMOT_ERR_REQUEST_USED_AT_MULTIPLE_THREADS,MPO_SourceLocation(sourceName,sourceLine));

        //Request already used -> give an error
        //b) log second direction
        stream.str("");
        stream.clear();
        stream
            << "Error on call "
            << pInfo->callName
            << std::endl
            << "Same request was used in two MPI Calls simultaneously (on different threads)."
            << std::endl
            << "Conflicting with: Thread "
            << threadNum
            << " Call "
            << callName
            << " ";

        MPO_SourceLocation(sourceName,sourceLine).debugout(stream);

        text.assign(&stream);
        MPO_MsgObject msg2 (MSG_TYPE_ERROR,0,text, pInfo->callName ,(MPO_CALL_ID) pInfo->callId,MARMOT_ERR_REQUEST_USED_AT_MULTIPLE_THREADS,pInfo->loc);
        msg2.setThreadRank(pInfo->threadId);

        //we have to make full synchonisation here as we are going to use something global!
        endCritical();
        leavePMPI();
        MPO_LogFile *log = MPO_Space::getLogFile();
        if (log)
            log->getMsg(msg1);
        if (log)
            log->getMsg(msg2);
        enterPMPI();
        beginCritical();
    }

    // registerCollComm
    //========================================================================
    int MPO_ThreadSynchronisationSyncBase::registerCollComm (MPI_Comm comm, int sourceLine,const char *sourceName, int threadNum, int callId, std::string callName)
    {
        int rvalue;

#ifdef MARMOT_TCHECK
        //write - write Data race for TCheck in order to detect usage of a request
        //in two thread-parallel MPI calls
        //therfore we have to shutdown race condition protection !!!
        beginCritical();

        int commID;

        std::map<MPI_Comm,int>::iterator itCommPos = my_commToIntConversion.find(comm);
        if (itCommPos != my_commToIntConversion.end())
        {
            commID = itCommPos->second;
        }
        else
        {
            my_commToIntConversion.insert(std::make_pair(comm,my_nextCommId));
            commID = my_nextCommId;
            my_nextCommId++;

            if (my_nextCommId >= TCHECK_NUM_COMM_SLOTS)
                my_nextCommId = 0;
        }

        endCritical();

        MPO_TCheck_Races::getInstance()->writeIntForComm(commID);
#endif

        //start critical section
        beginCritical();

        //is the communicator already used ?
        for (my_mapTypeComm::iterator i = my_registeredComms.begin(); i != my_registeredComms.end() ; i++)
        {
            MPI_Comm tempComm = i->second.first;
            if (comm == tempComm)
            {
                logCommUsageError(comm,&(i->second.second), sourceLine, sourceName, threadNum, callId, callName);
            }
        }

        //insert an empty vector with the next free id and the source information
        sourceInfo s;
        s.loc = MPO_SourceLocation(sourceName, sourceLine);
        s.threadId = threadNum;
        s.callId = callId;
        s.callName = callName;
        my_registeredComms.insert(std::make_pair(threadNum, std::make_pair(comm, s)));

        rvalue = threadNum;

        //end critical section
        endCritical();

        return rvalue;
    }

    // unregisterCollComm
    //========================================================================
    void MPO_ThreadSynchronisationSyncBase::unregisterCollComm (int id)
    {
        //start critical section
        beginCritical();

        //remove the entry
        my_registeredComms.erase(id);

        //end critical section
        endCritical();
    }

#ifdef MARMOT_TCHECK
    // unregisterCommToIntEntry
    //========================================================================
    void MPO_ThreadSynchronisationSyncBase::unregisterCommToIntEntry(MPI_Comm comm)
    {
        beginCritical();
        my_commToIntConversion.erase(comm);
        endCritical();
    }
#endif

    // logRequestUsageError
    //========================================================================
    void MPO_ThreadSynchronisationSyncBase::logCommUsageError (MPI_Comm comm,sourceInfo *pInfo, int sourceLine,const char *sourceName, int threadNum, int callId, std::string callName)
    {
        //Request already used -> give an error
        //a) log first direction
        std::stringstream stream;
        stream
                << "Error on call "
                << callName
                << ". This Process used the same Communicator in two collective calls executed on two different threads."
                << std::endl
                << "Conflicting with: Thread "
                << pInfo->threadId
                << " Call "
                << pInfo->callName
                << " ";

        pInfo->loc.debugout(stream);

        MPO_String text;
        text.assign(&stream);
        MPO_MsgObject msg1 (MSG_TYPE_ERROR,0,text, callName ,(MPO_CALL_ID) callId,MARMOT_ERR_COMM_USED_IN_MULTIPLE_COLLECTIVES,MPO_SourceLocation(sourceName,sourceLine));

        //Request already used -> give an error
        //b) log second direction
        stream.str("");
        stream.clear();
        stream
            << "Error on call "
            << pInfo->callName
            << ". This Process used the same Communicator in two collective calls executed on two different threads."
            << std::endl
            << "Conflicting with: Thread "
            << threadNum
            << " Call "
            << callName
            << " ";

        MPO_SourceLocation(sourceName,sourceLine).debugout(stream);

        text.assign(&stream);
        MPO_MsgObject msg2 (MSG_TYPE_ERROR,0,text, pInfo->callName ,(MPO_CALL_ID) pInfo->callId,MARMOT_ERR_COMM_USED_IN_MULTIPLE_COLLECTIVES,pInfo->loc);
        msg2.setThreadRank(pInfo->threadId);

        //we have to make full synchonisation here as we are going to use something global!
        endCritical();
        leavePMPI();
        MPO_LogFile *log = MPO_Space::getLogFile();
        if (log)
            log->getMsg(msg1);
        if (log)
            log->getMsg(msg2);
        enterPMPI();
        beginCritical();
   }

    // getMinThreadLvl
    //========================================================================
    int MPO_ThreadSynchronisationSyncBase::getMinThreadLvl (void)
    {
        return my_minThreadLevel;
    }

    // getViolationInfoForThreadLvl
    //========================================================================
    MPO_ThreadViolationInfo MPO_ThreadSynchronisationSyncBase::getViolationInfoForThreadLvl (int threadLevel)
    {
        if (threadLevel >= my_minThreadLevel)
            return MPO_ThreadViolationInfo();

        return my_threadViolationInfos[threadLevel];
    }

}//end of namespace MPO







