/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of the base class MPO_Typeconstructors.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: MPO_TypeConstructors.cc 851 2008-07-17 08:34:15Z hpchimml $
 */


#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"

namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Typeconstructors::MPO_Typeconstructors()
        : my_count(-1),
          my_oldtype(MPI_DATATYPE_NULL), 
          my_id(-1)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Typeconstructors::MPO_Typeconstructors(int count,
                                               MPO_Datatype oldtype,
                                               int id)
        : my_count(count),
          my_oldtype(oldtype), 
          my_id(id)
    {
    }


    // destructor
    //========================================================================
    MPO_Typeconstructors::~MPO_Typeconstructors()
    {
    }


    /** There are some pure virtual methods, see mpo.h.
     */


    // getFamilyId
    //========================================================================
    int MPO_Typeconstructors::getFamilyId() const
    {
        // returns the unique family identity set by enum MPO_FAMILY_ID
        return MPO_CALLOBJECT_ID;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Typeconstructors::checkPrerequisites()
    {
        // check correctness of parameters

        checkCount(MSG_COUNT);

        // For Type_struct, we check datatypes separately.
        if (my_id != MPO_TYPE_STRUCT_ID)
        {
            checkDatatype();
        }

        return 0;
    }


    // checkCount
    //========================================================================
    int MPO_Typeconstructors::checkCount(const string KIND_OF_COUNT)
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if count is negative.
        if ((our_errcodes.at(MARMOT_ERR_COUNT_NEG))&&(my_count < 0) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": "
                << KIND_OF_COUNT
                << " = "
                << my_count
                << " less than 0 !" 
                << std::endl;
            rvalue = MARMOT_ERR_COUNT_NEG;
            
            our_datatype_traces->findAndPrintTraceNamed(my_oldtype,stream,"oldtype");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (2) Check if count is 0.
        if ((our_errcodes.at(MARMOT_WARN_COUNT_ZERO))&&(my_count == 0) && 
            (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "WARNING: " 
                << getCallname() 
                << ": "
                << KIND_OF_COUNT
                << "= 0!" 
                << std::endl;
            rvalue = MARMOT_WARN_COUNT_ZERO;
            
            our_datatype_traces->findAndPrintTraceNamed(my_oldtype,stream,"oldtype");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        return rvalue;
    }


    // checkDatatype
    //========================================================================
    int MPO_Typeconstructors::checkDatatype()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        std::map < MPO_Datatype, enum KIND_OF_DATATYPE >::iterator pos;
        pos = our_datatypes.find(my_oldtype);

        // (1) Datatype is not valid, i.e. not registered in
        // our MPO datatypes.
        if ((our_errcodes.at(MARMOT_ERR_TYPE_NOT_VALID))&&(pos == our_datatypes.end()) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": oldtype is not valid!" 
                << std::endl;
            rvalue = MARMOT_ERR_TYPE_NOT_VALID;
            
            our_datatype_traces->findAndPrintTraceNamed(my_oldtype,stream,"oldtype");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (2) Datatype is registered in our MPO datatypes.
        // In this case, we have to check different violations 
        // in the switch statement.
        else if ((pos != our_datatypes.end()) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            
            std::stringstream stream;
            MPO_MSG_TYPE type = MSG_TYPE_ERROR;
            
            // all the cases from enum KIND_OF_DATATYPE
            switch (pos->second)
            {
                case (DATATYPE_NULL):
			if(our_errcodes.at(MARMOT_ERR_TYPE_NULL))
		    {
                    	stream
                        << "ERROR: "
                        << getCallname() 
                        << ": oldtype is MPI_DATATYPE_NULL!" 
                        << std::endl;
                    	rvalue = MARMOT_ERR_TYPE_NULL;
		    }
                    break;
                case (DATATYPE_CREATED_NOT_COMMITTED):
                	//see the MPI Standard ! this isn`t even a Warning !
       /*             stream
                        << "ERROR: "
                        << getCallname() 
                        << ": oldtype is not committed!" 
                        << std::endl;
                    rvalue = MARMOT_ERR_TYPE_NOT_COMMITTED;*/
                    break;
                case (DATATYPE_COMMITTED_C):
                    if ((our_errcodes.at(MARMOT_ERR_TYPE_C))&&(F_INTERFACE == INTERFACE_MODE))
                    {
                        stream
                            << "ERROR: "
                            << getCallname() 
                            << ": oldtype is C-Type!" 
                            << std::endl;
                        rvalue = MARMOT_ERR_TYPE_C;
                    }
                    break;
                case (DATATYPE_COMMITTED_C_AND_FORTRAN):
                    // Do nothing, because everything is fine.
                    break;
                case (DATATYPE_COMMITTED_FORTRAN):
                    if ((our_errcodes.at(MARMOT_ERR_TYPE_F))&&(C_INTERFACE == INTERFACE_MODE))
                    {
                        stream
                            << "ERROR: "
                            << getCallname() 
                            << ": oldtype is Fortran-Type!" 
                            << std::endl;
                        rvalue = MARMOT_ERR_TYPE_F;
                    }
                    break;
                case (DATATYPE_REDUCTION_C):
                    if ((our_errcodes.at(MARMOT_WARN_TYPE_REDUCTION_C))&&(MSG_ERR_WARN <= DEBUG_MODE) &&
                        (C_INTERFACE == INTERFACE_MODE))
                    {
                        stream 
                            << "WARNING: "
                            << getCallname()
                            << ": oldtype is for reduction functions (C)!" 
                            << std::endl;
                        rvalue = MARMOT_WARN_TYPE_REDUCTION_C;
                        type = MSG_TYPE_WARNING;
                    }
                    if ((our_errcodes.at(MARMOT_ERR_TYPE_REDUCTION_C))&&(F_INTERFACE == INTERFACE_MODE))
                    {
                        stream
                            << "ERROR: "
                            << getCallname()
                            << ": oldtype is for reduction functions (C)!" 
                            << std::endl;
                        rvalue = MARMOT_ERR_TYPE_REDUCTION_C;
                    }
                    break;
                case (DATATYPE_REDUCTION_FORTRAN):
                    if ((our_errcodes.at(MARMOT_ERR_TYPE_REDUCTION_F))&&(C_INTERFACE == INTERFACE_MODE))
                    {
                        stream
                            << "ERROR: "
                            << getCallname()
                            << ": oldtype is for reduction functions "
                            << "(Fortran)!" 
                            << std::endl;
                        rvalue = MARMOT_ERR_TYPE_REDUCTION_F;
                    }
                    if ((our_errcodes.at(MARMOT_WARN_TYPE_REDUCTION_F))&&(MSG_ERR_WARN <= DEBUG_MODE) &&
                        (F_INTERFACE == INTERFACE_MODE))
                    {
                        stream
                            << "WARNING: "
                            << getCallname()
                            << ": oldtype is for reduction functions "
                            << "(Fortran)!" 
                            << std::endl;
                        rvalue = MARMOT_WARN_TYPE_REDUCTION_F;
                        type = MSG_TYPE_WARNING;
                    }
                    break;
                case (DATATYPE_OPTIONAL_C):
                    if ((our_errcodes.at(MARMOT_WARN_TYPE_OPTIONAL_C))&&(MSG_ERR_WARN <= DEBUG_MODE) &&
                        (C_INTERFACE == INTERFACE_MODE))
                    {
                        stream
                            << "WARNING: "
                            << getCallname() 
                            << ": oldtype is only optional!" 
                            << std::endl;
                        rvalue = MARMOT_WARN_TYPE_OPTIONAL_C;
                        type = MSG_TYPE_WARNING;
                    }
                    if ((our_errcodes.at(MARMOT_ERR_TYPE_OPTIONAL_C))&&(F_INTERFACE == INTERFACE_MODE))
                    {
                        stream
                            << "ERROR: "
                            << getCallname() 
                            << ": oldtype is optional C-Type!" 
                            << std::endl;
                        rvalue = MARMOT_ERR_TYPE_OPTIONAL_C;
                    }
                    break;
                case (DATATYPE_OPTIONAL_FORTRAN):
                    if ((our_errcodes.at(MARMOT_ERR_TYPE_OPTIONAL_F))&&(C_INTERFACE == INTERFACE_MODE))
                    {
                        stream
                            << "ERROR: "
                            << getCallname() 
                            << ": oldtype is optional Fortran-type!" 
                            << std::endl;
                        rvalue = MARMOT_ERR_TYPE_OPTIONAL_F;
                    }
                    if ((our_errcodes.at(MARMOT_WARN_TYPE_OPTIONAL_F))&&(MSG_ERR_WARN <= DEBUG_MODE) && 
                        (F_INTERFACE == INTERFACE_MODE))
                    {
                        stream
                            << "WARNING: "
                            << getCallname() 
                            << ": oldtype is only optional!" 
                            << std::endl;
                        rvalue = MARMOT_WARN_TYPE_OPTIONAL_F;
                        type = MSG_TYPE_WARNING;
                    }
                    break;
                case (DATATYPE_CONSTRUCT_DERIVED_TYPE):
                    // Do nothing, because everything is fine.
                    break;
                default:
		    if(our_errcodes.at(MARMOT_ERR_MARMOT))
		    {
                    	stream
                        << "Check mapping of datatypes."
                        << std::endl
                        << "Check implementation of switch statement."
                        << std::endl;
                    	rvalue = MARMOT_ERR_MARMOT;
		    }
                    break;
            }                   // end switch
            
            if (rvalue != MARMOT_SUCCESS)
            {
            	our_datatype_traces->findAndPrintTraceNamed(my_oldtype,stream,"oldtype");
            	
                MPO_String text;
                text.assign(&stream);      
                MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
                our_log->getMsg(msg);   
            }
        }                       // end else if (2)

        return rvalue;
    }                           // end checkDatatype

 
}                               // end of namespace MPO
