/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definitions of MPO_State and MPO_Statecollection.
 *
 *  This file contains all definitions of classes that take care of 
 *  state management of the clients.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller
 * 
 *  $Id: mpo-state.cc 838 2008-06-13 10:44:34Z hpchimml $
 */


#include <assert.h>
#include <iostream>


#include "mpo-state.h"
#include "mpo-binaryIO.h"


namespace MPO
{

    //***********************************************************************
    //
    //                          MPO_State
    //   
    /** This class represents the state of an MPI process. */

    // proper constructor
    //========================================================================
    MPO_State::MPO_State(MPO_STATE_ID state)
        : my_id(state)
    {
    }


    // destructor
    //========================================================================
    MPO_State::~MPO_State()
    {
    }


    // getId
    //========================================================================
    int MPO_State::getId() const
    {
        return my_id;
    }


    // getFamilyId
    //========================================================================
    int MPO_State::getFamilyId() const
    {
        return MPO_STATEOBJECT_ID;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_State::debugout(std::ostream & output) const
    {
        output << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_State::save(std::ostream & out) const
    {
        out < my_id;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_State::restore(std::istream & in)
    {
        in > my_id;
        assert(my_id > 0);
        assert(my_id < MPO_LAST_STATE_ID);

        return in;
    }


    // operator ==
    //========================================================================
    bool operator==(const MPO_State & lhs, const MPO_State & rhs)
    {
        return lhs.getId() == rhs.getId();
    }


    // operator !=
    //========================================================================
    bool operator!=(const MPO_State & lhs, const MPO_State & rhs)
    {
        return lhs.getId() != rhs.getId();
    }



    //************************************************************************
    //
    //                          CallTimestampPair
    //   

    // default constructor
    //========================================================================
    CallTimestampPair::CallTimestampPair()
        : my_call_p(0)
    {
    }


    // destructor
    //========================================================================
    CallTimestampPair::~CallTimestampPair()
    {
    }



    //***********************************************************************
    //
    //                          MPO_Clientstate
    //   
    /** This class represents the detailed state of a client. 
     *
     *  We don't derive it from MPO_Object, because it is not intended to 
     *  send it between clients and server.
     *  We also forbid to copy and assign these objects by declaring 
     *  the copy constructor and assignment operator private.
     */

    // default constructor
    //========================================================================
    MPO_Clientstate::MPO_Clientstate()
        : my_state(MPO_LAST_STATE_ID)
    {
        my_call_pair.my_call_p = 0;
    }


    // proper constructor
    //========================================================================
    MPO_Clientstate::MPO_Clientstate(const MPO_Clientstate & rhs)
        : my_state(rhs.
                   my_state)
    {
        my_call_pair.my_call_p = rhs.my_call_pair.my_call_p;
        my_call_pair.my_timestamp = rhs.my_call_pair.my_timestamp;
    }


    // destructor
    //========================================================================
    /**
     *  We do not delete the pointer here. 
     *  The responsibility is in the hand of the user!
     *  We only delete the pointer in setCallTimestampPair().
     */
    MPO_Clientstate::~MPO_Clientstate()
    {
        //delete my_call_p;
    }


    // getState
    //========================================================================
    MPO_State MPO_Clientstate::getState()const
    {
        return my_state;
    }


    // getClientname
    //========================================================================
    const MPO_Callobject & MPO_Clientstate::getClientname() const
    {
        return *my_call_pair.my_call_p;
    }


    // getTimestamp
    //========================================================================
    MPO_Timestamp MPO_Clientstate::getTimestamp() const
    {
        return my_call_pair.my_timestamp;
    }


    // setState
    //========================================================================
    void MPO_Clientstate::setState(MPO_State new_state)
    {
        my_state = new_state;
    }


    // setCallTimestampPair
    //========================================================================
    /** We take the full responsibility of the object. 
     *  Every time we receive a new object we delete the old one. 
     */
    void MPO_Clientstate::setCallTimestampPair(CallTimestampPair 
                                                 new_call_pair)
    {
        my_call_pair = new_call_pair;
        //my_call_pair.my_call_p = new_call_p;
        //my_call_pair.my_timestamp = new_timestamp;
    }


    // operator =
    //========================================================================
    const MPO_Clientstate &
    MPO_Clientstate::operator=(const MPO_Clientstate & rhs)
    {
        // check for assignment to self
        if (&rhs == this)
        {
            return *this;
        }

        my_state = rhs.my_state;
        my_call_pair.my_call_p = rhs.my_call_pair.my_call_p;
        my_call_pair.my_timestamp = rhs.my_call_pair.my_timestamp;

        return *this;
    }

}                               // end of namespace MPO
