/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of MPO_Abort class.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Abort.cc 1067 2010-02-08 12:48:48Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Abort::MPO_Abort()
        : my_comm(MPI_COMM_WORLD)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Abort::MPO_Abort(MPO_Comm comm,
                         int errorcode)
        : my_comm(comm),
          my_errorcode(errorcode)
    {
    }


    // destructor
    //========================================================================
    MPO_Abort::~MPO_Abort()
    {
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Abort::checkPrerequisites()
    {

        // check correctness of parameters
        checkComm();
        checkCommWorld();


        our_comm_traces->findAndAddUsage (my_comm,getSourceLocation());


        return 0;
    }


    // checkComm
    //========================================================================
    int MPO_Abort::checkComm()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if communicator is MPI_COMM_NULL.
        if ((our_errcodes.at(MARMOT_ERR_COMM_NULL))&&(MPI_COMM_NULL == my_comm) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Communicator is MPI_COMM_NULL!"
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NULL;

            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        // (2) Check if communicator is valid.
        else if ((our_errcodes.at(MARMOT_ERR_COMM_NOT_VALID))&&(our_comms.end() == our_comms.find(my_comm)) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << " : Communicator is not valid!"
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NOT_VALID;

            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }                           // end checkComm


    // checkCommWorld
    //========================================================================
    int MPO_Abort::checkCommWorld()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        // Check if communicator is MPI_COMM_WORLD.
        if ((our_errcodes.at(MARMOT_WARN_COMM_NOT_WORLD))&&(my_comm != our_mpi_comm_world) &&
            (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "WARNING: "
                << getCallname()
                << ": Communicator is not MPI_COMM_WORLD,"
                << " but MPI_Implementation may use MCW here!"
                << std::endl;
            rvalue = MARMOT_WARN_COMM_NOT_WORLD;

            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }                           // end checkCommWorld


    // getId
    //========================================================================
    int MPO_Abort::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_ABORT_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Abort::checkPost(void)
    {

        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Abort::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Abort(my_comm, my_errorcode);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Abort::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Abort";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_ABORT";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Abort::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(comm = "
            << my_comm
            << ", errorcode = "
            << my_errorcode;

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Abort::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_comm;
        out < my_errorcode;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Abort::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_comm;
        in > my_errorcode;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/env/abort.c.
/*@
  MPI_Abort - Terminates MPI execution environment

  Input Parameters:
  + comm - communicator of tasks to abort
  - errorcode - error code to return to invoking environment

  Notes:
  Terminates all MPI processes associated with the communicator 'comm'; in
  most systems (all to date), terminates `all` processes.

  .N fortran

  int MPI_Abort(
  MPI_Comm comm,
  int errorcode )
  @ */

