/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of MPO_Allgather class.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Allgather.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Allgather::MPO_Allgather()
        : MPO_Collective(MPI_COMM_NULL),
          my_sendbuf(0),
          my_sendcount(-1),
          my_sendtype(MPI_DATATYPE_NULL),
          my_recvbuf(0),
          my_recvcount(-1),
          my_recvtype(MPI_DATATYPE_NULL)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Allgather::MPO_Allgather(void *sendbuf,
                                 int sendcount,
                                 MPO_Datatype sendtype,
                                 void *recvbuf,
                                 int recvcount,
                                 MPO_Datatype recvtype,
                                 MPO_Comm comm)
        : MPO_Collective(comm),
          my_sendbuf(sendbuf),
          my_sendcount(sendcount),
          my_sendtype(sendtype),
          my_recvbuf(recvbuf),
          my_recvcount(recvcount),
          my_recvtype(recvtype)
    {
    }


    // destructor
    //========================================================================
    MPO_Allgather::~MPO_Allgather()
    {
    }


    // checkPrerequisites2
    //========================================================================
    int MPO_Allgather::checkPrerequisites2()
    {
        // check correctness of parameters
        checkCount(my_sendcount, MSG_SENDCOUNT);
        checkCount(my_recvcount, MSG_RECVCOUNT);
        checkDatatypeSendRecv(my_sendcount, my_sendtype,
                              IS_SENDCOUNT, IS_NOT_REDUCE_ALLREDUCE);
        checkDatatypeSendRecv(my_recvcount, my_recvtype,
                              IS_RECVCOUNT, IS_NOT_REDUCE_ALLREDUCE);

        return 0;
    }


    // getId
    //========================================================================
    int MPO_Allgather::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_ALLGATHER_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Allgather::checkPost(void)
    {

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Allgather::checkPrerequisites(void)
    {


        MPO_Collective::checkPrerequisites();
        checkBufferInUse(my_sendbuf,my_sendcount,my_sendtype);
        checkBufferInUse(my_recvbuf,my_recvcount,my_recvtype);

        our_datatype_traces->findAndAddUsage(my_sendtype,getSourceLocation());
        our_datatype_traces->findAndAddUsage(my_sendtype,getSourceLocation());
        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Allgather::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Allgather(my_sendbuf, my_sendcount, my_sendtype,
                              my_recvbuf,
                              my_recvcount, my_recvtype, my_comm);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Allgather::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Allgather";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_ALLGATHER";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Allgather::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(*sendbuf, "
            << "sendcount = "
            << my_sendcount
            << ", sendtype = "
            << my_sendtype
            << ", *recvbuf, "
            << "recvcount = "
            << my_recvcount
            << ", recvtype = "
            << my_recvtype
            << ", comm = "
            << my_comm;

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Allgather::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_sendbuf;
        out < my_sendcount;
        out < my_sendtype;
        // out < my_recvbuf;
        out < my_recvcount;
        out < my_recvtype;
        out < my_comm;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Allgather::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_sendbuf;
        in > my_sendcount;
        in > my_sendtype;
        // in > my_recvbuf;
        in > my_recvcount;
        in > my_recvtype;
        in > my_comm;

        return in;
    }

}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/coll/allgather.c.
/*@

MPI_Allgather - Gathers data from all tasks and distribute it to all

Input Parameters:
+ sendbuf - starting address of send buffer (choice)
. sendcount - number of elements in send buffer (integer)
. sendtype - data type of send buffer elements (handle)
. recvcount - number of elements received from any process (integer)
. recvtype - data type of receive buffer elements (handle)
- comm - communicator (handle)

Output Parameter:
. recvbuf - address of receive buffer (choice)

Notes:
The MPI standard (1.0 and 1.1) says that

The jth block of data sent from
each proess is received by every process and placed in the jth block of the
buffer 'recvbuf'.

This is misleading; a better description is

The block of data sent from the jth process is received by every
process and placed in the jth block of the buffer 'recvbuf'.

This text was suggested by Rajeev Thakur.

.N fortran

.N Errors
.N MPI_ERR_COMM
.N MPI_ERR_COUNT
.N MPI_ERR_TYPE
.N MPI_ERR_BUFFER

int MPI_Allgather (
void *sendbuf,
int sendcount,
MPI_Datatype sendtype,
void *recvbuf,
int recvcount,
MPI_Datatype recvtype,
MPI_Comm comm )
@ */

