/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of MPO_Allgatherv class.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Allgatherv.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Allgatherv::MPO_Allgatherv()
        : MPO_Collective(MPI_COMM_NULL),
          my_sendbuf(0),
          my_sendcount(-1),
          my_sendtype(MPI_DATATYPE_NULL),
          my_recvbuf(0),
          my_recvcounts(0),
          my_displs(0),
          my_recvtype(MPI_DATATYPE_NULL)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Allgatherv::MPO_Allgatherv(void *sendbuf,
                                   int sendcount,
                                   MPO_Datatype sendtype,
                                   void *recvbuf,
                                   int *recvcounts,
                                   int *displs,
                                   MPO_Datatype recvtype,
                                   MPO_Comm comm)
        : MPO_Collective(comm),
          my_sendbuf(sendbuf),
          my_sendcount(sendcount),
          my_sendtype(sendtype),
          my_recvbuf(recvbuf),
          my_recvcounts(recvcounts),
          my_displs(displs),
          my_recvtype(recvtype)
    {
    }


    // destructor
    //========================================================================
    MPO_Allgatherv::~MPO_Allgatherv()
    {
    }


    // checkPrerequisites2
    //========================================================================
    int MPO_Allgatherv::checkPrerequisites2()
    {
        // check correctness of parameters
        checkCountArray(my_recvcounts, MSG_RECVCOUNT);
        checkDisplacementArrayOfReceiver();
        checkDatatypeOfReceiver();
        checkCount(my_sendcount, MSG_SENDCOUNT);
        checkDatatypeSendRecv(my_sendcount, my_sendtype,
                              IS_SENDCOUNT, IS_NOT_REDUCE_ALLREDUCE);

        //check buffers
        checkBufferInUse(my_sendbuf,my_sendcount,my_sendtype);
        int size;
        PMPI_Comm_size(my_comm,&size);
        for (int i = 0;i < size;i++)
            checkBufferInUse(my_recvbuf,my_displs[i]+my_recvcounts[i],my_recvtype);

        return 0;
    }


    // checkDisplacementArrayOfReceiver
    //========================================================================
    int MPO_Allgatherv::checkDisplacementArrayOfReceiver()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

    if(!(our_errcodes.at(MARMOT_ERR_DISPLACEMENT_OVERRIDES_DATA)||our_errcodes.at(MARMOT_NOTE_DISPLACEMENT_OVERRIDES_DATA)))
        return rvalue;

        int size = -1;
        PMPI_Comm_size(my_comm, &size);

        int rank = -1;
        PMPI_Comm_rank(my_comm, &rank);

        // compilation error:  MPI_Aint typeextent may be int or long, so
        // use instead of
        //     int used = 0;
        // an MPI_Aint and cast typesize to MPI_Aint in the if condition
        // below.
        MPI_Aint used = 0;

        MPI_Aint typeextent = 0;
        PMPI_Type_extent(my_recvtype, &typeextent);

        int typesize = -1;
        PMPI_Type_size(my_recvtype, &typesize);

        // (1) Here we have a compact datatype without holes
        // so overriding are errors.
        if (typeextent == (MPI_Aint) typesize)
        {
            //test all pairs of intervals
            for (int i = 0; i < size;i++)
            {
                for (int j = i+1; j < size;j++)
                {
                    //some variables to describe the intervals
                    // (s_i - start of interval i, e_i - end of interval i, ...)
                    int s_i = my_displs[i],
                        s_j = my_displs[j],
                        e_i = s_i +  my_recvcounts[i],
                        e_j = s_j +  my_recvcounts[j];

                    //are the two intervals intersecting ?
                    if ((our_errcodes.at(MARMOT_ERR_DISPLACEMENT_OVERRIDES_DATA))&& (e_i > s_j)             &&
                         (s_i < e_j)             &&
                         (MSG_ERR <= DEBUG_MODE)    )
                    {
                        //yes they intersect each other - give an error
                        std::stringstream stream;
                        stream
                            << "ERROR: "
                            << getCallname()
                            << ": On rank "
                            << rank
                            << ": "
                            << "recieve buffer location writen more than once."
                            << "(Conflict: displs["
                            << i
                            << "] = "
                            << my_displs[i]
                            << ", recvcounts["
                            << i
                            << "] = "
                            << my_recvcounts[i]
                            << ", displs["
                            << j
                            << "] = "
                            << my_displs[j]
                            << ", recvcounts["
                            << j
                            << "] = "
                            << my_recvcounts[j]
                            << ")"
                            << std::endl;

                        rvalue = MARMOT_ERR_DISPLACEMENT_OVERRIDES_DATA;

                        our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                        our_datatype_traces->findAndPrintTraceNamed(my_sendtype,stream,"sendtype");
                        our_datatype_traces->findAndPrintTraceNamed(my_recvtype,stream,"recvtype");

                        MPO_String text;
                        text.assign(&stream);
                        MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                        our_log->getMsg(msg);
                    }//if intersection
                }//for j
            }//for i
        }

        // (2) Here we have a datatype with holes
        // so we don't check whether it is correct.
        else
        {
            // Check array of displacements.
            for (int i = 0; i < size; i++)
            {
                if ((our_errcodes.at(MARMOT_NOTE_DISPLACEMENT_OVERRIDES_DATA))&&(my_displs[i] < used) &&
                    (MSG_ERR_WARN_NOTE <= DEBUG_MODE))
                {
                    std::stringstream stream;
                    stream
                        << "NOTE: "
                        << getCallname()
                        << ": Datatype has holes, "
                        << "so we don't check whether it's correct,"
                        << "but on rank "
                        << rank
                        << ": displs["
                        << i
                        << "] overrides data / holes from other nodes!"
                        << std::endl;
                    rvalue = MARMOT_NOTE_DISPLACEMENT_OVERRIDES_DATA;

                    our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                    our_datatype_traces->findAndPrintTraceNamed(my_sendtype,stream,"sendtype");
                    our_datatype_traces->findAndPrintTraceNamed(my_recvtype,stream,"recvtype");

                    MPO_String text;
                    text.assign(&stream);
                    MPO_MsgObject msg (MSG_TYPE_NOTE,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                    our_log->getMsg(msg);
                }
                used = my_displs[i] + (typeextent * my_recvcounts[i]);
            }
        }

        return rvalue;
    }                           // end checkDisplacementArrayOfReceiver


    // checkDatatypeOfReceiver
    //========================================================================
    int MPO_Allgatherv::checkDatatypeOfReceiver()
    {

        /** \todo @creator of this method: what about return values ? */

        if (my_recvtype == MPI_DATATYPE_NULL)
        {
            if (MSG_ERR <= DEBUG_MODE)
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname()
                    << ": Recvtype is MPI_DATATYPE_NULL!"
                    << std::endl;

                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                our_datatype_traces->findAndPrintTraceNamed(my_sendtype,stream,"sendtype");
                our_datatype_traces->findAndPrintTraceNamed(my_recvtype,stream,"recvtype");

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_NULL,getSourceLocation());
                our_log->getMsg(msg);
            }
        }
        else
        {
            if (our_datatypes.find(my_recvtype) == our_datatypes.end())
            {
                if (MSG_ERR <= DEBUG_MODE)
                {
                    std::stringstream stream;
                    stream
                        << "ERROR: "
                        << getCallname()
                        << ": Recvtype is not valid!"
                        << std::endl;

                    our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                    our_datatype_traces->findAndPrintTraceNamed(my_sendtype,stream,"sendtype");
                    our_datatype_traces->findAndPrintTraceNamed(my_recvtype,stream,"recvtype");

                    MPO_String text;
                    text.assign(&stream);
                    MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_NOT_VALID,getSourceLocation());
                    our_log->getMsg(msg);
                }
            }
            else if ((our_datatypes.find(my_recvtype)->second ==
                      DATATYPE_CREATED_NOT_COMMITTED) &&
                     (MSG_ERR <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname()
                    << ": Recvtype is not committed!"
                    << std::endl;

                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                our_datatype_traces->findAndPrintTraceNamed(my_sendtype,stream,"sendtype");
                our_datatype_traces->findAndPrintTraceNamed(my_recvtype,stream,"recvtype");

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_NOT_COMMITTED,getSourceLocation());
                our_log->getMsg(msg);
            }
        }
        return 1;
    }                           // end checkDatatypeOfReceiver


    // getId
    //========================================================================
    int MPO_Allgatherv::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_ALLGATHERV_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Allgatherv::checkPost(void)
    {

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Allgatherv::checkPrerequisites(void)
    {


        MPO_Collective::checkPrerequisites();
        checkPrerequisites2();

        our_datatype_traces->findAndAddUsage(my_sendtype,getSourceLocation());
        our_datatype_traces->findAndAddUsage(my_recvtype,getSourceLocation());
        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Allgatherv::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue =PMPI_Allgatherv(my_sendbuf, my_sendcount, my_sendtype,
                               my_recvbuf, my_recvcounts, my_displs,
                               my_recvtype, my_comm);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Allgatherv::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Allgatherv";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_ALLGATHERV";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Allgatherv::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(*sendbuf, "
            << "sendcount = "
            << my_sendcount
            << ", sendtype = "
            << my_sendtype
            << ", *recvbuf, "
            << "*recvcounts,"
            << " *displs, "
            << "recvtype = "
            << my_recvtype
            << ", comm = "
            << my_comm;

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Allgatherv::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_sendbuf;
        out < my_sendcount;
        out < my_sendtype;
        // out < my_recvbuf;
        // out < my_recvcounts;
        // out < my_displs;
        out < my_recvtype;
        out < my_comm;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Allgatherv::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_sendbuf;
        in > my_sendcount;
        in > my_sendtype;
        // in > my_recvbuf;
        // in > my_recvcounts;
        // in > my_displs;
        in > my_recvtype;
        in > my_comm;

        return in;
    }

}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/coll/allgatherv.c.
/*@

MPI_Allgatherv - Gathers data from all tasks and deliver it to all

Input Parameters:
+ sendbuf - starting address of send buffer (choice)
. sendcount - number of elements in send buffer (integer)
. sendtype - data type of send buffer elements (handle)
. recvcounts - integer array (of length group size)
containing the number of elements that are received from each process
. displs - integer array (of length group size). Entry
'i'  specifies the displacement (relative to recvbuf ) at
which to place the incoming data from process  'i'
. recvtype - data type of receive buffer elements (handle)
- comm - communicator (handle)

Output Parameter:
. recvbuf - address of receive buffer (choice)

Notes:
The MPI standard (1.0 and 1.1) says that

The jth block of data sent from
each proess is received by every process and placed in the jth block of the
buffer 'recvbuf'.

This is misleading; a better description is

The block of data sent from the jth process is received by every
process and placed in the jth block of the buffer 'recvbuf'.

This text was suggested by Rajeev Thakur.

.N fortran

.N Errors
.N MPI_ERR_BUFFER
.N MPI_ERR_COUNT
.N MPI_ERR_TYPE

int MPI_Allgatherv (
void *sendbuf,
int sendcount,
MPI_Datatype sendtype,

void *recvbuf,
int *recvcounts,
int *displs,
MPI_Datatype recvtype,
MPI_Comm comm )
@ */

