/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of MPO_Alltoallw class.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Alltoallw.cc 909 2008-12-17 09:48:19Z tobias $
 */


#include "mpo-objects.h"
#include "mpo-2-extcoll.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"

#ifdef IS_MPI_ALLTOALLW_PRESENT

namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Alltoallw::MPO_Alltoallw()
        : MPO_Collective(MPI_COMM_NULL),
          my_sendbuf(0),
          my_sendcounts(0),
          my_sdispls(0),
          my_sendtypes(0),
          my_recvbuf(0),
          my_recvcounts(0),
          my_rdispls(0),
          my_recvtypes(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Alltoallw::MPO_Alltoallw(void *sendbuf,
                                 int *sendcounts,
                                 int *sdispls,
                                 MPI_Datatype* sendtypes,
                                 void *recvbuf,
                                 int *recvcounts,
                                 int *rdispls,
                                 MPI_Datatype* recvtypes,
                                 MPO_Comm comm)
        : MPO_Collective(comm),
          my_sendbuf(sendbuf),
          my_sendcounts(sendcounts),
          my_sdispls(sdispls),
          my_sendtypes(sendtypes),
          my_recvbuf(recvbuf),
          my_recvcounts(recvcounts),
          my_rdispls(rdispls),
          my_recvtypes(recvtypes)
    {
    }


    // destructor
    //========================================================================
    MPO_Alltoallw::~MPO_Alltoallw()
    {
    }


    // checkPrerequisites2
    //========================================================================
    int MPO_Alltoallw::checkPrerequisites2(int *counts,
                                           MPI_Datatype* datatype,
                                           int type)
    {

        // check correctness of parameters
        checkDisplacementArrayOfReceiver();
        checkCountArray(my_sendcounts, MSG_SENDCOUNT);
        checkCountArray(my_recvcounts, MSG_RECVCOUNT);
        checkDatatypeOfSender(counts,
                              datatype,
                              type);
        
        int size;
        PMPI_Comm_size(my_comm,&size);
        for (int i = 0;i < size;i++)
        {

            //check buffers
            checkBufferInUse(my_sendbuf,my_sdispls[i]+my_sendcounts[i],my_sendtypes[i]);
            checkBufferInUse(my_recvbuf,my_rdispls[i]+my_recvcounts[i],my_sendtypes[i]);
        }
        return 0;
    }


    // checkDisplacementArrayOfReceiver
    //========================================================================
    int MPO_Alltoallw::checkDisplacementArrayOfReceiver()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        int size = -1;
        PMPI_Comm_size(my_comm, &size);

        int rank = -1;
        PMPI_Comm_rank(my_comm, &rank);

        for ( int k = 0; k < size; k++)
        {
            // compilation error:  MPI_Aint typeextent may be int or long, so
            // use instead of
            //     int used = 0;
            // an MPI_Aint and cast typesize to MPI_Aint in the if condition
            // below.
            MPI_Aint used = 0;

            MPI_Aint typeextent = 0;
            PMPI_Type_extent(my_recvtypes[k], &typeextent);

            int typesize = -1;
            PMPI_Type_size(my_recvtypes[k], &typesize);

            // (1) Here we have a compact datatype without holes
            // so overriding are errors.
            if (typeextent == (MPI_Aint) typesize)
            {
                //test all pairs of intervals
                for (int i = 0; i < size;i++)
                {
                    for (int j = i+1; j < size;j++)
                    {
                        //some variables to describe the intervals
                        // (s_i - start of interval i, e_i - end of interval i, ...)
                        int s_i = my_rdispls[i],
                            s_j = my_rdispls[j],
                            e_i = s_i +  my_recvcounts[i],
                            e_j = s_j +  my_recvcounts[j];

                        //are the two intervals intersecting ?
                        if ( (e_i > s_j)             &&
                             (s_i < e_j)             &&
                             (MSG_ERR <= DEBUG_MODE)    )
                        {
                            //yes they intersect each other - give an error
                            std::stringstream stream;
                            stream
                                << "ERROR: "
                                << getCallname()
                                << ": On rank "
                                << rank
                                << ": "
                                << "recieve buffer location writen more than once."
                                << "(Conflict: rdispls["
                                << i
                                << "] = "
                                << my_rdispls[i]
                                << ", recvcounts["
                                << i
                                << "] = "
                                << my_recvcounts[i]
                                << ", rdispls["
                                << j
                                << "] = "
                                << my_rdispls[j]
                                << ", recvcounts["
                                << j
                                << "] = "
                                << my_recvcounts[j]
                                << ")"
                                << std::endl;

                            rvalue = MARMOT_ERR_DISPLACEMENT_OVERRIDES_DATA;

                            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                            our_datatype_traces->findAndPrintTraceNamed(my_sendtypes[j],stream,"sendtype");
                            our_datatype_traces->findAndPrintTraceNamed(my_recvtypes[j],stream,"recvtype");

                            MPO_String text;
                            text.assign(&stream);
                            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                            our_log->getMsg(msg);
                        }//if intersection
                    }//for j
                }//for i
            }
            // (2) Here we have a datatype with holes
            // so we don't check whether it is correct.
            else
            {
                // Check array of displacements.
                for (int i = 0; i < size; i++)
                {
                    if ((my_rdispls[i] < used) &&
                        (MSG_ERR_WARN_NOTE <= DEBUG_MODE))
                    {
                        std::stringstream stream;
                        stream
                            << "REMARK: "
                            << getCallname()
                            << ": Datatype has holes, "
                            << "so we don't check whether it's correct,"
                            << "but on rank "
                            << rank
                            << ": displs["
                            << i
                            << "] reads data/ holes read from other nodes!"
                            << std::endl;
                        rvalue = MARMOT_NOTE_DISPLACEMENT_READS_DATA;

                        our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                        our_datatype_traces->findAndPrintTraceNamed(my_sendtypes[i],stream,"sendtype");
                        our_datatype_traces->findAndPrintTraceNamed(my_recvtypes[i],stream,"recvtype");

                        MPO_String text;
                        text.assign(&stream);
                        MPO_MsgObject msg (MSG_TYPE_NOTE,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                        our_log->getMsg(msg);
                    }
                    used = my_rdispls[i] + (typeextent * my_recvcounts[i]);
                }
            }
        }//for k

        return rvalue;
    }                           // end checkDisplacementArray


    // checkDatatypeOfSender
    //========================================================================
    int MPO_Alltoallw::checkDatatypeOfSender(int *counts,
                                             MPI_Datatype* datatype,
                                             int type)
    {
        /** \todo @creator of this method: what about return values, I think there are casses missing when using FORTRAN like see MARMOT_ERR_TYPE_C,...! */

        int size = -1;
        PMPI_Comm_size(my_comm, &size);

        for ( int i = 0; i < size; i++)
        {
            if (datatype[i] == MPI_DATATYPE_NULL)
            {
                if (MSG_ERR <= DEBUG_MODE)
                {
                    if (type == 1)
                    {
                        std::stringstream stream;
                        stream
                            << "ERROR: "
                            << getCallname()
                            << ": Sendtype is MPI_DATATYPE_NULL!"
                            << std::endl;

                        our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                        our_datatype_traces->findAndPrintTraceNamed(my_sendtypes[i],stream,"sendtype");
                        our_datatype_traces->findAndPrintTraceNamed(my_recvtypes[i],stream,"recvtype");

                        MPO_String text;
                        text.assign(&stream);
                        MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_NULL,getSourceLocation());
                        our_log->getMsg(msg);
                    }
                    else
                    {
                        std::stringstream stream;
                        stream
                            << "ERROR: "
                            << getCallname()
                            << ": Recvtype is MPI_DATATYPE_NULL!"
                            << std::endl;

                        our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                        our_datatype_traces->findAndPrintTraceNamed(my_sendtypes[i],stream,"sendtype");
                        our_datatype_traces->findAndPrintTraceNamed(my_recvtypes[i],stream,"recvtype");

                        MPO_String text;
                        text.assign(&stream);
                        MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_NULL,getSourceLocation());
                        our_log->getMsg(msg);
                    }
                }
            }
            else
            {
                if (our_datatypes.find(datatype[i]) == our_datatypes.end())
                {
                    if (MSG_ERR <= DEBUG_MODE)
                    {
                        if (type == 1)
                        {
                            std::stringstream stream;
                            stream
                                << "ERROR: "
                                << getCallname()
                                << ": Sendtype is not valid!"
                                << std::endl;

                            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                            our_datatype_traces->findAndPrintTraceNamed(my_sendtypes[i],stream,"sendtype");
                            our_datatype_traces->findAndPrintTraceNamed(my_recvtypes[i],stream,"recvtype");

                            MPO_String text;
                            text.assign(&stream);
                            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_NOT_VALID,getSourceLocation());
                            our_log->getMsg(msg);
                        }
                        else
                        {
                            std::stringstream stream;
                            stream
                                << "ERROR: "
                                << getCallname()
                                << ": Recvtype is not valid!"
                                << std::endl;

                            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                            our_datatype_traces->findAndPrintTraceNamed(my_sendtypes[i],stream,"sendtype");
                            our_datatype_traces->findAndPrintTraceNamed(my_recvtypes[i],stream,"recvtype");

                            MPO_String text;
                            text.assign(&stream);
                            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_NOT_VALID,getSourceLocation());
                            our_log->getMsg(msg);
                        }
                    }
                }
                else
                {
                    switch (our_datatypes.find(datatype[i])->second)
                    {
                    case (DATATYPE_CREATED_NOT_COMMITTED):
                        if (MSG_ERR <= DEBUG_MODE)
                        {
                            if (type == 1)
                            {
                                std::stringstream stream;
                                stream
                                    << "ERROR: "
                                    << getCallname()
                                    << ": Sendtype is not committed!"
                                    << std::endl;

                                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                                our_datatype_traces->findAndPrintTraceNamed(my_sendtypes[i],stream,"sendtype");
                                our_datatype_traces->findAndPrintTraceNamed(my_recvtypes[i],stream,"recvtype");

                                MPO_String text;
                                text.assign(&stream);
                                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_NOT_COMMITTED,getSourceLocation());
                                our_log->getMsg(msg);
                            }
                            else
                            {
                                std::stringstream stream;
                                stream
                                    << "ERROR: "
                                    << getCallname()
                                    << ": Recvtype is not committed!"
                                    << std::endl;

                                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                                our_datatype_traces->findAndPrintTraceNamed(my_sendtypes[i],stream,"sendtype");
                                our_datatype_traces->findAndPrintTraceNamed(my_recvtypes[i],stream,"recvtype");

                                MPO_String text;
                                text.assign(&stream);
                                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_NOT_COMMITTED,getSourceLocation());
                                our_log->getMsg(msg);
                            }
                        }
                        break;
                    case (DATATYPE_REDUCTION_C):
                        if (MSG_ERR_WARN <= DEBUG_MODE)
                        {
                            if (type == 1)
                            {
                                std::stringstream stream;
                                stream
                                    << "WARNING: "
                                    << getCallname()
                                    << ": Sendtype is for reduction functions!"
                                    << std::endl;

                                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                                our_datatype_traces->findAndPrintTraceNamed(my_sendtypes[i],stream,"sendtype");
                                our_datatype_traces->findAndPrintTraceNamed(my_recvtypes[i],stream,"recvtype");

                                MPO_String text;
                                text.assign(&stream);
                                MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_WARN_TYPE_REDUCTION_C,getSourceLocation());
                                our_log->getMsg(msg);
                            }
                            else
                            {
                                std::stringstream stream;
                                stream
                                    << "ERROR: "
                                    << getCallname()
                                    << ": Recvtype is for reduction functions!"
                                    << std::endl;

                                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                                our_datatype_traces->findAndPrintTraceNamed(my_sendtypes[i],stream,"sendtype");
                                our_datatype_traces->findAndPrintTraceNamed(my_recvtypes[i],stream,"recvtype");

                                MPO_String text;
                                text.assign(&stream);
                                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_WARN_TYPE_REDUCTION_C,getSourceLocation());
                                our_log->getMsg(msg);
                            }
                        }
                        break;
                    case (DATATYPE_OPTIONAL_C):
                        if (MSG_ERR_WARN <= DEBUG_MODE)
                        {
                            if (type == 1)
                            {
                                std::stringstream stream;
                                stream
                                    << "WARNING: "
                                    << getCallname()
                                    << ": Sendtype is optional!"
                                    << std::endl;

                                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                                our_datatype_traces->findAndPrintTraceNamed(my_sendtypes[i],stream,"sendtype");
                                our_datatype_traces->findAndPrintTraceNamed(my_recvtypes[i],stream,"recvtype");

                                MPO_String text;
                                text.assign(&stream);
                                MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_WARN_TYPE_OPTIONAL_C,getSourceLocation());
                                our_log->getMsg(msg);
                            }
                            else
                            {
                                std::stringstream stream;
                                stream
                                    << "WARNING: "
                                    << getCallname()
                                    << ": Recvtype is optional!"
                                    << std::endl;

                                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                                our_datatype_traces->findAndPrintTraceNamed(my_sendtypes[i],stream,"sendtype");
                                our_datatype_traces->findAndPrintTraceNamed(my_recvtypes[i],stream,"recvtype");

                                MPO_String text;
                                text.assign(&stream);
                                MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_WARN_TYPE_OPTIONAL_C,getSourceLocation());
                                our_log->getMsg(msg);
                            }
                        }
                        break;
                    case (DATATYPE_COMMITTED_FORTRAN):
                        if ((MSG_ERR <= DEBUG_MODE) &&
                            (C_INTERFACE == INTERFACE_MODE))
                        {
                            if (type == 1)
                            {
                                std::stringstream stream;
                                stream
                                    << "ERROR: "
                                    << getCallname()
                                    << ": Sendtype is Fortran-type!"
                                    << std::endl;

                                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                                our_datatype_traces->findAndPrintTraceNamed(my_sendtypes[i],stream,"sendtype");
                                our_datatype_traces->findAndPrintTraceNamed(my_recvtypes[i],stream,"recvtype");

                                MPO_String text;
                                text.assign(&stream);
                                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_F,getSourceLocation());
                                our_log->getMsg(msg);
                            }
                            else
                            {
                                std::stringstream stream;
                                stream
                                    << "ERROR: "
                                    << getCallname()
                                    << ": Recvtype is Fortran-type!"
                                    << std::endl;

                                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
                                our_datatype_traces->findAndPrintTraceNamed(my_sendtypes[i],stream,"sendtype");
                                our_datatype_traces->findAndPrintTraceNamed(my_recvtypes[i],stream,"recvtype");

                                MPO_String text;
                                text.assign(&stream);
                                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_F,getSourceLocation());
                                our_log->getMsg(msg);
                            }
                        }
                        break;

                    default:
                        // do nothing
                        break;

                    }
                }
            }
        } // for i

        return 1;
    }


    // getId
    //========================================================================
    int MPO_Alltoallw::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_ALLTOALLW_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Alltoallw::checkPost(void)
    {

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Alltoallw::checkPrerequisites(void)
    {
        int size = -1;
        PMPI_Comm_size(my_comm, &size);

        // ??? what about that ???  MPO_Collective::checkPrerequisites();
        checkPrerequisites2(my_recvcounts, my_recvtypes, 2);
            
        for (int i=0; i<size; i++)
        {
            our_datatype_traces->findAndAddUsage(my_sendtypes[i],getSourceLocation());
            our_datatype_traces->findAndAddUsage(my_recvtypes[i],getSourceLocation());
        }
        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());

        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Alltoallw::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Alltoallw(my_sendbuf, my_sendcounts, my_sdispls,
                              my_sendtypes,
                              my_recvbuf, my_recvcounts, my_rdispls,
                              my_recvtypes,
                              my_comm);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Alltoallw::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Alltoallw";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_ALLTOALLW";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Alltoallw::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(*sendbuf, "
            << " *sendcounts, "
            << "*sdispls, "
            << "sendtype = "
            << my_sendtypes
            << ", *recvbuf, "
            << "*recvcounts, "
            << "*rdispls, "
            << "recvtype = "
            << my_recvtypes
            << ", comm = "
            << my_comm;

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Alltoallw::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_sendbuf
        // out < my_sendcounts;
        // out < my_displs;
        // out < my_sendtypes;
        // out < my_recvbuf
        // out < my_recvcounts;
        // out < my_rdispls;
        // out < my_recvtypes;
        out < my_comm;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Alltoallw::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_sendbuf
        // in > my_sendcounts;
        // in > my_displs;
        // in > my_sendtypes;
        // in > my_recvbuf
        // in > my_recvcounts;
        // in > my_rdispls;
        // in > my_recvtypes;
        in > my_comm;

        return in;
    }

}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/coll/alltoallw.c.
/*@

   MPI_Alltoallw - Generalized all-to-all communication allowing different
   datatypes, counts, and displacements for each partner

   Input Parameters:
+ sendbuf - starting address of send buffer (choice)
. sendcounts - integer array equal to the group size specifying the number of
  elements to send to each processor (integer)
. sdispls - integer array (of length group size). Entry j specifies the
  displacement in bytes (relative to sendbuf) from which to take the outgoing
  data destined for process j
. sendtypes - array of datatypes (of length group size). Entry j specifies the
  type of data to send to process j (handle)
. recvcounts - integer array equal to the group size specifying the number of
   elements that can be received from each processor (integer)
. rdispls - integer array (of length group size). Entry i specifies the
  displacement in bytes (relative to recvbuf) at which to place the incoming
  data from process i
. recvtypes - array of datatypes (of length group size). Entry i specifies
  the type of data received from process i (handle)
- comm - communicator (handle)

 Output Parameter:
. recvbuf - address of receive buffer (choice)

.N ThreadSafe

.N Fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_ARG
.N MPI_ERR_COUNT
.N MPI_ERR_TYPE
@ */

#endif /*IS_MPI_ALLTOALLW_PRESENT*/

/*EOF*/
