/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Bcast.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Bcast.cc 1072 2010-02-09 09:08:23Z hpczink $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Bcast::MPO_Bcast()
        : MPO_Collective(MPI_COMM_NULL),
          my_buf(0),
          my_count(-1),
          my_datatype(MPI_DATATYPE_NULL),
          my_root(MPI_PROC_NULL)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Bcast::MPO_Bcast(void *buf,
                         int count,
                         MPO_Datatype datatype,
                         int root,
                         MPO_Comm comm)
        : MPO_Collective(comm),
          my_buf(buf),
          my_count(count),
          my_datatype(datatype),
          my_root(root)
    {
    }


    // destructor
    //========================================================================
    MPO_Bcast::~MPO_Bcast()
    {
    }


    // checkPrerequisites2
    //========================================================================
    int MPO_Bcast::checkPrerequisites2()
    {
        // check correctness of parameters
        checkCount(my_count, MSG_COUNT);
        checkDatatypeSendRecv(my_count, my_datatype, IS_COUNT, IS_NOT_REDUCE_ALLREDUCE);
        checkRoot(my_root);

        return 0;
    }


    // getId
    //========================================================================
    int MPO_Bcast::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_BCAST_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Bcast::checkPost(void)
    {

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Bcast::checkPrerequisites(void)
    {
        checkPrerequisites2();

        MPO_Collective::checkPrerequisites();
        checkBufferInUse(my_buf,my_count,my_datatype);

        our_datatype_traces->findAndAddUsage(my_datatype,getSourceLocation());
        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Bcast::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Bcast(my_buf, my_count, my_datatype, my_root, my_comm);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Bcast::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Bcast";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_BCAST";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Bcast::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(*buffer, "
            << "count = "
            << my_count
            << ", datatype = "
            << my_datatype
            << ", root = "
            << my_root
            << ", comm = "
            << my_comm;

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Bcast::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_buf;
        out < my_count;
        out < my_datatype;
        out < my_root;
        out < my_comm;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Bcast::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_buf;
        in > my_count;
        in > my_datatype;
        in > my_root;
        in > my_comm;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/coll/bcast.c
/*@

MPI_Bcast - Broadcasts a message from the process with rank "root" to
all other processes of the group.

Input/output Parameters:
+ buffer - starting address of buffer (choice)
. count - number of entries in buffer (integer)
. datatype - data type of buffer (handle)
. root - rank of broadcast root (integer)
- comm - communicator (handle)

Algorithm:
If the underlying device does not take responsibility, this function
uses a tree-like algorithm to broadcast the message to blocks of
processes.  A linear algorithm is then used to broadcast the message
from the first process in a block to all other processes.
'MPIR_BCAST_BLOCK_SIZE' determines the size of blocks.  If this is set
to 1, then this function is equivalent to using a pure tree algorithm.
If it is set to the size of the group or greater, it is a pure linear
algorithm.  The value should be adjusted to determine the most
efficient value on different machines.

.N fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_COUNT
.N MPI_ERR_TYPE
.N MPI_ERR_BUFFER
.N MPI_ERR_ROOT

int MPI_Bcast (
void *buffer,
int count,
MPI_Datatype datatype,
int root,
MPI_Comm comm )
@ */

