/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Bsend.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Bsend.cc 949 2009-04-24 08:57:57Z tobias $
 */


#include "mpo.h"
#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Bsend::MPO_Bsend()
        : MPO_Blocking(-1,
                       MPI_DATATYPE_NULL,
                       MPI_PROC_NULL,
                       MPI_ANY_TAG,
                       MPI_COMM_NULL),
          my_buf(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Bsend::MPO_Bsend(void *buf,
                         int count,
                         MPO_Datatype datatype,
                         int dest,
                         int tag,
                         MPO_Comm comm)
        : MPO_Blocking(count,
                       datatype,
                       dest,
                       tag,
                       comm),
          my_buf(buf)
    {
    }


    // destructor
    //========================================================================
    MPO_Bsend::~MPO_Bsend()
    {
    }


    // getId
    //========================================================================
    int MPO_Bsend::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_BSEND_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Bsend::checkPost(void)
    {

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Bsend::checkPrerequisites(void)
    {
        /*Check wheter a buffer is attached*/
        if ((!buffer_attached) &&
            (MSG_ERR <= DEBUG_MODE) )
        {
            std::stringstream stream;
            stream
            << "ERROR: "
            << getCallname()
            << ": No buffer is attached to MPI!"
            << std::endl
            << "Use MPI_Buffer_attach to specify a buffer."
            << std::endl;
            int rvalue = MARMOT_ERR_BUFFER_NOT_ATTACHED; //TODO use correct errorcode

            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        //TODO check for size of attached buffer !!!

        MPO_Blocking::checkPrerequisites();
        checkBufferInUse(my_buf,my_count,my_datatype);

        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());
        our_datatype_traces->findAndAddUsage(my_datatype,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Bsend::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Bsend(my_buf, my_count, my_datatype, my_rank, my_tag,
                          my_comm);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Bsend::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Bsend";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_BSEND";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Bsend::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(*buf, "
            << "count = "
            << my_count
            << ", datatype = "
            << my_datatype
            << ", dest = "
            << my_rank
            << ", tag = "
            << my_tag
            << ", comm = "
            << my_comm;

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Bsend::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_buf;
        out < my_count;
        out < my_datatype;
        out < my_rank;
        out < my_tag;
        out < my_comm;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Bsend::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_buf;
        in > my_count;
        in > my_datatype;
        in > my_rank;
        in > my_tag;
        in > my_comm;

        return in;
    }

}                               // end of namespace MPO


// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/pt2pt/bsend.c.
/*@
  MPI_Bsend - Basic send with user-specified buffering

  Input Parameters:
  + buf - initial address of send buffer (choice)
  . count - number of elements in send buffer (nonnegative integer)
  . datatype - datatype of each send buffer element (handle)
  . dest - rank of destination (integer)
  . tag - message tag (integer)
  - comm - communicator (handle)

  Notes:
  This send is provided as a convenience function; it allows the user to
  send messages without worring about where they are buffered (because the
  user `must` have provided buffer space with 'MPI_Buffer_attach').

  In deciding how much buffer space to allocate, remember that the buffer
  space is not available for reuse by subsequent 'MPI_Bsend's unless you
  are certain that the message has been received (not just that it should
  have been received).  For example,
  this code does not allocate enough buffer space
  .vb
  MPI_Buffer_attach( b, n*sizeof(double) + MPI_BSEND_OVERHEAD );
  for (i=0; i<m; i++) {
  MPI_Bsend( buf, n, MPI_DOUBLE, ... );
  }
  .ve
  because only enough buffer space is provided for a single send, and the
  loop may start a second 'MPI_Bsend' before the first is done making use
  of the buffer.

  In C, you can
  force the messages to be delivered by
  .vb
  MPI_Buffer_detach( &b, &n );
  MPI_Buffer_attach( b, n );
  .ve
  (The 'MPI_Buffer_detach' will not complete until all buffered messages
  are delivered.)

  .N fortran

  .N Errors
  .N MPI_SUCCESS
  .N MPI_ERR_COMM
  .N MPI_ERR_COUNT
  .N MPI_ERR_TYPE
  .N MPI_ERR_RANK
  .N MPI_ERR_TAG

  .seealso: MPI_Buffer_attach, MPI_Ibsend, MPI_Bsend_init

  int MPI_Bsend(
  void *buf,
  int count,
  MPI_Datatype datatype,
  int dest,
  int tag,
  MPI_Comm comm )
  @ */

