/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Buffer_detach.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Buffer_detach.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Buffer_detach::MPO_Buffer_detach()
        : my_buf(0),
          my_size(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Buffer_detach::MPO_Buffer_detach(void *buf,
                                         int *size)
        : my_buf(buf),
          my_size(size)
    {
    }


    // destructor
    //========================================================================
    MPO_Buffer_detach::~MPO_Buffer_detach()
    {
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Buffer_detach::checkPrerequisites()
    {


        // check correctness of parameters
        checkBufferAttached();



        return 0;
    }


    // checkBufferAttached
    //========================================================================
    int MPO_Buffer_detach::checkBufferAttached()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;
    if(!our_errcodes.at(MARMOT_ERR_BUFFER_NOT_ATTACHED))
        return rvalue;

        // Check if buffer is attached.
        if ((!buffer_attached) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": There's no buffer attached!"
                << std::endl;
            rvalue = MARMOT_ERR_BUFFER_NOT_ATTACHED;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        return 1;
    }                           // end checkBufferAttached


    // getId
    //========================================================================
    int MPO_Buffer_detach::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_BUFFER_DETACH_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Buffer_detach::checkPost(void)
    {

        buffer_attached = 0;

        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Buffer_detach::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Buffer_detach(my_buf, my_size);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Buffer_detach::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Buffer_detach";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_BUFFER_DETACH";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Buffer_detach::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(*buffer_adr, "
            << "*size";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Buffer_detach::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_buf;
        // out < my_size;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Buffer_detach::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_buf;
        // in > my_size;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/pt2pt/buffree.c.
/*@
  MPI_Buffer_detach - Removes an existing buffer (for use in MPI_Bsend etc)

  Output Parameters:
  + buffer - initial buffer address (choice)
  - size - buffer size, in bytes (integer)

  Notes:
  The reason that 'MPI_Buffer_detach' returns the address and size of the
  buffer being detached is to allow nested libraries to replace and restore
  the buffer.  For example, consider

  .vb
  int size, mysize, idummy;
  void *ptr, *myptr, *dummy;
  MPI_Buffer_detach( &ptr, &size );
  MPI_Buffer_attach( myptr, mysize );
  ...
  ... library code ...
  ...
  MPI_Buffer_detach( &dummy, &idummy );
  MPI_Buffer_attach( ptr, size );
  .ve

  This is much like the action of the Unix signal routine and has the same
  strengths (it is simple) and weaknesses (it only works for nested
  usages).

  Note that for this approach to work, MPI_Buffer_detach must return
  MPI_SUCCESS even when there is no buffer to detach.
  In that case, it returns a size of zero.  The MPI 1.1 standard for
  'MPI_BUFFER_DETACH' contains the text

  .vb
  The statements made in this section describe the behavior of MPI for
  buffered-mode sends. When no buffer is currently associated, MPI behaves
  as if a zero-sized buffer is associated with the process.
  .ve

  This could be read as applying only to the various Bsend routines.  This
  implementation takes the position that this applies to
  'MPI_BUFFER_DETACH' as well.

  .N fortran

  The Fortran binding for this routine is different.  Because Fortran
  does not have pointers, it is impossible to provide a way to use the
  output of this routine to exchange buffers.  In this case, only the
  size field is set.

  Notes for C:
  Even though the 'bufferptr' argument is declared as 'void *', it is
  really the address of a void pointer.  See the rationale in the
  standard for more details.

  int MPI_Buffer_detach(
  void *bufferptr,
  int *size )
  @ */

