/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Cart_map.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Cart_map.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Cart_map::MPO_Cart_map()
        : my_comm(MPI_COMM_NULL),
          my_ndims(-1), 
          my_dims(0), 
          my_periods(0), 
          my_newrank(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Cart_map::MPO_Cart_map(MPO_Comm comm,
                               int ndims,
                               int *dims,
                               int *periods,
                               int *newrank)
        : my_comm(comm),
          my_ndims(ndims), 
          my_dims(dims), 
          my_periods(periods), 
          my_newrank(newrank)
    {
    }


    // destructor
    //========================================================================
    MPO_Cart_map::~MPO_Cart_map()
    {
    }


    /** We overload MPO_Cartesian::checkPrerequisites to check 
     * other pre_requisites
     */
    // checkPrerequisites
    //========================================================================
    int MPO_Cart_map::checkPrerequisites()
    {


        // check correctness of parameters
        checkComm();
        checkTopologyCartNew();
    
        
        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());
        
        
        return 0;
    }

 
    // checkComm
    //========================================================================
    int MPO_Cart_map::checkComm()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if communicator is MPI_COMM_NULL.
        if ((our_errcodes.at(MARMOT_ERR_COMM_NULL))&&(MPI_COMM_NULL == my_comm) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": Communicator is MPI_COMM_NULL!" 
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NULL;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (2) Check if communicator is valid.
        else if ((our_errcodes.at(MARMOT_ERR_COMM_NOT_VALID))&&(our_comms.end() == our_comms.find(my_comm)) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << " : Communicator is not valid!" 
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NOT_VALID;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        return rvalue;
    }                           // end checkComm


    // checkTopologyCartNew
    //========================================================================
    int MPO_Cart_map::checkTopologyCartNew()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;
    if(!(our_errcodes.at(MARMOT_ERR_TOPOLOGY_CART_NEW_GT_COMM_OLD)||our_errcodes.at(MARMOT_WARN_TOPOLOGY_CART_NEW_LT_COMM_OLD)))
        return rvalue;

        int comm_size = -1;
        PMPI_Comm_size(my_comm, &comm_size);

        // Compute total size of cartesian grid
        int i = 0;
        int total_size = 1;

        for (i = 0; i <= my_ndims - 1; i++)
        {
            total_size = total_size * my_dims[i];
        }

        // (1) Check if total size > size of old communicator
        if ((our_errcodes.at(MARMOT_ERR_TOPOLOGY_CART_NEW_GT_COMM_OLD))&&(comm_size < total_size) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream    << "ERROR: "
                      << getCallname()
                      << ": New topology is larger than old communicator:"
                      << std::endl
                      << " comm size = "
                      << comm_size 
                      << ", total size = " 
                      << total_size 
                      << std::endl;
            rvalue = MARMOT_ERR_TOPOLOGY_CART_NEW_GT_COMM_OLD;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (2) Check if total_size < size of old communicator
        else if ((our_errcodes.at(MARMOT_WARN_TOPOLOGY_CART_NEW_LT_COMM_OLD))&&(comm_size > total_size) &&
                 (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "WARNING: "
                << getCallname()
                << ": New topology is smaller than old communicator:"
                << " Some nodes will get MPI_COMM_NULL." 
                << std::endl
                << " comm size = "
                << comm_size 
                << ", total size = " 
                << total_size 
                << std::endl;
            rvalue = MARMOT_WARN_TOPOLOGY_CART_NEW_LT_COMM_OLD;
            
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        return rvalue;
    }                           // end checkTopologyCartNew


    // getId
    //========================================================================
    int MPO_Cart_map::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_CART_MAP_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Cart_map::checkPost(void)
    {
            
        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Cart_map::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Cart_map(my_comm, my_ndims, my_dims, my_periods,
                               my_newrank);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Cart_map::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Cart_map";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_CART_MAP";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Cart_map::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output 
            << getCallname()
            << "(comm = "
            << my_comm
            << ", ndims = "
            << my_ndims 
            << ", *dims, "
            << "*periods, "
            << "*newrank";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"  
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Cart_map::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_comm;
        out < my_ndims;
        // out < my_dims;
        // out < my_periods;
        // out < my_newrank;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Cart_map::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_comm;
        in > my_ndims;
        // in > my_dims;
        // in > my_periods;
        // in > my_newrank;

        return in;
    }

}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/topol/cart_map.c.
/*@

MPI_Cart_map - Maps process to Cartesian topology information 

Input Parameters:
+ comm - input communicator (handle) 
. ndims - number of dimensions of Cartesian structure (integer) 
. dims - integer array of size 'ndims' specifying the number of processes  
in each coordinate direction 
- periods - logical array of size 'ndims' specifying the periodicity 
specification in each coordinate direction 

Output Parameter:
. newrank - reordered rank of the calling process; 'MPI_UNDEFINED' if 
calling process does not belong to grid (integer) 

.N fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_DIMS
.N MPI_ERR_ARG

int MPI_Cart_map ( 
MPI_Comm comm_old,
int ndims,
int *dims,
int *periods,
int *newrank)
@ */


