/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Comm_remote_group.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Comm_remote_group.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Comm_remote_group::
    MPO_Comm_remote_group()
        : MPO_Intercommunicators(MPI_COMM_NULL),
          my_group(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Comm_remote_group::MPO_Comm_remote_group(MPO_Comm communicator,
                                                 MPI_Group *
                                                 group)
        : MPO_Intercommunicators(communicator),
          my_group(group)
    {
    }


    // destructor
    //========================================================================
    MPO_Comm_remote_group::~MPO_Comm_remote_group()
    {
    }


    // getId
    //========================================================================
    int MPO_Comm_remote_group::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_COMM_REMOTE_GROUP_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Comm_remote_group::checkPost(void)
    {

        our_groups.insert(*my_group);

        our_group_traces->addAndSetCreation(*my_group,getSourceLocation());

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Comm_remote_group::checkPrerequisites(void)
    {


        MPO_Intercommunicators::checkPrerequisites();

        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Comm_remote_group::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Comm_remote_group(my_comm, my_group);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Comm_remote_group::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Comm_remote_group";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_COMM_REMOTE_GROUP";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream &
    MPO_Comm_remote_group::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(comm = "
            << my_comm
            << ", *group";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Comm_remote_group::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_comm;
        // out < my_group;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Comm_remote_group::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_comm;
        // in > my_group;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from
// mpich-1.2.4/src/context/comm_rgroup.c.
/*@

MPI_Comm_remote_group - Accesses the remote group associated with
the given inter-communicator

Input Parameter:
. comm - Communicator (must be intercommunicator)

Output Parameter:
. group - remote group of communicator

.N fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM

int MPI_Comm_remote_group (
MPI_Comm comm,
MPI_Group *group )
@ */

