/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Comm_test_inter.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Comm_test_inter.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Comm_test_inter::MPO_Comm_test_inter()
        : my_comm(MPI_COMM_NULL),
          my_flag(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Comm_test_inter::MPO_Comm_test_inter(MPO_Comm communicator,
                                             int *flag)
        : my_comm(communicator),
          my_flag(flag)
    {
    }


    // destructor
    //========================================================================
    MPO_Comm_test_inter::~MPO_Comm_test_inter()
    {
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Comm_test_inter::checkPrerequisites()
    {


        // check correctness of parameters
        checkComm();


        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());


        return 0;
    }


    // checkComm
    //========================================================================
    int MPO_Comm_test_inter::checkComm()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if communicator is MPI_COMM_NULL.
        if ((our_errcodes.at(MARMOT_ERR_COMM_NULL))&&(MPI_COMM_NULL == my_comm) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Communicator is MPI_COMM_NULL!"
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NULL;

            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        // (2) Check if communicator is valid.
        else if ((our_errcodes.at(MARMOT_ERR_COMM_NOT_VALID))&&(our_comms.end() == our_comms.find(my_comm)) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << " : Communicator is not valid!"
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NOT_VALID;

            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }                           // end checkComm


    // getId
    //========================================================================
    int MPO_Comm_test_inter::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_COMM_TEST_INTER_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Comm_test_inter::checkPost(void)
    {

        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Comm_test_inter::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Comm_test_inter(my_comm, my_flag);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Comm_test_inter::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Comm_test_inter";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_COMM_TEST_INTER";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Comm_test_inter::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(comm = "
            << my_comm
            << ", *flag";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Comm_test_inter::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_comm;
        // out < my_flag;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Comm_test_inter::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_comm;
        // in > my_flag;

        return in;
    }



}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from
// mpich-1.2.4/src/context/comm_testic.c.
/*@

MPI_Comm_test_inter - Tests to see if a comm is an inter-communicator

Input Parameter:
. comm - communicator (handle)

Output Parameter:
. flag - (logical)

.N fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_ARG

int MPI_Comm_test_inter (
MPI_Comm comm,
int *flag )
@ */


