/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Dims_create.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Dims_create.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Dims_create::MPO_Dims_create()
        : my_nnodes(-1), 
          my_ndims(-1), 
          my_dims(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Dims_create::MPO_Dims_create(int nnodes,
                                     int ndims,
                                     int *dims)
        : my_nnodes(nnodes),
          my_ndims(ndims), 
          my_dims(dims)
    {
    }


    // destructor
    //========================================================================
    MPO_Dims_create::~MPO_Dims_create()
    {
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Dims_create::checkPrerequisites()
    {


        // check correctness of parameters
        checkDimsArray();
    
        
        
        return 0;
    }


    // checkDimsArray
    //========================================================================
    int MPO_Dims_create::checkDimsArray()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS; 

        int prod = 1;
        int sum = 0;
        int nonzero = my_ndims;

        // Check array of dims.
        for (int i = 0; i < my_ndims; i++)
        {
            if (my_dims[i] != 0)
            {
                nonzero--;

                if ((our_errcodes.at(MARMOT_ERR_DIM_NEG))&&(my_dims[i] < 0) &&
                    (MSG_ERR <= DEBUG_MODE))
                {
                    std::stringstream stream;
                    stream
                        << "ERROR: "
                        << getCallname() 
                        << ": dim[" 
                        << i 
                        << "] less than 0 !" 
                        << std::endl;
                    rvalue = MARMOT_ERR_DIM_NEG;
            
                    MPO_String text;
                    text.assign(&stream);      
                    MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
                    our_log->getMsg(msg);   
                }
        
                else if ((my_dims[i] >= 0) &&
                         (MSG_ERR <= DEBUG_MODE))
                {                       
                    // so the entry is valid and will persist
                    prod *= my_dims[i];
                    sum += my_dims[i];

                    if ((our_errcodes.at(MARMOT_ERR_DIM_TOO_BIG))&&(sum > my_nnodes) && 
                        (MSG_ERR <= DEBUG_MODE))
                    {
                        std::stringstream stream;
                        stream
                            << "ERROR: "
                            << getCallname()
                            << ": Number of nodes in dims greater than nnodes!" 
                            << std::endl;
                        rvalue = MARMOT_ERR_DIM_TOO_BIG;
            
                        MPO_String text;
                        text.assign(&stream);      
                        MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
                        our_log->getMsg(msg);   
                    }
                }
            }
        }

        if ((our_errcodes.at(MARMOT_ERR_NODES_NEQ_PRODUCT_DIMS))&&(my_nnodes % prod != 0) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Nnodes="
                << my_nnodes
                << " is not a multiple of the product of entries in dims (="
                << prod 
                << ") !" 
                << std::endl;
            rvalue = MARMOT_ERR_NODES_NEQ_PRODUCT_DIMS;
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        if ((our_errcodes.at(MARMOT_WARN_DIM_ZERO))&&(nonzero == 0) && 
            (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "WARNING: "
                << getCallname() 
                << ": No entries in dims to be set!" 
                << std::endl;
            rvalue = MARMOT_WARN_DIM_ZERO;
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        return rvalue;
    }                           // end checkDimsArray


    // getId
    //========================================================================
    int MPO_Dims_create::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_DIMS_CREATE_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Dims_create::checkPost(void)
    {
    
        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Dims_create::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Dims_create(my_nnodes, my_ndims, my_dims);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Dims_create::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Dims_create";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_DIMS_CREATE";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Dims_create::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output 
            << getCallname()
            << "(nnodes = "
            << my_nnodes 
            << ", ndims = " 
            << my_ndims 
            << ", *dims";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"  
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Dims_create::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_nnodes;
        out < my_ndims;
        // out < my_dims;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Dims_create::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_nnodes;
        in > my_ndims;
        // in > my_dims;

        return in;
    }

}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied 
// from mpich-1.2.4/src/topol/dims_create.c.
/*@

MPI_Dims_create - Creates a division of processors in a cartesian grid

Input Parameters:
+ nnodes - number of nodes in a grid (integer) 
- ndims - number of cartesian dimensions (integer) 

In/Out Parameter:   
. dims - integer array of size  'ndims' specifying the number of nodes in 
each dimension  

.N fortran

int MPI_Dims_create(
int nnodes, 
int ndims, 
int *dims)
@ */

