/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Errhandler_create.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Errhandler_create.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Errhandler_create::MPO_Errhandler_create()
        : my_function(0),
          //bk// because of compiler warning
          //my_errhandler(MPI_ERRHANDLER_NULL)
          my_errhandler(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Errhandler_create::MPO_Errhandler_create(MPI_Handler_function *
                                                 function,
                                                 MPI_Errhandler *
                                                 errhandler)
        : my_function(function), 
          my_errhandler(errhandler)
    {
    }


    // destructor
    //========================================================================
    MPO_Errhandler_create::~MPO_Errhandler_create()
    {
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Errhandler_create::checkPrerequisites()
    {

        //checkPrerequisites();
        
        

        // currently no checks implemented    

        return 0;
    }


    // getId
    //========================================================================
    int MPO_Errhandler_create::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_ERRHANDLER_CREATE_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Errhandler_create::checkPost(void)
    {
        
        int tmp[2] = { 0, 0 };
        our_errhandlers.insert(std::make_pair((*my_errhandler), (int *) tmp));
        
        our_errhandler_traces->addAndSetCreation(*my_errhandler,getSourceLocation());
    
        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Errhandler_create::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Errhandler_create(my_function, my_errhandler);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Errhandler_create::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Errhandler_create";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_ERRHANDLER_CREATE";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & 
    MPO_Errhandler_create::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output 
            << getCallname() 
            << "(*function, "
            << "*errhandler";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"  
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Errhandler_create::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_function;
        // out < my_errhandler;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Errhandler_create::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_function;
        // in > my_errhandler;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/env/errcreate.c.
/*@
  MPI_Errhandler_create - Creates an MPI-style errorhandler

  Input Parameter:
  . function - user defined error handling procedure 

  Output Parameter:
  . errhandler - MPI error handler (handle) 

  Notes:
  The MPI Standard states that an implementation may make the output value 
  (errhandler) simply the address of the function.  However, the action of 
  'MPI_Errhandler_free' makes this impossible, since it is required to set 
  the value of the argument to 'MPI_ERRHANDLER_NULL'.  In addition, the 
  actual error handler must remain until all communicators that use it are 
  freed.

  .N fortran

  .N Errors
  .N MPI_SUCCESS
  .N MPI_ERR_EXHAUSTED

  int MPI_Errhandler_create( 
  MPI_Handler_function *function,
  MPI_Errhandler       *errhandler)
  @ */

