/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Errhandler_free.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Errhandler_free.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Errhandler_free::MPO_Errhandler_free()
        //bk// because of compiler warning
        //: my_errhandler(MPI_ERRHANDLER_NULL)
        : my_errhandler(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Errhandler_free::MPO_Errhandler_free(MPI_Errhandler *
                                             errhandler)
        : my_errhandler(errhandler)
    {
    }


    // destructor
    //========================================================================
    MPO_Errhandler_free::~MPO_Errhandler_free()
    {
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Errhandler_free::checkPrerequisites()
    {


        // check correctness of parameters
        checkAndEraseErrhandler();
    
        
        our_errhandler_traces->findAndSetFree(*my_errhandler,getSourceLocation());
        
        
        return 0;
    }


    // checkAndEraseErrhandler
    //========================================================================
    int MPO_Errhandler_free::checkAndEraseErrhandler()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if errhandler is valid.
        if ((our_errcodes.at(MARMOT_ERR_ERRHANDLER_NOT_VALID))&&(our_errhandlers.find(*my_errhandler) == our_errhandlers.end()) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": errhandler is not valid!" 
                << std::endl;
            rvalue = MARMOT_ERR_ERRHANDLER_NOT_VALID;
            
            our_errhandler_traces->findAndPrintTraceNamed(*my_errhandler,stream,"errhandler");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        (our_errhandlers.find(*my_errhandler)->second[1]) = 1;

        // (2) Erase errhandler from our map.
        if (our_errhandlers.find(*my_errhandler)->second[0] == 0)
        {
            our_errhandlers.erase(*my_errhandler);
        }

        return rvalue;
    }                           // end checkAndEraseErrhandler


    // getId
    //========================================================================
    int MPO_Errhandler_free::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_ERRHANDLER_FREE_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Errhandler_free::checkPost(void)
    {
            
        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Errhandler_free::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Errhandler_free(my_errhandler);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Errhandler_free::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Errhandler_free";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_ERRHANDLER_FREE";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Errhandler_free::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output 
            << getCallname() 
            << "(*errhandler";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"  
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Errhandler_free::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_errhandler;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Errhandler_free::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_errhandler;

        return in;
    }

}                               // end of namespace MPO




// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/env/errfree.c.
/*@
  MPI_Errhandler_free - Frees an MPI-style errorhandler

  Input Parameter:
  . errhandler - MPI error handler (handle).  Set to 'MPI_ERRHANDLER_NULL' 
  on exit.

  .N fortran

  .N Errors
  .N MPI_SUCCESS
  .N MPI_ERR_ARG

  int MPI_Errhandler_free( MPI_Errhandler *errhandler )
  @ */

