/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Errhandler_set.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Errhandler_set.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"
#include "config.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Errhandler_set::MPO_Errhandler_set()
        : my_comm(MPI_COMM_NULL),
          my_errhandler(MPI_ERRHANDLER_NULL)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Errhandler_set::MPO_Errhandler_set(MPO_Comm comm,
                                           MPI_Errhandler
                                           errhandler)
        : my_comm(comm),
          my_errhandler(errhandler)
    {
    }


    // destructor
    //========================================================================
    MPO_Errhandler_set::~MPO_Errhandler_set()
    {
    }


    // getId
    //========================================================================
    int MPO_Errhandler_set::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_ERRHANDLER_SET_ID;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Errhandler_set::checkPrerequisites()
    {


        // check correctness of parameters
        checkComm();
        checkErrhandler();
    
        
        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());
        our_errhandler_traces->findAndAddUsage(my_errhandler,getSourceLocation());
        
        
        return 0;
    }


    // checkComm
    //========================================================================
    int MPO_Errhandler_set::checkComm()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if communicator is MPI_COMM_NULL.
        if ((our_errcodes.at(MARMOT_ERR_COMM_NULL))&&(MPI_COMM_NULL == my_comm) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": Communicator is MPI_COMM_NULL!" 
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NULL;
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg); 
        }

        // (2) Check if communicator is valid.
        else if ((our_errcodes.at(MARMOT_ERR_COMM_NOT_VALID))&&(our_comms.end() == our_comms.find(my_comm)) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << " : Communicator is not valid!" 
                << std::endl;
            rvalue = MARMOT_ERR_COMM_NOT_VALID;
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg); 
        }

        return rvalue;
    }                           // end checkComm


    // checkErrhandler
    //========================================================================
    int MPO_Errhandler_set::checkErrhandler()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        // Check if errhandler is valid.
        if ((our_errcodes.at(MARMOT_ERR_ERRHANDLER_NOT_VALID))&&(our_errhandlers.find(my_errhandler) == our_errhandlers.end()) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": errhandler is not valid!" 
                << std::endl;
            rvalue = MARMOT_ERR_ERRHANDLER_NOT_VALID;
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg); 
        }

        return rvalue;
    }                           // end checkErrhandler


    // checkPost
    //========================================================================
    int MPO_Errhandler_set::checkPost(void)
    {
        
        (our_errhandlers.find(my_errhandler)->second[0])++;
    
        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Errhandler_set::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Errhandler_set(my_comm, my_errhandler);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Errhandler_set::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Errhandler_set";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_ERRHANDLER_SET";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Errhandler_set::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output 
            << getCallname()
            << "(comm = "
            << my_comm 
            << ", errhandler = " 
            << my_errhandler;

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output 
            << ")" 
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Errhandler_set::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_comm;
#ifndef MPI_ERRHANDLER_NON_BASIC
        out < my_errhandler;
#endif

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Errhandler_set::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_comm;
#ifndef MPI_ERRHANDLER_NON_BASIC
        in > my_errhandler;
#endif

        return in;
    }

}                               // end of namespace MPO


// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/env/errset.c.
/*@
  MPI_Errhandler_set - Sets the error handler for a communicator

  Input Parameters:
  + comm - communicator to set the error handler for (handle) 
  - errhandler - new MPI error handler for communicator (handle) 

  .N fortran

  .N Errors
  .N MPI_SUCCESS
  .N MPI_ERR_COMM
  .N MPI_ERR_ARG

  int MPI_Errhandler_set( 
  MPI_Comm comm, 
  MPI_Errhandler errhandler )
  @ */


