/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Error_class.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Error_class.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Error_class::MPO_Error_class()
        : my_errorcode(MPI_ERR_LASTCODE),
          my_errorclass(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Error_class::MPO_Error_class(int errorcode,
                                     int *errorclass)
        : my_errorcode(errorcode),
          my_errorclass(errorclass)
    {
    }


    // destructor
    //========================================================================
    MPO_Error_class::~MPO_Error_class()
    {
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Error_class::checkPrerequisites()
    {


        // check correctness of parameters
        checkErrorcode();
    
        
        
        return 0;
    }


    // checkErrorcode
    //========================================================================
    int MPO_Error_class::checkErrorcode()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        // Check if errorcode is valid.
        if (((my_errorcode < 0) ||
             (my_errorcode > MPI_ERR_LASTCODE)) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": errorcode is not valid!" 
                << std::endl;
            rvalue = MARMOT_ERR_ERRORCODE_NOT_VALID;
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        return rvalue;
    }                           // end checkErrorcode


    // getId
    //========================================================================
    int MPO_Error_class::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_ERROR_CLASS_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Error_class::checkPost(void)
    {
    
        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Error_class::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Error_class(my_errorcode, my_errorclass);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Error_class::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Error_class";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_ERROR_CLASS";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Error_class::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output 
            << getCallname()
            << "(errorcode = " 
            << my_errorcode 
            << ", *errorclass";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"  
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Error_class::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_errorcode;
        // out < my_errorclass;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Error_class::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_errorcode;
        // in > my_errorclass;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/env/errclass.c.
/*@
  MPI_Error_class - Converts an error code into an error class

  Input Parameter:
  . errorcode - Error code returned by an MPI routine 

  Output Parameter:
  . errorclass - Error class associated with 'errorcode'

  .N fortran

  int MPI_Error_class( 
  int errorcode, 
  int *errorclass)
  @ */

