/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Error_string.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Error_string.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Error_string::MPO_Error_string()
        : my_errorcode(MPI_ERR_LASTCODE),
          my_string(0), 
          my_resultlen(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Error_string::MPO_Error_string(int errorcode,
                                       char *string,
                                       int *resultlen)
        : my_errorcode(errorcode),
          my_string(string), 
          my_resultlen(resultlen)
    {
    }


    // destructor
    //========================================================================
    MPO_Error_string::~MPO_Error_string()
    {
    }


    // getId
    //========================================================================
    int MPO_Error_string::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_ERROR_STRING_ID;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Error_string::checkPrerequisites()
    {


        // check correctness of parameters
        checkErrorcode();
    
        
        
        return 0;
    }


    // checkErrorcode
    //========================================================================
    int MPO_Error_string::checkErrorcode()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        // Check if errorcode is valid.
        if (((my_errorcode < 0) ||
             (my_errorcode > MPI_ERR_LASTCODE)) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": errorcode is not valid!" 
                << std::endl;
            rvalue = MARMOT_ERR_ERRORCODE_NOT_VALID;
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        return rvalue;
    }                           // end checkErrorcode


    // checkPost
    //========================================================================
    int MPO_Error_string::checkPost(void)
    {
    
        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Error_string::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Error_string(my_errorcode, my_string, my_resultlen);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Error_string::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Error_string";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_ERROR_STRING";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Error_string::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output 
            << getCallname()
            << "(errcode = " 
            << my_errorcode 
            << ", *string, "
            << "*resultlen";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"  
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Error_string::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_errorcode;
        // out < my_string;
        // out < my_resultlen;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Error_string::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_errorcode;
        // in > my_string;
        // in > my_resultlen;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/env/errorstring.c.
/*@
  MPI_Error_string - Return a string for a given error code

  Input Parameters:
  . errorcode - Error code returned by an MPI routine or an MPI error class

  Output Parameter:
  + string - Text that corresponds to the errorcode 
  - resultlen - Length of string 

  Notes:  Error codes are the values return by MPI routines (in C) or in 
  the 'ierr' argument (in Fortran).  These can be converted into error 
  classes with the routine 'MPI_Error_class'.  

  .N fortran

  int MPI_Error_string( 
  int errorcode, 
  char *string, 
  int *resultlen )
  @ */

