/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Gather.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Gather.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Gather::MPO_Gather()
        : MPO_Collective(MPI_COMM_NULL),
          my_sendbuf(0),
          my_sendcount(-1),
          my_sendtype(MPI_DATATYPE_NULL),
          my_recvbuf(0),
          my_recvcount(-1),
          my_recvtype(MPI_DATATYPE_NULL),
          my_root(MPI_PROC_NULL)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Gather::MPO_Gather(void *sendbuf,
                           int sendcount,
                           MPO_Datatype sendtype,
                           void *recvbuf,
                           int recvcount,
                           MPO_Datatype recvtype,
                           int root,
                           MPO_Comm comm)
        : MPO_Collective(comm),
          my_sendbuf(sendbuf),
          my_sendcount(sendcount),
          my_sendtype(sendtype),
          my_recvbuf(recvbuf),
          my_recvcount(recvcount),
          my_recvtype(recvtype),
          my_root(root)
    {
    }


    // destructor
    //========================================================================
    MPO_Gather::~MPO_Gather()
    {
    }


    // checkPrerequisites2
    //========================================================================
    int MPO_Gather::checkPrerequisites2()
    {
        // check correctness of parameters
        // some checks are only for root
        checkCount(my_sendcount, MSG_SENDCOUNT);
        checkDatatypeSendRecv(my_sendcount, my_sendtype,
                              IS_SENDCOUNT, IS_NOT_REDUCE_ALLREDUCE);
        checkRoot(my_root);

        int testrank = 0;
        PMPI_Comm_rank(my_comm, &testrank);

        if (testrank == my_root)
        {
            checkCount(my_recvcount, MSG_RECVCOUNT);
            checkDatatypeSendRecv(my_recvcount, my_recvtype,
                                  IS_RECVCOUNT, IS_NOT_REDUCE_ALLREDUCE);
        }

        //check buffers
        checkBufferInUse(my_sendbuf,my_sendcount,my_sendtype);
        if (testrank == my_root)
            checkBufferInUse(my_recvbuf,my_recvcount,my_recvtype);

        return 0;
    }


    // getId
    //========================================================================
    int MPO_Gather::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_GATHER_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Gather::checkPost(void)
    {

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Gather::checkPrerequisites(void)
    {


        MPO_Collective::checkPrerequisites();
        checkPrerequisites2();

        our_datatype_traces->findAndAddUsage(my_sendtype,getSourceLocation());
        our_datatype_traces->findAndAddUsage(my_recvtype,getSourceLocation());
        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Gather::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Gather(my_sendbuf, my_sendcount, my_sendtype, my_recvbuf,
                           my_recvcount, my_recvtype, my_root, my_comm);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Gather::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Gather";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_GATHER";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Gather::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(*sendbuf, "
            << "sendcount = "
            << my_sendcount
            << ", sendtype = "
            << my_sendtype
            << ", *recvbuf, "
            << "recvcount = "
            << my_recvcount
            << ", recvtype = "
            << my_recvtype
            << ", root = "
            << my_root
            << ", comm = "
            << my_comm;

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Gather::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_sendbuf;
        out < my_sendcount;
        out < my_sendtype;
        // out < my_recvbuf
        out < my_recvcount;
        out < my_recvtype;
        out < my_root;
        out < my_comm;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Gather::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_sendbuf;
        in > my_sendcount;
        in > my_sendtype;
        // in > my_recvbuf
        in > my_recvcount;
        in > my_recvtype;
        in > my_root;
        in > my_comm;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/coll/gather.c.
/*@

MPI_Gather - Gathers together values from a group of processes

Input Parameters:
+ sendbuf - starting address of send buffer (choice)
. sendcount - number of elements in send buffer (integer)
. sendtype - data type of send buffer elements (handle)
. recvcount - number of elements for any single receive (integer,
significant only at root)
. recvtype - data type of recv buffer elements
(significant only at root) (handle)
. root - rank of receiving process (integer)
- comm - communicator (handle)

Output Parameter:
. recvbuf - address of receive buffer (choice, significant only at 'root')

.N fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_COUNT
.N MPI_ERR_TYPE
.N MPI_ERR_BUFFER

int MPI_Gather (
void *sendbuf,
int sendcnt,
MPI_Datatype sendtype,
void *recvbuf,
int recvcount,
MPI_Datatype recvtype,
int root,
MPI_Comm comm )
@ */

