/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Get_count.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Get_count.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Get_count::MPO_Get_count()
        : my_status(0),
          my_datatype(MPI_DATATYPE_NULL), 
          my_count(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Get_count::MPO_Get_count(MPI_Status * status,
                                 MPO_Datatype datatype,
                                 int *count)
        : my_status(status),
          my_datatype(datatype), 
          my_count(count)
    {
    }


    // destructor
    //========================================================================
    MPO_Get_count::~MPO_Get_count()
    {
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Get_count::checkPrerequisites()
    {


        // check correctness of parameters
        checkDatatype();
        checkCountUndefined();
    
        
        our_datatype_traces->findAndAddUsage(my_datatype,getSourceLocation());
        
        
        return 0;
    }


    // checkDatatype
    //========================================================================
    int MPO_Get_count::checkDatatype()
    {
        /** \todo @creator of this method: what about return values ? */
        
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        if ((my_datatype == MPI_DATATYPE_NULL) && 
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": Datatype is MPI_DATATYPE_NULL!" 
                << std::endl;
                
            our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_NULL,getSourceLocation()); 
            our_log->getMsg(msg);   
        }
        else
        {
            if ((our_datatypes.find(my_datatype) == our_datatypes.end()) &&
                (MSG_ERR <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname() 
                    << ": Datatype is not valid!" 
                    << std::endl;
                    
                our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");
            
                MPO_String text;
                text.assign(&stream);      
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_NOT_VALID,getSourceLocation()); 
                our_log->getMsg(msg);   
            }
            else
            {
                std::stringstream stream;
                MPO_MSG_TYPE type = MSG_TYPE_ERROR;
                                
                switch (our_datatypes.find(my_datatype)->second)
                {
                    case (DATATYPE_CREATED_NOT_COMMITTED):
                        if ((our_errcodes.at(MARMOT_ERR_TYPE_NOT_COMMITTED))&&(MSG_ERR <= DEBUG_MODE))
                        {
                            stream
                                << "ERROR: "
                                << getCallname() 
                                << ": Datatype is not committed!" 
                                << std::endl;
                            
                            //added by Tobias check for corectness !
                            rvalue = MARMOT_ERR_TYPE_NOT_COMMITTED;
                        }
                        break;
                    case (DATATYPE_REDUCTION_C):
                        if ((our_errcodes.at(MARMOT_WARN_TYPE_REDUCTION_C))&&(MSG_ERR_WARN <= DEBUG_MODE))
                        {
                            stream
                                << "WARNING: "
                                << getCallname()
                                << ": Datatype is for reduction functions only!" 
                                << std::endl;
                            
                            type = MSG_TYPE_WARNING;
                            //added by Tobias check for corectness !
                            rvalue = MARMOT_WARN_TYPE_REDUCTION_C;
                        }
                        break;
                    case (DATATYPE_OPTIONAL_C):
                        if ((our_errcodes.at(MARMOT_ERR_TYPE_OPTIONAL_C))&&(MSG_ERR_WARN <= DEBUG_MODE))
                        {
                            stream
                                << "WARNING: "
                                << getCallname() 
                                << ": Datatype is only optional!" 
                                << std::endl;
                            
                            type = MSG_TYPE_WARNING;
                            //added by Tobias check for corectness !
                            rvalue = MARMOT_ERR_TYPE_OPTIONAL_C;
                        }
                        break;
                    case (DATATYPE_COMMITTED_FORTRAN):
                        if ((our_errcodes.at(MARMOT_ERR_TYPE_F))&&(MSG_ERR <= DEBUG_MODE) && 
                            (C_INTERFACE == INTERFACE_MODE))
                        {
                            stream
                                << "ERROR: "
                                << getCallname() 
                                << ": Datatype is Fortran-Type!" 
                                << std::endl;
                            
                            //added by Tobias check for corectness !
                            rvalue = MARMOT_ERR_TYPE_F;
                        }
                        break;
                    default:
                        // do nothing
                        break;
                }//end switch
                
                            
                if (rvalue != MARMOT_SUCCESS)
                {
                    our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");
                    
                    MPO_String text;
                    text.assign(&stream);      
                    MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
                    our_log->getMsg(msg);  
                }
            }
        }

        return rvalue;
    }


    // checkCountUndefined
    //========================================================================
    int MPO_Get_count::checkCountUndefined()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS; 
    if(!our_errcodes.at(MARMOT_WARN_COUNT_UNDEFINED))
        return rvalue;

        // Check if count is MPI_UNDEFINED.
        if ((*my_count == MPI_UNDEFINED) &&
            (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "WARNING: "
                << getCallname() 
                << ": Count is MPI_UNDEFINED! " 
                << std::endl;
            rvalue = MARMOT_WARN_COUNT_UNDEFINED;
            
            our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"datatype");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);  
        }

        return rvalue;
    }


    // getId
    //========================================================================
    int MPO_Get_count::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_GET_COUNT_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Get_count::checkPost(void)
    {
    
        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Get_count::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Get_count(my_status, my_datatype, my_count);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Get_count::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Get_count";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_GET_COUNT";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Get_count::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output 
            << getCallname()
            << "(*status, "
            << "datatype = " 
            << my_datatype 
            << ", *count";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"  
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Get_count::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_status;
        out < my_datatype;
        // out < my_count;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Get_count::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_status;
        in > my_datatype;
        // in > my_count;

        return in;
    }



}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/pt2pt/getcount.c.
/*@

MPI_Get_count - Gets the number of "top level" elements

Input Parameters:
+ status - return status of receive operation (Status) 
- datatype - datatype of each receive buffer element (handle) 

Output Parameter:
. count - number of received elements (integer) 
Notes:
If the size of the datatype is zero, this routine will return a count of
zero.  If the amount of data in 'status' is not an exact multiple of the 
size of 'datatype' (so that 'count' would not be integral), a 'count' of
'MPI_UNDEFINED' is returned instead.

.N fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TYPE

int MPI_Get_count( 
MPI_Status *status, 
MPI_Datatype datatype, 
int *count )
@ */

