/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Graph_map.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Graph_map.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Graph_map::MPO_Graph_map()
        : my_comm(MPI_COMM_NULL),
          my_nnodes(-1),
          my_index(0),
          my_edges(0),
          my_newrank(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Graph_map::MPO_Graph_map(MPO_Comm comm,
                                 int nnodes,
                                 int *index,
                                 int *edges,
                                 int *newrank)
        : my_comm(comm),
          my_nnodes(nnodes),
          my_index(index),
          my_edges(edges),
          my_newrank(newrank)
    {
    }


    // destructor
    //========================================================================
    MPO_Graph_map::~MPO_Graph_map()
    {
    }


    /**
     *  We overload MPO_Graph::checkPrerequisites
     *  to check other prerequisites:
     *  Check if communicator, size and edges fit together.
     */
    // checkPrerequisites2
    //========================================================================
    int MPO_Graph_map::checkPrerequisites2()
    {
        checkTopologyNew();
        checkEdgeArray();

        return 0;
    }


    // checkTopologyNew
    //========================================================================
    int MPO_Graph_map::checkTopologyNew()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        int comm_size = -1;
        PMPI_Comm_size(my_comm, &comm_size);

        // (1) Check if nnodes > size of old communicator
        if ((our_errcodes.at(MARMOT_ERR_TOPOLOGY_NEW_GT_COMM_OLD))&&(comm_size < my_nnodes) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream    << "ERROR: "
                      << getCallname()
                      << ": New topology is larger than old communicator:"
                      << std::endl
                      << " comm_size = "
                      << comm_size
                      << ", nnodes = "
                      << my_nnodes
                      << std::endl;
            rvalue = MARMOT_ERR_TOPOLOGY_NEW_GT_COMM_OLD;

            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        // (2) Check if nnodes < size of old communicator
        else if ((our_errcodes.at(MARMOT_WARN_TOPOLOGY_NEW_LT_COMM_OLD))&&(comm_size > my_nnodes) &&
                 (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "WARNING: "
                << getCallname()
                << ": New topology is smaller than old communicator:"
                << " Some nodes will get MPI_COMM_NULL."
                << std::endl
                << " comm_size = "
                << comm_size
                << ", nnodes = "
                << my_nnodes
                << std::endl;
            rvalue = MARMOT_WARN_TOPOLOGY_NEW_LT_COMM_OLD;

            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }                           // end checkTopologyNew


    // checkEdgeArray
    //========================================================================
    int MPO_Graph_map::checkEdgeArray()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;
    if(!our_errcodes.at(MARMOT_ERR_EDGE_INVALID))
        return rvalue;

        int comm_size = -1;
        PMPI_Comm_size(my_comm, &comm_size);

        // Check if all entries in my_edges are valid
        for (int i = 0; i < my_index[my_nnodes - 1]; i++)
        {
            if ((my_edges[i] >= comm_size) &&
                (MSG_ERR <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname()
                    << ": Invalid rank: edges["
                    << i
                    << "] = "
                    << my_edges[i]
                    << std::endl;
                rvalue = MARMOT_ERR_EDGE_INVALID;

                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                our_log->getMsg(msg);
            }
        }

        return rvalue;
    }                           // end checkEdgeArray


    // getId
    //========================================================================
    int MPO_Graph_map::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_CART_MAP_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Graph_map::checkPost(void)
    {

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Graph_map::checkPrerequisites(void)
    {


        MPO_Graph::checkPrerequisites();
        checkPrerequisites2();

        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Graph_map::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Graph_map(my_comm, my_nnodes, my_index, my_edges,
                                my_newrank);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Graph_map::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Graph_map";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_GRAPH_MAP";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Graph_map::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(comm = "
            << my_comm
            << ", nnodes = "
            << my_nnodes
            << ", *index, "
            << "*edges, "
            << "*newrank";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Graph_map::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_comm;
        out < my_nnodes;
        // out < my_index;
        // out < my_edges;
        // out < my_newrank;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Graph_map::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_comm;
        in > my_nnodes;
        // in > my_index;
        // in > my_edges;
        // in > my_newrank;

        return in;
    }

}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/topol/graph_map.c.
/*@

MPI_Graph_map - Maps process to graph topology information

Input Parameters:
+ comm - input communicator (handle)
. nnodes - number of graph nodes (integer)
. index - integer array specifying the graph structure, see
'MPI_GRAPH_CREATE'
- edges - integer array specifying the graph structure

Output Parameter:
. newrank - reordered rank of the calling process; 'MPI_UNDEFINED' if the
calling process does not belong to graph (integer)

.N fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_TOPOLOGY
.N MPI_ERR_COMM
.N MPI_ERR_ARG

int MPI_Graph_map (
MPI_Comm comm_old,
int nnodes,
int *index,
int *edges,
int *newrank )
@ */

