/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Group_free.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Group_free.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Group_free::MPO_Group_free()
        : my_delgroup(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Group_free::MPO_Group_free(MPI_Group * delgroup)
        : my_delgroup(delgroup)
    {
    }


    // destructor
    //========================================================================
    MPO_Group_free::~MPO_Group_free()
    {
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Group_free::checkPrerequisites()
    {


        // check correctness of parameters
        checkGroupDel();
    
        our_groups.erase(*my_delgroup);
        our_group_traces->findAndSetFree(*my_delgroup,getSourceLocation());
        
        
        return 0;
    }


    // checkGroupDel
    //========================================================================
    int MPO_Group_free::checkGroupDel()
    {
        /** \todo @creator of this method: what about return values ? */
        
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        if (*my_delgroup == MPI_GROUP_NULL)
        {
            if (MSG_ERR_WARN <= DEBUG_MODE)
            {
                std::stringstream stream;
                stream
                    << "WARNING: "
                    << getCallname()
                    << ": Group to free is already MPI_GROUP_NULL !" 
                    << std::endl;
                    
                our_group_traces->findAndPrintTraceNamed(my_group,stream,"group");
                
                MPO_String text;
                text.assign(&stream);      
                MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_WARN_GROUP_NULL,getSourceLocation()); 
                our_log->getMsg(msg);   
            }
        }
        else if ((our_groups.find(*my_delgroup) == our_groups.end()) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": Group not valid!" 
                << std::endl;
                
            our_group_traces->findAndPrintTraceNamed(my_group,stream,"group");
                
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_GROUP_NOT_VALID,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        return rvalue;
    }


    // getId
    //========================================================================
    int MPO_Group_free::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_GROUP_FREE_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Group_free::checkPost(void)
    {
    
        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Group_free::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Group_free(my_delgroup);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Group_free::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Group_free";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_GROUP_FREE";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Group_free::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output 
            << getCallname() 
            << "(*group";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"  
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Group_free::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_delgroup;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Group_free::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_delgroup;

        return in;
    }

}                               // end namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from
// mpich-1.2.4/src/context/group_free.c.
/*@

MPI_Group_free - Frees a group

Input Parameter
. group - group (handle) 

Notes:
On output, group is set to 'MPI_GROUP_NULL'.

.N fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_ARG
.N MPI_ERR_PERM_GROUP

int MPI_Group_free ( MPI_Group *group )
@ */

