/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Group_incl.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Group_incl.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Group_incl::MPO_Group_incl()
        : MPO_Groupconstructor(MPI_GROUP_NULL,
                               MPI_GROUP_NULL,
                               0,
                               -1),
          my_count(-1),
          my_ranks(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Group_incl::MPO_Group_incl(MPO_Group group,
                                   int count,
                                   int *ranks,
                                   MPI_Group *
                                   newgroup)
        : MPO_Groupconstructor(group,
                               MPI_GROUP_NULL,
                               newgroup,
                               1),
          my_count(count),
          my_ranks(ranks)
    {
    }


    // destructor
    //========================================================================
    MPO_Group_incl::~MPO_Group_incl()
    {
    }


    // checkPrerequisites2
    //========================================================================
    int MPO_Group_incl::checkPrerequisites2()
    {
        checkCountGroup(MSG_COUNT, MSG_NEW_GROUP_EQ_EMPTY);
        checkRankArray();

        return 0;
    }


    // checkCountGroup
    //========================================================================
    int MPO_Group_incl::checkCountGroup(const string KIND_OF_COUNT,
                                        const string KIND_OF_NEW_GROUP)
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if count is negative.
        if ((our_errcodes.at(MARMOT_ERR_COUNT_NEG))&&(my_count < 0) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": "
                << KIND_OF_COUNT
                << " = "
                << my_count
                << " less than 0 !"
                << std::endl;
            rvalue = MARMOT_ERR_COUNT_NEG;

            our_group_traces->findAndPrintTraceNamed(my_group,stream,"group");

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        // (2) Check if count is 0.
        if ((our_errcodes.at(MARMOT_WARN_COUNT_ZERO)) &&(my_count == 0) &&
            (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "WARNING: "
                << getCallname()
                << ": "
                << KIND_OF_COUNT
                << "= 0!"
                << KIND_OF_NEW_GROUP
                << std::endl;
            rvalue = MARMOT_WARN_COUNT_ZERO;

            our_group_traces->findAndPrintTraceNamed(my_group,stream,"group");

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }                           // end checkCountGroup


    // checkRankArray
    //========================================================================
    int MPO_Group_incl::checkRankArray()
    {

        /** \todo @creator of this method: what about return values ? */

        int rvalue = 1;
        int i = 0;
        int j = 0;
        int out = 0;
        int size = 0;

        PMPI_Group_size(my_group, &size);
        if (my_count > size)
        {
            if (MSG_ERR <= DEBUG_MODE)
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname()
                    << ": More entries in ranks than nodes in first group!"
                    << std::endl;

                our_group_traces->findAndPrintTraceNamed(my_group,stream,"group");

                //I don`t know which Errorcode we will use here maybe we need a new one - I selected one until someone takes care of this
                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_RANK_TOO_BIG,getSourceLocation());
                our_log->getMsg(msg);
            }
        }
        else
        {
            std::stringstream stream;
            bool bError = false;

            for (i = 0; i <= my_count - 1; i++)
            {
                if ((my_ranks[i] < 0) &&
                    (MSG_ERR <= DEBUG_MODE))
                {
                    stream
                        << "ERROR:"
                        << getCallname()
                        << ": rank "
                        << i
                        << "less than 0 !"
                        << std::endl;

                    bError = true;
                }
                if (my_ranks[i] > size - 1)
                {
                    out++;
                }
                else
                {
                    for (j = i + 1; j <= my_count - 1; j++)
                        if ((my_ranks[i] == my_ranks[j]) &&
                            (MSG_ERR <= DEBUG_MODE))
                        {
                            stream
                                << "ERROR: "
                                << getCallname()
                                << ": Rank "
                                << my_ranks[i]
                                << " is not only once in array of ranks!"
                                << std::endl;

                            bError = true;
                        }
                }
            }

            if (bError)
            {
                our_group_traces->findAndPrintTraceNamed(my_group,stream,"group");

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_RANK_NEG,getSourceLocation());
                our_log->getMsg(msg);
            }

            if ((out > 0) &&
                (MSG_ERR <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname()
                    << ": "
                    << out
                    << " rank(s) out of group!"
                    << std::endl;

                our_group_traces->findAndPrintTraceNamed(my_group,stream,"group");

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_RANK_TOO_BIG,getSourceLocation());
                our_log->getMsg(msg);
            }
        }

        return rvalue;
    }


    // getId
    //========================================================================
    int MPO_Group_incl::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_GROUP_INCL_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Group_incl::checkPost(void)
    {

        insertGroup();

        our_group_traces->addAndSetCreation(*my_newgroup,getSourceLocation());
        our_group_traces->findAndTryAddDependendency(*my_newgroup,my_group);

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Group_incl::checkPrerequisites(void)
    {


        MPO_Groupconstructor::checkPrerequisites();
        checkPrerequisites2();

        our_group_traces->findAndAddUsage(my_group,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Group_incl::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Group_incl(my_group, my_count, my_ranks, my_newgroup);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Group_incl::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Group_incl";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_GROUP_INCL";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Group_incl::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(group = "
            << my_group
            << ", n = "
            << my_count
            << ", *ranks, "
            << "*newgroup";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Group_incl::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_group;
        out < my_count;
        // out < my_ranks;
        // out < my_newgroup;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Group_incl::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_group;
        in > my_count;
        // in > my_ranks;
        // in > my_newgroup;

        return in;
    }

}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from
// mpich-1.2.4/src/context/group_incl.c.
/*@

MPI_Group_incl - Produces a group by reordering an existing group and
taking only listed members

Input Parameters:
+ group - group (handle)
. n - number of elements in array 'ranks' (and size of newgroup ) (integer)
- ranks - ranks of processes in 'group' to appear in 'newgroup' (array of
integers)

Output Parameter:
. newgroup - new group derived from above, in the order defined by 'ranks'
(handle)

Note:
This implementation does not currently check to see that the list of
ranks to ensure that there are no duplicates.

.N fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_GROUP
.N MPI_ERR_ARG
.N MPI_ERR_EXHAUSTED
.N MPI_ERR_RANK

.seealso: MPI_Group_free

int MPI_Group_incl (
MPI_Group group,
int n,
int *ranks,
MPI_Group *group_out )
@ */
