/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Init_thread.
 *
 *  @author Tobias Hilbrich
 *
 *  $Id$
 */

#include "config.h"
#include "mpo-objects.h"
#include "mpo-2-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo.h"
#include "mpo-enums.h"
#include "mpo_LogASCII.h"
#include "mpo_LogToServer.h"
#include "mpo_LogToServerDDT.h"
#include "mpo_GenericLogFilter.h"
#include "mpo-fortran-datatypehelpers.h"
#include "MPO_ResourceGatherer.h"
#include "MPO_ResourceGathererFull.h"
#include "MPO_ResourceGathererDummy.h"
#include "MPO_ResourceGathererGraph.h"

namespace MPO
{

#ifdef MARMOT_THREADED

    // default constructor
    //========================================================================
    MPO_Init_thread::MPO_Init_thread()
    {
    }


    // proper constructor
    //========================================================================
    MPO_Init_thread::MPO_Init_thread(int *argc,
                       char ***argv,
                       int required,
                       int *provided)
        : my_provided(provided),
          my_required(required)
    {
    	my_argc = argc;
    	my_argv = argv;

        /* The call to the constructor of MPO_Init_thread is special because this is
         * the place to initialize everything. We call MPI_Init_thread here instead
         * of calling it in callNativeMPI().
         */
    	if (our_call_init)
    	{
#ifdef MARMOT_USE_THREADSYNC
    	    MPO_Space::getSyncObject()->enterPMPI();
#endif
    	    my_rvalue = PMPI_Init_thread(argc, argv, required, provided);
#ifdef MARMOT_USE_THREADSYNC
    	    MPO_Space::getSyncObject()->leavePMPI();
#endif
    	}/*our_call_init*/

        /**
         * Create a new MPI_COMM_WORLD which excludes the debug server.
         * Init some environmentals.
         */
        initCommAndEnv();

        /* Add all the predefined resources such as errhandlers, datatypes,
         * communicators, operations and groups to our maps.
         */
        initFactories();
        initErrhandlers();
        initDatatypes();
        initCommunicators();
        initOperations();
        initGroups();
        initLogfile();
        //we will call this in callDebugServer()
        //because we need to know wheter source locations are present or not
        //initResourceGatherers();

        /* Set buffer attached to 'false'. */
        buffer_attached = 0;

        /*
         * Set the required and provided threadlevel
         */
        our_threadlevel_required = required;
        our_threadlevel_provided = *provided;

        //now set MPI as initialized
        our_init_was_called = true;
    }                             // end constructor


    // destructor
    //========================================================================
    MPO_Init_thread::~MPO_Init_thread()
    {
    }

    // checkPrerequisites
    //========================================================================
    int MPO_Init_thread::checkPrerequisites()
    {
    	//do all the normal checks
    	int rvalue = MPO_Init::checkPrerequisites();

    	//check for the value of provided
        if ((our_errcodes.at(MARMOT_WARN_THREADLEVEL_GOT_LESS))&&(my_required > *my_provided) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "WARNING: "
                << getCallname()
                << ": provided is smaller than required"
                << "(required: "
                << MPO_Space::getThreadLevelName(my_required)
                << " provided: "
                << MPO_Space::getThreadLevelName(*my_provided)
                << ")"
                << std::endl;
            rvalue = MARMOT_WARN_THREADLEVEL_GOT_LESS;

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }

    // getId
    //========================================================================
    int MPO_Init_thread::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_INIT_THREAD_ID;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Init_thread::callNativeMPI()
    {
        // we do not call MPI_Init_thread here because it was already called in
        // the constructor
        checkPrerequisites();
        return my_rvalue;
    }

    int MPO_Init_thread::callDebugServer()
    {
    	//we call "callNativeMPI" here to perform checkPrerequisities
    	//"callNativeMPI" actually does not call MPI_Init_thread which
    	//was done before
    	//we will do the check only for process 0 !
    	int rvalue;
#ifdef MARMOT_USE_THREADSYNC
        MPO_Space::getSyncObject()->enterPMPI();
#endif
    	PMPI_Comm_rank(MPI_COMM_WORLD,&rvalue);
#ifdef MARMOT_USE_THREADSYNC
        MPO_Space::getSyncObject()->leavePMPI();
#endif
    	if (rvalue == 0)
    		callNativeMPI();

    	rvalue = MPO_Init::callDebugServer();

        return MPI_SUCCESS;
    }


    // getCallname
    //========================================================================
    string MPO_Init_thread::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Init_thread";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_INIT_THREAD";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Init_thread::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        if (C_INTERFACE == INTERFACE_MODE)
        {
            output
                << getCallname()
                << "(*argc, ***argv, "
                << my_required
                << ", *provided)"
                << std::endl;
        }

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output
                << getCallname()
                << "("
                << my_required
                << " , provided, ierror)"
                << std::endl;
        }

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Init_thread::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Init::save(out);
        out < my_required;
        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Init_thread::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Init::restore(in);
        in > my_required;
        return in;
    }

#else

	MPO_Init_thread::MPO_Init_thread(int *argc, char ***argv,int required, int *provided)
		:MPO_Init(argc,argv)
	{
		std::cerr << "WARNING: you built MARMOT without Threading support but called MPI_Init_thread"<<std::endl;
	}

	MPO_Init_thread::MPO_Init_thread(void)
	{
	    //nothing to do here
	}

#endif
//end of MARMOT_THREADED

}                               // end of namespace MPO


