/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Intercomm_merge.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Intercomm_merge.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Intercomm_merge::
    MPO_Intercomm_merge()
        : MPO_Intercommunicators(MPI_COMM_NULL),
          my_high(0),
          my_newintracomm(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Intercomm_merge::MPO_Intercomm_merge(MPO_Comm intercomm,
                                             int high,
                                             MPI_Comm *
                                             newintracomm)
        : MPO_Intercommunicators(intercomm),
          my_high(high),
          my_newintracomm(newintracomm)
    {
    }


    // destructor
    //========================================================================
    MPO_Intercomm_merge::~MPO_Intercomm_merge()
    {
    }


    // getId
    //========================================================================
    int MPO_Intercomm_merge::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_INTERCOMM_MERGE_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Intercomm_merge::checkPost(void)
    {

        our_comms.insert(*my_newintracomm);

        our_comm_traces->addAndSetCreation(*my_newintracomm,getSourceLocation());
        our_comm_traces->findAndTryAddDependendency(*my_newintracomm,my_comm);

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Intercomm_merge::checkPrerequisites(void)
    {


        MPO_Intercommunicators::checkPrerequisites();

        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Intercomm_merge::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Intercomm_merge(my_comm, my_high, my_newintracomm);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Intercomm_merge::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Intercomm_merge";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_INTERCOMM_MERGE";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Intercomm_merge::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(intercomm = "
            << my_comm
            << ", high = "
            << my_high
            << ", *newintracomm";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Intercomm_merge::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_comm;
        out < my_high;
        // out < my_newintracomm;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Intercomm_merge::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_comm;
        in > my_high;
        // in > my_intracomm;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/context/ic_merge.c.
/*@

MPI_Intercomm_merge - Creates an intracommuncator from an
intercommunicator

Input Parameters:
+ comm - Intercommunicator
- high - Used to order the groups of the two intracommunicators within
comm when creating the new communicator.

Output Parameter:
. comm_out - Created intracommunicator

.N fortran

Algorithm:
.vb
1) Allocate two contexts
2) Local and remote group leaders swap high values
3) Determine the high value.
4) Merge the two groups and make the intra-communicator
.ve

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_EXHAUSTED

.seealso: MPI_Intercomm_create, MPI_Comm_free

int MPI_Intercomm_merge (
MPI_Comm comm,
int high,
MPI_Comm *comm_out )
@ */

