/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 * @file
 * This Sourcefile implements the MPO_LogASCII object.
 * This object recieves MPO_MsgObject objects and creates a logfile from their data.
 * The logfile will be a pur ASCII-text file.
 * @author Tobias Hilbrich
 * @date 29.10.2005
 * $Id: mpo_LogASCII.cc 887 2008-11-10 16:19:05Z tobias $
 */

#include "mpo_LogASCII.h"
#include "mpo-binaryIO.h"

#include "config.h"
#include <time.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <sstream>
#include <ostream>

namespace MPO 
{
    
    /** 
     * Constructor that sets state to invalid. By Tobias, Date 19.11.2005.
     */
    MPO_LogASCII::MPO_LogASCII ()
    {
        m_bInitialized = 0;
        m_bHadAFlush = false;
    }
    
    /**
     * Constructor. By Tobias, Date 03.11.2005.
     * @param logFileName name of the output file.
     * @param MPIStandardFilename name of a file that contains information about the MPI-Standard.
     */
    MPO_LogASCII::MPO_LogASCII (string logFileName)
    {
        init(logFileName);
    } 
    
    /**
     * Constructor. By Tobias, Date 03.11.2005.
     * @param logFileName name of the output file.
     * @param MPIStandardFilename name of a file that contains information about the MPI-Standard.
     */
    bool MPO_LogASCII::init (string logFileName)
    {
        initMPIStandardReferencer ();
    
        m_bInitialized = 0;
        m_bHadAFlush = false;

        initSocket();

        if (!createOfstream(logFileName) || !printHeader ())
            return false;
      
        m_bInitialized = 1;
        
        return true;
    }
   
    /**
     * Standard Destructor. By Tobias, Date 03.11.2005.
     */
    MPO_LogASCII::~MPO_LogASCII ()
    {
        if (m_bInitialized) {
            closeLog (); //sets m_bInitialized to false and makes a call to PrintTrailer
        }
    }
   
    /**
     * Gets a message object, looks for MPI-Standard reference text and calls logMsg. 
     * By Tobias, Date 03.11.2005.
     * @param pMsg the Message that will be logged,
     * @return true when the operation succeeds false otherwise,
     */
    bool MPO_LogASCII::getMsg (MPO_MsgObject &pMsg)
    {
        if (!m_bInitialized)
            return (false);
      
        pMsg.debugout (*m_pLog);

        //print information on the MPI-Standard when possibile
        MARMOT_ERR_CODES err = pMsg.getErrorCode ();
          
        const char *temp = NULL;

        if ( (err != MARMOT_SUCCESS) &&
             (err != MARMOT_ERR_UNDEFINED) &&
             (err != MARMOT_ERR_MARMOT) &&
             (err != MARMOT_LAST_ERR_CODE)    )
        {
          temp = getLocalPathToMPIStandard(err);
          (*m_pLog) << " for MPI-Standard information see:" << temp << std::endl << std::endl;
        }

        // if there is a socket send msg
        pMsg.writeToSocket(getLogSocket(), temp);
        // wait for ACK from log-host
        if (pMsg.getMsgType() == MSG_TYPE_ERROR)
          CSocketLogger::instance()->waitForAck();

        if ((pMsg.getMsgType() == MSG_TYPE_ERROR) || m_bHadAFlush || (LOGFLUSH_TYPE == LOGFILE_FLUSH_TYPE_IMMEDIATE))
          flushOfstream();
      
        return (true);
    }
   
    /**
     * Initializes the ofstream for the Logfile. 
     * By Tobias, Date 03.11.2005.
     * @param logFileName the filename of the logfile that will be created.
     * @return true when the operation succeeds false otherwise,
     */
    bool MPO_LogASCII::createOfstream (string logFileName)
    {
        m_pLog = new ofstream (logFileName.data(),std::ios_base::out | std::ios_base::trunc);
      
        if (!m_pLog)
          return false;
        
        if (m_pLog->fail())
        {
          std::cerr << "Marmot: Could not create logfile (" << logFileName << ")" << std::endl;
          return false;
        }
     
        return true;
    }
   
    /**
     * Flushs the ofstream in order to prevent lose of data.
     * By Tobias, Date 03.11.2005.
     * @return true when the operation succeeds false otherwise,
     */
    bool MPO_LogASCII::flushOfstream (void)
    {
        if (!m_bInitialized)
            return (false);
      
        m_pLog->flush();

        m_bHadAFlush = true;

        return (true);
    }
   
    /**
     * Prints the Header of the Logfile like date and so on.
     * By Tobias, Date 03.11.2005.
     * @return true when the operation succeeds false otherwise,
     */
    bool MPO_LogASCII::printHeader (void)
    {
        if (!m_pLog)
            return (false);
      
        //read the system time
        //read the system time
        char buf[128];
      
        struct tm *ptr;
        time_t tm;
        tm = time(NULL);
        ptr = localtime(&tm);
        strftime(buf ,128 , "%c.\n",ptr);
      
        //print the header
        (*m_pLog) << "================================================================"<<std::endl;
#ifdef HAVE_UNISTD_H
        (*m_pLog) << "Marmot ASCII Logfile from: "<<buf;
#endif
#ifndef HAVE_UNISTD_H
        (*m_pLog) << "Marmot ASCII Logfile"<<std::endl;
#endif
        (*m_pLog) << "================================================================"<<std::endl;
      
        return (true);
    }
   
    /**
     * Prints the Trailer of the Logfile (maybe not needed). 
     * By Tobias, Date 03.11.2005.
     * @return true when the operation succeeds false otherwise,
     */
    bool MPO_LogASCII::printTrailer (void)
    {
        if (!m_bInitialized)
            return (false);
      
        (*m_pLog) << "================================================================"<<std::endl;
        (*m_pLog) << "End of Logfile"<<std::endl;
        (*m_pLog) << "================================================================"<<std::endl;
            
        return (true);
    }
   
    /**
     * Flushs the ofstream and closes it. By Tobias, Date 03.11.2005.
     * @return true when the operation succeeds false otherwise,
     */
    bool MPO_LogASCII::closeLog (void)
    {
        if (!m_bInitialized)
            return (false);

		/*make sure we also make our output available to connected clients*/
		CSocketLogger::instance()->waitForAck();
      
        printTrailer (); 
        m_pLog->close();
        m_bInitialized = 0;
        delete (m_pLog);
        m_pLog = NULL;

        return (true);
    }
   

}//end of namespace MPO







