/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 * @file 
 * This File implements the MPO_MsgObject class.
 * Message Objects are used for logging. They get Information about
 * the data to be logged and can be passed to a implementation of the
 * MPO_LogFile interface.
 * @author Tobias Hilbrich
 * @date 25.10.2005
 * $Id: mpo_MsgObject.cc 952 2009-04-28 15:43:19Z hpczink $
 */

#include "mpo-objects.h"
#include "mpo_MsgObject.h"
#include "mpo-binaryIO.h"
#include "SocketLib.h"
#include "mpo-debugserver.h"

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

//==================================================
// Initialization of static data
//==================================================

int MPO::MPO_MsgObject::m_GlobalStaticRank = -2; //we will use -1 for the debugserver

namespace MPO 
{	
		
    /**
     * Create invalid object constructor. By Tobias Hilbrich, Date 29.10.2005.
     */
    MPO_MsgObject::MPO_MsgObject () 
    {
        //sets some not error producing standard values
        /** might be removed to gain some speed up.*/
        m_MsgType            = MSG_TYPE_INFO;
        m_rank               = 0;
        m_OutputType         = OUTPUT_TYPE_PURE_TEXT;
        m_Callinfo           = "";
        m_Text               = "";
        m_CallID             = MPO_LAST_CALL_ID; //set invalid
        m_ErrorCode          = MARMOT_LAST_ERR_CODE; //set invalid
        m_Timestamp          = 0;
        m_ThreadRank         = 0;
    }

    //======================================
    // Constructor Versions with Rank
    //======================================
   
    /**
     * Constructor for OUTPUT_TYPE_PURE_TEXT Messages. By Tobias Hilbrich, Date 29.10.2005.
     * @param MsgType see MPO_MsgObject::initMsgObject.
     * @param Timestamp see MPO_MsgObject::initMsgObject.
     * @param rank see MPO_MsgObject::initMsgObject.
     * @param Text see MPO_MsgObject::initMsgObject.
     * @param ErrorCode see MPO_MsgObject::initMsgObject.
     */
    MPO_MsgObject::MPO_MsgObject (MPO_MSG_TYPE MsgType, long Timestamp, int rank, string Text, MARMOT_ERR_CODES ErrorCode)
    {
        initMsgObject (MsgType, OUTPUT_TYPE_PURE_TEXT, Timestamp, rank, Text, "" , MPO_LAST_CALL_ID, ErrorCode, m_CodeReference);
    }
	
    /**
     * Constructor for OUTPUT_TYPE_PURE_TEXT Messages with additional source code location. By Tobias Hilbrich, Date 29.10.2005.
     * @param MsgType see MPO_MsgObject::initMsgObject.
     * @param Timestamp see MPO_MsgObject::initMsgObject.
     * @param rank see MPO_MsgObject::initMsgObject.
     * @param Text see MPO_MsgObject::initMsgObject.
     * @param CodeReference see MPO_MsgObject::initMsgObject.
     * @param ErrorCode see MPO_MsgObject::initMsgObject.
     */
    MPO_MsgObject::MPO_MsgObject (MPO_MSG_TYPE MsgType, long Timestamp, int rank, string Text, MARMOT_ERR_CODES ErrorCode, MPO_SourceLocation CodeReference) 
    {
        initMsgObject (MsgType, OUTPUT_TYPE_PURE_TEXT, Timestamp, rank, Text, "" , MPO_LAST_CALL_ID, ErrorCode, CodeReference);
    }
	
    /**
     * Constructor for OUTPUT_TYPE_PURE_CALLINFO Messages. By Tobias Hilbrich, Date 29.10.2005.
     * @param MsgType see MPO_MsgObject::initMsgObject.
     * @param Timestamp see MPO_MsgObject::initMsgObject.
     * @param rank see MPO_MsgObject::initMsgObject.
     * @param Callinfo see MPO_MsgObject::initMsgObject.
     * @param CallID see MPO_MsgObject::initMsgObject.
     * @param ErrorCode see MPO_MsgObject::initMsgObject.
     */
    MPO_MsgObject::MPO_MsgObject (MPO_MSG_TYPE MsgType, long Timestamp, int rank, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode)
    {
        initMsgObject (MsgType, OUTPUT_TYPE_PURE_CALLINFO, Timestamp, rank, "", Callinfo , CallID, ErrorCode, m_CodeReference);
    }
	
    /**
     * Constructor for OUTPUT_TYPE_PURE_CALLINFO Messages with additional source code location. By Tobias Hilbrich, Date 29.10.2005.
     * @param MsgType see MPO_MsgObject::initMsgObject.
     * @param Timestamp see MPO_MsgObject::initMsgObject.
     * @param rank see MPO_MsgObject::initMsgObject.
     * @param Callinfo see MPO_MsgObject::initMsgObject.
     * @param CallID see MPO_MsgObject::initMsgObject.
     * @param ErrorCode see MPO_MsgObject::initMsgObject.
     * @param CodeReference see MPO_MsgObject::initMsgObject.
     */
    MPO_MsgObject::MPO_MsgObject (MPO_MSG_TYPE MsgType, long Timestamp, int rank, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode, MPO_SourceLocation CodeReference)
    {
        initMsgObject (MsgType, OUTPUT_TYPE_PURE_CALLINFO, Timestamp, rank, "", Callinfo , CallID, ErrorCode, CodeReference);
    }
	
    /**
     * Constructor for OUTPUT_TYPE_TEXT_AND_CALLINFO Messages. By Tobias Hilbrich, Date 29.10.2005.
     * @param MsgType see MPO_MsgObject::initMsgObject.
     * @param Timestamp see MPO_MsgObject::initMsgObject.
     * @param rank see MPO_MsgObject::initMsgObject.
     * @param Text see MPO_MsgObject::initMsgObject.
     * @param Callinfo see MPO_MsgObject::initMsgObject.
     * @param CallID see MPO_MsgObject::initMsgObject.
     * @param ErrorCode see MPO_MsgObject::initMsgObject.
     */
    MPO_MsgObject::MPO_MsgObject (MPO_MSG_TYPE MsgType, long Timestamp, int rank, string Text, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode)
    {
        initMsgObject (MsgType, OUTPUT_TYPE_TEXT_AND_CALLINFO, Timestamp, rank, Text, Callinfo , CallID, ErrorCode, m_CodeReference);
    }
	
    /**
     * Constructor for OUTPUT_TYPE_TEXT_AND_CALLINFO Messages  with additional source code location. By Tobias Hilbrich, Date 29.10.2005.
     * @param MsgType see MPO_MsgObject::initMsgObject.
     * @param Timestamp see MPO_MsgObject::initMsgObject.
     * @param rank see MPO_MsgObject::initMsgObject.
     * @param Text see MPO_MsgObject::initMsgObject.
     * @param Callinfo see MPO_MsgObject::initMsgObject.
     * @param CallID see MPO_MsgObject::initMsgObject.
     * @param ErrorCode see MPO_MsgObject::initMsgObject.
     * @param CodeReference see MPO_MsgObject::initMsgObject.
     */
    MPO_MsgObject::MPO_MsgObject (MPO_MSG_TYPE MsgType, long Timestamp, int rank, string Text, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode, MPO_SourceLocation CodeReference)
    {
        initMsgObject (MsgType, OUTPUT_TYPE_TEXT_AND_CALLINFO, Timestamp, rank, Text, Callinfo , CallID, ErrorCode, CodeReference);
    }
   
    //======================================
    // Constructor Versions without Rank
    //======================================
   
    /**
     * Constructor for OUTPUT_TYPE_PURE_TEXT Messages. By Tobias Hilbrich, Date 29.10.2005.
     * Rank is determined by the rank of the message creating process.
     * @param MsgType see MPO_MsgObject::initMsgObject.
     * @param Timestamp see MPO_MsgObject::initMsgObject.
     * @param Text see MPO_MsgObject::initMsgObject.
     * @param ErrorCode see MPO_MsgObject::initMsgObject.
     */
    MPO_MsgObject::MPO_MsgObject (MPO_MSG_TYPE MsgType, long Timestamp, string Text, MARMOT_ERR_CODES ErrorCode)
    {
        if (m_GlobalStaticRank < -1)
            initGlobalStaticRank();

        initMsgObject (MsgType, OUTPUT_TYPE_PURE_TEXT, Timestamp, m_GlobalStaticRank, Text, "" , MPO_LAST_CALL_ID, ErrorCode, m_CodeReference);
    }
	
    /**
     * Constructor for OUTPUT_TYPE_PURE_TEXT Messages with additional source code location. By Tobias Hilbrich, Date 29.10.2005.
     * Rank is determined by the rank of the message creating process.
     * @param MsgType see MPO_MsgObject::initMsgObject.
     * @param Timestamp see MPO_MsgObject::initMsgObject.
     * @param Text see MPO_MsgObject::initMsgObject.
     * @param CodeReference see MPO_MsgObject::initMsgObject.
     * @param ErrorCode see MPO_MsgObject::initMsgObject.
     */
    MPO_MsgObject::MPO_MsgObject (MPO_MSG_TYPE MsgType, long Timestamp, string Text, MARMOT_ERR_CODES ErrorCode, MPO_SourceLocation CodeReference) 
    {
        if (m_GlobalStaticRank < -1)
            initGlobalStaticRank();

        initMsgObject (MsgType, OUTPUT_TYPE_PURE_TEXT, Timestamp, m_GlobalStaticRank, Text, "" , MPO_LAST_CALL_ID, ErrorCode, CodeReference);
    }
	
    /**
     * Constructor for OUTPUT_TYPE_PURE_CALLINFO Messages. By Tobias Hilbrich, Date 29.10.2005.
     * Rank is determined by the rank of the message creating process.
     * @param MsgType see MPO_MsgObject::initMsgObject.
     * @param Timestamp see MPO_MsgObject::initMsgObject.
     * @param Callinfo see MPO_MsgObject::initMsgObject.
     * @param CallID see MPO_MsgObject::initMsgObject.
     * @param ErrorCode see MPO_MsgObject::initMsgObject.
     */
    MPO_MsgObject::MPO_MsgObject (MPO_MSG_TYPE MsgType, long Timestamp, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode)
    {
        if (m_GlobalStaticRank < -1)
            initGlobalStaticRank();

        initMsgObject (MsgType, OUTPUT_TYPE_PURE_CALLINFO, Timestamp, m_GlobalStaticRank, "", Callinfo , CallID, ErrorCode, m_CodeReference);
    }
	
    /**
     * Constructor for OUTPUT_TYPE_PURE_CALLINFO Messages with additional source code location. By Tobias Hilbrich, Date 29.10.2005.
     * Rank is determined by the rank of the message creating process.
     * @param MsgType see MPO_MsgObject::initMsgObject.
     * @param Timestamp see MPO_MsgObject::initMsgObject.
     * @param Callinfo see MPO_MsgObject::initMsgObject.
     * @param CallID see MPO_MsgObject::initMsgObject.
     * @param ErrorCode see MPO_MsgObject::initMsgObject.
     * @param CodeReference see MPO_MsgObject::initMsgObject.
     */
    MPO_MsgObject::MPO_MsgObject (MPO_MSG_TYPE MsgType, long Timestamp, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode, MPO_SourceLocation CodeReference)
    {
        if (m_GlobalStaticRank < -1)
            initGlobalStaticRank();

        initMsgObject (MsgType, OUTPUT_TYPE_PURE_CALLINFO, Timestamp, m_GlobalStaticRank, "", Callinfo , CallID, ErrorCode, CodeReference);
    }
	
    /**
     * Constructor for OUTPUT_TYPE_TEXT_AND_CALLINFO Messages. By Tobias Hilbrich, Date 29.10.2005.
     * Rank is determined by the rank of the message creating process.
     * @param MsgType see MPO_MsgObject::initMsgObject.
     * @param Timestamp see MPO_MsgObject::initMsgObject.
     * @param Text see MPO_MsgObject::initMsgObject.
     * @param Callinfo see MPO_MsgObject::initMsgObject.
     * @param CallID see MPO_MsgObject::initMsgObject.
     * @param ErrorCode see MPO_MsgObject::initMsgObject.
     */
    MPO_MsgObject::MPO_MsgObject (MPO_MSG_TYPE MsgType, long Timestamp, string Text, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode)
    {
        if (m_GlobalStaticRank < -1)
            initGlobalStaticRank();

        initMsgObject (MsgType, OUTPUT_TYPE_TEXT_AND_CALLINFO, Timestamp, m_GlobalStaticRank, Text, Callinfo , CallID, ErrorCode, m_CodeReference);
    }
	
    /**
     * Constructor for OUTPUT_TYPE_TEXT_AND_CALLINFO Messages  with additional source code location. By Tobias Hilbrich, Date 29.10.2005.
     * Rank is determined by the rank of the message creating process.
     * @param MsgType see MPO_MsgObject::initMsgObject.
     * @param Timestamp see MPO_MsgObject::initMsgObject.
     * @param Text see MPO_MsgObject::initMsgObject.
     * @param Callinfo see MPO_MsgObject::initMsgObject.
     * @param CallID see MPO_MsgObject::initMsgObject.
     * @param ErrorCode see MPO_MsgObject::initMsgObject.
     * @param CodeReference see MPO_MsgObject::initMsgObject.
     */
    MPO_MsgObject::MPO_MsgObject (MPO_MSG_TYPE MsgType, long Timestamp, string Text, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode, MPO_SourceLocation CodeReference)
    {
        if (m_GlobalStaticRank < -1)
            initGlobalStaticRank();

        initMsgObject (MsgType, OUTPUT_TYPE_TEXT_AND_CALLINFO, Timestamp, m_GlobalStaticRank, Text, Callinfo , CallID, ErrorCode, CodeReference);
    }
   
    //======================================
    // Init Methods with Rank
    //======================================
   
    /**
     * Init for OUTPUT_TYPE_PURE_TEXT Messages. By Tobias Hilbrich, Date 29.10.2005.
     * @param MsgType see MPO_MsgObject::initMsgObject.
     * @param Timestamp see MPO_MsgObject::initMsgObject.
     * @param rank see MPO_MsgObject::initMsgObject.
     * @param Text see MPO_MsgObject::initMsgObject.
     * @param ErrorCode see MPO_MsgObject::initMsgObject.
     */
    void MPO_MsgObject::initMsgObject (MPO_MSG_TYPE MsgType, long Timestamp, int rank, string Text, MARMOT_ERR_CODES ErrorCode)
    {
        initMsgObject (MsgType, OUTPUT_TYPE_PURE_TEXT, Timestamp, rank, Text, "" , MPO_LAST_CALL_ID, ErrorCode, m_CodeReference);
    }
	
    /**
     * Init for OUTPUT_TYPE_PURE_TEXT Messages with additional source code location. By Tobias Hilbrich, Date 29.10.2005.
     * @param MsgType see MPO_MsgObject::initMsgObject.
     * @param Timestamp see MPO_MsgObject::initMsgObject.
     * @param rank see MPO_MsgObject::initMsgObject.
     * @param Text see MPO_MsgObject::initMsgObject.
     * @param CodeReference see MPO_MsgObject::initMsgObject.
     * @param ErrorCode see MPO_MsgObject::initMsgObject.
     */
    void MPO_MsgObject::initMsgObject (MPO_MSG_TYPE MsgType, long Timestamp, int rank, string Text, MARMOT_ERR_CODES ErrorCode, MPO_SourceLocation CodeReference) 
    {
        initMsgObject (MsgType, OUTPUT_TYPE_PURE_TEXT, Timestamp, rank, Text, "" , MPO_LAST_CALL_ID, ErrorCode, CodeReference);
    }
	
    /**
     * Init for OUTPUT_TYPE_PURE_CALLINFO Messages. By Tobias Hilbrich, Date 29.10.2005.
     * @param MsgType see MPO_MsgObject::initMsgObject.
     * @param Timestamp see MPO_MsgObject::initMsgObject.
     * @param rank see MPO_MsgObject::initMsgObject.
     * @param Callinfo see MPO_MsgObject::initMsgObject.
     * @param CallID see MPO_MsgObject::initMsgObject.
     * @param ErrorCode see MPO_MsgObject::initMsgObject.
     */
    void MPO_MsgObject::initMsgObject (MPO_MSG_TYPE MsgType, long Timestamp, int rank, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode)
    {
        initMsgObject (MsgType, OUTPUT_TYPE_PURE_CALLINFO, Timestamp, rank, "", Callinfo , CallID, ErrorCode, m_CodeReference);
    }
	
    /**
     * Init for OUTPUT_TYPE_PURE_CALLINFO Messages with additional source code location. By Tobias Hilbrich, Date 29.10.2005.
     * @param MsgType see MPO_MsgObject::initMsgObject.
     * @param Timestamp see MPO_MsgObject::initMsgObject.
     * @param rank see MPO_MsgObject::initMsgObject.
     * @param Callinfo see MPO_MsgObject::initMsgObject.
     * @param CallID see MPO_MsgObject::initMsgObject.
     * @param ErrorCode see MPO_MsgObject::initMsgObject.
     * @param CodeReference see MPO_MsgObject::initMsgObject.
     */
    void MPO_MsgObject::initMsgObject (MPO_MSG_TYPE MsgType, long Timestamp, int rank, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode, MPO_SourceLocation CodeReference)
    {
        initMsgObject (MsgType, OUTPUT_TYPE_PURE_CALLINFO, Timestamp, rank, "", Callinfo , CallID, ErrorCode, CodeReference);
    }
	
    /**
     * Init for OUTPUT_TYPE_TEXT_AND_CALLINFO Messages. By Tobias Hilbrich, Date 29.10.2005.
     * @param MsgType see MPO_MsgObject::initMsgObject.
     * @param Timestamp see MPO_MsgObject::initMsgObject.
     * @param rank see MPO_MsgObject::initMsgObject.
     * @param Text see MPO_MsgObject::initMsgObject.
     * @param Callinfo see MPO_MsgObject::initMsgObject.
     * @param CallID see MPO_MsgObject::initMsgObject.
     * @param ErrorCode see MPO_MsgObject::initMsgObject.
     */
    void MPO_MsgObject::initMsgObject (MPO_MSG_TYPE MsgType, long Timestamp, int rank, string Text, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode)
    {
        initMsgObject (MsgType, OUTPUT_TYPE_TEXT_AND_CALLINFO, Timestamp, rank, Text, Callinfo , CallID, ErrorCode, m_CodeReference);
    }
	
    /**
     * Init for OUTPUT_TYPE_TEXT_AND_CALLINFO Messages  with additional source code location. By Tobias Hilbrich, Date 29.10.2005.
     * @param MsgType see MPO_MsgObject::initMsgObject.
     * @param Timestamp see MPO_MsgObject::initMsgObject.
     * @param rank see MPO_MsgObject::initMsgObject.
     * @param Text see MPO_MsgObject::initMsgObject.
     * @param Callinfo see MPO_MsgObject::initMsgObject.
     * @param CallID see MPO_MsgObject::initMsgObject.
     * @param ErrorCode see MPO_MsgObject::initMsgObject.
     * @param CodeReference see MPO_MsgObject::initMsgObject.
     */
    void MPO_MsgObject::initMsgObject (MPO_MSG_TYPE MsgType, long Timestamp, int rank, string Text, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode, MPO_SourceLocation CodeReference)
    {
        initMsgObject (MsgType, OUTPUT_TYPE_TEXT_AND_CALLINFO, Timestamp, rank, Text, Callinfo , CallID, ErrorCode, CodeReference);
    }
   
    //======================================
    // Init Methods without Rank
    //======================================
   
    /**
     * Init for OUTPUT_TYPE_PURE_TEXT Messages. By Tobias Hilbrich, Date 29.10.2005.
     * Rank is determined by the rank of the message creating process.
     * @param MsgType see MPO_MsgObject::initMsgObject.
     * @param Timestamp see MPO_MsgObject::initMsgObject.
     * @param Text see MPO_MsgObject::initMsgObject.
     * @param ErrorCode see MPO_MsgObject::initMsgObject.
     */
    void MPO_MsgObject::initMsgObject (MPO_MSG_TYPE MsgType, long Timestamp, string Text, MARMOT_ERR_CODES ErrorCode)
    {
        if (m_GlobalStaticRank < -1)
            initGlobalStaticRank();

        initMsgObject (MsgType, OUTPUT_TYPE_PURE_TEXT, Timestamp, m_GlobalStaticRank, Text, "" , MPO_LAST_CALL_ID, ErrorCode, m_CodeReference);
    }
	
    /**
     * Init for OUTPUT_TYPE_PURE_TEXT Messages with additional source code location. By Tobias Hilbrich, Date 29.10.2005.
     * Rank is determined by the rank of the message creating process.
     * @param MsgType see MPO_MsgObject::initMsgObject.
     * @param Timestamp see MPO_MsgObject::initMsgObject.
     * @param Text see MPO_MsgObject::initMsgObject.
     * @param CodeReference see MPO_MsgObject::initMsgObject.
     * @param ErrorCode see MPO_MsgObject::initMsgObject.
     */
    void MPO_MsgObject::initMsgObject (MPO_MSG_TYPE MsgType, long Timestamp, string Text, MARMOT_ERR_CODES ErrorCode, MPO_SourceLocation CodeReference) 
    {
        if (m_GlobalStaticRank < -1)
            initGlobalStaticRank();

        initMsgObject (MsgType, OUTPUT_TYPE_PURE_TEXT, Timestamp, m_GlobalStaticRank, Text, "" , MPO_LAST_CALL_ID, ErrorCode, CodeReference);
    }
	
    /**
     * Init for OUTPUT_TYPE_PURE_CALLINFO Messages. By Tobias Hilbrich, Date 29.10.2005.
     * Rank is determined by the rank of the message creating process.
     * @param MsgType see MPO_MsgObject::initMsgObject.
     * @param Timestamp see MPO_MsgObject::initMsgObject.
     * @param Callinfo see MPO_MsgObject::initMsgObject.
     * @param CallID see MPO_MsgObject::initMsgObject.
     * @param ErrorCode see MPO_MsgObject::initMsgObject.
     */
    void MPO_MsgObject::initMsgObject (MPO_MSG_TYPE MsgType, long Timestamp, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode)
    {
        if (m_GlobalStaticRank < -1)
            initGlobalStaticRank();

        initMsgObject (MsgType, OUTPUT_TYPE_PURE_CALLINFO, Timestamp, m_GlobalStaticRank, "", Callinfo , CallID, ErrorCode, m_CodeReference);
    }
	
    /**
     * Init for OUTPUT_TYPE_PURE_CALLINFO Messages with additional source code location. By Tobias Hilbrich, Date 29.10.2005.
     * Rank is determined by the rank of the message creating process.
     * @param MsgType see MPO_MsgObject::initMsgObject.
     * @param Timestamp see MPO_MsgObject::initMsgObject.
     * @param Callinfo see MPO_MsgObject::initMsgObject.
     * @param CallID see MPO_MsgObject::initMsgObject.
     * @param ErrorCode see MPO_MsgObject::initMsgObject.
     * @param CodeReference see MPO_MsgObject::initMsgObject.
     */
    void MPO_MsgObject::initMsgObject (MPO_MSG_TYPE MsgType, long Timestamp, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode, MPO_SourceLocation CodeReference)
    {
        if (m_GlobalStaticRank < -1)
            initGlobalStaticRank();

        initMsgObject (MsgType, OUTPUT_TYPE_PURE_CALLINFO, Timestamp, m_GlobalStaticRank , "", Callinfo , CallID, ErrorCode, CodeReference);
    }
	
    /**
     * Init for OUTPUT_TYPE_TEXT_AND_CALLINFO Messages. By Tobias Hilbrich, Date 29.10.2005.
     * Rank is determined by the rank of the message creating process.
     * @param MsgType see MPO_MsgObject::initMsgObject.
     * @param Timestamp see MPO_MsgObject::initMsgObject.
     * @param Text see MPO_MsgObject::initMsgObject.
     * @param Callinfo see MPO_MsgObject::initMsgObject.
     * @param CallID see MPO_MsgObject::initMsgObject.
     * @param ErrorCode see MPO_MsgObject::initMsgObject.
     */
    void MPO_MsgObject::initMsgObject (MPO_MSG_TYPE MsgType, long Timestamp, string Text, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode)
    {
        if (m_GlobalStaticRank < -1)
            initGlobalStaticRank();

        initMsgObject (MsgType, OUTPUT_TYPE_TEXT_AND_CALLINFO, Timestamp, m_GlobalStaticRank, Text, Callinfo , CallID, ErrorCode, m_CodeReference);
    }
	
    /**
     * Init for OUTPUT_TYPE_TEXT_AND_CALLINFO Messages  with additional source code location. By Tobias Hilbrich, Date 29.10.2005.
     * Rank is determined by the rank of the message creating process.
     * @param MsgType see MPO_MsgObject::initMsgObject.
     * @param Timestamp see MPO_MsgObject::initMsgObject.
     * @param Text see MPO_MsgObject::initMsgObject.
     * @param Callinfo see MPO_MsgObject::initMsgObject.
     * @param CallID see MPO_MsgObject::initMsgObject.
     * @param ErrorCode see MPO_MsgObject::initMsgObject.
     * @param CodeReference see MPO_MsgObject::initMsgObject.
     */
    void MPO_MsgObject::initMsgObject (MPO_MSG_TYPE MsgType, long Timestamp, string Text, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode, MPO_SourceLocation CodeReference)
    {
        if (m_GlobalStaticRank < -1)
            initGlobalStaticRank();

        initMsgObject (MsgType, OUTPUT_TYPE_TEXT_AND_CALLINFO, Timestamp, m_GlobalStaticRank, Text, Callinfo , CallID, ErrorCode, CodeReference);
    }
		
    /**
     * Standard destructor. By Tobias Hilbrich, Date 29.10.2005.
     */
    MPO_MsgObject::~MPO_MsgObject ()
    {
        //Nothing to be done here
    }
		
    /**
     * Init Method called by the different constructors. By Tobias Hilbrich, Date 29.10.2005.
     * @param MsgType the type of the Message see MPO_MSG_TYPE.
     * @param OutputType the content of the Message see MPO_OUTPUT_TYPE.
     * @param Timestamp globaly unique timestamp.
     * @param rank rank of the process, negative values (best -1) are used for MPI-Wide events like general Marmot Infos and Deadlocks.
     * @param Text message text i.e. "invalid tag, tag was < 0".
     * @param Callinfo text describing a call i.e. "MPI_Send(buf*,....)"
     * @param CallID id of the Call see MPO_CALL_ID.
     * @param ErrorCode error code that was produced see MARMOT_ERR_CODES.
     * @param CodeReference information about the source location of the message creating call.
     */
    void MPO_MsgObject::initMsgObject (MPO_MSG_TYPE MsgType, MPO_OUTPUT_TYPE OutputType, long Timestamp, int rank, string Text, string Callinfo, MPO_CALL_ID CallID, MARMOT_ERR_CODES ErrorCode, MPO_SourceLocation CodeReference)
    {
        //save all the passed values
        m_MsgType            = MsgType;
        m_rank               = rank;
        m_OutputType         = OutputType;
        m_Callinfo           = Callinfo;
        m_Text               = Text;
        m_CallID             = CallID;
        m_ErrorCode          = ErrorCode;
        m_Timestamp          = Timestamp;
        m_CodeReference      = CodeReference;
        m_ThreadRank         = MPO_Space::getSyncObject()->getThreadRank();
    }
		
    /**
     * Returns the type of this message. By Tobias Hilbrich, Date 29.10.2005.
     * @return the type of this message object.
     */
    MPO_MSG_TYPE MPO_MsgObject::getMsgType (void) const
    {
        return (m_MsgType);
    }
	
    /**
     * Returns the timestamp. By Tobias Hilbrich, Date 29.10.2005.
     * @return a long value represantation of the timestamp associated with this message.
     */
    long MPO_MsgObject::getTimestamp (void) const
    {
        return(m_Timestamp);
    }
	
    /**
     * Returns the rank. By Tobias Hilbrich, Date 29.10.2005.
     * @return the rank of the process that created this message.
     */
    int MPO_MsgObject::getRank (void) const
    {
        return(m_rank);
    }
	
    /**
     * Returns the output type. By Tobias Hilbrich, Date 29.10.2005.
     * @return the output type of this message object.
     */
    MPO_OUTPUT_TYPE MPO_MsgObject::getOutputType (void) const
    {
        return(m_OutputType);
    }
	
    /**
     * Returns the source location when present. By Tobias Hilbrich, Date 29.10.2005.
     * @return the sourcelocation information,
     */
    const MPO_SourceLocation& MPO_MsgObject::getSourceLocation (void) const
    {
        return (m_CodeReference);
    }
	
    /**
     * Returns the associated text. By Tobias Hilbrich, Date 29.10.2005.
     * @return the message text or "" when not present.
     */
    const string& MPO_MsgObject::getText (void) const
    {
        return (m_Text);
    }
	
    /**
     * Returns the associated callinfo. By Tobias Hilbrich, Date 29.10.2005.
     * @return the text describing the call or "" when not present.
     */
    const string& MPO_MsgObject::getCallinfo (void) const
    {
        return (m_Callinfo);
    }
	
    /**
     * Returns whether information for creating a reference to the MPI-Standard is present or not. By Tobias Hilbrich, Date 29.10.2005.
     * @return true when the errorcode of this message enables the creation of a reference to the mpi standard false otherwise.
     */
    bool MPO_MsgObject::hasMPIStandardReferenceInformation (void) const
    {
        return(  
            (m_CallID != MPO_LAST_CALL_ID) &&
            (m_ErrorCode != MARMOT_LAST_ERR_CODE) && 
            (m_ErrorCode != MARMOT_ERR_UNDEFINED) &&
            (m_ErrorCode != MARMOT_SUCCESS) &&
            (m_ErrorCode != MARMOT_ERR_MARMOT)        );
    }
	
    /**
     * Returns the refernce data. By Tobias Hilbrich, Date 29.10.2005.
     * @param pCallID pointer to a value that gets the id of the call associated with this message.
     * @param pErrorCode pointer to a value that gets the errorcode of this message.
     * @return true when see MPO_MsgObject::hasMPIStandardReferenceInformation would return true false otherwise.
     */
    bool MPO_MsgObject::getMPIStandardReferenceInformation (MPO_CALL_ID *pCallID, MARMOT_ERR_CODES *pErrorCode) const
    {
        if (!hasMPIStandardReferenceInformation())
            return false;
      
        if (pCallID)
            *pCallID = m_CallID;
      
        if (pErrorCode)
            *pErrorCode = m_ErrorCode;
      
        return(true);
    }
   
    /**
     * Returns a string that descibres the Messagetype (like "Error" or "Information"). By Tobias Hilbrich, Date 29.10.2005.
     * @return a string describing the message type in words.
     */
    string MPO_MsgObject::getMessageTypeString (void) const
    {
        switch (m_MsgType) 
        {
            case MSG_TYPE_INFO:
                return "Information";
            case MSG_TYPE_WARNING:
                return "Warning";
            case MSG_TYPE_NOTE:
                return "Note";
            case MSG_TYPE_ERROR:
                return "Error";
        }
      
        return ("");
    }
   
    /**
     * Returns a string that describes the Outputtype (like "Texutal information" or "Call information"). By Tobias Hilbrich, Date 29.10.2005.
     * @return a string decribing what conntents the message has in words.
     */
    string MPO_MsgObject::getOutputTypeString (void) const
    {
        switch (m_OutputType) 
        {
            case OUTPUT_TYPE_TEXT_AND_CALLINFO:
                return "Textual and MPI-Call describing information";
            case OUTPUT_TYPE_PURE_TEXT:
                return "Textual information";
            case OUTPUT_TYPE_PURE_CALLINFO:
                return "MPI-Call describing information";
        }
      
        return ("");
    }
   
    /**
     * Returns the ID of the Call assigned to this Object. By Tobias Hilbrich, Date 25.11.2005.
     * @return the callid associated with this object when the object does not contains any calldescription the return value is undefined.
     */
    MPO_CALL_ID MPO_MsgObject::getCallId (void)
    {
        return m_CallID;
    }
   
    /**
     * Returns the Errorcode assigned to this Object. By Tobias Hilbrich, Date 26.11.2005.
     * @return the errorcode associated with this object when the object does not contains any errorcode the return value is undefined.
     */
    MARMOT_ERR_CODES MPO_MsgObject::getErrorCode (void)
    {
        return m_ErrorCode;
    }
   
    /**
     * Sets the timestamp to the given value might be used to modify the timestamp after sending it with MPO_LogToServer.
     * By Tobias Hilbrich, Date 28.11.2005.
     * @param time the new timestamp as longvalue.
     * @return always true.
     */
    bool MPO_MsgObject::setTimestamp (long time)
    {
        m_Timestamp = time;
        return true;
    }
    /**
     * Initializes the static rank of the message creating process by using a MPI-Call.
     * By Tobias Hilbrich, Date 06.01.2006.
     */
    void MPO_MsgObject::initGlobalStaticRank()
    {
        PMPI_Comm_rank(MPI_COMM_WORLD,&m_GlobalStaticRank);
       
        int size;
        PMPI_Comm_size(MPI_COMM_WORLD,&size);
       
        //special case for Debugserver
        if (m_GlobalStaticRank == size-1)
            m_GlobalStaticRank = -1;
       
    }
		
    /**
     * see MPO_Object::getFamilyId By Tobias Hilbrich, Date 29.10.2005.
     */
    int MPO_MsgObject::getFamilyId() const
    {
        return (MPO_MESSAGEOBJECT_ID);
    }
   
    /**
     * see MPO_Object::getId By Tobias Hilbrich, Date 29.10.2005.
     */
    int MPO_MsgObject::getId() const
    {
        return(MPO_STANDARDMESSAGE);
    }
   
    /**
     * see MPO_Object::getCallname By Tobias Hilbrich, Date 29.10.2005.
     */
    string MPO_MsgObject::getCallname() const
    {
        return("MarmotMessageObject");
    }

    /**
     * see MPO_Object::debugout By Tobias Hilbrich, Date 29.10.2005.
     */
    std::ostream & MPO_MsgObject::debugout(std::ostream & output) const
    {

        string sMsgType = getMessageTypeString();
//      string sOutputType = getOutputTypeString();

        output << m_Timestamp
               << ": "
               << sMsgType;
             
        if (m_rank != -1)
            output  << " from rank " 
                    << m_rank
                    << "(Thread: "
                    << m_ThreadRank
                    << ")";
        else 
            output << " global message";
       
        output << " with ";
        //           << sOutputType;

        if (  (m_OutputType == OUTPUT_TYPE_TEXT_AND_CALLINFO) ||
              (m_OutputType == OUTPUT_TYPE_PURE_TEXT)            )
            output << " Text: "
                   << m_Text;

        if (  (m_OutputType == OUTPUT_TYPE_TEXT_AND_CALLINFO) ||
              (m_OutputType == OUTPUT_TYPE_PURE_CALLINFO)            ) {
         
            if (m_OutputType == OUTPUT_TYPE_TEXT_AND_CALLINFO)
                output << std::endl << "On Call: ";
         
            if (m_OutputType == OUTPUT_TYPE_PURE_CALLINFO)
                output << " Call: ";
               
            output << m_Callinfo;
        }
         
        if (m_CodeReference.isValidLocation())
        {
            output << " From: ";
            m_CodeReference.debugout(output);
        }

        return (output);
    }

    int MPO_MsgObject::writeToSocket(CTcpSocket *socketptr, const char *mpi_reference_path) const
    {
      if (socketptr == NULL)
        return -1;

      const char *delim="\xF8";
      char hstr[32];
      string mes, text;
      // Timestamp
      NetUtil::itoa(m_Timestamp, hstr, 10);
      mes = hstr;
      mes.append(delim);
      // Rank
      NetUtil::itoa(m_rank, hstr, 10);
      mes.append(hstr);
      mes.append(delim);
      // Threadrank
      NetUtil::itoa(m_ThreadRank, hstr, 10);
      mes.append(hstr);
      mes.append(delim);
      // MsgType
      NetUtil::itoa(m_MsgType, hstr, 10);
      mes.append(hstr);
      mes.append(delim);
      // Messagetext
      text = m_Text;
      if (!text.empty())
      {
        // replace "\n" with "<br>"
        for (string::size_type i=0, pos; (pos = text.find("\n", i)) != string::npos; i=pos + 1)
        {
          text.replace(pos, 1, "<br>");
        }
      }
      mes.append(text);
      mes.append(delim);
      // Location (File<delim>Line)
      if (m_CodeReference.isValidLocation())
      {
        mes.append(m_CodeReference.getFileName());
        mes.append(delim);
        NetUtil::itoa(m_CodeReference.getCodeline(), hstr, 10);
        mes.append(hstr);
        mes.append(delim);
      } else {
        // no location
        mes.append(delim);
        mes.append(delim);
      }
      
	    // CallID (the name of the MPI call if present)
      // remove newlines
      // TODO: try to avoid newlines in m_Callinfo!
      text = m_Callinfo;
      for (string::size_type i=0, pos; (pos = text.find("\n", i)) != string::npos; i=pos + 1)
      {
        text.replace(pos, 1, "");
      }
	    mes.append(text);
	    mes.append(delim);
	  
	    // ShortDesc (obtained from DebugServer using Code)
	    if ( (m_ErrorCode >= 0) &&
		     (m_ErrorCode < MARMOT_LAST_ERR_CODE) &&
		     (m_ErrorCode != MARMOT_SUCCESS) &&    /*Do not report silly error names*/
		     (m_ErrorCode != MARMOT_ERR_UNDEFINED))
	    {
		    std::stringstream stream;
		    DebugServer::outputErrorMessages (m_ErrorCode,stream);
        MPO_String temp;
        temp.assign(&stream);

		    if (!temp.empty())
		    {
		      // replace "\n" with "<br>"
		      for (string::size_type i=0, pos; (pos = temp.find("\n", i)) != string::npos; i=pos + 1)
		      {
		        temp.replace(pos, 1, "<br>");
		      }
		    }

		    mes.append(temp);
	    } else {
	      mes.append("Unknown");
	    }
      mes.append(delim);

      // MPI-Standard-Ref.
      if (mpi_reference_path)
	      mes.append(mpi_reference_path);
	    mes.append(delim);      

      // append terminating <CR><LF>
      mes.append("\r\n");
      // send data
      return socketptr->write((void *)mes.c_str(), mes.length());
    }

    /**
     * see MPO_Object::save By Tobias Hilbrich, Date 29.10.2005.
     */
    std::ostream & MPO_MsgObject::save(std::ostream & out) const
    {
        out < (unsigned int)m_OutputType;
        out < (unsigned int)m_MsgType;
        out < (unsigned int)m_CallID;
        out < (unsigned int)m_ErrorCode;
        out < m_Timestamp;
        out < m_rank;
        out < m_ThreadRank;
        out < m_CodeReference;
        out < m_Callinfo;
        out < m_Text;
      
        return (out);
    }

    /**
     * see MPO_Object::restore By Tobias Hilbrich, Date 29.10.2005.
     */
    std::istream & MPO_MsgObject::restore(std::istream & in)
    {
        unsigned int temp;
        in > temp;
        m_OutputType = (MPO_OUTPUT_TYPE)temp;
        in > temp;
        m_MsgType = (MPO_MSG_TYPE)temp;
        in > temp;
        m_CallID = (MPO_CALL_ID)temp;
        in > temp;
        m_ErrorCode = (MARMOT_ERR_CODES)temp;
        in > m_Timestamp;
        in > m_rank;
        in > m_ThreadRank;
        in > m_CodeReference;
        in > m_Callinfo;
        in > m_Text;
      
        return (in);
    }
    
   /**
    * Returns the rank of the thread that created this message.
    * @return the rank of the thread that created this message.
    */ 
   int MPO_MsgObject::getThreadRank(void)
   {
   	   return m_ThreadRank; 
   }
   
   /**
    * Sets the rank for the thread that created this message.
    * @param threadRank the id that will be set as thread rank.
    */ 
   void MPO_MsgObject::setThreadRank(int threadRank)
   {
   	   m_ThreadRank = threadRank;
   }
   
   /**
    * Appends the given text to the message.
    * @param append text to append.
    */
   void MPO_MsgObject::appendToText (std::string append)
   {
       m_Text.append(append);
   }

}//end of namespace MPO








