/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Op_create.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Op_create.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Op_create::MPO_Op_create()
        : my_function(0), 
          my_commute(-1),
          my_op(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Op_create::MPO_Op_create(MPI_User_function * function,
                                 int commute,
                                 MPI_Op * op)
        : my_function(function),
          my_commute(commute),
          my_op(op)
    {
    }


    // destructor
    //========================================================================
    MPO_Op_create::~MPO_Op_create()
    {
    }


    // insertOp
    //========================================================================
    int MPO_Op_create::insertOp()
    {
        our_ops.insert(*my_op);

        return 1;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Op_create::checkPrerequisites()
    {
        
        // currently no checks implemented    

        return 0;
    }


    // getId
    //========================================================================
    int MPO_Op_create::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_OP_CREATE_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Op_create::checkPost(void)
    {
    	insertOp();
		our_op_traces->addAndSetCreation(*my_op,getSourceLocation());
        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Op_create::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
		rvalue = PMPI_Op_create(my_function, my_commute, my_op);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Op_create::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Op_create";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_OP_CREATE";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Op_create::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output 
            << getCallname()
            << "(*function, "
            << "commute = " 
            << my_commute 
            << ", *op";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"  
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Op_create::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_function; 
        out < my_commute;
        // out < my_op; 

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Op_create::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_function; 
        in > my_commute;
        // in > my_op; 

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/coll/opcreate.c.
/*@
  MPI_Op_create - Creates a user-defined combination function handle

  Input Parameters:
  + function - user defined function (function) 
  - commute -  true if commutative;  false otherwise. 

  Output Parameter:
  . op - operation (handle) 

  Notes on the user function:
  The calling list for the user function type is
  .vb
  typedef void (MPI_User_function) ( void * a, 
  void * b, int * len, MPI_Datatype * ); 
  .ve
  where the operation is 'b[i] = a[i] op b[i]', for 'i=0,...,len-1'.  
  A pointer to the datatype given to the MPI collective computation routine
  (i.e., 'MPI_Reduce', 'MPI_Allreduce', 'MPI_Scan', or 
  'MPI_Reduce_scatter') 
  is also passed to the user-specified routine.

  .N fortran

  .N collops

  .N Errors
  .N MPI_SUCCESS
  .N MPI_ERR_EXHAUSTED

  .seealso: MPI_Op_free

  int MPI_Op_create( 
  MPI_User_function *function, 
  int commute, 
  MPI_Op *op )
  @ */

