/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Op_free.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Op_free.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Op_free::MPO_Op_free()
        : my_op(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Op_free::MPO_Op_free(MPI_Op * op)
        : my_op(op)
    {
    }


    // destructor
    //========================================================================
    MPO_Op_free::~MPO_Op_free()
    {
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Op_free::checkPrerequisites()
    {


        // check correctness of parameters
        checkOp();

        // erase op from our bookkeeping
        our_ops.erase(*my_op);
        our_op_traces->findAndSetFree(*my_op,getSourceLocation());

        return 0;
    }


    // checkOp
    //========================================================================
    int MPO_Op_free::checkOp()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;
    if(!our_errcodes.at(MARMOT_ERR_OP_NOT_VALID))
        return rvalue;

        // Check if op exists.
        if ((our_ops.find(*my_op) == our_ops.end()) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": Operation to free doesn't exist (or is already freed)!"
                << std::endl;
            rvalue = MARMOT_ERR_OP_NOT_VALID;

            if (my_op)
                our_op_traces->findAndPrintTraceNamed(*my_op,stream,"op");

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }                           // end checkOp


    // getId
    //========================================================================
    int MPO_Op_free::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_OP_FREE_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Op_free::checkPost(void)
    {
        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Op_free::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Op_free(my_op);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Op_free::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Op_free";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_OP_FREE";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Op_free::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(*op";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Op_free::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_op;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Op_free::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_op;

        return in;
    }

}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/coll/mpoOpFree.c.
/*@
  MPI_Op_free - Frees a user-defined combination function handle

  Input Parameter:
  . op - operation (handle)

  Notes:
  'op' is set to 'MPI_OP_NULL' on exit.

  .N NULL

  .N fortran

  .N Errors
  .N MPI_SUCCESS
  .N MPI_ERR_ARG
  .N MPI_ERR_PERM_OP

  .seealso: MPI_Op_create

  int MPI_Op_free( MPI_Op *op )
  @ */

