/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Rsend.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Rsend.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Rsend::MPO_Rsend()
        : MPO_Blocking(-1,
                       MPI_DATATYPE_NULL,
                       MPI_PROC_NULL,
                       MPI_ANY_TAG,
                       MPI_COMM_NULL),
          my_buf(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Rsend::MPO_Rsend(void *buf,
                         int count,
                         MPO_Datatype datatype,
                         int dest,
                         int tag,
                         MPO_Comm comm)
        : MPO_Blocking(count,
                       datatype,
                       dest,
                       tag,
                       comm),
          my_buf(buf)
    {
    }


    // destructor
    //========================================================================
    MPO_Rsend::~MPO_Rsend()
    {
    }


    // getId
    //========================================================================
    int MPO_Rsend::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_RSEND_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Rsend::checkPost(void)
    {

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Rsend::checkPrerequisites(void)
    {


        MPO_Blocking::checkPrerequisites();
        checkBufferInUse(my_buf,my_count,my_datatype);

        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());
        our_datatype_traces->findAndAddUsage(my_datatype,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Rsend::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Rsend(my_buf, my_count, my_datatype, my_rank, my_tag,
                          my_comm);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Rsend::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Rsend";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_RSEND";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Rsend::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(*buf, "
            << "count = "
            << my_count
            << ", datatype = "
            << my_datatype
            << ", dest = "
            << my_rank
            << ", tag = "
            << my_tag
            << ", comm = "
            << my_comm;

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Rsend::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_buf;
        out < my_count;
        out < my_datatype;
        out < my_rank;
        out < my_tag;
        out < my_comm;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Rsend::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_buf;
        in > my_count;
        in > my_datatype;
        in > my_rank;
        in > my_tag;
        in > my_comm;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/pt2pt/rsend.c.
/*@
  MPI_Rsend - Basic ready send

  Input Parameters:
  + buf - initial address of send buffer (choice)
  . count - number of elements in send buffer (nonnegative integer)
  . datatype - datatype of each send buffer element (handle)
  . dest - rank of destination (integer)
  . tag - message tag (integer)
  - comm - communicator (handle)

  .N fortran

  .N Errors
  .N MPI_SUCCESS
  .N MPI_ERR_COMM
  .N MPI_ERR_COUNT
  .N MPI_ERR_TYPE
  .N MPI_ERR_TAG
  .N MPI_ERR_RANK

  int MPI_Rsend(
  void *buf,
  int count,
  MPI_Datatype datatype,
  int dest,
  int tag,
  MPI_Comm comm )
  @ */

