/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Scatter.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Scatter.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Scatter::MPO_Scatter()
        : MPO_Collective(MPI_COMM_NULL),
          my_sendbuf(0),
          my_sendcount(-1),
          my_sendtype(MPI_DATATYPE_NULL),
          my_recvbuf(0),
          my_recvcount(-1),
          my_recvtype(MPI_DATATYPE_NULL),
          my_root(MPI_PROC_NULL)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Scatter::MPO_Scatter(void *sendbuf,
                             int sendcount,
                             MPO_Datatype sendtype,
                             void *recvbuf,
                             int recvcount,
                             MPO_Datatype recvtype,
                             int root,
                             MPO_Comm comm)
        : MPO_Collective(comm),
          my_sendbuf(sendbuf),
          my_sendcount(sendcount),
          my_sendtype(sendtype),
          my_recvbuf(recvbuf),
          my_recvcount(recvcount),
          my_recvtype(recvtype),
          my_root(root)
    {
    }


    // destructor
    //========================================================================
    MPO_Scatter::~MPO_Scatter()
    {
    }


    // checkPrerequisites2
    //========================================================================
    int MPO_Scatter::checkPrerequisites2()
    {
        // check correctness of parameters
        // some checks are only for root
        int rank = -1;
        PMPI_Comm_rank(my_comm, &rank);

        if (rank == my_root)
        {
            checkCount(my_sendcount, MSG_SENDCOUNT);
            checkDatatypeSendRecv(my_sendcount, my_sendtype,
                                  IS_SENDCOUNT, IS_NOT_REDUCE_ALLREDUCE);
        }

        checkCount(my_recvcount, MSG_RECVCOUNT);
        checkDatatypeSendRecv(my_recvcount, my_recvtype,
                              IS_RECVCOUNT, IS_NOT_REDUCE_ALLREDUCE);
        checkRoot(my_root);

        //check for buffer already in use
        checkBufferInUse(my_recvbuf,my_recvcount,my_recvtype);
        if (rank == my_root)
            checkBufferInUse(my_sendbuf,my_sendcount,my_sendtype);

        return 0;
    }


    // getId
    //========================================================================
    int MPO_Scatter::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_SCATTER_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Scatter::checkPost(void)
    {

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Scatter::checkPrerequisites(void)
    {


        MPO_Collective::checkPrerequisites();
        checkPrerequisites2();

        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());
        our_datatype_traces->findAndAddUsage(my_sendtype,getSourceLocation());
        our_datatype_traces->findAndAddUsage(my_recvtype,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Scatter::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Scatter(my_sendbuf, my_sendcount, my_sendtype, my_recvbuf,
                            my_recvcount, my_recvtype, my_root, my_comm);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Scatter::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Scatter";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_SCATTER";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Scatter::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(comm = "
            << my_comm
            << ", *coords, "
            << "*rank";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Scatter::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_sendbuf;
        out < my_sendcount;
        out < my_sendtype;
        // out < my_recvbuf;
        out < my_recvcount;
        out < my_recvtype;
        out < my_root;
        out < my_comm;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Scatter::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_sendbuf;
        in > my_sendcount;
        in > my_sendtype;
        // in > my_recvbuf;
        in > my_recvcount;
        in > my_recvtype;
        in > my_root;
        in > my_comm;

        return in;
    }

}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/coll/scatter.c
/*@

MPI_Scatter - Sends data from one task to all other tasks in a group

Input Parameters:
+ sendbuf - address of send buffer (choice, significant
only at 'root')
. sendcount - number of elements sent to each process
(integer, significant only at 'root')
. sendtype - data type of send buffer elements (significant only at
'root') (handle)
. recvcount - number of elements in receive buffer (integer)
. recvtype - data type of receive buffer elements (handle)
. root - rank of sending process (integer)
- comm - communicator (handle)

Output Parameter:
. recvbuf - address of receive buffer (choice)

.N fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_COUNT
.N MPI_ERR_TYPE
.N MPI_ERR_BUFFER

int MPI_Scatter (
void *sendbuf,
int sendcnt,
MPI_Datatype sendtype,
void *recvbuf,
int recvcnt,
MPI_Datatype recvtype,
int root,
MPI_Comm comm )
@ */

