/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Scatterv.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Scatterv.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Scatterv::MPO_Scatterv()
        : MPO_Collective(MPI_COMM_NULL),
          my_sendbuf(0),
          my_sendcounts(0),
          my_displs(0),
          my_sendtype(MPI_DATATYPE_NULL),
          my_recvbuf(0),
          my_recvcount(-1),
          my_recvtype(MPI_DATATYPE_NULL),
          my_root(MPI_PROC_NULL)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Scatterv::MPO_Scatterv(void *sendbuf,
                               int *sendcounts,
                               int *displs,
                               MPO_Datatype sendtype,
                               void *recvbuf,
                               int recvcount,
                               MPO_Datatype recvtype,
                               int root,
                               MPO_Comm comm)
        : MPO_Collective(comm),
          my_sendbuf(sendbuf),
          my_sendcounts(sendcounts),
          my_displs(displs),
          my_sendtype(sendtype),
          my_recvbuf(recvbuf),
          my_recvcount(recvcount),
          my_recvtype(recvtype),
          my_root(root)
    {
    }


    // destructor
    //========================================================================
    MPO_Scatterv::~MPO_Scatterv()
    {
    }


    // checkPrerequisites2
    //========================================================================
    int MPO_Scatterv::checkPrerequisites2()
    {
        int rank = -1;

        PMPI_Comm_rank(my_comm, &rank);

        // check correctness of parameters
        // some checks are only for the root
        if (rank == my_root)
        {
            checkDisplacementArrayOfSender();
            checkDatatypeArrayOfSender();
        }

        checkCount(my_recvcount, MSG_RECVCOUNT);
        checkCountArray(my_sendcounts, MSG_SENDCOUNT);
        checkDatatypeSendRecv(my_recvcount, my_recvtype,
                              IS_RECVCOUNT, IS_NOT_REDUCE_ALLREDUCE);
        checkRoot(my_root);

        //verify that buffers are not used yet
        checkBufferInUse(my_recvbuf,my_recvcount,my_recvtype);
        int size;
        PMPI_Comm_size(my_comm,&size);
        if (rank == my_root)
        {
            for (int i = 0; i < size;i++)
                checkBufferInUse(my_sendbuf,my_displs[i]+my_sendcounts[i],my_sendtype);
        }

        return 0;
    }


    // checkDisplacementArrayOfSender
    //========================================================================
    int MPO_Scatterv::checkDisplacementArrayOfSender()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;
    if(!(our_errcodes.at(MARMOT_NOTE_DISPLACEMENT_DATA_SENT_TWICE)||our_errcodes.at(MARMOT_NOTE_DISPLACEMENT_READS_DATA)))
        return rvalue;

        int size = -1;
        PMPI_Comm_size(my_comm, &size);

        int rank = -1;
        PMPI_Comm_rank(my_comm, &rank);

        // compilation error:  MPI_Aint typeextent may be int or long, so
        // use instead of
        //     int used = 0;
        // an MPI_Aint and cast typesize to MPI_Aint in the if condition
        // below.
        MPI_Aint used = 0;

        MPI_Aint typeextent = 0;
        PMPI_Type_extent(my_sendtype, &typeextent);

        int typesize = -1;
        PMPI_Type_size(my_sendtype, &typesize);

        // (1) Here we have a compact datatype without holes.
        if (typeextent == (MPI_Aint) typesize)
        {
            // Check array of displacements.
            for (int i = 0; i < size; i++)
            {
                if ((our_errcodes.at(MARMOT_NOTE_DISPLACEMENT_DATA_SENT_TWICE))&&(my_displs[i] < used) &&
                    (MSG_ERR_WARN_NOTE <= DEBUG_MODE))
                {
                    std::stringstream stream;
                    stream
                        << "REMARK: "
                        << getCallname()
                        << ": displs["
                        << i
                        << "] = "
                        << my_displs[i]
                        << "  (used = "
                        << used
                        << ") sends data already sent!"
                        << std::endl;
                    rvalue = MARMOT_NOTE_DISPLACEMENT_DATA_SENT_TWICE;

                    our_datatype_traces->findAndPrintTraceNamed(my_sendtype,stream,"sendtype");
                    our_datatype_traces->findAndPrintTraceNamed(my_recvtype,stream,"recvtype");
                    our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");

                    MPO_String text;
                    text.assign(&stream);
                    MPO_MsgObject msg (MSG_TYPE_NOTE,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                    our_log->getMsg(msg);
                }
            }
        }

        // (2) Here we have a datatype with holes.
        else
        {
            // Check array of displacements.
            for (int i = 0; i < size; i++)
            {
                if ((our_errcodes.at(MARMOT_NOTE_DISPLACEMENT_READS_DATA))&&(my_displs[i] < used) &&
                    (MSG_ERR_WARN_NOTE <= DEBUG_MODE))
                {
                    std::stringstream stream;
                    stream
                        << "REMARK: "
                        << getCallname()
                        << ": Datatype has holes, "
                        << "so we don't check whether it's correct,"
                        << "but on rank "
                        << rank
                        << ": displs["
                        << i
                        << "] reads data/ holes read from other nodes!"
                        << std::endl;
                    rvalue = MARMOT_NOTE_DISPLACEMENT_READS_DATA;

                    our_datatype_traces->findAndPrintTraceNamed(my_sendtype,stream,"sendtype");
                    our_datatype_traces->findAndPrintTraceNamed(my_recvtype,stream,"recvtype");
                    our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");

                    MPO_String text;
                    text.assign(&stream);
                    MPO_MsgObject msg (MSG_TYPE_NOTE,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                    our_log->getMsg(msg);
                }
                used = my_displs[i] + (typeextent * my_sendcounts[i]);
            }
        }

        return rvalue;
    }                           // end checkDisplacementArrayOfSender


    // checkDatatypeArrayOfSender
    //========================================================================
    int MPO_Scatterv::checkDatatypeArrayOfSender()
    {
        /** \todo @creator of this method: what about return values ? */

        if (my_sendtype == MPI_DATATYPE_NULL)
        {
            if (MSG_ERR <= DEBUG_MODE)
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname()
                    << ": Sendtype is MPI_DATATYPE_NULL!"
                    << std::endl;

                our_datatype_traces->findAndPrintTraceNamed(my_sendtype,stream,"sendtype");
                our_datatype_traces->findAndPrintTraceNamed(my_recvtype,stream,"recvtype");
                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_NULL,getSourceLocation());
                our_log->getMsg(msg);
            }
        }
        else
        {
            if (our_datatypes.find(my_sendtype) == our_datatypes.end())
            {
                if (MSG_ERR <= DEBUG_MODE)
                {
                    std::stringstream stream;
                    stream
                        << "ERROR: "
                        << getCallname()
                        << ": Sendtype is not valid!"
                        << std::endl;

                    our_datatype_traces->findAndPrintTraceNamed(my_sendtype,stream,"sendtype");
                    our_datatype_traces->findAndPrintTraceNamed(my_recvtype,stream,"recvtype");
                    our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");

                    MPO_String text;
                    text.assign(&stream);
                    MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_NOT_VALID,getSourceLocation());
                    our_log->getMsg(msg);
                }
            }
            else if ((our_datatypes.find(my_sendtype)->second ==
                      DATATYPE_CREATED_NOT_COMMITTED) &&
                     (MSG_ERR <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname()
                    << ": Sendtype is not committed!"
                        << std::endl;

                our_datatype_traces->findAndPrintTraceNamed(my_sendtype,stream,"sendtype");
                our_datatype_traces->findAndPrintTraceNamed(my_recvtype,stream,"recvtype");
                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_NOT_COMMITTED,getSourceLocation());
                our_log->getMsg(msg);
            }
        }

        return 1;
    }                           // end checkDatatypeArrayOfSender


    // getId
    //========================================================================
    int MPO_Scatterv::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_SCATTERV_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Scatterv::checkPost(void)
    {

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Scatterv::checkPrerequisites(void)
    {


        MPO_Collective::checkPrerequisites();
        checkPrerequisites2();

        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());
        our_datatype_traces->findAndAddUsage(my_sendtype,getSourceLocation());
        our_datatype_traces->findAndAddUsage(my_recvtype,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Scatterv::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Scatterv(my_sendbuf, my_sendcounts, my_displs, my_sendtype,
                             my_recvbuf, my_recvcount, my_recvtype, my_root,
                             my_comm);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Scatterv::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Scatterv";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_SCATTERV";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Scatterv::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(*sendbuf, "
            << "*sendcounts, "
            << "*displs, "
            << "sendtype = "
            << my_sendtype
            << ", *recvbuf, "
            << "recvcount = "
            << my_recvcount
            << ", recvtype = "
            << my_recvtype
            << ", root = "
            << my_root
            << ", comm = "
            << my_comm;

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Scatterv::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_sendbuf;
        // out < my_sendcounts;
        // out < my_displs;
        out < my_sendtype;
        // out < my_recvbuf;
        out < my_recvcount;
        out < my_recvtype;
        out < my_root;
        out < my_comm;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Scatterv::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_sendbuf;
        // in > my_sendcounts;
        // in > my_displs;
        in > my_sendtype;
        // in > my_recvbuf;
        in > my_recvcount;
        in > my_recvtype;
        in > my_root;
        in > my_comm;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/coll/scatterv.c.
/*@

MPI_Scatterv - Scatters a buffer in parts to all tasks in a group

Input Parameters:
+ sendbuf - address of send buffer (choice, significant only at 'root')
. sendcounts - integer array (of length group size)
specifying the number of elements to send to each processor
. displs - integer array (of length group size). Entry
'i'  specifies the displacement (relative to sendbuf  from
which to take the outgoing data to process  'i'
. sendtype - data type of send buffer elements (handle)
. recvcount - number of elements in receive buffer (integer)
. recvtype - data type of receive buffer elements (handle)
. root - rank of sending process (integer)
- comm - communicator (handle)

Output Parameter:
. recvbuf - address of receive buffer (choice)

.N fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_COUNT
.N MPI_ERR_TYPE
.N MPI_ERR_BUFFER

int MPI_Scatterv (
void *sendbuf,
int *sendcnts,
int *displs,
MPI_Datatype sendtype,
void *recvbuf,
int recvcnt,
MPI_Datatype recvtype,
int root,
MPI_Comm comm )
@ */

