/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Sendrecv.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Sendrecv.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Sendrecv::MPO_Sendrecv()
        : MPO_Sendrecvs(-1,
                        MPI_DATATYPE_NULL,
                        MPI_PROC_NULL,
                        MPI_ANY_TAG,
                        MPI_PROC_NULL,
                        MPI_ANY_TAG,
                        MPI_COMM_NULL),
          my_sendbuf(0),
          my_recvbuf(0),
          my_recvcount(-1),
          my_recvtype(MPI_DATATYPE_NULL),
          my_status(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Sendrecv::MPO_Sendrecv(void *sendbuf,
                               int sendcount,
                               MPO_Datatype datatype,
                               int dest,
                               int sendtag,
                               void *recvbuf,
                               int recvcount,
                               MPO_Datatype recvtype,
                               int source,
                               int recvtag,
                               MPO_Comm comm,
                               MPI_Status * status)
        : MPO_Sendrecvs(sendcount,
                        datatype,
                        dest,
                        sendtag,
                        source,
                        recvtag,
                        comm),
          my_sendbuf(sendbuf),
          my_recvbuf(recvbuf),
          my_recvcount(recvcount),
          my_recvtype(recvtype),
          my_status(status)
    {
    }


    // destructor
    //========================================================================
    MPO_Sendrecv::~MPO_Sendrecv()
    {
    }


    // checkPrerequisites2
    //========================================================================
    int MPO_Sendrecv::checkPrerequisites2()
    {
        checkCountOfReceiver(MSG_RECVCOUNT);
        checkDatatypeOfReceiver();

        return 0;
    }


    // checkCountOfReceiver
    //========================================================================
    int MPO_Sendrecv::checkCountOfReceiver(const string KIND_OF_COUNT)
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        // (1) Check if count is negative.
        if ((our_errcodes.at(MARMOT_ERR_COUNT_NEG))&&(my_recvcount < 0) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname()
                << ": "
                << KIND_OF_COUNT
                << " = "
                << my_recvcount
                << " less than 0 !"
                << std::endl;
            rvalue = MARMOT_ERR_COUNT_NEG;

            our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"sendtype");
            our_datatype_traces->findAndPrintTraceNamed(my_recvtype,stream,"recvtype");
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        // (2) Check if count is 0.
        if ((our_errcodes.at(MARMOT_WARN_COUNT_ZERO))&&(my_recvcount == 0) &&
            (MSG_ERR_WARN <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "WARNING: "
                << getCallname()
                << ": "
                << KIND_OF_COUNT
                << " = 0!"
                << std::endl;
            rvalue = MARMOT_WARN_COUNT_ZERO;

            our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"sendtype");
            our_datatype_traces->findAndPrintTraceNamed(my_recvtype,stream,"recvtype");
            our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");

            MPO_String text;
            text.assign(&stream);
            MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
            our_log->getMsg(msg);
        }

        return rvalue;
    }


    // checkDatatypeofReceiver
    //========================================================================
    int  MPO_Sendrecv::checkDatatypeOfReceiver()
    {

        if (my_recvtype == MPI_DATATYPE_NULL)
        {
            if (MSG_ERR <= DEBUG_MODE)
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname()
                    << ": recvtype is MPI_DATATYPE_NULL!"
                    << std::endl;

                our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"sendtype");
                our_datatype_traces->findAndPrintTraceNamed(my_recvtype,stream,"recvtype");
                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_NULL,getSourceLocation());
                our_log->getMsg(msg);
            }
        }
        else
        {
            if (our_datatypes.find(my_recvtype) == our_datatypes.end())
            {
                if (MSG_ERR <= DEBUG_MODE)
                {
                    std::stringstream stream;
                    stream
                        << "ERROR: "
                        << getCallname()
                        << ": recvtype is not valid!"
                        << std::endl;

                    our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"sendtype");
                    our_datatype_traces->findAndPrintTraceNamed(my_recvtype,stream,"recvtype");
                    our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");

                    MPO_String text;
                    text.assign(&stream);
                    MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_NOT_VALID,getSourceLocation());
                    our_log->getMsg(msg);
                }
            }
            else if ((our_datatypes.find(my_recvtype)->second ==
                      DATATYPE_CREATED_NOT_COMMITTED) &&
                     (MSG_ERR <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname()
                    << ": recvtype is not committed!"
                    << std::endl;

                our_datatype_traces->findAndPrintTraceNamed(my_datatype,stream,"sendtype");
                our_datatype_traces->findAndPrintTraceNamed(my_recvtype,stream,"recvtype");
                our_comm_traces->findAndPrintTraceNamed(my_comm,stream,"comm");

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),MARMOT_ERR_TYPE_NOT_COMMITTED,getSourceLocation());
                our_log->getMsg(msg);
            }
        }
        return 0;
    }


    // getId
    //========================================================================
    int MPO_Sendrecv::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_SENDRECV_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Sendrecv::checkPost(void)
    {

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Sendrecv::checkPrerequisites(void)
    {


        MPO_Sendrecvs::checkPrerequisites();
        checkPrerequisites2();
        checkBufferInUse(my_sendbuf,my_count,my_datatype);
        checkBufferInUse(my_recvbuf,my_recvcount,my_recvtype);

        our_datatype_traces->findAndAddUsage(my_datatype,getSourceLocation());
        our_datatype_traces->findAndAddUsage(my_recvtype,getSourceLocation());
        our_comm_traces->findAndAddUsage(my_comm,getSourceLocation());

        #ifdef MARMOT_USE_THREADSYNC
            testIfProbed (my_source, my_recvtag, my_comm, getSourceLocation().getCodeline(), getSourceLocation().getFileName(), MPO_Space::getSyncObject()->getThreadRank(), getId(), getCallname());
        #endif


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Sendrecv::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Sendrecv(my_sendbuf, my_count, my_datatype, my_dest,
                             my_sendtag, my_recvbuf, my_recvcount, my_recvtype,
                             my_source, my_recvtag, my_comm, my_status);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Sendrecv::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Sendrecv";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_SENDRECV";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Sendrecv::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(*sendbuf, "
            << "sendcount = "
            << my_count
            << ", sendtype = "
            << my_datatype
            << ", dest = "
            << my_dest
            << ", sendtag = "
            << my_sendtag
            << ", *recvbuf, "
            << "recvcount = "
            << my_recvcount
            << ", recvtype = "
            << my_recvtype
            << ", source = "
            << my_source
            << ", recvtag = "
            << my_recvtag
            << ", comm = "
            << my_comm
            << ", *status";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Sendrecv::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_sendbuf;
        out < my_count;
        out < my_datatype;
        out < my_dest;
        out < my_sendtag;
        // out < my_recvbuf;
        out < my_recvcount;
        out < my_recvtype;
        out < my_source;
        out < my_recvtag;
        out < my_comm;
        // out < my_status;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Sendrecv::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_sendbuf;
        in > my_count;
        in > my_datatype;
        in > my_dest;
        in > my_sendtag;
        // in > my_recvbuf;
        in > my_recvcount;
        in > my_recvtype;
        in > my_source;
        in > my_recvtag;
        in > my_comm;
        // in > my_status;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/pt2pt/sendrecv.c.
/*@
  MPI_Sendrecv - Sends and receives a message

  Input Parameters:
  + sendbuf - initial address of send buffer (choice)
  . sendcount - number of elements in send buffer (integer)
  . sendtype - type of elements in send buffer (handle)
  . dest - rank of destination (integer)
  . sendtag - send tag (integer)
  . recvcount - number of elements in receive buffer (integer)
  . recvtype - type of elements in receive buffer (handle)
  . source - rank of source (integer)
  . recvtag - receive tag (integer)
  - comm - communicator (handle)

  Output Parameters:
  + recvbuf - initial address of receive buffer (choice)
  - status - status object (Status).  This refers to the receive operation.

  .N fortran

  .N Errors
  .N MPI_SUCCESS
  .N MPI_ERR_COMM
  .N MPI_ERR_COUNT
  .N MPI_ERR_TYPE
  .N MPI_ERR_TAG
  .N MPI_ERR_RANK

  int MPI_Sendrecv(
  void *sendbuf,
  int sendcount,
  MPI_Datatype sendtype,
  int dest,
  int sendtag,
  void *recvbuf,
  int recvcount,
  MPI_Datatype recvtype,
  int source,
  int recvtag,
  MPI_Comm comm,
  MPI_Status *status )
  @ */


