/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Test.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Test.cc 964 2009-06-02 06:38:44Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Test::MPO_Test()
        : MPO_Tests(0,
                    0,
                    0),
          my_status(0)
    {
    }


    /** Constructor taking proper arguments.
     *
     *  The  constructor performs the following tasks:
     *  1. Check whether request is in our_requests
     *  - no: print warning to stderr
     *  - yes:
     *  2. Remove request from our_requests if test was successful
     *
     *  We store the pointer and not the request itself.
     *  Because the request may change during the message transfer
     *  (it is a opaque type and thus  may contain some internal status etc.)
     */
    // proper constructor
    //========================================================================
    MPO_Test::MPO_Test(MPI_Request * request,
                       int *flag,
                       MPI_Status * status)
        : MPO_Tests(1,
                    request,
                    flag),
          my_status(status)
    {
        if (!handlePreDataConstructor ())
        {
            my_temp_request = *request;
        }
    }


    // destructor
    //========================================================================
    MPO_Test::~MPO_Test()
    {
        if (!handlePreDataDestructor ())
        {
            //Nothing to delete here
        }
    }

    // setPreData
    //========================================================================
    void MPO_Test::setPreData (MPO_Callobject *obj)
    {
        my_temp_request = ((MPO_Test*)obj)->my_temp_request;
    }

    // clonePreData
    //========================================================================
    MPO_Callobject* MPO_Test::clonePreData (void)
    {
        MPO_Test *obj = new MPO_Test();
        obj->my_temp_request = my_temp_request;
        return obj;
    }

    // getId
    //========================================================================
    int MPO_Test::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_TEST_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Test::checkPost(void)
    {
        int my_outcount[1], my_index_array[1];

		/*TOBIAS: moved erase request into the if, we should only remove it if
                  the test was successfull. */
        if (*my_flags)
        {
        	my_outcount[0] = 1;
        	my_index_array[0] = 0;
        	eraseRequest(my_outcount, my_index_array, &my_temp_request);

            our_request_traces->findAndSetFree(&my_temp_request,getSourceLocation());
            unregisterBuffer(&my_temp_request);
        }

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Test::checkPrerequisites(void)
    {

        MPO_Tests::checkPrerequisites();

        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Test::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Test(my_request_array, my_flags, my_status);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Test::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Test";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_TEST";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Test::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(*request, "
            << "*flag, "
            << "*status";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Test::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_request_p;
        // out < my_flag;
        // out < my_status;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Test::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_request_p;
        // in > my_flag;
        // in > my_status;

        return in;
    }


}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/pt2pt/test.c.
/*@
  MPI_Test  - Tests for the completion of a send or receive

  Input Parameter:
  . request - communication request (handle)

  Output Parameter:
  + flag - true if operation completed (logical)
  - status - status object (Status).  May be 'MPI_STATUS_NULL'.

  .N waitstatus

  .N fortran

  .N Errors
  .N MPI_SUCCESS
  .N MPI_ERR_REQUEST
  .N MPI_ERR_ARG

  int MPI_Test (
  MPI_Request  *request,
  int          *flag,
  MPI_Status   *status)
  @ */

