/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Type_commit.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Type_commit.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Type_commit::MPO_Type_commit()
        : my_datatype(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Type_commit::MPO_Type_commit(MPI_Datatype* datatype)
        : my_datatype(datatype)
    {
    }


    // destructor
    //========================================================================
    MPO_Type_commit::~MPO_Type_commit()
    {
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Type_commit::checkPrerequisites()
    {

// TEST errorcodes

        // check correctness of parameters
        checkDatatypeCommitted();
    
// end TEST errorcodes

        
        return 0;
    }


// TEST errorcodes
    // checkPrerequisites
    //========================================================================
//     int MPO_Type_commit::checkPrerequisites()
//     {
//         // check correctness of parameters
//         return checkDatatypeCommitted();
//     }
// end TEST errorcodes


     // checkDatatypeCommitted
     //========================================================================
    int MPO_Type_commit::checkDatatypeCommitted()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
        int rvalue = MARMOT_SUCCESS;

        std::map < MPO_Datatype, enum KIND_OF_DATATYPE >::iterator pos;
        pos = our_datatypes.find(*my_datatype);

        // (1) Datatype is not valid, i.e. not registered in
        // our MPO datatypes.
        if ((our_errcodes.at(MARMOT_ERR_TYPE_NOT_VALID))&&(pos == our_datatypes.end()) &&
            (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            stream
                << "ERROR: "
                << getCallname() 
                << ": datatype is not valid!" 
                << std::endl;
            rvalue = MARMOT_ERR_TYPE_NOT_VALID;
            
            if (my_datatype)
                our_datatype_traces->findAndPrintTraceNamed(*my_datatype,stream,"datatype");
            
            MPO_String text;
            text.assign(&stream);      
            MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
            our_log->getMsg(msg);   
        }

        // (2) Datatype is registered in our MPO datatypes.
        // In this case, we have to check different violations 
        // in the switch statement.
        else if ((pos != our_datatypes.end()) &&
                 (MSG_ERR <= DEBUG_MODE))
        {
            std::stringstream stream;
            MPO_MSG_TYPE type = MSG_TYPE_ERROR;
            
            // all the cases from enum KIND_OF_DATATYPE
            switch (pos->second)
            {
                case (DATATYPE_NULL):
                    if ((our_errcodes.at(MARMOT_NOTE_TYPE_NULL))&&(MSG_ERR_WARN_NOTE <= DEBUG_MODE))
                    {
                        stream
                            << "NOTE: "
                            << getCallname() 
                            << ": datatype is MPI_DATATYPE_NULL!" 
                            << std::endl;
                        rvalue = MARMOT_NOTE_TYPE_NULL;
                        type = MSG_TYPE_NOTE;
                    }
                    break;
                case (DATATYPE_CREATED_NOT_COMMITTED):
                    // Do nothing, because everything is fine.
                    break;
                case (DATATYPE_COMMITTED_C):
                    if ((our_errcodes.at(MARMOT_NOTE_TYPE_ALREADY_COMMITTED_C))&&(MSG_ERR_WARN_NOTE <= DEBUG_MODE))
                    {
                        stream
                            << "NOTE: "
                            << getCallname() 
                            << ": Datatype already committed!" 
                            << std::endl;
                        rvalue = MARMOT_NOTE_TYPE_ALREADY_COMMITTED_C;
                        type = MSG_TYPE_NOTE;
                    }
                    break;
                case (DATATYPE_COMMITTED_C_AND_FORTRAN):
                    if ((our_errcodes.at(MARMOT_NOTE_TYPE_ALREADY_COMMITTED_C_AND_F))&&(MSG_ERR_WARN_NOTE <= DEBUG_MODE))
                    {
                        stream
                            << "NOTE: "
                            << getCallname() 
                            << ": Datatype already committed!" 
                            << std::endl;
                        rvalue = MARMOT_NOTE_TYPE_ALREADY_COMMITTED_C_AND_F;
                        type = MSG_TYPE_NOTE;
                    }                    
                    break;
                case (DATATYPE_COMMITTED_FORTRAN):
                    if ((our_errcodes.at(MARMOT_NOTE_TYPE_ALREADY_COMMITTED_F))&&(MSG_ERR_WARN_NOTE <= DEBUG_MODE))
                    {
                        stream
                            << "NOTE: "
                            << getCallname() 
                            << ": Datatype already committed!" 
                            << std::endl;
                        rvalue = MARMOT_NOTE_TYPE_ALREADY_COMMITTED_F;
                        type = MSG_TYPE_NOTE;
                    }                    
                    break;
                case (DATATYPE_REDUCTION_C):
                    // Do nothing.
                    break;
                case (DATATYPE_REDUCTION_FORTRAN):
                    // Do nothing.
                    break;
                case (DATATYPE_OPTIONAL_C):
                    // Do nothing.
                    break;
                case (DATATYPE_OPTIONAL_FORTRAN):
                    // Do nothing.
                    break;
                case (DATATYPE_CONSTRUCT_DERIVED_TYPE):
                    // Do nothing.
                    break;
                default:
            if(our_errcodes.at(MARMOT_ERR_MARMOT))
            {
                        stream
                        << "Check mapping of datatypes."
                        << std::endl
                        << "Check implementation of switch statement."
                        << std::endl;
                        rvalue = MARMOT_ERR_MARMOT;
            }
                    break;
            }                   // end switch
            
            if (rvalue != MARMOT_SUCCESS)
            {
                if (my_datatype)
                    our_datatype_traces->findAndPrintTraceNamed(*my_datatype,stream,"datatype");
                
                MPO_String text;
                text.assign(&stream);      
                MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation()); 
                our_log->getMsg(msg);     
            }
        }                       // end else if (2)

        return rvalue;
    }                           // end checkDatatypeCommitted


// TEST errorcodes
    // checkDatatypeCommitted
    //========================================================================
//     int MPO_Type_commit::checkDatatypeCommitted()
//     {
//         // Return value is the unique error code set by enum MARMOT_ERR_CODES. 
//         int rvalue = MARMOT_SUCCESS;

//         std::map < MPO_Datatype, enum KIND_OF_DATATYPE >::iterator pos;
//         pos = our_datatypes.find(*my_datatype);

//         // (1) Datatype is not valid, i.e. not registered in
//         // our MPO datatypes.
//         if ((pos == our_datatypes.end()) &&
//             (MSG_ERR <= DEBUG_MODE))
//         {
//             rvalue = MARMOT_ERR_TYPE_NOT_VALID;
//         }

//         // (2) Datatype is registered in our MPO datatypes.
//         // In this case, we have to check different violations 
//         // in the switch statement.
//         else if ((pos != our_datatypes.end()) &&
//                  (MSG_ERR <= DEBUG_MODE))
//         {
//             // all the cases from enum KIND_OF_DATATYPE
//             switch (pos->second)
//             {
//                 case (DATATYPE_NULL):
//                     if (MSG_ERR_WARN_NOTE <= DEBUG_MODE)
//                     {
//                         rvalue = MARMOT_NOTE_TYPE_NULL;
//                     }
//                     break;
//                 case (DATATYPE_CREATED_NOT_COMMITTED):
//                     // Do nothing, because everything is fine.
//                     break;
//                 case (DATATYPE_COMMITTED_C):
//                     if (MSG_ERR_WARN_NOTE <= DEBUG_MODE)
//                     {
//                         rvalue = MARMOT_NOTE_TYPE_ALREADY_COMMITTED_C;
//                     }
//                     break;
//                 case (DATATYPE_COMMITTED_C_AND_FORTRAN):
//                     if (MSG_ERR_WARN_NOTE <= DEBUG_MODE)
//                     {
//                         rvalue = MARMOT_NOTE_TYPE_ALREADY_COMMITTED_C_AND_F;
//                     }                    
//                     break;
//                 case (DATATYPE_COMMITTED_FORTRAN):
//                     if (MSG_ERR_WARN_NOTE <= DEBUG_MODE)
//                     {
//                         rvalue = MARMOT_NOTE_TYPE_ALREADY_COMMITTED_F;
//                     }                    
//                     break;
//                 case (DATATYPE_REDUCTION_C):
//                     // Do nothing.
//                     break;
//                 case (DATATYPE_REDUCTION_FORTRAN):
//                     // Do nothing.
//                     break;
//                 case (DATATYPE_OPTIONAL_C):
//                     // Do nothing.
//                     break;
//                 case (DATATYPE_OPTIONAL_FORTRAN):
//                     // Do nothing.
//                     break;
//                 case (DATATYPE_CONSTRUCT_DERIVED_TYPE):
//                     // Do nothing.
//                     break;
//                 default:
//                     rvalue = MARMOT_ERR_MARMOT;
//                     break;
//             }                   // end switch
//         }                       // end else if (2)

//         return rvalue;
//     }                           // end checkDatatypeCommitted
// 
// end TEST errorcodes


    // getId
    //========================================================================
    int MPO_Type_commit::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_TYPE_COMMIT_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Type_commit::checkPost(void)
    {
        
        if (C_INTERFACE == INTERFACE_MODE)
        {
            our_datatypes.find(*my_datatype)->second = 
                DATATYPE_COMMITTED_C;
        }
        else
        {
            our_datatypes.find(*my_datatype)->second = 
                DATATYPE_COMMITTED_FORTRAN;
        }
        
        our_datatype_traces->findAndSetCommit(*my_datatype,getSourceLocation());
            
        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Type_commit::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Type_commit(my_datatype);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Type_commit::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Type_commit";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_TYPE_COMMIT";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Type_commit::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output 
            << getCallname() 
            << "(*datatype";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"  
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Type_commit::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_datatype;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Type_commit::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_datatype;
        
        return in;
    }

}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied
// from mpich-1.2.4/src/pt2pt/type_commit.c.
/*@
  MPI_Type_commit - Commits the datatype

  Input Parameter:
  . datatype - datatype (handle) 

  .N fortran

  .N Errors
  .N MPI_SUCCESS
  .N MPI_ERR_TYPE

  int MPI_Type_commit ( MPI_Datatype *datatype )
  @ */


