/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Type_contiguous.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Type_contiguous.cc 909 2008-12-17 09:48:19Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"

namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Type_contiguous::MPO_Type_contiguous()
        : MPO_Typeconstructors(-1,
                               MPI_DATATYPE_NULL,
                               -1),
          my_newtype(0)
    {
    }


    // proper constructor
    //========================================================================
    MPO_Type_contiguous::MPO_Type_contiguous(int count,
                                             MPO_Datatype oldtype,
                                             MPI_Datatype *
                                             newtype)
        : MPO_Typeconstructors(count,
                               oldtype,
                               MPO_TYPE_CONTIGUOUS_ID),
          my_newtype(newtype)
    {
    }


    // destructor
    //========================================================================
    MPO_Type_contiguous::~MPO_Type_contiguous()
    {
    }


    // getId
    //========================================================================
    int MPO_Type_contiguous::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_TYPE_CONTIGUOUS_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Type_contiguous::checkPost(void)
    {

        our_datatypes.insert(std::make_pair(MPO_Datatype(*my_newtype),
                                            DATATYPE_CREATED_NOT_COMMITTED));

        our_datatype_traces->addAndSetCreation(*my_newtype,getSourceLocation());
        our_datatype_traces->findAndTryAddDependendency(*my_newtype,my_oldtype);

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Type_contiguous::checkPrerequisites(void)
    {


        MPO_Typeconstructors::checkPrerequisites();

        our_datatype_traces->findAndAddUsage(my_oldtype,getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Type_contiguous::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Type_contiguous(my_count, my_oldtype, my_newtype);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Type_contiguous::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Type_contiguous";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_TYPE_CONTIGUOUS";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Type_contiguous::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(count = "
            << my_count
            << ", oldtype = "
            << my_oldtype
            << ", *newtype";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Type_contiguous::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_count;
        out < my_oldtype;
        // out < my_newtype;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Type_contiguous::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_count;
        in > my_oldtype;
        // in > my_newtype;

        return in;
    }

}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied
// from mpich-1.2.4/src/pt2pt/type_contig.c.
/*@
  MPI_Type_contiguous - Creates a contiguous datatype

  Input Parameters:
  + count - replication count (nonnegative integer)
  - oldtype - old datatype (handle)

  Output Parameter:
  . newtype - new datatype (handle)

  .N fortran

  .N Errors
  .N MPI_SUCCESS
  .N MPI_ERR_TYPE
  .N MPI_ERR_COUNT
  .N MPI_ERR_EXHAUSTED

  int MPI_Type_contiguous(
  int count,
  MPI_Datatype old_type,
  MPI_Datatype *newtype)
  @ */


