/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of class MPO_Type_struct.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Type_struct.cc 1077 2010-02-10 09:53:09Z hpczink $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"
#include <stdio.h>


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Type_struct::MPO_Type_struct()
        : MPO_Typeconstructors(-1,
                               MPI_DATATYPE_NULL,
                               -1),
          my_blocklength(0),
          my_displ_array(0),
          // compilation error:
          // my_oldtypes(MPI_DATATYPE_NULL),
          my_oldtypes(0),
          my_newtype(0)
    {
    }


    /** Constructor taking proper arguments.
     *  We treat oldtype locally because it's a pointer:
     */
    // proper constructor
    //========================================================================
    MPO_Type_struct::MPO_Type_struct(int count,
                                     int *blocklength,
                                     MPI_Aint * displ_array,
                                     MPI_Datatype * oldtype,
                                     MPI_Datatype *
                                     newtype)
        : MPO_Typeconstructors(count,
                               MPI_DATATYPE_NULL,
                               MPO_TYPE_STRUCT_ID),
          my_blocklength(blocklength),
          my_displ_array(displ_array),
          my_oldtypes(oldtype),
          my_newtype(newtype)
    {
    }


    // destructor
    //========================================================================
    MPO_Type_struct::~MPO_Type_struct()
    {
    }


    // checkPrerequisites2
    //========================================================================
    int MPO_Type_struct::checkPrerequisites2()
    {
        // check correctness of parameters
        checkBlocklengthArray();
        checkDatatypeArrayTypeStruct();

        return 0;
    }


    // checkBlocklenghtArray
    //========================================================================
    int MPO_Type_struct::checkBlocklengthArray()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

    if(our_errcodes.at(MARMOT_ERR_BLOCKLENGTH_NEG)||our_errcodes.at(MARMOT_WARN_BLOCKLENGTH_ZERO))
    {
        for (int i = 0; i < my_count; i++)
        {
            // (1) Check if blocklength is negative.
            if ((our_errcodes.at(MARMOT_ERR_BLOCKLENGTH_NEG))&&(my_blocklength[i] < 0) &&
                (MSG_ERR <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname()
                    << ": blocklength["
                    << i
                    << "] less than 0!"
                    << std::endl;
                rvalue = MARMOT_ERR_BLOCKLENGTH_NEG;

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                our_log->getMsg(msg);
            }

            // (2) Check if blocklength is 0.
            else if ((our_errcodes.at(MARMOT_WARN_BLOCKLENGTH_ZERO))&&(my_blocklength[i] == 0) &&
                     (MSG_ERR_WARN <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream
                    << "WARNING: "
                    << getCallname()
                    << ": blocklength["
                    << i
                    << "] = 0!"
                    << std::endl;
                rvalue = MARMOT_WARN_BLOCKLENGTH_ZERO;

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_WARNING,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                our_log->getMsg(msg);
            }
        }                       // end for
    }

        return rvalue;
    }


    // checkDatatypeArrayTypeStruct
    //========================================================================
    int MPO_Type_struct::checkDatatypeArrayTypeStruct()
    {
        // Return value is the unique error code set by enum MARMOT_ERR_CODES.
        int rvalue = MARMOT_SUCCESS;

        std::map < MPO_Datatype, enum KIND_OF_DATATYPE >::iterator pos;

        // Check array of datatypes
        for (int i = 0; i < my_count; i++)
        {
            pos = our_datatypes.find(my_oldtypes[i]);

            // (1) Datatype is not valid, i.e. not registered in
            // our MPO datatypes.
            if ((our_errcodes.at(MARMOT_ERR_TYPE_NOT_VALID))&&(pos == our_datatypes.end()) &&
                (MSG_ERR <= DEBUG_MODE))
            {
                std::stringstream stream;
                stream
                    << "ERROR: "
                    << getCallname()
                    << ": datatype["
                    << i
                    << "] is not valid!"
                    << std::endl;
                rvalue = MARMOT_ERR_TYPE_NOT_VALID;

                std::string s ("oldtype[");
                char buf[64];
                sprintf (buf,"%d]",i);
                s.append(buf);
                our_datatype_traces->findAndPrintTraceNamed(my_oldtypes[i],stream,s);

                MPO_String text;
                text.assign(&stream);
                MPO_MsgObject msg (MSG_TYPE_ERROR,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                our_log->getMsg(msg);
            }

            // (2) Datatype is registered in our MPO datatypes.
            // In this case, we have to check different violations
            // in the switch statement.
            else if ((pos != our_datatypes.end()) &&
                     (MSG_ERR <= DEBUG_MODE))
            {

                std::stringstream stream;
                MPO_MSG_TYPE type = MSG_TYPE_ERROR;

                // all the cases from enum KIND_OF_DATATYPE
                switch (pos->second)
                {
                    case (DATATYPE_NULL):
                        if(our_errcodes.at(MARMOT_ERR_TYPE_NULL))
                        {
                            stream
                            << "ERROR: "
                            << getCallname()
                            << ": datatype["
                            << i
                            << "] is MPI_DATATYPE_NULL!"
                            << std::endl;
                            rvalue = MARMOT_ERR_TYPE_NULL;
                        }
                        break;

                    case (DATATYPE_CREATED_NOT_COMMITTED):
/*                        if(our_errcodes.at(MARMOT_ERR_TYPE_NOT_COMMITTED))
                        {
                            stream
                            << "ERROR: "
                            << getCallname()
                            << ": datatype["
                            << i
                            << "] is not committed!"
                            << std::endl;
                            rvalue = MARMOT_ERR_TYPE_NOT_COMMITTED;
                        }
*/
                        break;

                    case (DATATYPE_COMMITTED_C):
                        if ((our_errcodes.at(MARMOT_ERR_TYPE_C))&&(F_INTERFACE == INTERFACE_MODE))
                        {
                            stream
                                << "ERROR: "
                                << getCallname()
                                << ": datatype["
                                << i
                                << "] is C-Type!"
                                << std::endl;
                            rvalue = MARMOT_ERR_TYPE_C;
                        }
                        break;
                    case (DATATYPE_COMMITTED_C_AND_FORTRAN):
                        // Do nothing, because everything is fine.
                        break;
                    case (DATATYPE_COMMITTED_FORTRAN):
                        if ((our_errcodes.at(MARMOT_ERR_TYPE_F))&&(C_INTERFACE == INTERFACE_MODE))
                        {
                            stream
                                << "ERROR: "
                                << getCallname()
                                << ": datatype ["
                                << i
                                << "] is Fortran-Type!"
                                << std::endl;
                            rvalue = MARMOT_ERR_TYPE_F;
                        }
                        break;
                    case (DATATYPE_REDUCTION_C):
                        if ((our_errcodes.at(MARMOT_WARN_TYPE_REDUCTION_C))&&(MSG_ERR_WARN <= DEBUG_MODE) &&
                            (C_INTERFACE == INTERFACE_MODE))
                        {
                            stream
                                << "WARNING: "
                                << getCallname()
                                << ": datatype ["
                                << i
                                << "] is for reduction functions (C)!"
                                << std::endl;
                            rvalue = MARMOT_WARN_TYPE_REDUCTION_C;
                            type = MSG_TYPE_WARNING;
                        }
                        if ((our_errcodes.at(MARMOT_ERR_TYPE_REDUCTION_C))&&(F_INTERFACE == INTERFACE_MODE))
                        {
                            stream
                                << "ERROR: "
                                << getCallname()
                                << ": datatype ["
                                << i
                                << "] is for reduction functions (C)!"
                                << std::endl;
                            rvalue = MARMOT_ERR_TYPE_REDUCTION_C;
                        }
                        break;
                    case (DATATYPE_REDUCTION_FORTRAN):
                        if ((our_errcodes.at(MARMOT_ERR_TYPE_REDUCTION_F))&&(C_INTERFACE == INTERFACE_MODE))
                        {
                            stream
                                << "ERROR: "
                                << getCallname()
                                << ": datatype ["
                                << i
                                << "] is for reduction functions "
                                << "(Fortran)!"
                                << std::endl;
                            rvalue = MARMOT_ERR_TYPE_REDUCTION_F;
                        }
                        if ((our_errcodes.at(MARMOT_WARN_TYPE_REDUCTION_F))&&(MSG_ERR_WARN <= DEBUG_MODE) &&
                            (F_INTERFACE == INTERFACE_MODE))
                        {
                            stream
                                << "WARNING: "
                                << getCallname()
                                << ": datatype ["
                                << i
                                << "] is for reduction functions "
                                << "(Fortran)!"
                                << std::endl;
                            rvalue = MARMOT_WARN_TYPE_REDUCTION_F;
                            type = MSG_TYPE_WARNING;
                        }
                        break;
                    case (DATATYPE_OPTIONAL_C):
                        if ((our_errcodes.at(MARMOT_WARN_TYPE_OPTIONAL_C))&&(MSG_ERR_WARN <= DEBUG_MODE) &&
                            (C_INTERFACE == INTERFACE_MODE))
                        {
                            stream
                                << "WARNING: "
                                << getCallname()
                                << ": datatype ["
                                << i
                                << "] is only optional!"
                                << std::endl;
                            rvalue = MARMOT_WARN_TYPE_OPTIONAL_C;
                            type = MSG_TYPE_WARNING;
                        }
                        if ((our_errcodes.at(MARMOT_ERR_TYPE_OPTIONAL_C))&&(F_INTERFACE == INTERFACE_MODE))
                        {
                            stream
                                << "ERROR: "
                                << getCallname()
                                << ": datatype ["
                                << i
                                << "] is optional C-Type!"
                                << std::endl;
                            rvalue = MARMOT_ERR_TYPE_OPTIONAL_C;
                        }
                        break;
                    case (DATATYPE_OPTIONAL_FORTRAN):
                        if ((our_errcodes.at(MARMOT_ERR_TYPE_OPTIONAL_F))&&(C_INTERFACE == INTERFACE_MODE))
                        {
                            stream
                                << "ERROR: "
                                << getCallname()
                                << ": datatype ["
                                << i
                                << "] is optional Fortran-type!"
                                << std::endl;
                            rvalue = MARMOT_ERR_TYPE_OPTIONAL_F;
                        }
                        if ((our_errcodes.at(MARMOT_WARN_TYPE_OPTIONAL_F))&&(MSG_ERR_WARN <= DEBUG_MODE) &&
                            (F_INTERFACE == INTERFACE_MODE))
                        {
                            stream
                                << "WARNING: "
                                << getCallname()
                                << ": datatype ["
                                << i
                                << "] is only optional!"
                                << std::endl;
                            rvalue = MARMOT_WARN_TYPE_OPTIONAL_F;
                            type = MSG_TYPE_WARNING;
                        }
                        break;
                    case (DATATYPE_CONSTRUCT_DERIVED_TYPE):
                        if ((our_errcodes.at(MARMOT_ERR_TYPE_LB))&&(my_oldtypes[i] == MPI_LB) &&
                            (i != 0) &&
                            (MSG_ERR <= DEBUG_MODE))
                        {
                            stream
                                << "ERROR: "
                                << getCallname()
                                << ": MPI_LB only allowed on entry[0]! "
                                << "Not on ["
                                << i
                                << "] !"
                                << std::endl;
                            rvalue = MARMOT_ERR_TYPE_LB;
                        }
                        else if ((our_errcodes.at(MARMOT_ERR_TYPE_UB))&&(my_oldtypes[i] == MPI_UB) &&
                                 (i != (my_count - 1)) &&
                                 (MSG_ERR <= DEBUG_MODE))
                        {
                            stream
                                << "ERROR: "
                                << getCallname()
                                << ": MPI_UB only allowed on last entry! "
                                << "Not on ["
                                << i
                                << "] !"
                                << std::endl;
                            rvalue = MARMOT_ERR_TYPE_UB;
                        }
                        break;
                    default:
            if(our_errcodes.at(MARMOT_ERR_MARMOT))
            {
                            stream
                            << "Check mapping of datatypes."
                            << std::endl
                            << "Check implementation of switch statement."
                            << std::endl;
                            rvalue = MARMOT_ERR_MARMOT;
            }
                        break;
                }               // end switch

                if (rvalue != MARMOT_SUCCESS)
                {
                    std::string s ("oldtype[");
                    char buf[64];
                    sprintf (buf,"%d]",i);
                    s.append(buf);
                    our_datatype_traces->findAndPrintTraceNamed(my_oldtypes[i],stream,s);

                    MPO_String text;
                    text.assign(&stream);
                    MPO_MsgObject msg (type,0,text, getCallname() ,(MPO_CALL_ID) getId(),(MARMOT_ERR_CODES)rvalue,getSourceLocation());
                    our_log->getMsg(msg);
                }

            }                   // end else if (2)
        }                       // end for

        return rvalue;
    }                           // end checkDatatypeArrayTypeStruct


    // getId
    //========================================================================
    int MPO_Type_struct::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_TYPE_STRUCT_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Type_struct::checkPost(void)
    {

        our_datatypes.insert(std::make_pair(MPO_Datatype(*my_newtype),
                                            DATATYPE_CREATED_NOT_COMMITTED));

        our_datatype_traces->addAndSetCreation(*my_newtype,getSourceLocation());
        for (int i = 0; i < my_count;i++)
            our_datatype_traces->findAndTryAddDependendency(*my_newtype,my_oldtypes[i]);

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Type_struct::checkPrerequisites(void)
    {

        MPO_Typeconstructors::checkPrerequisites();
        checkPrerequisites2();

        for (int i = 0; i < my_count;i++)
            our_datatype_traces->findAndAddUsage(my_oldtypes[i],getSourceLocation());


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Type_struct::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Type_struct(my_count, my_blocklength, my_displ_array,
                                  my_oldtypes, my_newtype);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Type_struct::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Type_struct";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_TYPE_STRUCT";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Type_struct::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(count = "
            << my_count
            << ", *array_of_blocklengths, "
            << "*array_of_displacements, "
            << "*array_of_types, "
            << "*newtype";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Type_struct::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        out < my_count;
        // out < my_blocklength;
        // out < my_displ_array;
        // out < my_oldtypes;
        // out < my_newtype;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Type_struct::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        in > my_count;
        // in > my_blocklength;
        // in > my_displ_array;
        // in > my_oldtypes;
        // in > my_newtype;

        return in;
    }

}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied
// from mpich-1.2.4/src/pt2pt/type_struct.c.
/*@
  MPI_Type_struct - Creates a struct datatype

  Input Parameters:
  + count - number of blocks (integer) -- also number of
  entries in arrays array_of_types ,
  array_of_displacements  and array_of_blocklengths
  . blocklens - number of elements in each block (array)
  . indices - byte displacement of each block (array)
  - old_types - type of elements in each block (array
  of handles to datatype objects)

  Output Parameter:
  . newtype - new datatype (handle)

  Notes:
  If an upperbound is set explicitly by using the MPI datatype 'MPI_UB',
  the corresponding index must be positive.

  The MPI standard originally made vague statements about padding and
  alignment; this was intended to allow the simple definition of structures
  that could be sent with a count greater than one.  For example,
  .vb
  struct { int a; char b; } foo;
  .ve
  may have 'sizeof(foo) > sizeof(int) + sizeof(char)'; for example,
  'sizeof(foo) == 2*sizeof(int)'.  The initial version of the MPI standard
  defined the extent of a datatype as including an `epsilon` that would
  have allowed an implementation to make the extent an MPI datatype
  for this structure equal to '2*sizeof(int)'.
  However, since different systems might define different paddings, there
  was much discussion by the MPI Forum about what was the correct value of
  epsilon, and one suggestion was to define epsilon as zero.
  This would have been the best thing to do in MPI 1.0, particularly since
  the 'MPI_UB' type allows the user to easily set the end of the structure.
  Unfortunately, this change did not make it into the final document.
  Currently, this routine does not add any padding, since the amount of
  padding needed is determined by the compiler that the user is using to
  build their code, not the compiler used to construct the MPI library.
  A later version of MPICH may provide for some natural choices of padding
  (e.g., multiple of the size of the largest basic member), but users are
  advised to never depend on this, even with vendor MPI implementations.
  Instead, if you define a structure datatype and wish to send or receive
  multiple items, you should explicitly include an 'MPI_UB' entry as the
  last member of the structure.  For example, the following code can be
  used for the structure foo
  .vb
  blen[0] = 1; indices[0] = 0; oldtypes[0] = MPI_INT;
  blen[1] = 1; indices[1] = &foo.b - &foo; oldtypes[1] = MPI_CHAR;
  blen[2] = 1; indices[2] = sizeof(foo); oldtypes[2] = MPI_UB;
  MPI_Type_struct( 3, blen, indices, oldtypes, &newtype );
  .ve

  .N fortran

  .N Errors
  .N MPI_SUCCESS
  .N MPI_ERR_TYPE
  .N MPI_ERR_COUNT
  .N MPI_ERR_EXHAUSTED

  int MPI_Type_struct(
  int count,
  int blocklens[],
  MPI_Aint indices[],
  MPI_Datatype old_types[],
  MPI_Datatype *newtype )
  @ */

