/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of MPO_Wait class.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Wait.cc 963 2009-05-27 12:37:15Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Wait::MPO_Wait()
        : MPO_Completion(0,
                         0),
          my_status(0)
    {
    }


    // proper constructor
    //========================================================================
    /** Constructor taking proper arguments.
     *
     *  The  constructor performs the following tasks:
     *  1. Check whether request is in our_requests
     *  - no: print warning to stderr
     *  - yes:
     *  2. Remove request from our_requests
     *
     *  We store the pointer and not the request itself.
     *  Because the request may change during the message transfer
     *  (it is a opaque type and thus  may contain some internal status etc.)
     */
    MPO_Wait::MPO_Wait(MPI_Request * request,
                       MPI_Status * status)
        : MPO_Completion(1,
                         request),
          my_status(status)
    {
    }


    // destructor
    //========================================================================
    MPO_Wait::~MPO_Wait()
    {
    }


    // getId
    //========================================================================
    int MPO_Wait::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_WAIT_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Wait::checkPost(void)
    {

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Wait::checkPrerequisites(void)
    {

        int outcount = 1;
        int index[1];
        index[0] = 0;

        MPO_Completion::checkPrerequisites();
        eraseRequest(&outcount, index, my_request_array);
        our_request_traces->findAndSetFree(my_request_array,getSourceLocation());
        unregisterBuffer(my_request_array);


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Wait::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Wait(my_request_array, my_status);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Wait::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Wait";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_WAIT";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Wait::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(*request, "
            << "*status";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Wait::save(std::ostream & out) const
    {
        // we don't save and restore pointers
        MPO_Callobject::save(out);
        // out < my_request_p;
        // out < my_status;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Wait::restore(std::istream & in)
    {
        // we don't save and restore pointers
        MPO_Callobject::restore(in);
        // in > my_request_p;
        // in > my_status;

        return in;
    }

}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/pt2pt/wait.c.
/*@
  MPI_Wait  - Waits for an MPI send or receive to complete

  Input Parameter:
  . request - request (handle)

  Output Parameter:
  . status - status object (Status) .  May be 'MPI_STATUS_NULL'.

  .N waitstatus

  .N fortran

  .N Errors
  .N MPI_SUCCESS
  .N MPI_ERR_REQUEST
  .N MPI_ERR_ARG

  int MPI_Wait (
  MPI_Request  *request,
  MPI_Status   *status)
  @ */
