/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of MPO_Waitany class.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Waitany.cc 964 2009-06-02 06:38:44Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Waitany::MPO_Waitany()
        : MPO_Completion(-1,
                         0),
          my_index(0),
          my_status_array(0),
          my_temp_request_array(NULL)
    {
    }


    // proper constructor
    //========================================================================
    /** Constructor taking proper arguments.
     *
     *  The  constructor performs the following tasks:
     *  1. Check whether requests are in our_requests
     *  - no: print warning to stderr
     *  - yes:
     *  2. Remove requests from our_request
     *
     */
    MPO_Waitany::MPO_Waitany(int count,
                             MPI_Request * request_array,
                             int *index,
                             MPI_Status * status_array)
        : MPO_Completion(count,
                         request_array),
          my_index(index),
          my_status_array(status_array)
    {
        //copy requests into temporary array
        /**
         * TODO
         * I think this does not works with requests as
         * pointers. Though thats something that we don't
         * use anymore ....
         * (Address will change, thus the issue)
         */
        if (!handlePreDataConstructor ())
        {
            my_temp_request_array = NULL;
            if (count >= 0)
            {
                if (request_array)
                {
                    my_temp_request_array = new MPI_Request[count];
                    for (int i = 0; i < count; i++)
                        my_temp_request_array[i] = request_array[i];
                }
            }
        }
    }


    // destructor
    //========================================================================
    MPO_Waitany::~MPO_Waitany()
    {
        if (!handlePreDataDestructor ())
        {
            if (my_temp_request_array)
                delete my_temp_request_array;
            my_temp_request_array = NULL;
        }
    }

    // setPreData
    //========================================================================
    void MPO_Waitany::setPreData (MPO_Callobject *obj)
    {
        my_temp_request_array = ((MPO_Waitany*)obj)->my_temp_request_array;
        ((MPO_Waitany*)obj)->my_temp_request_array = NULL;
    }

    // clonePreData
    //========================================================================
    MPO_Callobject* MPO_Waitany::clonePreData (void)
    {
        MPO_Waitany *obj = new MPO_Waitany();
        obj->my_temp_request_array = my_temp_request_array;
        return obj;
    }

    // getId
    //========================================================================
    int MPO_Waitany::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_WAITANY_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Waitany::checkPost(void)
    {
		/*
         *  If there are no active handles in the list, index is set to
         *  MPI_UNDEFINED, in MPI_Waitsome outcount is set to MPI_UNDEFINED.
         *  We check outcount, so here we have to set this to MPI_UNDEFINED.
         */

        int outcount[1] = { 0 };
        int index_array[1] = { 0 };

        if (*my_index != MPI_UNDEFINED)
        {
            outcount[0] = 1;
        }
        else
        {
            outcount[0] = MPI_UNDEFINED;
        }
        index_array[0] = *my_index;
        eraseRequest(outcount, index_array, my_temp_request_array);

        if (*my_index != MPI_UNDEFINED)
        {
            our_request_traces->findAndSetFree(&(my_temp_request_array[*my_index]),getSourceLocation());
            unregisterBuffer(&(my_temp_request_array[*my_index]));
        }

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Waitany::checkPrerequisites(void)
    {

        MPO_Completion::checkPrerequisites();

        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Waitany::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Waitany(my_count, my_request_array, my_index,
                              my_status_array);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Waitany::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Waitany";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_WAITANY";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Waitany::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(count = "
            << my_count
            << ", *array_of_requests, "
            << "*index, "
            << "*status";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Waitany::save(std::ostream & out) const
    {
        MPO_Callobject::save(out);
        out < my_count;
        // we don't save and restore pointers
        // out < my_request_array;
        // out < my_index;
        // out < my_status_array;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Waitany::restore(std::istream & in)
    {
        MPO_Callobject::restore(in);
        in > my_count;
        // we don't save and restore pointers
        // in > my_request_array;
        // in > my_index;
        // in > my_status_array;

        return in;
    }

}                               // end of namespace MPO



// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/pt2pt/waitany.c.
/*@
  MPI_Waitany - Waits for any specified send or receive to complete

  Input Parameters:
  + count - list length (integer)
  - array_of_requests - array of requests (array of handles)

  Output Parameters:
  + index - index of handle for operation that completed (integer).  In the
  range '0' to 'count-1'.  In Fortran, the range is '1' to 'count'.
  - status - status object (Status).  May be 'MPI_STATUS_NULL'.

  Notes:
  If all of the requests are 'MPI_REQUEST_NULL', then 'index' is returned
  as 'MPI_UNDEFINED', and 'status' is returned as an empty status.

  .N waitstatus

  .N fortran

  .N Errors
  .N MPI_SUCCESS
  .N MPI_ERR_REQUEST
  .N MPI_ERR_ARG

  int MPI_Waitany(
  int count,
  MPI_Request array_of_requests[],
  int *index,
  MPI_Status *status )

  @ */

