/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Definition of MPO_Waitsome class.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  $Id: mpo_Waitsome.cc 1108 2010-03-23 09:33:56Z tobias $
 */



#include "mpo-objects.h"
#include "mpo-binaryIO.h"
#include "mpo-debugserver.h"
#include "mpo_MsgObject.h"
#include "MPO_LogFile.h"
#include "MPO_ResourceGatherer.h"


namespace MPO
{

    // default constructor
    //========================================================================
    MPO_Waitsome::MPO_Waitsome()
        : MPO_Completion(-1,
                         0),
          my_outcount(0),
          my_index_array(0),
          my_status_array(0),
          my_temp_request_array(NULL)
    {
    }


    // proper constructor
    //========================================================================
    /** Constructor taking proper arguments.
     *
     *  The  constructor performs the following tasks:
     *  1. Check whether requests are in our_requests
     *  - no: print warning to stderr
     *  - yes:
     *  2. Remove requests from our_request
     */
    MPO_Waitsome::MPO_Waitsome(int incount,
                               MPI_Request * request_array,
                               int *outcount,
                               int *index_array,
                               MPI_Status *
                               status_array)
        : MPO_Completion(incount,
                         request_array),
          my_outcount(outcount),
          my_index_array(index_array),
          my_status_array(status_array)
    {
        //copy requests into temporary array
        /**
         * TODO
         * I think this does not works with requests as
         * pointers. Though thats something that we don't
         * use anymore ....
         * (Address will change, thus the issue)
         */
        if (!handlePreDataConstructor ())
        {
            my_temp_request_array = NULL;
            if (incount >= 0)
            {
                if (request_array)
                {
                    my_temp_request_array = new MPI_Request[incount];
                    for (int i = 0; i < incount; i++)
                        my_temp_request_array[i] = request_array[i];
                }
            }
        }
    }


    // destructor
    //========================================================================
    MPO_Waitsome::~MPO_Waitsome()
    {
        if (!handlePreDataDestructor ())
        {
            if (my_temp_request_array)
                delete [] my_temp_request_array;
            my_temp_request_array = NULL;
        }
    }

    // setPreData
    //========================================================================
    void MPO_Waitsome::setPreData (MPO_Callobject *obj)
    {
        my_temp_request_array = ((MPO_Waitsome*)obj)->my_temp_request_array;
        ((MPO_Waitsome*)obj)->my_temp_request_array = NULL;
    }

    // clonePreData
    //========================================================================
    MPO_Callobject* MPO_Waitsome::clonePreData (void)
    {
        MPO_Waitsome *obj = new MPO_Waitsome();
        obj->my_temp_request_array = my_temp_request_array;
        return obj;
    }

    // getId
    //========================================================================
    int MPO_Waitsome::getId() const
    {
        // returns the unique call identity set by enum MPO_CALL_ID
        return MPO_WAITSOME_ID;
    }


    // checkPost
    //========================================================================
    int MPO_Waitsome::checkPost(void)
    {

    	if (*my_outcount != MPI_UNDEFINED)
    	{
			eraseRequest(my_outcount, my_index_array, my_temp_request_array);

			for (int i = 0; i < *my_outcount;i++)
			{
				our_request_traces->findAndSetFree(&(my_temp_request_array[my_index_array[i]]),getSourceLocation());
				unregisterBuffer(&(my_temp_request_array[my_index_array[i]]));
			}
    	}

        return 0;
    }


    // checkPrerequisites
    //========================================================================
    int MPO_Waitsome::checkPrerequisites(void)
    {


        MPO_Completion::checkPrerequisites();


        return 0;
    }


    // callNativeMPI
    //========================================================================
    int MPO_Waitsome::callNativeMPI()
    {
        int rvalue = 1;

        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->enterPMPI();
        #endif
        rvalue = PMPI_Waitsome(my_count, my_request_array, my_outcount,
                               my_index_array, my_status_array);
        #ifdef MARMOT_USE_THREADSYNC
               MPO_Space::getSyncObject()->leavePMPI();
        #endif

        return rvalue;
    }


    // getCallname
    //========================================================================
    string MPO_Waitsome::getCallname() const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        string rvalue = "MPI_Waitsome";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            rvalue = "MPI_WAITSOME";
        }

        return rvalue;
    }


    // debugout
    //========================================================================
    std::ostream & MPO_Waitsome::debugout(std::ostream & output) const
    {
        /*  Distinguish between C and Fortran interface, default: C interface.
         */

        output
            << getCallname()
            << "(incount = "
            << my_count
            << ", *array_of_requests, "
            << "*outcount, "
            << "*array_of_indices, "
            << "*array_of_statuses";

        if (F_INTERFACE == INTERFACE_MODE)
        {
            output << ", ierror";
        }

        output
            << ")"
            << std::endl;

        return output;
    }


    // save
    //========================================================================
    std::ostream & MPO_Waitsome::save(std::ostream & out) const
    {
        MPO_Callobject::save(out);
        out < my_count;
        // we don't save and restore pointers
        // out < my_request_array;
        // out < my_outcount;
        // out < my_index_array;
        // out < my_status_array;

        return out;
    }


    // restore
    //========================================================================
    std::istream & MPO_Waitsome::restore(std::istream & in)
    {
        MPO_Callobject::restore(in);
        in > my_count;
        // we don't save and restore pointers
        // in > my_request_array;
        // in > my_outcount;
        // in > my_index_array;
        // in > my_status_array;

        return in;
    }

}                               // end of namespace MPO


// mpich implementation
//========================================================================
// The following comment is copied from mpich-1.2.4/src/pt2pt/waitsome.c.
/*@
  MPI_Waitsome - Waits for some given communications to complete

  Input Parameters:
  + incount - length of array_of_requests (integer)
  - array_of_requests - array of requests (array of handles)

  Output Parameters:
  + outcount - number of completed requests (integer)
  . array_of_indices - array of indices of operations that
  completed (array of integers)
  - array_of_statuses - array of status objects for
  operations that completed (array of Status).  May be 'MPI_STATUSES_NULL'.

  Notes:
  The array of indicies are in the range '0' to 'incount - 1' for C and
  in the range '1' to 'incount' for Fortran.

  Null requests are ignored; if all requests are null, then the routine
  returns with 'outcount' set to 'MPI_UNDEFINED'.

  .N waitstatus

  .N fortran

  .N Errors
  .N MPI_SUCCESS
  .N MPI_ERR_REQUEST
  .N MPI_ERR_ARG
  .N MPI_ERR_IN_STATUS

  int MPI_Waitsome(
  int incount,
  MPI_Request array_of_requests[],
  int *outcount,
  int array_of_indices[],
  MPI_Status array_of_statuses[] )

  @ */

