/**********************************************************************
 * @file   MarmotComm.cpp
 * @brief  Implementation of CMarmotComm
 * 
 * @author Blasius Czink
 * @date   26.02.2008
 **********************************************************************
 * Version History:
 * V 0.1  26.02.2008  : First Revision
 **********************************************************************/
#include <string>
#include <vector>
#include <sstream>
#include "MarmotComm.h"

#ifdef USE_MARMOT_DICE
#include "ToolWinComm.h"
#include "Win_DiceWindow.h"
#endif

using namespace std;

//
// string helper
//
// split a line into the first word, and rest-of-the-line
string GetWord(string &s, const string delim=" ", const bool trim_spaces=true)
{
  // find delimiter  
  string::size_type i(s.find(delim));
  // split into before and after delimiter
  string w(s.substr(0, i));
  // if no delimiter, remainder is empty
  if (i == string::npos)
  {
    s.erase();
  } else {
    // erase up to the delimiter
    s.erase(0, i+delim.size());
  }
  // return first word in line
  return w;
}

void StringToVector (const string s, vector<string> &v, const string delim=" ")
{
  // start with initial string, trimmed of leading/trailing spaces if required
  string s1(s);
  v.clear();
  if (s1.empty())
    return;
  // add to vector while we have a delimiter
  while (!s1.empty() && s1.find(delim) != string::npos)
    v.push_back(GetWord(s1, delim));
  // add final element
  v.push_back(s1);
}

// ------------------------------------------------------------------
// CMarmotComm -Impl.
// ------------------------------------------------------------------
const char CMarmotComm::MARMOT_MSG_DELIM[] = "\xF8";

CMutex CMarmotComm::m_output_mutex;
CMutex CMarmotComm::m_applaunch_mutex;

CMarmotComm::CMarmotComm(DTE2 *pDTE, OutputWindowPane *pMarmotOutput, OutputWindowPane *pBuildOutput) : 
            m_pDTE(pDTE), m_pMarmotOutput(pMarmotOutput), m_pBuildOutput(pBuildOutput), m_serversocket(NULL), m_channel(NULL),  m_poll_limit(500), m_finished(false), m_initialized(false)
{
}

CMarmotComm::~CMarmotComm()
{
  setRunning(false);
  waitForFinish();
  if (m_channel)
  {
    delete m_channel;  m_channel = NULL;
  }
  if (m_serversocket)
  {
    delete m_serversocket; m_serversocket = NULL;
  }
}

bool CMarmotComm::init()
{
  NetUtil::initialize();
/* ----------------------- getting addin load problems when accessing registry keys. why?? use env.var MARMOT_ADDIN_TIMEOUT instead --------------------
  // try to read timeout from registry
  LONG lRet;
  HKEY hKey, hCurrentUser;
  DWORD dwAddinTimeout=5000;
  DWORD dwBufLen=sizeof(DWORD);
  DWORD dwKeyType=REG_DWORD;
  lRet = RegOpenCurrentUser(KEY_READ, &hCurrentUser);
  if (lRet == ERROR_SUCCESS)
  {
    lRet = RegOpenKeyEx(hCurrentUser, TEXT("SOFTWARE\\HLRS\\Marmot"), 0, KEY_QUERY_VALUE, &hKey);
    if (lRet == ERROR_SUCCESS)
    {
      lRet = RegQueryValueEx(hKey, TEXT("AddinTimeout"), NULL, &dwKeyType, (LPBYTE)dwAddinTimeout, &dwBufLen);
      if ((lRet == ERROR_SUCCESS) && (dwKeyType == REG_DWORD))
      {
        m_poll_limit = dwAddinTimeout / 10;  // always sleeping for at least 10 msec.
        if (m_poll_limit == 0) m_poll_limit = 1;
      } else {
        m_poll_limit = 500;  // default 5 sec. timeout 
      }
      RegCloseKey(hKey);
    }
    RegCloseKey(hCurrentUser);
  }
*/
  char *marmotaddin_timeoutstr = getenv("MARMOT_ADDIN_TIMEOUT");
  if (marmotaddin_timeoutstr)
  {
    stringstream sstr;
    sstr << marmotaddin_timeoutstr;
    sstr >> m_poll_limit;
    m_poll_limit = m_poll_limit / 10;
    if (m_poll_limit == 0) m_poll_limit = 1;
  }

  // get listen port from env.var MARMOT_LOG_PORT (if not set use default 30777)
  m_marmot_log_port = 30777;
  char *portstr = getenv("MARMOT_LOG_PORT");
  if (portstr)
  {
    stringstream sstr;
    sstr << portstr;
    sstr >> m_marmot_log_port;
  }
  // do network init (increment port if it is already used by another instance of the addin)
  int max_port = m_marmot_log_port + 10;
  do
  {
    m_serversocket = new CTcpSocket(m_marmot_log_port);
    if (!m_serversocket)
    {
      ::MessageBox(NULL, "Could not create listen socket! Alloc failed!", "MarmotAddin - Error", MB_OK);
      return false;
    }
    if (!m_serversocket->isOk())
    {
      // is port in use by another addin instance?
      if (m_serversocket->getErrorCode() == NET_EADDRINUSE)
      {
        // increment port-no. and retry
        delete m_serversocket;  m_serversocket = NULL;
        m_marmot_log_port++;
        continue;
      } else {
        // other error bail out
        ::MessageBox(NULL, m_serversocket->getError().c_str(), "MarmotAddin - Error", MB_OK);
        return false;
      }
    } else {
      // socket ok!
      break;
    }
  } while (m_marmot_log_port < max_port);

  // set marmot_log_host
  vector<string> rout;
  if (NetUtil::resolve(NULL, rout))
  {
    if (!rout.empty())
    {
      m_marmot_log_host = rout[0];
    }
  } else {
    m_marmot_log_host = "127.0.0.1";
  }
  m_initialized = true;
  return true;
}

void CMarmotComm::waitForFinish()
{
  while (!m_finished)
    Sleep(500);
}

void CMarmotComm::setAppToLaunch(const char *cmdline)
{
  CLock applaunch_lock(m_applaunch_mutex);
  m_apptolaunch = cmdline;
}

bool CMarmotComm::appLaunch()
{
  CLock applaunch_lock(m_applaunch_mutex);
  if (m_apptolaunch.empty())
    return false;

  char buf[80];
/*
  // get current working dir
  TCHAR thecwd[MAX_PATH+4];
  memset(thecwd, 0, sizeof(thecwd));
  DWORD err = GetCurrentDirectory(sizeof(thecwd), thecwd);
  if ((err == 0) || (err > sizeof(thecwd)))
  {
    sprintf(buf, "GetCurrentDirectory() failed with code: %d\n", GetLastError());
    OutputVS(buf);
    return false;
  } else {
    sprintf(buf, "GetCurrentDirectory() -> %s\n", thecwd);
    OutputVS(buf);
  }
*/
  TCHAR cmdline[MAX_PATH];
  strcpy(cmdline, m_apptolaunch.c_str());
  // reset m_apptolaunch, so it does not get started all over again
  m_apptolaunch = "";
  
  STARTUPINFO si;
  PROCESS_INFORMATION pi;
  ZeroMemory(&si, sizeof(si));  si.cb = sizeof(si);
  ZeroMemory(&pi, sizeof(pi));

  // clear the DICE display
#ifdef USE_MARMOT_DICE
  CWin_DiceWindow::GetInstance()->putVisible();
  CToolWinComm::GetInstance()->pushClear();
#endif
  
  if (!CreateProcess(NULL, cmdline, NULL, NULL, FALSE, CREATE_NEW_CONSOLE, NULL, NULL, &si, &pi)) 
  {
    sprintf(buf, "CreateProcess() failed with code: %d\n", GetLastError());
    OutputVS(buf);
    return false;
  }

  // successful launch
  return true;
}

void CMarmotComm::run()
{
  // check
  if ((m_pDTE == NULL) || (m_pBuildOutput == NULL))
    return;
  
  string hstr, ack_mes("ACK");

  // do loop
  while (isRunning())
  {
    // init() called and server socket initialized?
    if (!m_initialized)
    {
      Sleep(100);
      continue;
    }
    
    // do app-launch (if something was set)
    if (!appLaunch())
    {
      Sleep(100);
      continue;
    }

    // something was launched. accept incoming connections...
    hstr = "going to accept incoming connections for ";
    char timeoutstr[20];
    NetUtil::itoa(m_poll_limit*10, timeoutstr, 10);
    hstr.append(timeoutstr);
    hstr.append(" msec.\n");
    OutputVS(hstr.c_str());
    unsigned int poll_counter = 0;
    while ((!m_serversocket->checkSocketReadable(0, 0)) && (poll_counter < m_poll_limit) && isRunning())
    {
      Sleep(10);
      poll_counter++;
    }
    if ((poll_counter < m_poll_limit) && isRunning()) // did not time out and user did not request addin removal
      m_channel = m_serversocket->accept();

    // Could we establish a client channel?
    if (m_channel == NULL)
    {
      
      OutputVS("Could not establish client-channel!\n");
      OutputVS(m_serversocket->getError().c_str());
      continue;
    }
    // get peer address
    CSocketAddr peer_sa = m_channel->getPeerAddress();
    hstr = "Client connection established from ";
    hstr.append(peer_sa.toString());
    hstr.append("\n");
    OutputVS(hstr.c_str());
    // set linger to 1 sec.
    m_channel->setLinger(true, 1);
    // read stuff from marmot debug-server
    do
    {
      string clientstr = m_channel->readln();
      if (!clientstr.empty())
      {
        // parse marmot message
        vector<string> msgparts;
        string outputmsg;
        StringToVector(clientstr, msgparts, MARMOT_MSG_DELIM);
        hstr = "\nGot marmot message with ";
        char msg_item_count[20];
        NetUtil::itoa(msgparts.size(), msg_item_count, 10);
        hstr.append(msg_item_count);
        hstr.append(" tokens\n");
        OutputVS(hstr.c_str());
#ifdef _DEBUG
        OutputVS(clientstr.c_str());
#endif
        // get message type
        int msg_type = 0;
        if (!msgparts[3].empty())
          sscanf(msgparts[3].c_str(), "%d", &msg_type);
        // get message location (line)
        int myline_no = 1;
        if (!msgparts[6].empty())
          sscanf(msgparts[6].c_str(), "%d", &myline_no);
        outputmsg = msgparts[4];
        // filter out <br>
        for (string::size_type i=0, pos; (pos = outputmsg.find("<br>", i)) != string::npos; i=pos + 1)
        {
          outputmsg.replace(pos, 4, " ");
        }
        // add html-reference
        if (!msgparts[9].empty())
        {
          outputmsg.append("  \"file://");
          outputmsg.append(msgparts[9]);
          outputmsg.append("\"");
        }

        // TODO: map relative path/filename to an absolute one
        //
        switch (msg_type)
        {
          case MSG_TYPE_WARNING:
            OutputMarmotMessage(vsTaskPriorityMedium, msgparts[5].c_str(), myline_no, outputmsg.c_str());
            break;
          case MSG_TYPE_ERROR:
            // send ACK message to debug-server
            m_channel->writeln(ack_mes);
            OutputMarmotMessage(vsTaskPriorityHigh, msgparts[5].c_str(), myline_no, outputmsg.c_str());
            break;
          case MSG_TYPE_NOTE:
            OutputMarmotMessage(vsTaskPriorityLow, msgparts[5].c_str(), myline_no, outputmsg.c_str());
            break;
        }

#ifdef USE_MARMOT_DICE
		//pass the message to the DICE window
		CWin_DiceWindow::GetInstance()->putVisible();
		CToolWinComm::GetInstance()->pushNewMarmotMsg(msgparts);
#endif
      }
    } while (m_channel->isOk());
    delete m_channel;  m_channel = NULL;
  }
  m_finished = true;
}

STDMETHODIMP CMarmotComm::OutputVS(LPCSTR mes)
{
  CLock  outputlock(m_output_mutex);

  if (m_pMarmotOutput)
  {
    // Write text to this new pane
    m_pMarmotOutput->OutputString(A2BSTR(mes));
    return S_OK;
  }
  return S_FALSE;
}

STDMETHODIMP CMarmotComm::OutputMarmotMessage(EnvDTE::vsTaskPriority priority, LPCSTR filename, long line_no, LPCSTR mes)
{
  CLock  outputlock(m_output_mutex);
 
  // get the Build Output Pane
  if (m_pBuildOutput)
  {
    char line_str[20];
    memset(line_str, 0, sizeof(line_str));
    _itoa_s(line_no, line_str, 20, 10);
    // if (strlen(filename) == 0)
    //   return S_FALSE;
    string prep_mes = filename + string("(") + string(line_str) + ")" + " : " + mes;
    //string prep_mes = string("c:\\MyProjects\\MARMOT_CmakeExamples\\InvalidRequest\\InvalidRequest.c") + string("(") + string(line_str) + ")" + string(" : ") + string("Error : muh");
    prep_mes.append("\n");
    // vsTaskCategoryMisc or vsTaskCategoryUser
    //m_pBuildOutput->OutputTaskItemString(A2BSTR(prep_mes.c_str()), priority, A2BSTR(vsTaskCategoryMisc), vsTaskIconCompile, A2BSTR(filename), line_no, A2BSTR(mes), false);
    m_pBuildOutput->OutputTaskItemString(A2BSTR(prep_mes.c_str()), priority, A2BSTR(vsTaskCategoryMisc), vsTaskIconCompile, A2BSTR(filename), line_no, A2BSTR(mes), false);
    m_pDTE->ExecuteCommand(A2BSTR("Build.Cancel"), A2BSTR(""));
    //m_pBuildOutput->OutputString(A2BSTR("\nOutput from MarmotAddin - done.\n\n"));
    m_pBuildOutput->ForceItemsToTaskList();
    return S_OK;
  }
  return S_FALSE;
}
