/**********************************************************************
 * @file   ThreadLib.cpp
 * @brief  Implementation of CThread, CMutex, CLock
 * 
 * @author Blasius Czink
 * @date   20.12.2003
 **********************************************************************
 * Version History:
 * V 0.1  20.12.2003  : First Revision
 **********************************************************************/

#include <stdio.h>
#ifdef _WIN32
#include <process.h>
#else
#include <unistd.h>
#endif
#include "ThreadLib.h"


// ------------------------------------------------------------------
// CThread -Impl.
// ------------------------------------------------------------------

CThread::CThread(bool release)
:m_thread(0)
,m_running(true)
,m_release(false)
,m_delete_on_exit(false)
,m_destructor(false)
{
#ifdef _WIN32
	m_thread = (HANDLE)_beginthreadex(NULL, 0, &startThread, this, 0, &m_dwThreadId);
#else
	pthread_attr_t attr;
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_DETACHED);
	if (pthread_create(&m_thread,&attr, StartThread,this) == -1)
	{
		perror("Could not create CThread!");
		setRunning(false);
	}
//	pthread_attr_destroy(&attr);
#endif
	m_release = release;
}

CThread::~CThread()
{
	m_destructor = true;
	if (m_running)
	{
		setRelease(true);
		setRunning(false);
#ifdef _WIN32
		Sleep(1000);
#else
		sleep(1);
#endif
	}
#ifdef _WIN32
	if (m_thread)
		::CloseHandle(m_thread);
#endif
}

threadfunc_t STDCALLPREFIX CThread::startThread(threadparam_t the_thread)
{
	CThread *p = (CThread *)the_thread;
	while (p->m_running && !p->m_release)
	{
#ifdef _WIN32
		Sleep(1000);
#else
		sleep(1);
#endif
	}
	if (p->m_running)
	{
		p->run();
	}
	p->setRunning(false); // if return
	if (p->isDeletedOnExit() && !p->isDestructor())
	{
		delete p;
	}
#ifdef _WIN32
	_endthreadex(0);
#endif
	return (threadfunc_t)NULL;
}


bool CThread::isRunning(void)
{
 	return m_running;
}


void CThread::setRunning(bool x)
{
 	m_running = x;
}


bool CThread::isReleased(void)
{
 	return m_release;
}


void CThread::setRelease(bool x)
{
 	m_release = x;
}


bool CThread::isDeletedOnExit(void)
{
	return m_delete_on_exit;
}


void CThread::setDeletedOnExit(bool x)
{
	m_delete_on_exit = x;
}


bool CThread::isDestructor(void)
{
	return m_destructor;
}

// ------------------------------------------------------------------
// CMutex -Impl.
// ------------------------------------------------------------------
CMutex::CMutex()
{
#ifdef _WIN32
	m_mutex = ::CreateMutex(NULL, FALSE, NULL);
#else
	pthread_mutex_init(&m_mutex, NULL);
#endif
}

CMutex::~CMutex()
{
#ifdef _WIN32
	::CloseHandle(m_mutex);
#else
	pthread_mutex_destroy(&m_mutex);
#endif
}

void CMutex::lock()
{
#ifdef _WIN32
	DWORD d = WaitForSingleObject(m_mutex, INFINITE);
	// TODO: check result value
#else
	pthread_mutex_lock(&m_mutex);
#endif
}

void CMutex::unlock()
{
#ifdef _WIN32
	::ReleaseMutex(m_mutex);
#else
	pthread_mutex_unlock(&m_mutex);
#endif
}

// ------------------------------------------------------------------
// CLock -Impl.
// ------------------------------------------------------------------
CLock::CLock(CMutex& m) : m_mutex(m)
{
	m_mutex.lock();
}

CLock::~CLock()
{
	m_mutex.unlock();
}
