/**********************************************************************
 * @file   inifile.cpp
 * @brief  Implementation of the class CIniFile 
 * 
 * A helper class to handle access to ini-files.
 * The structure of an ini-file is quite simple, so no further
 * description is necessary here.
 *
 * @author Blasius Czink
 * @date   12.10.2003
 **********************************************************************
 * Version History:
 * V 0.1  12.10.2003  : First Revision
 * V 0.2  03.12.2008  : Changed to work with QT4
 **********************************************************************/
#include <QtCore>
#include <QTextStream>
#include <QDateTime>
#include "inifile.h"

#define DBGOUT qDebug

#if !defined(DEBUG) && !defined(_DEBUG)
#define QT_NO_DEBUG_OUTPUT
#endif

/**
 * @brief Default - Constructor
 */
CIniFile::CIniFile()
{
//  m_data.setAutoDelete(true);
  m_total_num_keys = 0;
  m_content_changed = false;
}

/**
 * @brief Constructor that sets the path/filename of the ini-file
 * @param inipath  : Reference to a QString which holds path/filename of the ini-file
 * @see setPath
 */
CIniFile::CIniFile(QString &inipath)
{
	m_filename = inipath;
  m_total_num_keys = 0;
//  m_data.setAutoDelete(true);
  m_content_changed = false;
}


/**
 * @brief Destructor
 */
CIniFile::~CIniFile()
{
}

/**
 * @brief Reads ini-file
 * 
 * Reads ini-file and places data in internal structures
 *
 * @return Returns true if the ini-file could be read without any errors, false otherwise
 * If there were errors a detailed error description can be obtained with getLastError().
 * @see writeFile
 */
bool CIniFile::readFile()
{
  QFile file(m_filename);
  m_data.clear();
  m_content_changed = false;

	//CFileStatus status;
  if (!file.open(QFile::ReadOnly | QFile::Text))
  {
    m_error = "Could not open INI-File (" + m_filename + ") for reading !";
    DBGOUT(qPrintable(m_error));
    return false;
  }

	QTextStream t(&file);
	QString s;
	QString sectionname, keyname, value;
  INILineType lt;
  int line_count = 1;
  CSection *hptr = NULL;

  while (!t.atEnd())
  {
    s = t.readLine();
    lt = (INILineType) parseIniLine(line_count, s, sectionname, keyname, value);
    switch (lt)
    {
     case SECTION_LINE:
       m_data.append(new CSection(sectionname));
     break;
     case KEY_LINE:
       hptr = m_data.last();
       if (hptr == NULL)
       {
         // KEY_LINE without a parent-section !
         m_error = "Syntax-Error in INI-File (" + m_filename + QString(") in line %1 (No section for this line) !").arg(line_count);
         DBGOUT(qPrintable(m_error));
         file.close();
         return false;
       }
       // OK! Insert key and value !
       if (hptr->addKey(keyname, value))
       {
         m_total_num_keys++;
       } else { // ERROR: the same key-name twice in a section !!!
         m_error = "Syntax-Error in INI-File (" + m_filename + QString(") in line %1 (Key already in section) !").arg(line_count);
         DBGOUT(qPrintable(m_error));
       }
     break;
     case INVALID_LINE:
       file.close();
       return false;
     break;
    }
    line_count++;
  }
  file.close();
	return true;
}

/**
 * @brief Write data from memory to the ini-file
 * 
 * @return true if ini-file could be saved without any error, false otherwise
 * If there were errors a detailed error description can be obtained with getLastError().
 * @see writeFile
 */
bool CIniFile::writeFile()
{
  // Is an update necessary ?
  if (!m_content_changed) return true;

  // Write out file
	QFile file(m_filename);
  if (!file.open(QFile::WriteOnly | QFile::Text))
  {
    m_error = "Could not open INI-File (" + m_filename + ") for writing !";
    DBGOUT(qPrintable(m_error));
    return false;
  }

  QTextStream t(&file);
	QString s;
	QString sectionname, keyname, value, tmp;
  CSection *hptr = NULL;
  KeyMap::Iterator it;
  
  t << QString("# Automatically generated with CIniFile by B.C.\n");
  QDateTime mydatetime = QDateTime::currentDateTime();
  t << QString("# ") << mydatetime.toString() << QString("\n\n");

  //for (hptr=m_data.first(); hptr != NULL; hptr=m_data.next())
  foreach (hptr, m_data)
  {
    sectionname = hptr->getName();
    t << QChar('[') << sectionname << QString("]\n");
    for(it=hptr->getKeyMap().begin(); it != hptr->getKeyMap().end(); ++it )
    {
      t << it.key() << QString(" = ") << it.value() << QString("\n");
    }
    t << QString("\n");
  }
  file.close();
  m_content_changed = false;
  return true;
}

/**
 * @brief Clears all internal data
 *
 * After using this method you should either read an ini-file or build one.
 *
 * @see readFile, writeFile
 */
void CIniFile::reset()
{
  m_data.clear();
}

//returns number of keys stored in specified section

/**
 * @brief Returns the number of keys in a given section
 *
 * @param sectionname : name of the section
 * @return Number of keys or -1 if the given section was not found
 */
int CIniFile::getNumKeys(QString &sectionname)
{
  CSection *hptr;
  // for (hptr=m_data.first(); hptr != NULL; hptr=m_data.next())
  foreach (hptr, m_data)
  {
    if (hptr->getName() == sectionname)
    {
      return hptr->getCount();
    }
  }
  // Section not found
  return -1;
}

//gets value of [sectionname] keyname = 
//overloaded to return QString, int, and double,
//returns "", or 0 if key/value not found.  Sets error member to show problem

/**
 * @brief Returns data for a given key and section
 *
 * @param sectionname : name of the section
 * @param keyname : name of the key
 * @param value_str : Reference to a QString which receives the key-value
 * @return Returns false if the section and/or key were not found, otherwise true.
 * If successful value_str contains the key-value.
 * @see setValue
 */
bool CIniFile::getValue(const QString &sectionname, const QString &keyname, QString &value_str)
{
  CSection *hptr;
  // for (hptr=m_data.first(); hptr != NULL; hptr=m_data.next())
  foreach (hptr, m_data)
  {
    if (hptr->getName() == sectionname)
    {
      if (hptr->getKeyValue(keyname, value_str))
      {
        return true;
      }
      m_error = "Key (" + keyname + ") not found in section (" + sectionname + ") !";
      DBGOUT(qPrintable(m_error));
      return false;
    }
  }
  // Section not found
  m_error = "Unable to locate specified section (" + sectionname + ") !";
  DBGOUT(qPrintable(m_error));
  return false;
}

/**
 * @brief Returns data for a given key and section
 *
 * @param sectionname : name of the section
 * @param keyname : name of the key
 * @param value_int : Reference to an int, which receives the key-value
 * @return Returns false if the section and/or key were not found, otherwise true.
 * If successful value_int contains the key-value.
 * @see setValue
 */
bool CIniFile::getValue(const QString &sectionname, const QString &keyname, int &value_int)
{
  CSection *hptr;
  // for (hptr=m_data.first(); hptr != NULL; hptr=m_data.next())
  foreach (hptr, m_data)
  {
    if (hptr->getName() == sectionname)
    {
      QString hstr;
      if (hptr->getKeyValue(keyname, hstr))
      {
        bool bConvOk;
        value_int = hstr.toInt(&bConvOk, 10);
        if (!bConvOk)
        {
          m_error = "Error converting INI key-value ("+hstr+") to an integer!";
          DBGOUT(qPrintable(m_error));
          return false;
        }
        return true;
      } else {
        // key does not exist
        m_error = "Key (" + keyname + ") not found in section (" + sectionname + ") !";
        DBGOUT(qPrintable(m_error));
        value_int = 0;
        return false;
      }
    }
  }
  // Section not found
  m_error = "Unable to locate specified section (" + sectionname + ") !";
  DBGOUT(qPrintable(m_error));
  value_int = 0;
  return false;
}

/**
 * @brief Returns data for a given key and section
 *
 * @param sectionname : name of the section
 * @param keyname : name of the key
 * @param value_dbl : Reference to a double, which receives the key-value
 * @return Returns false if the section and/or key were not found, otherwise true.
 * If successful value_dbl contains the key-value.
 * @see setValue
 */
bool CIniFile::getValue(const QString &sectionname, const QString &keyname, double &value_dbl)
{
  CSection *hptr;
  // for (hptr=m_data.first(); hptr != NULL; hptr=m_data.next())
  foreach (hptr, m_data)
  {
    if (hptr->getName() == sectionname)
    {
      QString hstr;
      if (hptr->getKeyValue(keyname, hstr))
      {
        bool bConvOk;
        value_dbl = hstr.toDouble(&bConvOk);
        if (!bConvOk)
        {
          m_error = "Error converting INI key-value ("+hstr+") to a double!";
          DBGOUT(qPrintable(m_error));
          return false;
        }
        return true;
      } else {
        // key does not exist
        m_error = "Key (" + keyname + ") not found in section (" + sectionname + ") !";
        DBGOUT(qPrintable(m_error));
        value_dbl = 0.0;
        return false;
      }
    }
  }
  // Section not found
  m_error = "Unable to locate specified section ("+ sectionname +") !";
  DBGOUT(qPrintable(m_error));
  value_dbl = 0.0;
  return false;
}

/**
 * @brief Set data for a given section and key
 *
 * @param sectionname : Reference to a QString with the name of the section
 * @param keyname : Reference to a QString with the name of the key
 * @param value_str : New value of the key (as a QString)
 * @param create : Determines whether a section/key is created if it does not exist. Default is true.
 * @return Returns true if the key-value was set successfully. false is returned only if
 * create=false and the given section/key do not exist.
 * @see getValue
 */
bool CIniFile::setValue(QString &sectionname, QString &keyname, QString &value_str, bool create)
{
  CSection *hptr = NULL;
  // for (hptr=m_data.first(); hptr != NULL; hptr=m_data.next())
  foreach (hptr, m_data)
  {
    if (hptr->getName() == sectionname)
    { // section found
      if (create)
      {
        hptr->addKey(keyname, value_str);
        m_content_changed = true;
        return true;
      } else {
        // Could we set the key value
        m_content_changed = hptr->setKeyValue(keyname, value_str);
        return m_content_changed;
      }
    }
  }
  // section not found
  if (create)
  {
    hptr = new CSection(sectionname);
    hptr->addKey(keyname, value_str);
    m_data.append(hptr);
    m_content_changed = true;
    return true;
  }
  return false;
}

/**
 * @brief Set data for a given section and key
 *
 * @param sectionname : Reference to a QString with the name of the section
 * @param keyname : Reference to a QString with the name of the key
 * @param value_int : New value of the key (as an int)
 * @param create : Determines whether a section/key is created if it does not exist. Default is true.
 * @return Returns true if the key-value was set successfully. false is returned only if
 * create=false and the given section/key do not exist.
 * @see getValue
 */
bool CIniFile::setValue(QString &sectionname, QString &keyname, int value_int, bool create)
{
  CSection *hptr = NULL;
  QString tmp_num = QString::number(value_int);
  // for (hptr=m_data.first(); hptr != NULL; hptr=m_data.next())
  foreach (hptr, m_data)
  {
    if (hptr->getName() == sectionname)
    { // section found
      if (create)
      {
        hptr->addKey(keyname, tmp_num);
        m_content_changed = true;
        return true;
      } else { 
        m_content_changed = hptr->setKeyValue(keyname, tmp_num);
        return m_content_changed;
      }
    }
  }
  // section not found
  if (create)
  {
    hptr = new CSection(sectionname);
    hptr->addKey(keyname, tmp_num);
    m_data.append(hptr);
    m_content_changed = true;
    return true;
  }
  return false;
}

/**
 * @brief Set data for a given section and key
 *
 * @param sectionname : Reference to a QString with the name of the section
 * @param keyname : Reference to a QString with the name of the key
 * @param value_dbl : New value of the key (as a double)
 * @param create : Determines whether a section/key is created if it does not exist. Default is true.
 * @return Returns true if the key-value was set successfully. false is returned only if
 * create=false and the given section/key do not exist.
 * @see getValue
 */
bool CIniFile::setValue(QString &sectionname, QString &keyname, double value_dbl, bool create)
{
  CSection *hptr = NULL;
	QString   tmp_num = QString::number(value_dbl, 'g', 17);
  // for (hptr=m_data.first(); hptr != NULL; hptr=m_data.next())
  foreach (hptr, m_data)
  {
    if (hptr->getName() == sectionname)
    { // section found
      if (create)
      {
        hptr->addKey(keyname, tmp_num);
        m_content_changed = true;
        return true;
      } else {
        m_content_changed = hptr->setKeyValue(keyname, tmp_num);
        return m_content_changed;
      }
    }
  }
  // section not found
  if (create)
  {
    hptr = new CSection(sectionname);
    hptr->addKey(keyname, tmp_num);
    m_data.append(hptr);
    m_content_changed = true;
    return true;
  }
  return false;
}

/**
 * @brief Delete key from a given section
 *
 * @param sectionname : name of the section
 * @param keyname : name of the key
 * @return true if the key existed and was deleted successfully, false otherwise
 * @see deleteSection, setValue
 */
bool CIniFile::deleteKey(QString &sectionname, QString &keyname)
{
  CSection *hptr = NULL;
  // for (hptr=m_data.first(); hptr != NULL; hptr=m_data.next())
  foreach (hptr, m_data)
  {
    if (hptr->getName() == sectionname)
    { // section found
      // if successfully deleted then the content is also changed
      m_content_changed = hptr->deleteKey(keyname);
      return m_content_changed;
    }
  }
  // section not found
  return false;
}

/**
 * @brief Delete section with all its keys
 *
 * @param sectionname : name of the section
 * @return true if the given section existed and was deleted with all its keys, false otherwise
 * @see deleteKey, setValue
 */
bool CIniFile::deleteSection(QString &sectionname)
{
  CSection *hptr = NULL;
  int idx;
  // for (hptr=m_data.first(); hptr != NULL; hptr=m_data.next())
  foreach (hptr, m_data)
  {
    if (hptr->getName() == sectionname)
    { // section found
      //m_data.remove(hptr);
      idx = m_data.indexOf(hptr);
      m_data.removeAt(idx);
      //m_data.removeOne(hptr);
      m_content_changed = true;
      return true;
    }
  }
  // section not found
  return false;
}

// private stuff

/**
 * @brief Parse a line from an ini-file
 * 
 * Classify and read data in a line
 *
 * @param line_no : line number so error message may give better hints to locate an error in the ini-file
 * @param s : QString containing a line from the ini-file
 * @param sectionname : receives sectionname if the given line contains a sectionname
 * @param keyname : receives keyname if the line contains a key
 * @param value : receives the value of a key
 * @return Returns type of line or an error code (SECTION_LINE, KEY_LINE, EMPTY_LINE, COMMENT_LINE, INVALID_LINE)
 */
int CIniFile::parseIniLine(int line_no, QString &s, QString &sectionname, QString &keyname, QString &value)
{
  QString tmp = s.trimmed();

  if (tmp.isEmpty()) return EMPTY_LINE;

  if (tmp[0] == QChar('[')) // its a section
  {
    int sec_end = tmp.lastIndexOf(QChar(']'));
    if (sec_end == -1)
    {
      m_error = "Syntax-Error in INI-File (" + m_filename + QString(") in line %1 !").arg(line_no);
      DBGOUT(qPrintable(m_error));
      return INVALID_LINE;
    }
    sectionname = tmp.mid(1, sec_end-1);
    if (sectionname.isEmpty())
    {
      m_error = "Syntax-Error in INI-File (" + m_filename + QString(") in line %1 !").arg(line_no);
      DBGOUT(qPrintable(m_error));
      return INVALID_LINE;
    }
    keyname = "";
    value   = "";
    return SECTION_LINE;
  }
  if (tmp[0] == QChar('#')) // its a comment line
  {
    s = tmp;
    return COMMENT_LINE;
  }
  // must be a key line
  // but there could be a comment so remove it first
  int comment_begin = tmp.indexOf(QChar('#'));
  if (comment_begin != -1)
  {
    //tmp.remove(comment_begin, tmp.length());
    tmp.truncate(comment_begin);
    tmp = tmp.trimmed();
  }
  // no more comments in line
  int sep_pos = tmp.indexOf(QChar('='));
  if (sep_pos == -1)
  {
    m_error = "Syntax-Error in INI-File (" + m_filename + QString(") in line %1 !").arg(line_no);
    DBGOUT(qPrintable(m_error));
    return INVALID_LINE;
  }
  keyname = (tmp.left(sep_pos)).trimmed();
  value   = (tmp.mid(sep_pos+1)).trimmed();
  return KEY_LINE;
}
