/**********************************************************************
 * @file   inifile.h
 * @brief  Declaration of the class CIniFile
 * 
 * @author Blasius Czink
 * @date   12.10.2003
 **********************************************************************
 * Version History:
 * V 0.1  12.10.2003  : First Revision
 * V 0.2  03.12.2008  : Changed to work with QT4
 **********************************************************************/
#ifndef FW_INIFILE_H
#define FW_INIFILE_H

#include <qstring.h>
#include <qfile.h>
#include <qmap.h>
#include <qlist.h>
//#include "fwconfig.h"
//#include <qobject.h>
//#include <iostream.h>

/**
 * @class  CIniFile 
 * @brief  Class for the handling of ini-files
 * 
 * @author  B.C.
 * @date    12.10.2003 
 * @version 0.1
 * @since   0.1
 * @bug  none
 */
class CIniFile  
{
	//all private variables
private:

  // line types
  enum INILineType { SECTION_LINE, KEY_LINE, EMPTY_LINE, COMMENT_LINE, INVALID_LINE };
  typedef QMap<QString, QString> KeyMap;

  // class to help with the handling of sections
  class CSection {
  public:
    QString m_name;
    KeyMap  m_keys;

    CSection(QString &name) { m_name = name; };
    ~CSection() {  };

    void reset(void) { m_keys.clear(); };
    int  getCount(void) { return (m_keys.count()); };
    KeyMap &getKeyMap(void) { return m_keys; };
    QString &getName(void) { return m_name; };

    bool getKeyValue(const QString &keyname, QString &value) const
    {
      if (m_keys.contains(keyname))
      {
        value = m_keys[keyname];
        return TRUE;
      }
      value = QString("");
      return FALSE;
    };
    bool setKeyValue(const QString &keyname, const QString &value)
    {
      if (m_keys.contains(keyname))
      {
        m_keys.insert(keyname, value);
        return TRUE;
      }
      return FALSE;
    };
    bool addKey(QString &keyname, QString &value)
    {
      // just replace if key already there and return FALSE since no new key was inserted
      if (setKeyValue(keyname, value)) return FALSE;
      // we have to insert the key with the value
      m_keys.insert(keyname, value);
      return TRUE;
    };
    bool deleteKey(QString &keyname)
    {
      KeyMap::Iterator it;
      it = m_keys.find(keyname);
      if (it == m_keys.end()) return FALSE;
      m_keys.erase(it);
      return TRUE;
    };
  };
  
  /** 
   * @brief Contains all the data from the ini-file in a list of sections */
  QList<CSection *> m_data;
  /** 
   * @brief Contains the path/filename of the ini-file */
  QString m_filename;
  /** 
   * @brief Contains an error message (syntax errors etc.) */
  QString m_error;
  /** 
   * @brief Total number of keys in the ini-file */
  int m_total_num_keys;
  /** 
   * @brief Dirty-flag  */
  bool m_content_changed;

  //all private functions
  int parseIniLine(int line_no, QString &s, QString &sectionname, QString &keyname, QString &value);

  //public functions
public:
  CIniFile();
  CIniFile(QString &inipath);
  ~CIniFile();

  /**
   * @brief Sets path/filename of an ini-file
   * @param newpath  : Reference to the new path/filename
   */
  inline void setPath(const QString &newpath) { m_filename = newpath; }
  inline void setPath(const char *newpath) { m_filename = newpath; }
  inline QString getPath(void) { return m_filename; }

  bool readFile();
  bool writeFile();
  void reset();

  inline bool isDirty(void) { return m_content_changed; }
  int        getTotalNumKeys() { return m_total_num_keys; }
  int        getNumKeys(QString &sectionname);
  inline int getNumKeys(const char *sectionname) 
  { 
   	QString secname(sectionname);
	return getNumKeys(secname); 
  }

  bool getValue(const QString &sectionname, const QString &keyname, QString &value_str);
  bool getValue(const QString &sectionname, const QString &keyname, int &value_int);
  bool getValue(const QString &sectionname, const QString &keyname, double &value_dbl);
  // convenience get-methods
  inline bool getValue(const char *sectionname, const char *keyname, QString &value_str)
  {
   QString secname(sectionname), kname(keyname);
   return getValue(secname, kname, value_str);
  }
  inline bool getValue(const char *sectionname, const char *keyname, int &value_int)
  {
   QString secname(sectionname), kname(keyname);
   return getValue(secname, kname, value_int);
  }
  inline bool getValue(const char *sectionname, const char *keyname, double &value_dbl)
  {
   QString secname(sectionname), kname(keyname);
   return getValue(secname, kname, value_dbl);
  }

  bool setValue(QString &sectionname, QString &keyname, QString &value_str, bool create = TRUE);
  bool setValue(QString &sectionname, QString &keyname, int value_int, bool create = TRUE);
  bool setValue(QString &sectionname, QString &keyname, double value_dbl, bool create = TRUE);
  // convenience set-methods
  inline bool setValue(const char *sectionname, const char *keyname, QString &value_str, bool create = TRUE)
  {
   QString secname(sectionname), kname(keyname);
   return setValue(secname, kname, value_str, create);
  }
  inline bool setValue(const char *sectionname, const char *keyname, const char *value_str, bool create = TRUE)
  {
   QString secname(sectionname), kname(keyname), vstr(value_str);
   return setValue(secname, kname, vstr, create);
  }
  inline bool setValue(const char *sectionname, const char *keyname, int value_int, bool create = TRUE)
  {
   QString secname(sectionname), kname(keyname);
   return setValue(secname, kname, value_int, create);
  }
  inline bool setValue(const char *sectionname, const char *keyname, double value_dbl, bool create = TRUE)
  {
   QString secname(sectionname), kname(keyname);
   return setValue(secname, kname, value_dbl, create);
  }

  bool deleteKey(QString &sectionname, QString &keyname);
  bool deleteSection(QString &sectionname);
  // convenience delete-methods
  inline bool deleteKey(char *sectionname, char *keyname)
  {
   QString secname(sectionname), kname(keyname);
   return deleteKey(secname, kname);
  }
  inline bool deleteSection(char *sectionname)
  {
   QString secname(sectionname);
   return deleteSection(secname);
  }

  QString getLastError(void) { return m_error; }

  // Returns TRUE if specified key was found and its value or FALSE and an empty string.
  // The first matching key will be returned.
  //bool findKey(const QString &keyname, QString &value);

  // Returns index of the specified section, or -1 if not found
  //int findSection(const QString &sectionname);
};

#endif
