package de.hlrs.marmoclipse.log;

import java.io.*;
import java.net.*;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.*;
import org.eclipse.swt.widgets.Display;

import de.hlrs.marmoclipse.Activator;

/**
 * implements a socket listener so it is possible to receive the log information 
 * via a specific port. It extends class Job so that we can schedule the running
 * of the listener into background and avoid it blocking the user interface.
 * 
 * @author peng
 *
 */
public class LogSocketListener extends Job {
	private int port;
	private MarmotLogList logList;

	private ServerSocket server;
	private Socket client;
	private BufferedReader in;

	private String inputLine;
	private String commandLine;

	private static final String MPI_EXEC_ERROR = "MPI execution exits with return value: %d\n%s";

	public LogSocketListener(int port, MarmotLogList logList, String commandLine) {
		super("Listening on port " + port);
		this.port = port;
		this.logList = logList;
		this.inputLine = "";
		this.setCommandLine(commandLine);
	}

	public IStatus run(IProgressMonitor monitor) {
		if (server == null || server.isClosed()) {
			try {
				server = new ServerSocket(getPort());
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}

		Process proc;

		try {
			proc = Runtime.getRuntime().exec(commandLine);
		} catch (IOException e) {
			e.printStackTrace();
			return Status.CANCEL_STATUS;
		}

		String errLine;
		String err = "";
		BufferedReader std_err = new BufferedReader(new InputStreamReader(proc
			.getErrorStream()));

		// loop until mpi execution connects or encounters errors
		boolean isConnected = false;
		while (true) {
			try {
				errLine = std_err.readLine();
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
				return Status.CANCEL_STATUS;
			}
			if (errLine == null) {
				break;
			} else if (errLine.contains("Successfully connected to")) {
				isConnected = true;
				break;
			} else {
				err = err + errLine + "\n";
			}
		}

		if (isConnected) {
			try {
				client = server.accept();
				in = new BufferedReader(new InputStreamReader(client
					.getInputStream()));

				Runnable updateLog = new Runnable() {
					public void run() {
						logList.addLog(inputLine);
					}
				};

				while ((inputLine = in.readLine()) != null) {
					Display.getDefault().syncExec(updateLog);
				}

				client.close();
			} catch (IOException e) {
				System.out.println("Error during listening: " + e.getMessage());
				return Status.CANCEL_STATUS;
			} catch (IllegalArgumentException e) {
				System.out.println("Error during listening: " + e.getMessage());
				return Status.CANCEL_STATUS;
			}
		}
		
		try {
			proc.waitFor();
		} catch (InterruptedException e) {
			e.printStackTrace();
			return new Status(Status.CANCEL, Activator.PLUGIN_ID, e.getMessage(), null);
		}

		if (proc != null && proc.exitValue() != 0) {
			String message = String.format(MPI_EXEC_ERROR, proc.exitValue(), err);
			return new Status(Status.CANCEL, Activator.PLUGIN_ID, message, null);
		}

		return Status.OK_STATUS;
	}

	public void setPort(int port) {
		if (this.port != port) {
			this.port = port;
			try {
				server.close();
			} catch (IOException e) {
				System.out.println("Error during closing server on port "
					+ port);
			}
		}
	}

	public int getPort() {
		return port;
	}

	public void setCommandLine(String commandLine) {
		this.commandLine = commandLine;
	}

	public String getCommandLine() {
		return commandLine;
	}

	protected void finalize() throws Throwable {
		try {
			if (!client.isClosed())
				client.close();
			if (!server.isClosed())
				server.close();
		} catch (IOException e) {
			System.out.println("Cannot close socket.");
		}
	}
}