package de.hlrs.marmoclipse.views;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.text.SimpleDateFormat;
import java.util.Date;

import org.eclipse.core.runtime.jobs.IJobChangeEvent;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.core.runtime.jobs.JobChangeAdapter;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.resource.JFaceResources;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.FileDialog;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Spinner;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.ISharedImages;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.part.*;

import de.hlrs.marmoclipse.Activator;
import de.hlrs.marmoclipse.log.LogSocketListener;
import de.hlrs.marmoclipse.log.MarmotLog;

/**
 * creates the user interface of the Marmot Control View by implementing
 * all the UI components and their corresponding method. 
 * There is an in-line class RunSelectionListener triggered by 
 * the “Run” button when clicked. It does some sanity checks and assembly 
 * the command line from all the options chosen by the user and 
 * starts/schedules the socket listener before switching to the Marmot Log View
 * 
 * @author peng
 *
 */
public class MarmotControlView extends ViewPart {
	/**
	 * The ID of the view as specified by the extension.
	 */
	public static final String ID = "de.hlrs.marmoclipse.views.MarmotControlView";

	/**
	 * Widgets
	 */
	private Text execText;
	private Label execLabel;
	private Button execButton;

	private Label procLabel;
	private Spinner procSpinner;
	private Text argsText;
	private Label argsLabel;

	private Label hostPortLabel;
	private Text hostText;
	private Spinner portSpinner;
	private Label colonLabel;

	private Label logLevelLabel;
	private Label logFormatLabel;
	private Combo logLevelCombo;
	private Combo logFormatCombo;

	private Button runButton;
	private ImageDescriptor runImageDesc;
	private Image runImage;
	
	private LogSocketListener logListener;

	/**
	 * Constants
	 */
	private static final String EXEC_LABEL = "Executable:";
	private static final String PROC_LABEL = "Processes:";
	private static final String ARGS_LABEL = "Additional arguements:";
	private static final String HOST_PORT_LABEL = "Host (port):";
	private static final String LOG_LEVEL_LABEL = "Log Level:";
	private static final String LOG_FORMAT_LABEL = "Log Format:";

	private static final int DEFAULT_PROC = 3;
	private static final int DEFAULT_PORT = 30777;
	private static final String DEFAULT_HOST = "localhost";

	private static final String EXEC_BUTTON = "&Select";
	private static final String EXEC_TOOLTIP = "Select an executable file to run with marmot";
	private static final String PROC_TOOLTIP = "Number of parallel processes to run";
	private static final String ARGS_TOOLTIP = "Additional arguements for the executable";
	private static final String HOST_TOOLTIP = "The host name or IP address to monitor";
	private static final String LOG_LEVEL_TOOLTIP = "Level of debug message to receive";
	private static final String LOG_FORMAT_TOOLTIP = "The Format of the log file to be saved";

	private static final String LOG_LEVEL_ERROR = "Only Errors";
	private static final String LOG_LEVEL_WARNING = "Errors and Warnings";
	private static final String LOG_LEVEL_REMARK = "Everything";

	private static final String COMMAND_LINE = "mpiexec "
		+ "-x MARMOT_LOG_HOST=%1$s -x MARMOT_LOG_PORT=%2$s "
		+ "-x MARMOT_DEBUG_MODE=%3$s -x MARMOT_LOGFILE_TYPE=%4$s "
		+ "-x MARMOT_LOGFILE_NAME=%5$s -n %6$s %7$s %8$s";

	private static final String FILENAME_FORMAT = "%1$s"
		+ System.getProperty("file.separator") + "Marmot_%2$s_%3$s";

	private final int fontSize;

	/**
	 * The constructor.
	 */
	public MarmotControlView() {
		logListener = new LogSocketListener(DEFAULT_PORT, MarmotLogView.logs, "");
		logListener.addJobChangeListener(new JobChangeAdapter() {
			@Override
			public void done(final IJobChangeEvent event) {
				if (event.getResult().isOK()) {
					((MarmotLogView) getSite().getPage().findView(MarmotLogView.ID)).enableActions(); 
				} else {
					Display.getDefault().asyncExec(new Runnable() {
						public void run() {
							MessageDialog.openError(getSite()
								.getShell(), "Error", event.getResult()
								.getMessage());
						}
					});
				}
			}
		});

		int fontsize_pt = JFaceResources.getFontRegistry().defaultFont()
			.getFontData()[0].getHeight();
		int dpi = Display.getDefault().getDPI().x;
		fontSize = fontsize_pt * dpi / 72;
	}

	/**
	 * This is a callback that will allow us to create the viewer and initialize
	 * it.
	 */
	@Override
	public void createPartControl(Composite parent) {
		GridLayout gl = new GridLayout(2, false);
		gl.marginHeight = 10;
		gl.marginWidth = 10;
		parent.setLayout(gl);

		createExecLine(parent);
		createArgsLine(parent);
		createNetworkLine(parent);
		createLogLine(parent);
	}

	/**
	 * Passing the focus request to the viewer's control.
	 */
	@Override
	public void setFocus() {
		execText.setFocus();
	}

	/**
	 * Create the widgets on the executable selection line
	 * 
	 * @param parent
	 */
	private void createExecLine(Composite parent) {
		execLabel = new Label(parent, SWT.NONE);
		execLabel.setText(EXEC_LABEL);

		GridData gd = new GridData(GridData.FILL, GridData.BEGINNING, true,
			false);
		Composite execRest = new Composite(parent, SWT.NONE);
		execRest.setLayoutData(gd);
		execRest.setLayout(new GridLayout(2, false));

		gd = new GridData();
		gd.horizontalAlignment = GridData.FILL;
		gd.grabExcessHorizontalSpace = true;
		execText = new Text(execRest, SWT.SINGLE | SWT.BORDER);
		execText.setLayoutData(gd);
		execText.setToolTipText(EXEC_TOOLTIP);

		execButton = new Button(execRest, SWT.PUSH);
		execButton.setText(EXEC_BUTTON);
		execButton.setImage(PlatformUI.getWorkbench().getSharedImages()
			.getImage(ISharedImages.IMG_OBJ_FOLDER));
		execButton.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				FileDialog fd = new FileDialog(execButton.getShell(), SWT.OPEN);
				fd.setText("Select file ...");
				String fn = fd.open();
				if (fn != null) {
					execText.setText(fn);
				}
			}
		});
	}

	/**
	 * Create the second line of widgets including number of concurrent
	 * processes and additional arguments
	 * 
	 * @param parent
	 */
	private void createArgsLine(Composite parent) {
		procLabel = new Label(parent, SWT.NONE);
		procLabel.setText(PROC_LABEL);

		GridData gd = new GridData(GridData.FILL, GridData.BEGINNING, true,
			false);
		Composite argsRest = new Composite(parent, SWT.NONE);
		argsRest.setLayoutData(gd);
		argsRest.setLayout(new GridLayout(3, false));

		procSpinner = new Spinner(argsRest, SWT.BORDER);
		procSpinner.setValues(DEFAULT_PROC, 1, 999, 0, 1, 10);
		procSpinner.setToolTipText(PROC_TOOLTIP);

		gd = new GridData();
		gd.horizontalIndent = 2 * fontSize;
		argsLabel = new Label(argsRest, SWT.NONE);
		argsLabel.setText(ARGS_LABEL);
		argsLabel.setLayoutData(gd);

		gd = new GridData();
		gd.horizontalAlignment = SWT.FILL;
		gd.grabExcessHorizontalSpace = true;
		argsText = new Text(argsRest, SWT.SINGLE | SWT.BORDER);
		argsText.setLayoutData(gd);
		argsText.setToolTipText(ARGS_TOOLTIP);
	}

	/**
	 * Create the line containing host and port input widgets
	 * 
	 * @param parent
	 */
	private void createNetworkLine(Composite parent) {
		hostPortLabel = new Label(parent, SWT.NONE);
		hostPortLabel.setText(HOST_PORT_LABEL);

		GridData gd = new GridData(GridData.FILL, GridData.BEGINNING, true,
			false);
		Composite networkRest = new Composite(parent, SWT.NONE);
		networkRest.setLayoutData(gd);
		networkRest.setLayout(new GridLayout(5, false));

		String localHostname;
		try {
			InetAddress addr = InetAddress.getLocalHost();
			localHostname = addr.getHostName();
		} catch (UnknownHostException e) {
			localHostname = DEFAULT_HOST;
		}

		gd = new GridData();
		gd.widthHint = 8 * fontSize;
		hostText = new Text(networkRest, SWT.BORDER | SWT.RIGHT);
		hostText.setText(localHostname);
		hostText.setToolTipText(HOST_TOOLTIP);
		hostText.setLayoutData(gd);

		colonLabel = new Label(networkRest, SWT.NONE);
		colonLabel.setText(":");

		portSpinner = new Spinner(networkRest, SWT.BORDER);
		portSpinner.setValues(DEFAULT_PORT, 1, 65535, 0, 1, 10);
	}

	/**
	 * Create log level and format selection widgets, and attach the run button
	 * at the end.
	 * 
	 * @param parent
	 */
	private void createLogLine(Composite parent) {
		logLevelLabel = new Label(parent, SWT.NONE);
		logLevelLabel.setText(LOG_LEVEL_LABEL);

		GridData gd = new GridData(GridData.FILL, GridData.BEGINNING, true,
			false);
		Composite logRest = new Composite(parent, SWT.NONE);
		logRest.setLayoutData(gd);
		logRest.setLayout(new GridLayout(4, false));

		String[] levels = { LOG_LEVEL_ERROR, LOG_LEVEL_WARNING,
			LOG_LEVEL_REMARK };
		logLevelCombo = new Combo(logRest, SWT.DROP_DOWN | SWT.READ_ONLY);
		logLevelCombo.setItems(levels);
		logLevelCombo.select(1);
		logLevelCombo.setToolTipText(LOG_LEVEL_TOOLTIP);

		gd = new GridData();
		gd.horizontalIndent = 2 * fontSize;
		logFormatLabel = new Label(logRest, SWT.NONE);
		logFormatLabel.setText(LOG_FORMAT_LABEL);
		logFormatLabel.setLayoutData(gd);

		String[] formats = { 
			MarmotLog.Formats[0].getTitle(),
			MarmotLog.Formats[1].getTitle(), 
			MarmotLog.Formats[2].getTitle() 
		};
		logFormatCombo = new Combo(logRest, SWT.DROP_DOWN | SWT.READ_ONLY);
		logFormatCombo.setItems(formats);
		logFormatCombo.select(1);
		logFormatCombo.setToolTipText(LOG_FORMAT_TOOLTIP);

		createRunButton(logRest);
	}

	/**
	 * Create the "Run" button.
	 * 
	 * @param parent
	 */
	private void createRunButton(Composite parent) {
		GridData gd = new GridData();
		gd.horizontalAlignment = GridData.END;
		gd.grabExcessHorizontalSpace = true;
		runImageDesc = Activator.getImageDescriptor("./icons/run.gif");
		runImage = runImageDesc.createImage();
		runButton = new Button(parent, SWT.PUSH);
		runButton.setText("&Run");
		runButton.setLayoutData(gd);
		runButton.setImage(runImage);
		runButton.addSelectionListener(new RunSelectionListener());
	}

	/**
	 * Selection listener implementation for "Run" button
	 * 
	 * @author peng
	 * 
	 */
	private class RunSelectionListener extends SelectionAdapter {
		private static final String NO_EXEC_TITLE = "No file selected";
		private static final String NO_EXEC_ERROR = "Please specific an executable to run.";

		@Override
		public void widgetSelected(SelectionEvent e) {
			final MarmotLogView logView = (MarmotLogView) getSite().getPage().findView(MarmotLogView.ID);
			String execFileName = execText.getText();
			if (execFileName.isEmpty()) {
				MessageDialog.openError(getSite().getShell(),
					NO_EXEC_TITLE,
					NO_EXEC_ERROR);
				return;
			}
			
			if (logListener.getState() != Job.NONE) {
				return;
			}
			
			if (logView.reset()) {
				// the temporary file path
				SimpleDateFormat fmt = new SimpleDateFormat("yyyyMMdd_HHmmss");
				String[] parts = execFileName.split(System
					.getProperty("file.separator"));
				String logFileName = String.format(FILENAME_FORMAT,
					System.getProperty("java.io.tmpdir"),
					parts[parts.length - 1],
					fmt.format(new Date()));
				logView.setTmpFile(logFileName);

				// the temporary file format
				logView.setTmpFileFormat(MarmotLog.Formats[logFormatCombo.getSelectionIndex()]);

				// construct the command line
				String cmd = String.format(COMMAND_LINE,
					hostText.getText(),
					portSpinner.getSelection(),
					logLevelCombo.getSelectionIndex(),
					logFormatCombo.getSelectionIndex(),
					logFileName,
					procSpinner.getSelection(),
					execFileName,
					argsText.getText());
				System.out.println(cmd);

				// start log listener
				logListener.setPort(portSpinner.getSelection());
				logListener.setCommandLine(cmd);

				logListener.schedule();

				// switch to log view
				try {
					getSite().getPage().showView(MarmotLogView.ID);
				} catch (PartInitException e1) {
					e1.printStackTrace();
				}
			}
		}
	}

	@Override
	public void dispose() {
		runImage.dispose();
	}
}