/*
 * Copyright (c) 2002-2008 High Performance Computing Center Stuttgart,
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2006-2008 The Center for Information Services and HPC
 *                         Technical University of Dresden.
 *                         All rights reserved.
 *
 * $COPYRIGHT$
 *
 * Additional copyrights may follow
 *
 * $HEADER$
 */
/**
 *  @file
 *
 *  Declaration of all classes representing MPI-calls.
 *
 *  @author Bettina Krammer, Katrin Bidmon, Matthias Mueller, Tobias Hilbrich
 *
 *  @see mpo_*.cc (implementation of classes representing MPI-calls)
 *  @see mpo.h, MPO_Space.cc (base classes)
 *
 *  $Id: mpo-objects.h 1096 2010-03-09 17:03:04Z shiqing $
 */

#ifndef MPO_OBJECTS_H
#define MPO_OBJECTS_H

#include "config.h"
#include <iostream>
#include <string>
using std::string;

#include <mpi.h>
#include "mpi_undef_makros.h"

#include "mpo.h"


namespace MPO
{


    //=========================================================================
    //                          Environment
    //=========================================================================



    //*************************************************************************
    //
    //                          MPO_Init
    //
    /** This class represents MPI_Init. */
    class MPO_Init : public MPO_Environment
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object.
         *  It should never be called anyhow, because MPI_Init is only called
         *  once. And the corresponding objects are never transferred.
         */
        MPO_Init();

        /** Constructor taking proper arguments.
         *
         *  @param   argc - Pointer to the number of arguments
         *  @param   argv - Pointer to the argument vector
         *  @see     mk_fortran_wrapper.sh takes care that argument list is
         *           recovered when using mpich-p4.
         */
        MPO_Init(int *argc, char ***argv);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Init();

        //==========================
        // Methods
        //==========================

        /**
         * @todo a workaround, currently we call MPI_Init in its constructor,
         *       I forgot the reason for that, but there probably was one,
         *       if not we should call it in callNativeMPI and remove this
         *       workaround.
         * Disables the call to PMPI_Init in the constructor.
         */
        static void doNotCallInit (void);

		/** Initialize the new MPI_COMM_WORLD and some of the environmentals.
         *  @return  0
         */
        int initCommAndEnv();

        /** Initialize the mapping of all predefined errhandlers.
         *  All errhandlers listed in MPI-standard 1.2, pp. 211, are added
         *  to our map.
         *
         *  @return  0
         */
        int initErrhandlers();

        /** Initialize the mapping of all predefined requests.
         *
         *  @return  0
         */
        int initRequests();

        /** Initialize the mapping of all predefined datatypes.
         *  All datatypes listed in MPI-standard 1.2, pp. 211, are added
         *  to our map.
         *
         *  @return  0
         */
        int initDatatypes();

        /** Initialize the mapping of null handle.
         *  See MPI-standard 1.2, pp. 213.
         *
         *  @return  0
         */
        int initDatatypeNull();

        /** Initialize the mapping of elementary C datatypes.
         *  See MPI-standard 1.2, pp. 211.
         *
         *  @return  0
         */
        int initDatatypeElementaryC();

        /** Initialize the mapping of elementary Fortran datatypes.
         *  See MPI-standard 1.2, pp. 211.
         *
         *  @return  0
         */
        int initDatatypeElementaryFortran();

        /** Initialize the mapping of C datatypes for reduction functions.
         *  See MPI-standard 1.2, pp. 212.
         *
         *  @return  0
         */
        int initDatatypeReductionC();

        /** Initialize the mapping of Fortran
         *  datatypes for reduction functions.
         *  See MPI-standard 1.2, pp. 212.
         *
         *  @return  0
         */
        int initDatatypeReductionFortran();

        /** Initialize the mapping of optional C datatypes.
         *  See MPI-standard 1.2, pp. 212.
         *
         *  @return  0
         */
        int initDatatypeOptionalC();

        /** Initialize the mapping of optional Fortran datatypes.
         *  See MPI-standard 1.2, pp. 212.
         *
         *  @return  0
         */
        int initDatatypeOptionalFortran();

        /** Initialize the mapping of datatypes for constructing
         *  derived datatypes.
         *  See MPI-standard 1.2, pp. 212.
         *
         *  @return  0
         */
        int initDatatypeConstructDerivedDatatypes();

        /** Initialize the mapping of all predefined communicators.
         *  All communicators listed in MPI-standard 1.2, pp. 211, are added
         *  to our map.
         *
         *  @return  0
         */
        int initCommunicators();

        /** Initialize the mapping of all predefined operations.
         *  All operations listed in MPI-standard 1.2, pp. 211, are added
         *  to our map.
         *
         *  @return  0
         */
        int initOperations();

        /** Initialize the mapping of all predefined groups.
         *  All groups listed in MPI-standard 1.2, pp. 211, are added
         *  to our map.
         *
         *  @return  0
         */
        int initGroups();

#ifdef MARMOT_1SIDED
        /** Initialize the mapping of all predefined windows.
         *
         *  @return  0
         */
        int initWins();
#endif

        /** Initializes the per Process MPO_LogFile interface.
         *  The selected Implementation of the interface depends on the selected
         *  environment variable and wheter it is a normal MPI Process or the
         *  DebugServer.
         *  @return 0.
         */
        int initLogfile();

        /**
         * Initializes all the MPO_ResourceGatherer objects that store
         * traced information about resources.
         * @return 0.
         */
        int initResourceGatherers();

        /** We override callDebugServer(), because we need special treatment
         *  in MPI_Init.
         *
         *  @return  MPI_SUCCESS
         */
        virtual int callDebugServer();

// TODO // currently no checks implemented.
        /** Checks correctness of parameters etc.
         *  Currently no checks implemented.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  my_rvalue
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    protected:
        //==========================
        // Data
        //==========================

        int *my_argc;
        char ***my_argv;
        int my_rvalue;

        /** Used to disable calling PMPI_Init in the constructor. */
        static bool our_call_init;

    };                          // end class MPO_Init



    //*************************************************************************
    //
    //                          MPO_Finalize
    //
    /** This class represents MPI_Finalize. */
    class MPO_Finalize : public MPO_Environment
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** Default constructor. */
        MPO_Finalize();

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Finalize();

        //==========================
        // Methods
        //==========================

        /** Checks special prerequisites for MPI_Finalize.
         *  Checks whether
         *  - active requests are left
         *  - pending messages are left
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether requests are left.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_WARN_REQUESTS_LEFT
         */
        virtual int checkRequestsLeft();

        /** Checks whether pending messages are left.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_WARN_PENDING_MESSAGES_LEFT
         */
        virtual int checkPendingMessagesLeft();


        /** Checks wheter Datatypes are left.
         * Important: this causes no errors since, acording to the
         *            MPI-Standard, this is not necessary but it might
         * 			  help to find resource allocation Problems.
         *
         * @return unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_WARN_TYPE_LEFT
         */
        virtual int checkDatatypesLeft();


        /** Checks wheter Errorhandlers are left.
         * Important: this causes no errors since, acording to the
         *            MPI-Standard, this is not necessary but it might
         * 			  help to find resource allocation Problems.
         *
         * @return unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_WARN_ERRHANDLER_LEFT
         */
        virtual int checkErrhandlersLeft();


        /** Checks wheter Communicators are left.
         * Important: this causes no errors since, acording to the
         *            MPI-Standard, this is not necessary but it might
         * 			  help to find resource allocation Problems.
         *
         * @return unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_WARN_COMM_LEFT
         */
        virtual int checkCommsLeft();


        /** Checks wheter Groups are left.
         * Important: this causes no errors since, acording to the
         *            MPI-Standard, this is not necessary but it might
         * 			  help to find resource allocation Problems.
         *
         * @return unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_WARN_GROUP_LEFT
         */
        virtual int checkGroupsLeft();


        /** Checks wheter Operations are left.
         * Important: this causes no errors since, acording to the
         *            MPI-Standard, this is not necessary but it might
         * 			  help to find resource allocation Problems.
         *
         * @return unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_WARN_OP_LEFT
         */
        virtual int checkOperationsLeft();

        /** Prints files containing timed information about resource ussage.
         */
        virtual void printResourceGraphs();

        /**
         * If the MPI Implementation used supports the call MPI_Is_thread_main
         * this method checks whether the thread executing MPI_Finalize is the
         * main thread or not. If not a Warning is created.
         * @return MARMOT_SUCCESS or MARMOT_WARN_FINALIZE_NOT_THREAD_MAIN.
         */
        virtual int checkIsThreadMain (void);

        /**
         * Checks whether there are other Threads still performing MPI-Calls.
         * @return MARMOT_SUCCESS or MARMOT_WARN_FINALIZE_OTHER_THREADS_WORKING.
         */
        virtual int checkOtherThreadsDone(void);

        /**
         * Prints information about the minimal threadlevel used for this application.
         * It also describes why this threadlevel is necessary by listing the violations
         * to the lower threadlevels.
         * @return MARMOT_SUCCESS.
         */
        virtual int logMinimalThreadLevel(void);

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Finalize
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:

    };                          // end class MPO_Finalize



    //*************************************************************************
    //
    //                          MPO_Initialized
    //
    /** This class represents MPI_Initialized. */
    class MPO_Initialized : public MPO_Environment
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Initialized();

        /** Constructor taking proper arguments. */
        MPO_Initialized(int *flag);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Initialized();

        //==========================
        // Methods
        //==========================

// TODO // currently no checks implemented.
        /** Checks correctness of parameters etc.
         *  Currently no checks implemented.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();


        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Initialized.
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        int *my_flag;

    };                          // end class MPO_Initialized



    //=========================================================================
    //                          Communicatorhandlers
    //=========================================================================



    //*************************************************************************
    //
    //                          MPO_Comm_size
    //
    /** This class represents MPI_Comm_size. */
    class MPO_Comm_size : public MPO_Communicatorhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Comm_size();

        /** Constructor taking proper arguments. */
        MPO_Comm_size(MPO_Comm communicator, int *size);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Comm_size();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Checks correctness of parameters etc.
          *
          *  @return  0
          */
         virtual int checkPrerequisites();

         /** Does the MPI call after checking it.
          *
          *  @return  PMPI_Comm_size
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Comm my_comm;
        int *my_size;

    };                          // end class MPO_Comm_size



    //*************************************************************************
    //
    //                          MPO_Comm_rank
    //
    /** This class represents MPI_Comm_rank. */
    class MPO_Comm_rank : public MPO_Communicatorhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Comm_rank();

        /** Constructor taking proper arguments. */
        MPO_Comm_rank(MPO_Comm communicator, int *rank);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Comm_rank();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Checks correctness of parameters etc.
          *
          *  @return  0
          */
         virtual int checkPrerequisites();

         /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Comm_rank
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Comm my_comm;
        int *my_rank;

    };                          // end class MPO_Comm_rank



    //*************************************************************************
    //
    //                          MPO_Comm_compare
    //
    /** This class represents MPI_Comm_compare. */
    class MPO_Comm_compare : public MPO_Communicatorhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Comm_compare();

        /** Constructor taking proper arguments. */
        MPO_Comm_compare(MPO_Comm comm, MPO_Comm comm2, int *result);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Comm_compare();

        //==========================
        // Methods
        //==========================

        /** Checks whether second communicator is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_NULL
         *           - MARMOT_ERR_COMM_NOT_VALID
         */
        virtual int checkCommSecond();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Checks correctness of parameters etc.
          *
          *  @return  0
          */
         virtual int checkPrerequisites();

         /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Comm_compare
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Comm my_comm;
        MPO_Comm my_comm2;
        int *my_result;

    };                          // end class MPO_Comm_compare



    //*************************************************************************
    //
    //                          MPO_Comm_dup
    //
    /** This class represents MPI_Comm_dup. */
    class MPO_Comm_dup : public MPO_Communicatorhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Comm_dup();

        /** Constructor taking proper arguments. */
        MPO_Comm_dup(MPO_Comm comm, MPI_Comm * newcomm);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Comm_dup();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Comm_dup
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Comm my_comm;
        MPI_Comm * my_newcomm;

    };                          // end class MPO_Comm_dup



    //*************************************************************************
    //
    //                          MPO_Comm_create
    //
    /** This class represents MPI_Comm_create. */
    class MPO_Comm_create : public MPO_Communicatorhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Comm_create();

        /** Constructor taking proper arguments. */
        MPO_Comm_create(MPO_Comm comm, MPO_Group group, MPI_Comm * newcomm);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Comm_create();

        //==========================
        // Methods
        //==========================

        /** Checks whether group is valid and subset of communicator.
         */
        virtual int checkGroup();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Checks correctness of parameters etc.
          *
          *  @return  0
          */
         virtual int checkPrerequisites();

         /** Does the MPI call after checking it.
          *
          *  @return  PMPI_Comm_create
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Comm my_comm;
        MPO_Group my_group;
        MPI_Comm *my_newcomm;

    };                          // end class MPO_Comm_create



    //*************************************************************************
    //
    //                          MPO_Comm_split
    //
    /** This calls represents MPI_Comm_split. */
    class MPO_Comm_split : public MPO_Communicatorhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Comm_split();

        /** Constructor taking proper arguments. */
        MPO_Comm_split(MPO_Comm comm, int color, int key, MPI_Comm * newcomm);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Comm_split();

        //==========================
        // Methods
        //==========================

        /** Checks color. */
        virtual int checkColor();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Comm_split
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        int my_color;
        int my_key;
        MPI_Comm *my_newcomm;

    };                          // end class MPO_Comm_split



    //*************************************************************************
    //
    //                          MPO_Comm_free
    //
    /** This class represents MPI_Comm_free. */
    class MPO_Comm_free : public MPO_Communicatorhandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Comm_free();

        /** Constructor taking proper arguments. */
        MPO_Comm_free(MPI_Comm * comm);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Comm_free();

        //==========================
        // Methods
        //==========================

        /** Checks additional prerequisites only for this call.
         *  Checks whether
         *  - communicator is ready to free
         *  - erases communicator from our map
         *  - erases errhandler from our map
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether communicator is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_NOT_VALID
         *           - MARMOT_WARN_COMM_NULL
         */
        virtual int checkCommFreed();

        /** Checks whether error-handler is ready to free and erases it.
         *
         *  @return   unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         */
        virtual int eraseErrhandler();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Comm_free
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        MPI_Comm * my_comm;

    };                          // end class MPO_Comm_free



    //=========================================================================
    //                          Blocking
    //=========================================================================



    //*************************************************************************
    //
    //                          MPO_Send
    //
    /**  This class represents MPI_Send. */
    class MPO_Send : public MPO_Blocking
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Send();

        /** Constructor taking proper arguments. */
        MPO_Send(void *buf, int count, MPO_Datatype datatype, int dest,
                 int tag, MPO_Comm comm);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Send();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Send
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_buf;
        // int my_count;
        //MPO_Datatype my_datatype;
        // int my_dest;
        // int my_tag;
        //MPO_Comm my_comm;

    };                          // end class MPO_Send



    //*************************************************************************
    //
    //                          MPO_Recv
    //
    /** This class represents MPI_Recv. */
    class MPO_Recv : public MPO_Blocking
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Recv();

        /** Constructor taking proper arguments. */
        MPO_Recv(void *buf, int count, MPO_Datatype datatype, int source,
                 int tag, MPO_Comm comm, MPI_Status * status);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Recv();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Recv
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_buf;
        // int my_count;
        //MPO_Datatype my_datatype;
        // int my_source;
        // int my_tag;
        //MPO_Comm my_comm;
        MPI_Status *my_status;

    };                          // end class MPO_Recv



    //*************************************************************************
    //
    //                          MPO_Ssend
    //
    /** This class represents MPI_Ssend. */
    class MPO_Ssend : public MPO_Blocking
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Ssend();

        /** Constructor taking proper arguments. */
        MPO_Ssend(void *buf, int count, MPO_Datatype datatype, int dest,
                  int tag, MPO_Comm comm);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Ssend();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Ssend
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_buf;
        // int my_count;
        //MPO_Datatype my_datatype;
        // int my_dest;
        // int my_tag;
        //MPO_Comm my_comm;

    };                          // end class MPO_Ssend



    //*************************************************************************
    //
    //                          MPO_Bsend
    //
    /** This class represents MPI_Bsend. */
    class MPO_Bsend : public MPO_Blocking
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Bsend();

        /** Constructor taking proper arguments. */
        MPO_Bsend(void *buf, int count, MPO_Datatype datatype, int dest,
                  int tag, MPO_Comm comm);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Bsend();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Bsend
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_buf;
        // int my_count;
        //MPO_Datatype my_datatype;
        // int my_dest;
        // int my_tag;
        //MPO_Comm my_comm;

    };                          // end class MPO_Bsend



    //*************************************************************************
    //
    //                          MPO_Buffer_attach
    //
    /** This class represents MPI_Buffer_attach. */
    class MPO_Buffer_attach : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Buffer_attach();

        /** Constructor taking proper arguments. */
        MPO_Buffer_attach(void *buf, int size);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Buffer_attach();

        //==========================
        // Methods
        //==========================

        /** Checks corrrectness of
         *  - buffer
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether buffer is attached.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_BUFFER_ATTACHED
         */
        virtual int checkBufferAttached();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Buffer_attach
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_buf;
        int my_size;

    };                          // end class MPO_Buffer_attach



    //*************************************************************************
    //
    //                          MPO_Buffer_detach
    //
    /** This class represents MPI_Buffer_detach. */
    class MPO_Buffer_detach : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Buffer_detach();

        /** Constructor taking proper arguments. */
        MPO_Buffer_detach(void *buf, int *size);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Buffer_detach();

        //==========================
        // Methods
        //==========================

        /** Checks corrrectness of
         *  - buffer
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether buffer is attached.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_BUFFER_NOT_ATTACHED
         */
        virtual int checkBufferAttached();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Buffer_detach
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_buf;
        int *my_size;

    };                          // end class MPO_Buffer_detach



    //*************************************************************************
    //
    //                          MPO_Rsend
    //
    /** This class represents MPI_Rsend. */
    class MPO_Rsend : public MPO_Blocking
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Rsend();

        /** Constructor taking proper arguments. */
        MPO_Rsend(void *buf, int count, MPO_Datatype datatype, int dest,
                  int tag, MPO_Comm comm);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Rsend();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Rsend
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_buf;
        // int my_count;
        //MPO_Datatype my_datatype;
        // int my_dest;
        // int my_tag;
        //MPO_Comm my_comm;

    };                          // end class MPO_Rsend



    //=========================================================================
    //                          Sendrecvs
    //=========================================================================



    //*************************************************************************
    //
    //                          MPO_Sendrecv
    //
    /** This class represents MPI_Sendrecv. */
    class MPO_Sendrecv : public MPO_Sendrecvs
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Sendrecv();

        /** Constructor taking proper arguments. */
        MPO_Sendrecv(void *sendbuf, int sendcount,
                     MPO_Datatype datatype, int dest, int sendtag,
                     void *recvbuf, int recvcount, MPO_Datatype recvtype,
                     int source, int recvtag, MPO_Comm comm,
                     MPI_Status * status);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Sendrecv();

        //==========================
        // Methods
        //==========================

        /** Checks additional prerequisites only for this call.
         *  Checks whether count and datatype are valid.
         */
        virtual int checkPrerequisites2();

        /** Checks whether count is valid.
         *
         *  @param   KIND_OF_COUNT  specifies the kind of count for output
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COUNT_NEG
         *           - MARMOT_WARN_COUNT_ZERO
         */
        virtual int checkCountOfReceiver(const string KIND_OF_COUNT);

        /** Checks whether datatype is valid.
         */
        virtual int checkDatatypeOfReceiver();

         /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

         /** Checks correctness of parameters etc.
          *
          *  @return  0
          */
         virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Sendrecv
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_sendbuf;
        // int my_count;
        //MPO_Datatype my_sendtype;
        // int my_dest;
        // int my_sendtag;
        void *my_recvbuf;
        int my_recvcount;
        MPO_Datatype my_recvtype;
        // int my_source;
        // int my_recvtag;
        //MPO_Comm my_comm;
        MPI_Status *my_status;

    };                          // end class MPO_Sendrecv



    //*************************************************************************
    //
    //                          MPO_Sendrecv_replace
    //
    /** This class represents MPI_Sendrecv_replace. */
    class MPO_Sendrecv_replace : public MPO_Sendrecvs
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Sendrecv_replace();

        /** Constructor taking proper arguments. */
        MPO_Sendrecv_replace(void *buf, int count, MPO_Datatype datatype,
                             int dest,
                             int sendtag, int source, int recvtag,
                             MPO_Comm comm,
                             MPI_Status * status);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Sendrecv_replace();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

         /** Checks correctness of parameters etc.
          *
          *  @return  0
          */
         virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Sendrecv_replace
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_buf;
        // int my_count;
        //MPO_Datatype my_datatype;
        // int my_dest;
        // int my_sendtag;
        // int my_source;
        // int my_recvtag;
        //MPO_Comm my_comm;
        MPI_Status *my_status;

    };                          // end class MPO_Sendrecv_replace



    //=========================================================================
    //                          Collective
    //=========================================================================



    //*************************************************************************
    //
    //                          MPO_Barrier
    //
    /** This class represents MPI_Barrier. */
    class MPO_Barrier : public MPO_Collective
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Barrier();

        /** Constructor taking proper arguments. */
        MPO_Barrier(MPO_Comm comm);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Barrier();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();


        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Barrier
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Comm my_comm;

    };                          // end class MPO_Barrier



    //*************************************************************************
    //
    //                          MPO_Bcast
    //
    /** This class represents MPI_ Bcast. */
    class MPO_Bcast : public MPO_Collective
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Bcast();

        /** Constructor taking proper arguments. */
        MPO_Bcast(void *buf, int count, MPO_Datatype datatype, int root,
                  MPO_Comm comm);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Bcast();

        //==========================
        // Methods
        //==========================

        /** Checks additional prerequisites only for this call.
         *  Checks correctness of
         *  - count
         *  - datatype
         *  - root
         *
         *  @return  0
         */
        virtual int checkPrerequisites2();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Bcast
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_buf;
        int my_count;
        MPO_Datatype my_datatype;
        int my_root;
        //MPO_Comm my_comm;

    };                          // end  class MPO_Bcast



    //*************************************************************************
    //
    //                          MPO_Gather
    //
    /** This class represents MPI_ Gather. */
    class MPO_Gather : public MPO_Collective
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Gather();

        /** Constructor taking proper arguments. */
        MPO_Gather(void *sendbuf, int sendcount, MPO_Datatype sendtype,
                   void *recvbuf, int recvcount, MPO_Datatype recvtype,
                   int root, MPO_Comm comm);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Gather();

        //==========================
        // Methods
        //==========================

        /** Checks additional prerequisites only for this call.
         *  Checks correctness of
         *  - count
         *  - datatype
         *  - root
         *
         *  @return  0
         */
        virtual int checkPrerequisites2();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

         /** Checks correctness of parameters etc.
          *
          *  @return  0
          */
         virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Gather
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_sendbuf;
        int my_sendcount;
        MPO_Datatype my_sendtype;
        void *my_recvbuf;
        int my_recvcount;
        MPO_Datatype my_recvtype;
        int my_root;
        //MPO_Comm my_comm;

    };                          // end class MPO_Gather



    //*************************************************************************
    //
    //                          MPO_Gatherv
    //
    /** This class represents MPI_ Gatherv. */
    class MPO_Gatherv : public MPO_Collective
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Gatherv();

        /** Constructor taking proper arguments. */
        MPO_Gatherv(void *sendbuf, int sendcount, MPO_Datatype sendtype,
                    void *recvbuf, int *recvcounts, int *displs,
                    MPO_Datatype recvtype, int root, MPO_Comm comm);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Gatherv();

        //==========================
        // Methods
        //==========================

        /** Checks additional prerequisites only for this call.
         *  Checks whether
         *  - count
         *  - datatype
         *  - displacements
         *  - root
         *  are valid.
         *
         *  @return  0
         */
        virtual int checkPrerequisites2();

        /** Checks whether datatype is valid.
         */
        virtual int checkDatatypeOfReceiver();

        /** Checks array of displacements.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES:
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_DISPLACEMENT_OVERRIDES_DATA
         *           - MARMOT_NOTE_DISPLACEMENT_OVERRIDES_DATA
         */
        virtual int checkDisplacementArrayOfRoot();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Gatherv
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_sendbuf;
        int my_sendcount;
        MPO_Datatype my_sendtype;
        void *my_recvbuf;
        int *my_recvcounts;
        int *my_displs;
        MPO_Datatype my_recvtype;
        int my_root;
        //MPO_Comm my_comm;

    };                          // end class MPO_Gatherv



    //*************************************************************************
    //
    //                          MPO_Allgather
    //
    /** This class represents MPI_ Allgather. */
    class MPO_Allgather : public MPO_Collective
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Allgather();

        /** Constructor taking proper arguments. */
        MPO_Allgather(void *sendbuf, int sendcount, MPO_Datatype sendtype,
                      void *recvbuf, int recvcount, MPO_Datatype recvtype,
                      MPO_Comm comm);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Allgather();

        //==========================
        // Methods
        //==========================

        /** Checks additional prerequisites only for this call.
         *  Checks correctness of
         *  - count
         *  - datatype
         *
         *  @return  0
         */
        virtual int checkPrerequisites2();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Allgather
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_sendbuf;
        int my_sendcount;
        MPO_Datatype my_sendtype;
        void *my_recvbuf;
        int my_recvcount;
        MPO_Datatype my_recvtype;
        //MPO_Comm my_comm;

    };                          // end class MPO_Allgather



    //*************************************************************************
    //
    //                          MPO_Allgatherv
    //
    /** This class represents MPI_ Allgatherv. */
    class MPO_Allgatherv : public MPO_Collective
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Allgatherv();

        /** Constructor taking proper arguments. */
        MPO_Allgatherv(void *sendbuf, int sendcount, MPO_Datatype sendtype,
                       void *recvbuf, int *recvcounts, int *displs,
                       MPO_Datatype recvtype, MPO_Comm comm);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Allgatherv();

        //==========================
        // Methods
        //==========================

        /** Checks additional prerequisites only for this call.
         *  Checks correctness of displs, datatype and count.
         */
        virtual int checkPrerequisites2();

        /** Checks array of displacements.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES:
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_DISPLACEMENT_OVERRIDES_DATA
         *           - MARMOT_NOTE_DISPLACEMENT_OVERRIDES_DATA
         */
        virtual int checkDisplacementArrayOfReceiver();

        /** Checks array of receive datatypes. */
        virtual int checkDatatypeOfReceiver();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();


        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Allgatherv
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_sendbuf;
        int my_sendcount;
        MPO_Datatype my_sendtype;
        void *my_recvbuf;
        int *my_recvcounts;
        int *my_displs;
        MPO_Datatype my_recvtype;
        //MPO_Comm my_comm;

    };                          // end class MPO_Allgatherv



    //*************************************************************************
    //
    //                          MPO_Reduce
    //
    /** This class represents MPI_ Reduce. */
    class MPO_Reduce : public MPO_Collective
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Reduce();

        /** Constructor taking proper arguments. */
        MPO_Reduce(void *sendbuf, void *recvbuf, int count,
                   MPO_Datatype datatype, MPO_Op op, int root, MPO_Comm comm);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Reduce();

        //==========================
        // Methods
        //==========================

        /** Checks additional prerequisites only for this call.
         *  Checks correctness of
         *  - count
         *  - datatype
         *  - root
         *  - operation
         *
         *  @return  0
         */
        virtual int checkPrerequisites2();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Reduce
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_sendbuf;
        void *my_recvbuf;
        int my_count;
        MPO_Datatype my_datatype;
        MPO_Op my_op;
        int my_root;
        //MPO_Comm my_comm;

    };                          // end class MPO_Reduce



    //*************************************************************************
    //
    //                          MPO_Allreduce
    //
    /** This class represents MPI_ Allreduce. */
    class MPO_Allreduce : public MPO_Collective
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Allreduce();

        /** Constructor taking proper arguments. */
        MPO_Allreduce(void *sendbuf, void *recvbuf, int count,
                      MPO_Datatype datatype, MPO_Op op, MPO_Comm comm);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Allreduce();

        //==========================
        // Methods
        //==========================

        /** Checks additional prerequisites only for this call.
         *  Checks correctness of
         *  - count
         *  - datatype
         *  - operation
         *
         *  @return  0
         */
        virtual int checkPrerequisites2();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Allreduce
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_sendbuf;
        void *my_recvbuf;
        int my_count;
        MPO_Datatype my_datatype;
        MPO_Op my_op;
        //MPO_Comm my_comm;

    };                          // end class MPO_Allreduce



    //*************************************************************************
    //
    //                          MPO_Scatter
    //
    /** This class represents MPI_ Scatter. */
    class MPO_Scatter : public MPO_Collective
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Scatter();

        /** Constructor taking proper arguments. */
        MPO_Scatter(void *sendbuf, int sendcount, MPO_Datatype sendtype,
                    void *recvbuf, int recvcount, MPO_Datatype recvtype,
                    int root, MPO_Comm comm);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Scatter();

        //==========================
        // Methods
        //==========================

         /** Checks additional prerequisites only for this call.
         *  Checks correctness of
         *  - count
         *  - datatype
         *  - root
         *
         *  @return  0
         */
        virtual int checkPrerequisites2();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Scatter
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_sendbuf;
        int my_sendcount;
        MPO_Datatype my_sendtype;
        void *my_recvbuf;
        int my_recvcount;
        MPO_Datatype my_recvtype;
        int my_root;
        //MPO_Comm my_comm;

    };                          // end class MPO_Scatter



    //*************************************************************************
    //
    //                          MPO_Scatterv
    //
    /** This class represents MPI_ Scatterv. */
    class MPO_Scatterv : public MPO_Collective
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Scatterv();

        /** Constructor taking proper arguments. */
        MPO_Scatterv(void *sendbuf, int *sendcounts, int *displs,
                     MPO_Datatype sendtype, void *recvbuf, int recvcount,
                     MPO_Datatype recvtype, int root, MPO_Comm comm);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Scatterv();

        //==========================
        // Methods
        //==========================

        /** Checks additional prerequisites only for this call.
         *  Checks correctness of
         *  - displacements
         *  - count
         *  - datatype
         *  - root
         *
         *  @return  0
         */
        virtual int checkPrerequisites2();

        /** Checks array of displacements.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES:
         *           - MARMOT_SUCCESS
         *           - MARMOT_NOTE_DISPLACEMENT_DATA_SENT_TWICE
         *           - MARMOT_NOTE_DISPLACEMENT_READS_DATA
         */
        virtual int checkDisplacementArrayOfSender();

        virtual int checkDatatypeArrayOfSender();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();


        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Scatterv
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_sendbuf;
        int *my_sendcounts;
        int *my_displs;
        MPO_Datatype my_sendtype;
        void *my_recvbuf;
        int my_recvcount;
        MPO_Datatype my_recvtype;
        int my_root;
        //MPO_Comm my_comm;

    };                          // end class MPO_Scatterv



    //*************************************************************************
    //
    //                          MPO_Alltoall
    //
    /** This class represents MPI_ Alltoall. */
    class MPO_Alltoall : public MPO_Collective
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Alltoall();

        /** Constructor taking proper arguments. */
        MPO_Alltoall(void *sendbuf, int sendcount, MPO_Datatype sendtype,
                     void *recvbuf, int recvcount, MPO_Datatype recvtype,
                     MPO_Comm comm);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Alltoall();

        //==========================
        // Methods
        //==========================

        /** Checks correctness of parameters.
         *  Checks if
         *  - count
         *  - datatype
         *  are valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_WARN_COUNTS_NOT_EQUAL
         */
        virtual int checkPrerequisites2();

       /** Checks whether send and receive count are equal.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_WARN_COUNTS_NOT_EQUAL
         */
        virtual int checkCountsAreEqual();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Alltoall
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_sendbuf;
        int my_sendcount;
        MPO_Datatype my_sendtype;
        void *my_recvbuf;
        int my_recvcount;
        MPO_Datatype my_recvtype;
        //MPO_Comm my_comm;

    };                          // end class MPO_Alltoall



    //*************************************************************************
    //
    //                          MPO_Alltoallv
    //
    /** This class represents MPI_ Alltoallv. */
    class MPO_Alltoallv : public MPO_Collective
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Alltoallv();

        /** Constructor taking proper arguments. */
        MPO_Alltoallv(void *sendbuf, int *sendcounts, int *sdispls,
                      MPO_Datatype sendtype, void *recvbuf, int *recvcounts,
                      int *rdispls, MPO_Datatype recvtype, MPO_Comm comm);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Alltoallv();

        //==========================
        // Methods
        //==========================

        /** Checks whether sizes match. */
        virtual int checkPrerequisites2(int *counts,
                                                 MPO_Datatype datatype,
                                               int type);

        /** Checks array of displacements.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES:
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_DISPLACEMENT_OVERRIDES_DATA
         *           - MARMOT_NOTE_DISPLACEMENT_READS_DATA
         */
        virtual int checkDisplacementArrayOfReceiver();

        virtual int checkDatatypeOfSender(int *counts, MPO_Datatype datatype,
                int type);

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Alltallv
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_sendbuf;
        int *my_sendcounts;
        int *my_sdispls;
        MPO_Datatype my_sendtype;
        void *my_recvbuf;
        int *my_recvcounts;
        int *my_rdispls;
        MPO_Datatype my_recvtype;
        //MPO_Comm my_comm;

    };                          // end class MPO_Alltoallv



    //*************************************************************************
    //
    //                          MPO_Reduce_scatter
    //
    /** This class represents MPI_ Reduce_scatter. */
    class MPO_Reduce_scatter : public MPO_Collective
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Reduce_scatter();

        /** Constructor taking proper arguments. */
        MPO_Reduce_scatter(void *sendbuf, void *recvbuf, int *recvcounts,
                           MPO_Datatype datatype, MPO_Op op, MPO_Comm comm);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Reduce_scatter();

        //==========================
        // Methods
        //==========================

        /** Checks correctness of parameters.
         *  Checks if
         *  - counts
         *  - operation
         *  are valid.
         *
         *  @return  0
         */
        virtual int checkPrerequisites2();

        virtual int checkDatatypeArray();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Reduce_scatter
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_sendbuf;
        void *my_recvbuf;
        int *my_recvcounts;
        MPO_Datatype my_datatype;
        MPO_Op my_op;
        //MPO_Comm my_comm;

    };                          // end class MPO_Reduce_scatter



    //*************************************************************************
    //
    //                          MPO_Scan
    //
    /** This class represents MPI_Scan. */
    class MPO_Scan : public MPO_Collective
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Scan();

        /** Constructor taking proper arguments. */
        MPO_Scan(void *sendbuf, void *recvbuf, int count,
                 MPO_Datatype datatype,
                 MPO_Op op, MPO_Comm comm);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Scan();

        //==========================
        // Methods
        //==========================

        /** Checks correctness of parameters.
         *  Checks if
         *  - count
         *  - datatype
         *  - operation
         *  are valid.
         *
         *  @return  0
         */
        virtual int checkPrerequisites2();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Scan
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_sendbuf;
        void *my_recvbuf;
        int my_count;
        MPO_Datatype my_datatype;
        MPO_Op my_op;
        //MPO_Comm my_comm;

    };                          // end  class MPO_Scan



    //=========================================================================
    //                          Initiator
    //=========================================================================



    //*************************************************************************
    //
    //                          MPO_Isend
    //
    /** This class represents MPI_Isend. */
    class MPO_Isend : public MPO_Initiator
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Isend();

        /** Constructor taking proper arguments. */
        MPO_Isend(void *buf, int count, MPO_Datatype datatype, int rank,
                  int tag, MPO_Comm comm, MPI_Request * request);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Isend();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Isend
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_buf;
        // int my_count;
        //MPO_Datatype my_datatype;
        // int my_dest;
        // int my_tag;
        //MPO_Comm my_comm;
        //MPI_Request *my_request_p;

    };                          // end class MPO_Isend



    //*************************************************************************
    //
    //                          MPO_Ibsend
    //
    /** This class represents MPI_Ibsend. */
    class MPO_Ibsend : public MPO_Initiator
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Ibsend();

        /** Constructor taking proper arguments. */
        MPO_Ibsend(void *buf, int count, MPO_Datatype datatype, int rank,
                   int tag, MPO_Comm comm, MPI_Request * request);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Ibsend();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Ibsend
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_buf;
        // int my_count;
        //MPO_Datatype my_datatype;
        // int my_dest;
        // int my_tag;
        //MPO_Comm my_comm;
        //MPI_Request *my_request_p;

    };                          // end class MPO_Ibsend



    //*************************************************************************
    //
    //                          MPO_Issend
    //
    /** This class represents MPI_Issend. */
    class MPO_Issend : public MPO_Initiator
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Issend();

        /** Constructor taking proper arguments. */
        MPO_Issend(void *buf, int count, MPO_Datatype datatype, int rank,
                   int tag, MPO_Comm comm, MPI_Request * request);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Issend();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Issend
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_buf;
        // int my_count;
        //MPO_Datatype my_datatype;
        // int my_dest;
        // int my_tag;
        //MPO_Comm my_comm;
        //MPI_Request *my_request_p;

    };                          // end class MPO_Issend



    //*************************************************************************
    //
    //                          MPO_Irsend
    //
    /** This class represents MPI_Irsend. */
    class MPO_Irsend : public MPO_Initiator
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Irsend();

        /** Constructor taking proper arguments. */
        MPO_Irsend(void *buf, int count, MPO_Datatype datatype, int rank,
                   int tag, MPO_Comm comm, MPI_Request * request);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Irsend();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Irsend
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_buf;
        // int my_count;
        //MPO_Datatype my_datatype;
        // int my_dest;
        // int my_tag;
        //MPO_Comm my_comm;
        // MPI_Request *my_request_p;

    };                          // end class MPO_Irsend



    //*************************************************************************
    //
    //                          MPO_Irecv
    //
    /** This class represents MPI_Irecv. */
    class MPO_Irecv : public MPO_Initiator
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Irecv();

        /** Constructor taking proper arguments. */
        MPO_Irecv(void *buf, int count, MPO_Datatype datatype, int rank,
                  int tag,
                  MPO_Comm comm, MPI_Request * request);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Irecv();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Irecv
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_buf;
        // int my_count;
        //MPO_Datatype my_datatype;
        // int my_source;
        // int my_tag;
        //MPO_Comm my_comm;
        //MPI_Request *my_request_p;

    };                          // end class MPO_Irecv



    //*************************************************************************
    //
    //                          MPO_Send_init
    //
    /** This class represents MPI_Send_init. */
    class MPO_Send_init : public MPO_Initiator
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Send_init();

        /** Constructor taking proper arguments. */
        MPO_Send_init(void *buf, int count, MPO_Datatype datatype, int rank,
                      int tag, MPO_Comm comm, MPI_Request * request);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Send_init();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Send_init
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_buf;
        // int my_count;
        //MPO_Datatype my_datatype;
        // int my_dest;
        // int my_tag;
        //MPO_Comm my_comm;
        //MPI_Request *my_request_p;

    };                          // end class MPO_Send_init



    //*************************************************************************
    //
    //                          MPO_Bsend_init
    //
    /** This class represents MPI_Bsend_init. */
    class MPO_Bsend_init : public MPO_Initiator
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Bsend_init();

        /** Constructor taking proper arguments. */
        MPO_Bsend_init(void *buf, int count, MPO_Datatype datatype, int rank,
                       int tag, MPO_Comm comm, MPI_Request * request);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Bsend_init();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Bsend_init
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_buf;
        // int my_count;
        //MPO_Datatype my_datatype;
        // int my_dest;
        // int my_tag;
        //MPO_Comm my_comm;
        //MPI_Request *my_request_p;

    };                          // end class MPO_Bsend_init



    //*************************************************************************
    //
    //                          MPO_Ssend_init
    //
    /** This class represents MPI_Ssend_init. */
    class MPO_Ssend_init : public MPO_Initiator
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Ssend_init();

        /** Constructor taking proper arguments. */
        MPO_Ssend_init(void *buf, int count, MPO_Datatype datatype, int rank,
                       int tag, MPO_Comm comm, MPI_Request * request);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Ssend_init();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Ssend_init
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_buf;
        // int my_count;
        //MPO_Datatype my_datatype;
        // int my_dest;
        // int my_tag;
        //MPO_Comm my_comm;
        //MPI_Request *my_request_p;

    };                          // end class MPO_Ssend_init



    //*************************************************************************
    //
    //                          MPO_Rsend_init
    //
    /** This class represents MPI_Rsend_init. */
    class MPO_Rsend_init : public MPO_Initiator
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Rsend_init();

        /** Constructor taking proper arguments. */
        MPO_Rsend_init(void *buf, int count, MPO_Datatype datatype, int rank,
                       int tag, MPO_Comm comm, MPI_Request * request);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Rsend_init();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Rsend_init
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_buf;
        // int my_count;
        //MPO_Datatype my_datatype;
        // int my_dest;
        // int my_tag;
        //MPO_Comm my_comm;
        //MPI_Request *my_request_p;

    };                          // end class MPO_Rsend_init



    //*************************************************************************
    //
    //                          MPO_Recv_init
    //
    /** This class represents MPI_Recv_init. */
    class MPO_Recv_init : public MPO_Initiator
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Recv_init();

        /** Constructor taking proper arguments. */
        MPO_Recv_init(void *buf, int count, MPO_Datatype datatype, int rank,
                      int tag, MPO_Comm comm, MPI_Request * request);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Recv_init();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Recv_init
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_buf;
        // int my_count;
        //MPO_Datatype my_datatype;
        // int my_source;
        // int my_tag;
        //MPO_Comm my_comm;
        //MPI_Request *my_request_p;

    };                          // end class MPO_Recv_init



    //=========================================================================
    //                          Completion
    //=========================================================================



    //*************************************************************************
    //
    //                          MPO_Wait
    //
    /** This class represents MPI_Wait. */
    class MPO_Wait : public MPO_Completion
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Wait();

        /** Constructor taking proper arguments. */
        MPO_Wait(MPI_Request * request, MPI_Status * status);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Wait();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Wait
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPI_Request *my_request_p;
        MPI_Status * my_status;

    };                          // end class MPO_Wait



    //*************************************************************************
    //
    //                          MPO_Waitall
    //
    /** This class represents MPI_Waitall. */
    class MPO_Waitall : public MPO_Completion
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Waitall();

        /** Constructor taking proper arguments. */
        MPO_Waitall(int count, MPI_Request * request_array,
                    MPI_Status * status_array);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Waitall();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Waitall
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        // int my_count;
        //MPI_Request *my_request_array;
        MPI_Status * my_status_array;

    };                          // end class MPO_Waitall



    //*************************************************************************
    //
    //                          MPO_Waitany
    //
    /** This class represents MPI_Waitany. */
    class MPO_Waitany : public MPO_Completion
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Waitany();

        /** Constructor taking proper arguments. */
        MPO_Waitany(int count, MPI_Request * request_array, int *index,
                    MPI_Status * status_array);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Waitany();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Waitany
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /**
         * @see MPO_Callobject::setPreData
         */
        void setPreData (MPO_Callobject *obj);

        /**
         * @see MPO_Callobject::clonePreData
         */
        MPO_Callobject* clonePreData (void);

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        // int my_count;
        //MPI_Request *my_request_array;
        MPI_Request *my_temp_request_array;
        int *my_index;
        MPI_Status *my_status_array;

    };                          // end class MPO_Waitany



    //*************************************************************************
    //
    //                          MPO_Waitsome
    //
    /** This class represents MPI_Waitsome. */
    class MPO_Waitsome : public MPO_Completion
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Waitsome();

        /** Constructor taking proper arguments. */
        MPO_Waitsome(int incount, MPI_Request * request_array, int *outcount,
                     int *index_array, MPI_Status * status_array);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Waitsome();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Waitsome
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /**
         * @see MPO_Callobject::setPreData
         */
        void setPreData (MPO_Callobject *obj);

        /**
         * @see MPO_Callobject::clonePreData
         */
        MPO_Callobject* clonePreData (void);

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        // int my_count;
        //MPI_Request *my_request_array;
        MPI_Request *my_temp_request_array;
        int *my_outcount;
        int *my_index_array;
        MPI_Status *my_status_array;

    };                          // end class MPO_Waitsome



    //*************************************************************************
    //
    //                          MPO_Cancel
    //
    /** This class represents MPI_Cancel. */
    class MPO_Cancel : public MPO_Completion
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Cancel();

        /** Constructor taking proper arguments. */
        MPO_Cancel(MPI_Request * request);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Cancel();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Cancel
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPI_Request *my_request_p;

    };                          // end class MPO_Cancel



    //*************************************************************************
    //
    //                          MPO_Test_cancelled
    //
    /** This class represents MPI_Test_cancelled. */
    class MPO_Test_cancelled : public MPO_Completion
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Test_cancelled();

        /** Constructor taking proper arguments. */
        MPO_Test_cancelled(MPI_Status * status, int *flag);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Test_cancelled();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Test_cancelled
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        MPI_Status * my_status;
        int *my_flag;

    };                          // end class MPO_Test_cancelled



    //*************************************************************************
    //
    //                          MPO_Request_free
    //
    /** This class represents MPI_Request_free. */
    class MPO_Request_free : public MPO_Completion
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Request_free();

        /** Constructor taking proper arguments.
         *  The  constructor performs the following tasks:
         *  1. Check whether request is in our_requests
         *     - no: print warning to stderr
         *     - yes:
         *  2. Remove request from our_requests
         *
         *  We store the pointer and not the request itself.
         *  Because the request may change during the message transfer
         *  (it is a opaque type and thus  may contain some internal status
         *  etc.)
        */
        MPO_Request_free(MPI_Request * request);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Request_free();

        //==========================
        // Methods
        //==========================

        /** Checks prerequisites such as correctness of parameters.
         *  Checks whether request is valid.
         */
        virtual int checkPrerequisites();

        /** Checks additional prerequisites only for this call.
         *  Here we have to check in another way than in the other
         *  Completion-calls:
         *  The request we want to free should be INactive and may be
         *  persistent.
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_MARMOT
         *           - MARMOT_WARN_REQUEST_OUTCOUNT_UNDEFINED
         */
        virtual int checkRequest();

        /** Removes request from our_requests and frees it
         *  for further use.
         *  We can't use the eraseRequest from MPO_Completion:
         *  we need no arguments in this case, because we have to erase
         *  the request,
         *  no matter if it's persistent or not, and only one per call.
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_REQUEST_NOT_VALID
         */
        virtual int eraseRequest();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Request_free
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPI_Request *my_request_p;

    };                          // end class MPO_Request_free



    //=========================================================================
    //                          Tests
    //=========================================================================



    //*************************************************************************
    //
    //                          MPO_Test
    //
    /** This class represents MPI_Test. */
    class MPO_Test : public MPO_Tests
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Test();

        /** Constructor taking proper arguments. */
        MPO_Test(MPI_Request * request, int *flag, MPI_Status * status);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Test();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Test
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /**
         * @see MPO_Callobject::setPreData
         */
        void setPreData (MPO_Callobject *obj);

        /**
         * @see MPO_Callobject::clonePreData
         */
        MPO_Callobject* clonePreData (void);

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPI_Request *my_request_p;
        // int *my_flag;
        MPI_Request my_temp_request;
        MPI_Status * my_status;

    };                          // end class MPO_Test



    //*************************************************************************
    //
    //                          MPO_Testany
    //
    /** This class represents MPI_Testany. */
    class MPO_Testany : public MPO_Tests
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Testany();

        /** Constructor taking proper arguments. */
        MPO_Testany(int count, MPI_Request * request_array, int *index,
                    int *flag, MPI_Status * status);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Testany();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Testany
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /**
         * @see MPO_Callobject::setPreData
         */
        void setPreData (MPO_Callobject *obj);

        /**
         * @see MPO_Callobject::clonePreData
         */
        MPO_Callobject* clonePreData (void);

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        // int count;
        //MPI_Request *my_request_array;
        MPI_Request *my_temp_request_array;
        int *my_index;
        // int*  my_flag;
        MPI_Status *my_status;

    };                          // end class MPO_Testany



    //*************************************************************************
    //
    //                          MPO_Testall
    //
    /** This class represents MPI_Testall. */
    class MPO_Testall : public MPO_Tests
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Testall();

        /** Constructor taking proper arguments. */
        MPO_Testall(int count, MPI_Request * request_array, int *flag,
                    MPI_Status * status);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Testall();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Testall
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /**
         * @see MPO_Callobject::setPreData
         */
        void setPreData (MPO_Callobject *obj);

        /**
         * @see MPO_Callobject::clonePreData
         */
        MPO_Callobject* clonePreData (void);

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        // int count;
        //MPI_Request *my_request_array;
        MPI_Request *my_temp_request_array;
        // int *my_flag;
        MPI_Status * my_status;

    };                          // end class MPO_Testall



    //*************************************************************************
    //
    //                          MPO_Testsome
    //
    /** This class represents MPI_Testsome. */
    class MPO_Testsome : public MPO_Tests
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Testsome();

        /** Constructor taking proper arguments. */
        MPO_Testsome(int incount, MPI_Request * request_array, int *outcount,
                     int *index_array, MPI_Status * status);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Testsome();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Testsome
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /**
         * @see MPO_Callobject::setPreData
         */
        void setPreData (MPO_Callobject *obj);

        /**
         * @see MPO_Callobject::clonePreData
         */
        MPO_Callobject* clonePreData (void);

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        // int incount;
        //MPI_Request *my_request_array;
        MPI_Request *my_temp_request_array;
        int *my_outcount;
        int *my_index_array;
        MPI_Status *my_status;

    };                          // end class MPO_Testsome



    //=========================================================================
    //                          Starts
    //=========================================================================



    //*************************************************************************
    //
    //                          MPO_Start
    //
    /** This class represents MPI_Start. */
    class MPO_Start : public MPO_Starts
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Start();

        /** Constructor taking proper arguments. */
        MPO_Start(MPI_Request * request);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Start();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Start
         */
         virtual int callNativeMPI();

         /** Checks correctness of parameters etc.
          *  Executed after the actual MPI call.
          *
          *  @return  0
          */
         virtual int checkPost();

         /** Checks correctness of parameters etc.
          *
          *  @return  0
          */
         virtual int checkPrerequisites();

         /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPI_Request *my_request_p;

    };                          // end class MPO_Start



    //*************************************************************************
    //
    //                          MPO_Startall
    //
    /** This class represents MPI_Startall. */
    class MPO_Startall : public MPO_Starts
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Startall();

        /** Constructor taking proper arguments. */
        MPO_Startall(int count, MPI_Request * request_array);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Startall();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Startall
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        // int my_count;
        //MPI_Request *my_request_array;

    };                          // end class MPO_Startall



    //=========================================================================
    //                          Cartesian
    //=========================================================================



    //*************************************************************************
    //
    //                          MPO_Cart_create
    //
    /** This class represents MPI_Cart_create. */
    class MPO_Cart_create : public MPO_Cartesian
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Cart_create();

        /** Constructor taking proper arguments. */
        MPO_Cart_create(MPO_Comm comm_old, int ndims, int *dims, int *periods,
                        int reorder, MPI_Comm * comm_cart);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Cart_create();

        //==========================
        // Methods
        //==========================

        /** Checks correctness of parameters.
         *  Checks if
         *  - communicator is valid
         *  - size of input parameter matches size of new topology.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether communicator is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_NULL
         *           - MARMOT_ERR_COMM_NOT_VALID
         */
        virtual int checkComm();

        /** Checks whether new topology is correct.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_TOPOLOGY_CART_NEW_GT_COMM_OLD
         *           - MARMOT_WARN_TOPOLOGY_CART_NEW_LT_COMM_OLD
         */
        virtual int checkTopologyCartNew();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Cart_create
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        MPO_Comm my_comm_old;
        int my_ndims;
        int *my_dims;
        int *my_periods;
        int my_reorder;
        MPI_Comm *my_comm_cart;

    };                          // end  class MPO_Cart_create



    //*************************************************************************
    //
    //                          MPO_Dims_create
    //
    /** This class represents MPI_Dims_create. */
    class MPO_Dims_create : public MPO_Cartesian
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Dims_create();

        /** Constructor taking proper arguments. */
        MPO_Dims_create(int nnodes, int ndims, int *dims);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Dims_create();

        //==========================
        // Methods
        //==========================

        /** Checks if size of input parameter matches size of new topology.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks array of dimensions.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES:
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_DIM_NEG
         *           - MARMOT_ERR_DIM_TOO_BIG
         *           - MARMOT_ERR_NODES_NEQ_PRODUCT_DIMS
         *           - MARMOT_WARN_DIM_ZERO
         */
        virtual int checkDimsArray();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Dims_create
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        int my_nnodes;
        int my_ndims;
        int *my_dims;

    };                          // end class MPO_Dims_create



    //*************************************************************************
    //
    //                          MPO_Cartdim_get
    //
    /** This class represents MPI_Cartdim_get.*/
    class MPO_Cartdim_get : public MPO_Cartesian
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Cartdim_get();

        /** Constructor taking proper arguments. */
        MPO_Cartdim_get(MPO_Comm comm, int *ndims);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Cartdim_get();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Cartdim_get
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Comm my_comm;
        int *my_ndims;

    };                          // end class MPO_Cartdim_get



    //*************************************************************************
    //
    //                          MPO_Cart_get
    //
    /** This class represents MPI_Cart_get. */
    class MPO_Cart_get : public MPO_Cartesian
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Cart_get();

        /** Constructor taking proper arguments. */
        MPO_Cart_get(MPO_Comm comm, int maxdims, int *dims, int *periods,
                     int *coords);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Cart_get();

        //==========================
        // Methods
        //==========================

        /** Checks dimensions.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES:
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_DIM_NEG
         *           - MARMOT_WARN_DIM_ZERO
         */
        virtual int checkMaxDims();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Cart_get
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Comm my_comm;
        int my_maxdims;
        int *my_dims;
        int *my_periods;
        int *my_coords;

    };                          // end class MPO_Cart_get



    //*************************************************************************
    //
    //                          MPO_Cart_shift
    //
    /** This class represents MPI_Cart_shift. */
    class MPO_Cart_shift : public MPO_Cartesian
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Cart_shift();

        /** Constructor taking proper arguments. */
        MPO_Cart_shift(MPO_Comm comm, int direction, int disp,
                       int *rank_source,
                       int *rank_dest);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Cart_shift();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Cart_shift
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Comm my_comm;
        int my_direction;
        int my_disp;
        int *my_rank_source;
        int *my_rank_dest;

    };                          // end class MPO_Cart_shift



    //*************************************************************************
    //
    //                          MPO_Cart_sub
    //
    /** This class represents MPI_Cart_sub. */
    class MPO_Cart_sub : public MPO_Cartesian
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Cart_sub();

        /** Constructor taking proper arguments. */
        MPO_Cart_sub(MPO_Comm comm, int *remain_dims, MPI_Comm * new_comm);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Cart_sub();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Cart_sub
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Comm my_comm;
        int *my_remain_dims;
        MPI_Comm *my_new_comm;

    };                          // end class MPO_Cart_sub



    //*************************************************************************
    //
    //                          MPO_Cart_map
    //
    /** This class represents MPI_Cart_map. */
    class MPO_Cart_map : public MPO_Cartesian
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Cart_map();

        /** Constructor taking proper arguments. */
        MPO_Cart_map(MPO_Comm comm, int ndims, int *dims, int *periods,
                     int *newrank);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Cart_map();

        //==========================
        // Methods
        //==========================

        /** Checks additional prerequisites only for this call.
         *  Checks correctness of
         *  - communicator
         *  - topology
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether communicator is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_NULL
         *           - MARMOT_ERR_COMM_NOT_VALID
         */
        virtual int checkComm();

        /** Checks whether new topology is correct.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_TOPOLOGY_CART_NEW_GT_COMM_OLD
         *           - MARMOT_WARN_TOPOLOGY_CART_NEW_LT_COMM_OLD
         */
        virtual int checkTopologyCartNew();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Cart_map
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        MPO_Comm my_comm;
        int my_ndims;
        int *my_dims;
        int *my_periods;
        int *my_newrank;

    };                          // end class MPO_Cart_map



    //*************************************************************************
    //
    //                          MPO_Cart_coords
    //
    /** This class represents MPI_Cart_coords. */
    class MPO_Cart_coords : public MPO_Cartesian
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Cart_coords();

        /** Constructor taking proper arguments. */
        MPO_Cart_coords(MPO_Comm comm, int rank, int maxdims, int *coords);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Cart_coords();

        //==========================
        // Methods
        //==========================

        /** Checks if rank is valid.
         */
        virtual int checkRank();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Cart_coords
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Comm my_comm;
        int my_rank;
        int my_maxdims;
        int *my_coords;

    };                          // end class MPO_Cart_coords



    //*************************************************************************
    //
    //                          MPO_Cart_rank
    //
    /** This class represents MPI_Cart_rank. */
    class MPO_Cart_rank : public MPO_Cartesian
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Cart_rank();

        /** Constructor taking proper arguments. */
        MPO_Cart_rank(MPO_Comm comm, int *coords, int *rank);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Cart_rank();

        //==========================
        // Methods
        //==========================

        /** Checks whether coordinates are valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COORD_NEG
         *           - MARMOT_ERR_COORD_TOO_BIG
         */
        virtual int checkCoordinates();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Cart_rank
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Comm my_comm;
        int *my_coords;
        int *my_rank;

    };                          // end class MPO_Cart_rank



    //=========================================================================
    //                          Graph
    //=========================================================================



    //*************************************************************************
    //
    //                          MPO_Graph_create
    //
    /** This class represents MPI_Graph_create. */
    class MPO_Graph_create : public MPO_Graph
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Graph_create();

        /** Constructor taking proper arguments. */
        MPO_Graph_create(MPO_Comm comm_old, int nnodes, int *index,
                         int *edges, int reorder, MPI_Comm * comm_graph);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Graph_create();

        //==========================
        // Methods
        //==========================

        /** Checks whether input parameters are valid.
         *  We overload MPO_Graph::checkPrerequisites to check
         *  other prerequisites:
         *  Checks if
         *  - communicator
         *  - size
         *  - edges
         *  fit together.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether communicator is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_NULL
         *           - MARMOT_ERR_COMM_NOT_VALID
         */
        virtual int checkComm();

        /** Checks whether new topology is correct.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_TOPOLOGY_NEW_GT_COMM_OLD
         *           - MARMOT_WARN_TOPOLOGY_NEW_LT_COMM_OLD
         */
        virtual int checkTopologyNew();

        /** Checks array of edges.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES:
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_EDGE_INVALID
         */
        virtual int checkEdgeArray();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Graph_create
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        MPO_Comm my_comm_old;
        int my_nnodes;
        int *my_index;
        int *my_edges;
        int my_reorder;
        MPI_Comm *my_comm_graph;

    };                          // end class MPO_Graph_create



    //*************************************************************************
    //
    //                          MPO_Graph_get
    //
    /** This class represents MPI_Graph_get. */
    class MPO_Graph_get : public MPO_Graph
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Graph_get();

        /** Constructor taking proper arguments. */
        MPO_Graph_get(MPO_Comm comm, int maxindex, int maxedges,
                      int *index, int *edges);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Graph_get();

        //==========================
        // Methods
        //==========================

        /** Checks additional prerequisites only for this call.
         *  We overload MPO_Graph::checkPrerequisites to check
         *  other prerequisites:
         *  Checks whether index and edges match with communicator.
         */
        virtual int checkPrerequisites2();

        /** Checks maximal index.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_INDEX_NEG
         *           - MARMOT_WARN_INDEX_LT_SIZE
         */
        virtual int checkMaxIndex();

        /** Checks maximal number of edges.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_EDGE_NEG
         */
        virtual int checkMaxEdges();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Graph_get
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Comm my_comm;
        int my_maxindex;
        int my_maxedges;
        int *my_index;
        int *my_edges;

    };                          // end  class MPO_Graph_get



    //*************************************************************************
    //
    //                          MPO_Graphdims_get
    //
    /** This class represents MPI_Graphdims_get. */
    class MPO_Graphdims_get : public MPO_Graph
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Graphdims_get();

        /** Constructor taking proper arguments. */
        MPO_Graphdims_get(MPO_Comm comm, int *nnodes, int *nedges);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Graphdims_get();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Graphdims_get
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Comm my_comm;
        int *my_nnodes;
        int *my_nedges;

    };                          // end class MPO_Graphdims_get



    //*************************************************************************
    //
    //                          MPO_Graph_neighbors
    //
    /** This class represents MPI_Graph_neighbors. */
    class MPO_Graph_neighbors : public MPO_Graph
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Graph_neighbors();

        /** Constructor taking proper arguments. */
        MPO_Graph_neighbors(MPO_Comm comm, int rank, int maxneighbors,
                            int *neighbors);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Graph_neighbors();

        //==========================
        // Methods
        //==========================

        /** Checks additional prerequisites only for this call.
         *  We overload MPO_Graph::checkPrerequisites to check
         *  other prerequisites:
         *  Checks whether rank and neighbours match with communicator.
         */
        virtual int checkPrerequisites2();

        /** Checks rank. */
        virtual int checkRank();

        /** Checks maximal number of neighbors.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_NEIGHBORS_NEG
         */
        virtual int checkMaxNeighbors();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Graph_neighbors
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Comm my_comm;
        int my_rank;
        int my_maxneighbors;
        int *my_neighbors;

    };                          // end class MPO_Graph_neighbors



    //*************************************************************************
    //
    //                          MPO_Graph_map
    //
    /** This class represents MPI_Graph_map. */
    class MPO_Graph_map : public MPO_Graph
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Graph_map();

        /** Constructor taking proper arguments. */
        MPO_Graph_map(MPO_Comm comm, int nnodes, int *index, int *edges,
                      int *newrank);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Graph_map();

        //==========================
        // Methods
        //==========================

        /** Checks additional prerequisites only for this call.
         *  We overload MPO_Graph::checkPrerequisites to check
         *  other prerequisites:
         *  Checks if communicator, size and edges fit together.
         */
        virtual int checkPrerequisites2();

        /** Checks whether new topology is correct.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_TOPOLOGY_NEW_GT_COMM_OLD
         *           - MARMOT_WARN_TOPOLOGY_NEW_LT_COMM_OLD
         */
        virtual int checkTopologyNew();

        /** Checks array of edges.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES:
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_EDGE_INVALID
         */
        virtual int checkEdgeArray();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Graph_map
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        MPO_Comm my_comm;
        int my_nnodes;
        int *my_index;
        int *my_edges;
        int *my_newrank;

    };                          // end class MPO_Graph_map



    //*************************************************************************
    //
    //                          MPO_Graph_neighbors_count
    //
    /** This class represents MPI_Graph_neighbors_count. */
    class MPO_Graph_neighbors_count : public MPO_Graph
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Graph_neighbors_count();

        /** Constructor taking proper arguments. */
        MPO_Graph_neighbors_count(MPO_Comm comm, int rank, int *nneighbors);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Graph_neighbors_count();

        //==========================
        // Methods
        //==========================

        /** Checks additional prerequisites only for this call.
         *  We overload MPO_Graph::checkPrerequisites to check
         *  other prerequisites:
         *  Checks whether rank is valid.
         */
        virtual int checkRank();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Graph_neighbors_count
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Comm my_comm;
        int my_rank;
        int *my_nneighbors;

    };                          // end class MPO_Graph_neighbors_count



    //*************************************************************************
    //
    //                          MPO_Topo_test
    //
    /** This class represents MPI_Topo_test. */
    class MPO_Topo_test : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Topo_test();

        /** Constructor taking proper arguments. */
        MPO_Topo_test(MPO_Comm comm, int *status);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Topo_test();

        //==========================
        // Methods
        //==========================

// TODO // currently no checks implemented.
        /** Checks correctness of parameters etc.
         *  Currently no checks implemented.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Topo_test
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        MPO_Comm my_comm;
        int *my_status;

    };                          // end class MPO_Topo_test



    //=========================================================================
    //                          Intercomms
    //=========================================================================



    //*************************************************************************
    //
    //                          MPO_Comm_test_inter
    //
    /** This class represents MPI_Comm_test_inter. */
    class MPO_Comm_test_inter : public MPO_Intercommunicators
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Comm_test_inter();

        /** Constructor taking proper arguments. */
        MPO_Comm_test_inter(MPO_Comm comm, int *flag);


        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Comm_test_inter();

        //==========================
        // Methods
        //==========================

        /** Checks correctness of
         *  - comm
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether communicator is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_NULL
         *           - MARMOT_ERR_COMM_NOT_VALID
         */
        virtual int checkComm();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Comm_test_inter
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        MPO_Comm my_comm;
        int *my_flag;

    };                          // end class MPO_Comm_test_inter



    //*************************************************************************
    //
    //                          MPO_Comm_remote_size
    //
    /** This class represents MPI_Comm_remote_size. */
    class MPO_Comm_remote_size : public MPO_Intercommunicators
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Comm_remote_size();

        /** Constructor taking proper arguments. */
        MPO_Comm_remote_size(MPO_Comm comm, int *size);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Comm_remote_size();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Comm_remote_size
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Comm my_comm;
        int *my_size;

    };                          // end class MPO_Comm_remote_size



    //*************************************************************************
    //
    //                          MPO_Comm_remote_group
    //
    /** This class represents MPI_Comm_remote_group. */
    class MPO_Comm_remote_group : public MPO_Intercommunicators
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Comm_remote_group();

        /** Constructor taking proper arguments. */
        MPO_Comm_remote_group(MPO_Comm comm, MPI_Group * group);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Comm_remote_group();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Comm_remote_group
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Comm my_comm;
        MPI_Group * my_group;

    };                          // end class MPO_Comm_remote_group



    //*************************************************************************
    //
    //                          MPO_Intercomm_create
    //
    /** This class represents MPI_Intercomm_create. */
    class MPO_Intercomm_create : public MPO_Intercommunicators
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Intercomm_create();

        /** Constructor taking proper arguments. */
        MPO_Intercomm_create(MPO_Comm local_comm, int local_leader,
                             MPO_Comm peer_comm, int remote_leader, int tag,
                             MPI_Comm * newintercomm);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Intercomm_create();

        //==========================
        // Methods
        //==========================

        /** Checks correctness of parameters.
         *  Checks if
         *  - communicator
         *  - tag
         *  - local leader
         *  - remote leader
         *  are valid.
         *
         *  @return  0
         */
        virtual int checkPrerequisites2();

        /** Checks whether communicator is valid.
         *
         *  @param   my_comm    communicator
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_NULL
         *           - MARMOT_ERR_COMM_NOT_VALID
         */
        virtual int checkComm(MPO_Comm my_comm,string arg_name);

        /** Checks whether communicator is intra-communicator.
         *
         *  @param   my_comm    communicator
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_INTERCOMM
         */
        virtual int checkCommIntracomm(MPO_Comm my_comm,string arg_name);


        /** Checks whether tag is valid.
         *
         *  @param   KIND_OF_TAG  specifies the kind of count for output
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_TAG_NEG
         *           - MARMOT_ERR_TAG_TOO_BIG
         *           - MARMOT_ERR_TAG_ANY_TAG
         *           - MARMOT_WARN_TAG_TOO_BIG
         */
        virtual int checkTag(const string KIND_OF_TAG);

        /** Checks whether local leader is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_LEADER_ANY_SOURCE
         *           - MARMOT_ERR_LEADER_NEG
         *           - MARMOT_ERR_LEADER_OUT_OF_COMM
         */
        virtual int checkLeaderLocal();

        /** Checks whether remote leader is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_LEADER_ANY_SOURCE
         *           - MARMOT_ERR_LEADER_NEG
         *           - MARMOT_ERR_LEADER_OUT_OF_COMM
         */
        virtual int checkLeaderRemote();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Intercomm_create
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        MPO_Comm my_local_comm;
        int my_local_leader;
        MPO_Comm my_peer_comm;
        int my_remote_leader;
        int my_tag;
        MPI_Comm *my_newintercomm;

    };                          // end class MPO_Intercomm_create



    //*************************************************************************
    //
    //                          MPO_Intercomm_merge
    //
    /** This class represents MPI_Intercomm_merge. */
    class MPO_Intercomm_merge : public MPO_Intercommunicators
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Intercomm_merge();

        /** Constructor taking proper arguments. */
        MPO_Intercomm_merge(MPO_Comm intercomm, int high,
                            MPI_Comm * newintracomm);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Intercomm_merge();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Intercomm_merge
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Comm my_comm;
        int my_high;
        MPI_Comm *my_newintracomm;

    };                          // end class MPO_Intercomm_merge



    //=========================================================================
    //                          Probes
    //=========================================================================



    //*************************************************************************
    //
    //                          MPO_Iprobe
    //
    /** This class represents MPI_Iprobe. */
    class MPO_Iprobe : public MPO_Probes
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Iprobe();

        /** Constructor taking proper arguments. */
        MPO_Iprobe(int source, int tag, MPO_Comm comm, int *flag,
                   MPI_Status * status);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Iprobe();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Iprobe
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        // int my_source;
        // int my_tag;
        //MPO_Comm my_comm;
        int *my_flag;
        MPI_Status *my_status;

    };                          // end class MPO_Iprobe



    //*************************************************************************
    //
    //                          MPO_Probe
    //
    /** This class represents MPI_Probe. */
    class MPO_Probe : public MPO_Probes
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Probe();

        /** Constructor taking proper arguments. */
        MPO_Probe(int source, int tag, MPO_Comm comm, MPI_Status * status);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Probe();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Probe
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        // int my_source;
        // int my_tag;
        //MPO_Comm my_comm;
        MPI_Status * my_status;

    };                          // end class MPO_Probe



    //=========================================================================
    //                          Groupconstructors
    //=========================================================================



    //*************************************************************************
    //
    //                          MPO_Comm_group
    //
    /** This class represents MPI_Comm_group. */
    class MPO_Comm_group : public MPO_Groupconstructor
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Comm_group();

        /** Constructor taking proper arguments. */
        MPO_Comm_group(MPO_Comm comm, MPI_Group * newgroup);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Comm_group();

        //==========================
        // Methods
        //==========================

        /** Checks correctness of
         *  - comm
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether communicator is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_NULL
         *           - MARMOT_ERR_COMM_NOT_VALID
         */
        virtual int checkComm();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Comm_group
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        MPO_Comm my_comm;
        //MPI_Group* my_newgroup;

    };                          // end class MPO_Comm_group



    //*************************************************************************
    //
    //                          MPO_Group_union
    //
    /** This class represents MPI_Group_union. */
    class MPO_Group_union : public MPO_Groupconstructor
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Group_union();

        /** Constructor taking proper arguments. */
        MPO_Group_union(MPO_Group group, MPO_Group group2,
                        MPI_Group * newgroup);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Group_union();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Group_union
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Group my_group;
        //MPO_Group my_group2;
        //MPI_Group* my_newgroup;

    };                          // end  class MPO_Group_union



    //*************************************************************************
    //
    //                          MPO_Group_intersection
    //
    /** This class represents MPI_Group_intersection. */
    class MPO_Group_intersection : public MPO_Groupconstructor
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Group_intersection();

        /** Constructor taking proper arguments. */
        MPO_Group_intersection(MPO_Group group, MPO_Group group2,
                               MPI_Group * newgroup);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Group_intersection();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Group_intersection
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Group my_group;
        //MPO_Group my_group2;
        //MPI_Group* my_newgroup;

    };                          // end  class MPO_Group_intersection



    //*************************************************************************
    //
    //                          MPO_Group_difference
    //
    /** This class represents MPI_Group_difference. */
    class MPO_Group_difference : public MPO_Groupconstructor
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Group_difference();

        /** Constructor taking proper arguments. */
        MPO_Group_difference(MPO_Group group, MPO_Group group2,
                             MPI_Group * newgroup);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Group_difference();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Group_difference
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Group my_group;
        //MPO_Group my_group2;
        //MPI_Group* my_newgroup;

    };                          // end class MPO_Group_difference



    //*************************************************************************
    //
    //                          MPO_Group_incl
    //
    /** This class represents MPI_Group_incl. */
    class MPO_Group_incl : public MPO_Groupconstructor
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Group_incl();

        /** Constructor taking proper arguments. */
        MPO_Group_incl(MPO_Group group, int count, int *ranks,
                       MPI_Group * newgroup);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Group_incl();

        //==========================
        // Methods
        //==========================

        /** Checks additional prerequisites only for this call. */
        virtual int checkPrerequisites2();

        /** Checks whether count is valid.
         *
         *  @param   KIND_OF_COUNT  specifies the kind of count for output
         *  @param   KIND_OF_NEW_GROUP  specifies the kind of new group
         *                              for output
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COUNT_NEG
         *           - MARMOT_WARN_COUNT_ZERO
         */
        virtual int checkCountGroup(const string KIND_OF_COUNT,
                const string KIND_OF_NEW_GROUP);

        /** Checks array of ranks. */
        virtual int checkRankArray();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Group_incl
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Group my_group;
        int my_count;
        int *my_ranks;
        //MPI_Group* my_newgroup;

    };                          // end class MPO_Group_incl



    //*************************************************************************
    //
    //                          MPO_Group_excl
    //
    /** This class represents MPI_Group_excl. */
    class MPO_Group_excl : public MPO_Groupconstructor
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Group_excl();

        /** Constructor taking proper arguments. */
        MPO_Group_excl(MPO_Group group, int count, int *ranks,
                       MPI_Group * newgroup);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Group_excl();

        //==========================
        // Methods
        //==========================

        /** Checks additional prerequisites only for this call.
         *  Checks whether
         *  - ranks
         *  - number of ranks
         *  are valid.
         */
        virtual int checkPrerequisites2();

        /** Checks whether count is valid.
         *
         *  @param   KIND_OF_COUNT  specifies the kind of count for output
         *  @param   KIND_OF_NEW_GROUP  specifies the kind of new group
         *                              for output
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COUNT_NEG
         *           - MARMOT_WARN_COUNT_ZERO
         */
        virtual int checkCountGroup(const string KIND_OF_COUNT,
                                    const string KIND_OF_NEW_GROUP);

        /** Checks array of ranks. */
        virtual int checkRankArray();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Group_excl
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Group my_group;
        int my_count;
        int *my_ranks;
        //MPI_Group* my_newgroup;

    };                          // end  class MPO_Group_excl



    //*************************************************************************
    //
    //                          MPO_Group_range_incl
    //
    /** This class represents MPI_Group_range_incl. */
    class MPO_Group_range_incl : public MPO_Groupconstructor
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Group_range_incl();

        /** Constructor taking proper arguments. */
        MPO_Group_range_incl(MPO_Group group, int n, int ranges[][3],
                             MPI_Group * newgroup);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Group_range_incl();

        //==========================
        // Methods
        //==========================

        /** Checks correctness of parameters.
         *  Checks if
         *  - triples of range array
         *  - computed ranks
         *  are valid.
         *
         *  @return  0
         */
        virtual int checkPrerequisites2();

        /** Checks whether array of ranges is valid.
         *  Checks correctness of
         *  - first rank
         *  - last rank
         *  - stride
         *  - computed rank
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_RANK_FIRST_NEG
         *           - MARMOT_ERR_RANK_LAST_NEG
         *           - MARMOT_ERR_RANK_FIRST_TOO_BIG
         *           - MARMOT_ERR_RANK_LAST_TOO_BIG
         *           - MARMOT_ERR_STRIDE_NEG
         *           - MARMOT_ERR_STRIDE_ZERO
         *           - MARMOT_ERR_STRIDE_POS
         *           - MARMOT_ERR_RANK_COMPUTED_NEG
         *           - MARMOT_ERR_RANK_COMPUTED_TOO_BIG
         *           - MARMOT_ERR_RANK_COMPUTED_TWICE
         */
        virtual int checkRangeArray();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Group_range_incl
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Group my_group;
        int my_n;
        int (*my_ranges)[3];
        //MPI_Group* my_newgroup;

    };                          // end class MPO_Group_range_incl



    //*************************************************************************
    //
    //                          MPO_Group_range_excl
    //
    /** This class represents MPI_Group_range_excl. */
    class MPO_Group_range_excl : public MPO_Groupconstructor
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Group_range_excl();

        /** Constructor taking proper arguments. */
        MPO_Group_range_excl(MPO_Group group, int n, int ranges[][3],
                             MPI_Group * newgroup);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Group_range_excl();

        //==========================
        // Methods
        //==========================

        /** Checks correctness of parameters.
         *  Checks if
         *  - triples of range array
         *  - computed ranks
         *  are valid.
         *
         *  @return  0
         */
        virtual int checkPrerequisites2();

        /** Checks whether array of ranges is valid.
         *  Checks correctness of
         *  - first rank
         *  - last rank
         *  - stride
         *  - computed rank
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_RANK_FIRST_NEG
         *           - MARMOT_ERR_RANK_LAST_NEG
         *           - MARMOT_ERR_RANK_FIRST_TOO_BIG
         *           - MARMOT_ERR_RANK_LAST_TOO_BIG
         *           - MARMOT_ERR_STRIDE_NEG
         *           - MARMOT_ERR_STRIDE_ZERO
         *           - MARMOT_ERR_STRIDE_POS
         *           - MARMOT_ERR_RANK_COMPUTED_NEG
         *           - MARMOT_ERR_RANK_COMPUTED_TOO_BIG
         *           - MARMOT_ERR_RANK_COMPUTED_TWICE
         */
        virtual int checkRangeArray();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Group_range_excl
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Group my_group;
        int my_n;
        int (*my_ranges)[3];
        //MPI_Group* my_newgroup;

    };                          // end  class MPO_Group_range_excl



    //=========================================================================
    //                          Grouphandlers
    //=========================================================================



    //*************************************************************************
    //
    //                          MPO_Group_size
    //
    /** This class represents MPI_Group_size. */
    class MPO_Group_size : public MPO_Grouphandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Group_size();

        /** Constructor taking proper arguments. */
        MPO_Group_size(MPO_Group group, int *size);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Group_size();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Group_size
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Group my_group;
        int *my_size;

    };                          // end  class MPO_Group_size



    //*************************************************************************
    //
    //                          MPO_Group_rank
    //
    /** This class represents MPI_Group_rank. */
    class MPO_Group_rank : public MPO_Grouphandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Group_rank();

        /** Constructor taking proper arguments. */
        MPO_Group_rank(MPO_Group group, int *rank);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Group_rank();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Group_rank
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Group my_group;
        int *my_rank;

    };                          // end class MPO_Group_rank



    //*************************************************************************
    //
    //                          MPO_Group_translate_ranks
    //
    /** This class represents MPI_Group_translate_ranks. */
    class MPO_Group_translate_ranks : public MPO_Grouphandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Group_translate_ranks();

        /** Constructor taking proper arguments. */
        MPO_Group_translate_ranks(MPO_Group group, int num, int *ranks,
                                  MPO_Group group2, int *ranks2);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Group_translate_ranks();

        //==========================
        // Methods
        //==========================

        /** Checks array of ranks. */
        virtual int checkRankArray();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Group_translate_ranks
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Group my_group;
        int my_num;
        int *my_ranks;
        //MPO_Group my_group2;
        int *my_ranks2;

    };                          // end class MPO_Group_translate_ranks



    //*************************************************************************
    //
    //                          MPO_Group_compare
    //
    /** This class represents MPI_Group_compare. */
    class MPO_Group_compare : public MPO_Grouphandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Group_compare();

        /** Constructor taking proper arguments. */
        MPO_Group_compare(MPO_Group group, MPO_Group group2, int *result);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Group_compare();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Group_compare
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Group my_group;
        //MPO_Group my_group2;
        int *my_result;

    };                          // end  class MPO_Group_compare



    //*************************************************************************
    //
    //                          MPO_Group_free
    //
    /** This class represents MPI_Group_free. */
    class MPO_Group_free : public MPO_Grouphandler
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Group_free();

        /** Constructor taking proper arguments. */
        MPO_Group_free(MPI_Group * delgroup);


        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Group_free();

        //==========================
        // Methods
        //==========================

        /** Checks additional prerequisites only for this call. */
        virtual int checkPrerequisites();

// TODO
        virtual int checkGroupDel();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Group_free
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        MPI_Group * my_delgroup;

    };                          // end class MPO_Group_free



    //=========================================================================
    //                          Typeconstructors
    //=========================================================================



    //*************************************************************************
    //
    //                          MPO_Type_contiguous
    //
    /** This class represents MPI_Type_contiguous. */
    class MPO_Type_contiguous : public MPO_Typeconstructors
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Type_contiguous();

        /** Constructor taking proper arguments. */
        MPO_Type_contiguous(int count, MPO_Datatype oldtype,
                            MPI_Datatype * newtype);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Type_contiguous();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Type_contiguous
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        // int my_count;
        //MPO_Datatype my_oldtype;
        MPI_Datatype * my_newtype;

    };                          // end class MPO_Type_contiguous



    //*************************************************************************
    //
    //                          MPO_Type_vector
    //
    /** This class represents MPI_Type_vector.
     */
    class MPO_Type_vector : public MPO_Typeconstructors
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Type_vector();

        /** Constructor taking proper arguments. */
        MPO_Type_vector(int count, int blocklength, int stride,
                        MPO_Datatype oldtype, MPI_Datatype * newtype);


        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Type_vector();

        //==========================
        // Methods
        //==========================

        /** Checks whether blocklength is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_BLOCKLENGTH_NEG
         *           - MARMOT_WARN_BLOCKLENGTH_ZERO
         */
        virtual int checkBlocklength();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Type_vector
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        // int my_count;
        int my_blocklength;
        int my_stride;
        //MPO_Datatype my_oldtype;
        MPI_Datatype *my_newtype;

    };                          // end  class MPO_Type_vector



    //*************************************************************************
    //
    //                          MPO_Type_hvector
    //
    /** This class represents MPI_Type_hvector. */
    class MPO_Type_hvector : public MPO_Typeconstructors
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Type_hvector();

        /** Constructor taking proper arguments. */
        MPO_Type_hvector(int count, int blocklength, MPI_Aint stride,
                         MPO_Datatype oldtype, MPI_Datatype * newtype);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Type_hvector();

        //==========================
        // Methods
        //==========================

        /** Checks whether blocklength is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_BLOCKLENGTH_NEG
         *           - MARMOT_WARN_BLOCKLENGTH_ZERO
         */
        virtual int checkBlocklength();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Type_hvector
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        // int my_count;
        int my_blocklength;
        MPI_Aint my_stride;
        //MPO_Datatype my_oldtype;
        MPI_Datatype *my_newtype;

    };                          // end class MPO_Type_hvector



    //*************************************************************************
    //
    //                          MPO_Type_indexed
    //
    /** This class represents MPI_Type_indexed. */
    class MPO_Type_indexed : public MPO_Typeconstructors
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Type_indexed();

        /** Constructor taking proper arguments. */
        MPO_Type_indexed(int count, int *blocklength, int *displ_array,
                         MPO_Datatype oldtype, MPI_Datatype * newtype);


        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Type_indexed();

        //==========================
        // Methods
        //==========================

        /** Checks whether array of blocklengths is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_BLOCKLENGTH_NEG
         *           - MARMOT_WARN_BLOCKLENGTH_ZERO
         */
        virtual int checkBlocklengthArray();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Type_indexed
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        // int my_count;
        int *my_blocklength;
        int *my_displ_array;
        //MPO_Datatype my_oldtype;
        MPI_Datatype *my_newtype;

    };                          // end class MPO_Type_indexed



    //*************************************************************************
    //
    //                          MPO_Type_hindexed
    //
    /** This class represents MPI_Type_hindexed. */
    class MPO_Type_hindexed : public MPO_Typeconstructors
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Type_hindexed();

        /** Constructor taking proper arguments. */
        MPO_Type_hindexed(int count, int *blocklength, MPI_Aint * displ_array,
                          MPO_Datatype oldtype, MPI_Datatype * newtype);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Type_hindexed();

        //==========================
        // Methods
        //==========================


        /** Checks whether array of blocklengths is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_BLOCKLENGTH_NEG
         *           - MARMOT_WARN_BLOCKLENGTH_ZERO
         */
        virtual int checkBlocklengthArray();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Type_hindexed
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        // int my_count;
        int *my_blocklength;
        MPI_Aint *my_displ_array;
        //MPO_Datatype my_oldtype;
        MPI_Datatype *my_newtype;

    };                          // end  class MPO_Type_hindexed



    //*************************************************************************
    //
    //                          MPO_Type_struct
    //
    /** This class represents MPI_Type_struct. */
    class MPO_Type_struct : public MPO_Typeconstructors
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Type_struct();

        /** Constructor taking proper arguments. */
        MPO_Type_struct(int count, int *blocklength, MPI_Aint * displ_array,
                        MPI_Datatype * oldtype, MPI_Datatype * newtype);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Type_struct();

        //==========================
        // Methods
        //==========================

        /** Checks additional prerequisites only for this call.
         *  Checks if
         *  - blocklengths
         *  - datatype
         *  are valid.
         *
         *  @return  0
         */
        virtual int checkPrerequisites2();

        /** Checks whether array of blocklengths is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_BLOCKLENGTH_NEG
         *           - MARMOT_WARN_BLOCKLENGTH_ZERO
         */
        virtual int checkBlocklengthArray();

        /** Checks correctness of array of datatypes.
         *  We do NOT use the checkDatatype from our base class,
         *  but redefine it here.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES:
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_MARMOT
         *           - MARMOT_ERR_TYPE_NOT_VALID
         *           - MARMOT_ERR_TYPE_NULL
         *           - MARMOT_ERR_TYPE_NOT_COMMITTED
         *           - MARMOT_ERR_TYPE_C
         *           - MARMOT_ERR_TYPE_F
         *           - MARMOT_ERR_TYPE_REDUCTION_C
         *           - MARMOT_ERR_TYPE_REDUCTION_F
         *           - MARMOT_ERR_TYPE_OPTIONAL_C
         *           - MARMOT_ERR_TYPE_OPTIONAL_F
         *           - MARMOT_ERR_TYPE_LB
         *           - MARMOT_ERR_TYPE_UB
         *           - MARMOT_WARN_TYPE_REDUCTION_C
         *           - MARMOT_WARN_TYPE_REDUCTION_F
         *           - MARMOT_WARN_TYPE_OPTIONAL_C
         *           - MARMOT_WARN_TYPE_OPTIONAL_F
         */
        virtual int checkDatatypeArrayTypeStruct();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Type_struct
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        // int my_count;
        int *my_blocklength;
        MPI_Aint *my_displ_array;
        MPI_Datatype *my_oldtypes;
        MPI_Datatype *my_newtype;

    };                          // end class MPO_Type_struct



    //=========================================================================
    //                          Typehandlers
    //=========================================================================



    //*************************************************************************
    //
    //                          MPO_Type_commit
    //
    /** This class represents MPI_Type_commit. */
    class MPO_Type_commit : public MPO_Typehandlers
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Type_commit();

        /** Constructor taking proper arguments. */
        MPO_Type_commit(MPI_Datatype * datatype);


        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Type_commit();

        //==========================
        // Methods
        //==========================

        /** Checks additional prerequisites only for this call.
         *  Check correctness of
         *  - datatype
         *
         *  @return  0 */
        virtual int checkPrerequisites();

        /** Checks correctness datatype.
         *  We do NOT use the checkDatatype from our base class,
         *  but redefine it here.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES:
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_MARMOT
         *           - MARMOT_NOTE_TYPE_NULL
         *           - MARMOT_NOTE_TYPE_ALREADY_COMMITTED_C
         *           - MARMOT_NOTE_TYPE_ALREADY_COMMITTED_C_AND_F
         *           - MARMOT_NOTE_TYPE_ALREADY_COMMITTED_F
         */
        virtual int checkDatatypeCommitted();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Type_commit
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        MPI_Datatype * my_datatype;
//        int my_errorcode;

    };                          // end class MPO_Type_commit



    //*************************************************************************
    //
    //                          MPO_Type_free
    //
    /** This class represents MPI_Type_free. */
    class MPO_Type_free : public MPO_Typehandlers
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Type_free();

        /** Constructor taking proper arguments. */
        MPO_Type_free(MPI_Datatype * datatype);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Type_free();

        //==========================
        // Methods
        //==========================

        /** Checks additional prerequisites only for this call. */
        virtual int checkPrerequisites();

// TODO
        virtual int checkDatatype();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Type_free
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        MPI_Datatype * my_datatype;

    };                          // end class MPO_Type_free



    //*************************************************************************
    //
    //                          MPO_Address
    //
    /** This class represents MPI_Address. */
    class MPO_Address : public MPO_Typehandlers
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Address();

        /** Constructor taking proper arguments. */
        MPO_Address(void *location, MPI_Aint * address);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Address();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Address
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_location;
        MPI_Aint *my_address;

    };                          // end class MPO_Address



    //*************************************************************************
    //
    //                          MPO_Type_extent
    //
    /** This class represents MPI_Type_extent. */
    class MPO_Type_extent : public MPO_Typehandlers
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Type_extent();

        /** Constructor taking proper arguments. */
        MPO_Type_extent(MPO_Datatype datatype, MPI_Aint * extent);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Type_extent();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Type_extent
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Datatype my_datatype;
        MPI_Aint * my_extent;

    };                          // end class MPO_Type_extent



    //*************************************************************************
    //
    //                          MPO_Type_size
    //
    /** This class represents MPI_Type_size. */
    class MPO_Type_size : public MPO_Typehandlers
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Type_size();

        /** Constructor taking proper arguments. */
        MPO_Type_size(MPO_Datatype datatype, int *size);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Type_size();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Type_size
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Datatype my_datatype;
        int *my_size;

    };                          // end class MPO_Type_size



    //*************************************************************************
    //
    //                          MPO_Type_lb
    //
    /** This class represents MPI_Type_lb. */
    class MPO_Type_lb : public MPO_Typehandlers
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Type_lb();

        /** Constructor taking proper arguments. */
        MPO_Type_lb(MPO_Datatype datatype, MPI_Aint * displacement);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Type_lb();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Type_lb
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Datatype my_datatype;
        MPI_Aint * my_displacement;

    };                          // end class MPO_Type_lb



    //*************************************************************************
    //
    //                          MPO_Type_ub
    //
    /** This class represents MPI_Type_ub. */
    class MPO_Type_ub : public MPO_Typehandlers
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Type_ub();

        /** Constructor taking proper arguments. */
        MPO_Type_ub(MPO_Datatype datatype, MPI_Aint * displacement);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Type_ub();

        //==========================
        // Methods
        //==========================

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Checks correctness of parameters etc.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Type_ub
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        //MPO_Datatype my_datatype;
        MPI_Aint * my_displacement;

    };                          // end class MPO_Type_ub



    //=========================================================================
    //                          Errhandlers
    //=========================================================================



    //*************************************************************************
    //
    //                          MPO_Errhandler_create
    //
    /** This class represents MPI_Errhandler_create. */
    class MPO_Errhandler_create : public MPO_Errhandlers
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Errhandler_create();

        /** Constructor taking proper arguments. */
        MPO_Errhandler_create(MPI_Handler_function * function,
                              MPI_Errhandler * errhandler);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Errhandler_create();

        //==========================
        // Methods
        //==========================

// TODO // currently no checks implemented.
        /** Checks correctness of parameters etc.
         *  Currently no checks implemented.
         */
        virtual int checkPrerequisites();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Errhandler_create
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        MPI_Handler_function * my_function;
        MPI_Errhandler *my_errhandler;

    };                          // end class MPO_Errhandler_create



    //*************************************************************************
    //
    //                          MPO_Errhandler_set
    //
    /** This class represents MPI_Errhandler_set. */
    class MPO_Errhandler_set : public MPO_Errhandlers
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Errhandler_set();

        /** Constructor taking proper arguments. */
        MPO_Errhandler_set(MPO_Comm comm, MPI_Errhandler errhandler);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Errhandler_set();

        //==========================
        // Methods
        //==========================

        /** Checks correctness of
         *  - communicator
         *  - errhandler
         */
        virtual int checkPrerequisites();

        /** Checks whether communicator is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_NULL
         *           - MARMOT_ERR_COMM_NOT_VALID
         */
        virtual int checkComm();

        /** Checks if errhandler is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES:
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_ERRHANDLER_NOT_VALID
         */
        virtual int checkErrhandler();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Errhandler_set
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        MPO_Comm my_comm;
        MPI_Errhandler my_errhandler;

    };                          // end class MPO_Errhandler_set



    //*************************************************************************
    //
    //                          MPO_Errhandler_get
    //
    /** This class represents MPI_Errhandler_get. */
    class MPO_Errhandler_get : public MPO_Errhandlers
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Errhandler_get();

        /** Constructor taking proper arguments. */
        MPO_Errhandler_get(MPO_Comm comm, MPI_Errhandler * errhandler);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Errhandler_get();

        //==========================
        // Methods
        //==========================

        /** Checks correctness of
         *  - communicator
         *  - errhandler
         */
        virtual int checkPrerequisites();

        /** Checks whether communicator is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_NULL
         *           - MARMOT_ERR_COMM_NOT_VALID
         */
        virtual int checkComm();

        /** Checks if errhandler is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES:
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_ERRHANDLER_NOT_VALID
         */
        virtual int checkErrhandler();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Errhandler_get
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        MPO_Comm my_comm;
        MPI_Errhandler *my_errhandler;

    };                          // end class MPO_Errhandler_get



    //*************************************************************************
    //
    //                          MPO_Errhandler_free
    //
    /** This class represents MPI_Errhandler_free. */
    class MPO_Errhandler_free : public MPO_Errhandlers
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Errhandler_free();

        /** Constructor taking proper arguments. */
        MPO_Errhandler_free(MPI_Errhandler * errhandler);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Errhandler_free();

        //==========================
        // Methods
        //==========================

        /** Checks correctness of
         *  - errhandler
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks if errhandler is valid and erases it from our map.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES:
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_ERRHANDLER_NOT_VALID
         */
        virtual int checkAndEraseErrhandler();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Errhandler_free
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        MPI_Errhandler * my_errhandler;

    };                          // end class MPO_Errhandler_free



    //*************************************************************************
    //
    //                          MPO_Error_string
    //
    /** This class represents MPI_Error_string. */
    class MPO_Error_string : public MPO_Errhandlers
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Error_string();

        /** Constructor taking proper arguments. */
        MPO_Error_string(int errorcode, char *string, int *resultlen);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Error_string();

        //==========================
        // Methods
        //==========================

        /** Checks correctness of
         *  - errorcode
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks if errorcode is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES:
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_ERRORCODE_NOT_VALID
         */
        virtual int checkErrorcode();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Error_string
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        int my_errorcode;
        char *my_string;
        int *my_resultlen;

    };                          // end class MPO_Error_string



    //*************************************************************************
    //
    //                          MPO_Error_class
    //
    /** This class represents MPI_Error_class. */
    class MPO_Error_class : public MPO_Errhandlers
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Error_class();

        /** Constructor taking proper arguments. */
        MPO_Error_class(int errorcode, int *errorclass);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Error_class();

        //==========================
        // Methods
        //==========================

        /** Checks correctness of
         *  - errorcode
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks if errorcode is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES:
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_ERRORCODE_NOT_VALID
         */
        virtual int checkErrorcode();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Error_class
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        int my_errorcode;
        int *my_errorclass;

    };                          // end class MPO_Error_class



    //=========================================================================
    //                          Misc
    //=========================================================================



    //*************************************************************************
    //
    //                          MPO_Op_create
    //
    /** This class represents MPI_Op_create. */
    class MPO_Op_create : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Op_create();

        /** Constructor taking proper arguments. */
        MPO_Op_create(MPI_User_function * function, int commute, MPI_Op * op);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Op_create();

        //==========================
        // Methods
        //==========================

        /** This is to insert the new Operator in our_ops. */
        virtual int insertOp();

// TODO // currently no checks implemented.
        /** Checks correctness of parameters etc.
         *  Currently no checks implemented.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Op_create
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        MPI_User_function * my_function;
        int my_commute;
        MPI_Op *my_op;

    };                          // end  class MPO_Op_create



    //*************************************************************************
    //
    //                          MPO_Op_free
    //
    /** This class represents MPI_Op_free. */
    class MPO_Op_free : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Op_free();

        /** Constructor taking proper arguments. */
        MPO_Op_free(MPI_Op * op);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Op_free();

        //==========================
        // Methods
        //==========================

        /** Checks correctness of
         *  - op
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether op is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_OP_NOT_VALID
         */
        virtual int checkOp();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Op_free
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        MPI_Op * my_op;

    };                          // end class MPO_Op_free



    //*************************************************************************
    //
    //                          MPO_Abort
    //
    /** This class represents MPI_Abort. */
    class MPO_Abort : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Abort();

        /** Constructor taking proper arguments. */
        MPO_Abort(MPO_Comm comm, int errorcode);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Abort();

        //==========================
        // Methods
        //==========================

        /** Checks correctness of
         *  - comm
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether communicator is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_NULL
         *           - MARMOT_ERR_COMM_NOT_VALID
         */
        virtual int checkComm();

        /** Checks whether communicator is MPI_COMM_WORLD.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_WARN_COMM_NOT_WORLD
         */
        virtual int checkCommWorld();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Abort
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        MPO_Comm my_comm;
        int my_errorcode;

    };                          // end class MPO_Abort



    //*************************************************************************
    //
    //                          MPO_Keyval_create
    //
    /** This class represents MPI_Keyval_create. */
    class MPO_Keyval_create : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Keyval_create();

        /** Constructor taking proper arguments. */
        MPO_Keyval_create(MPI_Copy_function * copy_fn, MPI_Delete_function
                          * delete_fn, int *keyval, void *extra_state);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Keyval_create();

        //==========================
        // Methods
        //==========================

// TODO // currently no checks implemented.
        /** Checks correctness of parameters etc.
         *  Currently no checks implemented.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Keyval_create
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        MPI_Copy_function * my_copy_fn;
        MPI_Delete_function *my_delete_fn;
        int *my_keyval;
        void *my_extra_state;

    };                          // end class MPO_Keyval_create



    //*************************************************************************
    //
    //                          MPO_Keyval_free
    //
    /** This class represents MPI_Keyval_free. */
    class MPO_Keyval_free : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Keyval_free();

        /** Constructor taking proper arguments. */
        MPO_Keyval_free(int *keyval);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Keyval_free();

        //==========================
        // Methods
        //==========================

// TODO // currently no checks implemented.
        /** Checks correctness of parameters etc.
         *  Currently no checks implemented.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Keyval_free
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        int *my_keyval;

    };                          // end class MPO_Keyval_free



    //*************************************************************************
    //
    //                          MPO_Attr_put
    //
    /** This class represents MPI_Attr_put. */
    class MPO_Attr_put : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Attr_put();

        /** Constructor taking proper arguments. */
        MPO_Attr_put(MPO_Comm comm, int keyval, void *attribute_val);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Attr_put();

        //==========================
        // Methods
        //==========================

        /** Checks correctness of
         *  - communicator
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether communicator is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_NULL
         *           - MARMOT_ERR_COMM_NOT_VALID
         */
        virtual int checkComm();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Attr_put
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        MPO_Comm my_comm;
        int my_keyval;
        void *my_attribute_val;

    };                          // end class MPO_Attr_put



    //*************************************************************************
    //
    //                          MPO_Attr_get
    //
    /** This class represents MPI_Attr_get. */
    class MPO_Attr_get : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Attr_get();

        /** Constructor taking proper arguments. */
        MPO_Attr_get(MPO_Comm comm, int keyval, void *attribute_val,
                     int *flag);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Attr_get();

        //==========================
        // Methods
        //==========================

        /** Checks correctness of
         *  - communicator
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether communicator is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_NULL
         *           - MARMOT_ERR_COMM_NOT_VALID
         */
        virtual int checkComm();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Attr_get
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        MPO_Comm my_comm;
        int my_keyval;
        void *my_attribute_val;
        int *my_flag;

    };                          // end class MPO_Attr_get



    //*************************************************************************
    //
    //                          MPO_Attr_delete
    //
    /** This class represents MPI_Attr_delete. */
    class MPO_Attr_delete : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Attr_delete();

        /** Constructor taking proper arguments. */
        MPO_Attr_delete(MPO_Comm comm, int keyval);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Attr_delete();

        //==========================
        // Methods
        //==========================

        /** Checks correctness of
         *  - communicator
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether communicator is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_NULL
         *           - MARMOT_ERR_COMM_NOT_VALID
         */
        virtual int checkComm();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Attr_delete
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        MPO_Comm my_comm;
        int my_keyval;

    };                          // end class MPO_Attr_delete



    //*************************************************************************
    //
    //                          MPO_Get_count
    //
    /** This class represents MPI_Get_count. */
    class MPO_Get_count : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Get_count();

        /** Constructor taking proper arguments. */
        MPO_Get_count(MPI_Status * status, MPO_Datatype datatype, int *count);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Get_count();

        //==========================
        // Methods
        //==========================

        /** Checks correctness of
         *  - datatype
         *  - count
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

// TODO
        virtual int checkDatatype();

        /** Checks count returned.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_WARN_COUNT_UNDEFINED
         */
        virtual int checkCountUndefined();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Get_count
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        MPI_Status * my_status;
        MPO_Datatype my_datatype;
        int *my_count;

    };                          // end class MPO_Get_count



    //*************************************************************************
    //
    //                          MPO_Get_elements
    //
    /** This class represents MPI_Get_elements. */
    class MPO_Get_elements : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Get_elements();

        /** Constructor taking proper arguments. */
        MPO_Get_elements(MPI_Status * status, MPO_Datatype datatype,
                         int *count);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Get_elements();

        //==========================
        // Methods
        //==========================

        /** Checks correctness of
         *  - datatype
         *  - count
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

// TODO
        virtual int checkDatatype();

        /** Checks count returned.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_WARN_COUNT_UNDEFINED
         */
        virtual int checkCountUndefined();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Get_elements
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        MPI_Status * my_status;
        MPO_Datatype my_datatype;
        int *my_count;

    };                          // end class MPO_Get_elements



    //*************************************************************************
    //
    //                          MPO_Get_processor_name
    //
    /** This class represents MPI_Get_processor_name. */
    class MPO_Get_processor_name : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Get_processor_name();

        /** Constructor taking proper arguments. */
        MPO_Get_processor_name(char *name, int *resultlen);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Get_processor_name();

        //==========================
        // Methods
        //==========================

// TODO // currently no checks implemented.
        /** Checks correctness of parameters etc.
         *  Currently no checks implemented.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Get_processor_name
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        char *my_name;
        int *my_resultlen;

    };                          // end MPO_Get_processor_name



    //*************************************************************************
    //
    //                          MPO_Get_version
    //
    /** This class represents MPI_Get_version. */
    class MPO_Get_version : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Get_version();

        /** Constructor taking proper arguments. */
        MPO_Get_version(int *version, int *subversion);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Get_version();

        //==========================
        // Methods
        //==========================

// TODO // currently no checks implemented.
        /** Checks correctness of parameters etc.
         *  Currently no checks implemented.
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Get_version
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        int *my_version;
        int *my_subversion;

    };                          // end class MPO_Get_version



    //*************************************************************************
    //
    //                          MPO_Pack
    //
    /** This class represents MPI_Pack. */
    class MPO_Pack : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Pack();

        /** Constructor taking proper arguments. */
        MPO_Pack(void *inbuf, int incount, MPO_Datatype datatye,
                 void *outbuf, int outsize, int *position, MPO_Comm comm);


        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Pack();

        //==========================
        // Methods
        //==========================


        /** Checks correctness of
         *  - size
         *  - count
         *  - communicator
         *  - datatype
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether count is valid.
         *
         *  @param   KIND_OF_COUNT  specifies the kind of count for output
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COUNT_NEG
         *           - MARMOT_WARN_COUNT_ZERO
         */
        virtual int checkCountIn(const string KIND_OF_COUNT);

        /** Checks whether size is valid.
         *
         *  @param   KIND_OF_SIZE  specifies the kind of size for output
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_SIZE_NEG
         *           - MARMOT_WARN_SIZE_ZERO
         */
        virtual int checkSizeOut(const string KIND_OF_SIZE);

        /** Checks whether communicator is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_NULL
         *           - MARMOT_ERR_COMM_NOT_VALID
         */
        virtual int checkComm();

// TODO
        virtual int checkDatatype();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Pack
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_inbuf;
        int my_incount;
        MPO_Datatype my_datatype;
        void *my_outbuf;
        int my_outsize;
        int *my_position;
        MPO_Comm my_comm;

    };                          // end class MPO_Pack



    //*************************************************************************
    //
    //                          MPO_Unpack
    //
    /** This class represents MPI_Unpack. */
    class MPO_Unpack : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Unpack();

        /** Constructor taking proper arguments. */
        MPO_Unpack(void *inbuf, int insize, int *position, void *outbuf,
                   int outcount, MPO_Datatype datatype, MPO_Comm comm);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Unpack();

        //==========================
        // Methods
        //==========================

        /** Checks correctness of
         *  - size
         *  - count
         *  - communicator
         *  - datatype
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether size is valid.
         *
         *  @param   KIND_OF_SIZE  specifies the kind of size for output
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_SIZE_NEG
         *           - MARMOT_WARN_SIZE_ZERO
         */
        virtual int checkSizeIn(const string KIND_OF_SIZE);

        /** Checks whether count is valid.
         *
         *  @param   KIND_OF_COUNT  specifies the kind of count for output
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COUNT_NEG
         *           - MARMOT_WARN_COUNT_ZERO
         */
        virtual int checkCountOut(const string KIND_OF_COUNT);

        /** Checks whether communicator is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_NULL
         *           - MARMOT_ERR_COMM_NOT_VALID
         */
        virtual int checkComm();

// TODO
        virtual int checkDatatype();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
         virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Unpack
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        void *my_inbuf;
        int my_insize;
        int *my_position;
        void *my_outbuf;
        int my_outcount;
        MPO_Datatype my_datatype;
        MPO_Comm my_comm;

    };                          // end class MPO_Unpack



    //*************************************************************************
    //
    //                          MPO_Pack_size
    //
    /** This class represents MPI_Pack_size. */
    class MPO_Pack_size : public MPO_Callobject
    {
    public:
        //==========================
        // Constructors/Destructors
        //==========================

        /** The default constructor initializes an invalid object. */
        MPO_Pack_size();

        /** Constructor taking proper arguments. */
        MPO_Pack_size(int incount, MPO_Datatype datatype, MPO_Comm comm,
                      int *size);

        /** The destructor is virtual because there are virtual members here.*/
        virtual ~MPO_Pack_size();

        //==========================
        // Methods
        //==========================

        /** Checks correctness of
         *  - count
         *  - communicator
         *  - datatype
         *
         *  @return  0
         */
        virtual int checkPrerequisites();

        /** Checks whether count is valid.
         *
         *  @param   KIND_OF_COUNT  specifies the kind of count for output
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COUNT_NEG
         *           - MARMOT_WARN_COUNT_ZERO
         */
        virtual int checkCountIn(const string KIND_OF_COUNT);

        /** Checks whether communicator is valid.
         *
         *  @return  unique error code set by enum MARMOT_ERR_CODES
         *           - MARMOT_SUCCESS
         *           - MARMOT_ERR_COMM_NULL
         *           - MARMOT_ERR_COMM_NOT_VALID
         */
        virtual int checkComm();

// TODO
        virtual int checkDatatype();

        /** @return  the unique call identity set by enum MPO_CALL_ID */
        virtual int getId() const;

        /** Does the MPI call after checking it.
         *
         *  @return  PMPI_Pack_size
         */
        virtual int callNativeMPI();

        /** Checks correctness of parameters etc.
         *  Executed after the actual MPI call.
         *
         *  @return  0
         */
        virtual int checkPost();

        /** @return  the name of the MPI-call (C or Fortran spelling). */
        virtual string getCallname() const;

        /** Overloads << to print call-description. */
        virtual std::ostream & debugout(std::ostream & output) const;

        /** Serializes an object to ostream.*/
        virtual std::ostream & save(std::ostream & out) const;

        /** Deserializes an object from istream.*/
        virtual std::istream & restore(std::istream & in);

    private:
        //==========================
        // Data
        //==========================

        int my_incount;
        MPO_Datatype my_datatype;
        MPO_Comm my_comm;
        int *my_size;

    };                          // end class MPO_Pack_size


}                               // end of namespace MPO

#endif
